/*=========================================================================

  Program: GDCM (Grassroots DICOM). A DICOM library

  Copyright (c) 2006-2011 Mathieu Malaterre
  All rights reserved.
  See Copyright.txt or http://gdcm.sourceforge.net/Copyright.html for details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/
/* Hardcode the path to GDCM_DATA_ROOT */
static const char GDCM_DATA_ROOT[] = "GDCM_DATA_ROOT-NOTFOUND";

static const char GDCM_DATA_EXTRA_ROOT[] = "GDCM_DATA_EXTRA_ROOT-NOTFOUND";

static const char GDCM_PIXEL_SPACING_DATA_ROOT[] = "";

/* Hardcode the path to GDCM_TEMP_DIRECTORY */
static const char GDCM_TEMP_DIRECTORY[] = "/build/gdcm/src/build/Testing/Temporary";


static const char * const gdcmDataFileNames[] = {
  
0 };

static const char * const gdcmBlackListReaderDataFileNames[] = {
  
0 };

//const char * const gdcmBlackListWriterDataImages[] = {
//  
//0 };
