
// GENERATED FILE DO NOT EDIT
// $ xsltproc DefaultPrivateDicts.xsl privatedicts.xml > gdcmPrivateDefaultDicts.cxx

/*=========================================================================

  Program: GDCM (Grassroots DICOM). A DICOM library

  Copyright (c) 2006-2011 Mathieu Malaterre
  All rights reserved.
  See Copyright.txt or http://gdcm.sourceforge.net/Copyright.html for details.

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/

#ifndef GDCMPRIVATEDEFAULTDICTS_CXX
#define GDCMPRIVATEDEFAULTDICTS_CXX

#include "gdcmDicts.h"
#include "gdcmVR.h"
#include "gdcmDict.h"
#include "gdcmDictEntry.h"

namespace {
using namespace gdcm;
using DICT_ENTRY = struct
{
  uint16_t group;
  uint16_t element;
  const char *owner;
  VR::VRType vr;
  VM::VMType vm;
  const char *name;
  bool ret;
};

static const DICT_ENTRY DICOMV3DataDict [] = {
  // Major MRI vendors typically use identifiable creator names for a good reason...
  // BEGIN UIH
  // See https://github.com/rordenlab/dcm2niix/issues/225
  {0x0019,0x0018,"EPI Private Header",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0028,"EPI Private Header",VR::FD,VM::VM1,"?",false },
  {0x0051,0x0011,"EPI Private Header",VR::LO,VM::VM1,"?",false },
  {0x0061,0x0002,"DB Dicom Private Creator",VR::US,VM::VM1,"Generate Private",false },
  {0x0061,0x0004,"EditStatus",VR::US,VM::VM1,"?",false },
  {0x0061,0x0001,"Common Private Header",VR::SH,VM::VM1,"?",false },
  {0x0061,0x0002,"Common Private Header",VR::SH,VM::VM1,"FOV",false },
  {0x0061,0x0003,"Common Private Header",VR::SH,VM::VM1,"?",false },
  {0x0061,0x0006,"PrivateCreator16",VR::ST,VM::VM1,"Radiopharmaceutical Information JSON",false },
  {0x0065,0x0000,"Image Private Header",VR::UL,VM::VM1,"MeasurmentUID",false },
  {0x0065,0x0002,"Image Private Header",VR::LO,VM::VM1,"ImageOrientationDisplayed",false },
  {0x0065,0x0003,"Image Private Header",VR::LO,VM::VM1,"ReceiveCoil",false },
  {0x0065,0x0004,"Image Private Header",VR::SH,VM::VM1,"Interpolation",false },
  {0x0065,0x0005,"Image Private Header",VR::SH,VM::VM1,"PE Direction Displayed",false },
  {0x0065,0x0006,"Image Private Header",VR::IS,VM::VM1,"Slice Group ID",false },
  {0x0065,0x0007,"Image Private Header",VR::OB,VM::VM1,"Uprotocol",false },
  {0x0065,0x0009,"Image Private Header",VR::FD,VM::VM1,"BActualValue",false },
  {0x0065,0x000a,"Image Private Header",VR::FD,VM::VM1,"BUserValue",false },
  {0x0065,0x000d,"Image Private Header",VR::SH,VM::VM1,"Parallel Information",false },
  {0x0065,0x000f,"Image Private Header",VR::SH,VM::VM1,"Slice Position",false },
  {0x0065,0x0011,"Image Private Header",VR::SH,VM::VM1,"Sections",false },
  {0x0065,0x0012,"Image Private Header",VR::LO,VM::VM1,"?",false },
  {0x0065,0x0013,"Image Private Header",VR::FD,VM::VM1,"InPlaneRotAngle",false },
  {0x0065,0x0014,"Image Private Header",VR::DS,VM::VM3,"SliceNormalVector",false },
  {0x0065,0x0015,"Image Private Header",VR::DS,VM::VM3,"SliceCenterPosition",false },
  {0x0065,0x0016,"Image Private Header",VR::UL,VM::VM1,"PixelRotateModel",false },
  {0x0065,0x0017,"Image Private Header",VR::LO,VM::VM1,"SAR Model",false },
  {0x0065,0x0018,"Image Private Header",VR::LO,VM::VM1,"dB/dt Model",false },
  {0x0065,0x0021,"Image Private Header",VR::SH,VM::VM1,"?",false },
  {0x0065,0x0023,"Image Private Header",VR::LO,VM::VM1,"TablePosition",false },
  {0x0065,0x0025,"Image Private Header",VR::DS,VM::VM1,"Slice Gap",false },
  {0x0065,0x0029,"Image Private Header",VR::SH,VM::VM1,"AcquisitionDuration",false },
  {0x0065,0x002b,"Image Private Header",VR::LT,VM::VM1,"ApplicationCategory",false },
  {0x0065,0x002c,"Image Private Header",VR::IS,VM::VM1,"RepeatitionIndex",false },
  {0x0065,0x002d,"Image Private Header",VR::ST,VM::VM1,"SequenceDisplayName",false },
  {0x0065,0x002e,"Image Private Header",VR::LO,VM::VM1,"NoiseDecovarFlag",false },
  {0x0065,0x002f,"Image Private Header",VR::FL,VM::VM1,"ScaleFactor",false },
  {0x0065,0x0030,"Image Private Header",VR::LO,VM::VM1,"?Customer?",false },
  {0x0065,0x0031,"Image Private Header",VR::SH,VM::VM1,"MRSequenceVariant",false },
  {0x0065,0x0032,"Image Private Header",VR::SH,VM::VM1,"MRKSpaceFilter",false },
  {0x0065,0x0033,"Image Private Header",VR::SH,VM::VM1,"MRTableMode",false },
  {0x0065,0x0036,"Image Private Header",VR::OB,VM::VM1,"MRDiscoParameter",false },
  {0x0065,0x0037,"Image Private Header",VR::FD,VM::VM3,"MRDiffusionGradOrientation",false },
  {0x0065,0x0038,"Image Private Header",VR::FD,VM::VM1,"MRPerfusionNoiseLevel",false },
  {0x0065,0x0039,"Image Private Header",VR::SH,VM::VM6,"MRGradRange",false },
  {0x0065,0x0040,"Image Private Header",VR::LO,VM::VM1,"?",false },
  {0x0065,0x0041,"Image Private Header",VR::LT,VM::VM1,"?NIN?",false },
  {0x0065,0x0048,"Image Private Header",VR::OB,VM::VM1,"?",false },
  {0x0065,0x0049,"Image Private Header",VR::DS,VM::VM3,"?",false },
  {0x0065,0x0050,"Image Private Header",VR::DS,VM::VM1,"MR Number Of Slice In Volume",false },
  {0x0065,0x0051,"Image Private Header",VR::SQ,VM::VM1,"MR VFrame Sequence",false },
  {0x0065,0x0052,"Image Private Header",VR::LO,VM::VM1_n,"?",false },
  {0x0065,0x0053,"Image Private Header",VR::DS,VM::VM3,"?",false },
  {0x0065,0x0055,"Image Private Header",VR::DS,VM::VM1,"?",false },
  {0x0065,0x0058,"Image Private Header",VR::IS,VM::VM1,"?",false },
  {0x0065,0x0059,"Image Private Header",VR::LO,VM::VM1,"?",false },
  {0x0065,0x0060,"Image Private Header",VR::LO,VM::VM1,"?",false },
  {0x0065,0x006a,"Image Private Header",VR::LO,VM::VM1,"?",false },
  {0x0065,0x006c,"Image Private Header",VR::OB,VM::VM1,"UIH serialized protocol",false },
  {0x0065,0x0071,"Image Private Header",VR::US,VM::VM1,"?",false },
  {0x0065,0x0073,"Image Private Header",VR::OB,VM::VM1,"?",false },
  {0x0065,0x0076,"Image Private Header",VR::SH,VM::VM1,"?",false },
  {0x0065,0x0077,"Image Private Header",VR::LO,VM::VM1,"?",false },
  {0x0065,0x0078,"Image Private Header",VR::LO,VM::VM1,"?",false },
  {0x0067,0x0004,"PrivateCreator10",VR::FD,VM::VM1,"?",false },
  {0x0067,0x0005,"PrivateCreator10",VR::FD,VM::VM2,"?",false },
  {0x0067,0x0007,"PrivateCreator10",VR::CS,VM::VM3,"?",false },
  {0x0067,0x0008,"PrivateCreator10",VR::LO,VM::VM1,"?",false },
  {0x0067,0x0009,"PrivateCreator10",VR::IS,VM::VM1,"?",false },
  {0x0067,0x000a,"PrivateCreator10",VR::IS,VM::VM1,"?",false },
  {0x0067,0x000b,"PrivateCreator10",VR::IS,VM::VM1,"?",false },
  {0x0067,0x000d,"PrivateCreator10",VR::SH,VM::VM1,"?",false },
  {0x0067,0x0010,"PrivateCreator10",VR::DT,VM::VM1,"?",false },
  {0x0067,0x0012,"PrivateCreator10",VR::DT,VM::VM1,"?",false },
  {0x0067,0x0014,"PrivateCreator10",VR::CS,VM::VM1,"?",false },
  {0x0067,0x0015,"PrivateCreator10",VR::CS,VM::VM1,"?",false },
  {0x0067,0x0017,"PrivateCreator10",VR::CS,VM::VM1,"?",false },
  {0x0067,0x0019,"PrivateCreator10",VR::CS,VM::VM1,"?",false },
  {0x0067,0x001a,"PrivateCreator10",VR::CS,VM::VM1,"?",false },
  {0x0067,0x001b,"PrivateCreator10",VR::CS,VM::VM1,"?",false },
  {0x0067,0x001c,"PrivateCreator10",VR::CS,VM::VM1,"?",false },
  {0x0067,0x001d,"PrivateCreator10",VR::IS,VM::VM1,"?",false },
  {0x0067,0x0020,"PrivateCreator10",VR::UT,VM::VM1,"?",false },
  {0x0067,0x0021,"PrivateCreator10",VR::SQ,VM::VM1,"?",false },
  {0x0067,0x002c,"PrivateCreator10",VR::ST,VM::VM1,"?",false },
  {0x0067,0x002f,"PrivateCreator10",VR::DS,VM::VM1,"?",false },
  {0x0067,0x0031,"PrivateCreator10",VR::UT,VM::VM1,"?",false },
  {0x0067,0x0001,"PrivateCreator11",VR::TM,VM::VM1,"?",false },
  {0x0067,0x0002,"PrivateCreator11",VR::DA,VM::VM1,"?",false },
  {0x0067,0x0003,"PrivateCreator11",VR::DS,VM::VM1,"?",false },
  {0x0067,0x0004,"PrivateCreator11",VR::TM,VM::VM1,"?",false },
  {0x0067,0x0005,"PrivateCreator11",VR::DA,VM::VM1,"?",false },
  {0x0067,0x0006,"PrivateCreator11",VR::DS,VM::VM1,"?",false },
  {0x0067,0x0007,"PrivateCreator11",VR::LO,VM::VM1,"?",false },
  // END UIH
  {0x0033,0x0001,"KONICA MINOLTA QA 1.4",VR::SQ,VM::VM1,"?Possibly PHI?",false },
  {0x0033,0x0004,"KONICA MINOLTA QA 1.4",VR::SS,VM::VM1,"?",false },
  {0x0033,0x0006,"KONICA MINOLTA QA 1.4",VR::SS,VM::VM1,"?",false },
  {0x0033,0x0008,"KONICA MINOLTA QA 1.4",VR::SS,VM::VM1,"?",false },
  {0x0033,0x0010,"KONICA MINOLTA QA 1.4",VR::SS,VM::VM1,"?",false },
  {0x0033,0x0011,"KONICA MINOLTA QA 1.4",VR::SS,VM::VM1,"?",false },
  {0x0033,0x0014,"KONICA MINOLTA QA 1.4",VR::SS,VM::VM1,"?",false },
  {0x0033,0x0016,"KONICA MINOLTA QA 1.4",VR::SS,VM::VM1,"?",false },
  {0x0049,0x0010,"AIR",VR::LO,VM::VM1,"?Possibly PHI?",false },
  {0x0049,0x0011,"AIR",VR::LO,VM::VM1,"?number?",false },
  {0x0049,0x0013,"AIR",VR::LT,VM::VM1,"?Long desc possibly PHI?",false },
  {0x0021,0x0010,"SIEMENS MR FMRI",VR::SQ,VM::VM1,"?",false },
  {0x0021,0x0011,"SIEMENS MR FMRI",VR::LO,VM::VM1,"?",false },
  {0x0021,0x0012,"SIEMENS MR FMRI",VR::SQ,VM::VM1,"?",false },
  {0x0021,0x0013,"SIEMENS MR FMRI",VR::US,VM::VM1,"?",false },
  {0x0021,0x0014,"SIEMENS MR FMRI",VR::SQ,VM::VM1,"?",false },
  {0x0021,0x0015,"SIEMENS MR FMRI",VR::US,VM::VM1,"?",false },
  {0x0021,0x0016,"SIEMENS MR FMRI",VR::LO,VM::VM1,"?",false },
  {0x0021,0x0017,"SIEMENS MR FMRI",VR::US,VM::VM1,"?",false },
  {0x0021,0x0018,"SIEMENS MR FMRI",VR::UL,VM::VM1,"?",false },
  {0x0021,0x0019,"SIEMENS MR FMRI",VR::FL,VM::VM1,"?",false },
  {0x0021,0x0031,"SIEMENS MR FMRI",VR::SQ,VM::VM1,"?",false },
  {0x0021,0x0040,"SIEMENS MR FMRI",VR::SQ,VM::VM1,"?",false },
  {0x0021,0x0041,"SIEMENS MR FMRI",VR::LO,VM::VM1,"?",false },
  {0x0021,0x0042,"SIEMENS MR FMRI",VR::SQ,VM::VM1,"?",false },
  {0x0021,0x0043,"SIEMENS MR FMRI",VR::LO,VM::VM1,"?",false },
  {0x0021,0x0044,"SIEMENS MR FMRI",VR::US,VM::VM1,"?",false },
  {0x0021,0x0045,"SIEMENS MR FMRI",VR::LO,VM::VM1_n,"?",false },
  {0x0021,0x0046,"SIEMENS MR FMRI",VR::SQ,VM::VM1,"?",false },
  {0x0021,0x0047,"SIEMENS MR FMRI",VR::UL,VM::VM1,"?",false },
  {0x0021,0x0048,"SIEMENS MR FMRI",VR::LO,VM::VM1,"?",false },
  {0x0021,0x0049,"SIEMENS MR FMRI",VR::FD,VM::VM1,"?",false },
  {0x0021,0x004a,"SIEMENS MR FMRI",VR::FD,VM::VM1_n,"?",false },
  {0x0021,0x004b,"SIEMENS MR FMRI",VR::FL,VM::VM1,"?",false },
  {0x0009,0x0002,"CANON_MEC_MR3^10",VR::SH,VM::VM1,"?",false },
  {0x0011,0x0003,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0011,0x0004,"CANON_MEC_MR3^10",VR::SQ,VM::VM1,"?",false },
  {0x0011,0x0005,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0011,0x0006,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0011,0x0007,"CANON_MEC_MR3^10",VR::UL,VM::VM1_n,"?",false },
  {0x0011,0x0008,"CANON_MEC_MR3^10",VR::SH,VM::VM1,"?",false },
  {0x0011,0x000c,"CANON_MEC_MR3^10",VR::SS,VM::VM1,"?",false },
  {0x0011,0x000e,"CANON_MEC_MR3^10",VR::US,VM::VM2,"?",false },
  {0x0019,0x0000,"CANON_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0001,"CANON_MEC_MR3^11",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0001,"CANON_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0001,"CANON_MEC_MR3^10",VR::UL,VM::VM1,"?",false },
  {0x0019,0x0002,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0002,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0002,"CANON_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0002,"CANON_MEC_MR3^13",VR::US,VM::VM1,"?",false },
  {0x0019,0x0003,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0003,"CANON_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0003,"CANON_MEC_MR3^13",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0003,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0004,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0004,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0004,"CANON_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0004,"CANON_MEC_MR3^13",VR::US,VM::VM1,"?",false },
  {0x0019,0x0005,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0005,"CANON_MEC_MR3^12",VR::SH,VM::VM1,"?",false },
  {0x0019,0x0005,"CANON_MEC_MR3^13",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0006,"CANON_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0006,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0007,"CANON_MEC_MR3^10",VR::FL,VM::VM3,"?",false },
  {0x0019,0x0007,"CANON_MEC_MR3^12",VR::UL,VM::VM1,"?",false },
  {0x0019,0x0008,"CANON_MEC_MR3^11",VR::SH,VM::VM1,"?",false },
  {0x0019,0x0008,"CANON_MEC_MR3^13",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0008,"CANON_MEC_MR3^12",VR::UL,VM::VM1,"?",false },
  {0x0019,0x0008,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0009,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0009,"CANON_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0009,"CANON_MEC_MR3^11",VR::SL,VM::VM2,"?",false },
  {0x0019,0x0009,"CANON_MEC_MR3^13",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x000a,"CANON_MEC_MR3^13",VR::FD,VM::VM3,"?",false },
  {0x0019,0x000a,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x000a,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x000a,"CANON_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x000b,"CANON_MEC_MR3^13",VR::FD,VM::VM1,"?",false },
  {0x0019,0x000b,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x000b,"CANON_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x000c,"CANON_MEC_MR3^13",VR::FD,VM::VM1,"?",false },
  {0x0019,0x000c,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x000c,"CANON_MEC_MR3^12",VR::UL,VM::VM1,"?",false },
  {0x0019,0x000c,"CANON_MEC_MR3^10",VR::US,VM::VM3,"?",false },
  {0x0019,0x000d,"CANON_MEC_MR3^10",VR::SL,VM::VM6,"?",false },
  {0x0019,0x000d,"CANON_MEC_MR3^12",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x000e,"CANON_MEC_MR3^10",VR::SL,VM::VM6,"?",false },
  {0x0019,0x000e,"CANON_MEC_MR3^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x000f,"CANON_MEC_MR3^13",VR::SH,VM::VM1,"?",false },
  {0x0019,0x000f,"CANON_MEC_MR3^10",VR::SL,VM::VM6,"?",false },
  {0x0019,0x0010,"CANON_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0010,"CANON_MEC_MR3^10",VR::US,VM::VM2,"?",false },
  {0x0019,0x0011,"CANON_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0011,"CANON_MEC_MR3^10",VR::SL,VM::VM2,"?",false },
  {0x0019,0x0012,"CANON_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0012,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0013,"CANON_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0013,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0014,"CANON_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0014,"CANON_MEC_MR3^13",VR::SH,VM::VM1,"?",false },
  {0x0019,0x0014,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0015,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0016,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0017,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0018,"CANON_MEC_MR3^12",VR::UL,VM::VM1,"?",false },
  {0x0019,0x0018,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0019,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0019,"CANON_MEC_MR3^11",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x0019,"CANON_MEC_MR3^13",VR::US,VM::VM1,"?",false },
  {0x0019,0x001a,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x001a,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x001b,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x001b,"CANON_MEC_MR3^13",VR::US,VM::VM1,"?",false },
  {0x0019,0x001c,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x001d,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x001e,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x001f,"CANON_MEC_MR3^13",VR::US,VM::VM1,"?",false },
  {0x0019,0x0022,"CANON_MEC_MR3^13",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0024,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0025,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0026,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0027,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0028,"CANON_MEC_MR3^12",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0028,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0028,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0029,"CANON_MEC_MR3^10",VR::SH,VM::VM1,"?",false },
  {0x0019,0x0029,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x002a,"CANON_MEC_MR3^12",VR::DS,VM::VM3,"?",false },
  {0x0019,0x002a,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x002a,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x002b,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x002b,"CANON_MEC_MR3^10",VR::SH,VM::VM1,"?",false },
  {0x0019,0x002b,"CANON_MEC_MR3^12",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x002c,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x002c,"CANON_MEC_MR3^11",VR::SL,VM::VM3,"?",false },
  {0x0019,0x002d,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x002d,"CANON_MEC_MR3^11",VR::SL,VM::VM3,"?",false },
  {0x0019,0x002e,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x002e,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x002f,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0030,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0030,"CANON_MEC_MR3^10",VR::UL,VM::VM1,"?",false },
  {0x0019,0x0031,"CANON_MEC_MR3^10",VR::OF,VM::VM1_n,"?",false },
  {0x0019,0x0031,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0032,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0032,"CANON_MEC_MR3^12",VR::SH,VM::VM1,"?",false },
  {0x0019,0x0032,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0033,"CANON_MEC_MR3^10",VR::SL,VM::VM3,"?",false },
  {0x0019,0x0034,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0034,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0034,"CANON_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0035,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0035,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0036,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0036,"CANON_MEC_MR3^10",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x0037,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0038,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0038,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0039,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0039,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x003a,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x003a,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x003b,"CANON_MEC_MR3^10",VR::SL,VM::VM2,"?",false },
  {0x0019,0x003c,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x003d,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x003d,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x003e,"CANON_MEC_MR3^11",VR::FL,VM::VM9,"?",false },
  {0x0019,0x003e,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x003f,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x003f,"CANON_MEC_MR3^11",VR::UI,VM::VM1,"?",false },
  {0x0019,0x0040,"CANON_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x0041,"CANON_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x0042,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0042,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0043,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0043,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0044,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0044,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0045,"CANON_MEC_MR3^10",VR::SS,VM::VM1,"?",false },
  {0x0019,0x0046,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0046,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0047,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0048,"CANON_MEC_MR3^11",VR::LT,VM::VM1,"?",false },
  {0x0019,0x0048,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0049,"CANON_MEC_MR3^11",VR::LT,VM::VM1,"?",false },
  {0x0019,0x0049,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x004b,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x004c,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x004c,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x004d,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x004e,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x004e,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x004f,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0050,"CANON_MEC_MR3^10",VR::SL,VM::VM2,"?",false },
  {0x0019,0x0051,"CANON_MEC_MR3^10",VR::SL,VM::VM2,"?",false },
  {0x0019,0x0052,"CANON_MEC_MR3^12",VR::FL,VM::VM2,"?",false },
  {0x0019,0x0052,"CANON_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0054,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0055,"CANON_MEC_MR3^11",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0055,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0056,"CANON_MEC_MR3^10",VR::FL,VM::VM3,"?",false },
  {0x0019,0x0057,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0057,"CANON_MEC_MR3^10",VR::FL,VM::VM3,"?",false },
  {0x0019,0x0058,"CANON_MEC_MR3^10",VR::FL,VM::VM3,"?",false },
  {0x0019,0x0058,"CANON_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x0059,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0059,"CANON_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x005a,"CANON_MEC_MR3^10",VR::FL,VM::VM1_n,"?",false },
  {0x0019,0x005d,"CANON_MEC_MR3^10",VR::FL,VM::VM1_n,"?",false },
  {0x0019,0x005d,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x005e,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x005e,"CANON_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x005f,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x005f,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0060,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0060,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0061,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0061,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0062,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0062,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0063,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0063,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0065,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0065,"CANON_MEC_MR3^10",VR::SL,VM::VM2,"?",false },
  {0x0019,0x0066,"CANON_MEC_MR3^11",VR::SL,VM::VM6,"?",false },
  {0x0019,0x0068,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0069,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x006a,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x006b,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x006b,"CANON_MEC_MR3^11",VR::SL,VM::VM6,"?",false },
  {0x0019,0x006d,"CANON_MEC_MR3^11",VR::SH,VM::VM1,"?",false },
  {0x0019,0x006d,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x006e,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x006e,"CANON_MEC_MR3^11",VR::SH,VM::VM1,"?",false },
  {0x0019,0x006f,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x006f,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0070,"CANON_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0071,"CANON_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0071,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0072,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0073,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0073,"CANON_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0075,"CANON_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0076,"CANON_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0076,"CANON_MEC_MR3^12",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0077,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0078,"CANON_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0078,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0079,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0079,"CANON_MEC_MR3^12",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x007a,"CANON_MEC_MR3^11",VR::SL,VM::VM2,"?",false },
  {0x0019,0x007b,"CANON_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0019,0x007c,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x007c,"CANON_MEC_MR3^11",VR::FL,VM::VM3,"?",false },
  {0x0019,0x007d,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x007d,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x007e,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x007e,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x007f,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x007f,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0080,"CANON_MEC_MR3^12",VR::FD,VM::VM1,"?",false },
  {0x0019,0x0080,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0083,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0085,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0086,"CANON_MEC_MR3^12",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0086,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0087,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0087,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0087,"CANON_MEC_MR3^12",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x0088,"CANON_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0088,"CANON_MEC_MR3^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x0089,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0089,"CANON_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x008a,"CANON_MEC_MR3^10",VR::SL,VM::VM1_n,"?",false },
  {0x0019,0x008a,"CANON_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x008a,"CANON_MEC_MR3^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x008b,"CANON_MEC_MR3^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x008c,"CANON_MEC_MR3^10",VR::UL,VM::VM1,"?",false },
  {0x0019,0x008d,"CANON_MEC_MR3^10",VR::TM,VM::VM1,"?",false },
  {0x0019,0x008e,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x008f,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0092,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0092,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0093,"CANON_MEC_MR3^10",VR::FL,VM::VM3,"?",false },
  {0x0019,0x0093,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0093,"CANON_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0094,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0094,"CANON_MEC_MR3^12",VR::UI,VM::VM1,"?",false },
  {0x0019,0x0095,"CANON_MEC_MR3^12",VR::UI,VM::VM1,"?",false },
  {0x0019,0x0096,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0097,"CANON_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0098,"CANON_MEC_MR3^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x0099,"CANON_MEC_MR3^12",VR::LO,VM::VM1,"?",false },
  {0x0019,0x009a,"CANON_MEC_MR3^12",VR::DS,VM::VM1,"?",false },
  {0x0019,0x009a,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x009b,"CANON_MEC_MR3^10",VR::FL,VM::VM3,"?",false },
  {0x0019,0x009b,"CANON_MEC_MR3^12",VR::IS,VM::VM1,"?",false },
  {0x0019,0x009b,"CANON_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x009c,"CANON_MEC_MR3^12",VR::DS,VM::VM1,"?",false },
  {0x0019,0x009c,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x009c,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x009d,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x009d,"CANON_MEC_MR3^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x009e,"CANON_MEC_MR3^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x009e,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x009f,"CANON_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x009f,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00a0,"CANON_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a0,"CANON_MEC_MR3^10",VR::SL,VM::VM4,"?",false },
  {0x0019,0x00a1,"CANON_MEC_MR3^10",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00a1,"CANON_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a1,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00a2,"CANON_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a2,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00a2,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00a3,"CANON_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a3,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00a3,"CANON_MEC_MR3^10",VR::SL,VM::VM3,"?",false },
  {0x0019,0x00a4,"CANON_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a4,"CANON_MEC_MR3^10",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00a5,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a5,"CANON_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a5,"CANON_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x00a6,"CANON_MEC_MR3^12",VR::FL,VM::VM2_2n,"?",false },
  {0x0019,0x00a6,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00a6,"CANON_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x00a7,"CANON_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a8,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a8,"CANON_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a8,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00a9,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00aa,"CANON_MEC_MR3^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00ab,"CANON_MEC_MR3^10",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00ab,"CANON_MEC_MR3^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00ac,"CANON_MEC_MR3^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00ac,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x00ad,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00ad,"CANON_MEC_MR3^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00ad,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00ae,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00ae,"CANON_MEC_MR3^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00ae,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00af,"CANON_MEC_MR3^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00af,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00b0,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00b0,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00b1,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00b1,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00b2,"CANON_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x00b3,"CANON_MEC_MR3^12",VR::FD,VM::VM1,"?",false },
  {0x0019,0x00b3,"CANON_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x00b4,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00b4,"CANON_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x00b5,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00b5,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00b6,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00b6,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00b7,"CANON_MEC_MR3^10",VR::FL,VM::VM2,"?",false },
  {0x0019,0x00b7,"CANON_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x00b9,"CANON_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x00ba,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00bb,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00bb,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00bc,"CANON_MEC_MR3^10",VR::FL,VM::VM2,"?",false },
  {0x0019,0x00bc,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00bd,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00bd,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00be,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00be,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00bf,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00c0,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00c1,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00c1,"CANON_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x00c2,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00c2,"CANON_MEC_MR3^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x00c3,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00c3,"CANON_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00c3,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00c4,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00c4,"CANON_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00c5,"CANON_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00c6,"CANON_MEC_MR3^10",VR::FL,VM::VM2,"?",false },
  {0x0019,0x00c6,"CANON_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00c7,"CANON_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00c9,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00ca,"CANON_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0019,0x00cb,"CANON_MEC_MR3^10",VR::UL,VM::VM1,"?",false },
  {0x0019,0x00cc,"CANON_MEC_MR3^10",VR::UL,VM::VM1,"?",false },
  {0x0019,0x00cd,"CANON_MEC_MR3^12",VR::LO,VM::VM1,"?",false },
  {0x0019,0x00cd,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00cd,"CANON_MEC_MR3^11",VR::SL,VM::VM3,"?",false },
  {0x0019,0x00ce,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00ce,"CANON_MEC_MR3^12",VR::UL,VM::VM1,"?",false },
  {0x0019,0x00cf,"CANON_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00cf,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00d0,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x00d1,"CANON_MEC_MR3^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00d1,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00d2,"CANON_MEC_MR3^10",VR::SL,VM::VM3,"?",false },
  {0x0019,0x00d2,"CANON_MEC_MR3^12",VR::US,VM::VM3,"?",false },
  {0x0019,0x00d3,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00d3,"CANON_MEC_MR3^12",VR::SS,VM::VM1,"?",false },
  {0x0019,0x00d4,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00d4,"CANON_MEC_MR3^12",VR::US,VM::VM2,"?",false },
  {0x0019,0x00d5,"CANON_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00d6,"CANON_MEC_MR3^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00d7,"CANON_MEC_MR3^12",VR::US,VM::VM2,"?",false },
  {0x0019,0x00d8,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x00d9,"CANON_MEC_MR3^10",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x00da,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x00db,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00dc,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00de,"CANON_MEC_MR3^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x00df,"CANON_MEC_MR3^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x00e0,"CANON_MEC_MR3^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x00e1,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00e1,"CANON_MEC_MR3^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00e1,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00e2,"CANON_MEC_MR3^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00e3,"CANON_MEC_MR3^11",VR::SL,VM::VM3,"?",false },
  {0x0019,0x00e3,"CANON_MEC_MR3^12",VR::TM,VM::VM1,"?",false },
  {0x0019,0x00e4,"CANON_MEC_MR3^11",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00e4,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00e4,"CANON_MEC_MR3^12",VR::US,VM::VM2,"?",false },
  {0x0019,0x00e5,"CANON_MEC_MR3^10",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x00e5,"CANON_MEC_MR3^11",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x00e6,"CANON_MEC_MR3^10",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x00e6,"CANON_MEC_MR3^11",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x00e9,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00ea,"CANON_MEC_MR3^10",VR::UI,VM::VM1,"?",false },
  {0x0019,0x00eb,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00eb,"CANON_MEC_MR3^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x00ec,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x00ec,"CANON_MEC_MR3^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x00ee,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00ef,"CANON_MEC_MR3^12",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00ef,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00f0,"CANON_MEC_MR3^12",VR::LO,VM::VM1,"?",false },
  {0x0019,0x00f0,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f1,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00f1,"CANON_MEC_MR3^12",VR::LO,VM::VM1,"?",false },
  {0x0019,0x00f1,"CANON_MEC_MR3^11",VR::SL,VM::VM3,"Diffusion b-value x Toshiba Orientation",false },
  {0x0019,0x00f2,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00f2,"CANON_MEC_MR3^11",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00f3,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f3,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f4,"CANON_MEC_MR3^10",VR::FD,VM::VM1,"?",false },
  {0x0019,0x00f4,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f5,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f5,"CANON_MEC_MR3^11",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x00f6,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f6,"CANON_MEC_MR3^10",VR::UI,VM::VM1,"?",false },
  {0x0019,0x00f7,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f8,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f8,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x00f9,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f9,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00fa,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00fa,"CANON_MEC_MR3^12",VR::LT,VM::VM1,"?",false },
  {0x0019,0x00fa,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00fb,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00fb,"CANON_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00fb,"CANON_MEC_MR3^10",VR::UL,VM::VM1,"?",false },
  {0x0019,0x00fc,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00fc,"CANON_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00fc,"CANON_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00fd,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00fd,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00fd,"CANON_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00fe,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00fe,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00fe,"CANON_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00ff,"CANON_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00ff,"CANON_MEC_MR3^10",VR::SL,VM::VM3,"?",false },
  {0x0021,0x0002,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0021,0x0003,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0021,0x0004,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0021,0x0008,"CANON_MEC_MR3^10",VR::DS,VM::VM3,"?",false },
  {0x0021,0x0009,"CANON_MEC_MR3^10",VR::DS,VM::VM3,"?",false },
  {0x0021,0x000a,"CANON_MEC_MR3^10",VR::DS,VM::VM3,"?",false },
  {0x0021,0x000b,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0021,0x000c,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0021,0x000d,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0021,0x000f,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0021,0x0012,"CANON_MEC_MR3^10",VR::DS,VM::VM3,"?",false },
  {0x0021,0x0014,"CANON_MEC_MR3^10",VR::SQ,VM::VM1,"?",false },
  {0x0021,0x0015,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0021,0x0016,"CANON_MEC_MR3^10",VR::FL,VM::VM2,"?",false },
  {0x0021,0x0018,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0021,0x001d,"CANON_MEC_MR3^10",VR::IS,VM::VM1,"?",false },
  {0x0021,0x001f,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0021,0x0022,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0021,0x0023,"CANON_MEC_MR3^10",VR::SS,VM::VM1,"?",false },
  {0x0029,0x0001,"CANON_MEC_MR3^10",VR::SL,VM::VM3,"Diffusion Gradient Orientation / Toshiba Orientation",false },
  {0x0029,0x0005,"CANON_MEC_MR3^10",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0006,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0029,0x0007,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0029,0x001c,"CANON_MEC_MR3^10",VR::FL,VM::VM2,"?",false },
  {0x0029,0x0021,"CANON_MEC_MR3^10",VR::SL,VM::VM3,"?",false },
  {0x0029,0x004e,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0029,0x0050,"CANON_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0051,"CANON_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0052,"CANON_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0054,"CANON_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0067,"CANON_MEC_MR3^10",VR::FD,VM::VM4,"Series Diffusion Info (b-value,x,y,z)",false },
  {0x0029,0x0068,"CANON_MEC_MR3^10",VR::SS,VM::VM1,"?",false },
  {0x0029,0x006e,"CANON_MEC_MR3^10",VR::OB,VM::VM1,"?",false },
  {0x0029,0x0001,"CANON_MEC_MR3",VR::SQ,VM::VM1,"Other Private Data",false },
  {0x0029,0x0089,"CANON_MEC_MR3",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0090,"CANON_MEC_MR3",VR::OB,VM::VM1,"Reversed EVRLE DataSet",false },
  {0x700d,0x0000,"CANON_MEC_MR3",VR::DS,VM::VM1,"Scale Factor",false },
  {0x700d,0x0005,"CANON_MEC_MR3",VR::DS,VM::VM4,"FOV",false },
  {0x700d,0x000c,"CANON_MEC_MR3",VR::CS,VM::VM1,"Receiver Gain Correction Check Flag",false },
  {0x700d,0x0020,"CANON_MEC_MR3",VR::SH,VM::VM1,"Identification Flag of 3D GDC",false },
  {0x700d,0x0010,"CANON_MEC_MR3^10",VR::DS,VM::VM1,"2nd Flip Angle [degree]",false },
  {0x700d,0x0011,"CANON_MEC_MR3^10",VR::US,VM::VM2,"Acquisition Inner Matrix",false },
  {0x700d,0x0012,"CANON_MEC_MR3^10",VR::US,VM::VM1,"MP2RAGE Flag",false },
  {0x700d,0x0013,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"Inversion efficiency of inversion recovery pulse",false },
  {0x700d,0x0014,"CANON_MEC_MR3^10",VR::SL,VM::VM1,"Number of dummy shot",false },
  {0x700d,0x0015,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"FFE total repetition time[s]",false },
  {0x700d,0x0016,"CANON_MEC_MR3^10",VR::LO,VM::VM3,"PAS Name",false },
  {0x700d,0x0017,"CANON_MEC_MR3^10",VR::LT,VM::VM1,"Intended Processing",false },
  {0x700d,0x0018,"CANON_MEC_MR3^10",VR::SS,VM::VM1,"Scanned Orientation IDs",false },
  {0x700d,0x0019,"CANON_MEC_MR3^10",VR::OB,VM::VM1,"PAS Reproduct Information",false },
  {0x700d,0x001a,"CANON_MEC_MR3^10",VR::DS,VM::VM1,"ASTAR Inversion Time",false },
  {0x700d,0x001b,"CANON_MEC_MR3^10",VR::DS,VM::VM1,"Saturation Recovery Time",false },
  {0xe201,0x0002,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0xe301,0x0000,"CANON_MEC_MR3^10",VR::SH,VM::VM1,"?",false },
  {0xe401,0x0000,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0xe401,0x0001,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0xe401,0x0002,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0xe401,0x0003,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0xe401,0x0004,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0xe401,0x0005,"CANON_MEC_MR3^10",VR::OB,VM::VM1,"?",false },
  {0xe401,0x0007,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0xe401,0x0008,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0xe401,0x0012,"CANON_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0xe401,0x0013,"CANON_MEC_MR3^10",VR::OB,VM::VM1,"?",false },
  {0xe401,0x0020,"CANON_MEC_MR3^10",VR::SQ,VM::VM1,"?",false },
  {0xe401,0x0021,"CANON_MEC_MR3^10",VR::SQ,VM::VM1,"?",false },
  {0xe401,0x0022,"CANON_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0xe401,0x0023,"CANON_MEC_MR3^10",VR::SH,VM::VM1,"?",false },
  {0xe401,0x0024,"CANON_MEC_MR3^10",VR::OB,VM::VM1,"?",false },
  {0xe403,0x0000,"CANON_MEC_MR3^13",VR::SH,VM::VM1,"?",false },
  {0xe403,0x0000,"CANON_MEC_MR3^10",VR::SQ,VM::VM1,"?",false },
  {0xe403,0x0001,"CANON_MEC_MR3^13",VR::OB,VM::VM1,"?",false },
  {0xe403,0x0002,"CANON_MEC_MR3^13",VR::OB,VM::VM1,"?",false },
  {0xe403,0x0003,"CANON_MEC_MR3^13",VR::OB,VM::VM1,"?",false },
  {0xee01,0x0000,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0xee01,0x0001,"CANON_MEC_MR3^10",VR::OB,VM::VM1,"?",false },
  {0xee01,0x0002,"CANON_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x6001,0x0003,"Merge eFilm Fusion",VR::CS,VM::VM1,"?",false },
  {0x6001,0x0030,"Merge eFilm Fusion",VR::US,VM::VM1,"?",false },
  {0x6001,0x0032,"Merge eFilm Fusion",VR::US,VM::VM1,"?",false },
  {0x0031,0x0000,"GEMS_MRPT_01",VR::SH,VM::VM1,"?",false },
  {0x0027,0x0001,"SIEMENS SYNGO ENHANCED IDATASET API",VR::CS,VM::VM1,"Business Unit Code",false },
  {0x0027,0x0002,"SIEMENS SYNGO ENHANCED IDATASET API",VR::LO,VM::VM1,"Application Type",false },
  {0x0027,0x0003,"SIEMENS SYNGO ENHANCED IDATASET API",VR::SQ,VM::VM1,"Application Attributes Sequence",false },
  {0x0021,0x0001,"SIEMENS MR IMA",VR::SQ,VM::VM1,"?",false },
  {0x0071,0x0002,"SIEMENS MED PT MU MAP",VR::UI,VM::VM1,"?",false },
  {0x0071,0x0021,"Visage",VR::UT,VM::VM1,"?",false },
  {0x0071,0x0022,"Visage",VR::DT,VM::VM1,"?",false },
  {0x0009,0x008c,"SIEMENS SYNGO INDEX SERVICE",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0000,"MORPHING REPLAY INFO",VR::UN,VM::VM1,"?",false }, // wtf: VR is SQ but contains non-SQ nested sub-elements...
  {0x0029,0x0031,"FujiFILM TM",VR::SQ,VM::VM1,"? PHI found ?",false }, //
  {0x7fd9,0x0060,"agfa/studyUpdate",VR::DT,VM::VM1,"?",false },
  {0x7fdd,0x0013,"agfa/originalStudyUID",VR::UI,VM::VM1,"?",false }, // most likely UI ??
  {0x7fdb,0x0012,"agfa/seriesHeader",VR::UN,VM::VM1,"?",false },
  {0x7fdb,0x0053,"agfa/seriesHeader",VR::UN,VM::VM1,"?",false },
  {0x7fdb,0x0054,"agfa/seriesHeader",VR::UN,VM::VM1,"?",false },
  {0x7fdb,0x0055,"agfa/seriesHeader",VR::UN,VM::VM1,"?",false },
  {0x7fdd,0x0042,"Thinking Systems Private Creator",VR::UL,VM::VM1,"?",false },
  {0x7fdd,0x0093,"Thinking Systems Private Creator",VR::LO,VM::VM1,"?",false },
  {0x7fdd,0x0094,"Thinking Systems Private Creator",VR::LO,VM::VM1,"?",false },
  {0x7fdd,0x0097,"Thinking Systems Private Creator",VR::LO,VM::VM1,"?",false },
  {0x0025,0x0001,"SIEMENS MR EXTRACTED CSA HEADER",VR::SQ,VM::VM1,"?",false },
  {0x0025,0x0002,"SIEMENS MR EXTRACTED CSA HEADER",VR::LO,VM::VM1,"?",false },
  {0x0025,0x0003,"SIEMENS MR EXTRACTED CSA HEADER",VR::AT,VM::VM1,"?",false },
  {0x0025,0x0004,"SIEMENS MR EXTRACTED CSA HEADER",VR::SS,VM::VM1,"?",false },
  {0x0009,0x0002,"TOSHIBA_MEC_MR3^10",VR::SH,VM::VM1,"?",false },
  {0x0011,0x0003,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0011,0x0004,"TOSHIBA_MEC_MR3^10",VR::SQ,VM::VM1,"?",false },
  {0x0011,0x0005,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0011,0x0006,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0011,0x0007,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1_n,"?",false },
  {0x0011,0x0008,"TOSHIBA_MEC_MR3^10",VR::SH,VM::VM1,"?",false },
  {0x0011,0x000c,"TOSHIBA_MEC_MR3^10",VR::SS,VM::VM1,"?",false },
  {0x0011,0x000e,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM2,"?",false },
  {0x0019,0x0000,"TOSHIBA_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0001,"TOSHIBA_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0001,"TOSHIBA_MEC_MR3^10",VR::UL,VM::VM1,"?",false },
  {0x0019,0x0002,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0002,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0002,"TOSHIBA_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0002,"TOSHIBA_MEC_MR3^13",VR::US,VM::VM1,"?",false },
  {0x0019,0x0003,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0003,"TOSHIBA_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0003,"TOSHIBA_MEC_MR3^13",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0003,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0004,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0004,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0004,"TOSHIBA_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0004,"TOSHIBA_MEC_MR3^13",VR::US,VM::VM1,"?",false },
  {0x0019,0x0005,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0005,"TOSHIBA_MEC_MR3^12",VR::SH,VM::VM1,"?",false },
  {0x0019,0x0005,"TOSHIBA_MEC_MR3^13",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0006,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0006,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0007,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM3,"?",false },
  {0x0019,0x0007,"TOSHIBA_MEC_MR3^12",VR::UL,VM::VM1_n,"?",false },
  {0x0019,0x0008,"TOSHIBA_MEC_MR3^11",VR::SH,VM::VM1,"?",false },
  {0x0019,0x0008,"TOSHIBA_MEC_MR3^13",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0008,"TOSHIBA_MEC_MR3^12",VR::UL,VM::VM1_n,"?",false },
  {0x0019,0x0008,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0009,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0009,"TOSHIBA_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0009,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM2,"?",false },
  {0x0019,0x0009,"TOSHIBA_MEC_MR3^13",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x000a,"TOSHIBA_MEC_MR3^13",VR::FD,VM::VM3,"?",false },
  {0x0019,0x000a,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x000a,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x000a,"TOSHIBA_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x000b,"TOSHIBA_MEC_MR3^13",VR::FD,VM::VM1,"?",false },
  {0x0019,0x000b,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x000b,"TOSHIBA_MEC_MR3^11",VR::OB,VM::VM1,"?",false },
  {0x0019,0x000b,"TOSHIBA_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x000c,"TOSHIBA_MEC_MR3^13",VR::FD,VM::VM1,"?",false },
  {0x0019,0x000c,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x000c,"TOSHIBA_MEC_MR3^12",VR::UL,VM::VM1,"?",false },
  {0x0019,0x000c,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM3,"?",false },
  {0x0019,0x000d,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM6,"?",false },
  {0x0019,0x000d,"TOSHIBA_MEC_MR3^12",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x000e,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM6,"?",false },
  {0x0019,0x000e,"TOSHIBA_MEC_MR3^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x000f,"TOSHIBA_MEC_MR3^13",VR::SH,VM::VM1,"?",false },
  {0x0019,0x000f,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM6,"?",false },
  {0x0019,0x0010,"TOSHIBA_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0010,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM2,"?",false },
  {0x0019,0x0011,"TOSHIBA_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0011,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM2,"?",false },
  {0x0019,0x0012,"TOSHIBA_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0012,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0013,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0013,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0014,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0014,"TOSHIBA_MEC_MR3^13",VR::SH,VM::VM1,"?",false },
  {0x0019,0x0014,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0015,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0016,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0017,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0018,"TOSHIBA_MEC_MR3^12",VR::UL,VM::VM1,"?",false },
  {0x0019,0x0018,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0019,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0019,"TOSHIBA_MEC_MR3^11",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x0019,"TOSHIBA_MEC_MR3^13",VR::US,VM::VM1,"?",false },
  {0x0019,0x001a,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x001b,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x001b,"TOSHIBA_MEC_MR3^13",VR::US,VM::VM1,"?",false },
  {0x0019,0x001c,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x001d,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x001e,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x001f,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x001f,"TOSHIBA_MEC_MR3^13",VR::US,VM::VM1,"?",false },
  {0x0019,0x0020,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM2,"?",false },
  {0x0019,0x0021,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0022,"TOSHIBA_MEC_MR3^13",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0022,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM2,"?",false },
  {0x0019,0x0023,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0024,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0024,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0025,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0025,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0026,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0026,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0027,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0028,"TOSHIBA_MEC_MR3^12",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0028,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0028,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0029,"TOSHIBA_MEC_MR3^10",VR::SH,VM::VM1,"?",false },
  {0x0019,0x0029,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x002a,"TOSHIBA_MEC_MR3^12",VR::DS,VM::VM3,"?",false },
  {0x0019,0x002a,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x002b,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x002b,"TOSHIBA_MEC_MR3^10",VR::SH,VM::VM1,"?",false },
  {0x0019,0x002b,"TOSHIBA_MEC_MR3^12",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x002c,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x002c,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM3,"?",false },
  {0x0019,0x002d,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x002d,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM3,"?",false },
  {0x0019,0x002e,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x002e,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x002f,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0030,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0030,"TOSHIBA_MEC_MR3^10",VR::UL,VM::VM1,"?",false },
  {0x0019,0x0031,"TOSHIBA_MEC_MR3^10",VR::OF,VM::VM1_n,"?",false },
  {0x0019,0x0031,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0032,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0032,"TOSHIBA_MEC_MR3^12",VR::SH,VM::VM1,"?",false },
  {0x0019,0x0032,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0033,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM3,"?",false },
  {0x0019,0x0034,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0034,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0034,"TOSHIBA_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0035,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0035,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0036,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0036,"TOSHIBA_MEC_MR3^10",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x0037,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0038,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0038,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0039,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x003a,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x003a,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x003b,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM2,"?",false },
  {0x0019,0x003c,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM3,"?",false },
  {0x0019,0x003c,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x003d,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x003e,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM9,"?",false },
  {0x0019,0x003e,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x003f,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x003f,"TOSHIBA_MEC_MR3^11",VR::UI,VM::VM1,"?",false },
  {0x0019,0x0040,"TOSHIBA_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x0041,"TOSHIBA_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x0042,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0042,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0043,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0043,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0044,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0044,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0045,"TOSHIBA_MEC_MR3^10",VR::SS,VM::VM1,"?",false },
  {0x0019,0x0046,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0046,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0047,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0048,"TOSHIBA_MEC_MR3^11",VR::LT,VM::VM1,"?",false },
  {0x0019,0x0048,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0049,"TOSHIBA_MEC_MR3^11",VR::LT,VM::VM1,"?",false },
  {0x0019,0x0049,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x004b,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x004c,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x004c,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x004d,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x004d,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x004e,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x004e,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x004f,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0050,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM2,"?",false },
  {0x0019,0x0051,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM2,"?",false },
  {0x0019,0x0052,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM2,"?",false },
  {0x0019,0x0052,"TOSHIBA_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0054,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0055,"TOSHIBA_MEC_MR3^11",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0055,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0057,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0057,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM3,"?",false },
  {0x0019,0x0058,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM3,"?",false },
  {0x0019,0x0058,"TOSHIBA_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x0059,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0059,"TOSHIBA_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x005a,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1_n,"?",false },
  {0x0019,0x005c,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x005d,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1_n,"?",false },
  {0x0019,0x005d,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x005e,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x005f,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x005f,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0060,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0060,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0061,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0061,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0062,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0062,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0063,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0063,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0064,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0065,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM2,"?",false },
  {0x0019,0x0067,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0068,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0069,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x006a,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x006b,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x006c,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x006d,"TOSHIBA_MEC_MR3^11",VR::SH,VM::VM1,"?",false },
  {0x0019,0x006d,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x006e,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x006e,"TOSHIBA_MEC_MR3^11",VR::SH,VM::VM1,"?",false },
  {0x0019,0x006f,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x006f,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0070,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0070,"TOSHIBA_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0071,"TOSHIBA_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0071,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0072,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0073,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0073,"TOSHIBA_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0075,"TOSHIBA_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0076,"TOSHIBA_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0076,"TOSHIBA_MEC_MR3^12",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0077,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0078,"TOSHIBA_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0078,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0079,"TOSHIBA_MEC_MR3^12",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x007a,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM2,"?",false },
  {0x0019,0x007b,"TOSHIBA_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0019,0x007c,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x007c,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM3,"?",false },
  {0x0019,0x007d,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x007d,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x007e,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x007e,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x007f,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x007f,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0080,"TOSHIBA_MEC_MR3^12",VR::FD,VM::VM1,"?",false },
  {0x0019,0x0080,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0083,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0084,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0085,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0085,"TOSHIBA_MEC_MR3^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x0086,"TOSHIBA_MEC_MR3^12",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0086,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0087,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0087,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0087,"TOSHIBA_MEC_MR3^12",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x0088,"TOSHIBA_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0088,"TOSHIBA_MEC_MR3^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x0089,"TOSHIBA_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x008a,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1_n,"?",false },
  {0x0019,0x008a,"TOSHIBA_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x008a,"TOSHIBA_MEC_MR3^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x008b,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x008b,"TOSHIBA_MEC_MR3^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x008c,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x008c,"TOSHIBA_MEC_MR3^10",VR::UL,VM::VM1,"?",false },
  {0x0019,0x008d,"TOSHIBA_MEC_MR3^10",VR::TM,VM::VM1,"?",false },
  {0x0019,0x008e,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x008f,"TOSHIBA_MEC_MR3^12",VR::OB,VM::VM1,"?",false },
  {0x0019,0x008f,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0091,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0092,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0092,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0093,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM3,"?",false },
  {0x0019,0x0093,"TOSHIBA_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0094,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0094,"TOSHIBA_MEC_MR3^12",VR::UI,VM::VM1,"?",false },
  {0x0019,0x0095,"TOSHIBA_MEC_MR3^12",VR::UI,VM::VM1,"?",false },
  {0x0019,0x0096,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0098,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x0099,"TOSHIBA_MEC_MR3^12",VR::LO,VM::VM1,"?",false },
  {0x0019,0x009a,"TOSHIBA_MEC_MR3^12",VR::DS,VM::VM1,"?",false },
  {0x0019,0x009a,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x009b,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM3,"?",false },
  {0x0019,0x009b,"TOSHIBA_MEC_MR3^12",VR::IS,VM::VM1,"?",false },
  {0x0019,0x009b,"TOSHIBA_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x009c,"TOSHIBA_MEC_MR3^12",VR::DS,VM::VM1,"?",false },
  {0x0019,0x009c,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x009c,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x009d,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x009d,"TOSHIBA_MEC_MR3^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x009e,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x009e,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x009f,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x009f,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00a0,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a0,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00a0,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM4,"?",false },
  {0x0019,0x00a1,"TOSHIBA_MEC_MR3^10",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00a1,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a2,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a2,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00a3,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a3,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM3,"?",false },
  {0x0019,0x00a4,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a4,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00a5,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a5,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a5,"TOSHIBA_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x00a6,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM1_n,"?",false },
  {0x0019,0x00a6,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00a6,"TOSHIBA_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x00a7,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a8,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a8,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a8,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00a9,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00aa,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00ab,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00ab,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00ab,"TOSHIBA_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x00ac,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00ac,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x00ad,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00ad,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00ad,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00ae,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00ae,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00ae,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00af,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00af,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00b0,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00b0,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00b1,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00b1,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00b2,"TOSHIBA_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x00b3,"TOSHIBA_MEC_MR3^12",VR::FD,VM::VM1,"?",false },
  {0x0019,0x00b3,"TOSHIBA_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x00b4,"TOSHIBA_MEC_MR3^12",VR::LO,VM::VM1,"?",false },
  {0x0019,0x00b4,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00b4,"TOSHIBA_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x00b5,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00b5,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00b6,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00b6,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00b7,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM2,"?",false },
  {0x0019,0x00b7,"TOSHIBA_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x00b9,"TOSHIBA_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x00ba,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00bb,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00bb,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00bc,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM2,"?",false },
  {0x0019,0x00bc,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00bd,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00bd,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00be,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00be,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00bf,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00c0,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00c1,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00c1,"TOSHIBA_MEC_MR3^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x00c2,"TOSHIBA_MEC_MR3^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x00c3,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00c3,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00c3,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00c4,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00c4,"TOSHIBA_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00c5,"TOSHIBA_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00c6,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM2,"?",false },
  {0x0019,0x00c6,"TOSHIBA_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00c7,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00c9,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00ca,"TOSHIBA_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0019,0x00cb,"TOSHIBA_MEC_MR3^10",VR::UL,VM::VM1,"?",false },
  {0x0019,0x00cc,"TOSHIBA_MEC_MR3^10",VR::UL,VM::VM1,"?",false },
  {0x0019,0x00cd,"TOSHIBA_MEC_MR3^12",VR::LO,VM::VM1,"?",false },
  {0x0019,0x00cd,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00cd,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM3,"?",false },
  {0x0019,0x00ce,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00ce,"TOSHIBA_MEC_MR3^12",VR::UL,VM::VM1,"?",false },
  {0x0019,0x00cf,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00cf,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00d0,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x00d1,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00d1,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00d2,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM3,"?",false },
  {0x0019,0x00d2,"TOSHIBA_MEC_MR3^12",VR::US,VM::VM3,"?",false },
  {0x0019,0x00d3,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00d3,"TOSHIBA_MEC_MR3^12",VR::SS,VM::VM1,"?",false },
  {0x0019,0x00d4,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00d4,"TOSHIBA_MEC_MR3^12",VR::US,VM::VM2,"?",false },
  {0x0019,0x00d5,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00d6,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00d7,"TOSHIBA_MEC_MR3^12",VR::US,VM::VM2,"?",false },
  {0x0019,0x00d8,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM1_n,"?",false },
  {0x0019,0x00d8,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x00d9,"TOSHIBA_MEC_MR3^10",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x00da,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x00db,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00dc,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00de,"TOSHIBA_MEC_MR3^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x00df,"TOSHIBA_MEC_MR3^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x00e0,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00e0,"TOSHIBA_MEC_MR3^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x00e1,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00e1,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00e1,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00e2,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00e3,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM3,"?",false },
  {0x0019,0x00e3,"TOSHIBA_MEC_MR3^12",VR::TM,VM::VM1,"?",false },
  {0x0019,0x00e4,"TOSHIBA_MEC_MR3^11",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00e4,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00e4,"TOSHIBA_MEC_MR3^12",VR::US,VM::VM2,"?",false },
  {0x0019,0x00e5,"TOSHIBA_MEC_MR3^10",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x00e5,"TOSHIBA_MEC_MR3^11",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x00e6,"TOSHIBA_MEC_MR3^10",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x00e6,"TOSHIBA_MEC_MR3^11",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x00e9,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00ea,"TOSHIBA_MEC_MR3^10",VR::UI,VM::VM1,"?",false },
  {0x0019,0x00eb,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00eb,"TOSHIBA_MEC_MR3^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x00ec,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x00ec,"TOSHIBA_MEC_MR3^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x00ee,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00ef,"TOSHIBA_MEC_MR3^12",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00ef,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00f0,"TOSHIBA_MEC_MR3^12",VR::LO,VM::VM1,"?",false },
  {0x0019,0x00f0,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f0,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM4,"?",false },
  {0x0019,0x00f1,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00f1,"TOSHIBA_MEC_MR3^12",VR::LO,VM::VM1,"?",false },
  {0x0019,0x00f1,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM3,"Diffusion b-value x Toshiba Orientation",false },
  {0x0019,0x00f2,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00f2,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00f3,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f3,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f4,"TOSHIBA_MEC_MR3^10",VR::FD,VM::VM1,"?",false },
  {0x0019,0x00f4,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f5,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f5,"TOSHIBA_MEC_MR3^11",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x00f6,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f6,"TOSHIBA_MEC_MR3^10",VR::UI,VM::VM1,"?",false },
  {0x0019,0x00f7,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f8,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f8,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x00f9,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f9,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00fa,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00fa,"TOSHIBA_MEC_MR3^12",VR::LT,VM::VM1,"?",false },
  {0x0019,0x00fa,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00fb,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00fb,"TOSHIBA_MEC_MR3^10",VR::UL,VM::VM1,"?",false },
  {0x0019,0x00fc,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00fc,"TOSHIBA_MEC_MR3^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00fc,"TOSHIBA_MEC_MR3^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00fd,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00fd,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00fd,"TOSHIBA_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00fe,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00fe,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00fe,"TOSHIBA_MEC_MR3^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00ff,"TOSHIBA_MEC_MR3^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00ff,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM3,"?",false },
  {0x0021,0x0002,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0021,0x0003,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0021,0x0004,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0021,0x0008,"TOSHIBA_MEC_MR3^10",VR::DS,VM::VM3,"?",false },
  {0x0021,0x0009,"TOSHIBA_MEC_MR3^10",VR::DS,VM::VM3,"?",false },
  {0x0021,0x000a,"TOSHIBA_MEC_MR3^10",VR::DS,VM::VM3,"?",false },
  {0x0021,0x000b,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0021,0x000c,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0021,0x000d,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0021,0x000f,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0021,0x0012,"TOSHIBA_MEC_MR3^10",VR::DS,VM::VM3,"?",false },
  {0x0021,0x0014,"TOSHIBA_MEC_MR3^10",VR::SQ,VM::VM1,"?",false },
  {0x0021,0x0015,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0021,0x0016,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM2,"?",false },
  {0x0021,0x0018,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0021,0x001a,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0021,0x001d,"TOSHIBA_MEC_MR3^10",VR::IS,VM::VM1,"?",false },
  {0x0021,0x001f,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0021,0x0022,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0021,0x0023,"TOSHIBA_MEC_MR3^10",VR::SS,VM::VM1,"?",false },
  {0x0029,0x0001,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM3,"Diffusion Gradient Orientation / Toshiba Orientation",false },
  {0x0029,0x0005,"TOSHIBA_MEC_MR3^10",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0006,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0029,0x0007,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0x0029,0x001c,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM2,"?",false },
  {0x0029,0x0021,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM3,"?",false },
  {0x0029,0x004e,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"?",false },
  {0x0029,0x0050,"TOSHIBA_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0051,"TOSHIBA_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0052,"TOSHIBA_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0054,"TOSHIBA_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0067,"TOSHIBA_MEC_MR3^10",VR::FD,VM::VM4,"Series Diffusion Info (b-value,x,y,z)",false },
  {0x0029,0x0068,"TOSHIBA_MEC_MR3^10",VR::SS,VM::VM1,"?",false },
  {0x0029,0x006e,"TOSHIBA_MEC_MR3^10",VR::OB,VM::VM1,"?",false },
  {0x0029,0x0089,"TOSHIBA_MEC_MR3",VR::LO,VM::VM1,"?",false },
  {0x700d,0x0010,"TOSHIBA_MEC_MR3^10",VR::DS,VM::VM1,"2nd Flip Angle [degree]",false },
  {0x700d,0x0011,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM2,"Acquisition Inner Matrix",false },
  {0x700d,0x0012,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"MP2RAGE Flag",false },
  {0x700d,0x0013,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"Inversion efficiency of inversion recovery pulse",false },
  {0x700d,0x0014,"TOSHIBA_MEC_MR3^10",VR::SL,VM::VM1,"Number of dummy shot",false },
  {0x700d,0x0015,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"FFE total repetition time[s]",false },
  {0x700d,0x0016,"TOSHIBA_MEC_MR3^10",VR::LO,VM::VM3,"PAS Name",false },
  {0x700d,0x0017,"TOSHIBA_MEC_MR3^10",VR::LT,VM::VM1,"Intended Processing",false },
  {0x700d,0x0018,"TOSHIBA_MEC_MR3^10",VR::SS,VM::VM1_n,"Scanned Orientation IDs",false },
  {0x700d,0x0019,"TOSHIBA_MEC_MR3^10",VR::OB,VM::VM1,"PAS Reproduct Information",false },
  {0x700d,0x0080,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0xe201,0x0002,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0xe301,0x0000,"TOSHIBA_MEC_MR3^10",VR::SH,VM::VM1,"?",false },
  {0xe401,0x0000,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0xe401,0x0001,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0xe401,0x0002,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0xe401,0x0003,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0xe401,0x0004,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0xe401,0x0005,"TOSHIBA_MEC_MR3^10",VR::OB,VM::VM1,"?",false },
  {0xe401,0x0007,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0xe401,0x0008,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0xe401,0x0012,"TOSHIBA_MEC_MR3^10",VR::FL,VM::VM1,"?",false },
  {0xe401,0x0013,"TOSHIBA_MEC_MR3^10",VR::OB,VM::VM1,"?",false },
  {0xe401,0x0020,"TOSHIBA_MEC_MR3^10",VR::SQ,VM::VM1,"?",false },
  {0xe401,0x0021,"TOSHIBA_MEC_MR3^10",VR::SQ,VM::VM1,"?",false },
  {0xe401,0x0022,"TOSHIBA_MEC_MR3^10",VR::LO,VM::VM1,"?",false },
  {0xe401,0x0023,"TOSHIBA_MEC_MR3^10",VR::SH,VM::VM1,"?",false },
  {0xe401,0x0024,"TOSHIBA_MEC_MR3^10",VR::OB,VM::VM1,"?",false },
  {0xe403,0x0000,"TOSHIBA_MEC_MR3^13",VR::SH,VM::VM1,"?",false },
  {0xe403,0x0000,"TOSHIBA_MEC_MR3^10",VR::SQ,VM::VM1,"?",false },
  {0xe403,0x0001,"TOSHIBA_MEC_MR3^13",VR::OB,VM::VM1,"?",false },
  {0xe403,0x0002,"TOSHIBA_MEC_MR3^13",VR::OB,VM::VM1,"?",false },
  {0xe403,0x0003,"TOSHIBA_MEC_MR3^13",VR::OB,VM::VM1,"?",false },
  {0xee01,0x0000,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0xee01,0x0001,"TOSHIBA_MEC_MR3^10",VR::OB,VM::VM1,"?",false },
  {0xee01,0x0002,"TOSHIBA_MEC_MR3^10",VR::US,VM::VM1,"?",false },
  {0x0029,0x0001,"TOSHIBA_MEC_MR3",VR::SQ,VM::VM1,"Other Private Data",false},
  {0x0029,0x0002,"TOSHIBA_MEC_MR3",VR::SQ,VM::VM1,"Other Private Data 2",false},
  {0x0029,0x0090,"TOSHIBA_MEC_MR3",VR::OB,VM::VM1,"Reversed EVRLE DataSet",false},
  {0x7fdf,0x0050,"TomTec",VR::OB,VM::VM1,"Bookmark Information",false},
  {0x7fdf,0x0051,"TomTec",VR::OB,VM::VM1,"Bookmark Content",false},
  {0x2013,0x0010,"BioDICOMizer",VR::LO,VM::VM1,"Custom Storage Version",false},
  {0x2013,0x0011,"BioDICOMizer",VR::ST,VM::VM1,"Custom Storage Former Filename",false},
  {0x2013,0x0012,"BioDICOMizer",VR::LO,VM::VM1,"Custom Storage Mime-Type",false},
  {0x2013,0x0013,"BioDICOMizer",VR::UL,VM::VM1,"Custom Storage File Length",false},
  {0x2013,0x0016,"BioDICOMizer",VR::ST,VM::VM1,"Custom Storage Information Description",false},
  {0x2013,0x0017,"BioDICOMizer",VR::UL,VM::VM1,"Custom Storage Number Of Segments",false},
  {0x2013,0x0018,"BioDICOMizer",VR::IS,VM::VM1,"Custom Storage Last Byte Padded",false},
  {0x2013,0x0010,"BioClinicaImaging",VR::LO,VM::VM1,"Custom Storage Version",false},
  {0x2013,0x0011,"BioClinicaImaging",VR::ST,VM::VM1,"Custom Storage Former Filename",false},
  {0x2013,0x0012,"BioClinicaImaging",VR::LO,VM::VM1,"Custom Storage Mime-Type",false},
  {0x2013,0x0013,"BioClinicaImaging",VR::UL,VM::VM1,"Custom Storage File Length",false},
  {0x2013,0x0016,"BioClinicaImaging",VR::ST,VM::VM1,"Custom Storage Information Description",false},
  {0x2013,0x0017,"BioClinicaImaging",VR::UL,VM::VM1,"Custom Storage Number Of Segments",false},
  {0x2013,0x0018,"BioClinicaImaging",VR::IS,VM::VM1,"Custom Storage Last Byte Padded",false},
  {0x2013,0x0001,"BioClinicaImagingSegmentGroup",VR::OB,VM::VM1,"Custom Storage Segment Blob #1",false},
  {0x2013,0x0002,"BioClinicaImagingSegmentGroup",VR::OB,VM::VM1,"Custom Storage Segment Blob #2",false},
  {0x2013,0x0003,"BioClinicaImagingSegmentGroup",VR::OB,VM::VM1,"Custom Storage Segment Blob #3",false},
  {0x2013,0x0004,"BioClinicaImagingSegmentGroup",VR::OB,VM::VM1,"Custom Storage Segment Blob #4",false},
  {0x2013,0x0005,"BioClinicaImagingSegmentGroup",VR::OB,VM::VM1,"Custom Storage Segment Blob #5",false},
  {0x2013,0x0006,"BioClinicaImagingSegmentGroup",VR::OB,VM::VM1,"Custom Storage Segment Blob #6",false},
  {0x2013,0x0007,"BioClinicaImagingSegmentGroup",VR::OB,VM::VM1,"Custom Storage Segment Blob #7",false},
  {0x2013,0x0008,"BioClinicaImagingSegmentGroup",VR::OB,VM::VM1,"Custom Storage Segment Blob #8",false},
  {0x2013,0x0009,"BioClinicaImagingSegmentGroup",VR::OB,VM::VM1,"Custom Storage Segment Blob #9",false},
  {0x2213,0x0001,"Clario",VR::UT,VM::VM1,"AES tag",false},
  {0x2213,0x0002,"Clario",VR::LO,VM::VM1,"PX version",false},
  {0x2213,0x0003,"Clario",VR::LO,VM::VM1,"Mask Name",false},
  {0x7fe1,0x0001,"Bioclinica",VR::UT,VM::VM1,"AES tag",false},
  {0x7fe1,0x0002,"Bioclinica",VR::LO,VM::VM1,"PX version",false},
  {0x7fe1,0x0003,"Bioclinica",VR::LO,VM::VM1,"Mask Name",false},
  {0x5533,0x0033,"Visus Change",VR::SQ,VM::VM1,"Visus Data Save Sequence",false},
  {0x5533,0x0035,"Visus Change",VR::DA,VM::VM1,"Visus Data Save Date",false},
  {0x5533,0x0037,"Visus Change",VR::LO,VM::VM1,"Visus Data Save Originator",false},
  {0x5533,0x0039,"Visus Change",VR::FD,VM::VM1,"Visus Data Save ID",false},
  {0x5533,0x003b,"Visus Change",VR::TM,VM::VM1,"?Visus Data Save Time?",false},
  {0x0071,0x0021,"SIEMENS MED PT",VR::UI,VM::VM1,"Registration Matrix UID",false},
  {0x0071,0x0022,"SIEMENS MED PT",VR::DT,VM::VM1,"Decay Correction DateTime",false},
  {0x0071,0x0023,"SIEMENS MED PT",VR::FD,VM::VM16,"Registration Matrix",false},
  {0x0071,0x0024,"SIEMENS MED PT",VR::CS,VM::VM1,"Table Motion",false},
  {0x0021,0x0001,"SIEMENS MR SDR 01",VR::LO,VM::VM1,"??",false},
  {0x0021,0x0001,"SIEMENS MR SDS 01",VR::IS,VM::VM1,"??",false},
  {0x0021,0x0002,"SIEMENS MR SDR 01",VR::LO,VM::VM1_n,"??",false},
  {0x0021,0x0003,"SIEMENS MR SDR 01",VR::LO,VM::VM1,"?",false},
  {0x0021,0x0003,"SIEMENS MR SDI 02",VR::DS,VM::VM1,"FrameAcquisitionDuration",false},
  {0x0021,0x0004,"SIEMENS MR SDI 02",VR::DS,VM::VM1,"??",false},
  {0x0021,0x0004,"SIEMENS MR SDS 01",VR::DS,VM::VM1,"??",false},
  {0x0021,0x0005,"SIEMENS MR SDI 02",VR::IS,VM::VM1,"ICE_Dims",false},
  {0x0021,0x0005,"SIEMENS MR SDS 01",VR::IS,VM::VM3,"??",false},
  {0x0021,0x0006,"SIEMENS MR SDI 02",VR::LO,VM::VM1,"ICE_Dims",false},
  {0x0021,0x0006,"SIEMENS MR SDS 01",VR::LO,VM::VM1,"CoilForGradient",false},
  {0x0021,0x0007,"SIEMENS MR SDS 01",VR::LO,VM::VM1,"LongModelName",false},
  {0x0021,0x0008,"SIEMENS MR SDS 01",VR::SH,VM::VM1,"GradientMode",false},
  {0x0021,0x0009,"SIEMENS MR SDS 01",VR::LO,VM::VM1,"PAT mode",false},
  {0x0021,0x000a,"SIEMENS MR SDS 01",VR::DS,VM::VM1,"??",false},
  {0x0021,0x000c,"SIEMENS MR SDS 01",VR::SH,VM::VM1,"?Orientation?",false},
  {0x0021,0x000d,"SIEMENS MR SDS 01",VR::US,VM::VM1,"??",false},
  {0x0021,0x000f,"SIEMENS MR SDS 01",VR::DS,VM::VM3,"?Stim_lim?",false},
  {0x0021,0x0010,"SIEMENS MR SDS 01",VR::IS,VM::VM1,"PhoenixMetaProtocolVersion",false},
  {0x0021,0x0011,"SIEMENS MR SDS 01",VR::DS,VM::VM1,"??",false},
  {0x0021,0x0012,"SIEMENS MR SDS 01",VR::FD,VM::VM1,"??",false},
  {0x0021,0x0013,"SIEMENS MR SDS 01",VR::DS,VM::VM1,"??",false},
  {0x0021,0x0014,"SIEMENS MR SDS 01",VR::IS,VM::VM1,"??",false},
  {0x0021,0x0016,"SIEMENS MR SDS 01",VR::DS,VM::VM1,"??",false},
  {0x0021,0x0017,"SIEMENS MR SDS 01",VR::DS,VM::VM1,"??",false},
  {0x0021,0x0018,"SIEMENS MR SDS 01",VR::SH,VM::VM1,"?RFSWDMostCriticalAspect?",false},
  {0x0021,0x0019,"SIEMENS MR SDS 01",VR::OB,VM::VM1,"PhoenixMetaProtocol",false},
  {0x0021,0x001a,"SIEMENS MR SDI 02",VR::SH,VM::VM1,"?RFSWDDataType?",false},
  {0x0021,0x001a,"SIEMENS MR SDS 01",VR::LO,VM::VM1,"??",false},
  {0x0021,0x001b,"SIEMENS MR SDS 01",VR::DS,VM::VM1,"??",false},
  {0x0021,0x001c,"SIEMENS MR SDI 02",VR::IS,VM::VM1,"??",false},
  {0x0021,0x001c,"SIEMENS MR SDS 01",VR::DS,VM::VM3,"??",false},
  {0x0021,0x001d,"SIEMENS MR SDS 01",VR::IS,VM::VM1,"??",false},
  {0x0021,0x0021,"SIEMENS MR SDI 02",VR::CS,VM::VM1,"?DiffusionDirectionality?",false},
  {0x0021,0x0022,"SIEMENS MR SDS 01",VR::SH,VM::VM1,"Manufacturer",false},
  {0x0021,0x0023,"SIEMENS MR SDI 02",VR::FD,VM::VM6,"?",false },
  {0x0021,0x0023,"SIEMENS MR SDS 01",VR::IS,VM::VM1,"??",false},
  {0x0021,0x0024,"SIEMENS MR SDI 02",VR::IS,VM::VM1,"??",false},
  {0x0021,0x0025,"SIEMENS MR SDS 01",VR::SL,VM::VM3,"?ImaAbsTablePosition?",false},
  {0x0021,0x0026,"SIEMENS MR SDS 01",VR::IS,VM::VM1_n,"?MiscSequenceParam?",false},
  {0x0021,0x0027,"SIEMENS MR SDS 01",VR::US,VM::VM1,"??",false},
  {0x0021,0x002a,"SIEMENS MR SDS 01",VR::IS,VM::VM1_n,"?CoilId?",false},
  {0x0021,0x002b,"SIEMENS MR SDS 01",VR::ST,VM::VM1,"?ImaAbsTablePosition?",false},
  {0x0021,0x002c,"SIEMENS MR SDS 01",VR::DS,VM::VM3,"??",false},
  {0x0021,0x002d,"SIEMENS MR SDS 01",VR::DS,VM::VM3,"??",false},
  {0x0021,0x002e,"SIEMENS MR SDS 01",VR::IS,VM::VM1,"??",false},
  {0x0021,0x002f,"SIEMENS MR SDS 01",VR::DS,VM::VM3,"??",false},
  {0x0021,0x0030,"SIEMENS MR SDS 01",VR::DS,VM::VM1,"??",false},
  {0x0021,0x0031,"SIEMENS MR SDS 01",VR::IS,VM::VM1,"PtabAbsStartPosZ",false},
  {0x0021,0x0032,"SIEMENS MR SDS 01",VR::SS,VM::VM1,"??",false},
  {0x0021,0x0033,"SIEMENS MR SDI 02",VR::IS,VM::VM1,"?CoilForGradient2?",false},
  {0x0021,0x0033,"SIEMENS MR SDS 01",VR::SH,VM::VM1,"??",false},
  {0x0021,0x0034,"SIEMENS MR SDS 01",VR::DS,VM::VM1,"??",false},
  {0x0021,0x0035,"SIEMENS MR SDI 02",VR::IS,VM::VM1,"??",false},
  {0x0021,0x0035,"SIEMENS MR SDS 01",VR::DS,VM::VM1,"??",false},
  {0x0021,0x0036,"SIEMENS MR SDS 01",VR::DS,VM::VM1,"??",false},
  {0x0021,0x0037,"SIEMENS MR SDS 01",VR::SH,VM::VM1,"??",false},
  {0x0021,0x0038,"SIEMENS MR SDS 01",VR::DS,VM::VM1,"??",false},
  {0x0021,0x003b,"SIEMENS MR SDS 01",VR::DS,VM::VM1,"??",false},
  {0x0021,0x003d,"SIEMENS MR SDS 01",VR::CS,VM::VM1,"?Image Acquisition Param?",false},
  {0x0021,0x003f,"SIEMENS MR SDS 01",VR::UT,VM::VM1,"?",false },
  {0x0021,0x0041,"SIEMENS MR SDI 02",VR::SH,VM::VM1,"?GSWDDataType?",false},
  {0x0021,0x0042,"SIEMENS MR SDI 02",VR::IS,VM::VM1,"?RealDwellTime?",false},
  {0x0021,0x0044,"SIEMENS MR SDS 01",VR::DS,VM::VM2,"??",false},
  {0x0021,0x0045,"SIEMENS MR SDI 02",VR::SL,VM::VM3,"?Table Position?",false},
  {0x0021,0x0045,"SIEMENS MR SDS 01",VR::CS,VM::VM1,"??",false},
  {0x0021,0x0046,"SIEMENS MR SDI 02",VR::FD,VM::VM3,"?",false },
  {0x0021,0x0046,"SIEMENS MR SDS 01",VR::DS,VM::VM1,"??",false},
  {0x0021,0x0047,"SIEMENS MR SDS 01",VR::CS,VM::VM1,"??",false},
  {0x0021,0x0048,"SIEMENS MR SDI 02",VR::IS,VM::VM1,"??",false},
  {0x0021,0x0048,"SIEMENS MR SDS 01",VR::CS,VM::VM1,"??",false},
  {0x0021,0x0049,"SIEMENS MR SDI 02",VR::IS,VM::VM1,"??",false},
  {0x0021,0x0049,"SIEMENS MR SDS 01",VR::CS,VM::VM1,"??",false},
  {0x0021,0x004e,"SIEMENS MR SDI 02",VR::IS,VM::VM1,"Actual 3D Ima Part Number",false},
  {0x0021,0x004f,"SIEMENS MR SDI 02",VR::LO,VM::VM1,"ImaCoilString",false},
  {0x0021,0x0050,"SIEMENS MR SDS 01",VR::US,VM::VM1,"??",false},
  {0x0021,0x0051,"SIEMENS MR SDI 02",VR::UL,VM::VM1,"?SequenceMask?",false},
  {0x0021,0x0051,"SIEMENS MR SDS 01",VR::US,VM::VM1,"??",false},
  {0x0021,0x0053,"SIEMENS MR SDI 02",VR::FD,VM::VM1,"?BandwidthPerPixelPhaseEncode?",false},
  {0x0021,0x0053,"SIEMENS MR SDS 01",VR::CS,VM::VM1,"??",false},
  {0x0021,0x0054,"SIEMENS MR SDI 02",VR::US,VM::VM1,"??",false},
  {0x0021,0x0056,"SIEMENS MR SDI 02",VR::LO,VM::VM1,"?PAT Mode?",false},
  {0x0021,0x0058,"SIEMENS MR SDI 02",VR::SH,VM::VM1,"AcquisitionMatrixText",false},
  {0x0021,0x0059,"SIEMENS MR SDI 02",VR::IS,VM::VM3,"?ImaRelTablePosition?",false},
  {0x0021,0x005a,"SIEMENS MR SDS 01",VR::CS,VM::VM1,"?Sequence Variant 1?",false},
  {0x0021,0x005b,"SIEMENS MR SDI 02",VR::FD,VM::VM3,"Image Position (Patient)",false},
  {0x0021,0x005b,"SIEMENS MR SDS 01",VR::CS,VM::VM1_n,"?Sequence Variant 2?",false},
  {0x0021,0x005c,"SIEMENS MR SDS 01",VR::CS,VM::VM1_n,"MTR flag. MT => ON",false},
  {0x0021,0x005d,"SIEMENS MR SDS 01",VR::SL,VM::VM1,"??",false},
  {0x0021,0x005e,"SIEMENS MR SDI 02",VR::IS,VM::VM1,"?FrameIndex?",false},
  {0x0021,0x005e,"SIEMENS MR SDS 01",VR::LO,VM::VM1,"?",false},
  {0x0021,0x005f,"SIEMENS MR SDS 01",VR::SH,VM::VM1,"?FrameIndex?",false},
  {0x0021,0x0060,"SIEMENS MR SDS 01",VR::DT,VM::VM1,"?",false},
  {0x0021,0x0061,"SIEMENS MR SDS 01",VR::SH,VM::VM1,"?",false},
  {0x0021,0x0062,"SIEMENS MR SDS 01",VR::FL,VM::VM1,"?",false},
  {0x0021,0x0063,"SIEMENS MR SDI 02",VR::IS,VM::VM1,"?Acquisition Number?",false},
  {0x0021,0x0067,"SIEMENS MR SDI 02",VR::ST,VM::VM1,"?DataRole XML?",false },
  {0x0021,0x0068,"SIEMENS MR SDI 02",VR::OB,VM::VM1,"?MRDiffusionSequence as base64 encoded?",false },
  {0x0021,0x0071,"SIEMENS MR SDI 02",VR::UT,VM::VM1,"?may contain PHI?",false},
  {0x0021,0x0075,"SIEMENS MR SDI 02",VR::CS,VM::VM1_n,"Image Type",false},
  {0x0021,0x0076,"SIEMENS MR SDI 02",VR::LO,VM::VM1_n,"Acquisition Info",false},
  {0x0021,0x0077,"SIEMENS MR SDI 02",VR::LO,VM::VM1,"Sequence Name",false},
  {0x0021,0x0078,"SIEMENS MR SDI 02",VR::CS,VM::VM1,"Acquisition Type 1",false},
  {0x0021,0x0079,"SIEMENS MR SDI 02",VR::CS,VM::VM1,"Acquisition Type 2",false},
  {0x0021,0x0081,"SIEMENS MR SDI 02",VR::LO,VM::VM1,"?",false},
  {0x0021,0x0088,"SIEMENS MR SDI 02",VR::DS,VM::VM1,"Slice Location",false},
  {0x0021,0x0089,"SIEMENS MR SDI 02",VR::SH,VM::VM1,"?",false },
  {0x0021,0x008a,"SIEMENS MR SDI 02",VR::IS,VM::VM1,"Instance/Frame Number",false},
  {0x0021,0x008b,"SIEMENS MR SDI 02",VR::SH,VM::VM1,"?",false},
  {0x0021,0x008c,"SIEMENS MR SDI 02",VR::FD,VM::VM1,"?",false},
  {0x0021,0x008d,"SIEMENS MR SDI 02",VR::FD,VM::VM1,"?",false},
  {0x0021,0x008e,"SIEMENS MR SDI 02",VR::ST,VM::VM1,"?",false},
  {0x0021,0x00fe,"SIEMENS MR SDI 02",VR::SQ,VM::VM1,"??",false},
  {0x0021,0x00fe,"SIEMENS MR SDS 01",VR::SQ,VM::VM1,"Series Data Sequence",false},
  {0x0089,0x0054,"SYNGO_IMAGING",VR::OB,VM::VM1,"??",false},
  {0x0095,0x00fa,"SIENET",VR::PN,VM::VM1,"?Some kind of Patient Name?",false},
  {0x8ff1,0x0010,"SSI Image enhancement Group",VR::LO,VM::VM1,"SSI Image enhancement Group Version",false },
  {0x8ff1,0x0020,"SSI Image enhancement Group",VR::IS,VM::VM256,"RGB LUT",false },
  {0x8ff1,0x0024,"SSI Image enhancement Group",VR::LO,VM::VM1,"LUT Description",false },
  {0x0009,0x00ee,"MEDISO-1",VR::DT,VM::VM1,"?",false }, // ftp://dicom.nema.org/MEDICAL/dicom/DataSets/WG30/Mediso/PETCT/Rat%20C-2-3%20PETCT/Compliance/testratc23enhancedpet2.dcm.dcdump.txt
  {0x0009,0x00ef,"MEDISO-1",VR::DT,VM::VM1,"?",false },
  {0x0009,0x00f0,"MEDISO-1",VR::FD,VM::VM1,"?",false },
  {0x0009,0x00f1,"MEDISO-1",VR::FD,VM::VM1,"?",false },
  {0x0009,0x00f2,"MEDISO-1",VR::FD,VM::VM1,"?",false },
  {0x0009,0x00f3,"MEDISO-1",VR::FD,VM::VM1,"?",false },
  {0x0009,0x00fa,"MEDISO-1",VR::ST,VM::VM1,"?",false },
  {0x0009,0x00fb,"MEDISO-1",VR::US,VM::VM1,"?",false },
  {0x0011,0x0006,"MEDISO-1",VR::LO,VM::VM1,"UUID (variant=RFC 4122,version=4)",false },
  {0x0075,0x0000,"LLTech, Light-CT Viewer group",VR::CS,VM::VM1,"Enable MPR",false },
  {0x0075,0x0001,"LLTech, Light-CT Viewer group",VR::LO,VM::VM1,"Imaged surface",false },
  {0x0075,0x0002,"LLTech, Light-CT Viewer group",VR::LO,VM::VM1,"Imaged depth",false },
  {0x0075,0x0003,"LLTech, Light-CT Viewer group",VR::LO,VM::VM1,"Sample description",false },
  {0x0075,0x0004,"LLTech, Light-CT Viewer group",VR::SH,VM::VM1,"Number of accumulations",false }, // UL ?
  {0x0857,0x0000,"Nautilus Medical",VR::LO,VM::VM1,"?Product Name. Eg. DICOM_RAY?",false },
  {0x0857,0x0001,"Nautilus Medical",VR::LO,VM::VM1,"?Source Application Entity Title?",false },
  {0x0857,0x0002,"Nautilus Medical",VR::LO,VM::VM1,"?Burn date ISO 8601 with decimal second fraction?",false },
  {0x0857,0x0003,"Nautilus Medical",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0004,"TELEMIS",VR::US,VM::VM1,"?",false },
  {0x3129,0x0010,"RamSoft Race Identifier",VR::LO,VM::VM1,"?",false },
  {0x7501,0x0001,"CoActiv_PvtTags",VR::LO,VM::VM1,"?",false },
  {0x7501,0x0002,"CoActiv_PvtTags",VR::LO,VM::VM1,"?IP Address?",false },
  {0x7501,0x0003,"CoActiv_PvtTags",VR::UI,VM::VM1,"?Transfer Syntax UID?",false },
  {0x7501,0x0004,"CoActiv_PvtTags",VR::LO,VM::VM1,"?Transfer Syntax Name?",false },
  {0x0009,0x0002,"PMTF INFORMATION DATA^10",VR::SH,VM::VM1,"Device Serial Number",false },
  {0x0011,0x0003,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0011,0x0004,"PMTF INFORMATION DATA^10",VR::SQ,VM::VM1,"?",false },
  {0x0011,0x0005,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0011,0x0006,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0011,0x0007,"PMTF INFORMATION DATA^10",VR::US,VM::VM1_n,"?",false },
  {0x0011,0x0008,"PMTF INFORMATION DATA^10",VR::SH,VM::VM1,"?",false },
  {0x0019,0x0000,"PMTF INFORMATION DATA^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0001,"PMTF INFORMATION DATA^11",VR::OB,VM::VM1,"?character set?",false },
  {0x0019,0x0001,"PMTF INFORMATION DATA^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0001,"PMTF INFORMATION DATA^10",VR::UL,VM::VM1,"?",false },
  {0x0019,0x0002,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0002,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0002,"PMTF INFORMATION DATA^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0003,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0003,"PMTF INFORMATION DATA^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0003,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0004,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0004,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0004,"PMTF INFORMATION DATA^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0005,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0005,"PMTF INFORMATION DATA^12",VR::SH,VM::VM1,"?Sequence Variant?",false },
  {0x0019,0x0006,"PMTF INFORMATION DATA^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0006,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0007,"PMTF INFORMATION DATA^10",VR::FL,VM::VM3,"?",false },
  {0x0019,0x0007,"PMTF INFORMATION DATA^11",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0007,"PMTF INFORMATION DATA^12",VR::UL,VM::VM1_2,"?",false },
  {0x0019,0x0008,"PMTF INFORMATION DATA^11",VR::SH,VM::VM1,"?",false },
  {0x0019,0x0008,"PMTF INFORMATION DATA^12",VR::UL,VM::VM1_2,"?",false },
  {0x0019,0x0008,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0009,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0009,"PMTF INFORMATION DATA^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0009,"PMTF INFORMATION DATA^11",VR::SL,VM::VM2,"?",false },
  {0x0019,0x000a,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x000a,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x000a,"PMTF INFORMATION DATA^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x000b,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x000b,"PMTF INFORMATION DATA^11",VR::OB,VM::VM1,"?",false },
  {0x0019,0x000b,"PMTF INFORMATION DATA^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x000c,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x000c,"PMTF INFORMATION DATA^12",VR::UL,VM::VM1,"?",false },
  {0x0019,0x000c,"PMTF INFORMATION DATA^10",VR::US,VM::VM3,"?",false },
  {0x0019,0x000d,"PMTF INFORMATION DATA^10",VR::SL,VM::VM6,"?",false },
  {0x0019,0x000d,"PMTF INFORMATION DATA^12",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x000e,"PMTF INFORMATION DATA^10",VR::SL,VM::VM6,"?",false },
  {0x0019,0x000e,"PMTF INFORMATION DATA^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x000f,"PMTF INFORMATION DATA^10",VR::SL,VM::VM6,"?",false },
  {0x0019,0x0010,"PMTF INFORMATION DATA^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0010,"PMTF INFORMATION DATA^10",VR::US,VM::VM2,"?",false },
  {0x0019,0x0011,"PMTF INFORMATION DATA^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0011,"PMTF INFORMATION DATA^10",VR::SL,VM::VM2,"?",false },
  {0x0019,0x0012,"PMTF INFORMATION DATA^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0012,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0013,"PMTF INFORMATION DATA^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0013,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0014,"PMTF INFORMATION DATA^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0014,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0015,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0016,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0017,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0018,"PMTF INFORMATION DATA^12",VR::UL,VM::VM1,"?",false },
  {0x0019,0x0018,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0019,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0019,"PMTF INFORMATION DATA^11",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x001a,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x001a,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x001b,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x001c,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x001d,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x001e,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x001f,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0020,"PMTF INFORMATION DATA^10",VR::FL,VM::VM2,"?",false },
  {0x0019,0x0021,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0022,"PMTF INFORMATION DATA^10",VR::SL,VM::VM2,"?",false },
  {0x0019,0x0023,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0024,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0024,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0025,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0025,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0026,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0026,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0027,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0028,"PMTF INFORMATION DATA^12",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0028,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0028,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0029,"PMTF INFORMATION DATA^10",VR::SH,VM::VM1,"?",false },
  {0x0019,0x0029,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x002a,"PMTF INFORMATION DATA^12",VR::DS /* TM*/,VM::VM3 /* 1 */,"?",false }, /* dual TM,1 or DS,3 !!! */
  {0x0019,0x002a,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x002a,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x002b,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x002b,"PMTF INFORMATION DATA^10",VR::SH,VM::VM1,"?",false },
  {0x0019,0x002b,"PMTF INFORMATION DATA^12",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x002c,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x002c,"PMTF INFORMATION DATA^11",VR::SL,VM::VM3,"?",false },
  {0x0019,0x002d,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x002d,"PMTF INFORMATION DATA^11",VR::SL,VM::VM3,"?",false },
  {0x0019,0x002e,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x002e,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x002f,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0030,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0030,"PMTF INFORMATION DATA^10",VR::UL,VM::VM1,"?",false },
  {0x0019,0x0031,"PMTF INFORMATION DATA^10",VR::OF,VM::VM1_n,"?",false },
  {0x0019,0x0031,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0032,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0032,"PMTF INFORMATION DATA^12",VR::SH,VM::VM1,"Sequence Name",false },
  {0x0019,0x0032,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0033,"PMTF INFORMATION DATA^10",VR::SL,VM::VM3,"?",false },
  {0x0019,0x0034,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0034,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0034,"PMTF INFORMATION DATA^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0035,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0035,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0036,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0036,"PMTF INFORMATION DATA^10",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x0037,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0038,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0038,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0039,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0039,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x003a,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"Repetition Time (/100)",false },
  {0x0019,0x003a,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x003b,"PMTF INFORMATION DATA^10",VR::SL,VM::VM2,"?",false },
  {0x0019,0x003c,"PMTF INFORMATION DATA^11",VR::FL,VM::VM3,"?",false },
  {0x0019,0x003c,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x003d,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x003d,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x003e,"PMTF INFORMATION DATA^11",VR::FL,VM::VM9,"?",false },
  {0x0019,0x003e,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x003f,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x003f,"PMTF INFORMATION DATA^11",VR::UI,VM::VM1,"?",false },
  {0x0019,0x0040,"PMTF INFORMATION DATA^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x0041,"PMTF INFORMATION DATA^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x0042,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0042,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0043,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0043,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0044,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0044,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0045,"PMTF INFORMATION DATA^10",VR::SS,VM::VM1,"?",false },
  {0x0019,0x0046,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0046,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0047,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0048,"PMTF INFORMATION DATA^11",VR::LT,VM::VM1,"?QD Whole Body?",false },
  {0x0019,0x0048,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0049,"PMTF INFORMATION DATA^11",VR::LT,VM::VM1,"?Atlas Head?",false },
  {0x0019,0x0049,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x004b,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x004c,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x004c,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x004d,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x004d,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?postproc series?",false },
  {0x0019,0x004e,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x004e,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x004f,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0050,"PMTF INFORMATION DATA^10",VR::SL,VM::VM2,"?",false },
  {0x0019,0x0051,"PMTF INFORMATION DATA^10",VR::SL,VM::VM2,"?",false },
  {0x0019,0x0052,"PMTF INFORMATION DATA^12",VR::FL,VM::VM2,"?",false },
  {0x0019,0x0052,"PMTF INFORMATION DATA^10",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0054,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0055,"PMTF INFORMATION DATA^11",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0055,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0056,"PMTF INFORMATION DATA^10",VR::FL,VM::VM3,"?",false },
  {0x0019,0x0057,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0057,"PMTF INFORMATION DATA^10",VR::FL,VM::VM3,"?",false },
  {0x0019,0x0058,"PMTF INFORMATION DATA^10",VR::FL,VM::VM3,"?",false },
  {0x0019,0x0058,"PMTF INFORMATION DATA^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x0059,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0059,"PMTF INFORMATION DATA^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x005a,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1_n,"?",false },
  {0x0019,0x005c,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x005d,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1_n,"?",false },
  {0x0019,0x005d,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x005e,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x005f,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x005f,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0060,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0060,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0061,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0061,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0062,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0062,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0063,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0063,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0064,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0065,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0065,"PMTF INFORMATION DATA^10",VR::SL,VM::VM2,"?",false },
  {0x0019,0x0067,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0068,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0069,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x006a,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x006b,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x006c,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x006d,"PMTF INFORMATION DATA^11",VR::SH,VM::VM1,"?",false },
  {0x0019,0x006d,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x006e,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x006e,"PMTF INFORMATION DATA^11",VR::SH,VM::VM1,"Sequence Name",false },
  {0x0019,0x006f,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x006f,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0070,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0070,"PMTF INFORMATION DATA^10",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0071,"PMTF INFORMATION DATA^10",VR::LO,VM::VM1,"?Sequence Options?",false },
  {0x0019,0x0071,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0072,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0073,"PMTF INFORMATION DATA^10",VR::LO,VM::VM1,"?Sequence Options?",false },
  {0x0019,0x0075,"PMTF INFORMATION DATA^10",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0076,"PMTF INFORMATION DATA^10",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0076,"PMTF INFORMATION DATA^12",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0078,"PMTF INFORMATION DATA^10",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0078,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0079,"PMTF INFORMATION DATA^12",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x007a,"PMTF INFORMATION DATA^11",VR::SL,VM::VM2,"?",false },
  {0x0019,0x007b,"PMTF INFORMATION DATA^10",VR::LO,VM::VM1,"?",false },
  {0x0019,0x007c,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x007c,"PMTF INFORMATION DATA^11",VR::FL,VM::VM3,"?",false },
  {0x0019,0x007d,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x007d,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x007e,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x007e,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x007f,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x007f,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0080,"PMTF INFORMATION DATA^12",VR::FD,VM::VM1,"Acquisition Duration",false },
  {0x0019,0x0080,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0084,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0085,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0085,"PMTF INFORMATION DATA^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x0086,"PMTF INFORMATION DATA^12",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0086,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0087,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0087,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0087,"PMTF INFORMATION DATA^12",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x0088,"PMTF INFORMATION DATA^10",VR::LO,VM::VM1,"?RxGain?",false },
  {0x0019,0x0088,"PMTF INFORMATION DATA^12",VR::US,VM::VM1,"Series Number Global",false },
  {0x0019,0x0089,"PMTF INFORMATION DATA^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x008a,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1_n,"?",false },
  {0x0019,0x008a,"PMTF INFORMATION DATA^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x008a,"PMTF INFORMATION DATA^12",VR::US,VM::VM1,"Series Number Global",false },
  {0x0019,0x008b,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x008b,"PMTF INFORMATION DATA^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x008c,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x008c,"PMTF INFORMATION DATA^10",VR::UL,VM::VM1,"?",false },
  {0x0019,0x008d,"PMTF INFORMATION DATA^10",VR::TM,VM::VM1,"?",false },
  {0x0019,0x008e,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x008f,"PMTF INFORMATION DATA^12",VR::OB,VM::VM1,"?",false },
  {0x0019,0x008f,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0090,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0091,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0092,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0092,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0093,"PMTF INFORMATION DATA^10",VR::FL,VM::VM3,"?",false },
  {0x0019,0x0093,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0093,"PMTF INFORMATION DATA^12",VR::SL,VM::VM1,"?num?",false },
  {0x0019,0x0094,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x0094,"PMTF INFORMATION DATA^12",VR::UI,VM::VM1,"Referenced SOP Instance UID",false },
  {0x0019,0x0095,"PMTF INFORMATION DATA^12",VR::UI,VM::VM1,"?",false },
  {0x0019,0x0096,"PMTF INFORMATION DATA^12",VR::LO,VM::VM1,"?PASTA?",false },
  {0x0019,0x0096,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x0098,"PMTF INFORMATION DATA^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x0099,"PMTF INFORMATION DATA^12",VR::LO,VM::VM1,"Sequence Name",false },
  {0x0019,0x009a,"PMTF INFORMATION DATA^12",VR::DS,VM::VM1,"Series Number Local",false },
  {0x0019,0x009a,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x009b,"PMTF INFORMATION DATA^12",VR::DS,VM::VM1,"?",false }, // Sometimes IS
  {0x0019,0x009b,"PMTF INFORMATION DATA^10",VR::FL,VM::VM3,"?",false },
  {0x0019,0x009b,"PMTF INFORMATION DATA^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x009c,"PMTF INFORMATION DATA^12",VR::DS,VM::VM1,"?num?",false },
  {0x0019,0x009c,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x009c,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x009d,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x009d,"PMTF INFORMATION DATA^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x009e,"PMTF INFORMATION DATA^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x009e,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x009f,"PMTF INFORMATION DATA^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x009f,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00a0,"PMTF INFORMATION DATA^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a0,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00a0,"PMTF INFORMATION DATA^10",VR::SL,VM::VM4,"?",false },
  {0x0019,0x00a1,"PMTF INFORMATION DATA^10",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00a1,"PMTF INFORMATION DATA^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a1,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00a2,"PMTF INFORMATION DATA^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a2,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00a3,"PMTF INFORMATION DATA^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a3,"PMTF INFORMATION DATA^10",VR::SL,VM::VM3,"?",false },
  {0x0019,0x00a4,"PMTF INFORMATION DATA^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a4,"PMTF INFORMATION DATA^10",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00a5,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a5,"PMTF INFORMATION DATA^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a5,"PMTF INFORMATION DATA^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x00a6,"PMTF INFORMATION DATA^12",VR::FL,VM::VM2_2n,"?",false },
  {0x0019,0x00a6,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00a6,"PMTF INFORMATION DATA^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x00a7,"PMTF INFORMATION DATA^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a8,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?Repetition Time (/100)",false },
  {0x0019,0x00a8,"PMTF INFORMATION DATA^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00a8,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00a9,"PMTF INFORMATION DATA^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00a9,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00aa,"PMTF INFORMATION DATA^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00ab,"PMTF INFORMATION DATA^10",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00ab,"PMTF INFORMATION DATA^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00ab,"PMTF INFORMATION DATA^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x00ac,"PMTF INFORMATION DATA^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00ac,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x00ad,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00ad,"PMTF INFORMATION DATA^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00ad,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00ae,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00ae,"PMTF INFORMATION DATA^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00ae,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00af,"PMTF INFORMATION DATA^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00af,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00b0,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00b0,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00b1,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00b1,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00b2,"PMTF INFORMATION DATA^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x00b3,"PMTF INFORMATION DATA^12",VR::FD,VM::VM1,"Acquisition Duration",false },
  {0x0019,0x00b3,"PMTF INFORMATION DATA^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x00b4,"PMTF INFORMATION DATA^12",VR::LO,VM::VM1,"?",false },
  {0x0019,0x00b4,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00b4,"PMTF INFORMATION DATA^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x00b5,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00b5,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00b6,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00b6,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00b7,"PMTF INFORMATION DATA^10",VR::FL,VM::VM2,"?",false },
  {0x0019,0x00b7,"PMTF INFORMATION DATA^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x00b9,"PMTF INFORMATION DATA^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x00ba,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00bb,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00bb,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00bc,"PMTF INFORMATION DATA^10",VR::FL,VM::VM2,"?",false },
  {0x0019,0x00bc,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00bd,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00bd,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00be,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1_n,"?",false },
  {0x0019,0x00be,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"Acquisition Duration",false },
  {0x0019,0x00bf,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00c0,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00c1,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00c1,"PMTF INFORMATION DATA^11",VR::US,VM::VM1,"?",false },
  {0x0019,0x00c2,"PMTF INFORMATION DATA^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x00c3,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00c3,"PMTF INFORMATION DATA^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00c3,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00c4,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00c4,"PMTF INFORMATION DATA^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00c5,"PMTF INFORMATION DATA^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00c6,"PMTF INFORMATION DATA^10",VR::FL,VM::VM2,"?",false },
  {0x0019,0x00c6,"PMTF INFORMATION DATA^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00c7,"PMTF INFORMATION DATA^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00c9,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00ca,"PMTF INFORMATION DATA^10",VR::LO,VM::VM1,"Sequence Name",false },
  {0x0019,0x00cb,"PMTF INFORMATION DATA^10",VR::UL,VM::VM1,"?",false },
  {0x0019,0x00cc,"PMTF INFORMATION DATA^12",VR::SS,VM::VM1_n,"?",false },
  {0x0019,0x00cc,"PMTF INFORMATION DATA^10",VR::UL,VM::VM1,"?",false },
  {0x0019,0x00cd,"PMTF INFORMATION DATA^12",VR::LO,VM::VM1,"?",false },
  {0x0019,0x00cd,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00cd,"PMTF INFORMATION DATA^11",VR::SL,VM::VM3,"?",false },
  {0x0019,0x00ce,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00ce,"PMTF INFORMATION DATA^12",VR::UL,VM::VM1,"?",false },
  {0x0019,0x00cf,"PMTF INFORMATION DATA^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00cf,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?acquired series (vs postproc)?",false },
  {0x0019,0x00d0,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x00d1,"PMTF INFORMATION DATA^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00d1,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00d2,"PMTF INFORMATION DATA^10",VR::SL,VM::VM3,"?",false },
  {0x0019,0x00d2,"PMTF INFORMATION DATA^12",VR::US,VM::VM3,"?",false },
  {0x0019,0x00d3,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00d3,"PMTF INFORMATION DATA^12",VR::SS,VM::VM1,"?",false },
  {0x0019,0x00d4,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00d4,"PMTF INFORMATION DATA^12",VR::US,VM::VM2,"?",false },
  {0x0019,0x00d5,"PMTF INFORMATION DATA^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00d6,"PMTF INFORMATION DATA^12",VR::FL,VM::VM3,"?",false },
  {0x0019,0x00d7,"PMTF INFORMATION DATA^12",VR::US,VM::VM2,"?vect2?",false },
  {0x0019,0x00d8,"PMTF INFORMATION DATA^12",VR::FL,VM::VM1_n,"?",false },
  {0x0019,0x00d8,"PMTF INFORMATION DATA^10",VR::US,VM::VM1_n,"?",false },
  {0x0019,0x00d9,"PMTF INFORMATION DATA^10",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x00da,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x00db,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00dc,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00de,"PMTF INFORMATION DATA^12",VR::US,VM::VM1,"?num?",false },
  {0x0019,0x00df,"PMTF INFORMATION DATA^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x00e0,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00e0,"PMTF INFORMATION DATA^12",VR::US,VM::VM1,"?",false },
  {0x0019,0x00e1,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00e1,"PMTF INFORMATION DATA^12",VR::FL,VM::VM3,"?vect3?",false },
  {0x0019,0x00e1,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00e2,"PMTF INFORMATION DATA^12",VR::FL,VM::VM3,"?vect3?",false },
  {0x0019,0x00e3,"PMTF INFORMATION DATA^11",VR::SL,VM::VM3,"?",false },
  {0x0019,0x00e3,"PMTF INFORMATION DATA^12",VR::TM,VM::VM1,"?time?",false },
  {0x0019,0x00e4,"PMTF INFORMATION DATA^11",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00e4,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00e4,"PMTF INFORMATION DATA^12",VR::US,VM::VM2,"?vect2?",false },
  {0x0019,0x00e5,"PMTF INFORMATION DATA^10",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x00e5,"PMTF INFORMATION DATA^11",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x00e6,"PMTF INFORMATION DATA^10",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x00e6,"PMTF INFORMATION DATA^11",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x00e9,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00ea,"PMTF INFORMATION DATA^10",VR::UI,VM::VM1,"?",false },
  {0x0019,0x00eb,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00eb,"PMTF INFORMATION DATA^12",VR::US,VM::VM1,"?num?",false },
  {0x0019,0x00ec,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x00ec,"PMTF INFORMATION DATA^12",VR::US,VM::VM1,"?num?",false },
  {0x0019,0x00ee,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00ef,"PMTF INFORMATION DATA^12",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00ef,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00f0,"PMTF INFORMATION DATA^12",VR::LO,VM::VM1,"?",false },
  {0x0019,0x00f0,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f0,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1_n,"?",false },
  {0x0019,0x00f1,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00f1,"PMTF INFORMATION DATA^12",VR::LO,VM::VM1,"?",false },
  {0x0019,0x00f1,"PMTF INFORMATION DATA^11",VR::SL,VM::VM3,"Diffusion b-value x Toshiba Orientation",false },
  {0x0019,0x00f2,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00f2,"PMTF INFORMATION DATA^11",VR::FL,VM::VM3,"Specific Absorption Rate Values",false },
  {0x0019,0x00f3,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f3,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f4,"PMTF INFORMATION DATA^10",VR::FD,VM::VM1,"?",false },
  {0x0019,0x00f4,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f5,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f5,"PMTF INFORMATION DATA^11",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x00f6,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f6,"PMTF INFORMATION DATA^10",VR::UI,VM::VM1,"?",false },
  {0x0019,0x00f7,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f8,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f8,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0019,0x00f9,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00f9,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00fa,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00fa,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00fb,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00fb,"PMTF INFORMATION DATA^10",VR::UL,VM::VM1,"?",false },
  {0x0019,0x00fc,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00fc,"PMTF INFORMATION DATA^11",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00fc,"PMTF INFORMATION DATA^12",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00fd,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00fd,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00fd,"PMTF INFORMATION DATA^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00fe,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00fe,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00fe,"PMTF INFORMATION DATA^12",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00ff,"PMTF INFORMATION DATA^11",VR::SL,VM::VM1,"?",false },
  {0x0019,0x00ff,"PMTF INFORMATION DATA^10",VR::SL,VM::VM3,"?",false },
  {0x0021,0x0002,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?num?",false },
  {0x0021,0x0003,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?num?",false },
  {0x0021,0x0004,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?num?",false },
  {0x0021,0x0008,"PMTF INFORMATION DATA^10",VR::DS,VM::VM3,"?vect1?",false },
  {0x0021,0x0009,"PMTF INFORMATION DATA^10",VR::DS,VM::VM3,"?vect2?",false },
  {0x0021,0x000a,"PMTF INFORMATION DATA^10",VR::DS,VM::VM3,"?vect3?",false },
  {0x0021,0x000b,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0021,0x000c,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0021,0x000d,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0021,0x000f,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0x0021,0x0012,"PMTF INFORMATION DATA^10",VR::DS,VM::VM3,"?vect4?",false },
  {0x0021,0x0014,"PMTF INFORMATION DATA^10",VR::SQ,VM::VM1,"?matrix?",false },
  {0x0021,0x0015,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0x0021,0x0016,"PMTF INFORMATION DATA^10",VR::FL,VM::VM2,"?",false },
  {0x0021,0x0018,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"Series Number Local",false },
  {0x0021,0x001a,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?num2?",false },
  {0x0021,0x001d,"PMTF INFORMATION DATA^10",VR::IS,VM::VM1,"?num3?",false },
  {0x0021,0x001f,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?num?",false },
  {0x0029,0x0001,"PMTF INFORMATION DATA^10",VR::SL,VM::VM3,"Diffusion Gradient Orientation / Toshiba Orientation",false },
  {0x0029,0x0005,"PMTF INFORMATION DATA^10",VR::SQ,VM::VM1,"?seq?",false },
  {0x0029,0x0006,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?num?",false },
  {0x0029,0x0007,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?num?",false },
  {0x0029,0x001c,"PMTF INFORMATION DATA^10",VR::FL,VM::VM2,"?vect2?",false },
  {0x0029,0x0021,"PMTF INFORMATION DATA^10",VR::SL,VM::VM3,"?vect3?",false },
  {0x0029,0x004e,"PMTF INFORMATION DATA^10",VR::SL,VM::VM1,"?num?",false },
  {0x0029,0x0050,"PMTF INFORMATION DATA^10",VR::LO,VM::VM1,"?root filepath?",false },
  {0x0029,0x0051,"PMTF INFORMATION DATA^10",VR::LO,VM::VM1,"?series filepath?",false },
  {0x0029,0x0052,"PMTF INFORMATION DATA^10",VR::LO,VM::VM1,"?instance filepath?",false },
  {0x0029,0x0054,"PMTF INFORMATION DATA^10",VR::LO,VM::VM1,"?study filepath?",false },
  {0x0029,0x0067,"PMTF INFORMATION DATA^10",VR::FD,VM::VM4,"Series Diffusion Info (b-value,x,y,z)",false },
  {0x0029,0x0068,"PMTF INFORMATION DATA^10",VR::SS,VM::VM1,"?num 1?",false },
  {0x0029,0x006e,"PMTF INFORMATION DATA^10",VR::OB,VM::VM1,"?",false },
  //{0x0029,0x0089,"PMTF INFORMATION DATA",VR::LO,VM::VM1,"?",false }, /* defined below */
  {0xe201,0x0002,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0xe301,0x0000,"PMTF INFORMATION DATA^10",VR::SH,VM::VM1,"Series Description",false },
  {0xe401,0x0000,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?num?",false },
  {0xe401,0x0001,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?num?",false },
  {0xe401,0x0002,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?num?",false },
  {0xe401,0x0003,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?num?",false },
  {0xe401,0x0004,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?num?",false },
  {0xe401,0x0005,"PMTF INFORMATION DATA^10",VR::OB,VM::VM1,"?blob?",false },
  {0xe401,0x0007,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?",false },
  {0xe401,0x0008,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?ADC factor?",false },
  {0xe401,0x0012,"PMTF INFORMATION DATA^10",VR::FL,VM::VM1,"?num?",false },
  {0xe401,0x0013,"PMTF INFORMATION DATA^10",VR::OB,VM::VM1,"Icon DIB (Microsoft Windows 3.X Packed Device-Independent Bitmap)",false },
  {0xe401,0x0020,"PMTF INFORMATION DATA^10",VR::SQ,VM::VM1,"?seq?",false },
  {0xe401,0x0021,"PMTF INFORMATION DATA^10",VR::SQ,VM::VM1,"?seq?",false },
  {0xe401,0x0022,"PMTF INFORMATION DATA^10",VR::LO,VM::VM1,"?MR.gdc?",false },
  {0xe401,0x0023,"PMTF INFORMATION DATA^10",VR::SH,VM::VM1,"Software Version(s)",false },
  {0xe401,0x0024,"PMTF INFORMATION DATA^10",VR::OB,VM::VM1,"?blob?",false },
  {0xe403,0x0000,"PMTF INFORMATION DATA^13",VR::SH,VM::VM1,"?VIEW?",false },
  {0xe403,0x0000,"PMTF INFORMATION DATA^10",VR::SQ,VM::VM1,"?seq?",false },
  {0xe403,0x0001,"PMTF INFORMATION DATA^13",VR::OB,VM::VM1,"?blob0?",false },
  {0xe403,0x0002,"PMTF INFORMATION DATA^13",VR::OB,VM::VM1,"?blob1?",false },
  {0xe403,0x0003,"PMTF INFORMATION DATA^13",VR::OB,VM::VM1,"?blob2?",false },
  {0xee01,0x0000,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },
  {0xee01,0x0001,"PMTF INFORMATION DATA^10",VR::OB,VM::VM1,"?",false },
  {0xee01,0x0002,"PMTF INFORMATION DATA^10",VR::US,VM::VM1,"?",false },

  {0x3001,0x0001,"Philips3D",VR::UL,VM::VM1,"Number of Time Phases",false },
  {0x3001,0x0002,"Philips3D",VR::US,VM::VM1,"Physical Units Z Direction",false },
  {0x3001,0x0003,"Philips3D",VR::FD,VM::VM1,"Physical Delta Z",false },
  {0x3001,0x0004,"Philips3D",VR::UI,VM::VM1,"?UID?",false },
  {0x6301,0x0001,"KRETZ_PRIVATE",VR::IS,VM::VM1,"?",false },
  {0x6301,0x0002,"KRETZ_PRIVATE",VR::IS,VM::VM1,"?",false },
  {0x0029,0x0000,"VOLCANO-PCDE 1.0",VR::DS,VM::VM1,"Pullback Rate",false },
  {0x0029,0x0001,"VOLCANO-PCDE 1.0",VR::FD,VM::VM1,"B Gain",false },
  {0x0029,0x0002,"VOLCANO-PCDE 1.0",VR::US,VM::VM1,"B Persistence Index",false },
  {0x0029,0x0003,"VOLCANO-PCDE 1.0",VR::FD,VM::VM1,"B ROI Diameter",false },
  {0x0029,0x0004,"VOLCANO-PCDE 1.0",VR::US,VM::VM1,"CF Sensitivity Index",false },
  {0x0029,0x0005,"VOLCANO-PCDE 1.0",VR::FD,VM::VM1,"CF ROI Diameter",false },
  {0x0029,0x0006,"VOLCANO-PCDE 1.0",VR::US,VM::VM1,"Frame Capture Interleave Rate",false },
  {0x0029,0x0007,"VOLCANO-PCDE 1.0",VR::US,VM::VM1,"Ringdown Subtraction",false },
  {0x0029,0x0008,"VOLCANO-PCDE 1.0",VR::US,VM::VM1,"Graticule Spacing",false },
  {0x0029,0x0010,"VOLCANO-PCDE 1.0",VR::SH,VM::VM1,"PIM Serial Number",false },
  {0x0029,0x0011,"VOLCANO-PCDE 1.0",VR::OB,VM::VM1,"Color Palette",false },
  {0x0029,0x0012,"VOLCANO-PCDE 1.0",VR::UT,VM::VM1,"Measurement Data",false },
  {0x0029,0x0015,"VOLCANO-PCDE 1.0",VR::US,VM::VM1,"Still Image Number",false },
  {0x0029,0x0016,"VOLCANO-PCDE 1.0",VR::US,VM::VM1,"Video Loop Number",false },
  {0x0029,0x0030,"VOLCANO-PCDE 1.0",VR::SS,VM::VM1,"Catheter Boot Mode",false },
  {0x0029,0x0010,"ShowcaseAppearance",VR::DS,VM::VM1,"?",false },
  {0x0029,0x0011,"ShowcaseAppearance",VR::DS,VM::VM1,"?",false },
  {0x0029,0x0012,"ShowcaseAppearance",VR::DS,VM::VM1,"?",false },
  {0x0029,0x0013,"ShowcaseAppearance",VR::DS,VM::VM1,"?",false },
  {0x0029,0x0014,"ShowcaseAppearance",VR::SQ,VM::VM1,"?",false },
  {0x2111,0x0010,"Viztek OPALRAD PrivateGroup",VR::UT,VM::VM1,"XML annotations",false },
  {0x0caf,0x0011,"Xinapse Systems",VR::LO,VM::VM1,"Private attributes Identification Code",false },
  {0x0caf,0x0001,"Xinapse Systems",VR::OB,VM::VM1,"Regions of interest",false },
  {0x0043,0x0014,"dcm4che/archive",VR::AE,VM::VM1,"ae name 1",false },
  {0x0043,0x0015,"dcm4che/archive",VR::AE,VM::VM1,"ae name 2",false },
  {0x0009,0x0001,"DicomUtils 20100512",VR::SQ,VM::VM1,"DICOMUTILS_PROCESSING_METHODS",false },
  {0x0009,0x0002,"DicomUtils 20100512",VR::LO,VM::VM1,"DICOMUTILS_VERSION",false },
  {0x0009,0x0003,"DicomUtils 20100512",VR::LT,VM::VM1,"DICOMUTILS_ARGS",false },
  {0x0009,0x0004,"DicomUtils 20100512",VR::LT,VM::VM1,"DICOMUTILS_COMMENT",false },
  {0x0009,0x0005,"DicomUtils 20100512",VR::LT,VM::VM1,"DICOMUTILS_SITE",false },
  {0x0053,0x0001,"3Z_IMAGE_IMPORT_CLIENT_INFO",VR::LO,VM::VM1,"?Date+Time?",false },
  {0x0053,0x0002,"3Z_IMAGE_IMPORT_CLIENT_INFO",VR::LO,VM::VM1,"?Org Name?",false },
  {0x0053,0x0003,"3Z_IMAGE_IMPORT_CLIENT_INFO",VR::LO,VM::VM1,"?Software Version?",false },
  {0x0053,0x0004,"3Z_IMAGE_IMPORT_CLIENT_INFO",VR::LO,VM::VM1,"?ID?",false },
  {0x0053,0x0005,"3Z_IMAGE_IMPORT_CLIENT_INFO",VR::LO,VM::VM1,"?IP Address?",false },
  {0x0053,0x0006,"3Z_IMAGE_IMPORT_CLIENT_INFO",VR::LO,VM::VM1,"?AE Title?",false },
  {0x0055,0x0001,"3Z_IMAGE_IMPORT_GENERAL_HIST",VR::LO,VM::VM1_n,"Copy of Specific Character Set (0008,0005)",false },
  {0x0055,0x0003,"3Z_IMAGE_IMPORT_GENERAL_HIST",VR::LO,VM::VM3,"Copy of Study Instance UID (0020,000d)",false },
  {0x0055,0x0004,"3Z_IMAGE_IMPORT_GENERAL_HIST",VR::LO,VM::VM3,"Copy of Series Instance UID (0020,000e)",false },
  {0x0055,0x0005,"3Z_IMAGE_IMPORT_GENERAL_HIST",VR::LO,VM::VM3,"Copy of SOP Instance UID (0008,0018)",false },
  {0x0055,0x0006,"3Z_IMAGE_IMPORT_GENERAL_HIST",VR::LO,VM::VM3,"Copy of Patient ID (0010,0010)",false },
  {0x0055,0x0007,"3Z_IMAGE_IMPORT_GENERAL_HIST",VR::LO,VM::VM3,"Copy of Patient's Name (0010,0010)",false },
  {0x0055,0x0008,"3Z_IMAGE_IMPORT_GENERAL_HIST",VR::LO,VM::VM3,"Copy of Patient's Birth Date (0010,1030)",false },
  {0x0055,0x0009,"3Z_IMAGE_IMPORT_GENERAL_HIST",VR::LO,VM::VM3,"Copy of Patient's Sex (0010,0040)",false },
  {0x0055,0x0012,"3Z_IMAGE_IMPORT_GENERAL_HIST",VR::LO,VM::VM3,"Copy of Study Description (0008,1030)",false },
  {0x0055,0x0013,"3Z_IMAGE_IMPORT_GENERAL_HIST",VR::LO,VM::VM3,"Copy of Accession Number (0008,0050)",false },
  {0x0029,0x0000,"1.2.840.113663.1",VR::US,VM::VM1,"?",false },
  {0x0029,0x0001,"1.2.840.113663.1",VR::US,VM::VM1,"?",false },
  {0x0019,0x0010,"1.2.840.113681",VR::ST,VM::VM1,"CR Image Params Common",false },
  {0x0019,0x0011,"1.2.840.113681",VR::ST,VM::VM1,"CR Image IP Params Single",false },
  {0x0019,0x0012,"1.2.840.113681",VR::ST,VM::VM1,"CR Image IP Params Left",false },
  {0x0019,0x0013,"1.2.840.113681",VR::ST,VM::VM1,"CR Image IP Params Right",false },
  {0x0087,0x0010,"1.2.840.113708.794.1.1.2.0",VR::CS,VM::VM1,"Media Type",false },
  {0x0087,0x0020,"1.2.840.113708.794.1.1.2.0",VR::CS,VM::VM1,"Media Location",false },
  {0x0087,0x0030,"1.2.840.113708.794.1.1.2.0",VR::ST,VM::VM1,"Storage File ID",false },
  {0x0087,0x0040,"1.2.840.113708.794.1.1.2.0",VR::DS,VM::VM1,"Study or Image Size in MB",false },
  {0x0087,0x0050,"1.2.840.113708.794.1.1.2.0",VR::IS,VM::VM1,"Estimated Retrieve Time",false },
  {0x0029,0x0020,"2.16.840.1.114059.1.1.6.1.50.1",VR::LT,VM::VM1,"Description",false },
  {0x0029,0x0021,"2.16.840.1.114059.1.1.6.1.50.1",VR::ST,VM::VM1,"Orientation",false },
  {0x0029,0x0022,"2.16.840.1.114059.1.1.6.1.50.1",VR::ST,VM::VM1,"Parameter 1",false },
  {0x0029,0x0023,"2.16.840.1.114059.1.1.6.1.50.1",VR::ST,VM::VM1,"Parameter 2",false },
  {0x0029,0x0024,"2.16.840.1.114059.1.1.6.1.50.1",VR::LO,VM::VM1,"Teeth",false },
  {0x0029,0x0025,"2.16.840.1.114059.1.1.6.1.50.1",VR::LO,VM::VM1,"Jaw",false },
  {0x0029,0x0026,"2.16.840.1.114059.1.1.6.1.50.1",VR::LO,VM::VM1,"Quadrant",false },
  {0x0029,0x0027,"2.16.840.1.114059.1.1.6.1.50.1",VR::LO,VM::VM1,"CRC",false },
  {0x3711,0x0001,"A.L.I. Technologies, Inc.",VR::LO,VM::VM1,"Filename",false },
  {0x3711,0x0002,"A.L.I. Technologies, Inc.",VR::OB,VM::VM1,"Data Blob of a Visit",false },
  {0x3711,0x0003,"A.L.I. Technologies, Inc.",VR::US,VM::VM1,"Revision Number",false },
  {0x3711,0x0004,"A.L.I. Technologies, Inc.",VR::UL,VM::VM1,"Unix Timestamp",false },
  {0x3711,0x0005,"A.L.I. Technologies, Inc.",VR::IS,VM::VM1,"Bag ID",false },
  {0x3711,0x000c,"A.L.I. Technologies, Inc.",VR::UI,VM::VM1,"Original Study UID",false },
  {0x3711,0x000d,"A.L.I. Technologies, Inc.",VR::US,VM::VM1,"Overlay Grayscale Value",false },
  {0x3711,0x000e,"A.L.I. Technologies, Inc.",VR::CS,VM::VM1,"Anonymization Status",false },
  {0x0009,0x0000,"ACUSON",VR::IS,VM::VM1,"?",false },
  {0x0009,0x0001,"ACUSON",VR::IS,VM::VM1,"?",false },
  {0x0009,0x0002,"ACUSON",VR::UN,VM::VM1,"?",false },
  {0x0009,0x0003,"ACUSON",VR::UN,VM::VM1,"?",false },
  {0x0009,0x0004,"ACUSON",VR::UN,VM::VM1,"?",false },
  {0x0009,0x0005,"ACUSON",VR::UN,VM::VM1,"?",false },
  {0x0009,0x0006,"ACUSON",VR::UN,VM::VM1,"?",false },
  {0x0009,0x0007,"ACUSON",VR::UN,VM::VM1,"?",false },
  {0x0009,0x0008,"ACUSON",VR::LT,VM::VM1,"?",false },
  {0x0009,0x0009,"ACUSON",VR::LT,VM::VM1,"?",false },
  {0x0009,0x000a,"ACUSON",VR::IS,VM::VM1,"?",false },
  {0x0009,0x000b,"ACUSON",VR::IS,VM::VM1,"?",false },
  {0x0009,0x000c,"ACUSON",VR::IS,VM::VM1,"?",false },
  {0x0009,0x000d,"ACUSON",VR::IS,VM::VM1,"?",false },
  {0x0009,0x000e,"ACUSON",VR::IS,VM::VM1,"?",false },
  {0x0009,0x000f,"ACUSON",VR::UN,VM::VM1,"?",false },
  {0x0009,0x0010,"ACUSON",VR::IS,VM::VM1,"?",false },
  {0x0009,0x0011,"ACUSON",VR::UN,VM::VM1,"?",false },
  {0x0009,0x0012,"ACUSON",VR::IS,VM::VM1,"?",false },
  {0x0009,0x0013,"ACUSON",VR::IS,VM::VM1,"?",false },
  {0x0009,0x0014,"ACUSON",VR::LT,VM::VM1,"?",false },
  {0x0009,0x0015,"ACUSON",VR::UN,VM::VM1,"?",false },
  {0x7fdf,0x0000,"ACUSON: 1.2.840.11386.1.0",VR::IS,VM::VM1,"Lossy Compression Ratio",false },
  {0x7fdf,0x0001,"ACUSON: 1.2.840.11386.1.0",VR::US,VM::VM1,"Image Format",false },
  {0x7fdf,0x0002,"ACUSON: 1.2.840.11386.1.0",VR::US,VM::VM1,"Acuson Region Type",false },
  {0x7fdf,0x000b,"ACUSON: 1.2.840.11386.1.0",VR::UL,VM::VM1,"Acuson Image Apex X",false },
  {0x7fdf,0x000c,"ACUSON: 1.2.840.11386.1.0",VR::UL,VM::VM1,"Acuson Image Apex Y",false },
  {0x7fdf,0x000d,"ACUSON: 1.2.840.11386.1.0",VR::IS,VM::VM1,"B-Color-On Flag",false },
  {0x0009,0x0020,"ACUSON:1.2.840.113680.1.0:0921",VR::UN,VM::VM1,"View Name",false },
  {0x0009,0x002a,"ACUSON:1.2.840.113680.1.0:0921",VR::UN,VM::VM1,"View List",false },
  {0x7fdf,0x0000,"ACUSON:1.2.840.113680.1.0:7f10",VR::IS,VM::VM1,"Lossy Compression Ratio",false },
  {0x7fdf,0x0001,"ACUSON:1.2.840.113680.1.0:7f10",VR::US,VM::VM1,"Image Format",false },
  {0x7fdf,0x0002,"ACUSON:1.2.840.113680.1.0:7f10",VR::US,VM::VM1,"Acuson Region Type",false },
  {0x7fdf,0x000b,"ACUSON:1.2.840.113680.1.0:7f10",VR::UL,VM::VM1,"Acuson Image Apex X",false },
  {0x7fdf,0x000c,"ACUSON:1.2.840.113680.1.0:7f10",VR::UL,VM::VM1,"Acuson Image Apex Y",false },
  {0x7fdf,0x000d,"ACUSON:1.2.840.113680.1.0:7f10",VR::IS,VM::VM1,"B-Color-On Flag",false },
  {0x7fdf,0x000e,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Acuson Mechanical Apex X",false },
  {0x7fdf,0x000f,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Acuson Mechanical Apex Y",false },
  {0x7fdf,0x0010,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Acquisition Type:",false },
  {0x7fdf,0x0018,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Transformation Matrix Sequence",false },
  {0x7fdf,0x0020,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Left angle",false },
  {0x7fdf,0x0022,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Right angle",false },
  {0x7fdf,0x0024,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Color Map Family",false },
  {0x7fdf,0x0025,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Full Colormap.",false },
  {0x7fdf,0x0026,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Color Invert",false },
  {0x7fdf,0x0027,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Color Baseline",false },
  {0x7fdf,0x0028,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"CD Color Mix Points X1",false },
  {0x7fdf,0x0029,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"CD Color Mix Points Y1",false },
  {0x7fdf,0x002a,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"CD Color Mix Points X2",false },
  {0x7fdf,0x002b,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"CD Color Mix Points Y2",false },
  {0x7fdf,0x002c,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Color Accent",false },
  {0x7fdf,0x0030,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Persistence SQ",false },
  {0x7fdf,0x0031,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Persistence Mode",false },
  {0x7fdf,0x0032,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Persistence Coefficient Mode",false },
  {0x7fdf,0x0033,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Alpha coefficient",false },
  {0x7fdf,0x0034,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Gamma coefficient",false },
  {0x7fdf,0x0035,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Persistence Time Flag",false },
  {0x7fdf,0x0036,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Persistence adaptive flag",false },
  {0x7fdf,0x0037,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Persistence Frame Rate",false },
  {0x7fdf,0x0038,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Persistence ID",false },
  {0x7fdf,0x0040,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Observation Date Time SQ",false },
  {0x7fdf,0x0050,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Capture Type Name",false },
  {0x7fdf,0x0052,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Capture Type Number",false },
  {0x7fdf,0x0054,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Number of Capture Types",false },
  {0x7fdf,0x0060,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"CD Steering Angle",false },
  {0x7fdf,0x0061,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"CD PRI",false },
  {0x7fdf,0x0062,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"CD Dynamic Range",false },
  {0x7fdf,0x0063,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"CD Velocity Scale Min",false },
  {0x7fdf,0x0064,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"CD Velocity Scale Max",false },
  {0x7fdf,0x0065,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"CD Color Mode",false },
  {0x7fdf,0x0066,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"CD Frequency",false },
  {0x7fdf,0x0067,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"CD Balance",false },
  {0x7fdf,0x0068,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"CD Delta",false },
  {0x7fdf,0x0069,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"CD Pan Box Min X0",false },
  {0x7fdf,0x006a,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"CD Pan Box Min Y0",false },
  {0x7fdf,0x006b,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"CD Pan Box Min X1",false },
  {0x7fdf,0x006c,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"CD Pan Box Min Y1",false },
  {0x7fdf,0x006d,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"CPS Map Type",false },
  {0x7fdf,0x006e,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"CPS Map Data",false },
  {0x7fdf,0x006f,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"CPS Balance Setting",false },
  {0x7fdf,0x0070,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"3DCard Step Angle",false },
  {0x7fdf,0x0071,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"3DCard Xdcr Angle",false },
  {0x7fdf,0x0072,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"B-mode Frequency",false },
  {0x7fdf,0x0073,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"B-mode Dynamic Range",false },
  {0x7fdf,0x0074,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"B-mode Frame Rate",false },
  {0x7fdf,0x0075,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"B-mode Space Time",false },
  {0x7fdf,0x0076,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"B-mode Persistence",false },
  {0x7fdf,0x0077,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"B-mode Display Depth Start",false },
  {0x7fdf,0x0078,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"B-mode Display Depth End",false },
  {0x7fdf,0x0079,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"B-mode Res Mode",false },
  {0x7fdf,0x007a,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"B-mode Preset Application",false },
  {0x7fdf,0x007b,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Image Spec Name",false },
  {0x7fdf,0x007c,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"B Preset Image Look",false },
  {0x7fdf,0x007d,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"B-mode Post Processing",false },
  {0x7fdf,0x007e,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"B Edge",false },
  {0x7fdf,0x007f,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"B Delta",false },
  {0x7fdf,0x0080,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"B-mode 1D Post Processing Curve",false },
  {0x7fdf,0x0081,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"B-mode Delta (ECRI) Map Diagonal",false },
  {0x7fdf,0x0082,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Bytes Per Timestamp",false },
  {0x7fdf,0x0083,"ACUSON:1.2.840.113680.1.0:7f10",VR::UL,VM::VM1,"Microseconds in unit timestamp",false },
  {0x7fdf,0x0084,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Start Stopwatch Timestamp",false },
  {0x7fdf,0x0085,"ACUSON:1.2.840.113680.1.0:7f10",VR::OB,VM::VM1,"Acoustic Frame Timestamp",false },
  {0x7fdf,0x0086,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"R-Wave Timestamp",false },
  {0x7fdf,0x0087,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Last Destruction Timestamp",false },
  {0x7fdf,0x0088,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Pixels Per Second",false },
  {0x7fdf,0x0089,"ACUSON:1.2.840.113680.1.0:7f10",VR::OB,VM::VM1,"ECG Reference Timestamp",false },
  {0x7fdf,0x008a,"ACUSON:1.2.840.113680.1.0:7f10",VR::FD,VM::VM1,"ECG Sampling Interval (milliseconds)",false },
  {0x7fdf,0x008b,"ACUSON:1.2.840.113680.1.0:7f10",VR::UL,VM::VM1,"ECG Sample Count",false },
  {0x7fdf,0x008c,"ACUSON:1.2.840.113680.1.0:7f10",VR::UL,VM::VM1,"ECG Sample Size",false },
  {0x7fdf,0x008d,"ACUSON:1.2.840.113680.1.0:7f10",VR::OB,VM::VM1,"ECG Data Value",false },
  {0x7fdf,0x008e,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Contrast/Active Image Indicator",false },
  {0x7fdf,0x008f,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Live Dual Mode Indicator",false },
  {0x7fdf,0x0090,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"3DCard Clipset ID",false },
  {0x7fdf,0x0091,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"3DCard HRWave Min",false },
  {0x7fdf,0x0092,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"3DCard HRWave Max",false },
  {0x7fdf,0x0093,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Perspective Capture Type",false },
  {0x7fdf,0x00f1,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Trigger Mask.",false },
  {0x7fdf,0x00f2,"ACUSON:1.2.840.113680.1.0:7f10",VR::LO,VM::VM1,"Study Directory",false },
  {0x7fdf,0x00f3,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Last Modify Date",false },
  {0x7fdf,0x00f4,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Last Modify Time",false },
  {0x7fdf,0x00f5,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Teaching Study",false },
  {0x7fdf,0x00f6,"ACUSON:1.2.840.113680.1.0:7f10",VR::UN,VM::VM1,"Series Base UID",false },
  {0x7fdf,0x0000,"ACUSON:1.2.840.113680.1.0:7ffe",VR::OB,VM::VM1,"Data Padding",false },
  {0x0019,0x0002,"ADAC_IMG",VR::IS,VM::VM1,"Ver200 ADAC Pegasys File Size",false },
  {0x0019,0x0010,"ADAC_IMG",VR::LO,VM::VM2,"ADAC Header Signature",false },
  {0x0019,0x0011,"ADAC_IMG",VR::US,VM::VM1,"Number of ADAC Headers",false },
  {0x0019,0x0012,"ADAC_IMG",VR::IS,VM::VM1_n,"ADAC Header/Image Sizes",false },
  {0x0019,0x0020,"ADAC_IMG",VR::OB,VM::VM1,"ADAC Pegasys Headers",false },
  {0x0019,0x0021,"ADAC_IMG",VR::US,VM::VM1,"Ver200 Number of ADAC Headers",false },
  {0x0019,0x0041,"ADAC_IMG",VR::IS,VM::VM1_n,"Ver200 ADAC Header/Image Size",false },
  {0x0019,0x0061,"ADAC_IMG",VR::OB,VM::VM1,"Ver200 ADAC Pegasys Headers",false },
  {0x7043,0x0000,"ADAC_IMG",VR::SH,VM::VM1,"Cardiac Stress State",false },
  {0x7043,0x0010,"ADAC_IMG",VR::LO,VM::VM1,"Philips NM Private Group",false },
  {0x0003,0x0000,"AEGIS_DICOM_2.00",VR::US,VM::VM1_n,"?",false },
  {0x0005,0x0000,"AEGIS_DICOM_2.00",VR::US,VM::VM1_n,"?",false },
  {0x0009,0x0000,"AEGIS_DICOM_2.00",VR::US,VM::VM1_n,"?",false },
  {0x0019,0x0000,"AEGIS_DICOM_2.00",VR::US,VM::VM1_n,"?",false },
  {0x0029,0x0000,"AEGIS_DICOM_2.00",VR::US,VM::VM1_n,"?",false },
  {0x1369,0x0000,"AEGIS_DICOM_2.00",VR::US,VM::VM1_n,"?",false },
  {0x0009,0x0010,"AGFA",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0011,"AGFA",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0013,"AGFA",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0014,"AGFA",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0015,"AGFA",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0010,"AGFA",VR::SH,VM::VM1,"Private Identification Code",false },
  {0x0019,0x0011,"AGFA",VR::LO,VM::VM3,"Identification Data (Note 2)",false },
  {0x0019,0x0013,"AGFA",VR::LO,VM::VM1,"Sensitometry Name",false },
  {0x0019,0x0014,"AGFA",VR::ST,VM::VM3,"Window/Level List (Note 3)",false },
  {0x0019,0x0015,"AGFA",VR::LO,VM::VM1,"Dose Monitoring List",false },
  {0x0019,0x0016,"AGFA",VR::LO,VM::VM3,"Other Info (Note 5)",false },
  {0x0019,0x001a,"AGFA",VR::LO,VM::VM1,"Clipped Exposure Deviation",false },
  {0x0019,0x001b,"AGFA",VR::LO,VM::VM1,"Logarithmic PLT Full Scale",false },
  {0x0019,0x0060,"AGFA",VR::US,VM::VM1,"Total number of series",false },
  {0x0019,0x0061,"AGFA",VR::SH,VM::VM1,"Session Number",false },
  {0x0019,0x0062,"AGFA",VR::SH,VM::VM1,"ID Station name",false },
  {0x0019,0x0065,"AGFA",VR::US,VM::VM1,"Number of images in study to be transmitted (only sent with autoverify: on)",false },
  {0x0019,0x0070,"AGFA",VR::US,VM::VM1,"Total number of images",false },
  {0x0019,0x0080,"AGFA",VR::ST,VM::VM1,"Geometrical Transformations",false },
  {0x0019,0x0081,"AGFA",VR::ST,VM::VM1,"Roam Origin",false },
  {0x0019,0x0082,"AGFA",VR::US,VM::VM1,"Zoom factor",false },
  {0x0019,0x0093,"AGFA",VR::CS,VM::VM1,"Status",false },
  {0x0031,0x0000,"AGFA PACS Archive Mirroring 1.0",VR::CS,VM::VM1,"study_status",false },
  /* 001310_IMPAX_6.5.x_Solution_DICOM_Conformance_Statement.pdf states is is VR:CS */
  {0x0031,0x0001,"AGFA PACS Archive Mirroring 1.0",VR::UL,VM::VM1,"date_time_verified",false },
  {0x0035,0x0000,"AGFA STUDY COMMENTS 1.0",VR::LT,VM::VM1,"Array of Comments (illegal XML)",false },
  {0x0071,0x0018,"AGFA-AG_HPState",VR::SQ,VM::VM1,"?",false },
  {0x0071,0x0019,"AGFA-AG_HPState",VR::SQ,VM::VM1,"?",false },
  {0x0071,0x001a,"AGFA-AG_HPState",VR::SQ,VM::VM1,"?",false },
  {0x0071,0x001c,"AGFA-AG_HPState",VR::SQ,VM::VM1,"?",false },
  {0x0071,0x001e,"AGFA-AG_HPState",VR::SQ,VM::VM1,"?",false },
  {0x0071,0x0020,"AGFA-AG_HPState",VR::FL,VM::VM1_n,"?",false },
  {0x0071,0x0021,"AGFA-AG_HPState",VR::FD,VM::VM1_n,"?",false },
  {0x0071,0x0022,"AGFA-AG_HPState",VR::FD,VM::VM1_n,"?",false },
  {0x0071,0x0023,"AGFA-AG_HPState",VR::FD,VM::VM1_n,"?",false },
  {0x0071,0x0024,"AGFA-AG_HPState",VR::FD,VM::VM1,"?",false },
  {0x0073,0x0023,"AGFA-AG_HPState",VR::SH,VM::VM1,"?",false },
  {0x0073,0x0024,"AGFA-AG_HPState",VR::SQ,VM::VM1,"?",false },
  {0x0073,0x0028,"AGFA-AG_HPState",VR::SQ,VM::VM1,"?",false },
  {0x0073,0x0080,"AGFA-AG_HPState",VR::FL,VM::VM1,"?",false },
  {0x0075,0x0010,"AGFA-AG_HPState",VR::LO,VM::VM1,"?",false },
  {0x0087,0x0001,"AGFA-AG_HPState",VR::LO,VM::VM1,"?",false },
  {0x0087,0x0002,"AGFA-AG_HPState",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0005,"AGFA_ADC_Compact",VR::ST,VM::VM1,"Data stream from cassette",false },
  {0x0019,0x0010,"AGFA_ADC_Compact",VR::LO,VM::VM1,"Private Identification Code",false },
  {0x0019,0x0030,"AGFA_ADC_Compact",VR::ST,VM::VM1,"Set of destination types",false },
  {0x0019,0x0040,"AGFA_ADC_Compact",VR::ST,VM::VM1,"Set of destination Ids",false },
  {0x0019,0x0050,"AGFA_ADC_Compact",VR::ST,VM::VM1,"Set of processing codes",false },
  {0x0019,0x0060,"AGFA_ADC_Compact",VR::US,VM::VM1,"Number of series in study",false },
  {0x0019,0x0061,"AGFA_ADC_Compact",VR::US,VM::VM1,"Session Number",false },
  {0x0019,0x0062,"AGFA_ADC_Compact",VR::SH,VM::VM1,"ID station name",false },
  {0x0019,0x0070,"AGFA_ADC_Compact",VR::US,VM::VM1,"Number of images in series",false },
  {0x0019,0x0071,"AGFA_ADC_Compact",VR::US,VM::VM1,"Break condition",false },
  {0x0019,0x0072,"AGFA_ADC_Compact",VR::US,VM::VM1,"Wait (or Hold) flag",false },
  {0x0019,0x0073,"AGFA_ADC_Compact",VR::US,VM::VM1,"ScanRes flag",false },
  {0x0019,0x0074,"AGFA_ADC_Compact",VR::SH,VM::VM1,"Operation code",false },
  {0x0019,0x0095,"AGFA_ADC_Compact",VR::CS,VM::VM1,"Image quality",false },
  {0x0009,0x0000,"ALOKA:1.2.392.200039.105.2",VR::SH,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x0004,"ALOKA:1.2.392.200039.105.2",VR::US,VM::VM1_n,"Private attribute without PHI",false },
  {0x0009,0x0006,"ALOKA:1.2.392.200039.105.2",VR::US,VM::VM1_n,"Private attribute without PHI",false },
  {0x0009,0x000a,"ALOKA:1.2.392.200039.105.2",VR::SH,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x0012,"ALOKA:1.2.392.200039.105.2",VR::US,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x0014,"ALOKA:1.2.392.200039.105.2",VR::FD,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x001a,"ALOKA:1.2.392.200039.105.2",VR::LO,VM::VM1,"??",false }, // FIXME ?
  {0x0009,0x0020,"ALOKA:1.2.392.200039.105.2",VR::CS,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x0022,"ALOKA:1.2.392.200039.105.2",VR::CS,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x0024,"ALOKA:1.2.392.200039.105.2",VR::CS,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x0026,"ALOKA:1.2.392.200039.105.2",VR::IS,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x0028,"ALOKA:1.2.392.200039.105.2",VR::IS,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x002a,"ALOKA:1.2.392.200039.105.2",VR::DS,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x002c,"ALOKA:1.2.392.200039.105.2",VR::FD,VM::VM6_6n,"Private attribute without PHI",false },
  {0x0009,0x002e,"ALOKA:1.2.392.200039.105.2",VR::FD,VM::VM3_3n,"Private attribute without PHI",false },
  {0x0009,0x0030,"ALOKA:1.2.392.200039.105.2",VR::FD,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x0032,"ALOKA:1.2.392.200039.105.2",VR::DS,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x0034,"ALOKA:1.2.392.200039.105.2",VR::CS,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x0036,"ALOKA:1.2.392.200039.105.2",VR::FD,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x0038,"ALOKA:1.2.392.200039.105.2",VR::CS,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x003a,"ALOKA:1.2.392.200039.105.2",VR::FD,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x003c,"ALOKA:1.2.392.200039.105.2",VR::FD,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x003e,"ALOKA:1.2.392.200039.105.2",VR::FD,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x0040,"ALOKA:1.2.392.200039.105.2",VR::FD,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x0042,"ALOKA:1.2.392.200039.105.2",VR::FD,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x0044,"ALOKA:1.2.392.200039.105.2",VR::FD,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x0046,"ALOKA:1.2.392.200039.105.2",VR::US,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x0048,"ALOKA:1.2.392.200039.105.2",VR::FD,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x004a,"ALOKA:1.2.392.200039.105.2",VR::FD,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x004c,"ALOKA:1.2.392.200039.105.2",VR::FD,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x004e,"ALOKA:1.2.392.200039.105.2",VR::FD,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x0050,"ALOKA:1.2.392.200039.105.2",VR::US,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x0052,"ALOKA:1.2.392.200039.105.2",VR::US,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x0054,"ALOKA:1.2.392.200039.105.2",VR::LO,VM::VM1,"Private attribute without PHI",false },
  {0x0019,0x0008,"ALOKA:1.2.392.200039.105.2",VR::FD,VM::VM1,"Private attribute without PHI",false },
  {0x0019,0x000c,"ALOKA:1.2.392.200039.105.2",VR::CS,VM::VM1,"Private attribute without PHI",false },
  {0x0019,0x000e,"ALOKA:1.2.392.200039.105.2",VR::DS,VM::VM1,"Private attribute without PHI",false },
  {0x0019,0x0018,"ALOKA:1.2.392.200039.105.2",VR::SL,VM::VM1,"Private attribute without PHI",false },
  {0x0019,0x001a,"ALOKA:1.2.392.200039.105.2",VR::SL,VM::VM1,"Private attribute without PHI",false },
  {0x0019,0x0040,"ALOKA:1.2.392.200039.105.2",VR::SS,VM::VM1,"Private attribute without PHI",false },
  {0x0019,0x0046,"ALOKA:1.2.392.200039.105.2",VR::US,VM::VM1,"Private attribute without PHI",false },
  {0x0019,0x0050,"ALOKA:1.2.392.200039.105.2",VR::SL,VM::VM1,"Private attribute without PHI",false },
  {0x0019,0x0052,"ALOKA:1.2.392.200039.105.2",VR::DS,VM::VM1,"Private attribute without PHI",false },
  {0x0019,0x0054,"ALOKA:1.2.392.200039.105.2",VR::DS,VM::VM1,"Private attribute without PHI",false },
  {0x0019,0x0056,"ALOKA:1.2.392.200039.105.2",VR::FD,VM::VM1,"Private attribute without PHI",false },
  {0x0019,0x0060,"ALOKA:1.2.392.200039.105.2",VR::US,VM::VM1,"Private attribute without PHI",false },
  {0x0019,0x0061,"ALOKA:1.2.392.200039.105.2",VR::UL,VM::VM1,"Private attribute without PHI",false },
  {0x0019,0x0062,"ALOKA:1.2.392.200039.105.2",VR::US,VM::VM1,"Private attribute without PHI",false },
  {0x0019,0x0064,"ALOKA:1.2.392.200039.105.2",VR::US,VM::VM1,"Private attribute without PHI",false },
  {0x0019,0x0066,"ALOKA:1.2.392.200039.105.2",VR::US,VM::VM1,"Private attribute without PHI",false },
  {0x0019,0x006c,"ALOKA:1.2.392.200039.105.2",VR::FD,VM::VM1,"Private attribute without PHI",false },
  {0x0019,0x006e,"ALOKA:1.2.392.200039.105.2",VR::FD,VM::VM1,"Private attribute without PHI",false },
  {0x0009,0x0000,"ALOKA:1.2.392.200039.103.2",VR::SH,VM::VM1,"?",false },
  {0x0009,0x0004,"ALOKA:1.2.392.200039.103.2",VR::US,VM::VM1_n,"?",false },
  {0x0009,0x0006,"ALOKA:1.2.392.200039.103.2",VR::US,VM::VM1_n,"?",false },
  {0x0009,0x000a,"ALOKA:1.2.392.200039.103.2",VR::SH,VM::VM1,"?",false },
  {0x0009,0x0020,"ALOKA:1.2.392.200039.103.2",VR::CS,VM::VM1,"?",false },
  {0x0009,0x0022,"ALOKA:1.2.392.200039.103.2",VR::CS,VM::VM1,"?",false },
  {0x0009,0x0024,"ALOKA:1.2.392.200039.103.2",VR::CS,VM::VM1,"?",false },
  {0x0009,0x0026,"ALOKA:1.2.392.200039.103.2",VR::IS,VM::VM1,"?",false },
  {0x0009,0x0028,"ALOKA:1.2.392.200039.103.2",VR::IS,VM::VM1,"?",false },
  {0x0009,0x002a,"ALOKA:1.2.392.200039.103.2",VR::DS,VM::VM1,"?",false },
  {0x0009,0x0030,"ALOKA:1.2.392.200039.103.2",VR::FD,VM::VM1,"?",false },
  {0x0009,0x0032,"ALOKA:1.2.392.200039.103.2",VR::DS,VM::VM1,"?",false },
  {0x0009,0x0034,"ALOKA:1.2.392.200039.103.2",VR::CS,VM::VM1,"?",false },
  {0x0019,0x0008,"ALOKA:1.2.392.200039.103.2",VR::FD,VM::VM1,"?",false },
  {0x0019,0x000c,"ALOKA:1.2.392.200039.103.2",VR::CS,VM::VM1,"?",false },
  {0x0019,0x000e,"ALOKA:1.2.392.200039.103.2",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0018,"ALOKA:1.2.392.200039.103.2",VR::SL,VM::VM1,"?",false },
  {0x0019,0x001a,"ALOKA:1.2.392.200039.103.2",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0040,"ALOKA:1.2.392.200039.103.2",VR::SS,VM::VM1,"?",false },
  {0x0019,0x0046,"ALOKA:1.2.392.200039.103.2",VR::US,VM::VM1,"?",false },
  {0x0019,0x0050,"ALOKA:1.2.392.200039.103.2",VR::SL,VM::VM1,"?",false },
  {0x0019,0x0052,"ALOKA:1.2.392.200039.103.2",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0054,"ALOKA:1.2.392.200039.103.2",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0056,"ALOKA:1.2.392.200039.103.2",VR::FD,VM::VM1,"?",false },
  {0x3101,0x0010,"AMI Annotations_01",VR::SQ,VM::VM1,"AMI Annotation Sequence (RET)",false },
  {0x3101,0x0020,"AMI Annotations_02",VR::SQ,VM::VM1,"AMI Annotation Sequence (RET)",false },
  {0x3107,0x00a0,"AMI ImageContextExt_01",VR::CS,VM::VM1,"AMI Window Function (RET)",false },
  {0x3107,0x00b0,"AMI ImageContextExt_01",VR::DS,VM::VM1,"AMI Window Slope (RET)",false },
  {0x3109,0x0010,"AMI ImageContext_01",VR::CS,VM::VM1,"AMI Window Invert (RET)",false },
  {0x3109,0x0020,"AMI ImageContext_01",VR::IS,VM::VM1,"AMI Window Center (RET)",false },
  {0x3109,0x0030,"AMI ImageContext_01",VR::IS,VM::VM1,"AMI Window Widith (RET)",false },
  {0x3109,0x0040,"AMI ImageContext_01",VR::CS,VM::VM1,"AMI Pixel Aspect Ratio Swap (RET)",false },
  {0x3109,0x0050,"AMI ImageContext_01",VR::CS,VM::VM1,"AMI Enable Averaging (RET)",false },
  {0x3109,0x0060,"AMI ImageContext_01",VR::CS,VM::VM1,"AMI Quality (RET)",false },
  {0x3109,0x0070,"AMI ImageContext_01",VR::CS,VM::VM1,"AMI Viewport Annotation Level (RET)",false },
  {0x3109,0x0080,"AMI ImageContext_01",VR::CS,VM::VM1,"AMI Show Image Annotation (RET)",false },
  {0x3109,0x0090,"AMI ImageContext_01",VR::CS,VM::VM1,"AMI Show Image Overlay (RET)",false },
  {0x3107,0x0010,"AMI ImageTransform_01",VR::DS,VM::VM1,"AMI Transformation Matrix (RET)",false },
  {0x3107,0x0020,"AMI ImageTransform_01",VR::DS,VM::VM1,"AMI Center Offset (RET)",false },
  {0x3107,0x0030,"AMI ImageTransform_01",VR::DS,VM::VM1,"AMI Magnification (RET)",false },
  {0x3107,0x0040,"AMI ImageTransform_01",VR::CS,VM::VM1,"AMI Magnification Type (RET)",false },
  {0x3107,0x0050,"AMI ImageTransform_01",VR::DS,VM::VM1,"AMI Displayed Area (RET)",false },
  {0x3107,0x0060,"AMI ImageTransform_01",VR::DS,VM::VM1,"AMI Calibration Factor (RET)",false },
  {0x3105,0x0010,"AMI Sequence AnnotElements_01",VR::DS,VM::VM1,"AMI Annotation Element Position",false },
  {0x3105,0x0020,"AMI Sequence AnnotElements_01",VR::LT,VM::VM1,"AMI Annotation Element Text",false },
  {0x3103,0x0010,"AMI Sequence Annotations_01",VR::CS,VM::VM1,"AMI Annotation Sequence (RET)",false },
  {0x3103,0x0020,"AMI Sequence Annotations_01",VR::UI,VM::VM1,"AMI Annotation UID (RET)",false },
  {0x3103,0x0030,"AMI Sequence Annotations_01",VR::US,VM::VM1,"AMI Annotation Color (RET)",false },
  {0x3103,0x0040,"AMI Sequence Annotations_01",VR::FD,VM::VM1,"FontSize",false },
  {0x3103,0x0050,"AMI Sequence Annotations_01",VR::CS,VM::VM1,"AMI Annotation Line Style (RET)",false },
  {0x3103,0x0060,"AMI Sequence Annotations_01",VR::SQ,VM::VM1,"AMI Annotation Elements (RET)",false },
  {0x3103,0x0070,"AMI Sequence Annotations_01",VR::SH,VM::VM1,"AMI Annotation Label (RET)",false },
  {0x3103,0x0080,"AMI Sequence Annotations_01",VR::PN,VM::VM1,"AMI Annotation Creator (RET)",false },
  {0x3103,0x0090,"AMI Sequence Annotations_01",VR::PN,VM::VM1,"AMI Annotation Modifiers (RET)",false },
  {0x3103,0x00a0,"AMI Sequence Annotations_01",VR::DA,VM::VM1,"AMI Annotation Creation Date (RET)",false },
  {0x3103,0x00b0,"AMI Sequence Annotations_01",VR::TM,VM::VM1,"AMI Annotation Creation Time (RET)",false },
  {0x3103,0x00c0,"AMI Sequence Annotations_01",VR::DA,VM::VM1,"AMI Annotation Modification Dates (RET)",false },
  {0x3103,0x00d0,"AMI Sequence Annotations_01",VR::TM,VM::VM1,"AMI Annotation Modification Times (RET)",false },
  {0x3103,0x00e0,"AMI Sequence Annotations_01",VR::US,VM::VM1,"AMI Annotation Frame Number (RET)",false },
  {0x3103,0x0010,"AMI Sequence Annotations_02",VR::CS,VM::VM1,"AMI Annotation Sequence (RET)",false },
  {0x3103,0x0020,"AMI Sequence Annotations_02",VR::UI,VM::VM1,"AMI Annotation UID (RET)",false },
  {0x3103,0x0030,"AMI Sequence Annotations_02",VR::US,VM::VM1,"AMI Annotation Color (RET)",false },
  {0x3103,0x0050,"AMI Sequence Annotations_02",VR::CS,VM::VM1,"AMI Annotation Line Style (RET)",false },
  {0x3103,0x0060,"AMI Sequence Annotations_02",VR::SQ,VM::VM1,"AMI Annotation Elements (RET)",false },
  {0x3103,0x0070,"AMI Sequence Annotations_02",VR::SH,VM::VM1,"AMI Annotation Label (RET)",false },
  {0x3103,0x0080,"AMI Sequence Annotations_02",VR::PN,VM::VM1,"AMI Annotation Creator (RET)",false },
  {0x3103,0x0090,"AMI Sequence Annotations_02",VR::PN,VM::VM1,"AMI Annotation Modifiers (RET)",false },
  {0x3103,0x00a0,"AMI Sequence Annotations_02",VR::DA,VM::VM1,"AMI Annotation Creation Date (RET)",false },
  {0x3103,0x00b0,"AMI Sequence Annotations_02",VR::TM,VM::VM1,"AMI Annotation Creation Time (RET)",false },
  {0x3103,0x00c0,"AMI Sequence Annotations_02",VR::DA,VM::VM1,"AMI Annotation Modification Dates (RET)",false },
  {0x3103,0x00d0,"AMI Sequence Annotations_02",VR::TM,VM::VM1,"AMI Annotation Modification Times (RET)",false },
  {0x3103,0x00e0,"AMI Sequence Annotations_02",VR::US,VM::VM1,"AMI Annotation Frame Number (RET)",false },
  {0x3111,0x0001,"AMI StudyExtensions_01",VR::UL,VM::VM1,"AMI Last Released Annot Label (RET)",false },
  {0x0023,0x0001,"AMICAS0",VR::UI,VM::VM1,"?UID?",false },
  {0x0023,0x0008,"AMICAS0",VR::US,VM::VM1,"?",false },
  {0x0023,0x0010,"AMICAS0",VR::US,VM::VM1,"?",false },
  {0x0023,0x0016,"AMICAS0",VR::SL,VM::VM1,"?",false },
  {0x0023,0x0021,"AMICAS0",VR::US,VM::VM1,"?",false },
  {0x0023,0x0022,"AMICAS0",VR::UL,VM::VM1,"?",false },
  {0x0023,0x0040,"AMICAS0",VR::US,VM::VM1,"?",false },
  {0x0023,0x0043,"AMICAS0",VR::US,VM::VM1,"?",false },
  {0x0023,0x0045,"AMICAS0",VR::US,VM::VM1,"?",false },
  {0x0023,0x0054,"AMICAS0",VR::ST,VM::VM1,"?Performed Station AE Title?",false },
  {0x0023,0x0055,"AMICAS0",VR::LO,VM::VM1,"?Patient's Birth Date?",false },
  {0x0023,0x0056,"AMICAS0",VR::LO,VM::VM1,"?Institution Name?",false },
  {0x0023,0x0057,"AMICAS0",VR::LO,VM::VM1,"?Truncated Study Time?",false },
  {0x0023,0x0065,"AMICAS0",VR::LO,VM::VM1,"?",false },
  {0x0027,0x0010,"APEX_PRIVATE",VR::LO,VM::VM1,"Private Creator",false },
  {0x0027,0x0011,"APEX_PRIVATE",VR::DS,VM::VM1,"Bed Position",false },
  {0x0009,0x0000,"ATL HDI V1.0",VR::UN,VM::VM1,"Private",false },
  {0x0009,0x0010,"ATL HDI V1.0",VR::UN,VM::VM1,"Private",false },
  {0x0009,0x0020,"ATL HDI V1.0",VR::UN,VM::VM1,"Private",false },
  {0x0009,0x0030,"ATL HDI V1.0",VR::UN,VM::VM1,"Private",false },
  {0x0009,0x0040,"ATL HDI V1.0",VR::UN,VM::VM1,"Private",false },
  {0x0009,0x0050,"ATL HDI V1.0",VR::UN,VM::VM1,"Private",false },
  {0x0009,0x0060,"ATL HDI V1.0",VR::UN,VM::VM1,"Private",false },
  {0x0009,0x0070,"ATL HDI V1.0",VR::UN,VM::VM1,"Private",false },
  {0x0009,0x0080,"ATL HDI V1.0",VR::UN,VM::VM1,"Private",false },
  {0x0009,0x0090,"ATL HDI V1.0",VR::UN,VM::VM1,"Private",false },
  {0x0009,0x0091,"ATL HDI V1.0",VR::UN,VM::VM1,"Private",false },
  {0x0029,0x0030,"ATL HDI V1.0",VR::UN,VM::VM1,"Loop Mode",false },
  {0x0029,0x0031,"ATL HDI V1.0",VR::UN,VM::VM1,"Trigger mode",false },
  {0x0029,0x0032,"ATL HDI V1.0",VR::UN,VM::VM1,"Number of Loops",false },
  {0x0029,0x0033,"ATL HDI V1.0",VR::UN,VM::VM1,"Loop Indexes",false },
  {0x0029,0x0034,"ATL HDI V1.0",VR::UN,VM::VM1,"Loop Heart Rates",false },
  {0x0029,0x0035,"ATL HDI V1.0",VR::UN,VM::VM1,"Medications",false },
  {0x0029,0x0030,"ATL PRIVATE TAGS",VR::UL,VM::VM1,"Loop Mode",false },
  {0x0029,0x0031,"ATL PRIVATE TAGS",VR::UL,VM::VM1,"Trigger mode",false },
  {0x0029,0x0032,"ATL PRIVATE TAGS",VR::UL,VM::VM1,"Number of Loops",false },
  {0x0029,0x0033,"ATL PRIVATE TAGS",VR::DS,VM::VM1_n,"Loop Indexes",false },
  {0x0029,0x0034,"ATL PRIVATE TAGS",VR::DS,VM::VM1_n,"Loop Heart Rates",false },
  {0x0029,0x0035,"ATL PRIVATE TAGS",VR::LO,VM::VM1,"Medications",false },
  {0x0009,0x0020,"Acuson X500",VR::UN,VM::VM1,"(a)View Name",false },
  {0x0009,0x002a,"Acuson X500",VR::UN,VM::VM1,"View List",false },
  {0x0011,0x0010,"Acuson X500",VR::UN,VM::VM1,"Siemens Medical",false },
  {0x0011,0x0011,"Acuson X500",VR::UN,VM::VM1,"DIMAQ Software",false },
  {0x0011,0x0020,"Acuson X500",VR::UN,VM::VM1,"Private Data",false },
  {0x0011,0x0021,"Acuson X500",VR::UN,VM::VM1,"Private Data",false },
  {0x0013,0x0010,"Acuson X500",VR::UN,VM::VM1,"Siemens Medical",false },
  {0x0013,0x0011,"Acuson X500",VR::UN,VM::VM1,"DIMAQ Software",false },
  {0x0013,0x0020,"Acuson X500",VR::UN,VM::VM1,"Private Data",false },
  {0x0015,0x0010,"Acuson X500",VR::UN,VM::VM1,"Siemens Medical",false },
  {0x0015,0x0011,"Acuson X500",VR::UN,VM::VM1,"DIMAQ Software",false },
  {0x0015,0x0020,"Acuson X500",VR::UN,VM::VM1,"Private Data",false },
  {0x0017,0x0010,"Acuson X500",VR::UN,VM::VM1,"Siemens Medical",false },
  {0x0017,0x0011,"Acuson X500",VR::UN,VM::VM1,"DIMAQ Software",false },
  {0x0017,0x0020,"Acuson X500",VR::UN,VM::VM1,"Private Data",false },
  {0x0019,0x0020,"Acuson X500",VR::UN,VM::VM1,"Import Structured",false },
  {0x0019,0x0007,"Agfa ADC NX",VR::CS,VM::VM1,"Is Speedclass Free ",false },
  {0x0019,0x0009,"Agfa ADC NX",VR::SQ,VM::VM1,"Collimator shape sequence ",false },
  {0x0019,0x0021,"Agfa ADC NX",VR::FL,VM::VM1,"Calibration Factor ",false },
  {0x0019,0x0028,"Agfa ADC NX",VR::CS,VM::VM1,"Is IPSR enabled",false },
  {0x0019,0x00f5,"Agfa ADC NX",VR::CS,VM::VM1,"Cassette orientation",false },
  {0x0019,0x00f6,"Agfa ADC NX",VR::DS,VM::VM1,"Image plate sensitivity",false },
  {0x0019,0x00f7,"Agfa ADC NX",VR::DS,VM::VM1,"Image plate erasability",false },
  {0x0019,0x00f8,"Agfa ADC NX",VR::IS,VM::VM1,"Breast density percentage",false },
  {0x0019,0x00fa,"Agfa ADC NX",VR::IS,VM::VM1,"Exposure index",false },
  {0x0019,0x00fb,"Agfa ADC NX",VR::FL,VM::VM1,"Deviation index",false },
  {0x0019,0x00fc,"Agfa ADC NX",VR::IS,VM::VM1,"Target exposure index",false },
  {0x0019,0x00fd,"Agfa ADC NX",VR::CS,VM::VM1,"Is overexposed",false },
  {0x0019,0x00fe,"Agfa ADC NX",VR::CS,VM::VM1,"Study Priority ID",false },
  {0x0029,0x0011,"AgilityRuntime",VR::CS,VM::VM1,"?",false },
  {0x0029,0x0012,"AgilityRuntime",VR::US,VM::VM1,"?",false },
  {0x0029,0x0013,"AgilityRuntime",VR::US,VM::VM1,"?",false },
  {0x0029,0x0014,"AgilityRuntime",VR::US,VM::VM1,"?",false },
  {0x0029,0x001f,"AgilityRuntime",VR::US,VM::VM1,"?",false },
  {0x4109,0x0001,"Applicare/Centricity Radiology Web/Version 1.0",VR::SH,VM::VM1,"Mammography Laterality",false },
  {0x4109,0x0002,"Applicare/Centricity Radiology Web/Version 1.0",VR::SH,VM::VM1,"Mammography View Name",false },
  {0x4109,0x0003,"Applicare/Centricity Radiology Web/Version 1.0",VR::SH,VM::VM1,"Mammography View Modifier",false },
  {0x4111,0x0001,"Applicare/Centricity Radiology Web/Version 2.0",VR::CS,VM::VM1,"Secondary Spine Label",false },
  {0x4111,0x0002,"Applicare/Centricity Radiology Web/Version 2.0",VR::IS,VM::VM1,"Additional tags for Presentation State",false },
  {0x4101,0x0001,"Applicare/Print/Version 5.1",VR::UL,VM::VM1,"<internal",false },
  {0x4101,0x0002,"Applicare/Print/Version 5.1",VR::SQ,VM::VM1,"<internal",false },
  {0x4101,0x0003,"Applicare/Print/Version 5.1",VR::LO,VM::VM1,"<internal",false },
  {0x4101,0x0004,"Applicare/Print/Version 5.1",VR::UL,VM::VM1,"<internal",false },
  {0x4101,0x0005,"Applicare/Print/Version 5.1",VR::FD,VM::VM1,"<internal",false },
  {0x4101,0x0006,"Applicare/Print/Version 5.1",VR::US,VM::VM1,"<internal",false },
  {0x4101,0x0007,"Applicare/Print/Version 5.1",VR::US,VM::VM1,"<internal",false },
  {0x4101,0x0008,"Applicare/Print/Version 5.1",VR::US,VM::VM1,"<internal",false },
  {0x4101,0x0009,"Applicare/Print/Version 5.1",VR::US,VM::VM1,"<internal",false },
  {0x3113,0x0001,"Applicare/RadStore/Version 1.0",VR::SL,VM::VM1,"OBSOLETE",false },
  {0x3113,0x0002,"Applicare/RadStore/Version 1.0",VR::SL,VM::VM1,"Id1",false },
  {0x3113,0x0003,"Applicare/RadStore/Version 1.0",VR::SL,VM::VM1,"Id2",false },
  {0x3113,0x0004,"Applicare/RadStore/Version 1.0",VR::SL,VM::VM1,"Id3",false },
  {0x3113,0x0011,"Applicare/RadStore/Version 1.0",VR::LO,VM::VM1,"OBSOLETE",false },
  {0x3113,0x0012,"Applicare/RadStore/Version 1.0",VR::CS,VM::VM1,"State",false },
  {0x3113,0x0013,"Applicare/RadStore/Version 1.0",VR::DT,VM::VM1,"DateLastModified",false },
  {0x3113,0x0014,"Applicare/RadStore/Version 1.0",VR::DT,VM::VM1,"DateLastAccessed",false },
  {0x3113,0x0015,"Applicare/RadStore/Version 1.0",VR::CS,VM::VM1,"OBSOLETE",false },
  {0x3113,0x0016,"Applicare/RadStore/Version 1.0",VR::FD,VM::VM1,"ByteSize",false },
  {0x3113,0x0017,"Applicare/RadStore/Version 1.0",VR::LO,VM::VM1,"LibraryId",false },
  {0x3113,0x0018,"Applicare/RadStore/Version 1.0",VR::LO,VM::VM1,"Pathnames",false },
  {0x3113,0x0019,"Applicare/RadStore/Version 1.0",VR::LO,VM::VM1,"DriverPath",false },
  {0x3113,0x001a,"Applicare/RadStore/Version 1.0",VR::LO,VM::VM1,"Source",false },
  {0x3113,0x001b,"Applicare/RadStore/Version 1.0",VR::LO,VM::VM1,"Destination",false },
  {0x3113,0x001c,"Applicare/RadStore/Version 1.0",VR::SL,VM::VM1,"MediumId",false },
  {0x3113,0x001d,"Applicare/RadStore/Version 1.0",VR::LO,VM::VM1,"ArchiveId",false },
  {0x3113,0x001e,"Applicare/RadStore/Version 1.0",VR::LO,VM::VM1,"Origin",false },
  {0x3113,0x0021,"Applicare/RadStore/Version 1.0",VR::SL,VM::VM1,"Version",false },
  {0x3113,0x0022,"Applicare/RadStore/Version 1.0",VR::SL,VM::VM1,"OBSOLETE",false },
  {0x3113,0x0023,"Applicare/RadStore/Version 1.0",VR::ST,VM::VM1,"InstanceFileLocation",false },
  {0x3113,0x0031,"Applicare/RadStore/Version 1.0",VR::IS,VM::VM1,"OBSOLETE",false },
  {0x3113,0x0032,"Applicare/RadStore/Version 1.0",VR::IS,VM::VM1,"OBSOLETE",false },
  {0x3113,0x0033,"Applicare/RadStore/Version 1.0",VR::IS,VM::VM1,"OBSOLETE",false },
  {0x3113,0x0035,"Applicare/RadStore/Version 1.0",VR::LO,VM::VM1,"ImageMediumLocation",false },
  {0x3113,0x0036,"Applicare/RadStore/Version 1.0",VR::LO,VM::VM1,"ImageMediumLabel",false },
  {0x3113,0x0037,"Applicare/RadStore/Version 1.0",VR::CS,VM::VM1,"ImageMediumState",false },
  {0x3113,0x0038,"Applicare/RadStore/Version 1.0",VR::LO,VM::VM1,"SeriesMediumLocation",false },
  {0x3113,0x0039,"Applicare/RadStore/Version 1.0",VR::LO,VM::VM1,"SeriesMediumLabel",false },
  {0x3113,0x003a,"Applicare/RadStore/Version 1.0",VR::CS,VM::VM1,"SeriesMediumState",false },
  {0x3113,0x003b,"Applicare/RadStore/Version 1.0",VR::LO,VM::VM1,"StudyMediumLocation",false },
  {0x3113,0x003c,"Applicare/RadStore/Version 1.0",VR::LO,VM::VM1,"StudyMediumLabel",false },
  {0x3113,0x003d,"Applicare/RadStore/Version 1.0",VR::CS,VM::VM1,"StudyMediumState",false },
  {0x3113,0x0052,"Applicare/RadStore/Version 1.0",VR::CS,VM::VM1,"StudyState",false },
  {0x3113,0x0053,"Applicare/RadStore/Version 1.0",VR::CS,VM::VM1,"SeriesState",false },
  {0x3113,0x0055,"Applicare/RadStore/Version 1.0",VR::CS,VM::VM1,"ImageStateText",false },
  {0x3113,0x0056,"Applicare/RadStore/Version 1.0",VR::CS,VM::VM1,"SeriesStateText",false },
  {0x3113,0x0057,"Applicare/RadStore/Version 1.0",VR::CS,VM::VM1,"StudyStateText",false },
  {0x3113,0x0060,"Applicare/RadStore/Version 1.0",VR::DT,VM::VM1,"Expiration",false },
  {0x3113,0x0069,"Applicare/RadStore/Version 1.0",VR::AT,VM::VM1,"DeletedTags",false },
  {0x3109,0x0001,"Applicare/RadWorks/Version 5.0",VR::ST,VM::VM1,"<internal",false },
  {0x3109,0x0002,"Applicare/RadWorks/Version 5.0",VR::SH,VM::VM1,"NEW/SEEN Status",false },
  {0x3109,0x0003,"Applicare/RadWorks/Version 5.0",VR::CS,VM::VM1,"Delete Lock",false },
  {0x3109,0x0004,"Applicare/RadWorks/Version 5.0",VR::CS,VM::VM1,"<internal",false },
  {0x3109,0x0005,"Applicare/RadWorks/Version 5.0",VR::CS,VM::VM1,"<internal",false },
  {0x3109,0x0006,"Applicare/RadWorks/Version 5.0",VR::CS,VM::VM1,"<internal",false },
  {0x3109,0x0007,"Applicare/RadWorks/Version 5.0",VR::UL,VM::VM1,"<internal",false },
  {0x3109,0x0008,"Applicare/RadWorks/Version 5.0",VR::LO,VM::VM1,"Receive Origin",false },
  {0x3109,0x0009,"Applicare/RadWorks/Version 5.0",VR::LO,VM::VM1,"Folder",false },
  {0x3109,0x000a,"Applicare/RadWorks/Version 5.0",VR::DA,VM::VM1,"Receive Date",false },
  {0x3109,0x000b,"Applicare/RadWorks/Version 5.0",VR::TM,VM::VM1,"Receive Time",false },
  {0x3109,0x000c,"Applicare/RadWorks/Version 5.0",VR::CS,VM::VM1,"Prior",false },
  {0x3109,0x000d,"Applicare/RadWorks/Version 5.0",VR::CS,VM::VM1,"STAT Study",false },
  {0x3109,0x000e,"Applicare/RadWorks/Version 5.0",VR::CS,VM::VM1,"<internal",false },
  {0x3109,0x0010,"Applicare/RadWorks/Version 5.0",VR::CS,VM::VM1,"<internal",false },
  {0x3109,0x0011,"Applicare/RadWorks/Version 5.0",VR::LO,VM::VM1,"<internal",false },
  {0x3109,0x0012,"Applicare/RadWorks/Version 5.0",VR::LO,VM::VM1,"<internal",false },
  {0x3109,0x0013,"Applicare/RadWorks/Version 5.0",VR::DA,VM::VM1,"<internal",false },
  {0x3109,0x0014,"Applicare/RadWorks/Version 5.0",VR::TM,VM::VM1,"<internal",false },
  {0x3109,0x0019,"Applicare/RadWorks/Version 5.0",VR::LO,VM::VM1,"<internal",false },
  {0x3109,0x0020,"Applicare/RadWorks/Version 5.0",VR::UI,VM::VM1,"<internal",false },
  {0x3109,0x0021,"Applicare/RadWorks/Version 5.0",VR::CS,VM::VM1,"<internal",false },
  {0x3109,0x0022,"Applicare/RadWorks/Version 5.0",VR::AE,VM::VM1,"<internal",false },
  {0x3109,0x0023,"Applicare/RadWorks/Version 5.0",VR::US,VM::VM1,"<internal",false },
  {0x3109,0x0024,"Applicare/RadWorks/Version 5.0",VR::LO,VM::VM1,"<internal",false },
  {0x3109,0x0025,"Applicare/RadWorks/Version 5.0",VR::LO,VM::VM1,"<internal",false },
  {0x3109,0x0026,"Applicare/RadWorks/Version 5.0",VR::UI,VM::VM1,"<internal",false },
  {0x3109,0x0027,"Applicare/RadWorks/Version 5.0",VR::SQ,VM::VM1,"<internal",false },
  {0x3109,0x0028,"Applicare/RadWorks/Version 5.0",VR::LO,VM::VM1,"<internal",false },
  {0x3109,0x0029,"Applicare/RadWorks/Version 5.0",VR::LO,VM::VM1,"<internal",false },
  {0x3109,0x002a,"Applicare/RadWorks/Version 5.0",VR::LO,VM::VM1,"Send Flag",false },
  {0x3109,0x002b,"Applicare/RadWorks/Version 5.0",VR::LO,VM::VM1,"Print Flag",false },
  {0x3109,0x002c,"Applicare/RadWorks/Version 5.0",VR::LO,VM::VM1,"Archive Flag",false },
  {0x3109,0x0030,"Applicare/RadWorks/Version 5.0",VR::LO,VM::VM1,"<internal",false },
  {0x3109,0x0031,"Applicare/RadWorks/Version 5.0",VR::CS,VM::VM1,"<internal",false },
  {0x3109,0x0032,"Applicare/RadWorks/Version 5.0",VR::CS,VM::VM1,"<internal",false },
  {0x3109,0x0033,"Applicare/RadWorks/Version 5.0",VR::CS,VM::VM1,"Request Storage Commitment",false },
  {0x3109,0x0034,"Applicare/RadWorks/Version 5.0",VR::CS,VM::VM1,"Requested Compression",false },
  {0x3109,0x0035,"Applicare/RadWorks/Version 5.0",VR::SQ,VM::VM1,"<internal",false },
  {0x3109,0x0036,"Applicare/RadWorks/Version 5.0",VR::LO,VM::VM1,"<internal",false },
  {0x3109,0x0037,"Applicare/RadWorks/Version 5.0",VR::UI,VM::VM1,"<internal",false },
  {0x3109,0x0038,"Applicare/RadWorks/Version 5.0",VR::SH,VM::VM1,"<internal",false },
  {0x3109,0x0039,"Applicare/RadWorks/Version 5.0",VR::SH,VM::VM1,"<internal",false },
  {0x3109,0x0040,"Applicare/RadWorks/Version 5.0",VR::IS,VM::VM1,"<internal",false },
  {0x3109,0x0041,"Applicare/RadWorks/Version 5.0",VR::CS,VM::VM1,"<internal",false },
  {0x3109,0x0042,"Applicare/RadWorks/Version 5.0",VR::CS,VM::VM1,"<internal",false },
  {0x3109,0x0043,"Applicare/RadWorks/Version 5.0",VR::CS,VM::VM1,"Archive Status",false },
  {0x3109,0x00ee,"Applicare/RadWorks/Version 5.0",VR::UI,VM::VM1,"<internal",false },
  {0x3109,0x00ef,"Applicare/RadWorks/Version 5.0",VR::CS,VM::VM1,"<internal",false },
  {0x4103,0x0001,"Applicare/RadWorks/Version 6.0",VR::AT,VM::VM1,"Internal: non-existent tags",false },
  {0x4103,0x0002,"Applicare/RadWorks/Version 6.0",VR::UI,VM::VM1,"Internal: non-existent objects",false },
  {0x4105,0x0001,"Applicare/RadWorks/Version 6.0",VR::CS,VM::VM1,"Annotation Type",false },
  {0x4105,0x0002,"Applicare/RadWorks/Version 6.0",VR::DS,VM::VM1,"Annotation Value",false },
  {0x4105,0x0003,"Applicare/RadWorks/Version 6.0",VR::UI,VM::VM1,"Cutline Image UID",false },
  {0x4105,0x0004,"Applicare/RadWorks/Version 6.0",VR::UI,VM::VM1,"Cutline Set UID",false },
  {0x4105,0x0005,"Applicare/RadWorks/Version 6.0",VR::US,VM::VM1,"Annotation Color",false },
  {0x4105,0x0006,"Applicare/RadWorks/Version 6.0",VR::CS,VM::VM1,"Annotation Line Style",false },
  {0x4105,0x0007,"Applicare/RadWorks/Version 6.0",VR::SH,VM::VM1,"Annotation Label",false },
  {0x4105,0x0008,"Applicare/RadWorks/Version 6.0",VR::PN,VM::VM1,"Annotation Creator",false },
  {0x4105,0x0009,"Applicare/RadWorks/Version 6.0",VR::DA,VM::VM1,"Annotation Creation Date",false },
  {0x4105,0x000a,"Applicare/RadWorks/Version 6.0",VR::TM,VM::VM1,"Annotation Creation Time",false },
  {0x4105,0x000b,"Applicare/RadWorks/Version 6.0",VR::SQ,VM::VM1,"Annotation Modification Sequence",false },
  {0x4105,0x000c,"Applicare/RadWorks/Version 6.0",VR::PN,VM::VM1,"Annotation Modifier",false },
  {0x4105,0x000d,"Applicare/RadWorks/Version 6.0",VR::DA,VM::VM1,"Annotation Modification Date",false },
  {0x4105,0x000e,"Applicare/RadWorks/Version 6.0",VR::TM,VM::VM1,"Annotation Modification Time",false },
  {0x4105,0x000f,"Applicare/RadWorks/Version 6.0",VR::LO,VM::VM1,"Palette Color LUT Name",false },
  {0x4105,0x0010,"Applicare/RadWorks/Version 6.0",VR::US,VM::VM1,"Annotation Number",false },
  {0x4107,0x0001,"Applicare/RadWorks/Version 6.0",VR::SQ,VM::VM1,"Requested Palette Color LUT (GSPS)",false },
  {0x3109,0x0011,"Applicare/RadWorks/Version 6.0/Summary",VR::ST,VM::VM1,"Receive Origin Description",false },
  {0x3109,0x0012,"Applicare/RadWorks/Version 6.0/Summary",VR::ST,VM::VM1,"Receive Origin Description",false },
  {0x3113,0x0001,"Applicare/Workflow/Version 1.0",VR::CS,VM::VM1,"<internal",false },
  {0x3113,0x0010,"Applicare/Workflow/Version 1.0",VR::SH,VM::VM1,"<internal",false },
  {0x3113,0x0011,"Applicare/Workflow/Version 1.0",VR::SH,VM::VM1,"<internal",false },
  {0x3113,0x0012,"Applicare/Workflow/Version 1.0",VR::LO,VM::VM1,"<internal",false },
  {0x3113,0x0015,"Applicare/Workflow/Version 1.0",VR::SH,VM::VM1,"<internal",false },
  {0x3113,0x0016,"Applicare/Workflow/Version 1.0",VR::SH,VM::VM1,"<internal",false },
  {0x3113,0x0017,"Applicare/Workflow/Version 1.0",VR::LO,VM::VM1,"<internal",false },
  {0x3113,0x0020,"Applicare/Workflow/Version 1.0",VR::SH,VM::VM1,"<internal",false },
  {0x3113,0x0021,"Applicare/Workflow/Version 1.0",VR::SH,VM::VM1,"<internal",false },
  {0x3113,0x0022,"Applicare/Workflow/Version 1.0",VR::LO,VM::VM1,"<internal",false },
  {0x3113,0x0025,"Applicare/Workflow/Version 1.0",VR::SH,VM::VM1,"<internal",false },
  {0x3113,0x0026,"Applicare/Workflow/Version 1.0",VR::SH,VM::VM1,"<internal",false },
  {0x3113,0x0027,"Applicare/Workflow/Version 1.0",VR::LO,VM::VM1,"<internal",false },
  {0x3113,0x0030,"Applicare/Workflow/Version 1.0",VR::UI,VM::VM1,"<internal",false },
  {0x3113,0x0031,"Applicare/Workflow/Version 1.0",VR::UI,VM::VM1,"<internal",false },
  {0x3113,0x00e0,"Applicare/Workflow/Version 1.0",VR::CS,VM::VM1,"<internal",false },
  {0x3113,0x00e1,"Applicare/Workflow/Version 1.0",VR::CS,VM::VM1,"<internal",false },
  {0x3113,0x00e2,"Applicare/Workflow/Version 1.0",VR::CS,VM::VM1,"<internal",false },
  {0x0021,0x0000,"BRIT Systems, Inc.",VR::SQ,VM::VM1,"Person Information Sequence",false },
  {0x0021,0x0001,"BRIT Systems, Inc.",VR::LO,VM::VM1,"Person ID",false },
  {0x0021,0x0002,"BRIT Systems, Inc.",VR::PN,VM::VM1,"Person Name",false },
  {0x0021,0x0003,"BRIT Systems, Inc.",VR::LO,VM::VM1,"Person Role",false },
  {0x0021,0x0004,"BRIT Systems, Inc.",VR::SH,VM::VM1,"Person Home Phone",false },
  {0x0021,0x0005,"BRIT Systems, Inc.",VR::SH,VM::VM1,"Person Work Phone",false },
  {0x0021,0x0006,"BRIT Systems, Inc.",VR::SH,VM::VM1,"Person Cell Phone",false },
  {0x0021,0x0007,"BRIT Systems, Inc.",VR::SH,VM::VM1,"Person Pager Phone",false },
  {0x0021,0x0008,"BRIT Systems, Inc.",VR::SH,VM::VM1,"Person Fax Phone",false },
  {0x0021,0x0009,"BRIT Systems, Inc.",VR::LO,VM::VM1,"Person EMail",false },
  {0x0021,0x000a,"BRIT Systems, Inc.",VR::ST,VM::VM1,"Person Address",false },
  {0x0021,0x000b,"BRIT Systems, Inc.",VR::LO,VM::VM1,"Person Password",false },
  {0x0021,0x000c,"BRIT Systems, Inc.",VR::SH,VM::VM1,"Person Emergency Phone",false },
  {0x0021,0x000d,"BRIT Systems, Inc.",VR::LO,VM::VM1,"Physician ID",false },
  {0x0021,0x0011,"BRIT Systems, Inc.",VR::LO,VM::VM1,"Original Patient ID",false },
  {0x0021,0x0012,"BRIT Systems, Inc.",VR::UI,VM::VM1,"Original Study Instance UID",false },
  {0x0021,0x0013,"BRIT Systems, Inc.",VR::UI,VM::VM1,"Original Series Instance UID",false },
  {0x0021,0x0014,"BRIT Systems, Inc.",VR::LO,VM::VM1,"Master Accession Number",false },
  {0x0021,0x0015,"BRIT Systems, Inc.",VR::LO,VM::VM1,"Order Category",false },
  {0x0021,0x0016,"BRIT Systems, Inc.",VR::LO,VM::VM1,"Patient ICN",false },
  {0x0021,0x0017,"BRIT Systems, Inc.",VR::LO,VM::VM1,"Patient DFS",false },
  {0x0021,0x0018,"BRIT Systems, Inc.",VR::LO,VM::VM1,"Patient Class",false },
  {0x0021,0x0019,"BRIT Systems, Inc.",VR::LO,VM::VM1,"Patient Type",false },
  {0x0021,0x001f,"BRIT Systems, Inc.",VR::LT,VM::VM1,"Generic String",false },
  {0x0021,0x0020,"BRIT Systems, Inc.",VR::LO,VM::VM1,"QC Study Assigned By",false },
  {0x0021,0x0021,"BRIT Systems, Inc.",VR::LO,VM::VM1,"QC Study Split By",false },
  {0x0021,0x0022,"BRIT Systems, Inc.",VR::LO,VM::VM1,"QC Study Moved By",false },
  {0x0021,0x0023,"BRIT Systems, Inc.",VR::LO,VM::VM1,"QC Study Edited By",false },
  {0x0021,0x0024,"BRIT Systems, Inc.",VR::LO,VM::VM1,"QC Series Split By",false },
  {0x0021,0x0025,"BRIT Systems, Inc.",VR::LO,VM::VM1,"QC Series Moved By",false },
  {0x0021,0x0026,"BRIT Systems, Inc.",VR::LO,VM::VM1,"QC Series Edited By",false },
  {0x0021,0x0027,"BRIT Systems, Inc.",VR::LO,VM::VM1,"QC Image Moved By",false },
  {0x0021,0x0028,"BRIT Systems, Inc.",VR::LO,VM::VM1,"QC Image Edited By",false },
  {0x0021,0x0030,"BRIT Systems, Inc.",VR::LO,VM::VM1,"QC Done Time",false },
  {0x0021,0x0031,"BRIT Systems, Inc.",VR::LO,VM::VM1,"QC Last Modification Time",false },
  {0x0021,0x0032,"BRIT Systems, Inc.",VR::LO,VM::VM1,"QC Image Accepted By",false },
  {0x0021,0x0033,"BRIT Systems, Inc.",VR::LO,VM::VM1,"QC Image Rejected By",false },
  {0x0021,0x0034,"BRIT Systems, Inc.",VR::DA,VM::VM1,"QC Done Date",false },
  {0x0021,0x0050,"BRIT Systems, Inc.",VR::LO,VM::VM1,"QC Deletion Requested",false },
  {0x0021,0x0090,"BRIT Systems, Inc.",VR::AE,VM::VM1,"Original Sender AE Title",false },
  {0x0021,0x0091,"BRIT Systems, Inc.",VR::LO,VM::VM1,"Software Title",false },
  {0x0021,0x0092,"BRIT Systems, Inc.",VR::SH,VM::VM1,"Software Version",false },
  {0x0021,0x0093,"BRIT Systems, Inc.",VR::LO,VM::VM1,"Serial Number",false },
  {0x0021,0x00a0,"BRIT Systems, Inc.",VR::SQ,VM::VM1,"Object Action Sequence",false },
  {0x0021,0x00a1,"BRIT Systems, Inc.",VR::ST,VM::VM1,"Object Action",false },
  {0x0021,0x00a2,"BRIT Systems, Inc.",VR::DA,VM::VM1,"Object Action Date",false },
  {0x0021,0x00a3,"BRIT Systems, Inc.",VR::TM,VM::VM1,"Object Action Time",false },
  {0x0021,0x00a5,"BRIT Systems, Inc.",VR::AE,VM::VM1,"Local AE Title",false },
  {0x0021,0x00a6,"BRIT Systems, Inc.",VR::SH,VM::VM1,"Local IP Address",false },
  {0x0021,0x00a7,"BRIT Systems, Inc.",VR::AE,VM::VM1,"Remote AE Title",false },
  {0x0021,0x00a8,"BRIT Systems, Inc.",VR::SH,VM::VM1,"Remote IP Address",false },
  {0x0009,0x0000,"BioPri",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0001,"BioPri",VR::UN,VM::VM1,"?",false },
  {0x0009,0x0002,"BioPri",VR::UN,VM::VM1,"?",false },
  {0x0009,0x0003,"BioPri",VR::LO,VM::VM1_n,"?",false },
  {0x0009,0x0004,"BioPri",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0005,"BioPri",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0007,"BioPri",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0008,"BioPri",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0009,"BioPri",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0010,"BioPri",VR::UN,VM::VM1,"?",false },
  {0x3411,0x0001,"BrainLAB_BeamProfile",VR::SQ,VM::VM1,"Beam Profile Sequence",false },
  {0x3411,0x0002,"BrainLAB_BeamProfile",VR::IS,VM::VM1,"Beam Profile Number",false },
  {0x3411,0x0003,"BrainLAB_BeamProfile",VR::SQ,VM::VM1,"Beam Parameter Sequence",false },
  {0x3411,0x0004,"BrainLAB_BeamProfile",VR::UT,VM::VM1,"Parameter Description",false },
  {0x3411,0x0005,"BrainLAB_BeamProfile",VR::OB,VM::VM1,"Parameter Data",false },
  {0x3411,0x0006,"BrainLAB_BeamProfile",VR::IS,VM::VM1,"Referenced Beam Profile Number",false },
  {0x0009,0x0001,"BrainLAB_Conversion",VR::LO,VM::VM1,"Export Platform Name",false },
  {0x0009,0x0002,"BrainLAB_Conversion",VR::OB,VM::VM1,"Export Platform Data",false },
  {0x3273,0x0000,"BrainLAB_PatientSetup",VR::DS,VM::VM3,"Isocenter Position",false },
  {0x3273,0x0001,"BrainLAB_PatientSetup",VR::CS,VM::VM1,"Patient Position",false },
  {0x0029,0x0010,"CAMTRONICS",VR::LT,VM::VM1,"Commentline",false },
  {0x0029,0x0020,"CAMTRONICS",VR::DS,VM::VM1,"Edge Enhancement Coefficient",false },
  {0x0029,0x0050,"CAMTRONICS",VR::LT,VM::VM1,"Scene Text",false },
  {0x0029,0x0060,"CAMTRONICS",VR::LT,VM::VM1,"Image Text",false },
  {0x0029,0x0070,"CAMTRONICS",VR::IS,VM::VM1,"Pixel Shift Horizontal",false },
  {0x0029,0x0080,"CAMTRONICS",VR::IS,VM::VM1,"Pixel Shift Vertical",false },
  {0x0029,0x0090,"CAMTRONICS",VR::IS,VM::VM1,"?",false },
  {0x0029,0x0010,"CAMTRONICS IP",VR::LT,VM::VM1,"?",false },
  {0x0029,0x0020,"CAMTRONICS IP",VR::UN,VM::VM1,"?",false },
  {0x0029,0x0030,"CAMTRONICS IP",VR::UN,VM::VM1,"?",false },
  {0x0029,0x0040,"CAMTRONICS IP",VR::UN,VM::VM1,"?",false },
  {0x0009,0x0000,"CARDIO-D.R. 1.0",VR::UL,VM::VM1,"File Location",false },
  {0x0009,0x0001,"CARDIO-D.R. 1.0",VR::UL,VM::VM1,"File Size",false },
  {0x0009,0x0040,"CARDIO-D.R. 1.0",VR::SQ,VM::VM1,"Alternate Image Sequence",false },
  {0x0019,0x0000,"CARDIO-D.R. 1.0",VR::CS,VM::VM1_2,"Image Blanking Shape",false },
  {0x0019,0x0002,"CARDIO-D.R. 1.0",VR::IS,VM::VM1,"Image Blanking Left Vertical Edge",false },
  {0x0019,0x0004,"CARDIO-D.R. 1.0",VR::IS,VM::VM1,"Image Blanking Right Vertical Edge",false },
  {0x0019,0x0006,"CARDIO-D.R. 1.0",VR::IS,VM::VM1,"Image Blanking Upper Horizontal Edge",false },
  {0x0019,0x0008,"CARDIO-D.R. 1.0",VR::IS,VM::VM1,"Image Blanking Lower Horizontal Edge",false },
  {0x0019,0x0010,"CARDIO-D.R. 1.0",VR::IS,VM::VM2,"Center Of Circular Image Blanking",false },
  {0x0019,0x0012,"CARDIO-D.R. 1.0",VR::IS,VM::VM1,"Radius Of Circular Image Blanking",false },
  {0x0019,0x0030,"CARDIO-D.R. 1.0",VR::UL,VM::VM1,"Maximum Frame Size",false },
  {0x0021,0x0013,"CARDIO-D.R. 1.0",VR::IS,VM::VM1,"Image Sequence Number",false },
  {0x0029,0x0000,"CARDIO-D.R. 1.0",VR::SQ,VM::VM1,"Standard Edge Enhancement Sequence",false },
  {0x0029,0x0001,"CARDIO-D.R. 1.0",VR::US,VM::VM2,"Convolution Kernel Size",false },
  {0x0029,0x0002,"CARDIO-D.R. 1.0",VR::US,VM::VM1_n,"Convolution Kernel Coefficients",false },
  {0x0029,0x0003,"CARDIO-D.R. 1.0",VR::FL,VM::VM1,"Edge Enhancement Gain",false },
  {0x0009,0x0002,"CARDIO-SMS 1.0",VR::OB,VM::VM1,"Private Data",false },
  {0x0009,0x0003,"CARDIO-SMS 1.0",VR::OB,VM::VM1,"Private Data",false },
  {0x0009,0x0004,"CARDIO-SMS 1.0",VR::OB,VM::VM1,"Private Data",false },
  {0x0009,0x0005,"CARDIO-SMS 1.0",VR::OB,VM::VM1,"Private Data",false },
  {0x0009,0x0006,"CARDIO-SMS 1.0",VR::OB,VM::VM1,"Private Data",false },
  {0x0009,0x0008,"CARDIO-SMS 1.0",VR::OB,VM::VM1,"ACOM flags",false },
  {0x0009,0x000a,"CARDIO-SMS 1.0",VR::UN,VM::VM1,"Private Data",false },
  {0x0009,0x000b,"CARDIO-SMS 1.0",VR::UN,VM::VM1,"Private Data",false },
  {0x0009,0x000c,"CARDIO-SMS 1.0",VR::UN,VM::VM1,"Private Data",false },
  {0x0029,0x0080,"CARDIO-SMS 1.0",VR::SQ,VM::VM1,"Private Data",false },
  {0x0031,0x0001,"CARDIO-SMS 1.0",VR::UN,VM::VM1,"Private Data stream",false },
  {0x0019,0x0000,"CEMAX-ICON",VR::LT,VM::VM1,"CR Processing Parameters",false },
  {0x0019,0x0010,"CEMAX-ICON",VR::LO,VM::VM1,"CR Exposure Menu Code",false },
  {0x0019,0x0020,"CEMAX-ICON",VR::LO,VM::VM1,"CR Exposure Menu String",false },
  {0x0019,0x0030,"CEMAX-ICON",VR::LO,VM::VM1,"CR EDR Mode",false },
  {0x0019,0x0040,"CEMAX-ICON",VR::LO,VM::VM1,"CR Latitude",false },
  {0x0019,0x0050,"CEMAX-ICON",VR::LO,VM::VM1,"CR Group Number",false },
  {0x0019,0x0060,"CEMAX-ICON",VR::LO,VM::VM1,"CR Equipment ID",false },
  {0x0019,0x0070,"CEMAX-ICON",VR::LO,VM::VM1,"CR Image Serial Number",false },
  {0x0019,0x0080,"CEMAX-ICON",VR::LO,VM::VM1,"CR Bar Code Number",false },
  {0x0019,0x0090,"CEMAX-ICON",VR::LO,VM::VM1,"CR Film Output Exposure",false },
  {0x0029,0x0000,"CEMAX-ICON",VR::LO,VM::VM1,"Key Image",false },
  {0x0029,0x0010,"CEMAX-ICON",VR::LT,VM::VM1_n,"Note",false },
  {0x0029,0x0020,"CEMAX-ICON",VR::SQ,VM::VM1,"Annotation Sequence",false },
  {0x0029,0x0030,"CEMAX-ICON",VR::IS,VM::VM1,"X Start Position",false },
  {0x0029,0x0040,"CEMAX-ICON",VR::IS,VM::VM1,"Y Start Position",false },
  {0x0029,0x0050,"CEMAX-ICON",VR::IS,VM::VM1,"X End Position",false },
  {0x0029,0x0060,"CEMAX-ICON",VR::IS,VM::VM1,"Y End Position",false },
  {0x0029,0x0070,"CEMAX-ICON",VR::LO,VM::VM1,"Annotation Text",false },
  {0x0029,0x0080,"CEMAX-ICON",VR::IS,VM::VM1,"Which Marker",false },
  {0x0029,0x0090,"CEMAX-ICON",VR::SQ,VM::VM1,"Annotation Sequence",false },
  {0x0029,0x0091,"CEMAX-ICON",VR::IS,VM::VM1_n,"X Coordinates",false },
  {0x0029,0x0092,"CEMAX-ICON",VR::IS,VM::VM1_n,"Y Coordinates",false },
  {0x0029,0x0093,"CEMAX-ICON",VR::IS,VM::VM1,"Annotation Type",false },
  {0x0029,0x0094,"CEMAX-ICON",VR::IS,VM::VM1,"Number of points",false },
  {0x0029,0x0095,"CEMAX-ICON",VR::IS,VM::VM1,"Fill Type",false },
  {0x3033,0x0000,"CEMAX-ICON",VR::CS,VM::VM1,"Study Priority Status",false },
  {0x0019,0x0010,"Canon Inc.",VR::OB,VM::VM1,"Canon Protocol Information",false },
  {0x0019,0x0013,"Canon Inc.",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0015,"Canon Inc.",VR::DS,VM::VM2,"Canon Imager Pixel Spacing",false },
  {0x0019,0x0016,"Canon Inc.",VR::LT,VM::VM1,"Canon Study Instance UID",false },
  {0x0019,0x0017,"Canon Inc.",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0018,"Canon Inc.",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0019,"Canon Inc.",VR::OB,VM::VM1,"?",false },
  {0x0019,0x001a,"Canon Inc.",VR::OB,VM::VM1,"?",false },
  {0x0019,0x001b,"Canon Inc.",VR::LO,VM::VM1,"Canon Patient Name",false },
  {0x0019,0x001c,"Canon Inc.",VR::IS,VM::VM1,"?",false },
  {0x0019,0x001e,"Canon Inc.",VR::IS,VM::VM1,"?",false },
  {0x0019,0x001f,"Canon Inc.",VR::OB,VM::VM1,"Canon Protocol Data",false },
  {0x0019,0x0021,"Canon Inc.",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0060,"Canon Inc.",VR::US,VM::VM1,"Performed number of series",false },
  {0x0019,0x0070,"Canon Inc.",VR::US,VM::VM1,"Performed number of images",false },
  {0x0019,0x007f,"Canon Inc.",VR::OB,VM::VM1,"Canon Internal Data 111",false },
  {0x0019,0x0000,"DIDI TO PCR 1.1",VR::LT,VM::VM1,"Post Mode String",false },
  {0x0019,0x0001,"DIDI TO PCR 1.1",VR::LT,VM::VM1,"Post Data",false },
  {0x0019,0x0010,"DIDI TO PCR 1.1",VR::LT,VM::VM1,"Image Header",false },
  {0x0019,0x0022,"DIDI TO PCR 1.1",VR::LO,VM::VM1,"Route AET",false },
  {0x0019,0x0023,"DIDI TO PCR 1.1",VR::DS,VM::VM1,"PCR Print Scale",false },
  {0x0019,0x0024,"DIDI TO PCR 1.1",VR::ST,VM::VM1,"PCR Print Job End",false },
  {0x0019,0x0025,"DIDI TO PCR 1.1",VR::IS,VM::VM1,"PCR No Film Copies",false },
  {0x0019,0x0026,"DIDI TO PCR 1.1",VR::IS,VM::VM1,"PCR Film Layout Position",false },
  {0x0019,0x0027,"DIDI TO PCR 1.1",VR::ST,VM::VM1,"PCR Print Report Name",false },
  {0x0019,0x0070,"DIDI TO PCR 1.1",VR::ST,VM::VM1,"RAD Protocol Printer",false },
  {0x0019,0x0071,"DIDI TO PCR 1.1",VR::ST,VM::VM1,"RAD Protocol Medium",false },
  {0x0019,0x0080,"DIDI TO PCR 1.1",VR::LO,VM::VM1,"Original Filename",false },
  {0x0019,0x0081,"DIDI TO PCR 1.1",VR::SH,VM::VM1,"Filter Type",false },
  {0x0019,0x0082,"DIDI TO PCR 1.1",VR::LT,VM::VM1,"Stitching",false },
  {0x0019,0x0089,"DIDI TO PCR 1.1",VR::IS,VM::VM1,"Exposure Index",false },
  {0x0019,0x008a,"DIDI TO PCR 1.1",VR::IS,VM::VM1,"Collimator X",false },
  {0x0019,0x008b,"DIDI TO PCR 1.1",VR::IS,VM::VM1,"Collimator Y",false },
  {0x0019,0x008c,"DIDI TO PCR 1.1",VR::LO,VM::VM1,"Print Marker",false },
  {0x0019,0x008d,"DIDI TO PCR 1.1",VR::LO,VM::VM1,"RGDV Name",false },
  {0x0019,0x008e,"DIDI TO PCR 1.1",VR::LO,VM::VM1,"Acqd Sensitivity",false },
  {0x0019,0x008f,"DIDI TO PCR 1.1",VR::LO,VM::VM1,"Processing Category",false },
  {0x0019,0x0090,"DIDI TO PCR 1.1",VR::LO,VM::VM1,"Unprocessed Flag",false },
  {0x0019,0x0091,"DIDI TO PCR 1.1",VR::DS,VM::VM2,"Key Values",false },
  {0x0019,0x0092,"DIDI TO PCR 1.1",VR::LO,VM::VM1,"Destination Postprocessing Function",false },
  {0x0019,0x00a0,"DIDI TO PCR 1.1",VR::LO,VM::VM1,"Version",false },
  {0x0019,0x00a1,"DIDI TO PCR 1.1",VR::LO,VM::VM1,"Ranging Mode",false },
  {0x0019,0x00a2,"DIDI TO PCR 1.1",VR::DS,VM::VM1,"Abdomen Brightness",false },
  {0x0019,0x00a3,"DIDI TO PCR 1.1",VR::DS,VM::VM1,"Fixed Brightness",false },
  {0x0019,0x00a4,"DIDI TO PCR 1.1",VR::DS,VM::VM1,"Detail Contrast",false },
  {0x0019,0x00a5,"DIDI TO PCR 1.1",VR::DA,VM::VM1,"Contrast Balance",false },
  {0x0019,0x00a6,"DIDI TO PCR 1.1",VR::DS,VM::VM1,"Structure Boost",false },
  {0x0019,0x00a7,"DIDI TO PCR 1.1",VR::DS,VM::VM1,"Structure Preference",false },
  {0x0019,0x00a8,"DIDI TO PCR 1.1",VR::DS,VM::VM1,"Noise Robustness",false },
  {0x0019,0x00a9,"DIDI TO PCR 1.1",VR::DS,VM::VM1,"Noise Dose Limit",false },
  {0x0019,0x00aa,"DIDI TO PCR 1.1",VR::DS,VM::VM1,"Noise Dose Step",false },
  {0x0019,0x00ab,"DIDI TO PCR 1.1",VR::DS,VM::VM1,"Noise Frequency Limit",false },
  {0x0019,0x00ac,"DIDI TO PCR 1.1",VR::DS,VM::VM1,"Weak Contrast Limit",false },
  {0x0019,0x00ad,"DIDI TO PCR 1.1",VR::DS,VM::VM1,"Strong Contrast Limit",false },
  {0x0019,0x00ae,"DIDI TO PCR 1.1",VR::DS,VM::VM1,"Structure Boost Offset",false },
  {0x0019,0x00af,"DIDI TO PCR 1.1",VR::LO,VM::VM1,"Smooth Gain",false },
  {0x0019,0x00b0,"DIDI TO PCR 1.1",VR::LO,VM::VM1,"Measure Field 1",false },
  {0x0019,0x00b1,"DIDI TO PCR 1.1",VR::LO,VM::VM1,"Measure Field 2",false },
  {0x0019,0x00b2,"DIDI TO PCR 1.1",VR::IS,VM::VM1,"Key Percentile 1",false },
  {0x0019,0x00b3,"DIDI TO PCR 1.1",VR::IS,VM::VM1,"Key Percentile 2",false },
  {0x0019,0x00b4,"DIDI TO PCR 1.1",VR::IS,VM::VM1,"Density LUT",false },
  {0x0019,0x00b5,"DIDI TO PCR 1.1",VR::DS,VM::VM1,"Brightness",false },
  {0x0019,0x00b6,"DIDI TO PCR 1.1",VR::DS,VM::VM1,"Gamma",false },
  {0x0089,0x0010,"DIDI TO PCR 1.1",VR::SQ,VM::VM1,"Stamp Image Sequence",false },
  {0x0029,0x0031,"DIGISCAN IMAGE",VR::US,VM::VM1_n,"?",false },
  {0x0029,0x0032,"DIGISCAN IMAGE",VR::US,VM::VM1_n,"?",false },
  {0x0029,0x0033,"DIGISCAN IMAGE",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0034,"DIGISCAN IMAGE",VR::LO,VM::VM1,"?",false },
  {0x0015,0x0001,"DLX_EXAMS_01",VR::DS,VM::VM1,"Stenos_calibr_ratio;",false },
  {0x0015,0x0002,"DLX_EXAMS_01",VR::DS,VM::VM1,"Stenos_magnification;",false },
  {0x0015,0x0003,"DLX_EXAMS_01",VR::DS,VM::VM1,"Cardiac_calibr_ratio;",false },
  {0x0011,0x0001,"DLX_PATNT_01",VR::LT,VM::VM1,"patient DOB",false },
  {0x0019,0x0001,"DLX_SERIE_01",VR::DS,VM::VM1,"Angle Value 1",false },
  {0x0019,0x0002,"DLX_SERIE_01",VR::DS,VM::VM1,"Angle Value 2",false },
  {0x0019,0x0003,"DLX_SERIE_01",VR::DS,VM::VM1,"Angle Value 3",false },
  {0x0019,0x0004,"DLX_SERIE_01",VR::CS,VM::VM1,"Angle Label 1",false },
  {0x0019,0x0005,"DLX_SERIE_01",VR::CS,VM::VM1,"Angle Label 2",false },
  {0x0019,0x0006,"DLX_SERIE_01",VR::CS,VM::VM1,"Angle Label 3",false },
  {0x0019,0x0007,"DLX_SERIE_01",VR::ST,VM::VM1,"Adx Procedure Name",false },
  {0x0019,0x0008,"DLX_SERIE_01",VR::ST,VM::VM1,"Adx Exam Name",false },
  {0x0019,0x0009,"DLX_SERIE_01",VR::SH,VM::VM1,"Adx Patient Size",false },
  {0x0019,0x000a,"DLX_SERIE_01",VR::IS,VM::VM1,"Adx Record View",false },
  {0x0019,0x0010,"DLX_SERIE_01",VR::DS,VM::VM1,"Adx Injector Delay",false },
  {0x0019,0x0011,"DLX_SERIE_01",VR::CS,VM::VM1,"Adx Auto Inject",false },
  {0x0019,0x0014,"DLX_SERIE_01",VR::IS,VM::VM1,"Adx Acq Mode",false },
  {0x0019,0x0015,"DLX_SERIE_01",VR::CS,VM::VM1,"Adx Camera Rotation Enable",false },
  {0x0019,0x0016,"DLX_SERIE_01",VR::CS,VM::VM1,"Adx Reverse Sweep",false },
  {0x0019,0x0017,"DLX_SERIE_01",VR::IS,VM::VM1,"User Spatial Filter Strength",false },
  {0x0019,0x0018,"DLX_SERIE_01",VR::IS,VM::VM1,"User Zoom Factor",false },
  {0x0019,0x0019,"DLX_SERIE_01",VR::IS,VM::VM1,"X Zoom",false },
  {0x0019,0x001a,"DLX_SERIE_01",VR::IS,VM::VM1,"Y Zoom",false },
  {0x0019,0x001b,"DLX_SERIE_01",VR::DS,VM::VM1,"Adx Focus",false },
  {0x0019,0x001c,"DLX_SERIE_01",VR::CS,VM::VM1,"Adx Dose",false },
  {0x0019,0x001d,"DLX_SERIE_01",VR::IS,VM::VM1,"Side Mark",false },
  {0x0019,0x001e,"DLX_SERIE_01",VR::IS,VM::VM1,"Percent Landscape",false },
  {0x0019,0x001f,"DLX_SERIE_01",VR::DS,VM::VM1,"Adx Exposure Duration",false },
  {0x0019,0x0020,"DLX_SERIE_01",VR::SH,VM::VM1,"Ip Address",false },
  {0x0019,0x0021,"DLX_SERIE_01",VR::DS,VM::VM1,"Table Vertical Position",false },
  {0x0019,0x0022,"DLX_SERIE_01",VR::DS,VM::VM1,"Table Longitudinal Position",false },
  {0x0019,0x0023,"DLX_SERIE_01",VR::DS,VM::VM1,"Table Lateral Position",false },
  {0x0019,0x0024,"DLX_SERIE_01",VR::DS,VM::VM1,"Lambda cm Pincushion Distortion",false },
  {0x0019,0x0025,"DLX_SERIE_01",VR::DS,VM::VM1,"Slope LV Regression",false },
  {0x0019,0x0026,"DLX_SERIE_01",VR::DS,VM::VM1,"Intercept LV Regression",false },
  {0x0019,0x0027,"DLX_SERIE_01",VR::DS,VM::VM1,"Image chain FWHM psf mm min",false },
  {0x0019,0x0028,"DLX_SERIE_01",VR::DS,VM::VM1,"Image chain FWHM psf mm max",false },
  {0x0015,0x008f,"DL_INTERNAL_USE",VR::IS,VM::VM1,"?",false },
  {0x4453,0x0000,"DR Systems, Inc.",VR::LO,VM::VM1,"?kind of identifier?",false },
  {0x4453,0x0001,"DR Systems, Inc.",VR::LO,VM::VM1,"DR Exam ID",false },
  {0x4453,0x0002,"DR Systems, Inc.",VR::LO,VM::VM1,"DR Image Type",false },
  {0x4453,0x0004,"DR Systems, Inc.",VR::LO,VM::VM1,"DR File Type",false },
  {0x4453,0x0005,"DR Systems, Inc.",VR::LO,VM::VM1,"DR File Suffix",false },
  {0x4453,0x000a,"DR Systems, Inc.",VR::UN,VM::VM1,"DR Annotation Type",false },
  {0x4453,0x000c,"DR Systems, Inc.",VR::SQ,VM::VM1,"DR Original Instance UID Sequence",false },
  {0x4453,0x000d,"DR Systems, Inc.",VR::PN,VM::VM1,"?",false },
  {0x4453,0x000e,"DR Systems, Inc.",VR::DA,VM::VM1,"?",false },
  {0x4453,0x000f,"DR Systems, Inc.",VR::CS,VM::VM1,"?",false },
  {0x4453,0x0010,"DR Systems, Inc.",VR::LO,VM::VM1,"?",false },
  {0x4453,0x0011,"DR Systems, Inc.",VR::DA,VM::VM1,"?",false },
  {0x4453,0x0012,"DR Systems, Inc.",VR::TM,VM::VM1,"?",false },
  {0x4453,0x0013,"DR Systems, Inc.",VR::LO,VM::VM1,"?",false },
  {0x7fa1,0x0000,"ELGEMS_XPERT",VR::OB,VM::VM1,"XPert Composite Report Item buffer",false },
  {0x7fa1,0x0010,"ELGEMS_XPERT",VR::OB,VM::VM1,"Xpert Composite Report Item Identificator",false },
  {0x0003,0x0001,"ELSCINT1",VR::OW,VM::VM1,"Offset List Structure",false },
  {0x00e1,0x0001,"ELSCINT1",VR::US,VM::VM1,"Data Dictionary Version",false },
  {0x00e1,0x0002,"ELSCINT1",VR::SH,VM::VM1,"?",false },
  {0x00e1,0x0005,"ELSCINT1",VR::IS,VM::VM1,"?",false },
  {0x00e1,0x0006,"ELSCINT1",VR::IS,VM::VM1,"?",false },
  {0x00e1,0x0007,"ELSCINT1",VR::IS,VM::VM1,"?",false },
  {0x00e1,0x0008,"ELSCINT1",VR::UI,VM::VM1,"SOP Instance UID",false },
  {0x00e1,0x0009,"ELSCINT1",VR::CS,VM::VM1,"Image Type[1]",false },
  {0x00e1,0x0014,"ELSCINT1",VR::CS,VM::VM1,"?",false },
  {0x00e1,0x0017,"ELSCINT1",VR::IS,VM::VM1,"Number of Series Related Instances",false },
  {0x00e1,0x0018,"ELSCINT1",VR::OB,VM::VM1,"?",false },
  {0x00e1,0x0019,"ELSCINT1",VR::ST,VM::VM1_8,"Referenced File ID",false },
  {0x00e1,0x0020,"ELSCINT1",VR::SQ,VM::VM1,"?Referenced Image Sequence?",false },
  {0x00e1,0x0021,"ELSCINT1",VR::DS,VM::VM1,"ELSCINT1_DLP_TOTAL (private)",false },
  {0x00e1,0x0022,"ELSCINT1",VR::DS,VM::VM2,"Presentation Relative Center",false },
  {0x00e1,0x0023,"ELSCINT1",VR::DS,VM::VM2,"Presentation Relative Part",false },
  {0x00e1,0x0024,"ELSCINT1",VR::CS,VM::VM1,"Presentation Horizontal Invert",false },
  {0x00e1,0x0025,"ELSCINT1",VR::CS,VM::VM1,"Presentation Vertical Invert",false },
  {0x00e1,0x0026,"ELSCINT1",VR::CS,VM::VM1,"? Orientation ?",false },
  {0x00e1,0x0027,"ELSCINT1",VR::DS,VM::VM1,"Image Position (Patient)[3]",false },
  {0x00e1,0x002a,"ELSCINT1",VR::DS,VM::VM1,"?",false },
  {0x00e1,0x0030,"ELSCINT1",VR::UI,VM::VM1,"?",false },
  {0x00e1,0x0031,"ELSCINT1",VR::CS,VM::VM1,"?",false },
  {0x00e1,0x0032,"ELSCINT1",VR::US,VM::VM2,"?",false },
  {0x00e1,0x0036,"ELSCINT1",VR::CS,VM::VM1,"?YES/NO?",false },
  {0x00e1,0x0037,"ELSCINT1",VR::DS,VM::VM1,"ELSCINT1_TOTAL_SAVING_DOSE",false },
  {0x00e1,0x0039,"ELSCINT1",VR::SQ,VM::VM1,"?",false },
  {0x00e1,0x003e,"ELSCINT1",VR::IS,VM::VM1,"Private (00E1,103E)",false },
  {0x00e1,0x003f,"ELSCINT1",VR::CS,VM::VM1,"?",false },
  {0x00e1,0x0040,"ELSCINT1",VR::SH,VM::VM1,"Image Label",false },
  {0x00e1,0x0041,"ELSCINT1",VR::DS,VM::VM1,"?",false },
  {0x00e1,0x0042,"ELSCINT1",VR::LO,VM::VM1,"?",false },
  {0x00e1,0x0043,"ELSCINT1",VR::IS,VM::VM1,"?",false },
  {0x00e1,0x0045,"ELSCINT1",VR::SL,VM::VM1,"?",false },
  {0x00e1,0x0046,"ELSCINT1",VR::OB,VM::VM1,"?",false },
  {0x00e1,0x0050,"ELSCINT1",VR::DS,VM::VM1,"Acquisition Duration",false },
  {0x00e1,0x0051,"ELSCINT1",VR::SH,VM::VM1,"?",false },
  {0x00e1,0x0060,"ELSCINT1",VR::CS,VM::VM1,"?",false },
  {0x00e1,0x0061,"ELSCINT1",VR::LO,VM::VM1,"Protocol File Name",false },
  {0x00e1,0x0062,"ELSCINT1",VR::CS,VM::VM1,"?",false },
  {0x00e1,0x0063,"ELSCINT1",VR::SH,VM::VM1,"Patient Language",false },
  {0x00e1,0x0065,"ELSCINT1",VR::LO,VM::VM1,"Patient Data Modification Date",false },
  {0x00e1,0x006a,"ELSCINT1",VR::IS,VM::VM1,"?",false },
  {0x00e1,0x006b,"ELSCINT1",VR::IS,VM::VM1,"?",false },
  {0x00e1,0x0084,"ELSCINT1",VR::OB,VM::VM1,"?Tracker Info?",false },
  {0x00e1,0x00a0,"ELSCINT1",VR::LO,VM::VM1,"?",false },
  {0x00e1,0x00c2,"ELSCINT1",VR::UI,VM::VM1,"PET-CT Multi Modality Name",false },
  {0x00e1,0x00c4,"ELSCINT1",VR::DS,VM::VM1,"Private (00E1,10C4)",false },
  {0x00e1,0x00e1,"ELSCINT1",VR::US,VM::VM1,"?",false },
  {0x00e1,0x00e3,"ELSCINT1",VR::US,VM::VM1,"?",false },
  {0x00e1,0x00ec,"ELSCINT1",VR::US,VM::VM1,"?",false },
  {0x00e3,0x0000,"ELSCINT1",VR::OB,VM::VM1,"?",false },
  {0x00e3,0x0018,"ELSCINT1",VR::OB,VM::VM1,"?",false },
  {0x00e3,0x001a,"ELSCINT1",VR::OB,VM::VM1,"?",false },
  {0x00e3,0x001f,"ELSCINT1",VR::OB,VM::VM1,"?",false },
  {0x01e1,0x0017,"ELSCINT1",VR::UI,VM::VM1,"?UI?",false },
  {0x01e1,0x0018,"ELSCINT1",VR::OB,VM::VM1,"Wave Information",false },
  {0x01e1,0x001f,"ELSCINT1",VR::SH,VM::VM1,"?",false },
  {0x01e1,0x0021,"ELSCINT1",VR::UI,VM::VM1,"CT Gating UID",false },
  {0x01e1,0x0023,"ELSCINT1",VR::SH,VM::VM1,"?",false },
  {0x01e1,0x0026,"ELSCINT1",VR::CS,VM::VM1,"PhantomType (private)",false },
  {0x01e1,0x0034,"ELSCINT1",VR::IS,VM::VM1,"?",false },
  {0x01e1,0x0040,"ELSCINT1",VR::UI,VM::VM1,"?",false },
  {0x01e1,0x0041,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f1,0x0001,"ELSCINT1",VR::CS,VM::VM1,"Acquisition Type",false },
  {0x01f1,0x0002,"ELSCINT1",VR::CS,VM::VM1,"Focal Spot Resolution",false },
  {0x01f1,0x0003,"ELSCINT1",VR::CS,VM::VM1,"Concurrent Slices Generation",false },
  {0x01f1,0x0004,"ELSCINT1",VR::CS,VM::VM1,"Angular Sampling Density",false },
  {0x01f1,0x0005,"ELSCINT1",VR::DS,VM::VM1,"Reconstruction Arc",false },
  {0x01f1,0x0006,"ELSCINT1",VR::DS,VM::VM1,"?",false },
  {0x01f1,0x0007,"ELSCINT1",VR::DS,VM::VM1,"Table Velocity",false },
  {0x01f1,0x0008,"ELSCINT1",VR::DS,VM::VM1,"Acquisition Length",false },
  {0x01f1,0x000a,"ELSCINT1",VR::US,VM::VM1,"Edge Enhancement Weight",false },
  {0x01f1,0x000b,"ELSCINT1",VR::CS,VM::VM2,"?",false },
  {0x01f1,0x000c,"ELSCINT1",VR::DS,VM::VM2,"Scanner Relative Center",false },
  {0x01f1,0x000d,"ELSCINT1",VR::DS,VM::VM1,"Rotation Angle",false },
  {0x01f1,0x000e,"ELSCINT1",VR::FL,VM::VM1,"Private (01F1,100E)",false },
  {0x01f1,0x000f,"ELSCINT1",VR::CS,VM::VM1,"?",false },
  {0x01f1,0x0026,"ELSCINT1",VR::DS,VM::VM1,"Pitch",false },
  {0x01f1,0x0027,"ELSCINT1",VR::DS,VM::VM1,"Rotation Time",false },
  {0x01f1,0x0028,"ELSCINT1",VR::DS,VM::VM1,"Table Increment",false },
  {0x01f1,0x0030,"ELSCINT1",VR::US,VM::VM1,"?",false },
  {0x01f1,0x0032,"ELSCINT1",VR::CS,VM::VM1,"Image View Convention",false },
  {0x01f1,0x0033,"ELSCINT1",VR::DS,VM::VM1,"Cycle Time",false },
  {0x01f1,0x0036,"ELSCINT1",VR::CS,VM::VM1,"?",false },
  {0x01f1,0x0037,"ELSCINT1",VR::DS,VM::VM1,"?",false },
  {0x01f1,0x0038,"ELSCINT1",VR::LO,VM::VM1,"CT Technique (eg. Bolus Tracking)",false },
  {0x01f1,0x0039,"ELSCINT1",VR::LO,VM::VM1,"?",false },
  {0x01f1,0x0040,"ELSCINT1",VR::CS,VM::VM1,"?",false },
  {0x01f1,0x0041,"ELSCINT1",VR::CS,VM::VM1,"?Percent?",false },
  {0x01f1,0x0042,"ELSCINT1",VR::SH,VM::VM1,"?",false },
  {0x01f1,0x0043,"ELSCINT1",VR::LO,VM::VM1,"?",false },
  {0x01f1,0x0044,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f1,0x0045,"ELSCINT1",VR::IS,VM::VM1,"Private (01F1,1045)",false },
  {0x01f1,0x0046,"ELSCINT1",VR::FL,VM::VM1,"?",false },
  {0x01f1,0x0047,"ELSCINT1",VR::SH,VM::VM1,"Acquisition Dimension (2D/3D)",false },
  {0x01f1,0x0049,"ELSCINT1",VR::DS,VM::VM1,"Private (01F1,1049)",false },
  {0x01f1,0x004a,"ELSCINT1",VR::SH,VM::VM1,"?",false },
  {0x01f1,0x004b,"ELSCINT1",VR::SH,VM::VM1,"Private (01F1,104B)",false },
  {0x01f1,0x004c,"ELSCINT1",VR::SH,VM::VM1,"Private (01F1,104C)",false },
  {0x01f1,0x004d,"ELSCINT1",VR::SH,VM::VM1,"Private (01F1,104D)",false },
  {0x01f1,0x004e,"ELSCINT1",VR::LO,VM::VM1,"Private (01F1,104E)",false },
  {0x01f1,0x0053,"ELSCINT1",VR::SH,VM::VM1,"?",false },
  {0x01f3,0x0001,"ELSCINT1",VR::SQ,VM::VM1,"?",false },
  {0x01f3,0x0002,"ELSCINT1",VR::SS,VM::VM1,"?",false },
  {0x01f3,0x0003,"ELSCINT1",VR::FL,VM::VM2,"?",false },
  {0x01f3,0x0004,"ELSCINT1",VR::FL,VM::VM1,"?",false },
  {0x01f3,0x0011,"ELSCINT1",VR::SQ,VM::VM1,"PS Sequence",false },
  {0x01f3,0x0012,"ELSCINT1",VR::SS,VM::VM1,"?",false },
  {0x01f3,0x0013,"ELSCINT1",VR::FL,VM::VM2,"?",false },
  {0x01f3,0x0014,"ELSCINT1",VR::FL,VM::VM1,"?",false },
  {0x01f3,0x0015,"ELSCINT1",VR::US,VM::VM1,"?",false },
  {0x01f3,0x0016,"ELSCINT1",VR::FL,VM::VM1,"?",false },
  {0x01f3,0x0017,"ELSCINT1",VR::FL,VM::VM1,"?",false },
  {0x01f3,0x0018,"ELSCINT1",VR::SH,VM::VM1,"?",false },
  {0x01f3,0x0019,"ELSCINT1",VR::FL,VM::VM1,"?",false },
  {0x01f3,0x0023,"ELSCINT1",VR::US,VM::VM1,"?",false },
  {0x01f3,0x0024,"ELSCINT1",VR::IS,VM::VM2,"?",false },
  {0x01f7,0x0010,"ELSCINT1",VR::OB,VM::VM1,"?",false },
  {0x01f7,0x0011,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x0013,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x0014,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x0015,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x0016,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x0017,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x0018,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x0019,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x001a,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x001b,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x001c,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x001e,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x001f,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x0022,"ELSCINT1",VR::UI,VM::VM1,"?UID?",false },
  {0x01f7,0x0023,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x0025,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x0026,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x0027,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x0028,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x0029,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x002b,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x002c,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x002d,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x002e,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x0030,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x0031,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x0070,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x0073,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x0074,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x0075,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x0077,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x007f,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x008b,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x0091,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x0095,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x01f7,0x0097,"ELSCINT1",VR::UN,VM::VM1,"?",false },
  {0x01f7,0x0099,"ELSCINT1",VR::UN,VM::VM1,"?",false },
  {0x01f7,0x009b,"ELSCINT1",VR::IS,VM::VM1,"iDose Level",false },
  {0x01f9,0x0001,"ELSCINT1",VR::LO,VM::VM1,"SP Filter (private)",false },
  {0x01f9,0x0004,"ELSCINT1",VR::IS,VM::VM1,"Adaptive Filter (private)",false },
  {0x01f9,0x0005,"ELSCINT1",VR::IS,VM::VM1,"Recon Increatment (private)",false },
  {0x01f9,0x0008,"ELSCINT1",VR::DS,VM::VM1,"Private (01F9,1008)",false },
  {0x01f9,0x0009,"ELSCINT1",VR::DS,VM::VM1,"Private (01F1,1009)",false },
  {0x0203,0x0001,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0205,0x001e,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0205,0x0032,"ELSCINT1",VR::CS,VM::VM2,"???",false },
  {0x0205,0x0035,"ELSCINT1",VR::UN,VM::VM4,"???",false },
  {0x0205,0x0036,"ELSCINT1",VR::UN,VM::VM4,"???",false },
  {0x0205,0x0037,"ELSCINT1",VR::UN,VM::VM4,"???",false },
  {0x0205,0x0065,"ELSCINT1",VR::UN,VM::VM2,"???",false },
  {0x0205,0x006d,"ELSCINT1",VR::CS,VM::VM1,"???",false },
  {0x0205,0x0079,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0205,0x00a6,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0207,0x0012,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0207,0x0013,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0207,0x0014,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0207,0x0015,"ELSCINT1",VR::IS,VM::VM1,"???",false },
  {0x0207,0x0016,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0207,0x0017,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0207,0x0018,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0207,0x0019,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0207,0x001a,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0207,0x001b,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0207,0x001c,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0207,0x001d,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0207,0x001e,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0207,0x001f,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0207,0x0023,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0207,0x0025,"ELSCINT1",VR::CS,VM::VM1,"???",false },
  {0x0207,0x0027,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0207,0x002a,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0207,0x002b,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0207,0x002c,"ELSCINT1",VR::CS,VM::VM1,"???",false },
  {0x0207,0x0034,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0207,0x0060,"ELSCINT1",VR::UI,VM::VM1,"???",false },
  {0x0207,0x0070,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0207,0x0072,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0207,0x0073,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0207,0x0075,"ELSCINT1",VR::CS,VM::VM1,"???",false },
  {0x0215,0x0010,"ELSCINT1",VR::CS,VM::VM1,"???",false },
  {0x0215,0x0011,"ELSCINT1",VR::CS,VM::VM1,"???",false },
  {0x0215,0x0013,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0215,0x0014,"ELSCINT1",VR::CS,VM::VM1,"???",false },
  {0x0215,0x0018,"ELSCINT1",VR::CS,VM::VM1,"???",false },
  {0x0215,0x0019,"ELSCINT1",VR::CS,VM::VM1,"???",false },
  {0x0215,0x001a,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0215,0x001b,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0215,0x0045,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0215,0x0046,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0215,0x0047,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0215,0x004b,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0215,0x004c,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0215,0x004d,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0215,0x0051,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0215,0x0053,"ELSCINT1",VR::CS,VM::VM1,"???",false },
  {0x0215,0x00a2,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0217,0x006e,"ELSCINT1",VR::CS,VM::VM1,"???",false },
  {0x0217,0x006f,"ELSCINT1",VR::CS,VM::VM1,"???",false },
  {0x0217,0x0091,"ELSCINT1",VR::CS,VM::VM1,"???",false },
  {0x0217,0x0093,"ELSCINT1",VR::CS,VM::VM1,"???",false },
  {0x0217,0x0094,"ELSCINT1",VR::CS,VM::VM1,"???",false },
  {0x0217,0x0097,"ELSCINT1",VR::CS,VM::VM1,"???",false },
  {0x0223,0x0003,"ELSCINT1",VR::CS,VM::VM1,"???",false },
  {0x0223,0x0010,"ELSCINT1",VR::CS,VM::VM1,"???",false },
  {0x0223,0x0012,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0223,0x0013,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0223,0x0026,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0223,0x0027,"ELSCINT1",VR::CS,VM::VM1,"???",false },
  {0x0223,0x002b,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0223,0x002c,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0223,0x002d,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0223,0x002e,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0227,0x0001,"ELSCINT1",VR::CS,VM::VM1,"???",false },
  {0x0227,0x0002,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0227,0x0003,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0227,0x0004,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0227,0x0005,"ELSCINT1",VR::CS,VM::VM1,"???",false },
  {0x0227,0x0006,"ELSCINT1",VR::CS,VM::VM1,"???",false },
  {0x0227,0x0007,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0227,0x0008,"ELSCINT1",VR::CS,VM::VM1,"???",false },
  {0x0227,0x0009,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0227,0x000a,"ELSCINT1",VR::UN,VM::VM1,"???",false },
  {0x0227,0x000b,"ELSCINT1",VR::CS,VM::VM1,"???",false },
  {0x0601,0x0000,"ELSCINT1",VR::SH,VM::VM1,"Implementation Version",false },
  {0x0601,0x0020,"ELSCINT1",VR::DS,VM::VM1,"Relative Table Position",false },
  {0x0601,0x0021,"ELSCINT1",VR::DS,VM::VM1,"Relative Table Height",false },
  {0x0601,0x0030,"ELSCINT1",VR::SH,VM::VM1,"Surview Direction",false },
  {0x0601,0x0031,"ELSCINT1",VR::DS,VM::VM1,"Surview Length",false },
  {0x0601,0x0050,"ELSCINT1",VR::SH,VM::VM1,"Image View Type",false },
  {0x0601,0x0070,"ELSCINT1",VR::DS,VM::VM1,"Batch Number",false },
  {0x0601,0x0071,"ELSCINT1",VR::DS,VM::VM1,"Batch Size",false },
  {0x0601,0x0072,"ELSCINT1",VR::DS,VM::VM1,"Batch Slice Number",false },
  {0x07a1,0x0001,"ELSCINT1",VR::UL,VM::VM1,"number of series in study",false },
  {0x07a1,0x0002,"ELSCINT1",VR::UL,VM::VM1,"number of images in series",false },
  {0x07a1,0x0003,"ELSCINT1",VR::UL,VM::VM1,"last update time",false },
  {0x07a1,0x0004,"ELSCINT1",VR::UL,VM::VM1,"last update date",false },
  {0x07a1,0x0007,"ELSCINT1",VR::US,VM::VM3,"?",false },
  {0x07a1,0x0008,"ELSCINT1",VR::DS,VM::VM1_n,"?",false },
  {0x07a1,0x0009,"ELSCINT1",VR::OW,VM::VM1,"?",false },
  {0x07a1,0x000a,"ELSCINT1",VR::OB,VM::VM1,"Tamar Compressed Pixel Data",false },
  {0x07a1,0x000c,"ELSCINT1",VR::US,VM::VM1,"?",false },
  {0x07a1,0x0010,"ELSCINT1",VR::LO,VM::VM1,"Tamar Software Version",false },
  {0x07a1,0x0011,"ELSCINT1",VR::CS,VM::VM1,"Tamar Compression Type",false },
  {0x07a1,0x0012,"ELSCINT1",VR::FL,VM::VM1_n,"?",false },
  {0x07a1,0x0013,"ELSCINT1",VR::UL,VM::VM1,"?",false },
  {0x07a1,0x0014,"ELSCINT1",VR::LO,VM::VM1,"protection flag",false },
  {0x07a1,0x0016,"ELSCINT1",VR::FL,VM::VM1_n,"?",false },
  {0x07a1,0x0018,"ELSCINT1",VR::SQ,VM::VM1,"?",false },
  {0x07a1,0x0019,"ELSCINT1",VR::FL,VM::VM1,"?",false },
  {0x07a1,0x001c,"ELSCINT1",VR::FL,VM::VM1_n,"?",false },
  {0x07a1,0x002a,"ELSCINT1",VR::CS,VM::VM1,"Tamar Study Status",false },
  {0x07a1,0x002b,"ELSCINT1",VR::CS,VM::VM1,"?",false },
  {0x07a1,0x0036,"ELSCINT1",VR::AE,VM::VM1,"Tamar Source Ae",false },
  {0x07a1,0x0040,"ELSCINT1",VR::CS,VM::VM1,"Tamar Study Body Part",false },
  {0x07a1,0x0042,"ELSCINT1",VR::SH,VM::VM1,"?",false },
  {0x07a1,0x0043,"ELSCINT1",VR::IS,VM::VM1,"?",false },
  {0x07a1,0x0047,"ELSCINT1",VR::CS,VM::VM1,"?orientation?",false },
  {0x07a1,0x004a,"ELSCINT1",VR::SH,VM::VM1,"?name?",false },
  {0x07a1,0x0050,"ELSCINT1",VR::US,VM::VM1,"Tamar Site Id",false },
  {0x07a1,0x0056,"ELSCINT1",VR::US,VM::VM1,"?",false },
  {0x07a1,0x0058,"ELSCINT1",VR::CS,VM::VM1,"?",false },
  {0x07a1,0x005a,"ELSCINT1",VR::CS,VM::VM1,"?",false },
  {0x07a1,0x005d,"ELSCINT1",VR::DT,VM::VM1,"?",false },
  {0x07a1,0x005f,"ELSCINT1",VR::CS,VM::VM1,"?",false },
  {0x07a1,0x0070,"ELSCINT1",VR::SH,VM::VM1,"?patient name?",false },
  {0x07a1,0x0071,"ELSCINT1",VR::SH,VM::VM1,"?num?",false },
  {0x07a1,0x0075,"ELSCINT1",VR::LO,VM::VM2,"?",false },
  {0x07a1,0x0085,"ELSCINT1",VR::UL,VM::VM1,"Tamar Translate Flags",false },
  {0x07a1,0x0088,"ELSCINT1",VR::CS,VM::VM1,"?Y/N?",false },
  {0x07a1,0x0098,"ELSCINT1",VR::CS,VM::VM1,"?Y/N?",false },
  {0x07a1,0x009f,"ELSCINT1",VR::CS,VM::VM1,"?",false },
  {0x07a1,0x00a7,"ELSCINT1",VR::CS,VM::VM1,"?",false },
  {0x07a1,0x00c0,"ELSCINT1",VR::ST,VM::VM1,"?",false },
  {0x07a1,0x00c1,"ELSCINT1",VR::ST,VM::VM1,"?",false },
  {0x07a1,0x00c2,"ELSCINT1",VR::ST,VM::VM1,"?status sent?",false },
  {0x07a1,0x00c8,"ELSCINT1",VR::ST,VM::VM1,"?",false },
  {0x07a1,0x00c9,"ELSCINT1",VR::ST,VM::VM1,"?",false },
  {0x07a1,0x00ca,"ELSCINT1",VR::ST,VM::VM1,"?",false },
  {0x07a1,0x00d0,"ELSCINT1",VR::LO,VM::VM1,"?",false },
  {0x07a3,0x0001,"ELSCINT1",VR::LO,VM::VM1,"Tamar Exe Software Version",false },
  {0x07a3,0x0003,"ELSCINT1",VR::CS,VM::VM1,"Tamar Study Has Sticky Note",false },
  {0x07a3,0x0005,"ELSCINT1",VR::CS,VM::VM1,"?",false },
  {0x07a3,0x0006,"ELSCINT1",VR::CS,VM::VM1,"?",false },
  {0x07a3,0x0010,"ELSCINT1",VR::LO,VM::VM1,"Tamar Original Curve Desc",false },
  {0x07a3,0x0013,"ELSCINT1",VR::SH,VM::VM1,"?num?",false },
  {0x07a3,0x0014,"ELSCINT1",VR::ST,VM::VM1,"?num?",false },
  {0x07a3,0x0015,"ELSCINT1",VR::ST,VM::VM1,"?hosp. service name?",false },
  {0x07a3,0x0017,"ELSCINT1",VR::SH,VM::VM1,"?num?",false },
  {0x07a3,0x0019,"ELSCINT1",VR::ST,VM::VM1,"?",false },
  {0x07a3,0x001a,"ELSCINT1",VR::ST,VM::VM1,"?",false },
  {0x07a3,0x001b,"ELSCINT1",VR::ST,VM::VM1,"?num?",false },
  {0x07a3,0x001c,"ELSCINT1",VR::ST,VM::VM1,"?patient name?",false },
  {0x07a3,0x001d,"ELSCINT1",VR::ST,VM::VM1,"?radiologist name?",false },
  {0x07a3,0x001e,"ELSCINT1",VR::ST,VM::VM1,"?num?",false },
  {0x07a3,0x001f,"ELSCINT1",VR::ST,VM::VM1,"?",false },
  {0x07a3,0x0020,"ELSCINT1",VR::DA,VM::VM1,"?",false },
  {0x07a3,0x0022,"ELSCINT1",VR::ST,VM::VM1,"?",false },
  {0x07a3,0x0023,"ELSCINT1",VR::ST,VM::VM1,"?+sign?",false },
  {0x07a3,0x0024,"ELSCINT1",VR::ST,VM::VM1,"?",false },
  {0x07a3,0x0034,"ELSCINT1",VR::SH,VM::VM1,"Tamar Study Age",false },
  {0x07a3,0x0039,"ELSCINT1",VR::UL,VM::VM1,"?",false },
  {0x07a3,0x0043,"ELSCINT1",VR::DS,VM::VM1_n,"?",false },
  {0x07a3,0x004c,"ELSCINT1",VR::ST,VM::VM1,"?",false },
  {0x07a3,0x0052,"ELSCINT1",VR::LO,VM::VM1,"?",false },
  {0x07a3,0x0055,"ELSCINT1",VR::SH,VM::VM1,"?",false },
  {0x07a3,0x005c,"ELSCINT1",VR::ST,VM::VM1,"?",false },
  {0x07a3,0x0061,"ELSCINT1",VR::LT,VM::VM1,"?",false },
  {0x07a3,0x0062,"ELSCINT1",VR::SQ,VM::VM1,"?",false },
  {0x07a3,0x0063,"ELSCINT1",VR::SQ,VM::VM1,"?",false },
  {0x07a3,0x0064,"ELSCINT1",VR::IS,VM::VM1_n,"?",false },
  {0x07a3,0x0065,"ELSCINT1",VR::CS,VM::VM1_n,"?yes/no bool?",false },
  {0x07a3,0x0066,"ELSCINT1",VR::IS,VM::VM1,"?",false },
  {0x07a3,0x008c,"ELSCINT1",VR::CS,VM::VM1,"?",false },
  {0x07a3,0x008f,"ELSCINT1",VR::CS,VM::VM1,"?",false },
  {0x07a3,0x0099,"ELSCINT1",VR::CS,VM::VM1,"?yes/no bool?",false },
  {0x07a3,0x009c,"ELSCINT1",VR::CS,VM::VM1,"?yes/no bool?",false },
  {0x07a3,0x009f,"ELSCINT1",VR::CS,VM::VM1,"?",false },
  {0x07a3,0x00b4,"ELSCINT1",VR::US,VM::VM1,"?num?",false },
  {0x07a3,0x00b9,"ELSCINT1",VR::CS,VM::VM1,"?yes/no bool?",false },
  {0x07a3,0x00bb,"ELSCINT1",VR::CS,VM::VM1,"?yes/no bool?",false },
  {0x07a3,0x00c1,"ELSCINT1",VR::LO,VM::VM1,"?PACS name?",false },
  {0x07a3,0x00c2,"ELSCINT1",VR::CS,VM::VM1,"?net name?",false },
  {0x07a3,0x00c3,"ELSCINT1",VR::CS,VM::VM1,"?secure?",false },
  {0x07a3,0x00c4,"ELSCINT1",VR::LO,VM::VM1,"Tamar Grid Token Next Host",false },
  {0x07a3,0x00c5,"ELSCINT1",VR::LO,VM::VM1,"Tamar Grid Token Next Host Port",false },
  {0x07a3,0x00c8,"ELSCINT1",VR::AE,VM::VM1,"Tamar Grid Token Ae",false },
  {0x07a3,0x00c9,"ELSCINT1",VR::CS,VM::VM1,"Tamar Grid Token Tunneled",false },
  {0x07a3,0x00c0,"ELSCINT1",VR::SQ,VM::VM1,"?network seq0?",false },
  {0x07a3,0x00c6,"ELSCINT1",VR::SQ,VM::VM1,"?network seq6?",false },
  {0x07a3,0x00ca,"ELSCINT1",VR::SQ,VM::VM1,"?network seq10?",false },
  {0x07a3,0x00cb,"ELSCINT1",VR::SQ,VM::VM1,"?network seq11?",false },
  {0x07a3,0x00cc,"ELSCINT1",VR::LO,VM::VM1,"Tamar Grid Token Version",false },
  {0x07a3,0x00ce,"ELSCINT1",VR::SQ,VM::VM1,"?",false },
  {0x07a3,0x00e3,"ELSCINT1",VR::LO,VM::VM1,"?name?",false },
  {0x07a3,0x00f2,"ELSCINT1",VR::CS,VM::VM1,"?",false },
  {0x07a3,0x00fb,"ELSCINT1",VR::DT,VM::VM1,"?",false },
  {0x07a5,0x0000,"ELSCINT1",VR::LO,VM::VM1,"?name?",false },
  {0x07a5,0x0010,"ELSCINT1",VR::UL,VM::VM1,"?",false },
  {0x07a5,0x0054,"ELSCINT1",VR::DT,VM::VM1,"?",false },
  {0x07a5,0x0056,"ELSCINT1",VR::CS,VM::VM1,"?",false },
  {0x07a5,0x0059,"ELSCINT1",VR::IS,VM::VM1,"?",false },
  {0x07a5,0x0063,"ELSCINT1",VR::CS,VM::VM1_n,"?",false },
  {0x07a5,0x0069,"ELSCINT1",VR::LO,VM::VM1,"?",false }, // AE sometimes ??
  {0x07a5,0x0072,"ELSCINT1",VR::DT,VM::VM1,"?",false },
  {0x07a5,0x00ae,"ELSCINT1",VR::IS,VM::VM1,"?",false },
  {0x07a5,0x00c8,"ELSCINT1",VR::CS,VM::VM1,"?",false },
  {0x07a5,0x00dc,"ELSCINT1",VR::UL,VM::VM1,"?",false },
  {0x07a5,0x00dd,"ELSCINT1",VR::FL,VM::VM1,"?",false },
  {0x07a5,0x00de,"ELSCINT1",VR::LO,VM::VM1,"?",false },
  {0x07a5,0x00e8,"ELSCINT1",VR::IS,VM::VM1,"?",false },
  {0x5001,0x0070,"ELSCINT1",VR::SQ,VM::VM1,"?",false },
  {0x5001,0x0071,"ELSCINT1",VR::SH,VM::VM1,"?",false },
  {0x5001,0x0080,"ELSCINT1",VR::SQ,VM::VM1,"?",false },
  {0x5001,0x0081,"ELSCINT1",VR::SH,VM::VM1,"?",false },
  {0x5001,0x0082,"ELSCINT1",VR::US,VM::VM3,"?",false },
  {0x5001,0x0083,"ELSCINT1",VR::FL,VM::VM1_n,"?",false },
  {0x5001,0x0084,"ELSCINT1",VR::SQ,VM::VM1,"?",false },
  {0x6001,0x0010,"ELSCINT1",VR::CS,VM::VM1,"Text Overlay Flag",false },
  {0x6003,0x0010,"ELSCINT1",VR::CS,VM::VM1,"Text Overlay Flag",false },
  {0x7fdf,0x00f0,"ELSCINT1",VR::OB,VM::VM1,"?",false },
  {0x7fdf,0x00ff,"ELSCINT1",VR::SH,VM::VM1,"?",false },
  {0x0009,0x0000,"EMAGEON JPEG2K INFO",VR::SQ,VM::VM1,"?",false },
  {0x0009,0x0001,"EMAGEON JPEG2K INFO",VR::DT,VM::VM1,"?",false },
  {0x0009,0x0000,"EMAGEON STUDY HOME",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0001,"EMAGEON STUDY HOME",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0030,"Eclipse 60",VR::UL,VM::VM1,"Private data",false },
  {0x0029,0x0031,"Eclipse 60",VR::UL,VM::VM1,"Private data",false },
  {0x0029,0x0032,"Eclipse 60",VR::UL,VM::VM1,"Private data",false },
  {0x0029,0x0033,"Eclipse 60",VR::DS,VM::VM1,"Private data",false },
  {0x0029,0x0034,"Eclipse 60",VR::DS,VM::VM1,"Private data",false },
  {0x0029,0x0036,"Eclipse 60",VR::SL,VM::VM1,"Private data",false },
  {0x0029,0x0040,"Eclipse 60",VR::CS,VM::VM1,"Private data",false },
  {0x0029,0x0050,"Eclipse 60",VR::CS,VM::VM1,"Private data",false },
  {0x0029,0x0051,"Eclipse 60",VR::UL,VM::VM1,"Private data",false },
  {0x0029,0x0052,"Eclipse 60",VR::UL,VM::VM1,"Private data",false },
  {0x0029,0x0053,"Eclipse 60",VR::DS,VM::VM1,"Private data",false },
  {0x0029,0x0054,"Eclipse 60",VR::UL,VM::VM1,"Private data",false },
  {0x0029,0x0055,"Eclipse 60",VR::FL,VM::VM1,"Private data",false },
  {0x0029,0x0056,"Eclipse 60",VR::US,VM::VM1,"Private data",false },
  {0x0029,0x0060,"Eclipse 60",VR::LO,VM::VM1,"Private data",false },
  {0x0031,0x0030,"Eclipse 60",VR::UL,VM::VM1,"Private data",false },
  {0x0031,0x0031,"Eclipse 60",VR::UL,VM::VM1,"Private data",false },
  {0x0031,0x0032,"Eclipse 60",VR::LO,VM::VM1,"Private data",false },
  {0x0031,0x0033,"Eclipse 60",VR::UL,VM::VM1,"Private data",false },
  {0x0031,0x0034,"Eclipse 60",VR::UL,VM::VM1,"Private data",false },
  {0x0031,0x0035,"Eclipse 60",VR::UL,VM::VM1,"Private data",false },
  {0x200d,0x0001,"Eclipse 60",VR::LO,VM::VM1,"Private data",false },
  {0x200d,0x0002,"Eclipse 60",VR::LO,VM::VM1,"Private data",false },
  {0x200d,0x0003,"Eclipse 60",VR::LO,VM::VM1,"Private data",false },
  {0x200d,0x0004,"Eclipse 60",VR::LO,VM::VM1,"Private data",false },
  {0x200d,0x0005,"Eclipse 60",VR::LO,VM::VM1,"Private data",false },
  {0x200d,0x0006,"Eclipse 60",VR::LO,VM::VM1,"Private data",false },
  {0x200d,0x0007,"Eclipse 60",VR::LO,VM::VM1,"Private data",false },
  {0x200d,0x0008,"Eclipse 60",VR::LO,VM::VM1,"Private data",false },
  {0x200d,0x0009,"Eclipse 60",VR::LO,VM::VM1,"Private data",false },
  {0x200d,0x000a,"Eclipse 60",VR::IS,VM::VM1,"Private data",false },
  {0x200d,0x000b,"Eclipse 60",VR::OB,VM::VM1,"Private data",false },
  {0x200d,0x000c,"Eclipse 60",VR::IS,VM::VM1,"Private data",false },
  {0x200d,0x000d,"Eclipse 60",VR::UL,VM::VM1,"Private data",false },
  {0x200d,0x000e,"Eclipse 60",VR::IS,VM::VM1,"Private data",false },
  {0x200d,0x000f,"Eclipse 60",VR::IS,VM::VM1,"Private data",false },
  {0x200d,0x0010,"Eclipse 60",VR::IS,VM::VM1,"Private data",false },
  {0x200d,0x0011,"Eclipse 60",VR::IS,VM::VM1,"Private data",false },
  {0x200d,0x0012,"Eclipse 60",VR::OB,VM::VM1,"Private data",false },
  {0x200d,0x0013,"Eclipse 60",VR::IS,VM::VM1,"Private data",false },
  {0x200d,0x0014,"Eclipse 60",VR::IS,VM::VM1,"Private data",false },
  {0x200d,0x0015,"Eclipse 60",VR::IS,VM::VM1,"Private data",false },
  {0x200d,0x001a,"Eclipse 60",VR::IS,VM::VM1,"Private data",false },
  {0x200d,0x001b,"Eclipse 60",VR::IS,VM::VM1,"Private data",false },
  {0x200d,0x0020,"Eclipse 60",VR::SQ,VM::VM1,"Private data",false },
  {0x200d,0x0027,"Eclipse 60",VR::FD,VM::VM1,"Private data",false },
  {0x200d,0x0028,"Eclipse 60",VR::FD,VM::VM1,"Private data",false },
  {0x200d,0x0030,"Eclipse 60",VR::LO,VM::VM1,"Private data",false },
  {0x200d,0x0031,"Eclipse 60",VR::LO,VM::VM1,"Private data",false },
  {0x200d,0x0032,"Eclipse 60",VR::LO,VM::VM1,"Private data",false },
  {0x200d,0x0033,"Eclipse 60",VR::LO,VM::VM1,"Private data",false },
  {0x200d,0x0034,"Eclipse 60",VR::LO,VM::VM1,"Private data",false },
  {0x200d,0x0035,"Eclipse 60",VR::LO,VM::VM1,"Private data",false },
  {0x200d,0x0036,"Eclipse 60",VR::LO,VM::VM1,"Private data",false },
  {0x200d,0x0037,"Eclipse 60",VR::LO,VM::VM1,"Private data",false },
  {0x200d,0x0038,"Eclipse 60",VR::LO,VM::VM1,"Private data",false },
  {0x200d,0x0039,"Eclipse 60",VR::LO,VM::VM1,"Private data",false },
  {0x200d,0x003a,"Eclipse 60",VR::LO,VM::VM1,"Private data",false },
  {0x200d,0x0050,"Eclipse 60",VR::CS,VM::VM1,"Private data",false },
  {0x200d,0x0051,"Eclipse 60",VR::CS,VM::VM1,"Private data",false },
  {0x200d,0x0052,"Eclipse 60",VR::FD,VM::VM1,"Private data",false },
  {0x200d,0x0053,"Eclipse 60",VR::CS,VM::VM1,"Private data",false },
  {0x200d,0x0054,"Eclipse 60",VR::CS,VM::VM1,"Private data",false },
  {0x200d,0x0055,"Eclipse 60",VR::FD,VM::VM1,"Private data",false },
  {0x200d,0x0056,"Eclipse 60",VR::FD,VM::VM1,"Private data",false },
  {0x200d,0x0057,"Eclipse 60",VR::FD,VM::VM1,"Private data",false },
  {0x200d,0x0058,"Eclipse 60",VR::IS,VM::VM1,"Private data",false },
  {0x200d,0x0059,"Eclipse 60",VR::FD,VM::VM1,"Private data",false },
  {0x200d,0x005a,"Eclipse 60",VR::FD,VM::VM1,"Private data",false },
  {0x200d,0x005d,"Eclipse 60",VR::CS,VM::VM1,"Private data",false },
  {0x200d,0x005e,"Eclipse 60",VR::CS,VM::VM1,"Private data",false },
  {0x200d,0x00ff,"Eclipse 60",VR::IS,VM::VM1,"Private data",false },
  {0x7777,0x0001,"Eclipse 60",VR::DA,VM::VM1,"Private data",false },
  {0x7777,0x0002,"Eclipse 60",VR::CS,VM::VM1,"Private data",false },
  {0x7777,0x0003,"Eclipse 60",VR::LO,VM::VM1,"Private data",false },
  {0x7777,0x0010,"Eclipse 60",VR::LO,VM::VM1,"Private data",false },
  {0x7777,0x0014,"Eclipse 60",VR::LT,VM::VM1,"Private data",false },
  {0x0859,0x0010,"ETIAM RE UIDENTIFY",VR::SQ,VM::VM1,"Some SQ",false },
  {0x0009,0x0004,"FDMS 1.0",VR::SH,VM::VM1,"Image Control Unit",false },
  {0x0009,0x0005,"FDMS 1.0",VR::OW,VM::VM1,"Image UID",false },
  {0x0009,0x0006,"FDMS 1.0",VR::OW,VM::VM1,"Route Image UID",false },
  {0x0009,0x0008,"FDMS 1.0",VR::UL,VM::VM1,"Image Display Information Version No.",false },
  {0x0009,0x0009,"FDMS 1.0",VR::UL,VM::VM1,"Patient Information Version No.",false },
  {0x0009,0x000c,"FDMS 1.0",VR::OW,VM::VM1,"Film UID",false },
  {0x0009,0x0010,"FDMS 1.0",VR::CS,VM::VM1,"Exposure Unit Type Code",false },
  {0x0009,0x0080,"FDMS 1.0",VR::LO,VM::VM1,"Kanji Hospital Name",false },
  {0x0009,0x0090,"FDMS 1.0",VR::ST,VM::VM1,"Distribution Code",false },
  {0x0009,0x0092,"FDMS 1.0",VR::SH,VM::VM1,"Kanji Department Name",false },
  {0x0009,0x00f0,"FDMS 1.0",VR::CS,VM::VM1,"Blackening Process Flag",false },
  {0x0009,0x00f1,"FDMS 1.0",VR::ST,VM::VM1,"Processing Information Flag",false },
  {0x0009,0x00f2,"FDMS 1.0",VR::CS,VM::VM1,"Normalization Flag",false },
  {0x0009,0x00f3,"FDMS 1.0",VR::CS,VM::VM1,"Tone characteristic",false },
  {0x0009,0x00f4,"FDMS 1.0",VR::CS,VM::VM1,"Window Value Fixed Flag",false },
  {0x0019,0x0015,"FDMS 1.0",VR::LO,VM::VM1,"Kanji Body Part for Exposure",false },
  {0x0019,0x0030,"FDMS 1.0",VR::LO,VM::VM1,"Menu Character String",false },
  {0x0019,0x0032,"FDMS 1.0",VR::LO,VM::VM1,"Kanji Menu Name",false },
  {0x0019,0x0040,"FDMS 1.0",VR::CS,VM::VM1,"Image Processing Type",false },
  {0x0019,0x0050,"FDMS 1.0",VR::CS,VM::VM1,"EDR Mode",false },
  {0x0019,0x0060,"FDMS 1.0",VR::SH,VM::VM1,"Radiographer's Code",false },
  {0x0019,0x0070,"FDMS 1.0",VR::IS,VM::VM1,"Split Exposure Format",false },
  {0x0019,0x0071,"FDMS 1.0",VR::IS,VM::VM1,"No. of Split Exposure Frames",false },
  {0x0019,0x0080,"FDMS 1.0",VR::IS,VM::VM1,"Reading Position Specification",false },
  {0x0019,0x0081,"FDMS 1.0",VR::IS,VM::VM1,"Reading Sensitivity Center",false },
  {0x0019,0x0090,"FDMS 1.0",VR::SH,VM::VM1,"Film Annotation Character String 1",false },
  {0x0019,0x0091,"FDMS 1.0",VR::SH,VM::VM1,"Film Annotation Character String 2",false },
  {0x0019,0x00a0,"FDMS 1.0",VR::SQ,VM::VM1,"Exposure Status Sequence",false },
  {0x0019,0x00a1,"FDMS 1.0",VR::CS,VM::VM1,"Exposure Status",false },
  {0x0019,0x00a2,"FDMS 1.0",VR::CS,VM::VM1,"Exposure Kind",false },
  {0x0021,0x0010,"FDMS 1.0",VR::CS,VM::VM1,"FCR Image ID",false },
  {0x0021,0x0030,"FDMS 1.0",VR::CS,VM::VM1,"Set No.",false },
  {0x0021,0x0040,"FDMS 1.0",VR::IS,VM::VM1,"Image No. in the Set",false },
  {0x0021,0x0050,"FDMS 1.0",VR::CS,VM::VM1,"Pair Processing Information",false },
  {0x0021,0x0070,"FDMS 1.0",VR::IS,VM::VM1,"Film Number within the Series",false },
  {0x0021,0x0080,"FDMS 1.0",VR::OB,VM::VM1,"Equipment Type-Specific Information",false },
  {0x0021,0x0090,"FDMS 1.0",VR::CS,VM::VM1,"LUT Number",false },
  {0x0023,0x0010,"FDMS 1.0",VR::SQ,VM::VM1,"Left-image Processing Parameter Sequence",false },
  {0x0023,0x0020,"FDMS 1.0",VR::SQ,VM::VM1,"Right-image Processing Parameter Sequence",false },
  {0x0023,0x0030,"FDMS 1.0",VR::SQ,VM::VM1,"Single-image Processing Parameter Sequence",false },
  {0x0025,0x0010,"FDMS 1.0",VR::US,VM::VM1,"Relative Light Emission Amount Sk",false },
  {0x0025,0x0011,"FDMS 1.0",VR::US,VM::VM1,"Term of Correction for Each IP Type St",false },
  {0x0025,0x0012,"FDMS 1.0",VR::US,VM::VM1,"Reading Gain Gp",false },
  {0x0025,0x0013,"FDMS 1.0",VR::US,VM::VM1,"Latitude",false },
  {0x0025,0x0015,"FDMS 1.0",VR::CS,VM::VM1,"Image Processing Selection Flag",false },
  {0x0025,0x0016,"FDMS 1.0",VR::US,VM::VM1,"Min",false },
  {0x0025,0x0017,"FDMS 1.0",VR::US,VM::VM1,"Max",false },
  {0x0025,0x0020,"FDMS 1.0",VR::US,VM::VM2,"Sensitivity Shift (S-SHFT)",false },
  {0x0025,0x0021,"FDMS 1.0",VR::US,VM::VM1,"Contrast Shift (C-SHIFT)",false },
  {0x0025,0x0030,"FDMS 1.0",VR::US,VM::VM1,"GT",false },
  {0x0025,0x0031,"FDMS 1.0",VR::SS,VM::VM1,"GA",false },
  {0x0025,0x0032,"FDMS 1.0",VR::US,VM::VM1,"GC",false },
  {0x0025,0x0033,"FDMS 1.0",VR::SS,VM::VM1,"GS",false },
  {0x0025,0x0034,"FDMS 1.0",VR::SS,VM::VM1,"tan θ",false },
  {0x0025,0x0040,"FDMS 1.0",VR::US,VM::VM1,"RT",false },
  {0x0025,0x0041,"FDMS 1.0",VR::US,VM::VM1,"RN",false },
  {0x0025,0x0042,"FDMS 1.0",VR::US,VM::VM1,"RE",false },
  {0x0025,0x0043,"FDMS 1.0",VR::US,VM::VM1,"N",false },
  {0x0025,0x0050,"FDMS 1.0",VR::US,VM::VM1,"DRT",false },
  {0x0025,0x0051,"FDMS 1.0",VR::US,VM::VM1,"DRN",false },
  {0x0025,0x0052,"FDMS 1.0",VR::US,VM::VM1,"DRE",false },
  {0x0025,0x0053,"FDMS 1.0",VR::US,VM::VM1,"DN",false },
  {0x0025,0x0060,"FDMS 1.0",VR::US,VM::VM1,"ORN",false },
  {0x0025,0x0061,"FDMS 1.0",VR::US,VM::VM1,"ORE",false },
  {0x0025,0x0062,"FDMS 1.0",VR::US,VM::VM1,"ON",false },
  {0x0025,0x0063,"FDMS 1.0",VR::CS,VM::VM1,"ORD",false },
  {0x0025,0x0070,"FDMS 1.0",VR::US,VM::VM1,"MDT",false },
  {0x0025,0x0071,"FDMS 1.0",VR::US,VM::VM1,"MDB",false },
  {0x0025,0x0072,"FDMS 1.0",VR::US,VM::VM1,"MDE",false },
  {0x0025,0x0073,"FDMS 1.0",VR::US,VM::VM1_n,"MFP-DRC Suppression Characteristics Parameters",false },
  {0x0025,0x0074,"FDMS 1.0",VR::US,VM::VM1_n,"MFP-DRC Frequency Characteristics Parameters",false },
  {0x0025,0x0080,"FDMS 1.0",VR::US,VM::VM1,"MRT",false },
  {0x0025,0x0081,"FDMS 1.0",VR::US,VM::VM1,"MRB",false },
  {0x0025,0x0082,"FDMS 1.0",VR::US,VM::VM1,"MRE",false },
  {0x0025,0x0083,"FDMS 1.0",VR::US,VM::VM1_n,"MFP-USM Suppression Characteristics Parameters",false },
  {0x0025,0x0084,"FDMS 1.0",VR::US,VM::VM1_n,"MFP-USM Frequency Characteristics Parameters",false },
  {0x0025,0x0090,"FDMS 1.0",VR::US,VM::VM1,"PTE (α edge)",false },
  {0x0025,0x0091,"FDMS 1.0",VR::US,VM::VM1,"PTR (α calc)",false },
  {0x0025,0x0092,"FDMS 1.0",VR::US,VM::VM1,"PRN",false },
  {0x0025,0x0093,"FDMS 1.0",VR::US,VM::VM1,"PRE",false },
  {0x0025,0x0094,"FDMS 1.0",VR::US,VM::VM1,"PEM Sstd",false },
  {0x0025,0x0095,"FDMS 1.0",VR::US,VM::VM1,"PEM Lstd",false },
  {0x0025,0x0096,"FDMS 1.0",VR::CS,VM::VM1,"PEM Unsharp Mask Value N",false },
  {0x0025,0x00a0,"FDMS 1.0",VR::US,VM::VM1,"MFP GT",false },
  {0x0025,0x00a1,"FDMS 1.0",VR::SS,VM::VM1,"MFP GA",false },
  {0x0025,0x00a2,"FDMS 1.0",VR::US,VM::VM1,"MFP GC",false },
  {0x0025,0x00a3,"FDMS 1.0",VR::SS,VM::VM1,"MFP GS",false },
  {0x0027,0x0010,"FDMS 1.0",VR::SQ,VM::VM1,"γ Raw Data Table Sequence",false },
  {0x0027,0x0020,"FDMS 1.0",VR::SQ,VM::VM1,"β Raw Data Table Sequence",false },
  {0x0027,0x0030,"FDMS 1.0",VR::SQ,VM::VM1,"DRC Raw Data Table Sequence",false },
  {0x0027,0x0040,"FDMS 1.0",VR::SQ,VM::VM1,"MFP-DRC Raw Data Table Sequence",false },
  {0x0027,0x0050,"FDMS 1.0",VR::SQ,VM::VM1,"MFP-USM Raw Data Table Sequence",false },
  {0x0027,0x0060,"FDMS 1.0",VR::SQ,VM::VM1,"PEM α Edge Raw Data Table Sequence",false },
  {0x0027,0x0070,"FDMS 1.0",VR::SQ,VM::VM1,"PEM α Calc Raw Data Table Sequence",false },
  {0x0027,0x0080,"FDMS 1.0",VR::SQ,VM::VM1,"MFP γ Raw Data Table Sequence",false },
  {0x0027,0x00a0,"FDMS 1.0",VR::IS,VM::VM1,"No. of Data",false },
  {0x0027,0x00a1,"FDMS 1.0",VR::CS,VM::VM2,"Data Input Type",false },
  {0x0027,0x00a2,"FDMS 1.0",VR::CS,VM::VM2,"Data Output Type",false },
  {0x0027,0x00a3,"FDMS 1.0",VR::US_SS,VM::VM1_n,"Raw Data",false },
  {0x0029,0x0020,"FDMS 1.0",VR::CS,VM::VM1,"Image Scanning Direction",false },
  {0x0029,0x0025,"FDMS 1.0",VR::CS,VM::VM1,"Image Rotation/Reversal Information",false },
  {0x0029,0x0030,"FDMS 1.0",VR::CS,VM::VM1,"Extended Reading Size Value",false },
  {0x0029,0x0034,"FDMS 1.0",VR::US,VM::VM1,"Mag./Reduc. Ratio",false },
  {0x0029,0x0044,"FDMS 1.0",VR::CS,VM::VM1,"Line Density Code",false },
  {0x0029,0x0050,"FDMS 1.0",VR::CS,VM::VM1,"Data Compression Code",false },
  {0x2011,0x0000,"FDMS 1.0",VR::CS,VM::VM1,"Trim Density",false },
  {0x2011,0x0001,"FDMS 1.0",VR::IS,VM::VM1,"Trim Width",false },
  {0x2011,0x0002,"FDMS 1.0",VR::CS,VM::VM1,"Image Mag./Reduc. Range",false },
  {0x2011,0x0010,"FDMS 1.0",VR::CS,VM::VM1,"Image Display Format",false },
  {0x2011,0x0011,"FDMS 1.0",VR::CS,VM::VM1,"Image Position Specifying Flag",false },
  {0x2011,0x0020,"FDMS 1.0",VR::OW,VM::VM1,"Interpolation A-VRS System Format",false },
  {0x50f1,0x0006,"FDMS 1.0",VR::CS,VM::VM1,"Energy Subtraction Param.",false },
  {0x50f1,0x0007,"FDMS 1.0",VR::CS,VM::VM1,"Subtraction Registration Result",false },
  {0x50f1,0x0008,"FDMS 1.0",VR::CS,VM::VM1,"Energy Subtraction Param. 2",false },
  {0x50f1,0x0009,"FDMS 1.0",VR::SL,VM::VM1,"Afin Conversion Coefficient",false },
  {0x50f1,0x000a,"FDMS 1.0",VR::ST,VM::VM1,"FNC Parameters",false },
  {0x50f1,0x000b,"FDMS 1.0",VR::SS,VM::VM1,"CRF Parameters",false },
  {0x50f1,0x0010,"FDMS 1.0",VR::CS,VM::VM1,"Film Output Format",false },
  {0x50f1,0x0020,"FDMS 1.0",VR::CS,VM::VM1,"Image Processing Modification Flag",false },
  {0x0009,0x0001,"FFP DATA",VR::UN,VM::VM1,"CR Header Information",false },
  {0x4d4d,0x0001,"GDCM CONFORMANCE TESTS",VR::SQ,VM::VM1,"Array Container Sequence",false },
  {0x4d4d,0x0002,"GDCM CONFORMANCE TESTS",VR::OB,VM::VM1,"0 Filled Array",false },
  {0x0019,0x0030,"GE ??? From Adantage Review CS",VR::LO,VM::VM1,"CR EDR Mode",false },
  {0x0019,0x0040,"GE ??? From Adantage Review CS",VR::LO,VM::VM1,"CR Latitude",false },
  {0x0019,0x0050,"GE ??? From Adantage Review CS",VR::LO,VM::VM1,"CR Group Number",false },
  {0x0019,0x0070,"GE ??? From Adantage Review CS",VR::LO,VM::VM1,"CR Image Serial Number",false },
  {0x0019,0x0080,"GE ??? From Adantage Review CS",VR::LO,VM::VM1,"CR Bar Code Number",false },
  {0x0019,0x0090,"GE ??? From Adantage Review CS",VR::LO,VM::VM1,"CR Film Output Exposures",false },
  {0x0045,0x0067,"GE LUT Asymmetry Parameter",VR::DS,VM::VM1,"LUT Asymmetry",false },
  {0x0053,0x0020,"GEHC_CT_ADVAPP_001",VR::IS,VM::VM1,"ShuttleFlag",false },
  {0x0053,0x0021,"GEHC_CT_ADVAPP_001",VR::IS,VM::VM1,"TableSpeedNotReachesTargetFlag",false },
  {0x0053,0x0040,"GEHC_CT_ADVAPP_001",VR::SH,VM::VM1,"IterativeReconAnnotation",false },
  {0x0053,0x0041,"GEHC_CT_ADVAPP_001",VR::SH,VM::VM1,"IterativeReconMode",false },
  {0x0053,0x0042,"GEHC_CT_ADVAPP_001",VR::LO,VM::VM1,"IterativeReconConfiguration",false },
  {0x0053,0x0043,"GEHC_CT_ADVAPP_001",VR::SH,VM::VM1,"IterativeReconLevel",false },
  {0x0053,0x0060,"GEHC_CT_ADVAPP_001",VR::SH,VM::VM1,"reconFlipRotateAnno",false },
  {0x0053,0x0061,"GEHC_CT_ADVAPP_001",VR::SH,VM::VM1,"highResolutionFlag",false },
  {0x0053,0x0062,"GEHC_CT_ADVAPP_001",VR::SH,VM::VM1,"RespiratoryFlag",false },
  {0x0053,0x0064,"GEHC_CT_ADVAPP_001",VR::IS,VM::VM1,"Shutter Mode",false },
  {0x0053,0x0065,"GEHC_CT_ADVAPP_001",VR::IS,VM::VM1,"Shutter Mode Percent",false },
  {0x0053,0x0066,"GEHC_CT_ADVAPP_001",VR::LO,VM::VM1,"Image Browser Annotation",false },
  {0x0053,0x0067,"GEHC_CT_ADVAPP_001",VR::IS,VM::VM1,"Overlapped Recon Flag",false },
  {0x0053,0x0068,"GEHC_CT_ADVAPP_001",VR::IS,VM::VM1,"Row Number Annotation Flag",false },
  {0x0053,0x006a,"GEHC_CT_ADVAPP_001",VR::IS,VM::VM1,"ODMFlag",false },
  {0x0053,0x006b,"GEHC_CT_ADVAPP_001",VR::IS,VM::VM1,"ODMReductionPercent",false },
  {0x0053,0x006c,"GEHC_CT_ADVAPP_001",VR::DS,VM::VM1,"?",false },
  {0x0053,0x006d,"GEHC_CT_ADVAPP_001",VR::DS,VM::VM1,"?",false },
  {0x0053,0x006f,"GEHC_CT_ADVAPP_001",VR::IS,VM::VM1,"ASiR Auto Guidance Dose Reduction Percent",false },
  {0x0053,0x009d,"GEHC_CT_ADVAPP_001",VR::LO,VM::VM1,"MARsAnnotation",false },
  {0x0009,0x0010,"GEIIS",VR::SQ,VM::VM1,"GE IIS Thumbnail Sequence",false },
  {0x0009,0x0012,"GEIIS",VR::IS,VM::VM1,"?",false },
  {0x0029,0x0010,"GEIIS",VR::UL,VM::VM1,"Shift Count",false },
  {0x0029,0x0012,"GEIIS",VR::UL,VM::VM1,"Offset",false },
  {0x0029,0x0014,"GEIIS",VR::UL,VM::VM1,"Actual Frame Number",false },
  {0x004b,0x0013,"GEIIS",VR::IS,VM::VM1,"?",false },
  {0x004b,0x0015,"GEIIS",VR::LT,VM::VM1,"?",false },
  {0x0905,0x0030,"GEIIS",VR::LO,VM::VM1,"Assigning Authority For Patient ID",false },
  {0x0907,0x0010,"GEIIS",VR::UI,VM::VM1,"Original Study Instance UID",false },
  {0x0907,0x0020,"GEIIS",VR::UI,VM::VM1,"Original Series Instance UID",false },
  {0x0907,0x0030,"GEIIS",VR::UI,VM::VM1,"Original SOP Instance UID",false },
  {0x7fd1,0x0010,"GEIIS",VR::UL,VM::VM1,"GE IIS Compression Type",false },
  {0x7fd1,0x0020,"GEIIS",VR::UL,VM::VM1_n,"Multiframe Offsets",false },
  {0x7fd1,0x0030,"GEIIS",VR::UL,VM::VM1,"?",false },
  {0x7fd1,0x0040,"GEIIS",VR::UL,VM::VM1_n,"?",false },
  {0x7fd1,0x0050,"GEIIS",VR::UL,VM::VM1_n,"?",false },
  {0x7fd1,0x0060,"GEIIS",VR::UL,VM::VM1_n,"?",false },
  {0x0903,0x0010,"GEIIS PACS",VR::US,VM::VM1,"Reject Image Flag",false },
  {0x0903,0x0011,"GEIIS PACS",VR::US,VM::VM1,"Significant Flag",false },
  {0x0903,0x0012,"GEIIS PACS",VR::US,VM::VM1,"Confidential Flag",false },
  {0x0903,0x0020,"GEIIS PACS",VR::CS,VM::VM1,"?",false },
  {0x0907,0x0021,"GEIIS PACS",VR::US,VM::VM1,"GEIIS_PREFETCH_ALGORITHM",false },
  {0x0907,0x0022,"GEIIS PACS",VR::US,VM::VM1,"GEIIS_LIMIT_RECENT_STUDIES",false },
  {0x0907,0x0023,"GEIIS PACS",VR::US,VM::VM1,"GEIIS_LIMIT_OLDEST_STUDIES",false },
  {0x0907,0x0024,"GEIIS PACS",VR::US,VM::VM1,"GEIIS_LIMIT_RECENT_MONTHS",false },
  {0x0907,0x0031,"GEIIS PACS",VR::UI,VM::VM1_n,"GEIIS_EXCLUDE_STUDY_UIDS",false },
  {0x4113,0x0010,"GEMS-IT/Centricity RA600/7.0",VR::UI,VM::VM1,"Number of images in study",false },
  {0x0039,0x0095,"GEMS_0039",VR::LO,VM::VM1,"?",false },
  {0x0047,0x00d6,"GEMS_3DSTATE_001",VR::ST,VM::VM1,"General Description",false },
  {0x0047,0x00d7,"GEMS_3DSTATE_001",VR::CS,VM::VM1,"TDRT",false },
  {0x0047,0x00d8,"GEMS_3DSTATE_001",VR::US,VM::VM1,"NVRP",false },
  {0x0047,0x00d9,"GEMS_3DSTATE_001",VR::US,VM::VM1,"CVRPN",false },
  {0x0047,0x00da,"GEMS_3DSTATE_001",VR::SQ,VM::VM1,"Volume Rendering Presets Sequence",false },
  {0x0047,0x00db,"GEMS_3DSTATE_001",VR::LO,VM::VM1,"Preset Name",false },
  {0x0047,0x00dc,"GEMS_3DSTATE_001",VR::SS,VM::VM1_n,"Opacity Curve X",false },
  {0x0047,0x00dd,"GEMS_3DSTATE_001",VR::FL,VM::VM1_n,"Opacity Curve Y",false },
  {0x0047,0x00de,"GEMS_3DSTATE_001",VR::US,VM::VM1,"NOCP",false },
  {0x0047,0x00df,"GEMS_3DSTATE_001",VR::SS,VM::VM1_n,"Color Curve X",false },
  {0x0047,0x00e1,"GEMS_3DSTATE_001",VR::US,VM::VM1,"NCCP",false },
  {0x0047,0x00e2,"GEMS_3DSTATE_001",VR::CS,VM::VM1,"GSA",false },
  {0x0047,0x00e3,"GEMS_3DSTATE_001",VR::CS,VM::VM1,"VRSF",false },
  {0x0047,0x00e4,"GEMS_3DSTATE_001",VR::FL,VM::VM1,"AF",false },
  {0x0047,0x00e5,"GEMS_3DSTATE_001",VR::FL,VM::VM1,"DF",false },
  {0x0047,0x00e6,"GEMS_3DSTATE_001",VR::FL,VM::VM1,"SCF",false },
  {0x0047,0x00e7,"GEMS_3DSTATE_001",VR::FL,VM::VM1,"SPF",false },
  {0x0047,0x00e8,"GEMS_3DSTATE_001",VR::FL,VM::VM24,"Orthogonal Clipping Planes",false },
  {0x0047,0x00e9,"GEMS_3DSTATE_001",VR::FL,VM::VM3,"CP",false },
  {0x0047,0x00ea,"GEMS_3DSTATE_001",VR::DS,VM::VM3,"CFP",false },
  {0x0047,0x00eb,"GEMS_3DSTATE_001",VR::DS,VM::VM3,"CVU",false },
  {0x0047,0x00ec,"GEMS_3DSTATE_001",VR::FL,VM::VM1,"RFOV",false },
  {0x0047,0x00ed,"GEMS_3DSTATE_001",VR::CS,VM::VM1,"PPRP",false },
  {0x0047,0x00ee,"GEMS_3DSTATE_001",VR::DS,VM::VM1,"3DWW",false },
  {0x0047,0x00ef,"GEMS_3DSTATE_001",VR::DS,VM::VM1,"3DWL",false },
  {0x0047,0x00f0,"GEMS_3DSTATE_001",VR::CS,VM::VM1,"BBV",false },
  {0x0047,0x00f1,"GEMS_3DSTATE_001",VR::CS,VM::VM1,"ERF",false },
  {0x0047,0x00f2,"GEMS_3DSTATE_001",VR::US,VM::VM1,"TDRMS",false },
  {0x0047,0x00f3,"GEMS_3DSTATE_001",VR::FL,VM::VM1,"TDSSS",false },
  {0x0009,0x0024,"GEMS_ACQU_01",VR::DS,VM::VM1,"?",false },
  {0x0009,0x0025,"GEMS_ACQU_01",VR::US,VM::VM1,"?",false },
  {0x0009,0x003e,"GEMS_ACQU_01",VR::US,VM::VM1,"?",false },
  {0x0009,0x003f,"GEMS_ACQU_01",VR::US,VM::VM1,"?",false },
  {0x0009,0x0042,"GEMS_ACQU_01",VR::US,VM::VM1,"?",false },
  {0x0009,0x0043,"GEMS_ACQU_01",VR::US,VM::VM1,"?",false },
  {0x0009,0x00f8,"GEMS_ACQU_01",VR::US,VM::VM1,"?",false },
  {0x0009,0x00fb,"GEMS_ACQU_01",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0001,"GEMS_ACQU_01",VR::LT,VM::VM1,"?",false },
  {0x0019,0x0002,"GEMS_ACQU_01",VR::SL,VM::VM1,"Detector Channel",false },
  {0x0019,0x0003,"GEMS_ACQU_01",VR::DS,VM::VM1,"Cell number at Theta",false },
  {0x0019,0x0004,"GEMS_ACQU_01",VR::DS,VM::VM1,"Cell spacing",false },
  {0x0019,0x0005,"GEMS_ACQU_01",VR::LT,VM::VM1,"?",false },
  {0x0019,0x0006,"GEMS_ACQU_01",VR::UN,VM::VM1,"?",false },
  {0x0019,0x000e,"GEMS_ACQU_01",VR::US,VM::VM1,"?",false },
  {0x0019,0x000f,"GEMS_ACQU_01",VR::DS,VM::VM1,"Horiz. Frame of ref.",false },
  {0x0019,0x0011,"GEMS_ACQU_01",VR::SS,VM::VM1,"Series contrast",false },
  {0x0019,0x0012,"GEMS_ACQU_01",VR::SS,VM::VM1,"Last pseq",false },
  {0x0019,0x0013,"GEMS_ACQU_01",VR::SS,VM::VM1,"Start number for baseline",false },
  {0x0019,0x0014,"GEMS_ACQU_01",VR::SS,VM::VM1,"End number for baseline",false },
  {0x0019,0x0015,"GEMS_ACQU_01",VR::SS,VM::VM1,"Start number for enhanced scans",false },
  {0x0019,0x0016,"GEMS_ACQU_01",VR::SS,VM::VM1,"End number for enhanced scans",false },
  {0x0019,0x0017,"GEMS_ACQU_01",VR::SS,VM::VM1,"Series plane",false },
  {0x0019,0x0018,"GEMS_ACQU_01",VR::LO,VM::VM1,"First scan ras",false },
  {0x0019,0x0019,"GEMS_ACQU_01",VR::DS,VM::VM1,"First scan location",false },
  {0x0019,0x001a,"GEMS_ACQU_01",VR::LO,VM::VM1,"Last scan ras",false },
  {0x0019,0x001b,"GEMS_ACQU_01",VR::DS,VM::VM1,"Last scan loc",false },
  {0x0019,0x001e,"GEMS_ACQU_01",VR::DS,VM::VM1,"Display field of view",false },
  {0x0019,0x0020,"GEMS_ACQU_01",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0022,"GEMS_ACQU_01",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0023,"GEMS_ACQU_01",VR::DS,VM::VM1,"Table Speed [mm/rotation]",false },
  {0x0019,0x0024,"GEMS_ACQU_01",VR::DS,VM::VM1,"Mid Scan Time [sec]",false },
  {0x0019,0x0025,"GEMS_ACQU_01",VR::SS,VM::VM1,"Mid scan flag",false },
  {0x0019,0x0026,"GEMS_ACQU_01",VR::SL,VM::VM1,"Tube Azimuth [degree]",false },
  {0x0019,0x0027,"GEMS_ACQU_01",VR::DS,VM::VM1,"Rotation Speed [msec]",false },
  {0x0019,0x002a,"GEMS_ACQU_01",VR::DS,VM::VM1,"x-ray On position",false },
  {0x0019,0x002b,"GEMS_ACQU_01",VR::DS,VM::VM1,"x-ray Off position",false },
  {0x0019,0x002c,"GEMS_ACQU_01",VR::SL,VM::VM1,"Number of triggers",false },
  {0x0019,0x002d,"GEMS_ACQU_01",VR::US,VM::VM1,"?",false },
  {0x0019,0x002e,"GEMS_ACQU_01",VR::DS,VM::VM1,"Angle of first view",false },
  {0x0019,0x002f,"GEMS_ACQU_01",VR::DS,VM::VM1,"Trigger frequency",false },
  {0x0019,0x0039,"GEMS_ACQU_01",VR::SS,VM::VM1,"SFOV Type",false },
  {0x0019,0x003a,"GEMS_ACQU_01",VR::IS,VM::VM1,"?",false },
  {0x0019,0x003b,"GEMS_ACQU_01",VR::LT,VM::VM1,"?",false },
  {0x0019,0x003c,"GEMS_ACQU_01",VR::UN,VM::VM1,"?",false },
  {0x0019,0x003e,"GEMS_ACQU_01",VR::UN,VM::VM1,"?",false },
  {0x0019,0x003f,"GEMS_ACQU_01",VR::UN,VM::VM1,"?",false },
  {0x0019,0x0040,"GEMS_ACQU_01",VR::SS,VM::VM1,"Stat recon flag",false },
  {0x0019,0x0041,"GEMS_ACQU_01",VR::SS,VM::VM1,"Compute type",false },
  {0x0019,0x0042,"GEMS_ACQU_01",VR::SS,VM::VM1,"Segment Number",false },
  {0x0019,0x0043,"GEMS_ACQU_01",VR::SS,VM::VM1,"Total Segments Required",false },
  {0x0019,0x0044,"GEMS_ACQU_01",VR::DS,VM::VM1,"Interscan delay",false },
  {0x0019,0x0047,"GEMS_ACQU_01",VR::SS,VM::VM1,"View compression factor",false },
  {0x0019,0x0048,"GEMS_ACQU_01",VR::US,VM::VM1,"?",false },
  {0x0019,0x0049,"GEMS_ACQU_01",VR::US,VM::VM1,"?",false },
  {0x0019,0x004a,"GEMS_ACQU_01",VR::SS,VM::VM1,"Total no. of ref channels",false },
  {0x0019,0x004b,"GEMS_ACQU_01",VR::SL,VM::VM1,"Data size for scan data",false },
  {0x0019,0x0052,"GEMS_ACQU_01",VR::SS,VM::VM1,"Recon post proc. Flag",false },
  {0x0019,0x0054,"GEMS_ACQU_01",VR::UN,VM::VM1,"?",false },
  {0x0019,0x0057,"GEMS_ACQU_01",VR::SS,VM::VM1,"CT water number",false },
  {0x0019,0x0058,"GEMS_ACQU_01",VR::SS,VM::VM1,"CT bone number",false },
  {0x0019,0x005a,"GEMS_ACQU_01",VR::FL,VM::VM1,"Acquisition Duration",false },
  {0x0019,0x005d,"GEMS_ACQU_01",VR::US,VM::VM1,"?",false },
  {0x0019,0x005e,"GEMS_ACQU_01",VR::SL,VM::VM1,"Number of channels (1...512)",false },
  {0x0019,0x005f,"GEMS_ACQU_01",VR::SL,VM::VM1,"Increment between channels",false },
  {0x0019,0x0060,"GEMS_ACQU_01",VR::SL,VM::VM1,"Starting view",false },
  {0x0019,0x0061,"GEMS_ACQU_01",VR::SL,VM::VM1,"Number of views",false },
  {0x0019,0x0062,"GEMS_ACQU_01",VR::SL,VM::VM1,"Increment between views",false },
  {0x0019,0x006a,"GEMS_ACQU_01",VR::SS,VM::VM1,"Dependent on #views processed",false },
  {0x0019,0x006b,"GEMS_ACQU_01",VR::SS,VM::VM1,"Field of view in detector cells",false },
  {0x0019,0x0070,"GEMS_ACQU_01",VR::SS,VM::VM1,"Value of back projection button",false },
  {0x0019,0x0071,"GEMS_ACQU_01",VR::SS,VM::VM1,"Set if fatq estimates were used",false },
  {0x0019,0x0072,"GEMS_ACQU_01",VR::DS,VM::VM1,"Z chan avg over views",false },
  {0x0019,0x0073,"GEMS_ACQU_01",VR::DS,VM::VM1,"Avg of left ref chans over views",false },
  {0x0019,0x0074,"GEMS_ACQU_01",VR::DS,VM::VM1,"Max left chan over views",false },
  {0x0019,0x0075,"GEMS_ACQU_01",VR::DS,VM::VM1,"Avg of right ref chans over views",false },
  {0x0019,0x0076,"GEMS_ACQU_01",VR::DS,VM::VM1,"Max right chan over views",false },
  {0x0019,0x007d,"GEMS_ACQU_01",VR::DS,VM::VM1,"Second echo",false },
  {0x0019,0x007e,"GEMS_ACQU_01",VR::SS,VM::VM1,"Number of echoes",false },
  {0x0019,0x007f,"GEMS_ACQU_01",VR::DS,VM::VM1,"Table delta",false },
  {0x0019,0x0081,"GEMS_ACQU_01",VR::SS,VM::VM1,"Contiguous",false },
  {0x0019,0x0082,"GEMS_ACQU_01",VR::US,VM::VM1,"?",false },
  {0x0019,0x0083,"GEMS_ACQU_01",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0084,"GEMS_ACQU_01",VR::DS,VM::VM1,"Peak SAR",false },
  {0x0019,0x0085,"GEMS_ACQU_01",VR::SS,VM::VM1,"Monitor SAR",false },
  {0x0019,0x0086,"GEMS_ACQU_01",VR::US,VM::VM1,"?",false },
  {0x0019,0x0087,"GEMS_ACQU_01",VR::DS,VM::VM1,"Cardiac repetition time",false },
  {0x0019,0x0088,"GEMS_ACQU_01",VR::SS,VM::VM1,"Images per cardiac cycle",false },
  {0x0019,0x008a,"GEMS_ACQU_01",VR::SS,VM::VM1,"Actual receive gain analog",false },
  {0x0019,0x008b,"GEMS_ACQU_01",VR::SS,VM::VM1,"Actual receive gain digital",false },
  {0x0019,0x008d,"GEMS_ACQU_01",VR::DS,VM::VM1,"Delay after trigger",false },
  {0x0019,0x008f,"GEMS_ACQU_01",VR::SS,VM::VM1,"Swappf",false },
  {0x0019,0x0090,"GEMS_ACQU_01",VR::SS,VM::VM1,"Pause Interval",false },
  {0x0019,0x0091,"GEMS_ACQU_01",VR::DS,VM::VM1,"Pause Time",false },
  {0x0019,0x0092,"GEMS_ACQU_01",VR::SL,VM::VM1,"Slice offset on freq axis",false },
  {0x0019,0x0093,"GEMS_ACQU_01",VR::DS,VM::VM1,"Auto Prescan Center Frequency",false },
  {0x0019,0x0094,"GEMS_ACQU_01",VR::SS,VM::VM1,"Auto Prescan Transmit Gain",false },
  {0x0019,0x0095,"GEMS_ACQU_01",VR::SS,VM::VM1,"Auto Prescan Analog receiver gain",false },
  {0x0019,0x0096,"GEMS_ACQU_01",VR::SS,VM::VM1,"Auto Prescan Digital receiver gain",false },
  {0x0019,0x0097,"GEMS_ACQU_01",VR::SL,VM::VM1,"Bitmap defining CVs",false },
  {0x0019,0x0098,"GEMS_ACQU_01",VR::SS,VM::VM1,"Center freq. Method",false },
  {0x0019,0x0099,"GEMS_ACQU_01",VR::US,VM::VM1,"?",false },
  {0x0019,0x009b,"GEMS_ACQU_01",VR::SS,VM::VM1,"Pulse Sequence Mode",false },
  {0x0019,0x009c,"GEMS_ACQU_01",VR::LO,VM::VM1,"Pulse Sequence Name",false },
  {0x0019,0x009d,"GEMS_ACQU_01",VR::DT,VM::VM1,"Pulse Sequence Date",false },
  {0x0019,0x009e,"GEMS_ACQU_01",VR::LO,VM::VM1,"Internal Pulse Sequence Name",false },
  {0x0019,0x009f,"GEMS_ACQU_01",VR::SS,VM::VM1,"Transmitting Coil Type",false },
  {0x0019,0x00a0,"GEMS_ACQU_01",VR::SS,VM::VM1,"Surface Coil Type",false },
  {0x0019,0x00a1,"GEMS_ACQU_01",VR::SS,VM::VM1,"Extremity Coil flag",false },
  {0x0019,0x00a2,"GEMS_ACQU_01",VR::SL,VM::VM1,"Raw data run number",false },
  {0x0019,0x00a3,"GEMS_ACQU_01",VR::UL,VM::VM1,"Calibrated Field strength",false },
  {0x0019,0x00a4,"GEMS_ACQU_01",VR::SS,VM::VM1,"SAT fat/water/bone",false },
  {0x0019,0x00a5,"GEMS_ACQU_01",VR::DS,VM::VM1,"Receive bandwidth",false },
  {0x0019,0x00a7,"GEMS_ACQU_01",VR::DS,VM::VM1,"User data 0",false },
  {0x0019,0x00a8,"GEMS_ACQU_01",VR::DS,VM::VM1,"User data 1",false },
  {0x0019,0x00a9,"GEMS_ACQU_01",VR::DS,VM::VM1,"User data 2",false },
  {0x0019,0x00aa,"GEMS_ACQU_01",VR::DS,VM::VM1,"User data 3",false },
  {0x0019,0x00ab,"GEMS_ACQU_01",VR::DS,VM::VM1,"User data 4",false },
  {0x0019,0x00ac,"GEMS_ACQU_01",VR::DS,VM::VM1,"User data 5",false },
  {0x0019,0x00ad,"GEMS_ACQU_01",VR::DS,VM::VM1,"User data 6",false },
  {0x0019,0x00ae,"GEMS_ACQU_01",VR::DS,VM::VM1,"User data 7",false },
  {0x0019,0x00af,"GEMS_ACQU_01",VR::DS,VM::VM1,"User data 8",false },
  {0x0019,0x00b0,"GEMS_ACQU_01",VR::DS,VM::VM1,"User data 9",false },
  {0x0019,0x00b1,"GEMS_ACQU_01",VR::DS,VM::VM1,"User data 10",false },
  {0x0019,0x00b2,"GEMS_ACQU_01",VR::DS,VM::VM1,"User data 11",false },
  {0x0019,0x00b3,"GEMS_ACQU_01",VR::DS,VM::VM1,"User data 12",false },
  {0x0019,0x00b4,"GEMS_ACQU_01",VR::DS,VM::VM1,"User data 13",false },
  {0x0019,0x00b5,"GEMS_ACQU_01",VR::DS,VM::VM1,"User data 14",false },
  {0x0019,0x00b6,"GEMS_ACQU_01",VR::DS,VM::VM1,"User data 15",false },
  {0x0019,0x00b7,"GEMS_ACQU_01",VR::DS,VM::VM1,"User data 16",false },
  {0x0019,0x00b8,"GEMS_ACQU_01",VR::DS,VM::VM1,"User data 17",false },
  {0x0019,0x00b9,"GEMS_ACQU_01",VR::DS,VM::VM1,"User data 18",false },
  {0x0019,0x00ba,"GEMS_ACQU_01",VR::DS,VM::VM1,"User data 19",false },
  {0x0019,0x00bb,"GEMS_ACQU_01",VR::DS,VM::VM1,"User data 20 (?Diffusion Direction X?)",false },
  {0x0019,0x00bc,"GEMS_ACQU_01",VR::DS,VM::VM1,"User data 21 (?Diffusion Direction Y?)",false },
  {0x0019,0x00bd,"GEMS_ACQU_01",VR::DS,VM::VM1,"User data 22 (?Diffusion Direction Z?)",false },
  {0x0019,0x00be,"GEMS_ACQU_01",VR::DS,VM::VM1,"Projection angle",false },
  {0x0019,0x00c0,"GEMS_ACQU_01",VR::SS,VM::VM1,"Saturation planes",false },
  {0x0019,0x00c1,"GEMS_ACQU_01",VR::SS,VM::VM1,"Surface coil intensity",false },
  {0x0019,0x00c2,"GEMS_ACQU_01",VR::SS,VM::VM1,"SAT location R",false },
  {0x0019,0x00c3,"GEMS_ACQU_01",VR::SS,VM::VM1,"SAT location L",false },
  {0x0019,0x00c4,"GEMS_ACQU_01",VR::SS,VM::VM1,"SAT location A",false },
  {0x0019,0x00c5,"GEMS_ACQU_01",VR::SS,VM::VM1,"SAT location P",false },
  {0x0019,0x00c6,"GEMS_ACQU_01",VR::SS,VM::VM1,"SAT location H",false },
  {0x0019,0x00c7,"GEMS_ACQU_01",VR::SS,VM::VM1,"SAT location F",false },
  {0x0019,0x00c8,"GEMS_ACQU_01",VR::SS,VM::VM1,"SAT thickness R/L",false },
  {0x0019,0x00c9,"GEMS_ACQU_01",VR::SS,VM::VM1,"SAT thickness A/P",false },
  {0x0019,0x00ca,"GEMS_ACQU_01",VR::SS,VM::VM1,"SAT thickness H/F",false },
  {0x0019,0x00cb,"GEMS_ACQU_01",VR::SS,VM::VM1,"Phase Contrast flow axis",false },
  {0x0019,0x00cc,"GEMS_ACQU_01",VR::SS,VM::VM1,"Velocity encoding",false },
  {0x0019,0x00cd,"GEMS_ACQU_01",VR::SS,VM::VM1,"Thickness disclaimer",false },
  {0x0019,0x00ce,"GEMS_ACQU_01",VR::SS,VM::VM1,"Prescan type",false },
  {0x0019,0x00cf,"GEMS_ACQU_01",VR::SS,VM::VM1,"Prescan status",false },
  {0x0019,0x00d0,"GEMS_ACQU_01",VR::SH,VM::VM1,"Raw data type",false },
  {0x0019,0x00d2,"GEMS_ACQU_01",VR::SS,VM::VM1,"Projection Algorithm",false },
  {0x0019,0x00d3,"GEMS_ACQU_01",VR::SH,VM::VM1,"Projection Algorithm Name",false },
  {0x0019,0x00d4,"GEMS_ACQU_01",VR::US,VM::VM1,"?",false },
  {0x0019,0x00d5,"GEMS_ACQU_01",VR::SS,VM::VM1,"Fractional echo",false },
  {0x0019,0x00d6,"GEMS_ACQU_01",VR::SS,VM::VM1,"Prep pulse",false },
  {0x0019,0x00d7,"GEMS_ACQU_01",VR::SS,VM::VM1,"Cardiac phase number",false },
  {0x0019,0x00d8,"GEMS_ACQU_01",VR::SS,VM::VM1,"Variable echoflag",false },
  {0x0019,0x00d9,"GEMS_ACQU_01",VR::DS,VM::VM1,"Concatenated SAT {# DTI Diffusion Dir., release 9.0 & below}",false },
  {0x0019,0x00da,"GEMS_ACQU_01",VR::SS,VM::VM1,"Reference channel used",false },
  {0x0019,0x00db,"GEMS_ACQU_01",VR::DS,VM::VM1,"Back projector coefficient",false },
  {0x0019,0x00dc,"GEMS_ACQU_01",VR::SS,VM::VM1,"Primary speed correction used",false },
  {0x0019,0x00dd,"GEMS_ACQU_01",VR::SS,VM::VM1,"Overrange correction used",false },
  {0x0019,0x00de,"GEMS_ACQU_01",VR::DS,VM::VM1,"Dynamic Z alpha value",false },
  {0x0019,0x00df,"GEMS_ACQU_01",VR::DS,VM::VM1,"User data 23 {# DTI Diffusion Dir., release 9.0 & below}",false },
  {0x0019,0x00e0,"GEMS_ACQU_01",VR::DS,VM::VM1,"User data 24 {# DTI Diffusion Dir., release 10.0 & above}",false },
  {0x0019,0x00e1,"GEMS_ACQU_01",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00e2,"GEMS_ACQU_01",VR::DS,VM::VM1,"Velocity Encode Scale",false },
  {0x0019,0x00e3,"GEMS_ACQU_01",VR::LT,VM::VM1,"?",false },
  {0x0019,0x00e4,"GEMS_ACQU_01",VR::LT,VM::VM1,"?",false },
  {0x0019,0x00e5,"GEMS_ACQU_01",VR::IS,VM::VM1,"?",false },
  {0x0019,0x00e6,"GEMS_ACQU_01",VR::US,VM::VM1,"?",false },
  {0x0019,0x00e8,"GEMS_ACQU_01",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00e9,"GEMS_ACQU_01",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00eb,"GEMS_ACQU_01",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00ec,"GEMS_ACQU_01",VR::US,VM::VM1,"?",false },
  {0x0019,0x00f0,"GEMS_ACQU_01",VR::UN,VM::VM1,"?",false },
  {0x0019,0x00f1,"GEMS_ACQU_01",VR::LT,VM::VM1,"?",false },
  {0x0019,0x00f2,"GEMS_ACQU_01",VR::SS,VM::VM1,"Fast phases",false },
  {0x0019,0x00f3,"GEMS_ACQU_01",VR::LT,VM::VM1,"?",false },
  {0x0019,0x00f4,"GEMS_ACQU_01",VR::LT,VM::VM1,"?",false },
  {0x0019,0x00f9,"GEMS_ACQU_01",VR::DS,VM::VM1,"Transmit gain",false },
  {0x0023,0x0000,"GEMS_ACRQA_1.0 BLOCK1",VR::LO,VM::VM1,"CR Exposure Menu Code",false },
  {0x0023,0x0010,"GEMS_ACRQA_1.0 BLOCK1",VR::LO,VM::VM1,"CR Exposure Menu String",false },
  {0x0023,0x0020,"GEMS_ACRQA_1.0 BLOCK1",VR::LO,VM::VM1,"CR EDR Mode",false },
  {0x0023,0x0030,"GEMS_ACRQA_1.0 BLOCK1",VR::LO,VM::VM1,"CR Latitude",false },
  {0x0023,0x0040,"GEMS_ACRQA_1.0 BLOCK1",VR::LO,VM::VM1,"CR Group Number",false },
  {0x0023,0x0050,"GEMS_ACRQA_1.0 BLOCK1",VR::US,VM::VM1,"CR Image Serial Number",false },
  {0x0023,0x0060,"GEMS_ACRQA_1.0 BLOCK1",VR::LO,VM::VM1,"CR Bar Code Number",false },
  {0x0023,0x0070,"GEMS_ACRQA_1.0 BLOCK1",VR::LO,VM::VM1,"CR Film Output Exposure",false },
  {0x0023,0x0080,"GEMS_ACRQA_1.0 BLOCK1",VR::LO,VM::VM1,"CR Film Format",false },
  {0x0023,0x0090,"GEMS_ACRQA_1.0 BLOCK1",VR::LO,VM::VM1,"CR S-Shift String",false },
  {0x0023,0x0000,"GEMS_ACRQA_1.0 BLOCK2",VR::US,VM::VM1,"CR S-Shift",false },
  {0x0023,0x0010,"GEMS_ACRQA_1.0 BLOCK2",VR::DS,VM::VM1,"CR C-Shift",false },
  {0x0023,0x0020,"GEMS_ACRQA_1.0 BLOCK2",VR::DS,VM::VM1,"CR GT",false },
  {0x0023,0x0030,"GEMS_ACRQA_1.0 BLOCK2",VR::DS,VM::VM1,"CR GA",false },
  {0x0023,0x0040,"GEMS_ACRQA_1.0 BLOCK2",VR::DS,VM::VM1,"CR GC",false },
  {0x0023,0x0050,"GEMS_ACRQA_1.0 BLOCK2",VR::DS,VM::VM1,"CR GS",false },
  {0x0023,0x0060,"GEMS_ACRQA_1.0 BLOCK2",VR::DS,VM::VM1,"CR RT",false },
  {0x0023,0x0070,"GEMS_ACRQA_1.0 BLOCK2",VR::DS,VM::VM1,"CR RE",false },
  {0x0023,0x0080,"GEMS_ACRQA_1.0 BLOCK2",VR::US,VM::VM1,"CR RN",false },
  {0x0023,0x0090,"GEMS_ACRQA_1.0 BLOCK2",VR::DS,VM::VM1,"CR DRT",false },
  {0x0023,0x0000,"GEMS_ACRQA_1.0 BLOCK3",VR::DS,VM::VM1,"CR DRE",false },
  {0x0023,0x0010,"GEMS_ACRQA_1.0 BLOCK3",VR::US,VM::VM1,"CR DRN",false },
  {0x0023,0x0020,"GEMS_ACRQA_1.0 BLOCK3",VR::DS,VM::VM1,"CR ORE",false },
  {0x0023,0x0030,"GEMS_ACRQA_1.0 BLOCK3",VR::US,VM::VM1,"CR ORN",false },
  {0x0023,0x0040,"GEMS_ACRQA_1.0 BLOCK3",VR::US,VM::VM1,"CR ORD",false },
  {0x0023,0x0050,"GEMS_ACRQA_1.0 BLOCK3",VR::LO,VM::VM1,"CR Cassette Size",false },
  {0x0023,0x0060,"GEMS_ACRQA_1.0 BLOCK3",VR::LO,VM::VM1,"CR Machine ID",false },
  {0x0023,0x0070,"GEMS_ACRQA_1.0 BLOCK3",VR::LO,VM::VM1,"CR Machine Type",false },
  {0x0023,0x0080,"GEMS_ACRQA_1.0 BLOCK3",VR::LO,VM::VM1,"CR Technician Code",false },
  {0x0023,0x0090,"GEMS_ACRQA_1.0 BLOCK3",VR::LO,VM::VM1,"CR ES Parameters",false },
  {0x0023,0x0000,"GEMS_ACRQA_2.0 BLOCK1",VR::LO,VM::VM1,"CR Exposure Menu Code",false },
  {0x0023,0x0010,"GEMS_ACRQA_2.0 BLOCK1",VR::LO,VM::VM1,"CR Exposure Menu String",false },
  {0x0023,0x0020,"GEMS_ACRQA_2.0 BLOCK1",VR::LO,VM::VM1,"CR EDR Mode",false },
  {0x0023,0x0030,"GEMS_ACRQA_2.0 BLOCK1",VR::LO,VM::VM1,"CR Latitude",false },
  {0x0023,0x0040,"GEMS_ACRQA_2.0 BLOCK1",VR::LO,VM::VM1,"CR Group Number",false },
  {0x0023,0x0050,"GEMS_ACRQA_2.0 BLOCK1",VR::US,VM::VM1,"CR Image Serial Number",false },
  {0x0023,0x0060,"GEMS_ACRQA_2.0 BLOCK1",VR::LO,VM::VM1,"CR Bar Code Number",false },
  {0x0023,0x0070,"GEMS_ACRQA_2.0 BLOCK1",VR::LO,VM::VM1,"CR Film Output Exposure",false },
  {0x0023,0x0080,"GEMS_ACRQA_2.0 BLOCK1",VR::LO,VM::VM1,"CR Film Format",false },
  {0x0023,0x0090,"GEMS_ACRQA_2.0 BLOCK1",VR::LO,VM::VM1,"CR S Shift String",false },
  {0x0023,0x0000,"GEMS_ACRQA_2.0 BLOCK2",VR::US,VM::VM1,"CR S Shift",false },
  {0x0023,0x0010,"GEMS_ACRQA_2.0 BLOCK2",VR::LO,VM::VM1,"CR C Shift",false },
  {0x0023,0x0020,"GEMS_ACRQA_2.0 BLOCK2",VR::LO,VM::VM1,"CR GT",false },
  {0x0023,0x0030,"GEMS_ACRQA_2.0 BLOCK2",VR::DS,VM::VM1,"CR GA",false },
  {0x0023,0x0040,"GEMS_ACRQA_2.0 BLOCK2",VR::DS,VM::VM1,"CR GC",false },
  {0x0023,0x0050,"GEMS_ACRQA_2.0 BLOCK2",VR::DS,VM::VM1,"CR GS",false },
  {0x0023,0x0060,"GEMS_ACRQA_2.0 BLOCK2",VR::LO,VM::VM1,"CR RT",false },
  {0x0023,0x0070,"GEMS_ACRQA_2.0 BLOCK2",VR::DS,VM::VM1,"CR RE",false },
  {0x0023,0x0080,"GEMS_ACRQA_2.0 BLOCK2",VR::US,VM::VM1,"CR RN",false },
  {0x0023,0x0090,"GEMS_ACRQA_2.0 BLOCK2",VR::DS,VM::VM1,"CR DRT",false },
  {0x0023,0x0000,"GEMS_ACRQA_2.0 BLOCK3",VR::DS,VM::VM1,"CR DRE",false },
  {0x0023,0x0010,"GEMS_ACRQA_2.0 BLOCK3",VR::US,VM::VM1,"CR DRN",false },
  {0x0023,0x0020,"GEMS_ACRQA_2.0 BLOCK3",VR::DS,VM::VM1,"CR ORE",false },
  {0x0023,0x0030,"GEMS_ACRQA_2.0 BLOCK3",VR::US,VM::VM1,"CR ORN",false },
  {0x0023,0x0040,"GEMS_ACRQA_2.0 BLOCK3",VR::US,VM::VM1,"CR ORD",false },
  {0x0023,0x0050,"GEMS_ACRQA_2.0 BLOCK3",VR::LO,VM::VM1,"CR Cassette Size",false },
  {0x0023,0x0060,"GEMS_ACRQA_2.0 BLOCK3",VR::LO,VM::VM1,"CR Machine ID",false },
  {0x0023,0x0070,"GEMS_ACRQA_2.0 BLOCK3",VR::LO,VM::VM1,"CR Machine Type",false },
  {0x0023,0x0080,"GEMS_ACRQA_2.0 BLOCK3",VR::LO,VM::VM1,"CR Technician Code",false },
  {0x0023,0x0090,"GEMS_ACRQA_2.0 BLOCK3",VR::LO,VM::VM1,"CR Energy Subtraction Parameters",false },
  {0x0023,0x00f0,"GEMS_ACRQA_2.0 BLOCK3",VR::LO,VM::VM1,"CR Distribution Code",false },
  {0x0023,0x00ff,"GEMS_ACRQA_2.0 BLOCK3",VR::US,VM::VM1,"CR Shutters Applied",false },
  {0x0047,0x0001,"GEMS_ADWSoft_3D1",VR::SQ,VM::VM1,"Reconstruction Parameters Sequence",false },
  {0x0047,0x0049,"GEMS_ADWSoft_3D1",VR::UL,VM::VM3_n,"Volume Color",false },
  {0x0047,0x0050,"GEMS_ADWSoft_3D1",VR::UL,VM::VM1,"Volume Voxel Count",false },
  {0x0047,0x0051,"GEMS_ADWSoft_3D1",VR::UL,VM::VM1_n,"Volume Segment Count",false },
  {0x0047,0x0053,"GEMS_ADWSoft_3D1",VR::US,VM::VM1,"Volume Slice Size",false },
  {0x0047,0x0054,"GEMS_ADWSoft_3D1",VR::US,VM::VM1,"Volume Slice Count",false },
  {0x0047,0x0055,"GEMS_ADWSoft_3D1",VR::SL,VM::VM1,"Volume Threshold Value",false },
  {0x0047,0x0057,"GEMS_ADWSoft_3D1",VR::DS,VM::VM1,"Volume Voxel Ratio",false },
  {0x0047,0x0058,"GEMS_ADWSoft_3D1",VR::DS,VM::VM1,"Volume Voxel Size",false },
  {0x0047,0x0059,"GEMS_ADWSoft_3D1",VR::US,VM::VM1,"Volume Z Position Size",false },
  {0x0047,0x0060,"GEMS_ADWSoft_3D1",VR::DS,VM::VM9,"Volume Base Line",false },
  {0x0047,0x0061,"GEMS_ADWSoft_3D1",VR::DS,VM::VM3,"Volume Center Point",false },
  {0x0047,0x0063,"GEMS_ADWSoft_3D1",VR::SL,VM::VM1,"Volume Skew Base",false },
  {0x0047,0x0064,"GEMS_ADWSoft_3D1",VR::DS,VM::VM9,"Volume Registration Transform Rotation Matrix",false },
  {0x0047,0x0065,"GEMS_ADWSoft_3D1",VR::DS,VM::VM3,"Volume Registration Transform Translation Vector",false },
  {0x0047,0x0066,"GEMS_ADWSoft_3D1",VR::DS,VM::VM1,"Volume Tilt",false },
  {0x0047,0x0070,"GEMS_ADWSoft_3D1",VR::DS,VM::VM1_n,"KPV List",false },
  {0x0047,0x0071,"GEMS_ADWSoft_3D1",VR::IS,VM::VM1_n,"X-Ray Tube Current List",false },
  {0x0047,0x0072,"GEMS_ADWSoft_3D1",VR::IS,VM::VM1_n,"Exposure List",false },
  {0x0047,0x0080,"GEMS_ADWSoft_3D1",VR::LO,VM::VM1,"Acquisition DLX Identifier",false },
  {0x0047,0x0081,"GEMS_ADWSoft_3D1",VR::IS,VM::VM1,"Acquisition DLX 2D Series Count",false },
  {0x0047,0x0085,"GEMS_ADWSoft_3D1",VR::SQ,VM::VM1,"Acquisition DLX 2D Series Sequence",false },
  {0x0047,0x0089,"GEMS_ADWSoft_3D1",VR::DS,VM::VM1_n,"Contrast Agent Volume List",false },
  {0x0047,0x008a,"GEMS_ADWSoft_3D1",VR::US,VM::VM1,"Number Of Injections",false },
  {0x0047,0x008b,"GEMS_ADWSoft_3D1",VR::US,VM::VM1,"Frame Count",false },
  {0x0047,0x0091,"GEMS_ADWSoft_3D1",VR::LO,VM::VM1,"XA 3D Reconstruction Algorithm Name",false },
  {0x0047,0x0092,"GEMS_ADWSoft_3D1",VR::CS,VM::VM1,"XA 3D Reconstruction Algorithm Version",false },
  {0x0047,0x0093,"GEMS_ADWSoft_3D1",VR::DA,VM::VM1,"DLX Calibration Date",false },
  {0x0047,0x0094,"GEMS_ADWSoft_3D1",VR::TM,VM::VM1,"DLX Calibration Time",false },
  {0x0047,0x0095,"GEMS_ADWSoft_3D1",VR::CS,VM::VM1,"DLX Calibration Status",false },
  {0x0047,0x0096,"GEMS_ADWSoft_3D1",VR::IS,VM::VM1_n,"Used Frames",false },
  {0x0047,0x0098,"GEMS_ADWSoft_3D1",VR::US,VM::VM1,"Transform Count",false },
  {0x0047,0x0099,"GEMS_ADWSoft_3D1",VR::SQ,VM::VM1,"Transform Sequence",false },
  {0x0047,0x009a,"GEMS_ADWSoft_3D1",VR::DS,VM::VM9,"Transform Rotation Matrix",false },
  {0x0047,0x009b,"GEMS_ADWSoft_3D1",VR::DS,VM::VM3,"Transform Translation Vector",false },
  {0x0047,0x009c,"GEMS_ADWSoft_3D1",VR::LO,VM::VM1,"Transform Label",false },
  {0x0047,0x00b0,"GEMS_ADWSoft_3D1",VR::SQ,VM::VM1,"Wireframe List",false },
  {0x0047,0x00b1,"GEMS_ADWSoft_3D1",VR::US,VM::VM1,"Wireframe Count",false },
  {0x0047,0x00b2,"GEMS_ADWSoft_3D1",VR::US,VM::VM1,"Location System",false },
  {0x0047,0x00b5,"GEMS_ADWSoft_3D1",VR::LO,VM::VM1,"Wireframe Name",false },
  {0x0047,0x00b6,"GEMS_ADWSoft_3D1",VR::LO,VM::VM1,"Wireframe Group Name",false },
  {0x0047,0x00b7,"GEMS_ADWSoft_3D1",VR::LO,VM::VM1,"Wireframe Color",false },
  {0x0047,0x00b8,"GEMS_ADWSoft_3D1",VR::SL,VM::VM1,"Wireframe Attributes",false },
  {0x0047,0x00b9,"GEMS_ADWSoft_3D1",VR::SL,VM::VM1,"Wireframe Point Count",false },
  {0x0047,0x00ba,"GEMS_ADWSoft_3D1",VR::SL,VM::VM1,"Wireframe Timestamp",false },
  {0x0047,0x00bb,"GEMS_ADWSoft_3D1",VR::SQ,VM::VM1,"Wireframe Point List",false },
  {0x0047,0x00bc,"GEMS_ADWSoft_3D1",VR::DS,VM::VM3,"Wireframe Points Coordinates",false },
  {0x0047,0x00c0,"GEMS_ADWSoft_3D1",VR::DS,VM::VM3,"Volume Upper Left High Corner RAS",false },
  {0x0047,0x00c1,"GEMS_ADWSoft_3D1",VR::DS,VM::VM9,"Volume Slice To RAS Rotation Matrix",false },
  {0x0047,0x00c2,"GEMS_ADWSoft_3D1",VR::DS,VM::VM1,"Volume Upper Left High Corner TLOC",false },
  {0x0047,0x00d1,"GEMS_ADWSoft_3D1",VR::OB,VM::VM1,"Volume Segment List",false },
  {0x0047,0x00d2,"GEMS_ADWSoft_3D1",VR::OB,VM::VM1,"Volume Gradient List",false },
  {0x0047,0x00d3,"GEMS_ADWSoft_3D1",VR::OB,VM::VM1,"Volume Density List",false },
  {0x0047,0x00d4,"GEMS_ADWSoft_3D1",VR::OB,VM::VM1,"Volume Z Position List",false },
  {0x0047,0x00d5,"GEMS_ADWSoft_3D1",VR::OB,VM::VM1,"Volume Original Index List",false },
  {0x0047,0x00f4,"GEMS_ADWSoft_3D1",VR::LO,VM::VM1_n,"Volume Name(s)",false },
  {0x0047,0x00f5,"GEMS_ADWSoft_3D1",VR::DS,VM::VM1_n,"Min original density",false },
  {0x0047,0x00f6,"GEMS_ADWSoft_3D1",VR::DS,VM::VM1_n,"Max original density",false },
  {0x0047,0x00f7,"GEMS_ADWSoft_3D1",VR::DS,VM::VM1_n,"Min Converted Density",false },
  {0x0047,0x00f8,"GEMS_ADWSoft_3D1",VR::DS,VM::VM1_n,"Max Converted Density",false },
  {0x0047,0x00f9,"GEMS_ADWSoft_3D1",VR::LO,VM::VM1,"Protocol Film Name",false },
  {0x0047,0x00fa,"GEMS_ADWSoft_3D1",VR::US,VM::VM1,"Protocol Resolution",false },
  {0x0047,0x00fb,"GEMS_ADWSoft_3D1",VR::US,VM::VM1,"Phase Number (percent)",false },
  {0x0047,0x00fc,"GEMS_ADWSoft_3D1",VR::OB,VM::VM1,"Volume midscan times list",false },
  {0x0047,0x00fd,"GEMS_ADWSoft_3D1",VR::OB,VM::VM1,"Volume Registered Phases List",false },
  {0x0047,0x00fe,"GEMS_ADWSoft_3D1",VR::LO,VM::VM1,"Protocol Name",false },
  {0x0047,0x00ff,"GEMS_ADWSoft_3D1",VR::LO,VM::VM1,"Protocol Title",false },
  {0x0057,0x0001,"GEMS_ADWSoft_3D2",VR::OB,VM::VM1,"Cardiac Reconstruction Algorithm List",false },
  {0x0057,0x0002,"GEMS_ADWSoft_3D2",VR::OB,VM::VM1,"Average Heart Rate for Image List",false },
  {0x0057,0x0003,"GEMS_ADWSoft_3D2",VR::OB,VM::VM1,"Temporal Resolution List",false },
  {0x0057,0x0004,"GEMS_ADWSoft_3D2",VR::LT,VM::VM1,"Layout View Preset",false },
  {0x0039,0x0080,"GEMS_ADWSoft_DPO",VR::IS,VM::VM1,"Private Entity Number",false },
  {0x0039,0x0085,"GEMS_ADWSoft_DPO",VR::DA,VM::VM1,"Private Entity Date",false },
  {0x0039,0x0090,"GEMS_ADWSoft_DPO",VR::TM,VM::VM1,"Private Entity Time",false },
  {0x0039,0x0095,"GEMS_ADWSoft_DPO",VR::LO,VM::VM1,"Private Entity Launch Command",false },
  {0x0039,0x00aa,"GEMS_ADWSoft_DPO",VR::CS,VM::VM1,"Private Entity Type",false },
  {0x0039,0x0080,"GEMS_ADWSoft_DPO1",VR::IS,VM::VM1,"Private Entity Number",false },
  {0x0039,0x0085,"GEMS_ADWSoft_DPO1",VR::DA,VM::VM1,"Private Entity Date",false },
  {0x0039,0x0090,"GEMS_ADWSoft_DPO1",VR::TM,VM::VM1,"Private Entity Time",false },
  {0x0039,0x0095,"GEMS_ADWSoft_DPO1",VR::LO,VM::VM1,"Private Entity Launch Command",false },
  {0x0039,0x00aa,"GEMS_ADWSoft_DPO1",VR::CS,VM::VM1,"Private Entity Type",false },
  {0x0039,0x0065,"GEMS_AWSOFT_CD1",VR::UI,VM::VM1,"Reference to Study UID",false },
  {0x0039,0x0070,"GEMS_AWSOFT_CD1",VR::UI,VM::VM1,"Reference to Series UID",false },
  {0x0039,0x0075,"GEMS_AWSOFT_CD1",VR::IS,VM::VM1,"Reference to Original Instance",false },
  {0x0039,0x0080,"GEMS_AWSOFT_CD1",VR::IS,VM::VM1,"DPO Number",false },
  {0x0039,0x0085,"GEMS_AWSOFT_CD1",VR::DA,VM::VM1,"DPO Date",false },
  {0x0039,0x0090,"GEMS_AWSOFT_CD1",VR::TM,VM::VM1,"DPO Time",false },
  {0x0039,0x0095,"GEMS_AWSOFT_CD1",VR::LO,VM::VM1,"DPO Invocation String",false },
  {0x0039,0x00aa,"GEMS_AWSOFT_CD1",VR::CS,VM::VM1,"DPO type",false },
  {0x0039,0x00ff,"GEMS_AWSOFT_CD1",VR::OB,VM::VM1,"DPO data",false },
  {0x0039,0x0050,"GEMS_AWSoft_SB1",VR::UI,VM::VM1,"Reference to Study UID",false },
  {0x0039,0x0051,"GEMS_AWSoft_SB1",VR::UI,VM::VM1,"Reference to Series UID",false },
  {0x0039,0x0052,"GEMS_AWSoft_SB1",VR::IS,VM::VM1,"Reference to Original Instance Number",false },
  {0x0039,0x0095,"GEMS_AWSoft_SB1",VR::LO,VM::VM1,"Private Entity Launch Command",false },
  {0x0033,0x0002,"GEMS_CTHD_01",VR::UN,VM::VM1,"?",false },
  {0x0049,0x0001,"GEMS_CT_CARDIAC_001",VR::SQ,VM::VM1,"CT Cardiac Sequence",false },
  {0x0049,0x0002,"GEMS_CT_CARDIAC_001",VR::CS,VM::VM1,"HeartRateAtConfirm",false },
  {0x0049,0x0003,"GEMS_CT_CARDIAC_001",VR::FL,VM::VM1,"AvgHeartRatePriorToConfirm",false },
  {0x0049,0x0004,"GEMS_CT_CARDIAC_001",VR::CS,VM::VM1,"MinHeartRatePriorToConfirm",false },
  {0x0049,0x0005,"GEMS_CT_CARDIAC_001",VR::CS,VM::VM1,"MaxHeartRatePriorToConfirm",false },
  {0x0049,0x0006,"GEMS_CT_CARDIAC_001",VR::FL,VM::VM1,"StdDevHeartRatePriorToConfirm",false },
  {0x0049,0x0007,"GEMS_CT_CARDIAC_001",VR::US,VM::VM1,"NumHeartRateSamplesPriorToConfirm",false },
  {0x0049,0x0008,"GEMS_CT_CARDIAC_001",VR::CS,VM::VM1,"AutoHeartRateDetectPredict",false },
  {0x0049,0x0009,"GEMS_CT_CARDIAC_001",VR::CS,VM::VM1,"SystemOptimizedHeartRate",false },
  {0x0049,0x000a,"GEMS_CT_CARDIAC_001",VR::ST,VM::VM1,"EkgMonitorType",false },
  {0x0049,0x000b,"GEMS_CT_CARDIAC_001",VR::CS,VM::VM1,"NumReconSectors",false },
  {0x0049,0x000c,"GEMS_CT_CARDIAC_001",VR::FL,VM::VM256,"RpeakTimeStamps",false },
  {0x004b,0x0001,"GEMS_CT_HINO_01",VR::DS,VM::VM1_n,"Beam Thickess",false },
  {0x004b,0x0002,"GEMS_CT_HINO_01",VR::DS,VM::VM1_n,"R Time",false },
  {0x004b,0x0003,"GEMS_CT_HINO_01",VR::IS,VM::VM1,"HBC Number",false },
  {0x0051,0x0001,"GEMS_CT_VES_01",VR::SQ,VM::VM1,"CTVESSequence",false },
  {0x0027,0x0001,"GEMS_DLX_DOSE_01",VR::SQ,VM::VM1,"Private Radiation Dose Sequence",false },
  {0x0027,0x0002,"GEMS_DLX_DOSE_01",VR::IS,VM::VM1,"Run Number",false },
  {0x0027,0x0003,"GEMS_DLX_DOSE_01",VR::TM,VM::VM1,"Run Time",false },
  {0x0027,0x0004,"GEMS_DLX_DOSE_01",VR::IS,VM::VM1,"Number of Frames",false },
  {0x0027,0x0005,"GEMS_DLX_DOSE_01",VR::DS,VM::VM1,"Frames per Second",false },
  {0x0027,0x0006,"GEMS_DLX_DOSE_01",VR::CS,VM::VM1,"Plane",false },
  {0x0027,0x0007,"GEMS_DLX_DOSE_01",VR::DS,VM::VM1,"KV",false },
  {0x0027,0x0008,"GEMS_DLX_DOSE_01",VR::DS,VM::VM1,"MA",false },
  {0x0027,0x0009,"GEMS_DLX_DOSE_01",VR::DS,VM::VM1,"Mas",false },
  {0x0027,0x0010,"GEMS_DLX_DOSE_01",VR::DS,VM::VM1,"Ms",false },
  {0x0027,0x0011,"GEMS_DLX_DOSE_01",VR::DS,VM::VM1,"Angulation",false },
  {0x0027,0x0012,"GEMS_DLX_DOSE_01",VR::DS,VM::VM1,"Rotation",false },
  {0x0027,0x0013,"GEMS_DLX_DOSE_01",VR::DS,VM::VM1,"Focal Distance",false },
  {0x0027,0x0014,"GEMS_DLX_DOSE_01",VR::DS,VM::VM1,"Image Intensifier Mode",false },
  {0x0025,0x0002,"GEMS_DL_FRAME_01",VR::IS,VM::VM1,"Frame Id",false },
  {0x0025,0x0003,"GEMS_DL_FRAME_01",VR::DS,VM::VM1,"Distance Source To Detector",false },
  {0x0025,0x0004,"GEMS_DL_FRAME_01",VR::DS,VM::VM1,"Distance Source To Patient",false },
  {0x0025,0x0005,"GEMS_DL_FRAME_01",VR::DS,VM::VM1,"Distance Source To Skin",false },
  {0x0025,0x0006,"GEMS_DL_FRAME_01",VR::DS,VM::VM1,"Positioner Primary Angle",false },
  {0x0025,0x0007,"GEMS_DL_FRAME_01",VR::DS,VM::VM1,"Positioner Secondary Angle",false },
  {0x0025,0x0008,"GEMS_DL_FRAME_01",VR::IS,VM::VM1,"Beam Orientation",false },
  {0x0025,0x0009,"GEMS_DL_FRAME_01",VR::DS,VM::VM1,"Larm Angle",false },
  {0x0025,0x000a,"GEMS_DL_FRAME_01",VR::SQ,VM::VM1,"Frame Sequence",false },
  {0x0025,0x000b,"GEMS_DL_FRAME_01",VR::DS,VM::VM1,"Target Cnr",false },
  {0x0025,0x0010,"GEMS_DL_FRAME_01",VR::DS,VM::VM1,"Pivot Angle",false },
  {0x0025,0x001a,"GEMS_DL_FRAME_01",VR::DS,VM::VM1,"Arc Angle",false },
  {0x0025,0x001b,"GEMS_DL_FRAME_01",VR::DS,VM::VM1,"Table Vertical Position",false },
  {0x0025,0x001c,"GEMS_DL_FRAME_01",VR::DS,VM::VM1,"Table Longitudinal Position",false },
  {0x0025,0x001d,"GEMS_DL_FRAME_01",VR::DS,VM::VM1,"Table Lateral Position",false },
  {0x0025,0x001e,"GEMS_DL_FRAME_01",VR::IS,VM::VM1,"Beam Cover Area",false },
  {0x0025,0x001f,"GEMS_DL_FRAME_01",VR::DS,VM::VM1,"Kvp Actual",false },
  {0x0025,0x0020,"GEMS_DL_FRAME_01",VR::DS,VM::VM1,"Mas Actual",false },
  {0x0025,0x0021,"GEMS_DL_FRAME_01",VR::DS,VM::VM1,"Pw Actual",false },
  {0x0025,0x0022,"GEMS_DL_FRAME_01",VR::DS,VM::VM1,"Kvp Commanded",false },
  {0x0025,0x0023,"GEMS_DL_FRAME_01",VR::DS,VM::VM1,"Mas Commanded",false },
  {0x0025,0x0024,"GEMS_DL_FRAME_01",VR::DS,VM::VM1,"Pw Commanded",false },
  {0x0025,0x0025,"GEMS_DL_FRAME_01",VR::CS,VM::VM1,"Grid",false },
  {0x0025,0x0026,"GEMS_DL_FRAME_01",VR::DS,VM::VM1,"Sensor Feedback",false },
  {0x0025,0x0027,"GEMS_DL_FRAME_01",VR::DS,VM::VM1,"Target Entrance Dose",false },
  {0x0025,0x0028,"GEMS_DL_FRAME_01",VR::DS,VM::VM1,"Cnr Commanded",false },
  {0x0025,0x0029,"GEMS_DL_FRAME_01",VR::DS,VM::VM1,"Contrast Commanded",false },
  {0x0025,0x002a,"GEMS_DL_FRAME_01",VR::DS,VM::VM1,"Ept Actual",false },
  {0x0025,0x002b,"GEMS_DL_FRAME_01",VR::IS,VM::VM1,"Spectral Filter Znb",false },
  {0x0025,0x002c,"GEMS_DL_FRAME_01",VR::DS,VM::VM1,"Spectral Filter Weight",false },
  {0x0025,0x002d,"GEMS_DL_FRAME_01",VR::DS,VM::VM1,"Spectral Filter Density",false },
  {0x0025,0x002e,"GEMS_DL_FRAME_01",VR::IS,VM::VM1,"Spectral Filter Thickness",false },
  {0x0025,0x002f,"GEMS_DL_FRAME_01",VR::IS,VM::VM1,"Spectral Filter Status",false },
  {0x0025,0x0030,"GEMS_DL_FRAME_01",VR::IS,VM::VM1,"Fov Dimension",false },
  {0x0025,0x0033,"GEMS_DL_FRAME_01",VR::IS,VM::VM1,"Fov Origin",false },
  {0x0025,0x0034,"GEMS_DL_FRAME_01",VR::IS,VM::VM1,"Collimator Left Vertical Edge",false },
  {0x0025,0x0035,"GEMS_DL_FRAME_01",VR::IS,VM::VM1,"Collimator Right Vertical Edge",false },
  {0x0025,0x0036,"GEMS_DL_FRAME_01",VR::IS,VM::VM1,"Collimator Up Horizontal Edge",false },
  {0x0025,0x0037,"GEMS_DL_FRAME_01",VR::IS,VM::VM1,"Collimator Low Horizontal Edge",false },
  {0x0025,0x0038,"GEMS_DL_FRAME_01",VR::IS,VM::VM1,"Vertices Polygonal Collimator",false },
  {0x0025,0x0039,"GEMS_DL_FRAME_01",VR::IS,VM::VM1,"Contour Filter Distance",false },
  {0x0025,0x003a,"GEMS_DL_FRAME_01",VR::UL,VM::VM1,"Contour Filter Angle",false },
  {0x0025,0x003b,"GEMS_DL_FRAME_01",VR::CS,VM::VM1,"Table Rotation Status",false },
  {0x0025,0x003c,"GEMS_DL_FRAME_01",VR::CS,VM::VM1,"Internal Label Frame",false },
  {0x0019,0x000b,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Fov Dimension Double",false },
  {0x0019,0x000c,"GEMS_DL_IMG_01",VR::FL,VM::VM2_2n,"LV Diastolic contour",false },
  {0x0019,0x000d,"GEMS_DL_IMG_01",VR::FL,VM::VM2_2n,"LV Systolic contour",false },
  {0x0019,0x002b,"GEMS_DL_IMG_01",VR::FL,VM::VM1,"Distance to table top",false },
  {0x0019,0x0030,"GEMS_DL_IMG_01",VR::LO,VM::VM1,"Image File Name",false },
  {0x0019,0x0031,"GEMS_DL_IMG_01",VR::IS,VM::VM1,"Default Spatial Filter Family",false },
  {0x0019,0x0032,"GEMS_DL_IMG_01",VR::IS,VM::VM1,"Default Spatial Filter Strength",false },
  {0x0019,0x0033,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Min Saturation Dose",false },
  {0x0019,0x0034,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Detector Gain",false },
  {0x0019,0x0035,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Patient Dose Limit",false },
  {0x0019,0x0036,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Preproc Image Rate Max",false },
  {0x0019,0x0037,"GEMS_DL_IMG_01",VR::CS,VM::VM1,"Sensor Roi Shape",false },
  {0x0019,0x0038,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Sensor Roi x Position",false },
  {0x0019,0x0039,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Sensor Roi y Position",false },
  {0x0019,0x003a,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Sensor Roi x Size",false },
  {0x0019,0x003b,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Sensor Roi y Size",false },
  {0x0019,0x003d,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Noise Sensitivity",false },
  {0x0019,0x003e,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Sharp Sensitivity",false },
  {0x0019,0x003f,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Contrast Sensitivity",false },
  {0x0019,0x0040,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Lag Sensitivity",false },
  {0x0019,0x0041,"GEMS_DL_IMG_01",VR::CS,VM::VM1,"Tube",false },
  {0x0019,0x0042,"GEMS_DL_IMG_01",VR::US,VM::VM1,"Detector Size Rows",false },
  {0x0019,0x0043,"GEMS_DL_IMG_01",VR::US,VM::VM1,"Detector Size Columns",false },
  {0x0019,0x0044,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Min Object Size",false },
  {0x0019,0x0045,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Max Object Size",false },
  {0x0019,0x0046,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Max Object Speed",false },
  {0x0019,0x0047,"GEMS_DL_IMG_01",VR::CS,VM::VM1,"Object Back Motion",false },
  {0x0019,0x0048,"GEMS_DL_IMG_01",VR::UL,VM::VM1,"Exposure Trajectory Family",false },
  {0x0019,0x0049,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Window Time Duration",false },
  {0x0019,0x004a,"GEMS_DL_IMG_01",VR::CS,VM::VM1,"Positioner Angle Display Mode",false },
  {0x0019,0x004b,"GEMS_DL_IMG_01",VR::IS,VM::VM1,"Detector Origin",false },
  {0x0019,0x004c,"GEMS_DL_IMG_01",VR::CS,VM::VM1,"?",false },
  {0x0019,0x004e,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Default Brightness Contrast",false },
  {0x0019,0x004f,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"User Brightness Contrast",false },
  {0x0019,0x0050,"GEMS_DL_IMG_01",VR::IS,VM::VM1,"Source Series Number",false },
  {0x0019,0x0051,"GEMS_DL_IMG_01",VR::IS,VM::VM1,"Source Image Number",false },
  {0x0019,0x0052,"GEMS_DL_IMG_01",VR::IS,VM::VM1,"Source Frame Number",false },
  {0x0019,0x0053,"GEMS_DL_IMG_01",VR::UI,VM::VM1,"Source Series Item Id",false },
  {0x0019,0x0054,"GEMS_DL_IMG_01",VR::UI,VM::VM1,"Source Image Item Id",false },
  {0x0019,0x0055,"GEMS_DL_IMG_01",VR::UI,VM::VM1,"Source Frame Item Id",false },
  {0x0019,0x0080,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Image Dose",false },
  {0x0019,0x0081,"GEMS_DL_IMG_01",VR::US,VM::VM1,"Calibration Frame",false },
  {0x0019,0x0082,"GEMS_DL_IMG_01",VR::CS,VM::VM1,"Calibration Object",false },
  {0x0019,0x0083,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Calibration Object Size mm",false },
  {0x0019,0x0084,"GEMS_DL_IMG_01",VR::FL,VM::VM1,"Calibration Factor",false },
  {0x0019,0x0085,"GEMS_DL_IMG_01",VR::DA,VM::VM1,"Calibration Date",false },
  {0x0019,0x0086,"GEMS_DL_IMG_01",VR::TM,VM::VM1,"Calibration Time",false },
  {0x0019,0x0087,"GEMS_DL_IMG_01",VR::US,VM::VM1,"Calibration Accuracy",false },
  {0x0019,0x0088,"GEMS_DL_IMG_01",VR::CS,VM::VM1,"Calibration Extended",false },
  {0x0019,0x0089,"GEMS_DL_IMG_01",VR::US,VM::VM1,"Calibration Image Original",false },
  {0x0019,0x008a,"GEMS_DL_IMG_01",VR::US,VM::VM1,"Calibration Frame Original",false },
  {0x0019,0x008b,"GEMS_DL_IMG_01",VR::US,VM::VM1,"Calibration Number Of Points Uif",false },
  {0x0019,0x008c,"GEMS_DL_IMG_01",VR::US,VM::VM1,"Calibration Points Row",false },
  {0x0019,0x008d,"GEMS_DL_IMG_01",VR::US,VM::VM1,"Calibration Points Column",false },
  {0x0019,0x008e,"GEMS_DL_IMG_01",VR::FL,VM::VM1,"Calibration Magnification Ratio",false },
  {0x0019,0x008f,"GEMS_DL_IMG_01",VR::LO,VM::VM1,"Calibration Sw Version",false },
  {0x0019,0x0090,"GEMS_DL_IMG_01",VR::LO,VM::VM1,"Extend Calibration Sw Version",false },
  {0x0019,0x0091,"GEMS_DL_IMG_01",VR::IS,VM::VM1,"Calibration Return Code",false },
  {0x0019,0x0092,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Detector Rotation Angle",false },
  {0x0019,0x0093,"GEMS_DL_IMG_01",VR::CS,VM::VM1,"Spatial Change",false },
  {0x0019,0x0094,"GEMS_DL_IMG_01",VR::CS,VM::VM1,"Inconsistent Flag",false },
  {0x0019,0x0095,"GEMS_DL_IMG_01",VR::CS,VM::VM1,"Image Sweep",false },
  {0x0019,0x0096,"GEMS_DL_IMG_01",VR::CS,VM::VM1,"Internal Label Image",false },
  {0x0019,0x0097,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Angle 1 Increment",false },
  {0x0019,0x0098,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Angle 2 Increment",false },
  {0x0019,0x0099,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Angle 3 Increment",false },
  {0x0019,0x009a,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Sensor Feedback",false },
  {0x0019,0x009b,"GEMS_DL_IMG_01",VR::CS,VM::VM1,"Grid",false },
  {0x0019,0x009c,"GEMS_DL_IMG_01",VR::FL,VM::VM1,"Default Mask Pixel Shift",false },
  {0x0019,0x009d,"GEMS_DL_IMG_01",VR::CS,VM::VM1,"Applicable Review Mode",false },
  {0x0019,0x009e,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Log Lut Control Points",false },
  {0x0019,0x009f,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Exp Lut Control Points",false },
  {0x0019,0x00a0,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"ABD Value",false },
  {0x0019,0x00a1,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Sub Window Center",false },
  {0x0019,0x00a2,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Sub Window Width",false },
  {0x0019,0x00a3,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Image Rotation",false },
  {0x0019,0x00a4,"GEMS_DL_IMG_01",VR::CS,VM::VM1,"Auto Injection Enabled",false },
  {0x0019,0x00a5,"GEMS_DL_IMG_01",VR::CS,VM::VM1,"Injection Phase",false },
  {0x0019,0x00a6,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Injection Delay",false },
  {0x0019,0x00a7,"GEMS_DL_IMG_01",VR::IS,VM::VM1,"Reference Injection Frame Number",false },
  {0x0019,0x00a8,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Injection Duration",false },
  {0x0019,0x00a9,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"EPT",false },
  {0x0019,0x00aa,"GEMS_DL_IMG_01",VR::CS,VM::VM1,"Can Downscan 512",false },
  {0x0019,0x00ab,"GEMS_DL_IMG_01",VR::IS,VM::VM1,"Current Spatial Filter Strength",false },
  {0x0019,0x00ac,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Brightness Sensitivity",false },
  {0x0019,0x00ad,"GEMS_DL_IMG_01",VR::DS,VM::VM1,"Exp Lut NOSUB Control Points",false },
  {0x0019,0x00ae,"GEMS_DL_IMG_01",VR::LO,VM::VM1_n,"SUB operator LUTs names",false },
  {0x0019,0x00af,"GEMS_DL_IMG_01",VR::DS,VM::VM1_n,"kVp actual vector",false },
  {0x0019,0x00b0,"GEMS_DL_IMG_01",VR::DS,VM::VM1_n,"mAs actual vector",false },
  {0x0019,0x00b1,"GEMS_DL_IMG_01",VR::LO,VM::VM1,"Acquisition Mode Description",false },
  {0x0019,0x00b2,"GEMS_DL_IMG_01",VR::LO,VM::VM1,"Acquisition Mode Display Label",false },
  {0x0019,0x00b3,"GEMS_DL_IMG_01",VR::LO,VM::VM1,"Acquisition Protocol User Name",false },
  {0x0019,0x00b8,"GEMS_DL_IMG_01",VR::FL,VM::VM1,"Recommended display frame rate float",false },
  {0x0019,0x00b9,"GEMS_DL_IMG_01",VR::FL,VM::VM1_n,"ABD Vector",false },
  {0x0019,0x00ba,"GEMS_DL_IMG_01",VR::CS,VM::VM1,"Acquisition region",false },
  {0x0019,0x00bb,"GEMS_DL_IMG_01",VR::CS,VM::VM1,"Acquisition SUB mode",false },
  {0x0019,0x00bc,"GEMS_DL_IMG_01",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00bd,"GEMS_DL_IMG_01",VR::CS,VM::VM1_n,"Table rotation status vector",false },
  {0x0019,0x00be,"GEMS_DL_IMG_01",VR::FL,VM::VM1_n,"Source to image distance per frame vector",false },
  {0x0019,0x00c2,"GEMS_DL_IMG_01",VR::DS,VM::VM1_n,"pw actual vector",false },
  {0x0019,0x00c4,"GEMS_DL_IMG_01",VR::IS,VM::VM1,"Spectral filter thickness",false },
  {0x0019,0x00c5,"GEMS_DL_IMG_01",VR::FL,VM::VM1,"Preselected pivot rotation speed",false },
  {0x0019,0x00c7,"GEMS_DL_IMG_01",VR::CS,VM::VM1,"Patient position per image",false },
  {0x0019,0x00c8,"GEMS_DL_IMG_01",VR::CS,VM::VM1,"3D structure of interest",false },
  {0x0019,0x00c9,"GEMS_DL_IMG_01",VR::CS,VM::VM1,"3D calibration out of date flag",false },
  {0x0019,0x00ca,"GEMS_DL_IMG_01",VR::IS,VM::VM1,"3Dspin expected number of frames",false },
  {0x0019,0x00d4,"GEMS_DL_IMG_01",VR::FL,VM::VM1,"Detection gain value",false },
  {0x0019,0x00d5,"GEMS_DL_IMG_01",VR::FL,VM::VM1,"mR mAs calibration value",false },
  {0x0019,0x00dc,"GEMS_DL_IMG_01",VR::LO,VM::VM1,"DRM LUT file name",false },
  {0x0019,0x00dd,"GEMS_DL_IMG_01",VR::DS,VM::VM1_n,"DRM Strength",false },
  {0x0019,0x00de,"GEMS_DL_IMG_01",VR::CS,VM::VM1,"Acquisition Plane",false },
  {0x0019,0x00df,"GEMS_DL_IMG_01",VR::FL,VM::VM1_n,"LP off longitudinal position Z",false },
  {0x0019,0x00e0,"GEMS_DL_IMG_01",VR::FL,VM::VM1,"DAP of current record",false },
  {0x0019,0x00e1,"GEMS_DL_IMG_01",VR::FL,VM::VM1,"Pivot Lateral Angle",false },
  {0x0019,0x00e2,"GEMS_DL_IMG_01",VR::FL,VM::VM1,"Carm Lateral Angle",false },
  {0x0019,0x00e3,"GEMS_DL_IMG_01",VR::FL,VM::VM1_n,"Pivot Lateral Angle increment",false },
  {0x0019,0x00e4,"GEMS_DL_IMG_01",VR::FL,VM::VM1_n,"Carm Lateral Angle increment",false },
  {0x0019,0x00e7,"GEMS_DL_IMG_01",VR::FL,VM::VM1,"LP off long pos Z first frame",false },
  {0x0019,0x00e8,"GEMS_DL_IMG_01",VR::FL,VM::VM1_n,"LP off long pos Z increment",false },
  {0x0019,0x00e9,"GEMS_DL_IMG_01",VR::FL,VM::VM1_n,"Source to detector distance per frame vector",false },
  {0x0019,0x00ea,"GEMS_DL_IMG_01",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00eb,"GEMS_DL_IMG_01",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00ec,"GEMS_DL_IMG_01",VR::FL,VM::VM1,"?",false },
  {0x0019,0x00ed,"GEMS_DL_IMG_01",VR::FL,VM::VM1,"?",false },
  {0x0011,0x0080,"GEMS_DL_PATNT_01",VR::UI,VM::VM1,"Patient Instance Uid",false },
  {0x0011,0x0081,"GEMS_DL_PATNT_01",VR::IS,VM::VM1,"Last Study Number",false },
  {0x0011,0x0082,"GEMS_DL_PATNT_01",VR::CS,VM::VM1,"Patient Repaired",false },
  {0x0011,0x0083,"GEMS_DL_PATNT_01",VR::CS,VM::VM1,"Lock Demographics",false },
  {0x0015,0x0087,"GEMS_DL_SERIES",VR::IS,VM::VM1,"Number of images",false },
  {0x0019,0x000b,"GEMS_DL_SERIES",VR::DS,VM::VM1_2,"fov dimension double",false },
  {0x0019,0x0031,"GEMS_DL_SERIES",VR::IS,VM::VM1,"default spatial filter family",false },
  {0x0019,0x0032,"GEMS_DL_SERIES",VR::IS,VM::VM1,"default spatial filter strength",false },
  {0x0019,0x004b,"GEMS_DL_SERIES",VR::IS,VM::VM2,"Detector origin",false },
  {0x0019,0x004e,"GEMS_DL_SERIES",VR::DS,VM::VM2,"def_bright_contr",false },
  {0x0019,0x004f,"GEMS_DL_SERIES",VR::DS,VM::VM2,"user_bright_contr",false },
  {0x0019,0x0081,"GEMS_DL_SERIES",VR::US,VM::VM1,"Calibration frame",false },
  {0x0019,0x0082,"GEMS_DL_SERIES",VR::CS,VM::VM1,"Calibration object",false },
  {0x0019,0x0083,"GEMS_DL_SERIES",VR::DS,VM::VM1,"Calibration object size mm",false },
  {0x0019,0x0084,"GEMS_DL_SERIES",VR::FL,VM::VM1,"Calibration factor",false },
  {0x0019,0x0085,"GEMS_DL_SERIES",VR::DA,VM::VM1,"Calibration date",false },
  {0x0019,0x0086,"GEMS_DL_SERIES",VR::TM,VM::VM1,"Calibration time",false },
  {0x0019,0x0087,"GEMS_DL_SERIES",VR::US,VM::VM1,"Calibration accuracy",false },
  {0x0019,0x0088,"GEMS_DL_SERIES",VR::CS,VM::VM1,"Calibration extended",false },
  {0x0019,0x0089,"GEMS_DL_SERIES",VR::US,VM::VM1,"Calibration image original",false },
  {0x0019,0x008a,"GEMS_DL_SERIES",VR::US,VM::VM1,"Calibration frame original",false },
  {0x0019,0x008b,"GEMS_DL_SERIES",VR::US,VM::VM1,"Calibration nb points uif",false },
  {0x0019,0x008c,"GEMS_DL_SERIES",VR::US,VM::VM1_n,"Calibration points row",false },
  {0x0019,0x008d,"GEMS_DL_SERIES",VR::US,VM::VM1_n,"Calibration points column",false },
  {0x0019,0x008e,"GEMS_DL_SERIES",VR::FL,VM::VM1,"Calibration mag ratio",false },
  {0x0019,0x008f,"GEMS_DL_SERIES",VR::LO,VM::VM1,"Calibration sw version",false },
  {0x0019,0x0090,"GEMS_DL_SERIES",VR::LO,VM::VM1,"Extend calib sw version",false },
  {0x0019,0x0091,"GEMS_DL_SERIES",VR::IS,VM::VM1,"Calibration return code",false },
  {0x0019,0x0092,"GEMS_DL_SERIES",VR::DS,VM::VM1,"detector_rot_angle",false },
  {0x0019,0x0093,"GEMS_DL_SERIES",VR::CS,VM::VM1,"Spatial change",false },
  {0x0019,0x0095,"GEMS_DL_SERIES",VR::CS,VM::VM2,"Image sweep",false },
  {0x0019,0x0097,"GEMS_DL_SERIES",VR::DS,VM::VM1_n,"Angle 1 increment",false },
  {0x0019,0x0098,"GEMS_DL_SERIES",VR::DS,VM::VM1_n,"Angle 2 increment",false },
  {0x0019,0x0099,"GEMS_DL_SERIES",VR::DS,VM::VM1_n,"Angle 3 increment",false },
  {0x0019,0x009a,"GEMS_DL_SERIES",VR::DS,VM::VM1_n,"sensor feedback",false },
  {0x0019,0x009d,"GEMS_DL_SERIES",VR::CS,VM::VM1,"applicable review mode",false },
  {0x0019,0x009e,"GEMS_DL_SERIES",VR::DS,VM::VM1_n,"log LUT control points",false },
  {0x0019,0x009f,"GEMS_DL_SERIES",VR::DS,VM::VM1_n,"exp LUT SUB control points",false },
  {0x0019,0x00a0,"GEMS_DL_SERIES",VR::DS,VM::VM1,"ABD value",false },
  {0x0019,0x00a1,"GEMS_DL_SERIES",VR::DS,VM::VM1,"Sub window center",false },
  {0x0019,0x00a2,"GEMS_DL_SERIES",VR::DS,VM::VM1,"Sub window width",false },
  {0x0019,0x00a4,"GEMS_DL_SERIES",VR::CS,VM::VM1,"Auto injection enabled",false },
  {0x0019,0x00a5,"GEMS_DL_SERIES",VR::CS,VM::VM1,"injection phase",false },
  {0x0019,0x00a6,"GEMS_DL_SERIES",VR::DS,VM::VM1,"injection delay",false },
  {0x0019,0x00a7,"GEMS_DL_SERIES",VR::IS,VM::VM1,"reference injection frame number",false },
  {0x0019,0x00a9,"GEMS_DL_SERIES",VR::DS,VM::VM1_n,"EPT",false },
  {0x0019,0x00aa,"GEMS_DL_SERIES",VR::CS,VM::VM1,"can downscan 512",false },
  {0x0019,0x00ab,"GEMS_DL_SERIES",VR::IS,VM::VM1,"current spatial filter strength",false },
  {0x0019,0x00ad,"GEMS_DL_SERIES",VR::DS,VM::VM1_n,"exp LUT NOSUB control points",false },
  {0x0019,0x00ae,"GEMS_DL_SERIES",VR::LO,VM::VM1_n,"SUB operator LUTs names",false },
  {0x0019,0x00c4,"GEMS_DL_SERIES",VR::IS,VM::VM1,"Default spatial filter family",false },
  {0x0019,0x00c5,"GEMS_DL_SERIES",VR::FL,VM::VM1,"Preselected pivot rotation speed",false },
  {0x0019,0x00c7,"GEMS_DL_SERIES",VR::CS,VM::VM1,"Patient position per image",false },
  {0x0019,0x00c8,"GEMS_DL_SERIES",VR::CS,VM::VM1,"3D structure of interest",false },
  {0x0019,0x00c9,"GEMS_DL_SERIES",VR::CS,VM::VM1,"3D calibration out of date flag",false },
  {0x0019,0x00ca,"GEMS_DL_SERIES",VR::IS,VM::VM1,"3Dspin expected number of frames",false },
  {0x0019,0x00d4,"GEMS_DL_SERIES",VR::FL,VM::VM1,"Detection gain value",false },
  {0x0019,0x00d5,"GEMS_DL_SERIES",VR::FL,VM::VM1,"mR mAs calibration value",false },
  {0x0019,0x00dc,"GEMS_DL_SERIES",VR::LO,VM::VM1,"DRM LUT file name",false },
  {0x0019,0x00dd,"GEMS_DL_SERIES",VR::DS,VM::VM1_n,"DRM Strength",false },
  {0x0015,0x0085,"GEMS_DL_SERIES_01",VR::LO,VM::VM1,"Series File Name",false },
  {0x0015,0x0087,"GEMS_DL_SERIES_01",VR::IS,VM::VM1,"Number Of Images",false },
  {0x0015,0x008c,"GEMS_DL_SERIES_01",VR::CS,VM::VM1,"Sent Flag",false },
  {0x0015,0x008d,"GEMS_DL_SERIES_01",VR::US,VM::VM1,"Item Locked",false },
  {0x0019,0x004c,"GEMS_DL_SERIES_01",VR::CS,VM::VM1,"internal label",false },
  {0x0019,0x004d,"GEMS_DL_SERIES_01",VR::CS,VM::VM1,"browser hide",false },
  {0x0015,0x0080,"GEMS_DL_STUDY_01",VR::DS,VM::VM1,"Study Dose",false },
  {0x0015,0x0081,"GEMS_DL_STUDY_01",VR::DS,VM::VM1,"Study Total Dap",false },
  {0x0015,0x0082,"GEMS_DL_STUDY_01",VR::DS,VM::VM1,"Study Fluoro Dap",false },
  {0x0015,0x0083,"GEMS_DL_STUDY_01",VR::IS,VM::VM1,"Study Fluoro Time",false },
  {0x0015,0x0084,"GEMS_DL_STUDY_01",VR::DS,VM::VM1,"Study Record Dap",false },
  {0x0015,0x0085,"GEMS_DL_STUDY_01",VR::IS,VM::VM1,"Study Record Time",false },
  {0x0015,0x0086,"GEMS_DL_STUDY_01",VR::IS,VM::VM1,"Last Xa Number",false },
  {0x0015,0x0088,"GEMS_DL_STUDY_01",VR::PN,VM::VM1,"Default Operator Name",false },
  {0x0015,0x0089,"GEMS_DL_STUDY_01",VR::PN,VM::VM1,"Default Performing Physician Name",false },
  {0x0015,0x008a,"GEMS_DL_STUDY_01",VR::CS,VM::VM1,"Default Patient Orientation",false },
  {0x0015,0x008b,"GEMS_DL_STUDY_01",VR::IS,VM::VM1,"Last Sc Number",false },
  {0x0015,0x008e,"GEMS_DL_STUDY_01",VR::UI,VM::VM1,"Common Series Instance Uid",false },
  {0x0015,0x008f,"GEMS_DL_STUDY_01",VR::IS,VM::VM1,"Study Number",false },
  {0x0015,0x0092,"GEMS_DL_STUDY_01",VR::FL,VM::VM1,"Study dose Frontal",false },
  {0x0015,0x0093,"GEMS_DL_STUDY_01",VR::FL,VM::VM1,"Study total dap Frontal",false },
  {0x0015,0x0094,"GEMS_DL_STUDY_01",VR::FL,VM::VM1,"Study fluoro dap frontal",false },
  {0x0015,0x0095,"GEMS_DL_STUDY_01",VR::IS,VM::VM1,"Study fluoro time frontal",false },
  {0x0015,0x0096,"GEMS_DL_STUDY_01",VR::FL,VM::VM1,"Study record dap frontal",false },
  {0x0015,0x0097,"GEMS_DL_STUDY_01",VR::IS,VM::VM1,"Study record time frontal",false },
  {0x0015,0x0098,"GEMS_DL_STUDY_01",VR::FL,VM::VM1,"Study dose lateral",false },
  {0x0015,0x0099,"GEMS_DL_STUDY_01",VR::FL,VM::VM1,"Study total dap lateral",false },
  {0x0015,0x009a,"GEMS_DL_STUDY_01",VR::FL,VM::VM1,"Study fluoro dap lateral",false },
  {0x0015,0x009b,"GEMS_DL_STUDY_01",VR::IS,VM::VM1,"Study fluoro time lateral",false },
  {0x0015,0x009c,"GEMS_DL_STUDY_01",VR::FL,VM::VM1,"Study record dap leteral",false },
  {0x0015,0x009d,"GEMS_DL_STUDY_01",VR::IS,VM::VM1,"Study record time lateral",false },
  {0x0037,0x0010,"GEMS_DRS_1",VR::LO,VM::VM1,"Referring Department",false },
  {0x0037,0x0020,"GEMS_DRS_1",VR::US,VM::VM1,"Screen Number",false },
  {0x0037,0x0040,"GEMS_DRS_1",VR::SH,VM::VM1,"Left Orientation",false },
  {0x0037,0x0042,"GEMS_DRS_1",VR::SH,VM::VM1,"Right Orientation",false },
  {0x0037,0x0050,"GEMS_DRS_1",VR::CS,VM::VM1,"Inversion",false },
  {0x0037,0x0060,"GEMS_DRS_1",VR::US,VM::VM1,"DSA",false },
  {0x0045,0x0055,"GEMS_FALCON_03",VR::DS,VM::VM8,"A_Coefficients",false },
  {0x0045,0x0062,"GEMS_FALCON_03",VR::IS,VM::VM1,"User Window Center",false },
  {0x0045,0x0063,"GEMS_FALCON_03",VR::IS,VM::VM1,"User Window Width",false },
  {0x0045,0x0065,"GEMS_FALCON_03",VR::IS,VM::VM1,"Requested Detector Entrance Dose",false },
  {0x0045,0x0067,"GEMS_FALCON_03",VR::DS,VM::VM3,"VOI LUT Asymmetry parameter beta",false },
  {0x0045,0x0069,"GEMS_FALCON_03",VR::IS,VM::VM1,"Collimator rotation",false },
  {0x0045,0x0072,"GEMS_FALCON_03",VR::DS,VM::VM1,"Collimator Width",false },
  {0x0045,0x0073,"GEMS_FALCON_03",VR::DS,VM::VM1,"Collimator Height",false },
  {0x0051,0x0001,"GEMS_FUNCTOOL_01",VR::LO,VM::VM1,"Functional Proc Group Name",false },
  {0x0051,0x0002,"GEMS_FUNCTOOL_01",VR::LO,VM::VM1,"Functional Processing Name",false },
  {0x0051,0x0003,"GEMS_FUNCTOOL_01",VR::SL,VM::VM1,"Bias of Functional Image",false },
  {0x0051,0x0004,"GEMS_FUNCTOOL_01",VR::FL,VM::VM1,"Scale of Functional Image",false },
  {0x0051,0x0005,"GEMS_FUNCTOOL_01",VR::SL,VM::VM1,"Length of Parameters String",false },
  {0x0051,0x0006,"GEMS_FUNCTOOL_01",VR::LT,VM::VM1,"Store Parameters string, delimited by character ESC=0x1B (27)",false },
  {0x0051,0x0007,"GEMS_FUNCTOOL_01",VR::LO,VM::VM1,"Functional Image Version",false },
  {0x0051,0x0008,"GEMS_FUNCTOOL_01",VR::SL,VM::VM1,"Store Color Ramp",false },
  {0x0051,0x0009,"GEMS_FUNCTOOL_01",VR::SL,VM::VM1,"Store Width of Functional Image",false },
  {0x0051,0x000a,"GEMS_FUNCTOOL_01",VR::SL,VM::VM1,"Store level of Functional Image",false },
  {0x0051,0x000b,"GEMS_FUNCTOOL_01",VR::FL,VM::VM1,"Store B-Value with Functional Image",false },
  {0x0051,0x000c,"GEMS_FUNCTOOL_01",VR::LO,VM::VM1,"Analysis Package",false },
  {0x0051,0x000e,"GEMS_FUNCTOOL_01",VR::SL,VM::VM1,"?",false },
  {0x7fdf,0x0010,"GEMS_GDXE_ATHENAV2_INTERNAL_USE",VR::LT,VM::VM1,"PPS Stream",false },
  {0x7fdf,0x0011,"GEMS_GDXE_ATHENAV2_INTERNAL_USE",VR::LT,VM::VM1,"Pixel Data References (temporary)",false },
  {0x7fdf,0x0020,"GEMS_GDXE_ATHENAV2_INTERNAL_USE",VR::SS,VM::VM1,"Auto Push Tag",false },
  {0x7fdf,0x0025,"GEMS_GDXE_ATHENAV2_INTERNAL_USE",VR::CS,VM::VM1,"PPS Status",false },
  {0x0011,0x0003,"GEMS_GDXE_FALCON_04",VR::UI,VM::VM1,"Processed Series UID",false },
  {0x0011,0x0004,"GEMS_GDXE_FALCON_04",VR::CS,VM::VM1,"Acquisition Type",false },
  {0x0011,0x0005,"GEMS_GDXE_FALCON_04",VR::UI,VM::VM1,"Acquisition UID",false },
  {0x0011,0x0006,"GEMS_GDXE_FALCON_04",VR::DS,VM::VM1,"Image Dose",false },
  {0x0011,0x0007,"GEMS_GDXE_FALCON_04",VR::FL,VM::VM1,"Study Dose",false },
  {0x0011,0x0008,"GEMS_GDXE_FALCON_04",VR::FL,VM::VM1,"Study DAP",false },
  {0x0011,0x0009,"GEMS_GDXE_FALCON_04",VR::SL,VM::VM1,"Non-Digital Exposures",false },
  {0x0011,0x0010,"GEMS_GDXE_FALCON_04",VR::SL,VM::VM1,"Total Exposures",false },
  {0x0011,0x0011,"GEMS_GDXE_FALCON_04",VR::LT,VM::VM1,"ROI",false },
  {0x0011,0x0012,"GEMS_GDXE_FALCON_04",VR::LT,VM::VM1,"Patient Size String",false },
  {0x0011,0x0013,"GEMS_GDXE_FALCON_04",VR::UI,VM::VM1,"SPS UID",false },
  {0x0011,0x0014,"GEMS_GDXE_FALCON_04",VR::UI,VM::VM1,"?",false },
  {0x0011,0x0015,"GEMS_GDXE_FALCON_04",VR::DS,VM::VM1,"Detector ARC Gain",false },
  {0x0011,0x0016,"GEMS_GDXE_FALCON_04",VR::LT,VM::VM1,"Processing Debug Info",false },
  {0x0011,0x0017,"GEMS_GDXE_FALCON_04",VR::CS,VM::VM1,"Override mode",false },
  {0x0011,0x0019,"GEMS_GDXE_FALCON_04",VR::DS,VM::VM1,"Film Speed Selection",false },
  {0x0011,0x0027,"GEMS_GDXE_FALCON_04",VR::FL,VM::VM1,"?",false },
  {0x0011,0x0028,"GEMS_GDXE_FALCON_04",VR::FL,VM::VM1,"?",false },
  {0x0011,0x0029,"GEMS_GDXE_FALCON_04",VR::SL,VM::VM1,"?",false },
  {0x0011,0x0030,"GEMS_GDXE_FALCON_04",VR::SL,VM::VM1,"?",false },
  {0x0011,0x0031,"GEMS_GDXE_FALCON_04",VR::IS,VM::VM8,"Detected Field of View",false },
  {0x0011,0x0032,"GEMS_GDXE_FALCON_04",VR::IS,VM::VM8,"Adjusted Field of View",false },
  {0x0011,0x0033,"GEMS_GDXE_FALCON_04",VR::DS,VM::VM1,"Detected Exposure Index",false },
  {0x0011,0x0034,"GEMS_GDXE_FALCON_04",VR::DS,VM::VM1,"Compensated Detector Exposure",false },
  {0x0011,0x0035,"GEMS_GDXE_FALCON_04",VR::DS,VM::VM1,"Uncompensated Detector Exposure",false },
  {0x0011,0x0036,"GEMS_GDXE_FALCON_04",VR::DS,VM::VM1,"Median Anatomy Count Value",false },
  {0x0011,0x0037,"GEMS_GDXE_FALCON_04",VR::DS,VM::VM2,"DEI lower and upper limit values",false },
  {0x0011,0x0038,"GEMS_GDXE_FALCON_04",VR::SL,VM::VM6,"Shift Vector for Pasting",false },
  {0x0011,0x0039,"GEMS_GDXE_FALCON_04",VR::CS,VM::VM1,"Image Number in Pasting",false },
  {0x0011,0x0040,"GEMS_GDXE_FALCON_04",VR::SL,VM::VM1,"Pasting Overlap",false },
  {0x0011,0x0041,"GEMS_GDXE_FALCON_04",VR::IS,VM::VM24,"Sub-image Collimator Vertices",false },
  {0x0011,0x0042,"GEMS_GDXE_FALCON_04",VR::LO,VM::VM1,"View IP",false },
  {0x0011,0x0043,"GEMS_GDXE_FALCON_04",VR::IS,VM::VM24,"Key-stone Coordinates",false },
  {0x0011,0x0044,"GEMS_GDXE_FALCON_04",VR::CS,VM::VM1,"Receptor Type",false },
  {0x0011,0x0046,"GEMS_GDXE_FALCON_04",VR::LO,VM::VM1_n,"?",false },
  {0x0011,0x0047,"GEMS_GDXE_FALCON_04",VR::DS,VM::VM1,"?",false },
  {0x0011,0x0059,"GEMS_GDXE_FALCON_04",VR::CS,VM::VM1,"?",false },
  {0x0011,0x0060,"GEMS_GDXE_FALCON_04",VR::CS,VM::VM1,"?",false },
  {0x0009,0x0001,"GEMS_GENIE_1",VR::SH,VM::VM1,"?",false },
  {0x0009,0x0010,"GEMS_GENIE_1",VR::LO,VM::VM1,"Study Name",false },
  {0x0009,0x0011,"GEMS_GENIE_1",VR::SL,VM::VM1,"Study Flags",false },
  {0x0009,0x0012,"GEMS_GENIE_1",VR::SL,VM::VM1,"Study Type",false },
  {0x0009,0x001e,"GEMS_GENIE_1",VR::UI,VM::VM1,"Dataset UID",false },
  {0x0009,0x0020,"GEMS_GENIE_1",VR::LO,VM::VM1,"Series Object Name",false },
  {0x0009,0x0021,"GEMS_GENIE_1",VR::SL,VM::VM1,"Series Flags",false },
  {0x0009,0x0022,"GEMS_GENIE_1",VR::SH,VM::VM1,"User Orientation",false },
  {0x0009,0x0023,"GEMS_GENIE_1",VR::SL,VM::VM1,"Initiation Type",false },
  {0x0009,0x0024,"GEMS_GENIE_1",VR::SL,VM::VM1,"Initiation Delay",false },
  {0x0009,0x0025,"GEMS_GENIE_1",VR::SL,VM::VM1,"Initiation Count Rate",false },
  {0x0009,0x0026,"GEMS_GENIE_1",VR::SL,VM::VM1,"Number Energy Sets",false },
  {0x0009,0x0027,"GEMS_GENIE_1",VR::SL,VM::VM1,"Number Detectors",false },
  {0x0009,0x0028,"GEMS_GENIE_1",VR::SL,VM::VM1,"Number RR Windows",false },
  {0x0009,0x0029,"GEMS_GENIE_1",VR::SL,VM::VM1,"Number MG Time Slots",false },
  {0x0009,0x002a,"GEMS_GENIE_1",VR::SL,VM::VM1,"Number View Sets",false },
  {0x0009,0x002b,"GEMS_GENIE_1",VR::LO,VM::VM1,"Trigger History UID",false },
  {0x0009,0x002c,"GEMS_GENIE_1",VR::LO,VM::VM1,"Series Comments",false },
  {0x0009,0x002d,"GEMS_GENIE_1",VR::SL,VM::VM1,"Track Beat Average",false },
  {0x0009,0x002e,"GEMS_GENIE_1",VR::FD,VM::VM1,"Distance Prescribed",false },
  {0x0009,0x002f,"GEMS_GENIE_1",VR::SL,VM::VM1,"Table Direction",false },
  {0x0009,0x0030,"GEMS_GENIE_1",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0033,"GEMS_GENIE_1",VR::FD,VM::VM1,"Rotational Continuous Speed",false },
  {0x0009,0x0034,"GEMS_GENIE_1",VR::SL,VM::VM1,"Gantry Motion Type (retired)",false },
  {0x0009,0x0035,"GEMS_GENIE_1",VR::UL,VM::VM1,"Gantry Locus Type",false },
  {0x0009,0x0037,"GEMS_GENIE_1",VR::SL,VM::VM1,"Starting Heart Rate",false },
  {0x0009,0x0038,"GEMS_GENIE_1",VR::SL,VM::VM1,"RR Window Width",false },
  {0x0009,0x0039,"GEMS_GENIE_1",VR::SL,VM::VM1,"RR Window Offset",false },
  {0x0009,0x003a,"GEMS_GENIE_1",VR::SL,VM::VM1,"Percent Cycle Imaged",false },
  {0x0009,0x0040,"GEMS_GENIE_1",VR::PN,VM::VM1,"Patient Object Name",false },
  {0x0009,0x0041,"GEMS_GENIE_1",VR::SL,VM::VM1,"Patient Flags",false },
  {0x0009,0x0042,"GEMS_GENIE_1",VR::DA,VM::VM1,"Patient Creation Date",false },
  {0x0009,0x0043,"GEMS_GENIE_1",VR::TM,VM::VM1,"Patient Creation Time",false },
  {0x0009,0x0044,"GEMS_GENIE_1",VR::SL,VM::VM1,"Num Views Acquired (retired)",false },
  {0x0009,0x0045,"GEMS_GENIE_1",VR::LT,VM::VM1,"Dataset UID List",false },
  {0x0009,0x0046,"GEMS_GENIE_1",VR::UI,VM::VM1,"?DataSet UID?",false },
  {0x0011,0x000a,"GEMS_GENIE_1",VR::SL,VM::VM1,"Series Type",false },
  {0x0011,0x000b,"GEMS_GENIE_1",VR::SL,VM::VM1,"Effective Series Duration",false },
  {0x0011,0x000c,"GEMS_GENIE_1",VR::SL,VM::VM1,"Num Beats",false },
  {0x0011,0x000d,"GEMS_GENIE_1",VR::LO,VM::VM1,"Radio Nuclide Name",false },
  {0x0011,0x0010,"GEMS_GENIE_1",VR::LO,VM::VM1,"Dataset Object Name",false },
  {0x0011,0x0011,"GEMS_GENIE_1",VR::SL,VM::VM1,"Dataset Modified",false },
  {0x0011,0x0012,"GEMS_GENIE_1",VR::LO,VM::VM1_n,"Dataset Name",false },
  {0x0011,0x0013,"GEMS_GENIE_1",VR::SL,VM::VM1,"Dataset Type",false },
  {0x0011,0x0014,"GEMS_GENIE_1",VR::LO,VM::VM1,"Completion Time",false },
  {0x0011,0x0015,"GEMS_GENIE_1",VR::SL,VM::VM1,"Detector Number",false },
  {0x0011,0x0016,"GEMS_GENIE_1",VR::SL,VM::VM1,"Energy Number",false },
  {0x0011,0x0017,"GEMS_GENIE_1",VR::SL,VM::VM1,"RR Interval Window Number",false },
  {0x0011,0x0018,"GEMS_GENIE_1",VR::SL,VM::VM1,"MG Bin Number",false },
  {0x0011,0x0019,"GEMS_GENIE_1",VR::FD,VM::VM1,"Radius Of Rotation",false },
  {0x0011,0x001a,"GEMS_GENIE_1",VR::SL,VM::VM1,"Detector Count Zone",false },
  {0x0011,0x001b,"GEMS_GENIE_1",VR::SL,VM::VM1,"Num Energy Windows",false },
  {0x0011,0x001c,"GEMS_GENIE_1",VR::SL,VM::VM1,"Energy Offset",false },
  {0x0011,0x001d,"GEMS_GENIE_1",VR::SL,VM::VM1,"Energy Range",false },
  {0x0011,0x001e,"GEMS_GENIE_1",VR::SL,VM::VM4,"Energy Width (retired)",false },
  {0x0011,0x001f,"GEMS_GENIE_1",VR::SL,VM::VM1,"Image Orientation",false },
  {0x0011,0x0021,"GEMS_GENIE_1",VR::DS,VM::VM1,"Acq Zoom (retired)",false },
  {0x0011,0x0022,"GEMS_GENIE_1",VR::DS,VM::VM1,"Acq Pan (retired)",false },
  {0x0011,0x0023,"GEMS_GENIE_1",VR::SL,VM::VM1,"Use FOV Mask",false },
  {0x0011,0x0024,"GEMS_GENIE_1",VR::SL,VM::VM1,"FOV Mask Y Cutoff Distance",false },
  {0x0011,0x0025,"GEMS_GENIE_1",VR::SL,VM::VM1,"FOV Mask Cutoff Angle",false },
  {0x0011,0x0026,"GEMS_GENIE_1",VR::SL,VM::VM1,"Table Orientation",false },
  {0x0011,0x0027,"GEMS_GENIE_1",VR::SL,VM::VM2,"ROI Top Left",false },
  {0x0011,0x0028,"GEMS_GENIE_1",VR::SL,VM::VM2,"ROI Bottom Right",false },
  {0x0011,0x0029,"GEMS_GENIE_1",VR::SL,VM::VM1,"Uniformity Mean",false },
  {0x0011,0x002a,"GEMS_GENIE_1",VR::FD,VM::VM1,"Phase Duration (retired)",false },
  {0x0011,0x002c,"GEMS_GENIE_1",VR::FD,VM::VM1,"View X Adjustment",false },
  {0x0011,0x002d,"GEMS_GENIE_1",VR::FD,VM::VM1,"View Y Adjustment",false },
  {0x0011,0x002e,"GEMS_GENIE_1",VR::SL,VM::VM1,"Pixel Overflow Flag",false },
  {0x0011,0x002f,"GEMS_GENIE_1",VR::SL,VM::VM1,"Overflow Level",false },
  {0x0011,0x0030,"GEMS_GENIE_1",VR::LO,VM::VM1,"Picture Object Name",false },
  {0x0011,0x0031,"GEMS_GENIE_1",VR::LO,VM::VM1,"Acquisition Parent UID",false },
  {0x0011,0x0032,"GEMS_GENIE_1",VR::LO,VM::VM1,"Processing Parent UID",false },
  {0x0011,0x0033,"GEMS_GENIE_1",VR::LO,VM::VM1,"Energy Correct Name",false },
  {0x0011,0x0034,"GEMS_GENIE_1",VR::LO,VM::VM1,"Spatial Correct Name",false },
  {0x0011,0x0035,"GEMS_GENIE_1",VR::LO,VM::VM1,"Tuning Calib Name",false },
  {0x0011,0x0036,"GEMS_GENIE_1",VR::LO,VM::VM1,"Uniformity Correct Name",false },
  {0x0011,0x0037,"GEMS_GENIE_1",VR::LT,VM::VM1,"Acquisition Specific Correction Name",false },
  {0x0011,0x0038,"GEMS_GENIE_1",VR::SL,VM::VM1,"Byte Order",false },
  {0x0011,0x0039,"GEMS_GENIE_1",VR::SL,VM::VM1,"Compression Type",false },
  {0x0011,0x003a,"GEMS_GENIE_1",VR::SL,VM::VM1,"Picture Format",false },
  {0x0011,0x003b,"GEMS_GENIE_1",VR::FD,VM::VM1,"Pixel Scale",false },
  {0x0011,0x003c,"GEMS_GENIE_1",VR::FD,VM::VM1,"Pixel Offset",false },
  {0x0011,0x003d,"GEMS_GENIE_1",VR::SL,VM::VM4,"Energy Peak (retired)",false },
  {0x0011,0x003e,"GEMS_GENIE_1",VR::SL,VM::VM1,"FOV Shape",false },
  {0x0011,0x003f,"GEMS_GENIE_1",VR::SL,VM::VM1,"Dataset Flags",false },
  {0x0011,0x0040,"GEMS_GENIE_1",VR::LO,VM::VM1,"Viewing Object Name",false },
  {0x0011,0x0041,"GEMS_GENIE_1",VR::SL,VM::VM1,"Orientation Angle",false },
  {0x0011,0x0042,"GEMS_GENIE_1",VR::FD,VM::VM1,"Rotation Angle",false },
  {0x0011,0x0043,"GEMS_GENIE_1",VR::SL,VM::VM1,"Window Inverse Flag",false },
  {0x0011,0x0044,"GEMS_GENIE_1",VR::FD,VM::VM1,"Threshold Center",false },
  {0x0011,0x0045,"GEMS_GENIE_1",VR::FD,VM::VM1,"Threshold Width",false },
  {0x0011,0x0046,"GEMS_GENIE_1",VR::SL,VM::VM1,"Interpolation Type",false },
  {0x0011,0x0050,"GEMS_GENIE_1",VR::LO,VM::VM1,"Where Object Name",false },
  {0x0011,0x0055,"GEMS_GENIE_1",VR::FD,VM::VM1,"Period",false },
  {0x0011,0x0056,"GEMS_GENIE_1",VR::FD,VM::VM1,"Elapsed Time",false },
  {0x0011,0x0057,"GEMS_GENIE_1",VR::FD,VM::VM2,"FOV",false },
  {0x0011,0x0061,"GEMS_GENIE_1",VR::SL,VM::VM3,"Image Size",false },
  {0x0011,0x0062,"GEMS_GENIE_1",VR::FD,VM::VM3,"Linear FOV",false },
  {0x0011,0x0063,"GEMS_GENIE_1",VR::FD,VM::VM3,"Spatial Offset",false },
  {0x0011,0x0064,"GEMS_GENIE_1",VR::FD,VM::VM3,"Spatial Orientation",false },
  {0x0011,0x0065,"GEMS_GENIE_1",VR::LO,VM::VM1,"Reference Dataset UID",false },
  {0x0011,0x0066,"GEMS_GENIE_1",VR::LO,VM::VM1,"Starcam Reference Dataset",false },
  {0x0011,0x0067,"GEMS_GENIE_1",VR::SL,VM::VM1,"Reference Frame Number",false },
  {0x0011,0x0068,"GEMS_GENIE_1",VR::SL,VM::VM1,"Cursor Length",false },
  {0x0011,0x0069,"GEMS_GENIE_1",VR::SL,VM::VM1,"Number of Cursors",false },
  {0x0011,0x006a,"GEMS_GENIE_1",VR::SL,VM::VM9,"Cursor Coordinates",false },
  {0x0011,0x006b,"GEMS_GENIE_1",VR::SL,VM::VM3,"Recon Options Flag",false },
  {0x0011,0x006c,"GEMS_GENIE_1",VR::FD,VM::VM3,"Motion Threshold",false },
  {0x0011,0x006d,"GEMS_GENIE_1",VR::UI,VM::VM1,"Motion Curve UID",false },
  {0x0011,0x006e,"GEMS_GENIE_1",VR::SL,VM::VM1,"Recon Type",false },
  {0x0011,0x006f,"GEMS_GENIE_1",VR::SL,VM::VM1,"Pre Filter Type",false },
  {0x0011,0x0071,"GEMS_GENIE_1",VR::SL,VM::VM1,"Back Proj Filter Type",false },
  {0x0011,0x0072,"GEMS_GENIE_1",VR::SL,VM::VM1,"Recon Arc",false },
  {0x0011,0x0073,"GEMS_GENIE_1",VR::FD,VM::VM1,"Recon Pan AP Offset",false },
  {0x0011,0x0074,"GEMS_GENIE_1",VR::FD,VM::VM1,"Recon Pan LR Offset",false },
  {0x0011,0x0075,"GEMS_GENIE_1",VR::FD,VM::VM1,"Recon Area",false },
  {0x0011,0x0076,"GEMS_GENIE_1",VR::SL,VM::VM1,"Start View",false },
  {0x0011,0x0077,"GEMS_GENIE_1",VR::SL,VM::VM1,"Attenuation Type",false },
  {0x0011,0x0078,"GEMS_GENIE_1",VR::SL,VM::VM1,"Dual Energy Processing",false },
  {0x0011,0x0079,"GEMS_GENIE_1",VR::SH,VM::VM1,"Pre Filter Param",false },
  {0x0011,0x007a,"GEMS_GENIE_1",VR::SH,VM::VM1,"Pre Filter Param 2",false },
  {0x0011,0x007b,"GEMS_GENIE_1",VR::SH,VM::VM1,"Back Proj Filter Param",false },
  {0x0011,0x007c,"GEMS_GENIE_1",VR::SH,VM::VM1,"Back Proj Filter Param 2",false },
  {0x0011,0x007d,"GEMS_GENIE_1",VR::SH,VM::VM1,"Attenuation Coef",false },
  {0x0011,0x007e,"GEMS_GENIE_1",VR::SL,VM::VM1,"Ref Slice Width",false },
  {0x0011,0x007f,"GEMS_GENIE_1",VR::FD,VM::VM1,"Ref Trans Pixel Volume",false },
  {0x0011,0x0081,"GEMS_GENIE_1",VR::SH,VM::VM1,"Attenuation Threshold",false },
  {0x0011,0x0082,"GEMS_GENIE_1",VR::FD,VM::VM1,"Interpolation Distance",false },
  {0x0011,0x0083,"GEMS_GENIE_1",VR::FD,VM::VM1,"Interpolation Center X",false },
  {0x0011,0x0084,"GEMS_GENIE_1",VR::FD,VM::VM1,"Interpolation Center Y",false },
  {0x0011,0x0085,"GEMS_GENIE_1",VR::SL,VM::VM1,"Quant Filter Flag",false },
  {0x0011,0x0086,"GEMS_GENIE_1",VR::SL,VM::VM1,"Head Conversion",false },
  {0x0011,0x0087,"GEMS_GENIE_1",VR::SL,VM::VM1,"Slice Width Pixels",false },
  {0x0011,0x0088,"GEMS_GENIE_1",VR::SL,VM::VM1,"Rfmtr Trans Ref",false },
  {0x0011,0x0089,"GEMS_GENIE_1",VR::FD,VM::VM1,"Rfmtr Trans Ref mm",false },
  {0x0011,0x008a,"GEMS_GENIE_1",VR::SL,VM::VM4,"Two Line Trans Ref",false },
  {0x0011,0x008b,"GEMS_GENIE_1",VR::SL,VM::VM3,"Three-D Zero",false },
  {0x0011,0x008c,"GEMS_GENIE_1",VR::SL,VM::VM3,"Three-D Zero Length",false },
  {0x0011,0x008d,"GEMS_GENIE_1",VR::SL,VM::VM1,"Three-D Zero In",false },
  {0x0013,0x0010,"GEMS_GENIE_1",VR::FD,VM::VM2,"Digital FOV",false },
  {0x0013,0x0011,"GEMS_GENIE_1",VR::SL,VM::VM1,"Source Translator",false },
  {0x0013,0x0012,"GEMS_GENIE_1",VR::SL,VM::VM1,"RAL Flags",false },
  {0x0013,0x0013,"GEMS_GENIE_1",VR::SQ,VM::VM1,"Xeleris Frame Sequence",false },
  {0x0013,0x0014,"GEMS_GENIE_1",VR::SL,VM::VM1_n,"Original Image Num",false },
  {0x0013,0x0015,"GEMS_GENIE_1",VR::FD,VM::VM1,"Fscalar",false },
  {0x0013,0x0016,"GEMS_GENIE_1",VR::SL,VM::VM1,"AutoTrack Peak",false },
  {0x0013,0x0017,"GEMS_GENIE_1",VR::SL,VM::VM1,"AutoTrack Width",false },
  {0x0013,0x0018,"GEMS_GENIE_1",VR::FD,VM::VM1,"Transmission Scan Time",false },
  {0x0013,0x0019,"GEMS_GENIE_1",VR::FD,VM::VM1,"Transmission Mask Width",false },
  {0x0013,0x001a,"GEMS_GENIE_1",VR::FD,VM::VM1,"Copper Attenuator Thickness",false },
  {0x0013,0x001b,"GEMS_GENIE_1",VR::FD,VM::VM1,"Det Ang Separation",false },
  {0x0013,0x001c,"GEMS_GENIE_1",VR::SL,VM::VM1,"Axial Acceptance Angle",false },
  {0x0013,0x001d,"GEMS_GENIE_1",VR::SL,VM::VM1,"Theta Acceptance Value",false },
  {0x0013,0x001e,"GEMS_GENIE_1",VR::FD,VM::VM1_n,"Tomo View Offset",false },
  {0x0013,0x0020,"GEMS_GENIE_1",VR::FD,VM::VM1,"Accepted Beats Time",false },
  {0x0013,0x0021,"GEMS_GENIE_1",VR::FD,VM::VM2,"Threshold",false },
  {0x0013,0x0022,"GEMS_GENIE_1",VR::FD,VM::VM2,"Linear Depth",false },
  {0x0013,0x0023,"GEMS_GENIE_1",VR::LO,VM::VM1,"Unif Date Time",false },
  {0x0013,0x0024,"GEMS_GENIE_1",VR::SL,VM::VM1,"Series Accepted Beats",false },
  {0x0013,0x0025,"GEMS_GENIE_1",VR::SL,VM::VM1,"Series Rejected Beats",false },
  {0x0013,0x0026,"GEMS_GENIE_1",VR::LT,VM::VM1,"Study Comments",false },
  {0x0015,0x0010,"GEMS_GENIE_1",VR::SL,VM::VM1,"Frame Termination Condition",false },
  {0x0015,0x0011,"GEMS_GENIE_1",VR::SL,VM::VM1,"Frame Termination Value",false },
  {0x0015,0x0012,"GEMS_GENIE_1",VR::SL,VM::VM1,"Num ECT Phases",false },
  {0x0015,0x0013,"GEMS_GENIE_1",VR::SL,VM::VM1,"Num WB Scans",false },
  {0x0015,0x0014,"GEMS_GENIE_1",VR::SL,VM::VM1,"ECT Phase Num",false },
  {0x0015,0x0015,"GEMS_GENIE_1",VR::SL,VM::VM1,"WB Scan Num",false },
  {0x0015,0x0016,"GEMS_GENIE_1",VR::SL,VM::VM1,"Comb Head Number",false },
  {0x0015,0x0017,"GEMS_GENIE_1",VR::UL,VM::VM1,"Preceding Beat",false },
  {0x0019,0x005f,"GEMS_GENIE_1",VR::SQ,VM::VM1,"Annotation Sequence",false },
  {0x0033,0x0007,"GEMS_GENIE_1",VR::LO,VM::VM1,"Orig SOP Instance UID",false },
  {0x0033,0x0008,"GEMS_GENIE_1",VR::CS,VM::VM1,"Xeleris Data Object Type",false },
  {0x0033,0x0010,"GEMS_GENIE_1",VR::SL,VM::VM1,"Modified",false },
  {0x0033,0x0011,"GEMS_GENIE_1",VR::LO,VM::VM1,"Name",false },
  {0x0033,0x0016,"GEMS_GENIE_1",VR::LO,VM::VM1,"ProtocolDataUID",false },
  {0x0033,0x0017,"GEMS_GENIE_1",VR::SH,VM::VM1,"Date",false },
  {0x0033,0x0018,"GEMS_GENIE_1",VR::SH,VM::VM1,"Time",false },
  {0x0033,0x0019,"GEMS_GENIE_1",VR::UL,VM::VM1,"ProtocoldataFlags",false },
  {0x0033,0x001a,"GEMS_GENIE_1",VR::UL,VM::VM1,"ProtocolName",false },
  {0x0033,0x001b,"GEMS_GENIE_1",VR::LO,VM::VM1,"RelevantDataUID",false },
  {0x0033,0x001c,"GEMS_GENIE_1",VR::LO,VM::VM1,"BulkData",false },
  {0x0033,0x001d,"GEMS_GENIE_1",VR::SL,VM::VM1_n,"IntData",false },
  {0x0033,0x001e,"GEMS_GENIE_1",VR::FD,VM::VM1_n,"DoubleData",false },
  {0x0033,0x001f,"GEMS_GENIE_1",VR::OB,VM::VM1,"StringData",false },
  {0x0033,0x0020,"GEMS_GENIE_1",VR::LT,VM::VM1_n,"BulkDataFormat",false },
  {0x0033,0x0021,"GEMS_GENIE_1",VR::LT,VM::VM1_n,"IntDataFormat",false },
  {0x0033,0x0022,"GEMS_GENIE_1",VR::LT,VM::VM1_n,"DoubleDataFormat",false },
  {0x0033,0x0023,"GEMS_GENIE_1",VR::LT,VM::VM1_n,"StringDataFormat",false },
  {0x0033,0x0024,"GEMS_GENIE_1",VR::LT,VM::VM1,"Description",false },
  {0x0033,0x0030,"GEMS_GENIE_1",VR::UL,VM::VM1,"Allocate trigger buffer",false },
  {0x0033,0x0033,"GEMS_GENIE_1",VR::UL,VM::VM1,"Number of triggers",false },
  {0x0033,0x0034,"GEMS_GENIE_1",VR::UL,VM::VM1,"Trigger size",false },
  {0x0033,0x0035,"GEMS_GENIE_1",VR::UL,VM::VM1,"Trigger data size",false },
  {0x0033,0x0036,"GEMS_GENIE_1",VR::OB,VM::VM1_n,"Trigger data",false },
  {0x0035,0x0001,"GEMS_GENIE_1",VR::FD,VM::VM1_n,"Start Angle",false },
  {0x0055,0x0012,"GEMS_GENIE_1",VR::SQ,VM::VM1,"Xeleris Energy Window Information Sequence",false },
  {0x0055,0x0013,"GEMS_GENIE_1",VR::SQ,VM::VM1,"Xeleris Energy Window Range Sequence",false },
  {0x0055,0x0022,"GEMS_GENIE_1",VR::SQ,VM::VM1,"Xeleris Detector Information Sequence",false },
  {0x0055,0x0052,"GEMS_GENIE_1",VR::SQ,VM::VM1,"Xeleris Rotation Information Sequence",false },
  {0x0055,0x0062,"GEMS_GENIE_1",VR::SQ,VM::VM1,"Xeleris Gated Information Sequence",false },
  {0x0055,0x0063,"GEMS_GENIE_1",VR::SQ,VM::VM1,"Xeleris Data Information Sequence",false },
  {0x0055,0x0064,"GEMS_GENIE_1",VR::SQ,VM::VM1,"SDODoubleData SQ",false },
  {0x0055,0x0065,"GEMS_GENIE_1",VR::SQ,VM::VM1,"Xeleris Frame Sequence",false },
  {0x5001,0x0001,"GEMS_GENIE_1",VR::SL,VM::VM1,"Modified",false },
  {0x5001,0x0002,"GEMS_GENIE_1",VR::LO,VM::VM1,"Name",false },
  {0x5001,0x0003,"GEMS_GENIE_1",VR::SL,VM::VM1,"Cid",false },
  {0x5001,0x0004,"GEMS_GENIE_1",VR::SL,VM::VM1,"Srid",false },
  {0x5001,0x0005,"GEMS_GENIE_1",VR::LO,VM::VM1,"SOPClassUID",false },
  {0x5001,0x0006,"GEMS_GENIE_1",VR::LO,VM::VM1,"SOPInstanceUID",false },
  {0x5001,0x0007,"GEMS_GENIE_1",VR::SL,VM::VM1,"CurveType",false },
  {0x5001,0x0008,"GEMS_GENIE_1",VR::SL,VM::VM1,"GraphType",false },
  {0x5001,0x0009,"GEMS_GENIE_1",VR::LO,VM::VM1,"Legend",false },
  {0x5001,0x000a,"GEMS_GENIE_1",VR::LO,VM::VM1,"XUnits",false },
  {0x5001,0x000b,"GEMS_GENIE_1",VR::LO,VM::VM1,"YUnits",false },
  {0x5001,0x000c,"GEMS_GENIE_1",VR::SL,VM::VM1,"Edit",false },
  {0x5001,0x000d,"GEMS_GENIE_1",VR::SL,VM::VM1,"Suspend",false },
  {0x5001,0x000e,"GEMS_GENIE_1",VR::SL,VM::VM1,"StyleLine",false },
  {0x5001,0x000f,"GEMS_GENIE_1",VR::SL,VM::VM1,"StyleFill",false },
  {0x5001,0x0010,"GEMS_GENIE_1",VR::LO,VM::VM1,"StyleColour",false },
  {0x5001,0x0011,"GEMS_GENIE_1",VR::SL,VM::VM1,"StyleWidth",false },
  {0x5001,0x0012,"GEMS_GENIE_1",VR::SL,VM::VM1,"StylePoint",false },
  {0x5001,0x0013,"GEMS_GENIE_1",VR::LO,VM::VM1,"StylePColour",false },
  {0x5001,0x0014,"GEMS_GENIE_1",VR::SL,VM::VM1,"StylePSize",false },
  {0x5001,0x0015,"GEMS_GENIE_1",VR::SL,VM::VM1,"Segments",false },
  {0x5001,0x0016,"GEMS_GENIE_1",VR::SL,VM::VM1,"SegType",false },
  {0x5001,0x0017,"GEMS_GENIE_1",VR::FD,VM::VM1_n,"SegStart",false },
  {0x5001,0x0018,"GEMS_GENIE_1",VR::FD,VM::VM1_n,"SegEnd",false },
  {0x5001,0x0019,"GEMS_GENIE_1",VR::SL,VM::VM1_n,"SegStyleLine",false },
  {0x5001,0x001a,"GEMS_GENIE_1",VR::SL,VM::VM1_n,"SegStyleFill",false },
  {0x5001,0x001b,"GEMS_GENIE_1",VR::LO,VM::VM1,"SegStyleColour",false },
  {0x5001,0x001c,"GEMS_GENIE_1",VR::SL,VM::VM1_n,"SegStyleWidth",false },
  {0x5001,0x001d,"GEMS_GENIE_1",VR::SL,VM::VM1_n,"SegStylePoint",false },
  {0x5001,0x001e,"GEMS_GENIE_1",VR::SL,VM::VM1,"SegStylePColour",false },
  {0x5001,0x001f,"GEMS_GENIE_1",VR::SL,VM::VM1,"SegStylePSize",false },
  {0x5001,0x0020,"GEMS_GENIE_1",VR::LO,VM::VM1,"SegName",false },
  {0x5001,0x0021,"GEMS_GENIE_1",VR::SL,VM::VM1_n,"SegAllowDirInt",false },
  {0x5001,0x0022,"GEMS_GENIE_1",VR::SL,VM::VM1,"TextAnnots",false },
  {0x5001,0x0023,"GEMS_GENIE_1",VR::FD,VM::VM1_n,"TxtX",false },
  {0x5001,0x0024,"GEMS_GENIE_1",VR::FD,VM::VM1_n,"TxtY",false },
  {0x5001,0x0025,"GEMS_GENIE_1",VR::LO,VM::VM1,"TxtText",false },
  {0x5001,0x0026,"GEMS_GENIE_1",VR::LO,VM::VM1,"TxtName",false },
  {0x5001,0x0030,"GEMS_GENIE_1",VR::LO,VM::VM1,"ROIName",false },
  {0x5001,0x0031,"GEMS_GENIE_1",VR::LO,VM::VM1,"DerivedFromImageUID",false },
  {0x5001,0x0032,"GEMS_GENIE_1",VR::SL,VM::VM1_n,"DerivedFromImages",false },
  {0x5001,0x0033,"GEMS_GENIE_1",VR::UL,VM::VM1,"CurveFlags",false },
  {0x5001,0x0034,"GEMS_GENIE_1",VR::LO,VM::VM1,"CurveName",false },
  {0x5001,0x0035,"GEMS_GENIE_1",VR::LO,VM::VM1,"DatasetName",false },
  {0x5001,0x0036,"GEMS_GENIE_1",VR::LO,VM::VM1,"CurveUID",false },
  {0x5001,0x0037,"GEMS_GENIE_1",VR::FD,VM::VM1,"ROIArea",false },
  {0x5001,0x0038,"GEMS_GENIE_1",VR::SL,VM::VM1,"Modified",false },
  {0x5001,0x0039,"GEMS_GENIE_1",VR::LO,VM::VM1,"Name",false },
  {0x5001,0x003a,"GEMS_GENIE_1",VR::LO,VM::VM1,"Software version",false },
  {0x5001,0x003b,"GEMS_GENIE_1",VR::SH,VM::VM1,"StartDate",false },
  {0x5001,0x003c,"GEMS_GENIE_1",VR::SH,VM::VM1,"CompletionDate",false },
  {0x5001,0x003d,"GEMS_GENIE_1",VR::LO,VM::VM1,"DetectorName",false },
  {0x5001,0x0041,"GEMS_GENIE_1",VR::SL,VM::VM1,"Modified",false },
  {0x5001,0x0042,"GEMS_GENIE_1",VR::LO,VM::VM1,"Name",false },
  {0x5001,0x0043,"GEMS_GENIE_1",VR::SL,VM::VM1,"Name",false },
  {0x5001,0x0044,"GEMS_GENIE_1",VR::SL,VM::VM1,"Name",false },
  {0x5001,0x0045,"GEMS_GENIE_1",VR::LO,VM::VM1,"SOPClassUID",false },
  {0x5001,0x0046,"GEMS_GENIE_1",VR::LO,VM::VM1,"SOPInstanceUID",false },
  {0x5001,0x0047,"GEMS_GENIE_1",VR::LO,VM::VM1,"NormalColor",false },
  {0x5001,0x0048,"GEMS_GENIE_1",VR::LT,VM::VM1,"NameFont",false },
  {0x5001,0x0049,"GEMS_GENIE_1",VR::SL,VM::VM1,"FillPattern",false },
  {0x5001,0x004a,"GEMS_GENIE_1",VR::SL,VM::VM1,"LineStyle",false },
  {0x5001,0x004b,"GEMS_GENIE_1",VR::SL,VM::VM1,"LineDashLength",false },
  {0x5001,0x004c,"GEMS_GENIE_1",VR::SL,VM::VM1,"LineThickness",false },
  {0x5001,0x004d,"GEMS_GENIE_1",VR::SL,VM::VM1,"Interactivity",false },
  {0x5001,0x004e,"GEMS_GENIE_1",VR::SL,VM::VM1,"NamePos",false },
  {0x5001,0x004f,"GEMS_GENIE_1",VR::SL,VM::VM1,"NameDisplay",false },
  {0x5001,0x0050,"GEMS_GENIE_1",VR::LO,VM::VM1,"Label",false },
  {0x5001,0x0051,"GEMS_GENIE_1",VR::SL,VM::VM1_n,"BpSeg",false },
  {0x5001,0x0052,"GEMS_GENIE_1",VR::US,VM::VM1_n,"BpSegPairs",false },
  {0x5001,0x0053,"GEMS_GENIE_1",VR::SL,VM::VM1,"SeedSpace",false },
  {0x5001,0x0054,"GEMS_GENIE_1",VR::FD,VM::VM1_n,"Seeds",false },
  {0x5001,0x0055,"GEMS_GENIE_1",VR::SL,VM::VM1_n,"Shape",false },
  {0x5001,0x0056,"GEMS_GENIE_1",VR::FD,VM::VM1_n,"ShapeTilt",false },
  {0x5001,0x0059,"GEMS_GENIE_1",VR::SL,VM::VM1_n,"ShapePtsSpace",false },
  {0x5001,0x005a,"GEMS_GENIE_1",VR::SL,VM::VM1_n,"ShapeCtrlPtsCounts",false },
  {0x5001,0x005b,"GEMS_GENIE_1",VR::FD,VM::VM1_n,"ShapeCtrlPts",false },
  {0x5001,0x005c,"GEMS_GENIE_1",VR::SL,VM::VM1,"ShapeCPSpace",false },
  {0x5001,0x005d,"GEMS_GENIE_1",VR::SL,VM::VM1,"ROIFlags",false },
  {0x5001,0x005e,"GEMS_GENIE_1",VR::SL,VM::VM1,"FrameNumber",false },
  {0x5001,0x005f,"GEMS_GENIE_1",VR::SL,VM::VM1,"Id",false },
  {0x5001,0x0060,"GEMS_GENIE_1",VR::LO,VM::VM1_n,"DatasetROIMapping",false },
  {0x0033,0x0001,"GEMS_GNHD_01",VR::UN,VM::VM1,"?",false },
  {0x0033,0x0002,"GEMS_GNHD_01",VR::UN,VM::VM1,"?",false },
  {0x0045,0x0001,"GEMS_HELIOS_01",VR::SS,VM::VM1,"Number of Macro Rows in Detector",false },
  {0x0045,0x0002,"GEMS_HELIOS_01",VR::FL,VM::VM1,"Macro width at ISO Center",false },
  {0x0045,0x0003,"GEMS_HELIOS_01",VR::SS,VM::VM1,"DAS type",false },
  {0x0045,0x0004,"GEMS_HELIOS_01",VR::SS,VM::VM1,"DAS gain",false },
  {0x0045,0x0005,"GEMS_HELIOS_01",VR::SS,VM::VM1,"DAS Temperature",false },
  {0x0045,0x0006,"GEMS_HELIOS_01",VR::CS,VM::VM1,"Table Direction",false },
  {0x0045,0x0007,"GEMS_HELIOS_01",VR::FL,VM::VM1,"Z smoothing Factor",false },
  {0x0045,0x0008,"GEMS_HELIOS_01",VR::SS,VM::VM1,"View Weighting Mode",false },
  {0x0045,0x0009,"GEMS_HELIOS_01",VR::SS,VM::VM1,"Sigma Row number",false },
  {0x0045,0x000a,"GEMS_HELIOS_01",VR::FL,VM::VM1,"Minimum DAS value",false },
  {0x0045,0x000b,"GEMS_HELIOS_01",VR::FL,VM::VM1,"Maximum Offset Value",false },
  {0x0045,0x000c,"GEMS_HELIOS_01",VR::SS,VM::VM1,"Number of Views shifted",false },
  {0x0045,0x000d,"GEMS_HELIOS_01",VR::SS,VM::VM1,"Z tracking Flag",false },
  {0x0045,0x000e,"GEMS_HELIOS_01",VR::FL,VM::VM1,"Mean Z error",false },
  {0x0045,0x000f,"GEMS_HELIOS_01",VR::FL,VM::VM1,"Z tracking Error",false },
  {0x0045,0x0010,"GEMS_HELIOS_01",VR::SS,VM::VM1,"Start View 2A",false },
  {0x0045,0x0011,"GEMS_HELIOS_01",VR::SS,VM::VM1,"Number of Views 2A",false },
  {0x0045,0x0012,"GEMS_HELIOS_01",VR::SS,VM::VM1,"Start View 1A",false },
  {0x0045,0x0013,"GEMS_HELIOS_01",VR::SS,VM::VM1,"Sigma Mode",false },
  {0x0045,0x0014,"GEMS_HELIOS_01",VR::SS,VM::VM1,"Number of Views 1A",false },
  {0x0045,0x0015,"GEMS_HELIOS_01",VR::SS,VM::VM1,"Start View 2B",false },
  {0x0045,0x0016,"GEMS_HELIOS_01",VR::SS,VM::VM1,"Number Views 2B",false },
  {0x0045,0x0017,"GEMS_HELIOS_01",VR::SS,VM::VM1,"Start View 1B",false },
  {0x0045,0x0018,"GEMS_HELIOS_01",VR::SS,VM::VM1,"Number of Views 1B",false },
  {0x0045,0x0019,"GEMS_HELIOS_01",VR::SS,VM::VM1,"Air Calibration Date",false },
  {0x0045,0x001a,"GEMS_HELIOS_01",VR::SS,VM::VM1,"Air Calibration Time",false },
  {0x0045,0x001b,"GEMS_HELIOS_01",VR::SS,VM::VM1,"Phantom Calibration Date",false },
  {0x0045,0x001c,"GEMS_HELIOS_01",VR::SS,VM::VM1,"Phantom Calibration Time",false },
  {0x0045,0x001d,"GEMS_HELIOS_01",VR::SS,VM::VM1,"Z Slope Calibration Date",false },
  {0x0045,0x001e,"GEMS_HELIOS_01",VR::SS,VM::VM1,"Z Slope Calibration Time",false },
  {0x0045,0x001f,"GEMS_HELIOS_01",VR::SS,VM::VM1,"Cross Talk Calibration Date",false },
  {0x0045,0x0020,"GEMS_HELIOS_01",VR::SS,VM::VM1,"Cross Talk Calibration Time",false },
  {0x0045,0x0021,"GEMS_HELIOS_01",VR::SS,VM::VM1,"Iterbone Flag",false },
  {0x0045,0x0022,"GEMS_HELIOS_01",VR::SS,VM::VM1,"Peristaltic Flag",false },
  {0x0045,0x0030,"GEMS_HELIOS_01",VR::CS,VM::VM1,"CardiacReconAlgorithm",false },
  {0x0045,0x0031,"GEMS_HELIOS_01",VR::CS,VM::VM1,"AvgHeartRateForImage",false },
  {0x0045,0x0032,"GEMS_HELIOS_01",VR::FL,VM::VM1,"Temporal Resolution",false },
  {0x0045,0x0033,"GEMS_HELIOS_01",VR::CS,VM::VM1,"PctRpeakDelay",false },
  {0x0045,0x0034,"GEMS_HELIOS_01",VR::CS,VM::VM1,"ActualPctRpeakDelay",false },
  {0x0045,0x0036,"GEMS_HELIOS_01",VR::CS,VM::VM1,"EkgFullMaStartPhase",false },
  {0x0045,0x0037,"GEMS_HELIOS_01",VR::CS,VM::VM1,"EkgFullMaEndPhase",false },
  {0x0045,0x0038,"GEMS_HELIOS_01",VR::CS,VM::VM1,"EkgModulationMaxMa",false },
  {0x0045,0x0039,"GEMS_HELIOS_01",VR::CS,VM::VM1,"EkgModulationMinMa",false },
  {0x0045,0x003b,"GEMS_HELIOS_01",VR::LO,VM::VM1,"Noise Reduction Image Filter Description",false },
  {0x0045,0x003f,"GEMS_HELIOS_01",VR::IS,VM::VM1,"RPeakTimeDelay",false },
  {0x0045,0x0044,"GEMS_HELIOS_01",VR::IS,VM::VM1,"ActualRPeakTimeDelay",false },
  {0x0045,0x0045,"GEMS_HELIOS_01",VR::ST,VM::VM1,"CardiacScanOptions",false },
  {0x0045,0x0050,"GEMS_HELIOS_01",VR::FD,VM::VM1,"Temporal Center View Angle",false },
  {0x0045,0x0051,"GEMS_HELIOS_01",VR::FD,VM::VM1,"Recon Center View Angle",false },
  {0x0045,0x0052,"GEMS_HELIOS_01",VR::CS,VM::VM1,"WideCone Masking",false },
  {0x0045,0x0053,"GEMS_HELIOS_01",VR::FD,VM::VM1,"WideCone Corner Blending Radius",false },
  {0x0045,0x0054,"GEMS_HELIOS_01",VR::FD,VM::VM1,"WideCone Corner Blending Radius Offset",false },
  {0x0045,0x0055,"GEMS_HELIOS_01",VR::CS,VM::VM1,"Internal Recon Algorithm",false },
  {0x004b,0x0001,"GEMS_HINO_CT_01",VR::DS,VM::VM1_n,"Beam Thickness",false },
  {0x004b,0x0002,"GEMS_HINO_CT_01",VR::DS,VM::VM1_n,"R Time",false },
  {0x004b,0x0003,"GEMS_HINO_CT_01",VR::IS,VM::VM1,"HBC number",false },
  {0x0009,0x0001,"GEMS_IDEN_01",VR::LO,VM::VM1,"Full fidelity",false },
  {0x0009,0x0002,"GEMS_IDEN_01",VR::SH,VM::VM1,"Suite id",false },
  {0x0009,0x0004,"GEMS_IDEN_01",VR::SH,VM::VM1,"Product id",false },
  {0x0009,0x0017,"GEMS_IDEN_01",VR::LT,VM::VM1,"Series Description",false },
  {0x0009,0x001a,"GEMS_IDEN_01",VR::US,VM::VM1,"?",false },
  {0x0009,0x0020,"GEMS_IDEN_01",VR::US,VM::VM1,"?",false },
  {0x0009,0x0027,"GEMS_IDEN_01",VR::SL,VM::VM1,"Image actual date",false },
  {0x0009,0x002f,"GEMS_IDEN_01",VR::SH,VM::VM1,"?",false },
  {0x0009,0x0030,"GEMS_IDEN_01",VR::SH,VM::VM1,"Service id",false },
  {0x0009,0x0031,"GEMS_IDEN_01",VR::SH,VM::VM1,"Mobile location number",false },
  {0x0009,0x00e2,"GEMS_IDEN_01",VR::LT,VM::VM1_n,"Image Type (ORIGINAL/PROSPECTIVE ...)",false },
  {0x0009,0x00e3,"GEMS_IDEN_01",VR::UI,VM::VM1,"Equipment UID",false },
  {0x0009,0x00e6,"GEMS_IDEN_01",VR::SH,VM::VM1,"Genesis Version - now",false },
  {0x0009,0x00e7,"GEMS_IDEN_01",VR::UL,VM::VM1,"Exam Record checksum",false },
  {0x0009,0x00e8,"GEMS_IDEN_01",VR::SH,VM::VM1,"Series Suite Id",false },
  {0x0009,0x00e9,"GEMS_IDEN_01",VR::SL,VM::VM1,"Actual series data time stamp",false },
  {0x0027,0x0006,"GEMS_IMAG_01",VR::SL,VM::VM1,"Image archive flag",false },
  {0x0027,0x0010,"GEMS_IMAG_01",VR::SS,VM::VM1,"Scout Type",false },
  {0x0027,0x001c,"GEMS_IMAG_01",VR::SL,VM::VM1,"Vma mamp",false },
  {0x0027,0x001d,"GEMS_IMAG_01",VR::SS,VM::VM1,"Vma phase",false },
  {0x0027,0x001e,"GEMS_IMAG_01",VR::SL,VM::VM1,"Vma mod",false },
  {0x0027,0x001f,"GEMS_IMAG_01",VR::SL,VM::VM1,"Vma clip",false },
  {0x0027,0x0020,"GEMS_IMAG_01",VR::SS,VM::VM1,"Smart scan ON/OFF flag",false },
  {0x0027,0x0030,"GEMS_IMAG_01",VR::SH,VM::VM1,"Foreign Image Revision",false },
  {0x0027,0x0031,"GEMS_IMAG_01",VR::SS,VM::VM1,"Imaging Mode",false },
  {0x0027,0x0032,"GEMS_IMAG_01",VR::SS,VM::VM1,"Pulse Sequence",false },
  {0x0027,0x0033,"GEMS_IMAG_01",VR::SL,VM::VM1,"Imaging Options",false }, // AMIInvalidPrivateDefinedLengthSQasUN.dcm => UL ?
  {0x0027,0x0035,"GEMS_IMAG_01",VR::SS,VM::VM1,"Plane Type",false },
  {0x0027,0x0036,"GEMS_IMAG_01",VR::SL,VM::VM1,"Oblique Plane",false },
  {0x0027,0x0040,"GEMS_IMAG_01",VR::SH,VM::VM1,"RAS letter of image location",false },
  {0x0027,0x0041,"GEMS_IMAG_01",VR::FL,VM::VM1,"Image location",false },
  {0x0027,0x0042,"GEMS_IMAG_01",VR::FL,VM::VM1,"Center R coord of plane image",false },
  {0x0027,0x0043,"GEMS_IMAG_01",VR::FL,VM::VM1,"Center A coord of plane image",false },
  {0x0027,0x0044,"GEMS_IMAG_01",VR::FL,VM::VM1,"Center S coord of plane image",false },
  {0x0027,0x0045,"GEMS_IMAG_01",VR::FL,VM::VM1,"Normal R coord",false },
  {0x0027,0x0046,"GEMS_IMAG_01",VR::FL,VM::VM1,"Normal A coord",false },
  {0x0027,0x0047,"GEMS_IMAG_01",VR::FL,VM::VM1,"Normal S coord",false },
  {0x0027,0x0048,"GEMS_IMAG_01",VR::FL,VM::VM1,"R Coord of Top Right Corner",false },
  {0x0027,0x0049,"GEMS_IMAG_01",VR::FL,VM::VM1,"A Coord of Top Right Corner",false },
  {0x0027,0x004a,"GEMS_IMAG_01",VR::FL,VM::VM1,"S Coord of Top Right Corner",false },
  {0x0027,0x004b,"GEMS_IMAG_01",VR::FL,VM::VM1,"R Coord of Bottom Right Corner",false },
  {0x0027,0x004c,"GEMS_IMAG_01",VR::FL,VM::VM1,"A Coord of Bottom Right Corner",false },
  {0x0027,0x004d,"GEMS_IMAG_01",VR::FL,VM::VM1,"S Coord of Bottom Right Corner",false },
  {0x0027,0x0050,"GEMS_IMAG_01",VR::FL,VM::VM1,"Scan Start Location",false },
  {0x0027,0x0051,"GEMS_IMAG_01",VR::FL,VM::VM1,"Scan End Location",false },
  {0x0027,0x0052,"GEMS_IMAG_01",VR::SH,VM::VM1,"RAS letter for side of image",false },
  {0x0027,0x0053,"GEMS_IMAG_01",VR::SH,VM::VM1,"RAS letter for anterior/posterior",false },
  {0x0027,0x0054,"GEMS_IMAG_01",VR::SH,VM::VM1,"RAS letter for scout start loc",false },
  {0x0027,0x0055,"GEMS_IMAG_01",VR::SH,VM::VM1,"RAS letter for scout end loc",false },
  {0x0027,0x0060,"GEMS_IMAG_01",VR::FL,VM::VM1,"Image dimension - X",false },
  {0x0027,0x0061,"GEMS_IMAG_01",VR::FL,VM::VM1,"Image dimension - Y",false },
  {0x0027,0x0062,"GEMS_IMAG_01",VR::FL,VM::VM1,"Number of Excitations",false },
  {0x0029,0x0015,"GEMS_IMAG_01",VR::SL,VM::VM1,"Lower range of Pixels1",false },
  {0x0029,0x0016,"GEMS_IMAG_01",VR::SL,VM::VM1,"Lower range of Pixels1",false },
  {0x0029,0x0017,"GEMS_IMAG_01",VR::SL,VM::VM1,"Lower range of Pixels2",false },
  {0x0029,0x0018,"GEMS_IMAG_01",VR::SL,VM::VM1,"Upper range of Pixels2",false },
  {0x0029,0x001a,"GEMS_IMAG_01",VR::SL,VM::VM1,"Len of tot hdr in bytes No longer supported",false },
  {0x0029,0x0026,"GEMS_IMAG_01",VR::SS,VM::VM1,"Version of the hdr struct",false },
  {0x0029,0x0034,"GEMS_IMAG_01",VR::SL,VM::VM1,"Advantage comp. Overflow",false },
  {0x0029,0x0035,"GEMS_IMAG_01",VR::SL,VM::VM1,"Advantage comp. Underflow",false },
  {0x0029,0x0004,"GEMS_IMPS_01",VR::SL,VM::VM1,"Lower range of Pixels1",false },
  {0x0029,0x0005,"GEMS_IMPS_01",VR::DS,VM::VM1,"Lower range of Pixels1",false },
  {0x0029,0x0006,"GEMS_IMPS_01",VR::DS,VM::VM1,"Lower range of Pixels1",false },
  {0x0029,0x0007,"GEMS_IMPS_01",VR::SL,VM::VM1,"Lower range of Pixels1",false },
  {0x0029,0x0008,"GEMS_IMPS_01",VR::SH,VM::VM1,"Lower range of Pixels1",false },
  {0x0029,0x0009,"GEMS_IMPS_01",VR::SH,VM::VM1,"Lower range of Pixels1",false },
  {0x0029,0x000a,"GEMS_IMPS_01",VR::SS,VM::VM1,"Lower range of Pixels1",false },
  {0x0029,0x0015,"GEMS_IMPS_01",VR::SL,VM::VM1,"Lower range of Pixels1",false },
  {0x0029,0x0016,"GEMS_IMPS_01",VR::SL,VM::VM1,"Upper range of Pixels1",false },
  {0x0029,0x0017,"GEMS_IMPS_01",VR::SL,VM::VM1,"Lower range of Pixels2",false },
  {0x0029,0x0018,"GEMS_IMPS_01",VR::SL,VM::VM1,"Upper range of Pixels2",false },
  {0x0029,0x001a,"GEMS_IMPS_01",VR::SL,VM::VM1,"Len of tot hdr in bytes",false },
  {0x0029,0x0026,"GEMS_IMPS_01",VR::SS,VM::VM1,"Version of the hdr struct",false },
  {0x0029,0x0034,"GEMS_IMPS_01",VR::SL,VM::VM1,"Advantage comp. Overflow",false },
  {0x0029,0x0035,"GEMS_IMPS_01",VR::SL,VM::VM1,"Advantage comp. Underflow",false },
  {0x0043,0x0013,"GEMS_IMPS_01",VR::SS,VM::VM5,"Decon kernel parameters",false },
  {0x0047,0x0001,"GEMS_IQTB_IDEN_47",VR::SH,VM::VM1,"Image Path",false },
  {0x0047,0x0002,"GEMS_IQTB_IDEN_47",VR::UL,VM::VM1,"Image Date/Time",false },
  {0x0047,0x0003,"GEMS_IQTB_IDEN_47",VR::UL,VM::VM1,"Queue Sequence Number",false },
  {0x0047,0x0004,"GEMS_IQTB_IDEN_47",VR::UL,VM::VM1,"Recon Sequence ID",false },
  {0x0045,0x0011,"GEMS_IT_US_REPORT",VR::OW,VM::VM1,"Vivid excel file",false },
  {0x0045,0x0012,"GEMS_IT_US_REPORT",VR::OW,VM::VM1,"Vivid CHM file",false },
  {0x0045,0x0013,"GEMS_IT_US_REPORT",VR::OW,VM::VM1,"Vivid PDF file",false },
  {0x7003,0x0001,"GEMS_LUNAR_RAW",VR::ST,VM::VM1,"enCORE File Name",false },
  {0x7003,0x0002,"GEMS_LUNAR_RAW",VR::OB,VM::VM1,"enCORE File Data",false },
  {0x7003,0x0003,"GEMS_LUNAR_RAW",VR::UL,VM::VM1,"enCORE File Length",false },
  {0x7003,0x0004,"GEMS_LUNAR_RAW",VR::LO,VM::VM1,"enCORE File Modified Time",false },
  {0x7001,0x0001,"GEMS_MR_RAW_01",VR::OB,VM::VM1,"rdb_hdr_rec",false },
  {0x7001,0x0002,"GEMS_MR_RAW_01",VR::OB,VM::VM1,"rdb_hdr_per_pass_tab",false },
  {0x7001,0x0003,"GEMS_MR_RAW_01",VR::OB,VM::VM1,"rdb_hdr_unlock_raw",false },
  {0x7001,0x0004,"GEMS_MR_RAW_01",VR::OB,VM::VM1,"rdb_hdr_data_acq_tab",false },
  {0x7001,0x0005,"GEMS_MR_RAW_01",VR::OB,VM::VM1,"rdb_hdr_nex_tab",false },
  {0x7001,0x0006,"GEMS_MR_RAW_01",VR::OB,VM::VM1,"rdb_hdr_nex_abort_tab",false },
  {0x7001,0x0007,"GEMS_MR_RAW_01",VR::OB,VM::VM1,"rdb_hdr_tool",false },
  {0x7001,0x0008,"GEMS_MR_RAW_01",VR::OB,VM::VM1,"rdb_raw_data",false },
  {0x7001,0x0009,"GEMS_MR_RAW_01",VR::OB,VM::VM1,"SSP save",false },
  {0x7001,0x000a,"GEMS_MR_RAW_01",VR::OB,VM::VM1,"UDA save",false },
  {0x7001,0x000b,"GEMS_MR_RAW_01",VR::OB,VM::VM1,"rdb_chemsat_data",false },
  {0x0043,0x0001,"GEMS_PARM_01",VR::SS,VM::VM1,"Bitmap of prescan options",false },
  {0x0043,0x0002,"GEMS_PARM_01",VR::SS,VM::VM1,"Gradient offset in X",false },
  {0x0043,0x0003,"GEMS_PARM_01",VR::SS,VM::VM1,"Gradient offset in Y",false },
  {0x0043,0x0004,"GEMS_PARM_01",VR::SS,VM::VM1,"Gradient offset in Z",false },
  {0x0043,0x0005,"GEMS_PARM_01",VR::SS,VM::VM1,"Img is original or unoriginal",false },
  {0x0043,0x0006,"GEMS_PARM_01",VR::SS,VM::VM1,"Number of EPI shots",false },
  {0x0043,0x0007,"GEMS_PARM_01",VR::SS,VM::VM1,"Views per segment",false },
  {0x0043,0x0008,"GEMS_PARM_01",VR::SS,VM::VM1,"Respiratory rate, bpm",false },
  {0x0043,0x0009,"GEMS_PARM_01",VR::SS,VM::VM1,"Respiratory trigger point",false },
  {0x0043,0x000a,"GEMS_PARM_01",VR::SS,VM::VM1,"Type of receiver used",false },
  {0x0043,0x000b,"GEMS_PARM_01",VR::DS,VM::VM1,"DB/dt Peak rate of change of gradient field",false },
  {0x0043,0x000c,"GEMS_PARM_01",VR::DS,VM::VM1,"dB/dt Limits in units of percent",false },
  {0x0043,0x000d,"GEMS_PARM_01",VR::DS,VM::VM1,"PSD estimated limit",false },
  {0x0043,0x000e,"GEMS_PARM_01",VR::DS,VM::VM1,"PSD estimated limit in tesla per second",false },
  {0x0043,0x000f,"GEMS_PARM_01",VR::DS,VM::VM1,"Saravghead",false },
  // (0043,1010) SS (US) 0                                    # 2,1 Window value
  // however http://www.gemed.com.ar/usen/interoperability/dicom/docs/DOC0636565r4_CS.pdf
  // indicates it to be US ! Is it a dual SS/US ?
  {0x0043,0x0010,"GEMS_PARM_01",VR::US,VM::VM1,"Window value",false },
  {0x0043,0x0011,"GEMS_PARM_01",VR::US,VM::VM1,"Total input views",false },
  {0x0043,0x0012,"GEMS_PARM_01",VR::SS,VM::VM3,"X-ray chain",false },
  {0x0043,0x0013,"GEMS_PARM_01",VR::SS,VM::VM5,"Decon kernel parameters",false },
  {0x0043,0x0014,"GEMS_PARM_01",VR::SS,VM::VM3,"Calibration parameters",false },
  {0x0043,0x0015,"GEMS_PARM_01",VR::SS,VM::VM3,"Total output views",false },
  {0x0043,0x0016,"GEMS_PARM_01",VR::SS,VM::VM5,"Number of overranges",false },
  {0x0043,0x0017,"GEMS_PARM_01",VR::DS,VM::VM1,"IBH image scale factors",false },
  {0x0043,0x0018,"GEMS_PARM_01",VR::DS,VM::VM3,"BBH coefficients",false },
  {0x0043,0x0019,"GEMS_PARM_01",VR::SS,VM::VM1,"Number of BBH chains to blend",false },
  {0x0043,0x001a,"GEMS_PARM_01",VR::SL,VM::VM1,"Starting channel number",false },
  {0x0043,0x001b,"GEMS_PARM_01",VR::SS,VM::VM1,"Ppscan parameters",false },
  {0x0043,0x001c,"GEMS_PARM_01",VR::SS,VM::VM1,"GE image integrity",false },
  {0x0043,0x001d,"GEMS_PARM_01",VR::SS,VM::VM1,"Level value",false },
  {0x0043,0x001e,"GEMS_PARM_01",VR::DS,VM::VM1,"Delta Start Time [msec]",false },
  {0x0043,0x001f,"GEMS_PARM_01",VR::SL,VM::VM1,"Max overranges in a view",false },
  {0x0043,0x0020,"GEMS_PARM_01",VR::DS,VM::VM1,"Avg overranges all views",false },
  {0x0043,0x0021,"GEMS_PARM_01",VR::SS,VM::VM1,"Corrected after glow terms",false },
  {0x0043,0x0025,"GEMS_PARM_01",VR::SS,VM::VM6,"Reference channels",false },
  {0x0043,0x0026,"GEMS_PARM_01",VR::US,VM::VM6,"No views ref chans blocked",false },
  {0x0043,0x0027,"GEMS_PARM_01",VR::SH,VM::VM1,"Scan Pitch Ratio",false },
  {0x0043,0x0028,"GEMS_PARM_01",VR::OB,VM::VM1,"Unique image iden",false },
  {0x0043,0x0029,"GEMS_PARM_01",VR::OB,VM::VM1,"Histogram tables",false },
  {0x0043,0x002a,"GEMS_PARM_01",VR::OB,VM::VM1,"User defined data",false },
  {0x0043,0x002b,"GEMS_PARM_01",VR::SS,VM::VM4,"Private Scan Options",false },
  {0x0043,0x002c,"GEMS_PARM_01",VR::SS,VM::VM1,"Effective echo spacing",false },
  {0x0043,0x002d,"GEMS_PARM_01",VR::SH,VM::VM1,"Filter Mode (String slop field 1 in legacy GE MR",false },
  {0x0043,0x002e,"GEMS_PARM_01",VR::SH,VM::VM1,"String slop field 2",false },
  {0x0043,0x002f,"GEMS_PARM_01",VR::SS,VM::VM1,"Image Type (real, imaginary, phase, magnitude)",false },
  {0x0043,0x0030,"GEMS_PARM_01",VR::SS,VM::VM1,"Vas collapse flag",false },
  {0x0043,0x0031,"GEMS_PARM_01",VR::DS,VM::VM2,"Recon Center Coordinates",false },
  {0x0043,0x0032,"GEMS_PARM_01",VR::SS,VM::VM1,"Vas flags",false },
  {0x0043,0x0033,"GEMS_PARM_01",VR::FL,VM::VM1,"Neg_scanspacing",false },
  {0x0043,0x0034,"GEMS_PARM_01",VR::IS,VM::VM1,"Offset Frequency",false },
  {0x0043,0x0035,"GEMS_PARM_01",VR::UL,VM::VM1,"User_usage_tag",false },
  {0x0043,0x0036,"GEMS_PARM_01",VR::UL,VM::VM1,"User_fill_map_MSW",false },
  {0x0043,0x0037,"GEMS_PARM_01",VR::UL,VM::VM1,"User_fill_map_LSW",false },
  {0x0043,0x0038,"GEMS_PARM_01",VR::FL,VM::VM24,"User data 25...User data 48 {User48=Effective Resolution for spiral}",false },
  {0x0043,0x0039,"GEMS_PARM_01",VR::IS,VM::VM4,"Slop_int_6... slop_int_9 (B_value/?/?/?)",false },
  {0x0043,0x0040,"GEMS_PARM_01",VR::FL,VM::VM4,"Trigger on position",false },
  {0x0043,0x0041,"GEMS_PARM_01",VR::FL,VM::VM4,"Degree of rotation",false },
  {0x0043,0x0042,"GEMS_PARM_01",VR::SL,VM::VM4,"DAS trigger source",false },
  {0x0043,0x0043,"GEMS_PARM_01",VR::SL,VM::VM4,"DAS fpa gain",false },
  {0x0043,0x0044,"GEMS_PARM_01",VR::SL,VM::VM4,"DAS output source",false },
  {0x0043,0x0045,"GEMS_PARM_01",VR::SL,VM::VM4,"DAS ad input",false },
  {0x0043,0x0046,"GEMS_PARM_01",VR::SL,VM::VM4,"DAS cal mode",false },
  {0x0043,0x0047,"GEMS_PARM_01",VR::SL,VM::VM4,"DAS cal frequency",false },
  {0x0043,0x0048,"GEMS_PARM_01",VR::SL,VM::VM4,"DAS reg xm",false },
  {0x0043,0x0049,"GEMS_PARM_01",VR::SL,VM::VM4,"DAS auto zero",false },
  {0x0043,0x004a,"GEMS_PARM_01",VR::SS,VM::VM4,"Starting channel of view",false },
  {0x0043,0x004b,"GEMS_PARM_01",VR::SL,VM::VM4,"DAS xm pattern",false },
  {0x0043,0x004c,"GEMS_PARM_01",VR::SS,VM::VM4,"TGGC trigger mode",false },
  {0x0043,0x004d,"GEMS_PARM_01",VR::FL,VM::VM4,"Start scan to X-ray on delay",false },
  {0x0043,0x004e,"GEMS_PARM_01",VR::FL,VM::VM4,"Duration of X-ray on",false },
  {0x0043,0x0060,"GEMS_PARM_01",VR::IS,VM::VM8,"Slop_int_10...slop_int_17",false },
  {0x0043,0x0061,"GEMS_PARM_01",VR::UI,VM::VM1,"Scanner Study Entity UID",false },
  {0x0043,0x0062,"GEMS_PARM_01",VR::SH,VM::VM1,"Scanner Study ID",false },
  {0x0043,0x0063,"GEMS_PARM_01",VR::SH,VM::VM1,"Raw Data ID",false },
  {0x0043,0x0064,"GEMS_PARM_01",VR::CS,VM::VM1_n,"Image Filter",false },
  {0x0043,0x0065,"GEMS_PARM_01",VR::US,VM::VM1,"Motion Correction Indicator",false },
  {0x0043,0x0066,"GEMS_PARM_01",VR::US,VM::VM1,"Helical Correction Indicator",false },
  {0x0043,0x0067,"GEMS_PARM_01",VR::US,VM::VM1,"IBO Correction Indicator",false },
  {0x0043,0x0068,"GEMS_PARM_01",VR::US,VM::VM1,"XT Correction Indicator",false },
  {0x0043,0x0069,"GEMS_PARM_01",VR::US,VM::VM1,"Q-cal Correction Indicator",false },
  {0x0043,0x006a,"GEMS_PARM_01",VR::US,VM::VM1,"AV Correction Indicator",false },
  {0x0043,0x006b,"GEMS_PARM_01",VR::US,VM::VM1,"L-MDK Correction Indicator",false },
  {0x0043,0x006c,"GEMS_PARM_01",VR::IS,VM::VM1,"Detector Row",false },
  {0x0043,0x006d,"GEMS_PARM_01",VR::DS,VM::VM1,"Area Size",false }, // Corrected IS -> DS
  {0x0043,0x006e,"GEMS_PARM_01",VR::SH,VM::VM1,"Auto mA Mode",false },
  {0x0043,0x006f,"GEMS_PARM_01",VR::DS,VM::VM3_4,"Scanner Table Entry (single gradient coil systems only)/Scanner Table Entry + Gradient Coil Selected",false },
  {0x0043,0x0070,"GEMS_PARM_01",VR::LO,VM::VM1,"Paradigm Name",false },
  {0x0043,0x0071,"GEMS_PARM_01",VR::ST,VM::VM1,"Paradigm Description",false },
  {0x0043,0x0072,"GEMS_PARM_01",VR::UI,VM::VM1,"Paradigm UID",false },
  {0x0043,0x0073,"GEMS_PARM_01",VR::US,VM::VM1,"Experiment Type",false },
  {0x0043,0x0074,"GEMS_PARM_01",VR::US,VM::VM1,"#rest volumes",false },
  {0x0043,0x0075,"GEMS_PARM_01",VR::US,VM::VM1,"#active volumes",false },
  {0x0043,0x0076,"GEMS_PARM_01",VR::US,VM::VM1,"#dummy scans",false },
  {0x0043,0x0077,"GEMS_PARM_01",VR::SH,VM::VM1,"Application Name",false },
  {0x0043,0x0078,"GEMS_PARM_01",VR::SH,VM::VM1,"Application Version",false },
  {0x0043,0x0079,"GEMS_PARM_01",VR::US,VM::VM1,"Slices Per Volume",false },
  {0x0043,0x007a,"GEMS_PARM_01",VR::US,VM::VM1,"Expected Time Points",false },
  {0x0043,0x007b,"GEMS_PARM_01",VR::FL,VM::VM1_n,"Regressor Values",false },
  {0x0043,0x007c,"GEMS_PARM_01",VR::FL,VM::VM1,"Delay after slice group",false },
  {0x0043,0x007d,"GEMS_PARM_01",VR::US,VM::VM1,"Recon mode flag word",false },
  {0x0043,0x007e,"GEMS_PARM_01",VR::LO,VM::VM1_n,"PACC specific information",false },
  {0x0043,0x007f,"GEMS_PARM_01",VR::DS,VM::VM1_n,"Reserved",false },
  {0x0043,0x0080,"GEMS_PARM_01",VR::LO,VM::VM1_n,"Coil ID Data",false },
  {0x0043,0x0081,"GEMS_PARM_01",VR::LO,VM::VM1,"GE Coil Name",false },
  {0x0043,0x0082,"GEMS_PARM_01",VR::LO,VM::VM1_n,"System Configuration Information",false },
  {0x0043,0x0083,"GEMS_PARM_01",VR::DS,VM::VM1_2,"Asset R Factors",false },
  {0x0043,0x0084,"GEMS_PARM_01",VR::LO,VM::VM1_n,"Additional Asset Data",false },
  {0x0043,0x0085,"GEMS_PARM_01",VR::UT,VM::VM1,"Debug Data (text format)",false },
  {0x0043,0x0086,"GEMS_PARM_01",VR::OB,VM::VM1,"Debug Data (binary format)",false },
  {0x0043,0x0087,"GEMS_PARM_01",VR::UT,VM::VM1,"Reserved",false },
  {0x0043,0x0088,"GEMS_PARM_01",VR::UI,VM::VM1,"PURE Acquisition Calibration Series UID",false },
  {0x0043,0x0089,"GEMS_PARM_01",VR::LO,VM::VM3,"Governing Body, dB/dt, and SAR definition",false },
  {0x0043,0x008a,"GEMS_PARM_01",VR::CS,VM::VM1,"Private In-Plane Phase Encoding Direction",false },
  {0x0043,0x008b,"GEMS_PARM_01",VR::OB,VM::VM1,"FMRI Binary Data Block",false },
  {0x0043,0x008c,"GEMS_PARM_01",VR::DS,VM::VM6,"Voxel Location",false },
  {0x0043,0x008d,"GEMS_PARM_01",VR::DS,VM::VM7_7n,"SAT Band Locations",false },
  {0x0043,0x008e,"GEMS_PARM_01",VR::DS,VM::VM3,"Spectro Prescan Values",false },
  {0x0043,0x008f,"GEMS_PARM_01",VR::DS,VM::VM3,"Spectro Parameters",false },
  {0x0043,0x0090,"GEMS_PARM_01",VR::LO,VM::VM1_n,"SAR Definition",false },
  {0x0043,0x0091,"GEMS_PARM_01",VR::DS,VM::VM1_n,"SAR value",false },
  {0x0043,0x0092,"GEMS_PARM_01",VR::LO,VM::VM1,"Image Error Text",false },
  {0x0043,0x0093,"GEMS_PARM_01",VR::DS,VM::VM1_n,"Spectro Quantitation Values",false },
  {0x0043,0x0094,"GEMS_PARM_01",VR::DS,VM::VM1_n,"Spectro Ratio Values",false },
  {0x0043,0x0095,"GEMS_PARM_01",VR::LO,VM::VM1,"Prescan Reuse String",false },
  {0x0043,0x0096,"GEMS_PARM_01",VR::CS,VM::VM1,"Content Qualification",false },
  {0x0043,0x0097,"GEMS_PARM_01",VR::LO,VM::VM1_n,"Image Filtering Parameters",false },
  {0x0043,0x0098,"GEMS_PARM_01",VR::UI,VM::VM1,"ASSET Acquisition Calibration Series UID",false },
  {0x0043,0x0099,"GEMS_PARM_01",VR::LO,VM::VM1_n,"Extended Options",false },
  {0x0043,0x009a,"GEMS_PARM_01",VR::IS,VM::VM1,"Rx Stack Identification",false },
  {0x0043,0x009b,"GEMS_PARM_01",VR::DS,VM::VM1,"NPW factor",false },
  {0x0043,0x009c,"GEMS_PARM_01",VR::OB,VM::VM1,"Research Tag 1",false },
  {0x0043,0x009d,"GEMS_PARM_01",VR::OB,VM::VM1,"Research Tag 2",false },
  {0x0043,0x009e,"GEMS_PARM_01",VR::OB,VM::VM1,"Research Tag 3",false },
  {0x0043,0x009f,"GEMS_PARM_01",VR::OB,VM::VM1,"Research Tag 4",false },
  {0x0043,0x00a0,"GEMS_PARM_01",VR::SQ,VM::VM1,"Spectroscopy Pixel Sequence",false },
  {0x0043,0x00a1,"GEMS_PARM_01",VR::SQ,VM::VM1,"Spectroscopy Default Display Sequence",false },
  {0x0043,0x00a2,"GEMS_PARM_01",VR::DS,VM::VM1_n,"MEG Data",false },
  {0x0043,0x00a3,"GEMS_PARM_01",VR::CS,VM::VM1,"ASL Contrast technique",false },
  {0x0043,0x00a4,"GEMS_PARM_01",VR::LO,VM::VM1,"Detailed text for ASL labeling technique",false },
  {0x0043,0x00a5,"GEMS_PARM_01",VR::IS,VM::VM1,"Duration of the label or control pulse",false },
  {0x0043,0x00a8,"GEMS_PARM_01",VR::DS,VM::VM3,"Dual Drive Mode, Amplitude Attenuation and Phase Offset ",false },
  {0x0043,0x00a9,"GEMS_PARM_01",VR::LO,VM::VM1,"?",false },
  {0x0043,0x00aa,"GEMS_PARM_01",VR::LO,VM::VM1_n,"Additional Filtering Parameters",false },
  {0x0043,0x00ab,"GEMS_PARM_01",VR::DS,VM::VM1_n,"Silenz Data",false },
  {0x0043,0x00ac,"GEMS_PARM_01",VR::DS,VM::VM1_n,"QMAP Delay Data (RESERVED for FUTURE USE)",false },
  {0x0043,0x00ad,"GEMS_PARM_01",VR::DS,VM::VM1_n,"Other Recovery Times Values",false },
  {0x0043,0x00ae,"GEMS_PARM_01",VR::LO,VM::VM1_n,"Other Recovery Times Labels",false },
  {0x0043,0x00af,"GEMS_PARM_01",VR::DS,VM::VM1_n,"Additional Echo Times",false },
  {0x0043,0x00b0,"GEMS_PARM_01",VR::FL,VM::VM1,"Rescan Time in Acquisition",false },
  {0x0043,0x00b1,"GEMS_PARM_01",VR::SS,VM::VM1,"Excitation Mode",false },
  {0x0043,0x00b2,"GEMS_PARM_01",VR::LO,VM::VM1_n,"MR Table Position Information",false },
  {0x0043,0x00b3,"GEMS_PARM_01",VR::DS,VM::VM1_n,"Advanced Eddy Correction",false },
  {0x0043,0x00b5,"GEMS_PARM_01",VR::LO,VM::VM1,"?",false },
  {0x0043,0x00b6,"GEMS_PARM_01",VR::LO,VM::VM1_n,"Multiband Parameters",false },
  {0x0043,0x00b7,"GEMS_PARM_01",VR::LO,VM::VM4,"Compressed Sensing Parameters",false },
  {0x0043,0x00b8,"GEMS_PARM_01",VR::DS,VM::VM1_n,"Grad Comp Parameters",false },
  {0x0043,0x00b9,"GEMS_PARM_01",VR::LO,VM::VM1,"Parallel Transmit Information",false },
  {0x0043,0x00ba,"GEMS_PARM_01",VR::DS,VM::VM1,"Echo Spacing",false },
  {0x0043,0x00bb,"GEMS_PARM_01",VR::LO,VM::VM1_n,"Pixel Information",false },
  {0x0043,0x00bc,"GEMS_PARM_01",VR::IS,VM::VM1,"Heart Beats pattern",false },
  {0x0043,0x00c0,"GEMS_PARM_01",VR::DS,VM::VM1_n,"?",false },
  {0x0043,0x00c1,"GEMS_PARM_01",VR::LO,VM::VM1,"?",false },
  {0x0043,0x00c2,"GEMS_PARM_01",VR::LO,VM::VM1,"?",false },
  {0x0043,0x00c3,"GEMS_PARM_01",VR::LO,VM::VM1,"?some kind of xml?",false },
  {0x0043,0x00c4,"GEMS_PARM_01",VR::LO,VM::VM1,"?some kind of xml?",false },
  {0x0043,0x00c5,"GEMS_PARM_01",VR::LO,VM::VM1,"?",false },
  {0x0043,0x00c6,"GEMS_PARM_01",VR::DS,VM::VM1_n,"?",false },
  {0x0043,0x00c7,"GEMS_PARM_01",VR::DS,VM::VM1_n,"?",false },
  {0x0043,0x00c8,"GEMS_PARM_01",VR::UL,VM::VM1,"?",false },
  {0x0043,0x00c9,"GEMS_PARM_01",VR::IS,VM::VM1,"?",false },
  {0x0043,0x00cc,"GEMS_PARM_01",VR::DS,VM::VM1,"?",false },
  {0x0043,0x00cd,"GEMS_PARM_01",VR::FL,VM::VM1,"?",false },
  {0x0051,0x000a,"GEMS_PARM_01",VR::SL,VM::VM1,"Store level of Functional Image",false },
  {0x0051,0x000b,"GEMS_PARM_01",VR::FL,VM::VM1,"Store B-Value with Functional Image",false },
  {0x2001,0x0010,"GEMS_PARM_01",VR::UI,VM::VM1,"DICOM Implementation UID",false },
  {0x2001,0x0011,"GEMS_PARM_01",VR::SH,VM::VM1,"DICOM Implementation Version",false },
  {0x2001,0x0012,"GEMS_PARM_01",VR::UI,VM::VM1,"Within-DICOM-Implementation SOP Instance UID",false },
  {0x2001,0x0013,"GEMS_PARM_01",VR::SH,VM::VM1,"Application Name",false },
  {0x2001,0x0014,"GEMS_PARM_01",VR::SH,VM::VM1,"Application Version",false },
  {0x2001,0x0015,"GEMS_PARM_01",VR::SH,VM::VM1,"Compatibility Version",false },
  {0x2001,0x0021,"GEMS_PARM_01",VR::UI,VM::VM1_n,"Referenced Series UID",false },
  {0x2001,0x0031,"GEMS_PARM_01",VR::US,VM::VM1,"Number of Objects Averaged",false },
  {0x2001,0x0041,"GEMS_PARM_01",VR::US,VM::VM1,"Number of Expected Time Points",false },
  {0x2001,0x0051,"GEMS_PARM_01",VR::US,VM::VM1,"Number of Slices Per Volume",false },
  {0x2001,0x0060,"GEMS_PARM_01",VR::US,VM::VM1,"BW Image Type",false },
  {0x2001,0x0061,"GEMS_PARM_01",VR::US,VM::VM1,"Experiment Type",false },
  {0x2001,0x0071,"GEMS_PARM_01",VR::UI,VM::VM1,"Paradigm UID",false },
  {0x2001,0x0072,"GEMS_PARM_01",VR::LO,VM::VM1,"Paradigm Name",false },
  {0x2001,0x0073,"GEMS_PARM_01",VR::ST,VM::VM1,"Paradigm Description",false },
  {0x2001,0x0080,"GEMS_PARM_01",VR::OB,VM::VM1,"Contrast",false },
  {0x2001,0x0081,"GEMS_PARM_01",VR::FL,VM::VM1_n,"Regressor Values",false },
  {0x2001,0x0086,"GEMS_PARM_01",VR::US,VM::VM1,"Number of Degrees of Freedom",false },
  {0x2001,0x008a,"GEMS_PARM_01",VR::FL,VM::VM1,"Z Threshold",false },
  {0x2001,0x008b,"GEMS_PARM_01",VR::FL,VM::VM1,"p Threshold",false },
  {0x2001,0x0090,"GEMS_PARM_01",VR::OB,VM::VM1,"Processing parameters",false },
  {0x2001,0x0091,"GEMS_PARM_01",VR::OB,VM::VM1,"Motion Plot",false },
  {0x2001,0x0092,"GEMS_PARM_01",VR::OB,VM::VM1,"ROIs",false },
  {0x2001,0x0093,"GEMS_PARM_01",VR::OB,VM::VM1,"Tracts",false },
  {0x2001,0x0094,"GEMS_PARM_01",VR::OB,VM::VM1,"Report",false },
  {0x2001,0x0095,"GEMS_PARM_01",VR::OB,VM::VM1,"Response Data",false },
  {0x2001,0x00a0,"GEMS_PARM_01",VR::FL,VM::VM1_n,"Motion Parameters",false },
  {0x2001,0x00a1,"GEMS_PARM_01",VR::FL,VM::VM1_n,"Registration Parameters",false },
  {0x2001,0x00a2,"GEMS_PARM_01",VR::FL,VM::VM1_n,"Subject Data",false },
  {0x2001,0x00b0,"GEMS_PARM_01",VR::OB,VM::VM1,"DTI Parameters",false },
  {0x2001,0x00c0,"GEMS_PARM_01",VR::OB,VM::VM1,"Paradigm Info",false },
  {0x0011,0x0010,"GEMS_PATI_01",VR::SS,VM::VM1,"Patient Status",false },
  {0x0011,0x0030,"GEMS_PATI_01",VR::DS,VM::VM1,"?",false },
  {0x0009,0x0001,"GEMS_PETD_01",VR::LO,VM::VM2,"GE Discovery PET Implementation Version Name",false },
  {0x0009,0x0002,"GEMS_PETD_01",VR::LO,VM::VM1,"PET patient_id",false },
  {0x0009,0x0003,"GEMS_PETD_01",VR::SH,VM::VM1,"PET compatible_version",false },
  {0x0009,0x0004,"GEMS_PETD_01",VR::SH,VM::VM1,"GE Advance Patient.software_version",false },
  {0x0009,0x0005,"GEMS_PETD_01",VR::DT,VM::VM1,"PET patient_datetime",false },
  {0x0009,0x0006,"GEMS_PETD_01",VR::SL,VM::VM1,"PET type",false },
  {0x0009,0x0007,"GEMS_PETD_01",VR::UI,VM::VM1,"PET exam_id",false },
  {0x0009,0x0008,"GEMS_PETD_01",VR::SH,VM::VM1,"PET compatible_version",false },
  {0x0009,0x0009,"GEMS_PETD_01",VR::SH,VM::VM1,"PET software_version",false },
  {0x0009,0x000a,"GEMS_PETD_01",VR::UI,VM::VM1,"PET scan_id",false },
  {0x0009,0x000b,"GEMS_PETD_01",VR::SH,VM::VM1,"PET compatible_version",false },
  {0x0009,0x000c,"GEMS_PETD_01",VR::SH,VM::VM1,"PET software_version",false },
  {0x0009,0x000d,"GEMS_PETD_01",VR::DT,VM::VM1,"PET scan_datetime",false },
  {0x0009,0x000e,"GEMS_PETD_01",VR::DT,VM::VM1,"PET scan_ready",false },
  {0x0009,0x000f,"GEMS_PETD_01",VR::ST,VM::VM1,"PET scan_description",false },
  {0x0009,0x0010,"GEMS_PETD_01",VR::LO,VM::VM1,"PET hospital_name",false },
  {0x0009,0x0011,"GEMS_PETD_01",VR::LO,VM::VM1,"PET scanner_desc",false },
  {0x0009,0x0012,"GEMS_PETD_01",VR::LO,VM::VM1,"PET manufacturer",false },
  {0x0009,0x0013,"GEMS_PETD_01",VR::UI,VM::VM1,"PET for_identifier",false },
  {0x0009,0x0014,"GEMS_PETD_01",VR::LO,VM::VM1,"PET landmark_name",false },
  {0x0009,0x0015,"GEMS_PETD_01",VR::SH,VM::VM1,"PET landmark_abbrev",false },
  {0x0009,0x0016,"GEMS_PETD_01",VR::SL,VM::VM1,"PET patient_position",false },
  {0x0009,0x0017,"GEMS_PETD_01",VR::SL,VM::VM1,"PET scan_perspective",false },
  {0x0009,0x0018,"GEMS_PETD_01",VR::SL,VM::VM1,"PET scan_type",false },
  {0x0009,0x0019,"GEMS_PETD_01",VR::SL,VM::VM1,"PET scan_mode",false },
  {0x0009,0x001a,"GEMS_PETD_01",VR::SL,VM::VM1,"PET start_condition",false },
  {0x0009,0x001b,"GEMS_PETD_01",VR::SL,VM::VM1,"PET start_cond_data",false },
  {0x0009,0x001c,"GEMS_PETD_01",VR::SL,VM::VM1,"PET sel_stop_cond",false },
  {0x0009,0x001d,"GEMS_PETD_01",VR::SL,VM::VM1,"PET sel_stop_cond_data",false },
  {0x0009,0x001e,"GEMS_PETD_01",VR::SL,VM::VM1,"PET collect_deadtime",false },
  {0x0009,0x001f,"GEMS_PETD_01",VR::SL,VM::VM1,"PET collect_singles",false },
  {0x0009,0x0020,"GEMS_PETD_01",VR::SL,VM::VM1,"PET collect_countrate",false },
  {0x0009,0x0021,"GEMS_PETD_01",VR::SL,VM::VM1,"PET countrate_period",false },
  {0x0009,0x0022,"GEMS_PETD_01",VR::SL,VM::VM1,"PET delayed_events",false },
  {0x0009,0x0023,"GEMS_PETD_01",VR::SL,VM::VM1,"PET delayed_bias",false },
  {0x0009,0x0024,"GEMS_PETD_01",VR::SL,VM::VM1,"PET word_size",false },
  {0x0009,0x0025,"GEMS_PETD_01",VR::SL,VM::VM1,"PET axial_acceptance",false },
  {0x0009,0x0026,"GEMS_PETD_01",VR::SL,VM::VM1,"PET axial_angle_3d",false },
  {0x0009,0x0027,"GEMS_PETD_01",VR::SL,VM::VM1,"PET theta_compression",false },
  {0x0009,0x0028,"GEMS_PETD_01",VR::SL,VM::VM1,"PET axial_compression",false },
  {0x0009,0x0029,"GEMS_PETD_01",VR::FL,VM::VM1,"PET gantry_tilt_angle",false },
  {0x0009,0x002a,"GEMS_PETD_01",VR::SL,VM::VM1,"PET collimation",false },
  {0x0009,0x002b,"GEMS_PETD_01",VR::SL,VM::VM1,"PET scan_fov",false },
  {0x0009,0x002c,"GEMS_PETD_01",VR::SL,VM::VM1,"PET axial_fov",false },
  {0x0009,0x002d,"GEMS_PETD_01",VR::SL,VM::VM1,"PET event_separation",false },
  {0x0009,0x002e,"GEMS_PETD_01",VR::SL,VM::VM1,"PET mask_width",false },
  {0x0009,0x002f,"GEMS_PETD_01",VR::SL,VM::VM1,"PET binning_mode",false },
  {0x0009,0x0030,"GEMS_PETD_01",VR::SL,VM::VM1,"PET trig_rej_method",false },
  {0x0009,0x0031,"GEMS_PETD_01",VR::SL,VM::VM1,"PET number_for_reject",false },
  {0x0009,0x0032,"GEMS_PETD_01",VR::SL,VM::VM1,"PET lower_reject_limit",false },
  {0x0009,0x0033,"GEMS_PETD_01",VR::SL,VM::VM1,"PET upper_reject_limit",false },
  {0x0009,0x0034,"GEMS_PETD_01",VR::SL,VM::VM1,"PET triggers_acquired",false },
  {0x0009,0x0035,"GEMS_PETD_01",VR::SL,VM::VM1,"PET triggers_rejected",false },
  {0x0009,0x0036,"GEMS_PETD_01",VR::LO,VM::VM1,"PET tracer_name",false },
  {0x0009,0x0037,"GEMS_PETD_01",VR::LO,VM::VM1,"PET batch_description",false },
  {0x0009,0x0038,"GEMS_PETD_01",VR::FL,VM::VM1,"PET tracer_activity",false },
  {0x0009,0x0039,"GEMS_PETD_01",VR::DT,VM::VM1,"PET meas_datetime",false },
  {0x0009,0x003a,"GEMS_PETD_01",VR::FL,VM::VM1,"PET pre_inj_volume",false },
  {0x0009,0x003b,"GEMS_PETD_01",VR::DT,VM::VM1,"PET admin_datetime",false },
  {0x0009,0x003c,"GEMS_PETD_01",VR::FL,VM::VM1,"PET post_inj_activity",false },
  {0x0009,0x003d,"GEMS_PETD_01",VR::DT,VM::VM1,"PET post_inj_datetime",false },
  {0x0009,0x003e,"GEMS_PETD_01",VR::SH,VM::VM1,"PET radionuclide_name",false },
  {0x0009,0x003f,"GEMS_PETD_01",VR::FL,VM::VM1,"PET half_life",false },
  {0x0009,0x0040,"GEMS_PETD_01",VR::FL,VM::VM1,"PET positron_fraction",false },
  {0x0009,0x0041,"GEMS_PETD_01",VR::SL,VM::VM1,"PET source1_holder",false },
  {0x0009,0x0042,"GEMS_PETD_01",VR::FL,VM::VM1,"PET source1_activity",false },
  {0x0009,0x0043,"GEMS_PETD_01",VR::DT,VM::VM1,"PET source1_meas_dt",false },
  {0x0009,0x0044,"GEMS_PETD_01",VR::SH,VM::VM1,"PET source1_radnuclide",false },
  {0x0009,0x0045,"GEMS_PETD_01",VR::FL,VM::VM1,"PET source1_half_life",false },
  {0x0009,0x0046,"GEMS_PETD_01",VR::SL,VM::VM1,"PET source2_holder",false },
  {0x0009,0x0047,"GEMS_PETD_01",VR::FL,VM::VM1,"PET source2_activity",false },
  {0x0009,0x0048,"GEMS_PETD_01",VR::DT,VM::VM1,"PET source2_meas_dt",false },
  {0x0009,0x0049,"GEMS_PETD_01",VR::SH,VM::VM1,"PET source2_radnuclide",false },
  {0x0009,0x004a,"GEMS_PETD_01",VR::FL,VM::VM1,"PET source2_half_life",false },
  {0x0009,0x004b,"GEMS_PETD_01",VR::SL,VM::VM1,"PET source_speed",false },
  {0x0009,0x004c,"GEMS_PETD_01",VR::FL,VM::VM1,"PET source_location",false },
  {0x0009,0x004d,"GEMS_PETD_01",VR::SL,VM::VM1,"PET emission_present",false },
  {0x0009,0x004e,"GEMS_PETD_01",VR::SL,VM::VM1,"PET lower_axial_acc",false },
  {0x0009,0x004f,"GEMS_PETD_01",VR::SL,VM::VM1,"PET upper_axial_acc",false },
  {0x0009,0x0050,"GEMS_PETD_01",VR::SL,VM::VM1,"PET lower_coinc_limit",false },
  {0x0009,0x0051,"GEMS_PETD_01",VR::SL,VM::VM1,"PET upper_coinc_limit",false },
  {0x0009,0x0052,"GEMS_PETD_01",VR::SL,VM::VM1,"PET coinc_delay_offset",false },
  {0x0009,0x0053,"GEMS_PETD_01",VR::SL,VM::VM1,"PET coinc_output_mode",false },
  {0x0009,0x0054,"GEMS_PETD_01",VR::SL,VM::VM1,"PET upper_energy_limit",false },
  {0x0009,0x0055,"GEMS_PETD_01",VR::SL,VM::VM1,"PET lower_energy_limit",false },
  {0x0009,0x0056,"GEMS_PETD_01",VR::UI,VM::VM1,"PET normal_cal_id",false },
  {0x0009,0x0057,"GEMS_PETD_01",VR::UI,VM::VM1,"PET normal_2d_cal_id",false },
  {0x0009,0x0058,"GEMS_PETD_01",VR::UI,VM::VM1,"PET blank_cal_id",false },
  {0x0009,0x0059,"GEMS_PETD_01",VR::UI,VM::VM1,"PET wc_cal_id",false },
  {0x0009,0x005a,"GEMS_PETD_01",VR::SL,VM::VM1,"PET derived",false },
  {0x0009,0x005b,"GEMS_PETD_01",VR::LO,VM::VM1,"PET contrast_agent",false },
  {0x0009,0x005c,"GEMS_PETD_01",VR::UI,VM::VM1,"PET frame_id",false },
  {0x0009,0x005d,"GEMS_PETD_01",VR::UI,VM::VM1,"PET scan_id",false },
  {0x0009,0x005e,"GEMS_PETD_01",VR::UI,VM::VM1,"PET exam_id",false },
  {0x0009,0x005f,"GEMS_PETD_01",VR::LO,VM::VM1,"PET patient_id",false },
  {0x0009,0x0060,"GEMS_PETD_01",VR::SH,VM::VM1,"PET compatible_version",false },
  {0x0009,0x0061,"GEMS_PETD_01",VR::SH,VM::VM1,"PET software_version",false },
  {0x0009,0x0062,"GEMS_PETD_01",VR::ST,VM::VM1,"PET where_is_frame",false },
  {0x0009,0x0063,"GEMS_PETD_01",VR::SL,VM::VM1,"PET frame_size",false },
  {0x0009,0x0064,"GEMS_PETD_01",VR::SL,VM::VM1,"PET file_exists",false },
  {0x0009,0x0065,"GEMS_PETD_01",VR::SL,VM::VM1,"PET patient_entry",false },
  {0x0009,0x0066,"GEMS_PETD_01",VR::FL,VM::VM1,"PET table_height",false },
  {0x0009,0x0067,"GEMS_PETD_01",VR::FL,VM::VM1,"PET table_z_position",false },
  {0x0009,0x0068,"GEMS_PETD_01",VR::DT,VM::VM1,"PET landmark_datetime",false },
  {0x0009,0x0069,"GEMS_PETD_01",VR::SL,VM::VM1,"PET slice_count",false },
  {0x0009,0x006a,"GEMS_PETD_01",VR::FL,VM::VM1,"PET start_location",false },
  {0x0009,0x006b,"GEMS_PETD_01",VR::SL,VM::VM1,"PET acq_delay",false },
  {0x0009,0x006c,"GEMS_PETD_01",VR::DT,VM::VM1,"PET acq_start",false },
  {0x0009,0x006d,"GEMS_PETD_01",VR::SL,VM::VM1,"PET acq_duration",false },
  {0x0009,0x006e,"GEMS_PETD_01",VR::SL,VM::VM1,"PET acq_bin_dur",false },
  {0x0009,0x006f,"GEMS_PETD_01",VR::SL,VM::VM1,"PET acq_bin_start",false },
  {0x0009,0x0070,"GEMS_PETD_01",VR::SL,VM::VM1,"PET actual_stop_cond",false },
  {0x0009,0x0071,"GEMS_PETD_01",VR::FD,VM::VM1,"PET total_prompts",false },
  {0x0009,0x0072,"GEMS_PETD_01",VR::FD,VM::VM1,"PET total_delays",false },
  {0x0009,0x0073,"GEMS_PETD_01",VR::SL,VM::VM1,"PET frame_valid",false },
  {0x0009,0x0074,"GEMS_PETD_01",VR::SL,VM::VM1,"PET validity_info",false },
  {0x0009,0x0075,"GEMS_PETD_01",VR::SL,VM::VM1,"PET archived",false },
  {0x0009,0x0076,"GEMS_PETD_01",VR::SL,VM::VM1,"PET compression",false },
  {0x0009,0x0077,"GEMS_PETD_01",VR::SL,VM::VM1,"PET uncompressed_size",false },
  {0x0009,0x0078,"GEMS_PETD_01",VR::SL,VM::VM1,"PET accum_bin_dur",false },
  {0x0009,0x0079,"GEMS_PETD_01",VR::SH,VM::VM1,"PET compatible_version",false },
  {0x0009,0x007a,"GEMS_PETD_01",VR::SH,VM::VM1,"PET software_version",false },
  {0x0009,0x007b,"GEMS_PETD_01",VR::DT,VM::VM1,"PET is_datetime",false },
  {0x0009,0x007c,"GEMS_PETD_01",VR::SL,VM::VM1,"PET is_source",false },
  {0x0009,0x007d,"GEMS_PETD_01",VR::SL,VM::VM1,"PET is_contents",false },
  {0x0009,0x007e,"GEMS_PETD_01",VR::SL,VM::VM1,"PET is_type",false },
  {0x0009,0x007f,"GEMS_PETD_01",VR::DS,VM::VM1_n,"PET is_reference",false },
  {0x0009,0x0080,"GEMS_PETD_01",VR::SL,VM::VM1,"PET multi_patient",false },
  {0x0009,0x0081,"GEMS_PETD_01",VR::SL,VM::VM1,"PET number_of_normals",false },
  {0x0009,0x0082,"GEMS_PETD_01",VR::UI,VM::VM1,"PET color_map_id",false },
  {0x0009,0x0083,"GEMS_PETD_01",VR::SL,VM::VM1,"PET window_level_type",false },
  {0x0009,0x0084,"GEMS_PETD_01",VR::FL,VM::VM1,"PET rotate",false },
  {0x0009,0x0085,"GEMS_PETD_01",VR::SL,VM::VM1,"PET flip",false },
  {0x0009,0x0086,"GEMS_PETD_01",VR::FL,VM::VM1,"PET zoom",false },
  {0x0009,0x0087,"GEMS_PETD_01",VR::SL,VM::VM1,"PET pan_x",false },
  {0x0009,0x0088,"GEMS_PETD_01",VR::SL,VM::VM1,"PET pan_y",false },
  {0x0009,0x0089,"GEMS_PETD_01",VR::FL,VM::VM1,"PET window_level_min",false },
  {0x0009,0x008a,"GEMS_PETD_01",VR::FL,VM::VM1,"PET window_level_max",false },
  {0x0009,0x008b,"GEMS_PETD_01",VR::SL,VM::VM1,"PET recon_method",false },
  {0x0009,0x008c,"GEMS_PETD_01",VR::SL,VM::VM1,"PET attenuation",false },
  {0x0009,0x008d,"GEMS_PETD_01",VR::FL,VM::VM1,"PET atten_coefficient",false },
  {0x0009,0x008e,"GEMS_PETD_01",VR::SL,VM::VM1,"PET bp_filter",false },
  {0x0009,0x008f,"GEMS_PETD_01",VR::FL,VM::VM1,"PET bp_filter_cutoff",false },
  {0x0009,0x0090,"GEMS_PETD_01",VR::SL,VM::VM1,"PET bp_filter_order",false },
  {0x0009,0x0091,"GEMS_PETD_01",VR::FL,VM::VM1,"PET bp_center_l",false },
  {0x0009,0x0092,"GEMS_PETD_01",VR::FL,VM::VM1,"PET bp_center_p",false },
  {0x0009,0x0093,"GEMS_PETD_01",VR::SL,VM::VM1,"PET atten_smooth",false },
  {0x0009,0x0094,"GEMS_PETD_01",VR::SL,VM::VM1,"PET atten_smooth_param",false },
  {0x0009,0x0095,"GEMS_PETD_01",VR::SL,VM::VM1,"PET angle_smooth_param",false },
  {0x0009,0x0096,"GEMS_PETD_01",VR::UI,VM::VM1,"PET wellcountercal_id",false },
  {0x0009,0x0097,"GEMS_PETD_01",VR::UI,VM::VM1,"PET trans_scan_id",false },
  {0x0009,0x0098,"GEMS_PETD_01",VR::UI,VM::VM1,"PET norm_cal_id",false },
  {0x0009,0x0099,"GEMS_PETD_01",VR::UI,VM::VM1,"PET blnk_cal_id",false },
  {0x0009,0x009a,"GEMS_PETD_01",VR::FL,VM::VM1,"PET cac_edge_threshold",false },
  {0x0009,0x009b,"GEMS_PETD_01",VR::FL,VM::VM1,"PET cac_skull_offset",false },
  {0x0009,0x009c,"GEMS_PETD_01",VR::UI,VM::VM1,"PET emiss_sub_id",false },
  {0x0009,0x009d,"GEMS_PETD_01",VR::SS,VM::VM1,"PET radial_filter_3d",false },
  {0x0009,0x009e,"GEMS_PETD_01",VR::FL,VM::VM1,"PET radial_cutoff_3d",false },
  {0x0009,0x009f,"GEMS_PETD_01",VR::SL,VM::VM1,"PET axial_filter_3d",false },
  {0x0009,0x00a0,"GEMS_PETD_01",VR::FL,VM::VM1,"PET axial_cutoff_3d",false },
  {0x0009,0x00a1,"GEMS_PETD_01",VR::FL,VM::VM1,"PET axial_start",false },
  {0x0009,0x00a2,"GEMS_PETD_01",VR::FL,VM::VM1,"PET axial_spacing",false },
  {0x0009,0x00a3,"GEMS_PETD_01",VR::SL,VM::VM1,"PET axial_angles_used",false },
  {0x0009,0x00a4,"GEMS_PETD_01",VR::SH,VM::VM1,"PET compatible_version",false },
  {0x0009,0x00a5,"GEMS_PETD_01",VR::SH,VM::VM1,"PET software_version",false },
  {0x0009,0x00a6,"GEMS_PETD_01",VR::SL,VM::VM1,"PET slice_number",false },
  {0x0009,0x00a7,"GEMS_PETD_01",VR::FL,VM::VM1,"PET total_counts",false },
  {0x0009,0x00a8,"GEMS_PETD_01",VR::OB,VM::VM1,"PET other_atts",false },
  {0x0009,0x00a9,"GEMS_PETD_01",VR::SL,VM::VM1,"PET other_atts_size",false },
  {0x0009,0x00aa,"GEMS_PETD_01",VR::SL,VM::VM1,"PET archived",false },
  {0x0009,0x00ab,"GEMS_PETD_01",VR::FL,VM::VM1,"PET bp_center_x",false },
  {0x0009,0x00ac,"GEMS_PETD_01",VR::FL,VM::VM1,"PET bp_center_y",false },
  {0x0009,0x00ad,"GEMS_PETD_01",VR::UI,VM::VM1,"PET trans_frame_id",false },
  {0x0009,0x00ae,"GEMS_PETD_01",VR::UI,VM::VM1,"PET tpluse_frame_id",false },
  {0x0009,0x00b0,"GEMS_PETD_01",VR::US,VM::VM1,"?",false },
  {0x0009,0x00b1,"GEMS_PETD_01",VR::FL,VM::VM1,"PET profile_spacing",false },
  {0x0009,0x00b2,"GEMS_PETD_01",VR::SS,VM::VM1,"PET ir_num_iterations",false },
  {0x0009,0x00b3,"GEMS_PETD_01",VR::SS,VM::VM1,"PET ir_num_subsets",false },
  {0x0009,0x00b4,"GEMS_PETD_01",VR::FL,VM::VM1,"PET ir_recon_fov",false },
  {0x0009,0x00b5,"GEMS_PETD_01",VR::SS,VM::VM1,"PET ir_corr_model",false },
  {0x0009,0x00b6,"GEMS_PETD_01",VR::SS,VM::VM1,"PET ir_loop_filter",false },
  {0x0009,0x00b7,"GEMS_PETD_01",VR::FL,VM::VM1,"PET ir_pre_filt_parm",false },
  {0x0009,0x00b8,"GEMS_PETD_01",VR::FL,VM::VM1,"PET ir_loop_filt_parm",false },
  {0x0009,0x00b9,"GEMS_PETD_01",VR::FL,VM::VM1,"PET response_filt_parm",false },
  {0x0009,0x00ba,"GEMS_PETD_01",VR::SS,VM::VM1,"PET post_filter",false },
  {0x0009,0x00bb,"GEMS_PETD_01",VR::FL,VM::VM1,"PET post_filt_parm",false },
  {0x0009,0x00bc,"GEMS_PETD_01",VR::SS,VM::VM1,"PET ir_regularize",false },
  {0x0009,0x00bd,"GEMS_PETD_01",VR::FL,VM::VM1,"PET regularize_parm",false },
  {0x0009,0x00be,"GEMS_PETD_01",VR::SS,VM::VM1,"PET ac_bp_filter",false },
  {0x0009,0x00bf,"GEMS_PETD_01",VR::FL,VM::VM1,"PET ac_bp_filt_cut_off",false },
  {0x0009,0x00c0,"GEMS_PETD_01",VR::SL,VM::VM1,"PET ac_bp_filt_order",false },
  {0x0009,0x00c1,"GEMS_PETD_01",VR::SS,VM::VM1,"PET ac_img_smooth",false },
  {0x0009,0x00c2,"GEMS_PETD_01",VR::FL,VM::VM1,"PET ac_img_smooth_parm",false },
  {0x0009,0x00c3,"GEMS_PETD_01",VR::SL,VM::VM1,"PET scatter_method",false },
  {0x0009,0x00c4,"GEMS_PETD_01",VR::SS,VM::VM1,"PET scatter_num_iter",false },
  {0x0009,0x00c5,"GEMS_PETD_01",VR::FL,VM::VM1,"PET scatter_parm",false },
  {0x0009,0x00c6,"GEMS_PETD_01",VR::FL,VM::VM1,"PET seg_qc_parm",false },
  {0x0009,0x00c7,"GEMS_PETD_01",VR::SL,VM::VM1,"PET overlap",false },
  {0x0009,0x00c8,"GEMS_PETD_01",VR::UI,VM::VM1,"PET ovlp_frm_id",false },
  {0x0009,0x00c9,"GEMS_PETD_01",VR::UI,VM::VM1,"PET ovlp_trans_frm_id",false },
  {0x0009,0x00ca,"GEMS_PETD_01",VR::UI,VM::VM1,"PET ovlp_tpulse_frm_id",false },
  {0x0009,0x00cb,"GEMS_PETD_01",VR::FL,VM::VM1,"PET vqc_x_axis_trans",false },
  {0x0009,0x00cc,"GEMS_PETD_01",VR::FL,VM::VM1,"PET vqc_x_axis_tilt",false },
  {0x0009,0x00cd,"GEMS_PETD_01",VR::FL,VM::VM1,"PET vqc_y_axis_trans",false },
  {0x0009,0x00ce,"GEMS_PETD_01",VR::FL,VM::VM1,"PET vqc_y_axis_swivel",false },
  {0x0009,0x00cf,"GEMS_PETD_01",VR::FL,VM::VM1,"PET vqc_z_axis_trans",false },
  {0x0009,0x00d0,"GEMS_PETD_01",VR::FL,VM::VM1,"PET vqc_z_axis_roll",false },
  {0x0009,0x00d1,"GEMS_PETD_01",VR::LO,VM::VM1,"PET ctac_conv_scale",false },
  {0x0009,0x00d2,"GEMS_PETD_01",VR::UI,VM::VM1,"PET image_set_id",false },
  {0x0009,0x00d3,"GEMS_PETD_01",VR::SL,VM::VM1,"PET constrast_route",false },
  {0x0009,0x00d4,"GEMS_PETD_01",VR::LO,VM::VM1,"PET ctac_conv_scale",false },
  {0x0009,0x00d5,"GEMS_PETD_01",VR::FL,VM::VM1,"PET loop_filter_parm",false },
  {0x0009,0x00d6,"GEMS_PETD_01",VR::FL,VM::VM1,"PET image_one_loc",false },
  {0x0009,0x00d7,"GEMS_PETD_01",VR::FL,VM::VM1,"PET image_index_loc",false },
  {0x0009,0x00d8,"GEMS_PETD_01",VR::SL,VM::VM1,"PET frame_number",false },
  {0x0009,0x00d9,"GEMS_PETD_01",VR::SL,VM::VM1,"PET list_file_exists",false },
  {0x0009,0x00da,"GEMS_PETD_01",VR::ST,VM::VM1,"PET where_is_list_frame",false },
  {0x0009,0x00db,"GEMS_PETD_01",VR::SL,VM::VM1,"PET ir_z_filter_flag",false },
  {0x0009,0x00dc,"GEMS_PETD_01",VR::FL,VM::VM1,"PET ir_z_filter_ratio",false },
  {0x0009,0x00dd,"GEMS_PETD_01",VR::US,VM::VM1,"PET num_of_rr_interval",false },
  {0x0009,0x00de,"GEMS_PETD_01",VR::US,VM::VM1,"PET num_of_time_slots",false },
  {0x0009,0x00df,"GEMS_PETD_01",VR::US,VM::VM1,"PET num_of_slices",false },
  {0x0009,0x00e0,"GEMS_PETD_01",VR::US,VM::VM1,"PET num_of_time_slices",false },
  {0x0009,0x00e1,"GEMS_PETD_01",VR::SL,VM::VM1,"PET unlisted_scan",false },
  {0x0009,0x00e2,"GEMS_PETD_01",VR::SL,VM::VM1,"PET rest_stress",false },
  {0x0009,0x00e3,"GEMS_PETD_01",VR::FL,VM::VM1,"PET phase percentage",false },
  {0x0009,0x00e4,"GEMS_PETD_01",VR::ST,VM::VM1,"Recon Protocol",false },
  {0x0009,0x00e5,"GEMS_PETD_01",VR::FL,VM::VM1,"PET left shift",false },
  {0x0009,0x00e6,"GEMS_PETD_01",VR::FL,VM::VM1,"PET posterior shift",false },
  {0x0009,0x00e7,"GEMS_PETD_01",VR::FL,VM::VM1,"PET superior shift",false },
  {0x0009,0x00e8,"GEMS_PETD_01",VR::SL,VM::VM1,"PET acq_bin_num",false },
  {0x0009,0x00e9,"GEMS_PETD_01",VR::FL,VM::VM1,"PET acq_bin_dur_percent",false },
  {0x0009,0x00ea,"GEMS_PETD_01",VR::SL,VM::VM1,"3D Filter flag",false },
  {0x0009,0x00eb,"GEMS_PETD_01",VR::FL,VM::VM1,"3D Filter cutoff",false },
  {0x0009,0x00ec,"GEMS_PETD_01",VR::SL,VM::VM1,"3D Filter order",false },
  {0x0009,0x00f0,"GEMS_PETD_01",VR::UI,VM::VM1,"Reformat group",false },
  {0x0009,0x00f1,"GEMS_PETD_01",VR::SL,VM::VM1,"PET prompt gamma",false },
  {0x0009,0x00f2,"GEMS_PETD_01",VR::UI,VM::VM1,"PET tracerInjectionUID",false },
  {0x0009,0x00f6,"GEMS_PETD_01",VR::SL,VM::VM1,"PET Q Static ScanMode",false },
  {0x0009,0x00f7,"GEMS_PETD_01",VR::SL,VM::VM1,"PET Q Static Frame",false },
  {0x0009,0x00f3,"GEMS_PETD_01",VR::FL,VM::VM1,"Blood Glucose Level",false },
  {0x0009,0x00f4,"GEMS_PETD_01",VR::SL,VM::VM1,"IsPatientDiabetic",false },
  {0x0009,0x00f5,"GEMS_PETD_01",VR::DA,VM::VM1,"Date of Last Treatment",false },
  {0x0009,0x00f8,"GEMS_PETD_01",VR::FL,VM::VM1,"Regularized Recon Beta",false },
  {0x0009,0x00f9,"GEMS_PETD_01",VR::FL,VM::VM1,"Regularized Recon Gamma",false },
  {0x0009,0x00fa,"GEMS_PETD_01",VR::SL,VM::VM1,"?",false },
  {0x0009,0x00fb,"GEMS_PETD_01",VR::SL,VM::VM1,"?",false },
  {0x0009,0x00fc,"GEMS_PETD_01",VR::SL,VM::VM1,"?",false },
  {0x0011,0x0001,"GEMS_PETD_01",VR::SQ,VM::VM1,"GE Advance ROI Sequence",false },
  {0x0011,0x0002,"GEMS_PETD_01",VR::UI,VM::VM1,"GE Advance ROI.roi_id",false },
  {0x0011,0x0003,"GEMS_PETD_01",VR::UI,VM::VM1,"GE Advance ROI.image_id",false },
  {0x0011,0x0004,"GEMS_PETD_01",VR::SH,VM::VM1,"GE Advance ROI.compatible_version",false },
  {0x0011,0x0005,"GEMS_PETD_01",VR::SH,VM::VM1,"GE Advance ROI.software_version",false },
  {0x0011,0x0006,"GEMS_PETD_01",VR::LO,VM::VM1,"GE Advance ROI.roi_name",false },
  {0x0011,0x0007,"GEMS_PETD_01",VR::DT,VM::VM1,"GE Advance ROI.roi_datetime",false },
  {0x0011,0x0008,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance ROI.roi_type",false },
  {0x0011,0x0009,"GEMS_PETD_01",VR::FL,VM::VM1,"GE Advance ROI.center_x",false },
  {0x0011,0x000a,"GEMS_PETD_01",VR::FL,VM::VM1,"GE Advance ROI.center_y",false },
  {0x0011,0x000b,"GEMS_PETD_01",VR::FL,VM::VM1,"GE Advance ROI.width",false },
  {0x0011,0x000c,"GEMS_PETD_01",VR::FL,VM::VM1,"GE Advance ROI.height",false },
  {0x0011,0x000d,"GEMS_PETD_01",VR::FL,VM::VM1,"GE Advance ROI.angle",false },
  {0x0011,0x000e,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance ROI.number_of_points",false },
  {0x0011,0x000f,"GEMS_PETD_01",VR::OB,VM::VM1,"GE Advance ROI.roi_data",false },
  {0x0011,0x0010,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance ROI.roi_size",false },
  {0x0011,0x0011,"GEMS_PETD_01",VR::LO,VM::VM1,"GE Advance ROI.color",false },
  {0x0011,0x0012,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance ROI.line_type",false },
  {0x0011,0x0013,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance ROI.line_width",false },
  {0x0011,0x0014,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance ROI.roi_number",false },
  {0x0011,0x0015,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance ROI.convex",false },
  {0x0011,0x0016,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance ROI.atten_corr_flag",false },
  {0x0011,0x0018,"GEMS_PETD_01",VR::OB,VM::VM1,"?",false },
  {0x0013,0x0001,"GEMS_PETD_01",VR::SQ,VM::VM1,"GE Advance Annotation Sequence",false },
  {0x0013,0x0002,"GEMS_PETD_01",VR::UI,VM::VM1,"GE Advance Annotation.annotation_id",false },
  {0x0013,0x0003,"GEMS_PETD_01",VR::UI,VM::VM1,"GE Advance Annotation.image_id",false },
  {0x0013,0x0004,"GEMS_PETD_01",VR::SH,VM::VM1,"GE Advance Annotation.compatible_version",false },
  {0x0013,0x0005,"GEMS_PETD_01",VR::SH,VM::VM1,"GE Advance Annotation.software_version",false },
  {0x0013,0x0006,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance Annotation.type",false },
  {0x0013,0x0007,"GEMS_PETD_01",VR::LO,VM::VM1,"GE Advance Annotation.font_name",false },
  {0x0013,0x0008,"GEMS_PETD_01",VR::SH,VM::VM1,"GE Advance Annotation.font_size",false },
  {0x0013,0x0009,"GEMS_PETD_01",VR::LO,VM::VM1,"GE Advance Annotation.foreground_color",false },
  {0x0013,0x000a,"GEMS_PETD_01",VR::LO,VM::VM1,"GE Advance Annotation.background_color",false },
  {0x0013,0x000b,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance Annotation.coordinate_system",false },
  {0x0013,0x000c,"GEMS_PETD_01",VR::FL,VM::VM1,"GE Advance Annotation.start_x",false },
  {0x0013,0x000d,"GEMS_PETD_01",VR::FL,VM::VM1,"GE Advance Annotation.start_y",false },
  {0x0013,0x000e,"GEMS_PETD_01",VR::FL,VM::VM1,"GE Advance Annotation.end_x",false },
  {0x0013,0x000f,"GEMS_PETD_01",VR::FL,VM::VM1,"GE Advance Annotation.end_y",false },
  {0x0013,0x0010,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance Annotation.start_symbol",false },
  {0x0013,0x0011,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance Annotation.end_symbol",false },
  {0x0013,0x0012,"GEMS_PETD_01",VR::OB,VM::VM1,"GE Advance Annotation.annotation_data",false },
  {0x0013,0x0013,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance Annotation.annotation_size",false },
  {0x0013,0x0014,"GEMS_PETD_01",VR::LO,VM::VM1,"GE Advance Annotation.label_id",false },
  {0x0015,0x003d,"GEMS_PETD_01",VR::UL,VM::VM1,"PET Sharp IR Flag",false },
  {0x0015,0x003e,"GEMS_PETD_01",VR::UL,VM::VM1,"PET Scatter Limit",false },
  {0x0015,0x003f,"GEMS_PETD_01",VR::SH,VM::VM1,"PET Q Static Recon",false },
  {0x0015,0x0039,"GEMS_PETD_01",VR::SL,VM::VM1,"Event histogram Format",false },
  {0x0015,0x003a,"GEMS_PETD_01",VR::SL,VM::VM1,"Number of Detector Rows",false },
  {0x0015,0x003b,"GEMS_PETD_01",VR::SL,VM::VM1,"Number of Detector Columns",false },
  {0x0015,0x001a,"GEMS_PETD_01",VR::SL,VM::VM1,"Physio Gating Type",false },
  {0x0015,0x001b,"GEMS_PETD_01",VR::SL,VM::VM1,"Total Number of Bins",false },
  {0x0015,0x001d,"GEMS_PETD_01",VR::SL,VM::VM1,"Phase Matched Series",false },
  {0x0015,0x001e,"GEMS_PETD_01",VR::SL,VM::VM1,"CTAC Percent Value",false },
  {0x0015,0x003c,"GEMS_PETD_01",VR::US,VM::VM1,"Recon Matrix Size",false },
  {0x0015,0x001f,"GEMS_PETD_01",VR::UL,VM::VM1,"PET Recon Parameters Exists",false },
  {0x0015,0x0020,"GEMS_PETD_01",VR::SL,VM::VM1,"First Packet Number",false },
  {0x0015,0x0021,"GEMS_PETD_01",VR::FL,VM::VM1,"IR Loop Filter Ratio",false },
  {0x0015,0x0022,"GEMS_PETD_01",VR::FL,VM::VM1,"IR Loop Filter Correction",false },
  {0x0015,0x0023,"GEMS_PETD_01",VR::UL,VM::VM1,"BP3d Filter FlagU",false },
  {0x0015,0x0024,"GEMS_PETD_01",VR::FL,VM::VM1,"BP3d Filter CutoffU",false },
  {0x0015,0x0025,"GEMS_PETD_01",VR::SL,VM::VM1,"BP3d Filter OrderU",false },
  {0x0015,0x0026,"GEMS_PETD_01",VR::UL,VM::VM1,"BP3d Filter FlagV",false },
  {0x0015,0x0027,"GEMS_PETD_01",VR::FL,VM::VM1,"BP3d Filter OrderV",false },
  {0x0015,0x0028,"GEMS_PETD_01",VR::SL,VM::VM1,"BP3d Filter CutoffV",false },
  {0x0015,0x0029,"GEMS_PETD_01",VR::UL,VM::VM1,"Decay Flag",false },
  {0x0015,0x002e,"GEMS_PETD_01",VR::UL,VM::VM1,"Image filter 3d flag",false },
  {0x0015,0x002f,"GEMS_PETD_01",VR::UL,VM::VM1,"WCC Cal Flag",false },
  {0x0015,0x0030,"GEMS_PETD_01",VR::UL,VM::VM1,"Emission Scatter Flag",false },
  {0x0015,0x0031,"GEMS_PETD_01",VR::UL,VM::VM1,"Recon Algorithm",false },
  {0x0015,0x0032,"GEMS_PETD_01",VR::UL,VM::VM1,"Normalization Flag",false },
  {0x0015,0x0033,"GEMS_PETD_01",VR::UL,VM::VM1,"Emission Deadtime Flag",false },
  {0x0015,0x0034,"GEMS_PETD_01",VR::UL,VM::VM1,"Emission Random Flag",false },
  {0x0015,0x0035,"GEMS_PETD_01",VR::UL,VM::VM1,"Blank Scan Flag",false },
  {0x0015,0x0036,"GEMS_PETD_01",VR::UL,VM::VM1,"?",false },
  {0x0015,0x0037,"GEMS_PETD_01",VR::UL,VM::VM1,"Trans Deadtime Flag",false },
  {0x0015,0x0038,"GEMS_PETD_01",VR::UL,VM::VM1,"Natural Logarithm Flag",false },
  {0x0017,0x0001,"GEMS_PETD_01",VR::UI,VM::VM1,"PET correction_cal_id",false },
  {0x0017,0x0002,"GEMS_PETD_01",VR::SH,VM::VM1,"PET compatible_version",false },
  {0x0017,0x0003,"GEMS_PETD_01",VR::SH,VM::VM1,"PET software_version",false },
  {0x0017,0x0004,"GEMS_PETD_01",VR::DT,VM::VM1,"PET cal_datetime",false },
  {0x0017,0x0005,"GEMS_PETD_01",VR::LO,VM::VM1,"PET cal_description",false },
  {0x0017,0x0006,"GEMS_PETD_01",VR::SL,VM::VM1,"PET cal_type",false },
  {0x0017,0x0007,"GEMS_PETD_01",VR::ST,VM::VM1,"PET where_is_corr",false },
  {0x0017,0x0008,"GEMS_PETD_01",VR::SL,VM::VM1,"PET corr_file_size",false },
  {0x0017,0x0009,"GEMS_PETD_01",VR::LO,VM::VM1,"PET scan_id",false },
  {0x0017,0x000a,"GEMS_PETD_01",VR::DT,VM::VM1,"PET scan_datetime",false },
  {0x0017,0x000b,"GEMS_PETD_01",VR::LO,VM::VM1,"PET norm_2d_cal_id",false },
  {0x0017,0x000c,"GEMS_PETD_01",VR::SH,VM::VM1,"PET hosp_identifier",false },
  {0x0017,0x000d,"GEMS_PETD_01",VR::SL,VM::VM1,"PET archived",false },
  {0x0019,0x0001,"GEMS_PETD_01",VR::UI,VM::VM1,"PET wc_cal_id",false },
  {0x0019,0x0002,"GEMS_PETD_01",VR::SH,VM::VM1,"PET compatible_version",false },
  {0x0019,0x0003,"GEMS_PETD_01",VR::SH,VM::VM1,"PET software_version",false },
  {0x0019,0x0004,"GEMS_PETD_01",VR::DT,VM::VM1,"PET cal_datetime",false },
  {0x0019,0x0005,"GEMS_PETD_01",VR::SL,VM::VM1,"PET cal_type",false },
  {0x0019,0x0006,"GEMS_PETD_01",VR::LO,VM::VM1,"PET cal_description",false },
  {0x0019,0x0007,"GEMS_PETD_01",VR::LO,VM::VM1,"PET cal_hardware",false },
  {0x0019,0x0008,"GEMS_PETD_01",VR::OB,VM::VM1,"PET coefficients",false },
  {0x0019,0x0009,"GEMS_PETD_01",VR::FL,VM::VM1,"PET activity_factor_hr",false },
  {0x0019,0x000a,"GEMS_PETD_01",VR::FL,VM::VM1,"PET activity_factor_hs",false },
  {0x0019,0x000b,"GEMS_PETD_01",VR::FL,VM::VM1,"PET activity_factor_3d",false },
  {0x0019,0x000c,"GEMS_PETD_01",VR::LO,VM::VM1,"PET scan_id",false },
  {0x0019,0x000d,"GEMS_PETD_01",VR::DT,VM::VM1,"PET scan_datetime",false },
  {0x0019,0x000e,"GEMS_PETD_01",VR::SH,VM::VM1,"PET hosp_identifier",false },
  {0x0019,0x000f,"GEMS_PETD_01",VR::FL,VM::VM1,"PET meas_activity",false },
  {0x0019,0x0010,"GEMS_PETD_01",VR::DT,VM::VM1,"PET meas_datetime",false },
  {0x0019,0x0011,"GEMS_PETD_01",VR::SL,VM::VM1,"PET axial_filter_3d",false },
  {0x0019,0x0012,"GEMS_PETD_01",VR::FL,VM::VM1,"PET axial_cutoff_3d",false },
  {0x0019,0x0013,"GEMS_PETD_01",VR::SL,VM::VM1,"PET default_flag",false },
  {0x0019,0x0014,"GEMS_PETD_01",VR::SL,VM::VM1,"PET archived",false },
  {0x0019,0x0015,"GEMS_PETD_01",VR::SL,VM::VM1,"PET wc_cal_rec_method",false },
  {0x0019,0x0016,"GEMS_PETD_01",VR::SL,VM::VM1,"PET activity_factor_2d",false },
  {0x0019,0x0017,"GEMS_PETD_01",VR::SL,VM::VM1,"PET isotope",false },
  {0x0021,0x0001,"GEMS_PETD_01",VR::US,VM::VM1,"PET raw_data_type",false },
  {0x0021,0x0002,"GEMS_PETD_01",VR::UL,VM::VM1,"PET raw_data_size",false },
  {0x0023,0x0001,"GEMS_PETD_01",VR::OB,VM::VM1,"raw_data_blob",false },
  {0x0023,0x0002,"GEMS_PETD_01",VR::OB,VM::VM1,"PET raw_data_blob",false },
  {0x0023,0x0017,"GEMS_PETD_01",VR::SQ,VM::VM1,"?",false },
  {0x0023,0x0019,"GEMS_PETD_01",VR::FL,VM::VM1,"?",false },
  {0x0023,0x001a,"GEMS_PETD_01",VR::FL,VM::VM1,"?",false },
  {0x0023,0x001b,"GEMS_PETD_01",VR::SL,VM::VM1,"?",false },
  {0x0023,0x001c,"GEMS_PETD_01",VR::SQ,VM::VM1,"?",false },
  {0x0023,0x001d,"GEMS_PETD_01",VR::SL,VM::VM1,"?",false },
  {0x0023,0x001e,"GEMS_PETD_01",VR::FL,VM::VM1,"?",false },
  {0x0023,0x002c,"GEMS_PETD_01",VR::UL,VM::VM1,"?",false },
  {0x0023,0x002d,"GEMS_PETD_01",VR::LO,VM::VM1,"?",false },
  {0x0023,0x002e,"GEMS_PETD_01",VR::LO,VM::VM1,"?",false },
  {0x0023,0x0031,"GEMS_PETD_01",VR::CS,VM::VM1,"?",false },
  {0x0023,0x0040,"GEMS_PETD_01",VR::SQ,VM::VM1,"?",false },
  {0x0023,0x0041,"GEMS_PETD_01",VR::LO,VM::VM1,"?",false },
  {0x0023,0x0042,"GEMS_PETD_01",VR::LO,VM::VM1,"?",false },
  {0x0023,0x0058,"GEMS_PETD_01",VR::SQ,VM::VM1,"?",false },
  {0x0023,0x0059,"GEMS_PETD_01",VR::LO,VM::VM1,"?",false },
  {0x0023,0x005a,"GEMS_PETD_01",VR::LO,VM::VM1,"?",false },
  {0x5001,0x0001,"GEMS_PETD_01",VR::UI,VM::VM1,"GE Advance Curve.curve_id",false },
  {0x5001,0x0002,"GEMS_PETD_01",VR::SH,VM::VM1,"GE Advance Curve.compatible_version",false },
  {0x5001,0x0003,"GEMS_PETD_01",VR::SH,VM::VM1,"GE Advance Curve.software_version",false },
  {0x5001,0x0004,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance Curve.statistics_type",false },
  {0x5001,0x0005,"GEMS_PETD_01",VR::LT,VM::VM1,"GE Advance Curve.how_derived",false },
  {0x5001,0x0006,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance Curve.how_derived_size",false },
  {0x5001,0x0007,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance Curve.multi_patient",false },
  {0x5001,0x0008,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance Curve.deadtime",false },
  {0x5003,0x0001,"GEMS_PETD_01",VR::SQ,VM::VM1,"GE Advance Graph Sequence",false },
  {0x5003,0x0002,"GEMS_PETD_01",VR::UI,VM::VM1,"GE Advance Graph.graph_id",false },
  {0x5003,0x0003,"GEMS_PETD_01",VR::SH,VM::VM1,"GE Advance Graph.compatible_version",false },
  {0x5003,0x0004,"GEMS_PETD_01",VR::SH,VM::VM1,"GE Advance Graph.software_version",false },
  {0x5003,0x0005,"GEMS_PETD_01",VR::LO,VM::VM1,"GE Advance Graph.title",false },
  {0x5003,0x0006,"GEMS_PETD_01",VR::DT,VM::VM1,"GE Advance Graph.graph_datetime",false },
  {0x5003,0x0007,"GEMS_PETD_01",VR::ST,VM::VM1,"GE Advance Graph.graph_description",false },
  {0x5003,0x0008,"GEMS_PETD_01",VR::LO,VM::VM1,"GE Advance Graph.title_font_name",false },
  {0x5003,0x0009,"GEMS_PETD_01",VR::SH,VM::VM1,"GE Advance Graph.title_font_size",false },
  {0x5003,0x000a,"GEMS_PETD_01",VR::LO,VM::VM1,"GE Advance Graph.footer",false },
  {0x5003,0x000b,"GEMS_PETD_01",VR::SH,VM::VM1,"GE Advance Graph.footer_font_size",false },
  {0x5003,0x000c,"GEMS_PETD_01",VR::LO,VM::VM1,"GE Advance Graph.foreground_color",false },
  {0x5003,0x000d,"GEMS_PETD_01",VR::LO,VM::VM1,"GE Advance Graph.background_color",false },
  {0x5003,0x000e,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance Graph.graph_border",false },
  {0x5003,0x000f,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance Graph.graph_width",false },
  {0x5003,0x0010,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance Graph.graph_height",false },
  {0x5003,0x0011,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance Graph.grid",false },
  {0x5003,0x0012,"GEMS_PETD_01",VR::LO,VM::VM1,"GE Advance Graph.label_font_name",false },
  {0x5003,0x0013,"GEMS_PETD_01",VR::SH,VM::VM1,"GE Advance Graph.label_font_size",false },
  {0x5003,0x0014,"GEMS_PETD_01",VR::LO,VM::VM1,"GE Advance Graph.axes_color",false },
  {0x5003,0x0015,"GEMS_PETD_01",VR::LO,VM::VM1,"GE Advance Graph.x_axis_label",false },
  {0x5003,0x0016,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance Graph.x_axis_units",false },
  {0x5003,0x0017,"GEMS_PETD_01",VR::FL,VM::VM1,"GE Advance Graph.x_major_tics",false },
  {0x5003,0x0018,"GEMS_PETD_01",VR::FL,VM::VM1,"GE Advance Graph.x_axis_min",false },
  {0x5003,0x0019,"GEMS_PETD_01",VR::FL,VM::VM1,"GE Advance Graph.x_axis_max",false },
  {0x5003,0x001a,"GEMS_PETD_01",VR::LO,VM::VM1,"GE Advance Graph.y_axis_label",false },
  {0x5003,0x001b,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance Graph.y_axis_units",false },
  {0x5003,0x001c,"GEMS_PETD_01",VR::FL,VM::VM1,"GE Advance Graph.y_major_tics",false },
  {0x5003,0x001d,"GEMS_PETD_01",VR::FL,VM::VM1,"GE Advance Graph.y_axis_min",false },
  {0x5003,0x001e,"GEMS_PETD_01",VR::FL,VM::VM1,"GE Advance Graph.y_axis_max",false },
  {0x5003,0x001f,"GEMS_PETD_01",VR::LO,VM::VM1,"GE Advance Graph.legend_font_name",false },
  {0x5003,0x0020,"GEMS_PETD_01",VR::SH,VM::VM1,"GE Advance Graph.legend_font_size",false },
  {0x5003,0x0021,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance Graph.legend_location_x",false },
  {0x5003,0x0022,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance Graph.legend_location_y",false },
  {0x5003,0x0023,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance Graph.legend_width",false },
  {0x5003,0x0024,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance Graph.legend_height",false },
  {0x5003,0x0025,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance Graph.legend_border",false },
  {0x5003,0x0026,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance Graph.multi_patient",false },
  {0x5005,0x0001,"GEMS_PETD_01",VR::SQ,VM::VM1,"GE Advance CurvePresentation Sequence",false },
  {0x5005,0x0002,"GEMS_PETD_01",VR::UI,VM::VM1,"GE Advance CurvePresentation.curvepresent_id",false },
  {0x5005,0x0003,"GEMS_PETD_01",VR::UI,VM::VM1,"GE Advance CurvePresentation.graph_id",false },
  {0x5005,0x0004,"GEMS_PETD_01",VR::UI,VM::VM1,"GE Advance CurvePresentation.curve_id",false },
  {0x5005,0x0005,"GEMS_PETD_01",VR::SH,VM::VM1,"GE Advance CurvePresentation.compatible_version",false },
  {0x5005,0x0006,"GEMS_PETD_01",VR::SH,VM::VM1,"GE Advance CurvePresentation.software_version",false },
  {0x5005,0x0007,"GEMS_PETD_01",VR::LO,VM::VM1,"GE Advance CurvePresentation.curve_label",false },
  {0x5005,0x0008,"GEMS_PETD_01",VR::LO,VM::VM1,"GE Advance CurvePresentation.color",false },
  {0x5005,0x0009,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance CurvePresentation.line_type",false },
  {0x5005,0x000a,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance CurvePresentation.line_width",false },
  {0x5005,0x000b,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance CurvePresentation.point_symbol",false },
  {0x5005,0x000c,"GEMS_PETD_01",VR::SL,VM::VM1,"GE Advance CurvePresentation.point_symbol_dim",false },
  {0x5005,0x000d,"GEMS_PETD_01",VR::LO,VM::VM1,"GE Advance CurvePresentation.point_color",false },
  {0x0009,0x0060,"GEMS_QVA_PHOTO_01",VR::FL,VM::VM1,"Dodge End Diastolic Volume ml",false },
  {0x0009,0x0061,"GEMS_QVA_PHOTO_01",VR::FL,VM::VM1,"Dodge End Systolic Volume ml",false },
  {0x0009,0x0062,"GEMS_QVA_PHOTO_01",VR::FL,VM::VM1,"Dodge Stroke Volume ml",false },
  {0x0009,0x0063,"GEMS_QVA_PHOTO_01",VR::IS,VM::VM1,"Dodge Ejection Fraction",false },
  {0x0009,0x0064,"GEMS_QVA_PHOTO_01",VR::FL,VM::VM1,"Simpson End Diastolic Volume ml",false },
  {0x0009,0x0065,"GEMS_QVA_PHOTO_01",VR::FL,VM::VM1,"Simpson End Systolic Volume ml",false },
  {0x0009,0x0066,"GEMS_QVA_PHOTO_01",VR::FL,VM::VM1,"Simpson Stroke Volume ml",false },
  {0x0009,0x0067,"GEMS_QVA_PHOTO_01",VR::IS,VM::VM1,"Simpson Ejection Fraction",false },
  {0x0009,0x0068,"GEMS_QVA_PHOTO_01",VR::FL,VM::VM1,"CFX Single Hypokinesia in Region",false },
  {0x0009,0x0069,"GEMS_QVA_PHOTO_01",VR::FL,VM::VM1,"CFX Single Hyperkinesia in Opposite Region",false },
  {0x0009,0x006a,"GEMS_QVA_PHOTO_01",VR::IS,VM::VM1,"CFX Single Total LV contour Percent",false },
  {0x0009,0x006b,"GEMS_QVA_PHOTO_01",VR::FL,VM::VM1,"CFX Multiple Hypokinesia in Region",false },
  {0x0009,0x006c,"GEMS_QVA_PHOTO_01",VR::FL,VM::VM1,"CFX Multiple Hyperkinesia in Opposite Region",false },
  {0x0009,0x006d,"GEMS_QVA_PHOTO_01",VR::IS,VM::VM1,"CFX Multiple Total LV contour Percent",false },
  {0x0009,0x006e,"GEMS_QVA_PHOTO_01",VR::FL,VM::VM1,"RCA Single Hypokinesia in Region",false },
  {0x0009,0x006f,"GEMS_QVA_PHOTO_01",VR::FL,VM::VM1,"RCA Single Hyperkinesia in Opposite Region",false },
  {0x0009,0x0070,"GEMS_QVA_PHOTO_01",VR::IS,VM::VM1,"RCA Single Total LV contour Percent",false },
  {0x0009,0x0071,"GEMS_QVA_PHOTO_01",VR::FL,VM::VM1,"RCA Multiple Hypokinesia in Region",false },
  {0x0009,0x0072,"GEMS_QVA_PHOTO_01",VR::FL,VM::VM1,"RCA Multiple Hyperkinesia in Opposite Region",false },
  {0x0009,0x0073,"GEMS_QVA_PHOTO_01",VR::IS,VM::VM1,"RCA Multiple Total LV contour Percent",false },
  {0x0009,0x0074,"GEMS_QVA_PHOTO_01",VR::FL,VM::VM1,"LAD Single Hypokinesia in Region",false },
  {0x0009,0x0075,"GEMS_QVA_PHOTO_01",VR::FL,VM::VM1,"LAD Single Hyperkinesia in Opposite Region",false },
  {0x0009,0x0076,"GEMS_QVA_PHOTO_01",VR::IS,VM::VM1,"LAD Single Total LV contour Percent",false },
  {0x0009,0x0077,"GEMS_QVA_PHOTO_01",VR::FL,VM::VM1,"LAD Multiple Hypokinesia in Region",false },
  {0x0009,0x0078,"GEMS_QVA_PHOTO_01",VR::FL,VM::VM1,"LAD Multiple Hyperkinesia in Opposite Region",false },
  {0x0009,0x0079,"GEMS_QVA_PHOTO_01",VR::IS,VM::VM1,"LAD Multiple Total LV contour Percent",false },
  {0x0009,0x007a,"GEMS_QVA_PHOTO_01",VR::FL,VM::VM1,"Dodge End Diastolic Volume ml/m2",false },
  {0x0009,0x007c,"GEMS_QVA_PHOTO_01",VR::FL,VM::VM1,"Dodge End Systolic Volume ml/m2",false },
  {0x0009,0x007e,"GEMS_QVA_PHOTO_01",VR::FL,VM::VM1,"Dodge Stroke Volume ml/m2",false },
  {0x0009,0x0080,"GEMS_QVA_PHOTO_01",VR::FL,VM::VM1,"Simpson End Diastolic Volume ml/m2",false },
  {0x0009,0x0082,"GEMS_QVA_PHOTO_01",VR::FL,VM::VM1,"Simpson End Systolic Volume ml/m2",false },
  {0x0009,0x0084,"GEMS_QVA_PHOTO_01",VR::FL,VM::VM1,"Simpson Stroke Volume ml/m2",false },
  {0x0021,0x0003,"GEMS_RELA_01",VR::SS,VM::VM1,"Series from which Prescribed",false },
  {0x0021,0x0005,"GEMS_RELA_01",VR::SH,VM::VM1,"Genesis Version - now",false },
  {0x0021,0x0007,"GEMS_RELA_01",VR::UL,VM::VM1,"Series Record checksum",false },
  {0x0021,0x0015,"GEMS_RELA_01",VR::US,VM::VM1,"?",false },
  {0x0021,0x0016,"GEMS_RELA_01",VR::SS,VM::VM1,"?",false },
  {0x0021,0x0018,"GEMS_RELA_01",VR::SH,VM::VM1,"Genesis version - Now",false },
  {0x0021,0x0019,"GEMS_RELA_01",VR::UL,VM::VM1,"Acq recon record checksum",false },
  {0x0021,0x0020,"GEMS_RELA_01",VR::DS,VM::VM1,"Table start location",false },
  {0x0021,0x0035,"GEMS_RELA_01",VR::SS,VM::VM1,"Series from which prescribed",false },
  {0x0021,0x0036,"GEMS_RELA_01",VR::SS,VM::VM1,"Image from which prescribed",false },
  {0x0021,0x0037,"GEMS_RELA_01",VR::SS,VM::VM1,"Screen Format",false },
  {0x0021,0x004a,"GEMS_RELA_01",VR::LO,VM::VM1,"Anatomical reference for scout",false },
  {0x0021,0x004e,"GEMS_RELA_01",VR::US,VM::VM1,"?",false },
  {0x0021,0x004f,"GEMS_RELA_01",VR::SS,VM::VM1,"Locations in acquisition",false },
  {0x0021,0x0050,"GEMS_RELA_01",VR::SS,VM::VM1,"Graphically prescribed",false },
  {0x0021,0x0051,"GEMS_RELA_01",VR::DS,VM::VM1,"Rotation from source x rot",false },
  {0x0021,0x0052,"GEMS_RELA_01",VR::DS,VM::VM1,"Rotation from source y rot",false },
  {0x0021,0x0053,"GEMS_RELA_01",VR::DS,VM::VM1,"Rotation from source z rot",false },
  {0x0021,0x0054,"GEMS_RELA_01",VR::SH,VM::VM3,"Image position",false },
  {0x0021,0x0055,"GEMS_RELA_01",VR::SH,VM::VM6,"Image orientation",false },
  {0x0021,0x0056,"GEMS_RELA_01",VR::SL,VM::VM1,"Num 3D slabs",false },
  {0x0021,0x0057,"GEMS_RELA_01",VR::SL,VM::VM1,"Locs per 3D slab",false },
  {0x0021,0x0058,"GEMS_RELA_01",VR::SL,VM::VM1,"Overlaps",false },
  {0x0021,0x0059,"GEMS_RELA_01",VR::SL,VM::VM1,"Image Filtering 0.5/0.2T",false },
  {0x0021,0x005a,"GEMS_RELA_01",VR::SL,VM::VM1,"Diffusion direction",false },
  {0x0021,0x005b,"GEMS_RELA_01",VR::DS,VM::VM1,"Tagging Flip Angle",false },
  {0x0021,0x005c,"GEMS_RELA_01",VR::DS,VM::VM1,"Tagging Orientation",false },
  {0x0021,0x005d,"GEMS_RELA_01",VR::DS,VM::VM1,"Tag Spacing",false },
  {0x0021,0x005e,"GEMS_RELA_01",VR::DS,VM::VM1,"RTIA_timer",false },
  {0x0021,0x005f,"GEMS_RELA_01",VR::DS,VM::VM1,"Fps",false },
  {0x0021,0x0070,"GEMS_RELA_01",VR::LT,VM::VM1,"?",false },
  {0x0021,0x0071,"GEMS_RELA_01",VR::LT,VM::VM1,"?",false },
  {0x0021,0x0081,"GEMS_RELA_01",VR::DS,VM::VM1,"Auto window/level alpha",false },
  {0x0021,0x0082,"GEMS_RELA_01",VR::DS,VM::VM1,"Auto window/level beta",false },
  {0x0021,0x0083,"GEMS_RELA_01",VR::DS,VM::VM1,"Auto window/level window",false },
  {0x0021,0x0084,"GEMS_RELA_01",VR::DS,VM::VM1,"Auto window/level level",false },
  {0x0021,0x0090,"GEMS_RELA_01",VR::SS,VM::VM1,"Tube focal spot position",false },
  {0x0021,0x0091,"GEMS_RELA_01",VR::SS,VM::VM1,"Biopsy position",false },
  {0x0021,0x0092,"GEMS_RELA_01",VR::FL,VM::VM1,"Biopsy T location",false },
  {0x0021,0x0093,"GEMS_RELA_01",VR::FL,VM::VM1,"Biopsy ref location",false },
  {0x0249,0x0011,"GEMS_RTEN_01",VR::DA,VM::VM1,"Plan Creation Date",false },
  {0x0249,0x0013,"GEMS_RTEN_01",VR::TM,VM::VM1,"Plan Creation Time",false },
  {0x0249,0x0014,"GEMS_RTEN_01",VR::PN,VM::VM1,"Operator Name",false },
  {0x0249,0x0016,"GEMS_RTEN_01",VR::LO,VM::VM1,"Plan Comment",false },
  {0x0249,0x0018,"GEMS_RTEN_01",VR::SQ,VM::VM1,"Plan Image Sequence",false },
  {0x0249,0x001a,"GEMS_RTEN_01",VR::LO,VM::VM1,"Exam/Series/Image Identifier",false },
  {0x0249,0x001b,"GEMS_RTEN_01",VR::IS,VM::VM1,"Additional Image Identifier",false },
  {0x0249,0x001c,"GEMS_RTEN_01",VR::CS,VM::VM1,"Build Resolution",false },
  {0x0249,0x0020,"GEMS_RTEN_01",VR::SQ,VM::VM1,"Structure Sequence",false },
  {0x0249,0x0022,"GEMS_RTEN_01",VR::SH,VM::VM1,"Structure Name",false },
  {0x0249,0x0024,"GEMS_RTEN_01",VR::CS,VM::VM1,"Structure Type",false },
  {0x0249,0x0028,"GEMS_RTEN_01",VR::CS,VM::VM1,"Structure Color",false },
  {0x0249,0x0030,"GEMS_RTEN_01",VR::SQ,VM::VM1,"Slab Sequence",false },
  {0x0249,0x0032,"GEMS_RTEN_01",VR::SQ,VM::VM1,"Slab Image Sequence",false },
  {0x0249,0x0036,"GEMS_RTEN_01",VR::DS,VM::VM1,"Z Plus Thickness",false },
  {0x0249,0x0038,"GEMS_RTEN_01",VR::DS,VM::VM1,"Z Minus Thickness",false },
  {0x0249,0x003a,"GEMS_RTEN_01",VR::DS,VM::VM3_n,"Slab Shape",false },
  {0x0249,0x0040,"GEMS_RTEN_01",VR::SQ,VM::VM1,"Marker Sequence",false },
  {0x0249,0x0042,"GEMS_RTEN_01",VR::SH,VM::VM1,"Marker Name",false },
  {0x0249,0x0044,"GEMS_RTEN_01",VR::CS,VM::VM1,"Marker Color",false },
  {0x0249,0x0046,"GEMS_RTEN_01",VR::DS,VM::VM3,"Marker Position",false },
  {0x0249,0x0050,"GEMS_RTEN_01",VR::SQ,VM::VM1,"Beam Group Sequence",false },
  {0x0249,0x0051,"GEMS_RTEN_01",VR::SH,VM::VM1,"Group Name",false },
  {0x0249,0x0052,"GEMS_RTEN_01",VR::CS,VM::VM1_n,"Group Properties",false },
  {0x0249,0x0054,"GEMS_RTEN_01",VR::SQ,VM::VM1,"Beam Sequence",false },
  {0x0249,0x0060,"GEMS_RTEN_01",VR::SQ,VM::VM1,"Machine Reference Sequence",false },
  {0x0249,0x0062,"GEMS_RTEN_01",VR::SH,VM::VM1,"Machine Name",false },
  {0x0249,0x0064,"GEMS_RTEN_01",VR::SH,VM::VM1,"Beam Name",false },
  {0x0249,0x0066,"GEMS_RTEN_01",VR::CS,VM::VM1,"Particle Type",false },
  {0x0249,0x0068,"GEMS_RTEN_01",VR::DS,VM::VM1,"Nominal Particle Energy",false },
  {0x0249,0x0070,"GEMS_RTEN_01",VR::SQ,VM::VM1,"Block Sequence",false },
  {0x0249,0x0072,"GEMS_RTEN_01",VR::LO,VM::VM1,"Block Name",false },
  {0x0249,0x0074,"GEMS_RTEN_01",VR::DS,VM::VM2_n,"Block Shape",false },
  {0x0249,0x0080,"GEMS_RTEN_01",VR::SQ,VM::VM1,"Cutout Sequence",false },
  {0x0249,0x0082,"GEMS_RTEN_01",VR::LO,VM::VM1,"Cutout Name",false },
  {0x0249,0x0084,"GEMS_RTEN_01",VR::DS,VM::VM2_n,"Cutout Shape",false },
  {0x0249,0x0090,"GEMS_RTEN_01",VR::SH,VM::VM1,"Collimator Name",false },
  {0x0249,0x0092,"GEMS_RTEN_01",VR::CS,VM::VM1,"Collimator Type",false },
  {0x0249,0x00a0,"GEMS_RTEN_01",VR::SQ,VM::VM1,"Dynamic Segment Sequence",false },
  {0x0249,0x00a2,"GEMS_RTEN_01",VR::DS,VM::VM1,"Table Angle",false },
  {0x0249,0x00a4,"GEMS_RTEN_01",VR::DS,VM::VM1,"Gantry Angle",false },
  {0x0249,0x00a6,"GEMS_RTEN_01",VR::DS,VM::VM1,"Collimator Angle",false },
  {0x0249,0x00a8,"GEMS_RTEN_01",VR::DS,VM::VM3,"Isocenter Position",false },
  {0x0249,0x00a9,"GEMS_RTEN_01",VR::DS,VM::VM1,"Source-Surface Distance",false },
  {0x0249,0x00aa,"GEMS_RTEN_01",VR::DS,VM::VM1,"X Symmetric Opening",false },
  {0x0249,0x00ac,"GEMS_RTEN_01",VR::DS,VM::VM1,"Y Symmetric Opening",false },
  {0x0249,0x00ae,"GEMS_RTEN_01",VR::DS,VM::VM1,"X Positive Jaw Position",false },
  {0x0249,0x00b0,"GEMS_RTEN_01",VR::DS,VM::VM1,"X Negative Jaw Position",false },
  {0x0249,0x00b2,"GEMS_RTEN_01",VR::DS,VM::VM1,"Y Positive Jaw Position",false },
  {0x0249,0x00b4,"GEMS_RTEN_01",VR::DS,VM::VM1,"Y Negative Jaw Position",false },
  {0x0249,0x00b6,"GEMS_RTEN_01",VR::DS,VM::VM2_n,"Leaf Positions",false },
  {0x0249,0x00d0,"GEMS_RTEN_01",VR::CS,VM::VM1,"Treatment Position",false },
  {0x0045,0x0004,"GEMS_SENO_02",VR::CS,VM::VM1,"AES",false },
  {0x0045,0x0006,"GEMS_SENO_02",VR::DS,VM::VM1,"Stereo angle",false },
  {0x0045,0x0009,"GEMS_SENO_02",VR::DS,VM::VM1,"Real Magnification Factor",false },
  {0x0045,0x000b,"GEMS_SENO_02",VR::CS,VM::VM1,"Senograph Type",false },
  {0x0045,0x000c,"GEMS_SENO_02",VR::DS,VM::VM1,"Integration Time",false },
  {0x0045,0x000d,"GEMS_SENO_02",VR::DS,VM::VM1,"ROI Origin X and Y",false },
  {0x0045,0x0011,"GEMS_SENO_02",VR::DS,VM::VM2,"Receptor Size cm X and Y",false },
  {0x0045,0x0012,"GEMS_SENO_02",VR::IS,VM::VM2,"Receptor Size Pixels X and Y",false },
  {0x0045,0x0013,"GEMS_SENO_02",VR::ST,VM::VM1,"Screen",false },
  {0x0045,0x0014,"GEMS_SENO_02",VR::DS,VM::VM1,"Pixel Pitch Microns",false },
  {0x0045,0x0015,"GEMS_SENO_02",VR::IS,VM::VM1,"Pixel Depth Bits",false },
  {0x0045,0x0016,"GEMS_SENO_02",VR::IS,VM::VM2,"Binning Factor X and Y",false },
  {0x0045,0x001b,"GEMS_SENO_02",VR::CS,VM::VM1,"Clinical View",false },
  {0x0045,0x001d,"GEMS_SENO_02",VR::DS,VM::VM1,"Mean Of Raw Gray Levels",false },
  {0x0045,0x001e,"GEMS_SENO_02",VR::DS,VM::VM1,"Mean Of Offset Gray Levels",false },
  {0x0045,0x001f,"GEMS_SENO_02",VR::DS,VM::VM1,"Mean Of Corrected Gray Levels",false },
  {0x0045,0x0020,"GEMS_SENO_02",VR::DS,VM::VM1,"Estimated Anat mean",false },
  {0x0045,0x0021,"GEMS_SENO_02",VR::DS,VM::VM1,"Mean Of Log Region Gray Levels",false },
  {0x0045,0x0022,"GEMS_SENO_02",VR::DS,VM::VM1,"Standard Deviation Of Raw Gray Levels",false },
  {0x0045,0x0023,"GEMS_SENO_02",VR::DS,VM::VM1,"Standard Deviation Of Corrected Gray Levels",false },
  {0x0045,0x0024,"GEMS_SENO_02",VR::DS,VM::VM1,"Standard Deviation Of Region Gray Levels",false },
  {0x0045,0x0025,"GEMS_SENO_02",VR::DS,VM::VM1,"Standard Deviation Of Log Region Gray Levels",false },
  {0x0045,0x0026,"GEMS_SENO_02",VR::OB,VM::VM1,"MAO Buffer",false },
  {0x0045,0x0027,"GEMS_SENO_02",VR::IS,VM::VM1,"Set Number",false },
  {0x0045,0x0028,"GEMS_SENO_02",VR::CS,VM::VM1,"WindowingType (LINEAR or GAMMA)",false },
  {0x0045,0x0029,"GEMS_SENO_02",VR::DS,VM::VM2,"Windowing parameters",false },
  {0x0045,0x002a,"GEMS_SENO_02",VR::IS,VM::VM1,"2DLocX",false },
  {0x0045,0x002b,"GEMS_SENO_02",VR::IS,VM::VM1,"2DLocY",false },
  {0x0045,0x0039,"GEMS_SENO_02",VR::US,VM::VM1,"Vignette Rows",false },
  {0x0045,0x003a,"GEMS_SENO_02",VR::US,VM::VM1,"Vignette Columns",false },
  {0x0045,0x003b,"GEMS_SENO_02",VR::US,VM::VM1,"Vignette Bits Allocated",false },
  {0x0045,0x003c,"GEMS_SENO_02",VR::US,VM::VM1,"Vignette Bits Stored",false },
  {0x0045,0x003d,"GEMS_SENO_02",VR::US,VM::VM1,"Vignette High Bit",false },
  {0x0045,0x003e,"GEMS_SENO_02",VR::US,VM::VM1,"Vignette Pixel Representation",false },
  {0x0045,0x003f,"GEMS_SENO_02",VR::OB,VM::VM1,"Vignette Pixel Data",false },
  {0x0045,0x0049,"GEMS_SENO_02",VR::DS,VM::VM1,"Radiological Thickness",false },
  {0x0045,0x0050,"GEMS_SENO_02",VR::UI,VM::VM1,"CR Instance UID",false },
  {0x0045,0x0051,"GEMS_SENO_02",VR::UI,VM::VM1,"CR series UID",false },
  {0x0045,0x0052,"GEMS_SENO_02",VR::IS,VM::VM1,"Raw Diagnostic Low",false },
  {0x0045,0x0053,"GEMS_SENO_02",VR::IS,VM::VM1,"Raw Diagnostic High",false },
  {0x0045,0x0054,"GEMS_SENO_02",VR::DS,VM::VM1,"Exponent",false },
  {0x0045,0x0055,"GEMS_SENO_02",VR::IS,VM::VM1_n,"A Coefficients",false },
  {0x0045,0x0056,"GEMS_SENO_02",VR::DS,VM::VM1,"Noise Reduction Sensitivity",false },
  {0x0045,0x0057,"GEMS_SENO_02",VR::DS,VM::VM1,"Noise Reduction Threshold",false },
  {0x0045,0x0058,"GEMS_SENO_02",VR::DS,VM::VM1,"mu",false },
  {0x0045,0x0059,"GEMS_SENO_02",VR::IS,VM::VM1,"Threshold",false },
  {0x0045,0x0060,"GEMS_SENO_02",VR::IS,VM::VM4,"Breast ROI X",false },
  {0x0045,0x0061,"GEMS_SENO_02",VR::IS,VM::VM4,"Breast ROI Y",false },
  {0x0045,0x0062,"GEMS_SENO_02",VR::IS,VM::VM1,"User Window Center",false },
  {0x0045,0x0063,"GEMS_SENO_02",VR::IS,VM::VM1,"User Window Width",false },
  {0x0045,0x0064,"GEMS_SENO_02",VR::IS,VM::VM1,"Segm Threshold",false },
  {0x0045,0x0065,"GEMS_SENO_02",VR::IS,VM::VM1,"Detector Entrance Dose",false },
  {0x0045,0x0066,"GEMS_SENO_02",VR::IS,VM::VM1,"Asymmetrical Collimation Information",false },
  {0x0045,0x0067,"GEMS_SENO_02",VR::DS,VM::VM3_4,"LUT Asymmetry parameter",false },
  {0x0045,0x0068,"GEMS_SENO_02",VR::CS,VM::VM3,"Acquisition Mode",false },
  {0x0045,0x0069,"GEMS_SENO_02",VR::IS,VM::VM1,"Collimator Rotation",false },
  {0x0045,0x0070,"GEMS_SENO_02",VR::DS,VM::VM1,"Brightness Adjustment",false },
  {0x0045,0x0071,"GEMS_SENO_02",VR::DS,VM::VM1,"Contrast Adjustment",false },
  {0x0045,0x0072,"GEMS_SENO_02",VR::DS,VM::VM1,"Collimator Width",false },
  {0x0045,0x0073,"GEMS_SENO_02",VR::DS,VM::VM1,"Collimator Height",false },
  {0x0025,0x0006,"GEMS_SERS_01",VR::SS,VM::VM1,"Last pulse sequence used",false },
  {0x0025,0x0007,"GEMS_SERS_01",VR::SL,VM::VM1,"Images in Series",false },
  {0x0025,0x0010,"GEMS_SERS_01",VR::SL,VM::VM1,"Landmark Counter",false },
  {0x0025,0x0011,"GEMS_SERS_01",VR::SS,VM::VM1,"Number of Acquisitions",false },
  {0x0025,0x0014,"GEMS_SERS_01",VR::SL,VM::VM1,"Indicates no. of updates to header",false },
  {0x0025,0x0017,"GEMS_SERS_01",VR::SL,VM::VM1,"Series Complete Flag",false },
  {0x0025,0x0018,"GEMS_SERS_01",VR::SL,VM::VM1,"Number of images archived",false },
  {0x0025,0x0019,"GEMS_SERS_01",VR::SL,VM::VM1,"Last image number used",false },
  {0x0025,0x001a,"GEMS_SERS_01",VR::SH,VM::VM1,"Primary Receiver Suite and Host",false },
  {0x0025,0x001b,"GEMS_SERS_01",VR::OB,VM::VM1,"Protocol Data Block (compressed)",false },
  {0x0023,0x0001,"GEMS_STDY_01",VR::SL,VM::VM1,"Number of series in Study",false },
  {0x0023,0x0002,"GEMS_STDY_01",VR::SL,VM::VM1,"Number of unarchived Series",false },
  {0x0023,0x0010,"GEMS_STDY_01",VR::SS,VM::VM1,"Reference image field",false },
  {0x0023,0x0050,"GEMS_STDY_01",VR::SS,VM::VM1,"Summary image",false },
  {0x0023,0x0070,"GEMS_STDY_01",VR::FD,VM::VM1,"Start time(secs) in first axial",false },
  {0x0023,0x0074,"GEMS_STDY_01",VR::SL,VM::VM1,"No. of updates to header",false },
  {0x0023,0x007d,"GEMS_STDY_01",VR::SS,VM::VM1,"Indicates study has complete info (DICOM/genesis)",false },
  {0x0023,0x0080,"GEMS_STDY_01",VR::SQ,VM::VM1,"Has MPPS related tags",false },
  {0x6003,0x0005,"GEMS_Ultrasound_ImageGroup_001",VR::LT,VM::VM1_n,"GEMS Image Information",false },
  {0x6003,0x0010,"GEMS_Ultrasound_ImageGroup_001",VR::SQ,VM::VM1,"GEMS Image Thumbnail Sequence",false },
  {0x6003,0x0011,"GEMS_Ultrasound_ImageGroup_001",VR::OB,VM::VM1,"GEMS Thumbnail Pixel Data (uncompressed)",false },
  {0x7fe1,0x0001,"GEMS_Ultrasound_MovieGroup_001",VR::SQ,VM::VM1,"?",false },
  {0x7fe1,0x0002,"GEMS_Ultrasound_MovieGroup_001",VR::LO,VM::VM1,"US MovieGroup Value 0008 Name",false },
  {0x7fe1,0x0003,"GEMS_Ultrasound_MovieGroup_001",VR::UL,VM::VM1,"?",false },
  {0x7fe1,0x0008,"GEMS_Ultrasound_MovieGroup_001",VR::SQ,VM::VM1,"US MovieGroup Value 0008 Sequence",false },
  {0x7fe1,0x0010,"GEMS_Ultrasound_MovieGroup_001",VR::SQ,VM::VM1,"?",false },
  {0x7fe1,0x0012,"GEMS_Ultrasound_MovieGroup_001",VR::LO,VM::VM1,"US MovieGroup Value 0018 Name",false },
  {0x7fe1,0x0018,"GEMS_Ultrasound_MovieGroup_001",VR::SQ,VM::VM1,"US MovieGroup Value 0018 Sequence",false },
  {0x7fe1,0x0020,"GEMS_Ultrasound_MovieGroup_001",VR::SQ,VM::VM1,"?",false },
  {0x7fe1,0x0024,"GEMS_Ultrasound_MovieGroup_001",VR::SH,VM::VM1,"US MovieGroup Value 0026 Name",false },
  {0x7fe1,0x0026,"GEMS_Ultrasound_MovieGroup_001",VR::SQ,VM::VM1,"US MovieGroup Value 0026 Sequence",false },
  {0x7fe1,0x0030,"GEMS_Ultrasound_MovieGroup_001",VR::LO,VM::VM1,"?",false },
  {0x7fe1,0x0032,"GEMS_Ultrasound_MovieGroup_001",VR::UL,VM::VM1,"?",false },
  {0x7fe1,0x0036,"GEMS_Ultrasound_MovieGroup_001",VR::SQ,VM::VM1,"?",false },
  {0x7fe1,0x0037,"GEMS_Ultrasound_MovieGroup_001",VR::UL,VM::VM1,"US MovieGroup Number of Frames",false },
  {0x7fe1,0x003a,"GEMS_Ultrasound_MovieGroup_001",VR::SQ,VM::VM1,"?",false },
  {0x7fe1,0x003c,"GEMS_Ultrasound_MovieGroup_001",VR::FD,VM::VM1,"?",false },
  {0x7fe1,0x0043,"GEMS_Ultrasound_MovieGroup_001",VR::OB,VM::VM1,"?",false },
  {0x7fe1,0x0048,"GEMS_Ultrasound_MovieGroup_001",VR::FD,VM::VM1_2,"US MovieGroup Value Index",false },
  {0x7fe1,0x0049,"GEMS_Ultrasound_MovieGroup_001",VR::UL,VM::VM1,"US MovieGroup Value UL/1",false },
  {0x7fe1,0x0051,"GEMS_Ultrasound_MovieGroup_001",VR::FL,VM::VM1,"US MovieGroup Value FL/1",false },
  {0x7fe1,0x0052,"GEMS_Ultrasound_MovieGroup_001",VR::FD,VM::VM1,"US MovieGroup Value FD/1",false },
  {0x7fe1,0x0053,"GEMS_Ultrasound_MovieGroup_001",VR::UL,VM::VM1,"US MovieGroup Value UL/1",false },
  {0x7fe1,0x0054,"GEMS_Ultrasound_MovieGroup_001",VR::SL,VM::VM1,"US MovieGroup Value SL/1",false },
  {0x7fe1,0x0055,"GEMS_Ultrasound_MovieGroup_001",VR::OB,VM::VM1,"US MovieGroup Value OB/1",false },
  {0x7fe1,0x0057,"GEMS_Ultrasound_MovieGroup_001",VR::LT,VM::VM1,"US MovieGroup Value LT/1",false },
  {0x7fe1,0x0060,"GEMS_Ultrasound_MovieGroup_001",VR::OB,VM::VM1,"US MovieGroup Image Data",false },
  {0x7fe1,0x0061,"GEMS_Ultrasound_MovieGroup_001",VR::OW,VM::VM1,"?",false },
  {0x7fe1,0x0069,"GEMS_Ultrasound_MovieGroup_001",VR::OW,VM::VM1,"?",false },
  {0x7fe1,0x0070,"GEMS_Ultrasound_MovieGroup_001",VR::SQ,VM::VM1,"US MovieGroup Dict",false },
  {0x7fe1,0x0071,"GEMS_Ultrasound_MovieGroup_001",VR::UL,VM::VM1,"US MovieGroup Dict Index",false },
  {0x7fe1,0x0072,"GEMS_Ultrasound_MovieGroup_001",VR::LO,VM::VM1,"US MovieGroup Dict Name",false },
  {0x7fe1,0x0073,"GEMS_Ultrasound_MovieGroup_001",VR::SQ,VM::VM1,"?",false },
  {0x7fe1,0x0074,"GEMS_Ultrasound_MovieGroup_001",VR::LO,VM::VM1,"?",false },
  {0x7fe1,0x0075,"GEMS_Ultrasound_MovieGroup_001",VR::SQ,VM::VM1,"?",false },
  {0x7fe1,0x0077,"GEMS_Ultrasound_MovieGroup_001",VR::FD,VM::VM1_n,"US MovieGroup Value FD/1-N",false },
  {0x7fe1,0x0079,"GEMS_Ultrasound_MovieGroup_001",VR::SL,VM::VM1_n,"US MovieGroup Value SL/1-N",false },
  {0x7fe1,0x0083,"GEMS_Ultrasound_MovieGroup_001",VR::SQ,VM::VM1,"?",false },
  {0x7fe1,0x0084,"GEMS_Ultrasound_MovieGroup_001",VR::LO,VM::VM1,"US MovieGroup Value 0085 Name",false },
  {0x7fe1,0x0085,"GEMS_Ultrasound_MovieGroup_001",VR::SQ,VM::VM1,"US MovieGroup Value 0085 Sequence",false },
  {0x7fe1,0x0086,"GEMS_Ultrasound_MovieGroup_001",VR::SL,VM::VM4,"US MovieGroup Value SL/4",false },
  {0x7fe1,0x0087,"GEMS_Ultrasound_MovieGroup_001",VR::FD,VM::VM4,"US MovieGroup Value FD/4",false },
  {0x7fe1,0x0088,"GEMS_Ultrasound_MovieGroup_001",VR::FD,VM::VM2,"US MovieGroup Value FD/2",false },
  {0x0057,0x0005,"GEMS_VXTLSTATE_001",VR::SQ,VM::VM1,"Volumes_info",false },
  {0x0057,0x0006,"GEMS_VXTLSTATE_001",VR::SQ,VM::VM1,"Image_UIDs",false },
  {0x0057,0x0007,"GEMS_VXTLSTATE_001",VR::DT,VM::VM1,"SUV_ScanTime",false },
  {0x0057,0x0008,"GEMS_VXTLSTATE_001",VR::DT,VM::VM1,"SUV_AdministredTime",false },
  {0x0057,0x0009,"GEMS_VXTLSTATE_001",VR::DT,VM::VM1,"SUV_MeasuredTime",false },
  {0x0057,0x0010,"GEMS_VXTLSTATE_001",VR::DT,VM::VM1,"SUV_PostInjectedTime",false },
  {0x0057,0x0011,"GEMS_VXTLSTATE_001",VR::FL,VM::VM1,"SUV_TracerActivity",false },
  {0x0057,0x0012,"GEMS_VXTLSTATE_001",VR::FL,VM::VM1,"SUV_PostInjectedActivity",false },
  {0x0057,0x0013,"GEMS_VXTLSTATE_001",VR::FL,VM::VM1,"SUV_HalfLife",false },
  {0x0057,0x0014,"GEMS_VXTLSTATE_001",VR::LO,VM::VM1,"SState_Version",false },
  {0x0057,0x0015,"GEMS_VXTLSTATE_001",VR::SQ,VM::VM1,"SegList_Seq",false },
  {0x0057,0x0016,"GEMS_VXTLSTATE_001",VR::IS,VM::VM1,"SegList_Count",false },
  {0x0057,0x0017,"GEMS_VXTLSTATE_001",VR::OB,VM::VM1,"SegList_List",false },
  {0x0057,0x0018,"GEMS_VXTLSTATE_001",VR::LO,VM::VM1,"SegList_Name",false },
  {0x0057,0x0019,"GEMS_VXTLSTATE_001",VR::LO,VM::VM1,"Volume_Filename",false },
  {0x0057,0x0020,"GEMS_VXTLSTATE_001",VR::LO,VM::VM1,"SegList_Label",false },
  {0x0057,0x0021,"GEMS_VXTLSTATE_001",VR::SQ,VM::VM1,"Bookmark_Seq",false },
  {0x0057,0x0022,"GEMS_VXTLSTATE_001",VR::LT,VM::VM1,"Bookmark",false },
  {0x0057,0x0023,"GEMS_VXTLSTATE_001",VR::FL,VM::VM3,"Cursor_position",false },
  {0x0057,0x0024,"GEMS_VXTLSTATE_001",VR::UL,VM::VM3_n,"Color_Value_Field",false },
  {0x0057,0x0025,"GEMS_VXTLSTATE_001",VR::IS,VM::VM1,"Count",false },
  {0x0057,0x0026,"GEMS_VXTLSTATE_001",VR::SQ,VM::VM1,"Views_info",false },
  {0x0057,0x0027,"GEMS_VXTLSTATE_001",VR::LT,VM::VM1,"View_Slot",false },
  {0x0057,0x0028,"GEMS_VXTLSTATE_001",VR::IS,VM::VM1,"Wireframe_Size",false },
  {0x0057,0x0029,"GEMS_VXTLSTATE_001",VR::UT,VM::VM1,"Wireframe",false },
  {0x0057,0x0030,"GEMS_VXTLSTATE_001",VR::IS,VM::VM1,"Annotation_Size",false },
  {0x0057,0x0031,"GEMS_VXTLSTATE_001",VR::LT,VM::VM1,"Annotation",false },
  {0x0057,0x0032,"GEMS_VXTLSTATE_001",VR::IS,VM::VM1,"Slider_State",false },
  {0x0057,0x0033,"GEMS_VXTLSTATE_001",VR::LO,VM::VM1,"Proto_Name",false },
  {0x0057,0x0034,"GEMS_VXTLSTATE_001",VR::LO,VM::VM1,"Proto_Title",false },
  {0x0057,0x0035,"GEMS_VXTLSTATE_001",VR::LO,VM::VM1,"Proto_FilmName",false },
  {0x0057,0x0036,"GEMS_VXTLSTATE_001",VR::LO,VM::VM1,"Proto_Step",false },
  {0x0057,0x0038,"GEMS_VXTLSTATE_001",VR::LT,VM::VM1,"SegList_Slots",false },
  {0x0057,0x0039,"GEMS_VXTLSTATE_001",VR::IS,VM::VM1,"VT_Points_Size",false },
  {0x0057,0x0040,"GEMS_VXTLSTATE_001",VR::LT,VM::VM1,"VT_Points",false },
  {0x0057,0x0041,"GEMS_VXTLSTATE_001",VR::IS,VM::VM1,"VT_Meas_Size",false },
  {0x0057,0x0042,"GEMS_VXTLSTATE_001",VR::LT,VM::VM1,"VT_Meas",false },
  {0x0057,0x0043,"GEMS_VXTLSTATE_001",VR::IS,VM::VM1,"VT_Tree_Size",false },
  {0x0057,0x0044,"GEMS_VXTLSTATE_001",VR::UT,VM::VM1,"VT_Tree",false },
  {0x0057,0x0046,"GEMS_VXTLSTATE_001",VR::IS,VM::VM1,"VT_Preset_Size",false },
  {0x0057,0x0047,"GEMS_VXTLSTATE_001",VR::LT,VM::VM1,"VT_Preset",false },
  {0x0057,0x0048,"GEMS_VXTLSTATE_001",VR::IS,VM::VM1,"VT_State_Size",false },
  {0x0057,0x0049,"GEMS_VXTLSTATE_001",VR::LT,VM::VM1,"VT_State",false },
  {0x0057,0x0050,"GEMS_VXTLSTATE_001",VR::IS,VM::VM1,"Preferences_Size",false },
  {0x0057,0x0051,"GEMS_VXTLSTATE_001",VR::LT,VM::VM1,"Preferences",false },
  {0x0057,0x0052,"GEMS_VXTLSTATE_001",VR::IS,VM::VM1,"Volume_ZComb_Filter",false },
  {0x0057,0x0053,"GEMS_VXTLSTATE_001",VR::IS,VM::VM1,"HTML_Page_Size",false },
  {0x0057,0x0054,"GEMS_VXTLSTATE_001",VR::LT,VM::VM1,"HTML_Page",false },
  {0x0057,0x0055,"GEMS_VXTLSTATE_001",VR::FL,VM::VM3,"Camera_Position",false },
  {0x0057,0x0056,"GEMS_VXTLSTATE_001",VR::IS,VM::VM1,"Slider_Size",false },
  {0x0057,0x0057,"GEMS_VXTLSTATE_001",VR::LT,VM::VM1,"Slider",false },
  {0x0057,0x0058,"GEMS_VXTLSTATE_001",VR::LO,VM::VM1,"SegList_Pretty_Name",false },
  {0x0047,0x0011,"GEMS_VXTL_USERDATA_01",VR::LT,VM::VM1,"?",false },
  {0x0033,0x0008,"GEMS_XELPRV_01",VR::CS,VM::VM1,"Object Type",false },
  {0x0033,0x0010,"GEMS_XELPRV_01",VR::SL,VM::VM1,"Modified",false },
  {0x0033,0x0011,"GEMS_XELPRV_01",VR::LO,VM::VM1,"Name",false },
  {0x0033,0x0014,"GEMS_XELPRV_01",VR::LO,VM::VM1,"StudyId",false },
  {0x0033,0x0016,"GEMS_XELPRV_01",VR::LO,VM::VM1,"ProtocolDataUID",false },
  {0x0033,0x0017,"GEMS_XELPRV_01",VR::SH,VM::VM1,"Date",false },
  {0x0033,0x0018,"GEMS_XELPRV_01",VR::SH,VM::VM1,"Time",false },
  {0x0033,0x0019,"GEMS_XELPRV_01",VR::UL,VM::VM1,"ProtocoldataFlags",false },
  {0x0033,0x001a,"GEMS_XELPRV_01",VR::LO,VM::VM1,"ProtocolName",false },
  {0x0033,0x001b,"GEMS_XELPRV_01",VR::LO,VM::VM1,"StudyId",false },
  {0x0033,0x001c,"GEMS_XELPRV_01",VR::OB,VM::VM1,"BulkData",false },
  {0x0033,0x001d,"GEMS_XELPRV_01",VR::SL,VM::VM1_n,"IntData",false },
  {0x0033,0x001e,"GEMS_XELPRV_01",VR::FD,VM::VM1_n,"DoubleData",false },
  {0x0033,0x001f,"GEMS_XELPRV_01",VR::OB,VM::VM1,"StringData",false },
  {0x0033,0x0020,"GEMS_XELPRV_01",VR::OB,VM::VM1,"BulkDataFormat",false },
  {0x0033,0x0021,"GEMS_XELPRV_01",VR::OB,VM::VM1,"IntDataFormat",false },
  {0x0033,0x0022,"GEMS_XELPRV_01",VR::OB,VM::VM1,"DoubleDataFormat",false },
  {0x0033,0x0023,"GEMS_XELPRV_01",VR::OB,VM::VM1,"StringDataFormat",false },
  {0x0033,0x0024,"GEMS_XELPRV_01",VR::LT,VM::VM1,"Description",false },
  {0x0033,0x0028,"GEMS_XELPRV_01",VR::LO,VM::VM1,"RTName",false },
  {0x0033,0x0029,"GEMS_XELPRV_01",VR::LT,VM::VM1,"RTSpecification",false },
  {0x0033,0x002a,"GEMS_XELPRV_01",VR::UL,VM::VM1,"ReviewTemplatesFlags",false },
  {0x0033,0x002b,"GEMS_XELPRV_01",VR::LT,VM::VM1,"DataValidationSpec",false },
  {0x0033,0x002c,"GEMS_XELPRV_01",VR::LT,VM::VM1,"Description",false },
  {0x0033,0x002d,"GEMS_XELPRV_01",VR::LT,VM::VM1,"IconDescription",false },
  {0x0033,0x0050,"GEMS_XELPRV_01",VR::SQ,VM::VM1,"ProtocoldataSQ",false },
  {0x0033,0x0051,"GEMS_XELPRV_01",VR::UI,VM::VM1,"PDOSOPClassUID",false },
  {0x0033,0x0052,"GEMS_XELPRV_01",VR::UI,VM::VM1,"PDOSOPInstanceUID",false },
  {0x0033,0x0060,"GEMS_XELPRV_01",VR::SQ,VM::VM1,"ReviewtemplateSQ",false },
  {0x0033,0x0061,"GEMS_XELPRV_01",VR::UI,VM::VM1,"RTOSOPClassUID",false },
  {0x0033,0x0062,"GEMS_XELPRV_01",VR::UI,VM::VM1,"RTOSOPInstanceUID",false },
  {0x0033,0x0070,"GEMS_XELPRV_01",VR::SQ,VM::VM1,"SeriesdataSQ",false },
  {0x0033,0x0071,"GEMS_XELPRV_01",VR::UI,VM::VM1,"SDOSOPClassUID",false },
  {0x0033,0x0072,"GEMS_XELPRV_01",VR::UI,VM::VM1,"SDOSOPInstanceUID",false },
  {0x0033,0x0073,"GEMS_XELPRV_01",VR::SQ,VM::VM1,"SDODoubleDataSQ",false },
  {0x0057,0x0001,"GEMS_XELPRV_01",VR::SQ,VM::VM1,"ROISQ",false },
  {0x0057,0x0002,"GEMS_XELPRV_01",VR::UI,VM::VM1,"SOPClassUID",false },
  {0x0057,0x0003,"GEMS_XELPRV_01",VR::UI,VM::VM1,"SOPInstanceUID",false },
  {0x0057,0x0010,"GEMS_XELPRV_01",VR::IS,VM::VM1,"ROINumber",false },
  {0x0057,0x0011,"GEMS_XELPRV_01",VR::UI,VM::VM1,"Dimensions",false },
  {0x0057,0x0012,"GEMS_XELPRV_01",VR::UI,VM::VM1,"Points",false },
  {0x0057,0x0013,"GEMS_XELPRV_01",VR::UI,VM::VM1,"Type",false },
  {0x0057,0x0014,"GEMS_XELPRV_01",VR::UI,VM::VM1,"Description",false },
  {0x0057,0x0015,"GEMS_XELPRV_01",VR::UI,VM::VM1,"Data value Representation",false },
  {0x0057,0x0016,"GEMS_XELPRV_01",VR::UI,VM::VM1,"Label",false },
  {0x0057,0x0017,"GEMS_XELPRV_01",VR::UI,VM::VM1,"Data",false },
  {0x0057,0x0041,"GEMS_XELPRV_01",VR::SL,VM::VM1,"Modified",false },
  {0x0057,0x0042,"GEMS_XELPRV_01",VR::LO,VM::VM1,"Name",false },
  {0x0057,0x0043,"GEMS_XELPRV_01",VR::SL,VM::VM1,"Rid",false },
  {0x0057,0x0045,"GEMS_XELPRV_01",VR::LO,VM::VM1,"SOPClassUID",false },
  {0x0057,0x0046,"GEMS_XELPRV_01",VR::LO,VM::VM1,"SOPInstanceUID",false },
  {0x0057,0x0047,"GEMS_XELPRV_01",VR::LO,VM::VM1,"Normal Colour",false },
  {0x0057,0x0048,"GEMS_XELPRV_01",VR::LT,VM::VM1,"NameFont",false },
  {0x0057,0x0049,"GEMS_XELPRV_01",VR::SL,VM::VM1,"FillPattern",false },
  {0x0057,0x004a,"GEMS_XELPRV_01",VR::SL,VM::VM1,"LineStyle",false },
  {0x0057,0x004b,"GEMS_XELPRV_01",VR::SL,VM::VM1,"LineDashLength",false },
  {0x0057,0x004c,"GEMS_XELPRV_01",VR::SL,VM::VM1,"LineThickness",false },
  {0x0057,0x004d,"GEMS_XELPRV_01",VR::SL,VM::VM1,"Interactivity",false },
  {0x0057,0x004e,"GEMS_XELPRV_01",VR::SL,VM::VM1,"NamePos",false },
  {0x0057,0x004f,"GEMS_XELPRV_01",VR::SL,VM::VM1,"NameDisplay",false },
  {0x0057,0x0050,"GEMS_XELPRV_01",VR::LO,VM::VM1,"Label",false },
  {0x0057,0x0051,"GEMS_XELPRV_01",VR::SL,VM::VM1_n,"BpSeg",false },
  {0x0057,0x0052,"GEMS_XELPRV_01",VR::US,VM::VM1_n,"BpSegpairs",false },
  {0x0057,0x0053,"GEMS_XELPRV_01",VR::SL,VM::VM1,"SeedSpace",false },
  {0x0057,0x0054,"GEMS_XELPRV_01",VR::FD,VM::VM1_n,"Seeds",false },
  {0x0057,0x0055,"GEMS_XELPRV_01",VR::SL,VM::VM1_n,"Shape",false },
  {0x0057,0x0056,"GEMS_XELPRV_01",VR::FD,VM::VM1_n,"ShapeTilt",false },
  {0x0057,0x0059,"GEMS_XELPRV_01",VR::SL,VM::VM1_n,"ShapePtsSpace",false },
  {0x0057,0x005a,"GEMS_XELPRV_01",VR::SL,VM::VM1_n,"ShapeCtrlPtsCount",false },
  {0x0057,0x005b,"GEMS_XELPRV_01",VR::FD,VM::VM1_n,"Shap CtrlPts",false },
  {0x0057,0x005c,"GEMS_XELPRV_01",VR::SL,VM::VM1,"ShapeCPSpace",false },
  {0x0057,0x005d,"GEMS_XELPRV_01",VR::UL,VM::VM1_n,"ROIFlags",false },
  {0x0057,0x005e,"GEMS_XELPRV_01",VR::UL,VM::VM1,"FrameNumber",false },
  {0x0057,0x0060,"GEMS_XELPRV_01",VR::LO,VM::VM1_n,"DatasetROIMapping",false },
  {0x0021,0x0001,"GEMS_XR3DCAL_01",VR::IS,VM::VM1,"3Dcal image rows",false },
  {0x0021,0x0002,"GEMS_XR3DCAL_01",VR::IS,VM::VM1,"3Dcal image columns",false },
  {0x0021,0x0003,"GEMS_XR3DCAL_01",VR::FL,VM::VM1,"3Dcal field of view",false },
  {0x0021,0x0004,"GEMS_XR3DCAL_01",VR::DA,VM::VM1,"3Dcal acquisition date",false },
  {0x0021,0x0005,"GEMS_XR3DCAL_01",VR::TM,VM::VM1,"3Dcal acquisition time",false },
  {0x0021,0x0006,"GEMS_XR3DCAL_01",VR::DA,VM::VM1,"3Dcal calibration processing date",false },
  {0x0021,0x0007,"GEMS_XR3DCAL_01",VR::TM,VM::VM1,"3Dcal calibration processing time",false },
  {0x0021,0x0008,"GEMS_XR3DCAL_01",VR::FL,VM::VM1,"3Dcal L arm angle",false },
  {0x0021,0x0009,"GEMS_XR3DCAL_01",VR::FL,VM::VM1_n,"3Dcal Pivot angle vector",false },
  {0x0021,0x000a,"GEMS_XR3DCAL_01",VR::FL,VM::VM1,"3Dcal C arm angle",false },
  {0x0021,0x000b,"GEMS_XR3DCAL_01",VR::SQ,VM::VM1,"3Dcal matrix sequence",false },
  {0x0021,0x000c,"GEMS_XR3DCAL_01",VR::LO,VM::VM1_n,"3Dcal matrix elements",false },
  {0x0021,0x000d,"GEMS_XR3DCAL_01",VR::LO,VM::VM1,"3Dcal algorithm version",false },
  {0x0021,0x000e,"GEMS_XR3DCAL_01",VR::FL,VM::VM1,"3Dcal 3D frame unit size",false },
  {0x0021,0x000f,"GEMS_XR3DCAL_01",VR::LO,VM::VM1,"3Dcal calibration mode",false },
  {0x0021,0x0010,"GEMS_XR3DCAL_01",VR::FL,VM::VM1,"3Dcal image frame origin row",false },
  {0x0021,0x0011,"GEMS_XR3DCAL_01",VR::FL,VM::VM1,"3Dcal image frame origin column",false },
  {0x0021,0x0012,"GEMS_XR3DCAL_01",VR::IS,VM::VM1,"3Dcal positioner pivot rotation speed",false },
  {0x0021,0x0013,"GEMS_XR3DCAL_01",VR::IS,VM::VM1,"3Dcal number of images",false },
  {0x0021,0x0014,"GEMS_XR3DCAL_01",VR::UI,VM::VM1,"3Dcal Instance UID",false },
  {0x0021,0x0015,"GEMS_XR3DCAL_01",VR::FL,VM::VM2,"3Dcal image pixel spacing",false },
  {0x0021,0x0016,"GEMS_XR3DCAL_01",VR::CS,VM::VM1,"3Dcal centering mode",false },
  {0x0033,0x0005,"GEMS_YMHD_01",VR::UN,VM::VM1,"?",false },
  {0x0033,0x0006,"GEMS_YMHD_01",VR::UN,VM::VM1,"?",false },
  {0x0019,0x0039,"GE_GENESIS_REV3.0",VR::SS,VM::VM1,"Axial Type",false },
  {0x0019,0x008f,"GE_GENESIS_REV3.0",VR::SS,VM::VM1,"Swap Phase / Frequency Axis",false },
  {0x0019,0x009c,"GE_GENESIS_REV3.0",VR::SS,VM::VM1,"Pulse Sequence Name",false },
  {0x0019,0x009f,"GE_GENESIS_REV3.0",VR::SS,VM::VM1,"Coil Type",false },
  {0x0019,0x00a4,"GE_GENESIS_REV3.0",VR::SS,VM::VM1,"SAT fat/water/none",false },
  {0x0019,0x00c0,"GE_GENESIS_REV3.0",VR::SS,VM::VM1,"Bitmap of SAT Selections",false },
  {0x0019,0x00c1,"GE_GENESIS_REV3.0",VR::SS,VM::VM1,"Surfacel Coil Iintensity Correction Flag",false },
  {0x0019,0x00cb,"GE_GENESIS_REV3.0",VR::SS,VM::VM1,"Phase Contrast Flow Axis",false },
  {0x0019,0x00cc,"GE_GENESIS_REV3.0",VR::SS,VM::VM1,"Phase Contrast Velocity Encoding",false },
  {0x0019,0x00d5,"GE_GENESIS_REV3.0",VR::SS,VM::VM1,"Fractional Echo",false },
  {0x0019,0x00d8,"GE_GENESIS_REV3.0",VR::SS,VM::VM1,"Variable Echo Flag",false },
  {0x0019,0x00d9,"GE_GENESIS_REV3.0",VR::DS,VM::VM1,"Concatenated Sat",false },
  {0x0019,0x00f2,"GE_GENESIS_REV3.0",VR::SS,VM::VM1,"Number of Phases",false },
  {0x0043,0x001e,"GE_GENESIS_REV3.0",VR::DS,VM::VM1,"Delta Start Time",false },
  {0x0043,0x0027,"GE_GENESIS_REV3.0",VR::SH,VM::VM1,"Pitch Ratio",false },
  {0x6005,0x0010,"GE_GROUP",VR::UT,VM::VM1,"?",false },
  {0x0009,0x0000,"HMC - CT - ID",VR::UN,VM::VM1,"Image ID Information (e.g. Patient Name, Patient ID)",false },
  {0x0009,0x0001,"HMC - CT - ID",VR::UN,VM::VM1,"Image ID Information (e.g. Patient Comment)",false },
  {0x0011,0x0000,"HOLOGIC",VR::LO,VM::VM1,"Hx Questionnaire",false },
  {0x0021,0x0001,"HOLOGIC",VR::LT,VM::VM1,"Image Analysis Data in XML",false },
  {0x0023,0x0000,"HOLOGIC",VR::LO,VM::VM1,"Encoding Scheme Version",false },
  {0x0023,0x0001,"HOLOGIC",VR::LO,VM::VM1,"P File Name",false },
  {0x0023,0x0002,"HOLOGIC",VR::OB,VM::VM1,"P File Data",false },
  {0x0023,0x0003,"HOLOGIC",VR::UL,VM::VM1,"P File Length",false },
  {0x0023,0x0004,"HOLOGIC",VR::OB,VM::VM1,"R File Data",false },
  {0x0023,0x0005,"HOLOGIC",VR::UL,VM::VM1,"R File Length",false },
  {0x0019,0x0006,"HOLOGIC, Inc.",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0007,"HOLOGIC, Inc.",VR::SH,VM::VM1,"?",false },
  {0x0019,0x0008,"HOLOGIC, Inc.",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0016,"HOLOGIC, Inc.",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0025,"HOLOGIC, Inc.",VR::SH,VM::VM1,"?",false },
  {0x0019,0x0026,"HOLOGIC, Inc.",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0027,"HOLOGIC, Inc.",VR::SH,VM::VM1,"?",false },
  {0x0019,0x0028,"HOLOGIC, Inc.",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0029,"HOLOGIC, Inc.",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0030,"HOLOGIC, Inc.",VR::US,VM::VM1,"?",false },
  {0x0019,0x0031,"HOLOGIC, Inc.",VR::US,VM::VM1,"?",false },
  {0x0019,0x0032,"HOLOGIC, Inc.",VR::US,VM::VM1,"?",false },
  {0x0019,0x0033,"HOLOGIC, Inc.",VR::US,VM::VM1,"?",false },
  {0x0019,0x0034,"HOLOGIC, Inc.",VR::US,VM::VM1,"?",false },
  {0x0019,0x0035,"HOLOGIC, Inc.",VR::US,VM::VM1,"?",false },
  {0x0019,0x0037,"HOLOGIC, Inc.",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0040,"HOLOGIC, Inc.",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0041,"HOLOGIC, Inc.",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0042,"HOLOGIC, Inc.",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0043,"HOLOGIC, Inc.",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0044,"HOLOGIC, Inc.",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0045,"HOLOGIC, Inc.",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0046,"HOLOGIC, Inc.",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0060,"HOLOGIC, Inc.",VR::LT,VM::VM1,"?",false },
  {0x0019,0x0061,"HOLOGIC, Inc.",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0062,"HOLOGIC, Inc.",VR::LT,VM::VM1,"?",false },
  {0x0019,0x0070,"HOLOGIC, Inc.",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0085,"HOLOGIC, Inc.",VR::SH,VM::VM1,"?",false },
  {0x0019,0x0087,"HOLOGIC, Inc.",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0088,"HOLOGIC, Inc.",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0089,"HOLOGIC, Inc.",VR::DS,VM::VM2,"?",false },
  {0x0019,0x008a,"HOLOGIC, Inc.",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x0090,"HOLOGIC, Inc.",VR::DS,VM::VM1,"?",false },
  {0x7e01,0x0001,"HOLOGIC, Inc.",VR::LO,VM::VM1,"?",false },
  {0x7e01,0x0002,"HOLOGIC, Inc.",VR::LO,VM::VM1,"?",false },
  {0x7e01,0x0010,"HOLOGIC, Inc.",VR::SQ,VM::VM1,"?",false },
  {0x7e01,0x0011,"HOLOGIC, Inc.",VR::SQ,VM::VM1,"?",false },
  {0x7e01,0x0012,"HOLOGIC, Inc.",VR::OB,VM::VM1,"?",false },
  {0x7f01,0x0001,"HOLOGIC, Inc.",VR::LO,VM::VM1,"?",false },
  {0x7f01,0x0002,"HOLOGIC, Inc.",VR::SH,VM::VM1,"?",false },
  {0x7f01,0x0010,"HOLOGIC, Inc.",VR::SQ,VM::VM1,"?",false },
  {0x7f01,0x0012,"HOLOGIC, Inc.",VR::OB,VM::VM1,"?",false },
/*  {0x0011,0x0000,"Hologic",VR::LO,VM::VM1,"Hx Questionnaire",false },
  {0x0021,0x0001,"Hologic",VR::LT,VM::VM1,"Image Analysis Data in XML",false },
  {0x0023,0x0000,"Hologic",VR::LO,VM::VM1,"Encoding Scheme Version",false },
  {0x0023,0x0001,"Hologic",VR::LO,VM::VM1,"P File Name",false },
  {0x0023,0x0002,"Hologic",VR::OB,VM::VM1,"P File Data",false },
  {0x0023,0x0003,"Hologic",VR::UL,VM::VM1,"P File Length",false },
  {0x0023,0x0004,"Hologic",VR::OB,VM::VM1,"R File Data",false },
  {0x0023,0x0005,"Hologic",VR::UL,VM::VM1,"R File Length",false }, */
  {0x0011,0x0000,"IDEXX",VR::LO,VM::VM1,"Breed Name",false },
  {0x0011,0x0001,"IDEXX",VR::LO,VM::VM1,"Species Name",false },
  {0x0011,0x0002,"IDEXX",VR::PN,VM::VM1,"Owner",false },
  {0x300b,0x0002,"IMPAC",VR::FL,VM::VM1,"?",false },
  {0x300b,0x0004,"IMPAC",VR::FL,VM::VM1,"?",false },
  {0x300b,0x000e,"IMPAC",VR::FL,VM::VM1,"?",false },
  {0x1269,0x0001,"IMS s.r.l. Biopsy Private Code",VR::IS,VM::VM1,"Biopsy Image",false },
  {0x1269,0x0010,"IMS s.r.l. Biopsy Private Code",VR::IS,VM::VM1_n,"Biopsy Markers X",false },
  {0x1269,0x0011,"IMS s.r.l. Biopsy Private Code",VR::IS,VM::VM1_n,"Biopsy Markers Y",false },
  {0x1269,0x0012,"IMS s.r.l. Biopsy Private Code",VR::IS,VM::VM1,"Biopsy Markers Number",false },
  {0x1269,0x0020,"IMS s.r.l. Biopsy Private Code",VR::IS,VM::VM1,"Biopsy Area Left Border",false },
  {0x1269,0x0021,"IMS s.r.l. Biopsy Private Code",VR::IS,VM::VM1,"Biopsy Area Right Border",false },
  {0x1269,0x0022,"IMS s.r.l. Biopsy Private Code",VR::IS,VM::VM1,"Biopsy Area Top Border",false },
  {0x1269,0x0023,"IMS s.r.l. Biopsy Private Code",VR::IS,VM::VM1,"Biopsy Area Bottom Border",false },
  {0x1271,0x0001,"IMS s.r.l. Mammography Private Code",VR::IS,VM::VM1,"?",false },
  {0x1271,0x0002,"IMS s.r.l. Mammography Private Code",VR::IS,VM::VM1,"?",false },
  {0x1271,0x0010,"IMS s.r.l. Mammography Private Code",VR::IS,VM::VM1,"?",false },
  {0x1271,0x0011,"IMS s.r.l. Mammography Private Code",VR::IS,VM::VM1,"?",false },
  {0x1271,0x0012,"IMS s.r.l. Mammography Private Code",VR::IS,VM::VM1,"?",false },
  {0x1271,0x0013,"IMS s.r.l. Mammography Private Code",VR::IS,VM::VM1,"?",false },
  {0x1271,0x0020,"IMS s.r.l. Mammography Private Code",VR::IS,VM::VM1,"?",false },
  {0x1271,0x0021,"IMS s.r.l. Mammography Private Code",VR::IS,VM::VM1,"?",false },
  {0x1271,0x0022,"IMS s.r.l. Mammography Private Code",VR::IS,VM::VM1,"?",false },
  {0x1271,0x0038,"IMS s.r.l. Mammography Private Code",VR::LO,VM::VM1,"?",false },
  {0x1271,0x0039,"IMS s.r.l. Mammography Private Code",VR::US,VM::VM1,"?",false },
  {0x1271,0x0040,"IMS s.r.l. Mammography Private Code",VR::US,VM::VM1,"?",false },
  {0x1271,0x0041,"IMS s.r.l. Mammography Private Code",VR::LO,VM::VM1,"?",false },
  {0x1271,0x0042,"IMS s.r.l. Mammography Private Code",VR::LT,VM::VM1,"?",false },
  {0x1271,0x0045,"IMS s.r.l. Mammography Private Code",VR::IS,VM::VM1,"?",false },
  {0x1271,0x0046,"IMS s.r.l. Mammography Private Code",VR::IS,VM::VM1,"?",false },
  {0x1271,0x0047,"IMS s.r.l. Mammography Private Code",VR::IS,VM::VM1,"?",false },
  {0x1271,0x0048,"IMS s.r.l. Mammography Private Code",VR::IS,VM::VM1,"?",false },
  {0x1271,0x0049,"IMS s.r.l. Mammography Private Code",VR::IS,VM::VM1,"?",false },
  {0x1271,0x0050,"IMS s.r.l. Mammography Private Code",VR::IS,VM::VM1,"?",false },
  {0x1271,0x0051,"IMS s.r.l. Mammography Private Code",VR::IS,VM::VM1,"?",false },
  {0x1271,0x0052,"IMS s.r.l. Mammography Private Code",VR::IS,VM::VM1,"?",false },
  {0x1271,0x0053,"IMS s.r.l. Mammography Private Code",VR::LO,VM::VM1,"?",false },
  {0x1271,0x0054,"IMS s.r.l. Mammography Private Code",VR::IS,VM::VM1,"?",false },
  {0x0015,0x0010,"INFINITT_FMX",VR::LO,VM::VM1,"?",false },
  {0x0015,0x0011,"INFINITT_FMX",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0008,"INTEGRIS 1.0",VR::CS,VM::VM1,"Exposure Channel",false },
  {0x0009,0x0032,"INTEGRIS 1.0",VR::TM,VM::VM1,"Exposure Start Time",false },
  {0x0019,0x0000,"INTEGRIS 1.0",VR::LO,VM::VM1,"APR Name",false },
  {0x0019,0x0040,"INTEGRIS 1.0",VR::DS,VM::VM1,"Frame Rate",false },
  {0x0021,0x0012,"INTEGRIS 1.0",VR::IS,VM::VM1,"Exposure Number",false },
  {0x0029,0x0000,"INTEGRIS 1.0",VR::SQ,VM::VM1,"Edge Enhancement Sequence",false },
  {0x0029,0x0001,"INTEGRIS 1.0",VR::US,VM::VM2,"Convolution Kernel Size",false },
  {0x0029,0x0002,"INTEGRIS 1.0",VR::US,VM::VM1_n,"Convolution Kernel Coefficients",false },
  {0x0029,0x0003,"INTEGRIS 1.0",VR::FL,VM::VM1,"Edge Enhancement Gain",false },
  {0x0029,0x0008,"INTEGRIS 1.0",VR::IS,VM::VM1,"Number of Exposure Results",false },
  {0x0041,0x0020,"INTEGRIS 1.0",VR::DS,VM::VM1,"Accumulated Fluoroscopy Dose",false },
  {0x0041,0x0030,"INTEGRIS 1.0",VR::DS,VM::VM1,"Accumulated Exposure Dose",false },
  {0x0041,0x0040,"INTEGRIS 1.0",VR::DS,VM::VM1,"Total Dose",false },
  {0x0041,0x0041,"INTEGRIS 1.0",VR::US,VM::VM1,"Total Number of Frames",false },
  {0x0041,0x0050,"INTEGRIS 1.0",VR::SQ,VM::VM1,"Exposure Information Sequence",false },
  {0x0029,0x0001,"INTELERAD MEDICAL SYSTEMS",VR::FD,VM::VM1,"Image Compression Fraction",false },
  {0x0029,0x0002,"INTELERAD MEDICAL SYSTEMS",VR::FD,VM::VM1,"Image Quality",false },
  {0x0029,0x0003,"INTELERAD MEDICAL SYSTEMS",VR::FD,VM::VM1,"Image Bytes Transferred",false },
  {0x0029,0x0008,"INTELERAD MEDICAL SYSTEMS",VR::SL,VM::VM4,"Image Bounding Box",false },
  {0x0029,0x0009,"INTELERAD MEDICAL SYSTEMS",VR::LO,VM::VM1,"Image Bounding Box Algorithm ID",false },
  {0x0029,0x0010,"INTELERAD MEDICAL SYSTEMS",VR::SH,VM::VM1,"J2C Parameter Type",false },
  {0x0029,0x0011,"INTELERAD MEDICAL SYSTEMS",VR::CS,VM::VM1,"J2C Pixel Representation",false },
  {0x0029,0x0012,"INTELERAD MEDICAL SYSTEMS",VR::US,VM::VM1,"J2C Bits Allocated",false },
  {0x0029,0x0013,"INTELERAD MEDICAL SYSTEMS",VR::US,VM::VM1,"J2C Pixel Shift Value",false },
  {0x0029,0x0014,"INTELERAD MEDICAL SYSTEMS",VR::US,VM::VM1,"J2C Planar Configuration",false },
  {0x0029,0x0015,"INTELERAD MEDICAL SYSTEMS",VR::DS,VM::VM1,"J2C Rescale Intercept",false },
  {0x0029,0x0016,"INTELERAD MEDICAL SYSTEMS",VR::US,VM::VM1,"J2c Smallest Pixel Value",false },
  {0x0029,0x0017,"INTELERAD MEDICAL SYSTEMS",VR::US,VM::VM1,"J2c Largest Pixel Value",false },
  {0x0029,0x0020,"INTELERAD MEDICAL SYSTEMS",VR::LO,VM::VM1,"Pixel Data MD5 Sum Values",false },
  {0x0029,0x0021,"INTELERAD MEDICAL SYSTEMS",VR::US,VM::VM1_n,"Pixel Data Histogram Percentiles",false },
  {0x0029,0x0022,"INTELERAD MEDICAL SYSTEMS",VR::FD,VM::VM1_n,"Pixel Data Histogram Percentile Values",false },
  {0x0029,0x0023,"INTELERAD MEDICAL SYSTEMS",VR::FD,VM::VM1,"VOI LUT Window Level Values",false },
  {0x0029,0x0024,"INTELERAD MEDICAL SYSTEMS",VR::FD,VM::VM1,"VOI LUT Real Pixel Value Range",false },
  {0x0029,0x0025,"INTELERAD MEDICAL SYSTEMS",VR::IS,VM::VM1,"VOI LUT Index",false },
  {0x0029,0x0026,"INTELERAD MEDICAL SYSTEMS",VR::DS,VM::VM1,"J2C Rescale Slope",false },
  {0x0029,0x0027,"INTELERAD MEDICAL SYSTEMS",VR::FD,VM::VM1,"Image Rotation Angle",true},
  {0x0029,0x0028,"INTELERAD MEDICAL SYSTEMS",VR::FD,VM::VM1,"Displayed Area Top Left-Hand Corner",true},
  {0x0029,0x0029,"INTELERAD MEDICAL SYSTEMS",VR::FD,VM::VM1,"Displayed Area Bottom Right-Hand Corner",true},
  {0x3f01,0x0001,"INTELERAD MEDICAL SYSTEMS",VR::LO,VM::VM1,"Institution",false },
  {0x3f01,0x0002,"INTELERAD MEDICAL SYSTEMS",VR::SH,VM::VM1,"AE Title Destination",false },
  {0x3f01,0x0003,"INTELERAD MEDICAL SYSTEMS",VR::LO,VM::VM1,"Source AE Title",false },
  {0x3f01,0x0004,"INTELERAD MEDICAL SYSTEMS",VR::SH,VM::VM1,"Deferred Validation",false },
  {0x3f01,0x0005,"INTELERAD MEDICAL SYSTEMS",VR::LO,VM::VM1,"Series Owner",false },
  {0x3f01,0x0006,"INTELERAD MEDICAL SYSTEMS",VR::LO,VM::VM1,"Group Number",false },
  {0x3f01,0x0007,"INTELERAD MEDICAL SYSTEMS",VR::SH,VM::VM1,"Stripped Pixel Data",false },
  {0x3f01,0x0008,"INTELERAD MEDICAL SYSTEMS",VR::SH,VM::VM1,"Pending Move Request",false },
  {0x3f01,0x0009,"INTELERAD MEDICAL SYSTEMS",VR::LO,VM::VM1,"Dicom Premap Sites",false },
  {0x3f01,0x000a,"INTELERAD MEDICAL SYSTEMS",VR::DA,VM::VM1,"Association Date",false },
  {0x3f01,0x000b,"INTELERAD MEDICAL SYSTEMS",VR::TM,VM::VM1,"Association Time",false },
  {0x3f01,0x000c,"INTELERAD MEDICAL SYSTEMS",VR::SH,VM::VM1,"Use Source AE",false },
  {0x3f01,0x000e,"INTELERAD MEDICAL SYSTEMS",VR::SH,VM::VM1,"Relay to AE",false },
  {0x3f01,0x000f,"INTELERAD MEDICAL SYSTEMS",VR::SH,VM::VM1,"ImsMoveRequestRedirectionCount",false },
  {0x3f01,0x0010,"INTELERAD MEDICAL SYSTEMS",VR::LO,VM::VM1,"ImsPostFix to AccessionNumber",false },
  {0x3f01,0x0011,"INTELERAD MEDICAL SYSTEMS",VR::US,VM::VM1,"Count of tags present when CTN wrote the file",false },
  {0x3f01,0x0012,"INTELERAD MEDICAL SYSTEMS",VR::UL,VM::VM1,"Association Epoch In Seconds",false },
  {0x3f01,0x0013,"INTELERAD MEDICAL SYSTEMS",VR::LO,VM::VM1,"?",false },
  {0x3f01,0x0014,"INTELERAD MEDICAL SYSTEMS",VR::LO,VM::VM1,"?",false },
  {0x3f01,0x001d,"INTELERAD MEDICAL SYSTEMS",VR::LO,VM::VM1,"?",false },
  {0x3f03,0x0001,"INTELERAD MEDICAL SYSTEMS",VR::SQ,VM::VM1,"Attribute History",false },
  {0x3f03,0x0002,"INTELERAD MEDICAL SYSTEMS",VR::DT,VM::VM1,"Change Timestamp",false },
  {0x3f03,0x0003,"INTELERAD MEDICAL SYSTEMS",VR::LO,VM::VM1,"Change initiator",false },
  {0x3f03,0x0004,"INTELERAD MEDICAL SYSTEMS",VR::ST,VM::VM1,"Change explanation",false },
  {0x6001,0x0001,"INTELERAD MEDICAL SYSTEMS",VR::US,VM::VM1,"Overlay bit position",false },
  {0x0071,0x0001,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::LO,VM::VM1,"InteleViewer Markup Type",false },
  {0x0071,0x0002,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::ST,VM::VM1,"InteleViewer Markup Text",false },
  {0x0071,0x0003,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::FD,VM::VM1,"InteleViewer Markup Text Offset",false },
  {0x0071,0x0004,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::FD,VM::VM1,"InteleViewer Markup Position",false },
  {0x0071,0x0005,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::FD,VM::VM1,"InteleViewer Markup Half-Vector",false },
  {0x0071,0x0006,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::FD,VM::VM1,"InteleViewer Markup Vector 1",false },
  {0x0071,0x0007,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::FD,VM::VM1,"InteleViewer Markup Vector 2",false },
  {0x0071,0x0008,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::FD,VM::VM1,"InteleViewer Markup Orthogonal Position",false },
  {0x0071,0x0009,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::FD,VM::VM1,"InteleViewer Markup Orthogonal Half-Vector",false },
  {0x0071,0x000a,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::FD,VM::VM1,"InteleViewer Markup Minimum Length",false },
  {0x0071,0x000b,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::SL,VM::VM1,"InteleViewer Markup Number of Polygons",false },
  {0x0071,0x000c,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::SL,VM::VM1,"InteleViewer Markup Length of Polygons",false },
  {0x0071,0x000d,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::FD,VM::VM1,"InteleViewer Markup Points",false },
  {0x0071,0x000e,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::FD,VM::VM1,"InteleViewer Markup Slice Separation",false },
  {0x0071,0x000f,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::US,VM::VM1,"InteleViewer Markup VOI Index",false },
  {0x0071,0x0010,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::FD,VM::VM1,"InteleViewer Markup Sagittal Text Offset",false },
  {0x0071,0x0011,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::FD,VM::VM1,"InteleViewer Markup Default Text Offset",false },
  {0x0071,0x0020,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::LO,VM::VM1,"InteleViewer Markup Dataset Tool ID",false },
  {0x0071,0x0021,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::FD,VM::VM1,"InteleViewer Markup Distance Calibration",false },
  {0x0071,0x0030,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::FD,VM::VM1,"InteleViewer Markup Image Grid Origin",false },
  {0x0071,0x0031,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::FD,VM::VM1,"InteleViewer Markup Image Grid Pixel Row Vector",false },
  {0x0071,0x0032,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::FD,VM::VM1,"InteleViewer Markup Image Grid Pixel Column Vector",false },
  {0x0071,0x0033,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::CS,VM::VM1,"InteleViewer Markup Image Grid Is Calibrated",false },
  {0x0071,0x0034,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::FD,VM::VM1,"Image Rotation Angle",false },
  {0x0071,0x0035,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::FD,VM::VM1,"Displayed Area Top Left-Hand Corner",false },
  {0x0071,0x0036,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::FD,VM::VM1,"Displayed Area Bottom Right-Hand Corner",false },
  {0x3f05,0x0001,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::LO,VM::VM1,"InteleViewer Source Name",false },
  {0x3f05,0x0002,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::LO,VM::VM1,"Preview Image Id",false },
  {0x3f0f,0x0001,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::US,VM::VM1,"InteleViewer Advanced Visualization Presentation Bin Size",false },
  {0x3f0f,0x0002,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::SQ,VM::VM1,"InteleViewer Advanced Visualization Presentation Bin",false },
  {0x3f0f,0x0003,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::LO,VM::VM1,"InteleViewer Advanced Visualization RenderType",false },
  {0x3f0f,0x0004,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::FD,VM::VM1,"InteleViewer Advanced Visualization Row Vector",false },
  {0x3f0f,0x0005,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::FD,VM::VM1,"InteleViewer Advanced Visualization Column Vector",false },
  {0x3f0f,0x0006,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::FD,VM::VM1,"InteleViewer Advanced Visualization Slab Thickness",false },
  {0x3f0f,0x0007,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::LO,VM::VM1,"InteleViewer Advanced Visualization Original Dataset Series UID",false },
  {0x3f0f,0x0008,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::LO,VM::VM1,"InteleViewer Advanced Visualization Transfer Function Type",false },
  {0x3f0f,0x0009,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::LO,VM::VM1,"InteleViewer Advanced Visualization Source Dataset Sub Volume First SOP UID",false },
  {0x3f0f,0x0010,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::LO,VM::VM1,"InteleViewer Advanced Visualization Source Dataset Sub Volume Last SOP UID",false },
  {0x3f0f,0x0011,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::LO,VM::VM1,"InteleViewer Advanced Visualization Sub volume creation algorithm",false },
  {0x3f0f,0x0012,"INTELERAD MEDICAL SYSTEMS INTELEVIEWER",VR::US,VM::VM1,"InteleViewer Advanced Visualization Number of images in Sub volume",false },
  {0x0065,0x0011,"ISG Image",VR::UL,VM::VM1,"Length in Byte of Image",false },
  {0x0029,0x0070,"ISG shadow",VR::IS,VM::VM1,"?",false },
  {0x0029,0x0080,"ISG shadow",VR::IS,VM::VM1,"?",false },
  {0x0029,0x0090,"ISG shadow",VR::IS,VM::VM1,"?",false },
  {0x0009,0x0001,"ISI",VR::SQ,VM::VM1,"SIENET General Purpose IMGF",false },
  {0x8003,0x0000,"Image (ID, Version, Size, Dump, GUID)",VR::LO,VM::VM1,"ID",false },
  {0x8003,0x0010,"Image (ID, Version, Size, Dump, GUID)",VR::LO,VM::VM1,"Version",false },
  {0x8003,0x0020,"Image (ID, Version, Size, Dump, GUID)",VR::UL,VM::VM1,"Size",false },
  {0x8003,0x0030,"Image (ID, Version, Size, Dump, GUID)",VR::OB,VM::VM1,"Dump",false },
  {0x8003,0x0040,"Image (ID, Version, Size, Dump, GUID)",VR::LO,VM::VM1,"GUID",false },
  {0x4321,0x0005,"Imaging Dynamics Company Ltd.",VR::UN,VM::VM1,"Custom Tag",false },
  {0x4321,0x0010,"Imaging Dynamics Company Ltd.",VR::UN,VM::VM1,"Custom Tag",false },
  {0x4321,0x0041,"Imaging Dynamics Company Ltd.",VR::CS,VM::VM1,"?",false },
  {0x4321,0x0042,"Imaging Dynamics Company Ltd.",VR::US,VM::VM1,"?",false },
  {0x4321,0x0050,"Imaging Dynamics Company Ltd.",VR::DS,VM::VM1,"Custom Tag",false },
  {0x4321,0x0051,"Imaging Dynamics Company Ltd.",VR::DS,VM::VM1,"Custom Tag",false },
  {0x4321,0x0064,"Imaging Dynamics Company Ltd.",VR::LO,VM::VM1,"Custom Tag",false },
  {0x0021,0x00a5,"KINETDX",VR::US,VM::VM1,"?",false },
  {0x0021,0x00a8,"KINETDX",VR::LO,VM::VM1,"?",false },
  {0x0021,0x00aa,"KINETDX",VR::OB,VM::VM1,"?",false },
  {0x0021,0x00a4,"KINETDX_GRAPHICS",VR::LT,VM::VM1,"?",false },
  {0x7fe1,0x0001,"KRETZ_US",VR::OB,VM::VM1,"?",false },
  {0x0029,0x0015,"Kodak Image Information",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0016,"Kodak Image Information",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0017,"Kodak Image Information",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0018,"Kodak Image Information",VR::UT,VM::VM1,"?",false },
  {0x0029,0x0019,"Kodak Image Information",VR::IS,VM::VM1,"?",false },
  {0x0029,0x001a,"Kodak Image Information",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0001,"LODOX_STATSCAN",VR::IS,VM::VM1_n,"?",false },
  {0x0019,0x0002,"LODOX_STATSCAN",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0003,"LODOX_STATSCAN",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0004,"LODOX_STATSCAN",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0005,"LODOX_STATSCAN",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0006,"LODOX_STATSCAN",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0007,"LODOX_STATSCAN",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0008,"LODOX_STATSCAN",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0006,"LORAD Selenia",VR::OB,VM::VM1,"Paddle ID",false },
  {0x0019,0x0007,"LORAD Selenia",VR::OB,VM::VM1,"Paddle Position",false },
  {0x0019,0x0008,"LORAD Selenia",VR::OB,VM::VM1,"Collimation Size",false },
  {0x0019,0x0026,"LORAD Selenia",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0027,"LORAD Selenia",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0028,"LORAD Selenia",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0029,"LORAD Selenia",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0030,"LORAD Selenia",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0031,"LORAD Selenia",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0032,"LORAD Selenia",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0033,"LORAD Selenia",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0034,"LORAD Selenia",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0035,"LORAD Selenia",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0040,"LORAD Selenia",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0041,"LORAD Selenia",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0050,"LORAD Selenia",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0051,"LORAD Selenia",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0052,"LORAD Selenia",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0053,"LORAD Selenia",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0070,"LORAD Selenia",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0071,"LORAD Selenia",VR::OB,VM::VM1,"?",false },
  {0x0037,0x0001,"MAROTECH Inc.",VR::LO,VM::VM1,"Some kind of version number",false },
  {0x0037,0x0021,"MAROTECH Inc.",VR::US,VM::VM1,"Icon Row",false },
  {0x0037,0x0022,"MAROTECH Inc.",VR::US,VM::VM1,"Icon Col",false },
  {0x0037,0x0023,"MAROTECH Inc.",VR::OB,VM::VM1,"JPEG Compressed Icon",false },
  {0x3005,0x0000,"MDS NORDION OTP ANATOMY MODELLING",VR::SQ,VM::VM1,"?",false },
  {0x3005,0x0002,"MDS NORDION OTP ANATOMY MODELLING",VR::CS,VM::VM1,"?",false },
  {0x3005,0x0004,"MDS NORDION OTP ANATOMY MODELLING",VR::DS,VM::VM1,"?",false },
  {0x3005,0x0006,"MDS NORDION OTP ANATOMY MODELLING",VR::DS,VM::VM1,"?",false },
  {0x3005,0x0008,"MDS NORDION OTP ANATOMY MODELLING",VR::DS,VM::VM1,"?",false },
  {0x3005,0x000a,"MDS NORDION OTP ANATOMY MODELLING",VR::CS,VM::VM1,"?",false },
  {0x3005,0x000c,"MDS NORDION OTP ANATOMY MODELLING",VR::CS,VM::VM1,"?",false },
  {0x3005,0x000e,"MDS NORDION OTP ANATOMY MODELLING",VR::CS,VM::VM1,"?",false },
  {0x3005,0x0010,"MDS NORDION OTP ANATOMY MODELLING",VR::DS,VM::VM1,"?",false },
  {0x3005,0x0012,"MDS NORDION OTP ANATOMY MODELLING",VR::DS,VM::VM1,"?",false },
  {0x0021,0x0001,"MEDIFACE",VR::SL,VM::VM1,"?",false },
  {0x0021,0x0010,"MEDIFACE",VR::SH,VM::VM1,"?",false },
  {0x0021,0x0011,"MEDIFACE",VR::LO,VM::VM1,"?",false },
  {0x0021,0x0012,"MEDIFACE",VR::UL,VM::VM1,"?",false },
  {0x0021,0x0021,"MEDIFACE",VR::LO,VM::VM1,"?",false },
  {0x0021,0x0022,"MEDIFACE",VR::LO,VM::VM1,"?",false },
  {0x0021,0x0030,"MEDIFACE",VR::DA,VM::VM1,"?",false },
  {0x0021,0x0031,"MEDIFACE",VR::IS,VM::VM1,"?",false },
  {0x0021,0x0040,"MEDIFACE",VR::SQ,VM::VM1,"?",false },
  {0x0021,0x0041,"MEDIFACE",VR::SH,VM::VM1,"?",false },
  {0x0021,0x0042,"MEDIFACE",VR::LO,VM::VM1,"?",false },
  {0x0021,0x0050,"MEDIFACE",VR::SQ,VM::VM1,"?",false },
  {0x0021,0x0051,"MEDIFACE",VR::SH,VM::VM1,"?",false },
  {0x0021,0x0052,"MEDIFACE",VR::LO,VM::VM1,"?",false },
  {0x0021,0x0060,"MEDIFACE",VR::SQ,VM::VM1,"?",false },
  {0x0021,0x00f1,"MEDIFACE",VR::LO,VM::VM1,"?",false },
  {0x0021,0x00f2,"MEDIFACE",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0001,"MEDIFACE",VR::UL,VM::VM1,"?",false },
  {0x0029,0x0010,"MEDIFACE",VR::DS,VM::VM1,"Window Center",false },
  {0x0029,0x0011,"MEDIFACE",VR::DS,VM::VM1,"Window Width",false },
  {0x0029,0x0020,"MEDIFACE",VR::DS,VM::VM1,"?",false },
  {0x0029,0x0021,"MEDIFACE",VR::UL,VM::VM1,"?",false },
  {0x0029,0x0022,"MEDIFACE",VR::DS,VM::VM2,"MEDIFACE Pixel Spacing",false },
  {0x0029,0x0030,"MEDIFACE",VR::LT,VM::VM1,"?",false },
  {0x0003,0x0000,"MEDIS_IS&C 1.0",VR::US,VM::VM1,"?",false },
  {0x0003,0x0003,"MEDIS_IS&C 1.0",VR::US,VM::VM1,"?",false },
  {0x0003,0x0004,"MEDIS_IS&C 1.0",VR::CS,VM::VM1,"?",false },
  {0x0003,0x0010,"MEDIS_IS&C 1.0",VR::DS,VM::VM1_n,"?",false },
  {0x0003,0x0020,"MEDIS_IS&C 1.0",VR::UL,VM::VM1,"?",false },
  {0x0003,0x0022,"MEDIS_IS&C 1.0",VR::UL,VM::VM1,"?",false },
  {0x0003,0x0030,"MEDIS_IS&C 1.0",VR::US,VM::VM1,"?",false },
  {0x0009,0x0000,"MERGE TECHNOLOGIES, INC.",VR::OB,VM::VM1,"?",false },
  {0x0031,0x0020,"MITRA LINKED ATTRIBUTES 1.0",VR::LO,VM::VM1,"Global Patient ID",false },
/*  {0x0029,0x0000,"MITRA MARKUP 1.0",VR::LO,VM::VM1_n,"?",false },
  {0x0029,0x0001,"MITRA MARKUP 1.0",VR::LO,VM::VM1_n,"?",false },*/
  {0x0033,0x0002,"MITRA OBJECT ATTRIBUTES 1.0",VR::LO,VM::VM1,"?",false },
  {0x0033,0x0004,"MITRA OBJECT ATTRIBUTES 1.0",VR::LO,VM::VM1,"?",false },
  {0x0033,0x0006,"MITRA OBJECT ATTRIBUTES 1.0",VR::LO,VM::VM1,"?",false },
  {0x0033,0x0008,"MITRA OBJECT ATTRIBUTES 1.0",VR::LO,VM::VM1,"?",false },
  {0x0033,0x000a,"MITRA OBJECT ATTRIBUTES 1.0",VR::LO,VM::VM1,"?",false },
//  {0x0029,0x0000,"MITRA OBJECT DOCUMENT 1.0",VR::OB,VM::VM1,"?",false },
  {0x0033,0x0002,"MITRA OBJECT UTF8 ATTRIBUTES 1.0",VR::UN,VM::VM1,"Patient's Name",false },
  {0x0033,0x0004,"MITRA OBJECT UTF8 ATTRIBUTES 1.0",VR::UN,VM::VM1,"Study Description",false },
  {0x0033,0x0006,"MITRA OBJECT UTF8 ATTRIBUTES 1.0",VR::UN,VM::VM1,"Referring Physician's Name",false },
  {0x0033,0x0008,"MITRA OBJECT UTF8 ATTRIBUTES 1.0",VR::UN,VM::VM1,"Requesting Physician's Name",false },
  {0x0033,0x000a,"MITRA OBJECT UTF8 ATTRIBUTES 1.0",VR::UN,VM::VM1,"Performing Physician's Name",false },
  {0x0033,0x000c,"MITRA OBJECT UTF8 ATTRIBUTES 1.0",VR::UN,VM::VM1,"Reason for Study",false },
  {0x0033,0x000e,"MITRA OBJECT UTF8 ATTRIBUTES 1.0",VR::UN,VM::VM1,"Study Comments",false },
  {0x0033,0x0013,"MITRA OBJECT UTF8 ATTRIBUTES 1.0",VR::UN,VM::VM1,"Patient's Name Single Byte",false },
  {0x0033,0x0014,"MITRA OBJECT UTF8 ATTRIBUTES 1.0",VR::UN,VM::VM1,"Patient's Name Ideographic",false },
  {0x0033,0x0015,"MITRA OBJECT UTF8 ATTRIBUTES 1.0",VR::UN,VM::VM1,"Patient's Name Phonetic",false },
  {0x0033,0x0016,"MITRA OBJECT UTF8 ATTRIBUTES 1.0",VR::UN,VM::VM1,"Referring Physician's Name Single Byte",false },
  {0x0033,0x0017,"MITRA OBJECT UTF8 ATTRIBUTES 1.0",VR::UN,VM::VM1,"Referring Physician's Name Ideographic",false },
  {0x0033,0x0018,"MITRA OBJECT UTF8 ATTRIBUTES 1.0",VR::UN,VM::VM1,"Referring Physician's Name Phonetic",false },
  {0x0033,0x0019,"MITRA OBJECT UTF8 ATTRIBUTES 1.0",VR::UN,VM::VM1,"Requesting Physician's Name Single Byte",false },
  {0x0033,0x001a,"MITRA OBJECT UTF8 ATTRIBUTES 1.0",VR::UN,VM::VM1,"Requesting Physician's Name Ideographic",false },
  {0x0033,0x001b,"MITRA OBJECT UTF8 ATTRIBUTES 1.0",VR::UN,VM::VM1,"Requesting Physician's Name Phonetic",false },
  {0x0033,0x001c,"MITRA OBJECT UTF8 ATTRIBUTES 1.0",VR::UN,VM::VM1,"Performing Physician's Name Single Byte",false },
  {0x0033,0x001d,"MITRA OBJECT UTF8 ATTRIBUTES 1.0",VR::UN,VM::VM1,"Performing Physician's Name Ideographic",false },
  {0x0033,0x001e,"MITRA OBJECT UTF8 ATTRIBUTES 1.0",VR::UN,VM::VM1,"Performing Physician's Name Phonetic",false },
  {0x0029,0x0000,"MITRA PRESENTATION 1.0",VR::CS,VM::VM1,"Mitra Rotation",false },
  {0x0029,0x0001,"MITRA PRESENTATION 1.0",VR::LO,VM::VM1,"Mitra Window Width",false },
  {0x0029,0x0002,"MITRA PRESENTATION 1.0",VR::LO,VM::VM1,"Mitra Window Centre",false },
  {0x0029,0x0003,"MITRA PRESENTATION 1.0",VR::IS,VM::VM1,"Mitra Invert",false },
  {0x0029,0x0004,"MITRA PRESENTATION 1.0",VR::IS,VM::VM1,"Mitra Has Tabstop",false },
  {0x0029,0x0005,"MITRA PRESENTATION 1.0",VR::CS,VM::VM1,"Smooth Rotation",false },
  {0x0029,0x0010,"MITRA PRESENTATION 1.0",VR::CS,VM::VM1,"Breast Box x0",false },
  {0x0029,0x0011,"MITRA PRESENTATION 1.0",VR::CS,VM::VM1,"Breast Box y0",false },
  {0x0029,0x0012,"MITRA PRESENTATION 1.0",VR::CS,VM::VM1,"Breast Box x1",false },
  {0x0029,0x0013,"MITRA PRESENTATION 1.0",VR::CS,VM::VM1,"Breast Box y1",false },
  {0x0009,0x0000,"MMCPrivate",VR::LO,VM::VM1,"Private Creator",false },
  {0x0009,0x0001,"MMCPrivate",VR::LO,VM::VM1,"Technologist",false },
  {0x0009,0x0002,"MMCPrivate",VR::LO,VM::VM1,"ScheduledStudyDateTime",false },
  {0x0009,0x0003,"MMCPrivate",VR::OB,VM::VM1,"StudyAppData",false },
  {0x0009,0x0004,"MMCPrivate",VR::UI,VM::VM1,"ProtocolObjectID",false },
  {0x0009,0x0005,"MMCPrivate",VR::LO,VM::VM1,"Name",false },
  {0x0009,0x0006,"MMCPrivate",VR::IS,VM::VM1,"Frequency",false },
  {0x0009,0x0007,"MMCPrivate",VR::SH,VM::VM1,"UpdateFlag",false },
  {0x0009,0x0008,"MMCPrivate",VR::SH,VM::VM1,"Directory",false },
  {0x0009,0x0009,"MMCPrivate",VR::LO,VM::VM1,"Comments",false },
  {0x0009,0x000a,"MMCPrivate",VR::LO,VM::VM1,"Region",false },
  {0x0009,0x000b,"MMCPrivate",VR::SH,VM::VM1,"Laterality",false },
  {0x0009,0x000c,"MMCPrivate",VR::TM,VM::VM1,"TotalScanTime",false },
  {0x0009,0x000d,"MMCPrivate",VR::LO,VM::VM1,"ContrastMedium",false },
  {0x0009,0x000e,"MMCPrivate",VR::LO,VM::VM1,"CreateDateTime",false },
  {0x0009,0x000f,"MMCPrivate",VR::LO,VM::VM1,"Creator",false },
  {0x0009,0x0010,"MMCPrivate",VR::LO,VM::VM1,"SiteName",false },
  {0x0009,0x0011,"MMCPrivate",VR::LO,VM::VM1,"ReferringPhysician",false },
  {0x0009,0x0012,"MMCPrivate",VR::LO,VM::VM1,"Radiologist",false },
  {0x0009,0x0013,"MMCPrivate",VR::LO,VM::VM1,"Technologist",false },
  {0x0009,0x0014,"MMCPrivate",VR::UI,VM::VM1,"ProtocolUid",false },
  {0x0009,0x0015,"MMCPrivate",VR::SH,VM::VM1,"IsInLibrary",false },
  {0x0009,0x0016,"MMCPrivate",VR::LO,VM::VM1,"Gating",false },
  {0x0009,0x0017,"MMCPrivate",VR::ST,VM::VM1,"Note",false },
  {0x0009,0x0018,"MMCPrivate",VR::IS,VM::VM1,"NumberOfTasks",false },
  {0x0009,0x0019,"MMCPrivate",VR::SH,VM::VM1,"IsFlagRaised",false },
  {0x0009,0x001a,"MMCPrivate",VR::SH,VM::VM1,"IsArchived",false },
  {0x0009,0x001b,"MMCPrivate",VR::SH,VM::VM1,"IsDefault",false },
  {0x0009,0x001c,"MMCPrivate",VR::OB,VM::VM1,"ProtocolAppData",false },
  {0x0009,0x001d,"MMCPrivate",VR::SH,VM::VM1,"IsAllowCascadeSave",false },
  {0x0009,0x001e,"MMCPrivate",VR::SH,VM::VM1,"IsAllowCascadeProtect",false },
  {0x0009,0x001f,"MMCPrivate",VR::SQ,VM::VM1,"TaskInfo",false },
  {0x0009,0x0020,"MMCPrivate",VR::UI,VM::VM1,"TaskInfoObjectID",false },
  {0x0009,0x0021,"MMCPrivate",VR::LO,VM::VM1,"Name",false },
  {0x0009,0x0022,"MMCPrivate",VR::SH,VM::VM1,"TaskStatus",false },
  {0x0009,0x0023,"MMCPrivate",VR::SH,VM::VM1,"TaskPriority",false },
  {0x0009,0x0024,"MMCPrivate",VR::SH,VM::VM1,"Leaf",false },
  {0x0009,0x0025,"MMCPrivate",VR::LO,VM::VM1,"TaskID",false },
  {0x0009,0x0026,"MMCPrivate",VR::IS,VM::VM1,"Frequency",false },
  {0x0009,0x0027,"MMCPrivate",VR::SH,VM::VM1,"UpdateFlag",false },
  {0x0009,0x0028,"MMCPrivate",VR::SH,VM::VM1,"Directory",false },
  {0x0009,0x0029,"MMCPrivate",VR::LO,VM::VM1,"Comments",false },
  {0x0009,0x002a,"MMCPrivate",VR::SH,VM::VM1,"Category",false },
  {0x0009,0x002b,"MMCPrivate",VR::LO,VM::VM1,"Region",false },
  {0x0009,0x002c,"MMCPrivate",VR::SH,VM::VM1,"Laterality",false },
  {0x0009,0x002d,"MMCPrivate",VR::TM,VM::VM1,"ScanTime",false },
  {0x0009,0x002e,"MMCPrivate",VR::LO,VM::VM1,"ContrastMedium",false },
  {0x0009,0x002f,"MMCPrivate",VR::LO,VM::VM1,"CreateDateTime",false },
  {0x0009,0x0030,"MMCPrivate",VR::LO,VM::VM1,"Creator",false },
  {0x0009,0x0031,"MMCPrivate",VR::LO,VM::VM1,"SiteName",false },
  {0x0009,0x0032,"MMCPrivate",VR::LO,VM::VM1,"ReferringPhysician",false },
  {0x0009,0x0033,"MMCPrivate",VR::LO,VM::VM1,"Radiologist",false },
  {0x0009,0x0034,"MMCPrivate",VR::LO,VM::VM1,"Technologist",false },
  {0x0009,0x0035,"MMCPrivate",VR::UI,VM::VM1,"TaskUid",false },
  {0x0009,0x0036,"MMCPrivate",VR::UI,VM::VM1,"TaskInfoUid",false },
  {0x0009,0x0037,"MMCPrivate",VR::SH,VM::VM1,"IsInLibrary",false },
  {0x0009,0x0038,"MMCPrivate",VR::IS,VM::VM1,"TaskOrder",false },
  {0x0009,0x0039,"MMCPrivate",VR::LO,VM::VM1,"Gating",false },
  {0x0009,0x003a,"MMCPrivate",VR::SH,VM::VM1,"Plane",false },
  {0x0009,0x003b,"MMCPrivate",VR::LO,VM::VM1,"SequenceType",false },
  {0x0009,0x003c,"MMCPrivate",VR::SH,VM::VM1,"IsExecutive",false },
  {0x0009,0x003d,"MMCPrivate",VR::ST,VM::VM1,"Note",false },
  {0x0009,0x003e,"MMCPrivate",VR::SH,VM::VM1,"AutoStart",false },
  {0x0009,0x003f,"MMCPrivate",VR::SH,VM::VM1,"AutoSave",false },
  {0x0009,0x0040,"MMCPrivate",VR::SH,VM::VM1,"AutoArchive",false },
  {0x0009,0x0041,"MMCPrivate",VR::IS,VM::VM1,"QueueGroupID",false },
  {0x0009,0x0042,"MMCPrivate",VR::SH,VM::VM1,"IsFlagRaised",false },
  {0x0009,0x0043,"MMCPrivate",VR::SH,VM::VM1,"IsArchived",false },
  {0x0009,0x0044,"MMCPrivate",VR::SH,VM::VM1,"IsDefault",false },
  {0x0009,0x0045,"MMCPrivate",VR::OB,VM::VM1,"TaskInfoAppData",false },
  {0x0009,0x0046,"MMCPrivate",VR::SH,VM::VM1,"IsAllowCascadeSave",false },
  {0x0009,0x0047,"MMCPrivate",VR::SH,VM::VM1,"IsAllowCascadeProtect",false },
  {0x0009,0x0048,"MMCPrivate",VR::LO,VM::VM1,"ProtocolName",false },
  {0x0009,0x004e,"MMCPrivate",VR::LO,VM::VM1,"Cms_BodyPartExamined",false },
  {0x0009,0x004f,"MMCPrivate",VR::LO,VM::VM1,"IsProtected",false },
  {0x0009,0x0050,"MMCPrivate",VR::OB,VM::VM1,"?StudyReservationInfo",false },
  {0x0009,0x0051,"MMCPrivate",VR::LO,VM::VM1,"Cmi_contrastBolusAgent",false },
  {0x0009,0x0052,"MMCPrivate",VR::LO,VM::VM1,"Cms_institutionName",false },
  {0x0009,0x0053,"MMCPrivate",VR::LO,VM::VM1,"Cms_institutionalDepartmentName",false },
  {0x0009,0x0054,"MMCPrivate",VR::LO,VM::VM1,"Cms_seriesDescription",false },
  {0x0009,0x0055,"MMCPrivate",VR::LO,VM::VM1,"Cms_operatorsName",false },
  {0x0009,0x0056,"MMCPrivate",VR::LO,VM::VM1,"Cms_PerformingPhysiciansName",false },
  {0x0009,0x0057,"MMCPrivate",VR::ST,VM::VM1,"Cms_institutionAddress",false },
  {0x0009,0x0058,"MMCPrivate",VR::LO,VM::VM1,"Cmi_imageComments",false },
  {0x0009,0x0059,"MMCPrivate",VR::LO,VM::VM1,"Cmi_instanceCreationDateTime",false },
  {0x0009,0x005a,"MMCPrivate",VR::LO,VM::VM1,"MppsStepStatus",false },
  {0x0009,0x005b,"MMCPrivate",VR::IS,VM::VM1,"FilmedCount",false },
  {0x0009,0x005c,"MMCPrivate",VR::LO,VM::VM1,"IsAllowCascadeSave",false },
  {0x0009,0x005d,"MMCPrivate",VR::LO,VM::VM1,"IsAllowCascadeProtect",false },
  {0x0009,0x005e,"MMCPrivate",VR::LO,VM::VM1,"IsDeleted",false },
  {0x0009,0x005f,"MMCPrivate",VR::UI,VM::VM1,"ProtocolObjectID",false },
  {0x0009,0x0060,"MMCPrivate",VR::OB,VM::VM1,"TaskInfoAppData",false },
  {0x0009,0x0072,"MMCPrivate",VR::UI,VM::VM1,"ProtocolTaskInfoObjectID",false },
  {0x0009,0x0073,"MMCPrivate",VR::IS,VM::VM1,"ProtocolTaskOrder",false },
  {0x0009,0x0074,"MMCPrivate",VR::UI,VM::VM1,"ProtocolTaskUid",false },
  {0x0009,0x0075,"MMCPrivate",VR::OB,VM::VM1,"ProtocolTaskAppData",false },
  {0x0009,0x0076,"MMCPrivate",VR::SH,VM::VM1,"ProtocolTaskIsAllowCascadeSave",false },
  {0x0009,0x0077,"MMCPrivate",VR::SH,VM::VM1,"ProtocolTaskIsAllowCascadeProtect",false },
  {0x0011,0x0000,"MMCPrivate",VR::LO,VM::VM1,"Private Creator",false },
  {0x0011,0x0001,"MMCPrivate",VR::LO,VM::VM1,"IsRapidRegistration",false },
  {0x0011,0x0002,"MMCPrivate",VR::LO,VM::VM1,"IsProtected",false },
  {0x0011,0x0003,"MMCPrivate",VR::IS,VM::VM1,"FilmedCount",false },
  {0x0011,0x0004,"MMCPrivate",VR::OB,VM::VM1,"ApplicationData",false },
  {0x0011,0x0005,"MMCPrivate",VR::LO,VM::VM1,"IsAllowCascadeSave",false },
  {0x0011,0x0006,"MMCPrivate",VR::LO,VM::VM1,"IsAllowCascadeProtect",false },
  {0x0011,0x0007,"MMCPrivate",VR::LO,VM::VM1,"IsDeleted",false },
  {0x0019,0x0000,"MMCPrivate",VR::LO,VM::VM1,"Private Creator",false },
  {0x0019,0x0001,"MMCPrivate",VR::LO,VM::VM1,"ProcType",false },
  {0x0019,0x0002,"MMCPrivate",VR::LO,VM::VM1,"Plane",false },
  {0x0019,0x0003,"MMCPrivate",VR::SH,VM::VM1,"IsSnapShotSeries",false },
  {0x0019,0x0004,"MMCPrivate",VR::DS,VM::VM1,"MaxFsColor",false },
  {0x0019,0x0005,"MMCPrivate",VR::LO,VM::VM1,"SeriesCategoryType",false },
  {0x0019,0x0007,"MMCPrivate",VR::LO,VM::VM1,"ImageContrastBolusAgent",false },
  {0x0019,0x0008,"MMCPrivate",VR::DS,VM::VM1,"ImageSliceThickness",false },
  {0x0019,0x0009,"MMCPrivate",VR::DS,VM::VM1,"ImageReconstructionDiameter",false },
  {0x0019,0x000a,"MMCPrivate",VR::LO,VM::VM1,"ImageEchoTime",false },
  {0x0019,0x000b,"MMCPrivate",VR::DS,VM::VM1,"ImageRepetitionTime",false },
  {0x0019,0x000c,"MMCPrivate",VR::LO,VM::VM1,"SequenceType",false },
  {0x0019,0x000d,"MMCPrivate",VR::LO,VM::VM1,"TaskUid",false },
  {0x0019,0x000e,"MMCPrivate",VR::OB,VM::VM1,"SeriesAppData",false },
  {0x0019,0x000f,"MMCPrivate",VR::IS,VM::VM1,"MultiSliceNumber",false },
  {0x0019,0x0010,"MMCPrivate",VR::LO,VM::VM1,"ImageScanTime",false },
  {0x0019,0x0011,"MMCPrivate",VR::LO,VM::VM1,"IsProtected",false },
  {0x0019,0x0012,"MMCPrivate",VR::IS,VM::VM1,"ImageIncrement",false },
  {0x0019,0x0013,"MMCPrivate",VR::LO,VM::VM1,"MppsStepStatus",false },
  {0x0019,0x0014,"MMCPrivate",VR::IS,VM::VM1,"StorageCommittedCount",false },
  {0x0019,0x0015,"MMCPrivate",VR::IS,VM::VM1,"ArchivedCount",false },
  {0x0019,0x0016,"MMCPrivate",VR::IS,VM::VM1,"TransferredCount",false },
  {0x0019,0x0017,"MMCPrivate",VR::LO,VM::VM1,"IsAllowCascadeSave",false },
  {0x0019,0x0018,"MMCPrivate",VR::LO,VM::VM1,"IsAllowCascadeProtect",false },
  {0x0019,0x0019,"MMCPrivate",VR::LO,VM::VM1,"IsDeleted",false },
  {0x0019,0x001a,"MMCPrivate",VR::UI,VM::VM1,"CharacterizedImageInstanceUid",false },
  {0x0019,0x001b,"MMCPrivate",VR::IS,VM::VM1,"CharacterizedImageCount",false },
  {0x0019,0x001c,"MMCPrivate",VR::LO,VM::VM1,"InternalWindowWidth",false },
  {0x0019,0x001d,"MMCPrivate",VR::LO,VM::VM1,"InternalWindowLevel",false },
  {0x0019,0x001e,"MMCPrivate",VR::LO,VM::VM1,"?FlipAngle?",false },
  {0x0019,0x001f,"MMCPrivate",VR::LO,VM::VM1,"?InversionTime?",false },
  {0x0019,0x0020,"MMCPrivate",VR::UI,VM::VM1,"MultiFrameSopInstanceUid",false },
  {0x0019,0x0021,"MMCPrivate",VR::OB,VM::VM1,"?ScanInterface.GroupParamArchive",false },
  {0x0029,0x0000,"MMCPrivate",VR::LO,VM::VM1,"Private Creator",false },
  {0x0029,0x0001,"MMCPrivate",VR::IS,VM::VM1,"SliceNumber",false },
  {0x0029,0x0002,"MMCPrivate",VR::IS,VM::VM1,"PhaseNumber",false },
  {0x0029,0x0003,"MMCPrivate",VR::LO,VM::VM1,"ProcType",false },
  {0x0029,0x0004,"MMCPrivate",VR::LO,VM::VM1,"StopwatchTime",false },
  {0x0029,0x0005,"MMCPrivate",VR::LO,VM::VM1,"Plane",false },
  {0x0029,0x0006,"MMCPrivate",VR::LO,VM::VM1,"ScanTime",false },
  {0x0029,0x0008,"MMCPrivate",VR::LO,VM::VM1,"DualSliceFlag",false },
  {0x0029,0x0009,"MMCPrivate",VR::LO,VM::VM1,"SspRatio",false },
  {0x0029,0x000a,"MMCPrivate",VR::LO,VM::VM1,"GatingSignalSource",false },
  {0x0029,0x000b,"MMCPrivate",VR::LO,VM::VM1,"Rephase",false },
  {0x0029,0x000c,"MMCPrivate",VR::LO,VM::VM1,"HalfEcho",false },
  {0x0029,0x000d,"MMCPrivate",VR::LO,VM::VM1,"RectFOVRatio",false },
  {0x0029,0x000e,"MMCPrivate",VR::LO,VM::VM1,"HalfScan",false },
  {0x0029,0x000f,"MMCPrivate",VR::LO,VM::VM1,"NumShots",false },
  {0x0029,0x0010,"MMCPrivate",VR::LO,VM::VM1,"ContrastAgent",false },
  {0x0029,0x0011,"MMCPrivate",VR::LO,VM::VM1,"EchoAllocation",false },
  {0x0029,0x0012,"MMCPrivate",VR::LO,VM::VM1,"NumEchoShift",false },
  {0x0029,0x0013,"MMCPrivate",VR::LO,VM::VM1,"FatSat",false },
  {0x0029,0x0014,"MMCPrivate",VR::LO,VM::VM1,"MTC",false },
  {0x0029,0x0015,"MMCPrivate",VR::LO,VM::VM1,"NumPreSat",false },
  {0x0029,0x0016,"MMCPrivate",VR::LO,VM::VM1,"TargetVelocity",false },
  {0x0029,0x0017,"MMCPrivate",VR::LO,VM::VM1,"VENCAxis",false },
  {0x0029,0x0018,"MMCPrivate",VR::LO,VM::VM1,"NumVENCDirection",false },
  {0x0029,0x001c,"MMCPrivate",VR::LO,VM::VM1,"IsScalableWindowLevel",false },
  {0x0029,0x001d,"MMCPrivate",VR::LO,VM::VM1,"ThreeDSettingLineAngle",false },
  {0x0029,0x001e,"MMCPrivate",VR::LO,VM::VM1,"MPGTotalAxis",false },
  {0x0029,0x001f,"MMCPrivate",VR::LO,VM::VM1,"MPGAxisNumber",false },
  {0x0029,0x0020,"MMCPrivate",VR::IS,VM::VM1,"MultiEchoNumber",false },
  {0x0029,0x0021,"MMCPrivate",VR::DS,VM::VM1,"NaviAverageGateWidth",false },
  {0x0029,0x0022,"MMCPrivate",VR::ST,VM::VM1,"ShimCompensateValue",false },
  {0x0029,0x0023,"MMCPrivate",VR::LO,VM::VM1,"GCOffset",false },
  {0x0029,0x0024,"MMCPrivate",VR::DS,VM::VM1,"NaviMaxGateWidth",false },
  {0x0029,0x0025,"MMCPrivate",VR::DS,VM::VM1,"NaviMinGateWidth",false },
  {0x0029,0x0026,"MMCPrivate",VR::DS,VM::VM1,"NaviMaxGatePosition",false },
  {0x0029,0x0027,"MMCPrivate",VR::DS,VM::VM1,"NaviMinGatePosition",false },
  {0x0029,0x0028,"MMCPrivate",VR::DS,VM::VM1,"TimeDuration",false },
  {0x0029,0x0029,"MMCPrivate",VR::DS,VM::VM1,"TablePosition",false },
  {0x0029,0x002a,"MMCPrivate",VR::DS,VM::VM1,"NaviInitialGateWidth",false },
  {0x0029,0x002b,"MMCPrivate",VR::DS,VM::VM1,"NaviFinalGateWidth",false },
  {0x0029,0x002c,"MMCPrivate",VR::DS,VM::VM1,"NaviInitialGatePosition",false },
  {0x0029,0x002d,"MMCPrivate",VR::DS,VM::VM1,"NaviFinalGatePosition",false },
  {0x0029,0x002e,"MMCPrivate",VR::DS,VM::VM1,"NaviAverageGatePosition",false },
  {0x0029,0x002f,"MMCPrivate",VR::OB,VM::VM1,"ImageAppData",false },
  {0x0029,0x0030,"MMCPrivate",VR::FD,VM::VM1,"DiffusionBValue",false },
  {0x0029,0x0031,"MMCPrivate",VR::SQ,VM::VM1,"SharedFunctionalGroupsSequence",false },
  {0x0029,0x0032,"MMCPrivate",VR::SQ,VM::VM1,"PerFrameFunctionalGroupsSequence",false },
  {0x0029,0x0033,"MMCPrivate",VR::DS,VM::VM1,"LossyImageCompressionRatio",false },
  {0x0029,0x0034,"MMCPrivate",VR::UI,VM::VM1,"InstanceCreatorUID",false },
  {0x0029,0x0035,"MMCPrivate",VR::UI,VM::VM1,"RelatedGeneralSOPClassUID",false },
  {0x0029,0x0036,"MMCPrivate",VR::UI,VM::VM1,"OriginalSpecializedSOPClassUID",false },
  {0x0029,0x0037,"MMCPrivate",VR::SH,VM::VM1,"TimezoneOffsetFromUTC",false },
  {0x0029,0x0038,"MMCPrivate",VR::CS,VM::VM1,"SOPInstanceStatus",false },
  {0x0029,0x0039,"MMCPrivate",VR::DT,VM::VM1,"SOPAuthorizationDateandTime",false },
  {0x0029,0x003a,"MMCPrivate",VR::LT,VM::VM1,"SOPAuthorizationComment",false },
  {0x0029,0x003b,"MMCPrivate",VR::LO,VM::VM1,"AuthorizationEquipmentCertificationNumber",false },
  {0x0029,0x003c,"MMCPrivate",VR::UL,VM::VM1,"ConcatenationFrameOffsetNumber",false },
  {0x0029,0x003d,"MMCPrivate",VR::US,VM::VM1,"RepresentativeFrameNumber",false },
  {0x0029,0x003e,"MMCPrivate",VR::UI,VM::VM1,"ConcatenationUID",false },
  {0x0029,0x003f,"MMCPrivate",VR::US,VM::VM1,"InConcatenationNumber",false },
  {0x0029,0x0040,"MMCPrivate",VR::CS,VM::VM1,"CardiacSynchronizationTechnique",false },
  {0x0029,0x0041,"MMCPrivate",VR::CS,VM::VM1,"CardiacSignalSource",false },
  {0x0029,0x0042,"MMCPrivate",VR::FD,VM::VM1,"CardiacRRIntervalSpecified",false },
  {0x0029,0x0043,"MMCPrivate",VR::CS,VM::VM1,"CardiacBeatRejectionTechnique",false },
  {0x0029,0x0044,"MMCPrivate",VR::IS,VM::VM1,"LowRRValue",false },
  {0x0029,0x0045,"MMCPrivate",VR::IS,VM::VM1,"HighRRValue",false },
  {0x0029,0x0046,"MMCPrivate",VR::IS,VM::VM1,"IntervalsAcquired",false },
  {0x0029,0x0047,"MMCPrivate",VR::IS,VM::VM1,"IntervalsRejected",false },
  {0x0029,0x0048,"MMCPrivate",VR::CS,VM::VM1,"RespiratoryMotionCompensationTechnique",false },
  {0x0029,0x0049,"MMCPrivate",VR::CS,VM::VM1,"RespiratorySignalSource",false },
  {0x0029,0x004a,"MMCPrivate",VR::CS,VM::VM1,"BulkMotionCompensationTechnique",false },
  {0x0029,0x004b,"MMCPrivate",VR::CS,VM::VM1,"BulkMotionSignalSource",false },
  {0x0029,0x004c,"MMCPrivate",VR::CS,VM::VM1,"PixelPresentation",false },
  {0x0029,0x004d,"MMCPrivate",VR::CS,VM::VM1,"VolumetricProperties",false },
  {0x0029,0x004e,"MMCPrivate",VR::CS,VM::VM1,"VolumeBasedCalculationTechnique",false },
  {0x0029,0x004f,"MMCPrivate",VR::ST,VM::VM1,"AcquisitionContextDescription",false },
  {0x0029,0x0050,"MMCPrivate",VR::SQ,VM::VM1,"ModalityLUTSequence",false },
  {0x0029,0x0051,"MMCPrivate",VR::LO,VM::VM1,"LUTDescriptor",false },
  {0x0029,0x0052,"MMCPrivate",VR::LO,VM::VM1,"LUTExplanation",false },
  {0x0029,0x0053,"MMCPrivate",VR::LO,VM::VM1,"LUTData",false },
  {0x0029,0x0054,"MMCPrivate",VR::CS,VM::VM1,"PresentationLUTShape",false },
  {0x0029,0x0055,"MMCPrivate",VR::SQ,VM::VM1,"FrameAnatomySequence",false },
  {0x0029,0x0056,"MMCPrivate",VR::CS,VM::VM1,"FrameLaterality",false },
  {0x0029,0x0057,"MMCPrivate",VR::SQ,VM::VM1,"AnatomicRegionSequence",false },
  {0x0029,0x0058,"MMCPrivate",VR::SH,VM::VM1,"AnatomicRegionCodeValue",false },
  {0x0029,0x0059,"MMCPrivate",VR::SH,VM::VM1,"AnatomicRegionCodingSchemeDesignator",false },
  {0x0029,0x005a,"MMCPrivate",VR::SH,VM::VM1,"AnatomicRegionCodingSchemeVersion",false },
  {0x0029,0x005b,"MMCPrivate",VR::LO,VM::VM1,"AnatomicRegionCodeMeaning",false },
  {0x0029,0x005c,"MMCPrivate",VR::SQ,VM::VM1,"PixelValueTransformationSequence",false },
  {0x0029,0x005d,"MMCPrivate",VR::LO,VM::VM1,"RescaleType",false },
  {0x0029,0x005e,"MMCPrivate",VR::SQ,VM::VM1,"CardiacTriggerSequence",false },
  {0x0029,0x005f,"MMCPrivate",VR::FD,VM::VM1,"TriggerDelayTime",false },
  {0x0029,0x0060,"MMCPrivate",VR::SQ,VM::VM1,"FrameVOILUTSequence",false },
  {0x0029,0x0061,"MMCPrivate",VR::LO,VM::VM1,"WindowCenterAndWidthExplanation",false },
  {0x0029,0x0062,"MMCPrivate",VR::CS,VM::VM1,"AcquisitionContrast",false },
  {0x0029,0x0063,"MMCPrivate",VR::SQ,VM::VM1,"MRModifierSequence",false },
  {0x0029,0x0064,"MMCPrivate",VR::CS,VM::VM1,"ParallelAcquisitionTechnic",false },
  {0x0029,0x0065,"MMCPrivate",VR::FD,VM::VM1,"ParallelReductionFactorSecIn",false },
  {0x0029,0x0066,"MMCPrivate",VR::CS,VM::VM1,"InversionRecovery",false },
  {0x0029,0x0067,"MMCPrivate",VR::CS,VM::VM1,"FlowCompensation",false },
  {0x0029,0x0068,"MMCPrivate",VR::CS,VM::VM1,"FlowCompensationDirection",false },
  {0x0029,0x0069,"MMCPrivate",VR::CS,VM::VM1,"SpatialPreSaturation",false },
  {0x0029,0x006a,"MMCPrivate",VR::CS,VM::VM1,"PartialFourier",false },
  {0x0029,0x006b,"MMCPrivate",VR::CS,VM::VM1,"PartialFourierDirection",false },
  {0x0029,0x006c,"MMCPrivate",VR::CS,VM::VM1,"ResonantNucleus",false },
  {0x0029,0x006d,"MMCPrivate",VR::CS,VM::VM1,"KSpaceFiltering",false },
  {0x0029,0x006e,"MMCPrivate",VR::CS,VM::VM1,"ApplicableSafetyStandardAgency",false },
  {0x0029,0x006f,"MMCPrivate",VR::LO,VM::VM1,"ApplicableSafetyStandardDescription",false },
  {0x0029,0x0070,"MMCPrivate",VR::SQ,VM::VM1,"MRReceiveCoilSequence",false },
  {0x0029,0x0071,"MMCPrivate",VR::LO,VM::VM1,"ReceiveCoilManufacturerName",false },
  {0x0029,0x0072,"MMCPrivate",VR::CS,VM::VM1,"ReceiveCoilType",false },
  {0x0029,0x0073,"MMCPrivate",VR::CS,VM::VM1,"QuadratureReceiveCoil",false },
  {0x0029,0x0074,"MMCPrivate",VR::LO,VM::VM1,"MultiCoilConfiguration",false },
  {0x0029,0x0075,"MMCPrivate",VR::CS,VM::VM1,"ComplexImageComponent",false },
  {0x0029,0x0076,"MMCPrivate",VR::SH,VM::VM1,"PulseSequenceName",false },
  {0x0029,0x0077,"MMCPrivate",VR::CS,VM::VM1,"EchoPulseSequence",false },
  {0x0029,0x0078,"MMCPrivate",VR::CS,VM::VM1,"MultipleSpinEcho",false },
  {0x0029,0x0079,"MMCPrivate",VR::CS,VM::VM1,"MultiPlanarExcitation",false },
  {0x0029,0x007a,"MMCPrivate",VR::CS,VM::VM1,"PhaseContrast",false },
  {0x0029,0x007b,"MMCPrivate",VR::CS,VM::VM1,"TimeOfFlightContrast",false },
  {0x0029,0x007c,"MMCPrivate",VR::CS,VM::VM1,"SteadyStatePulseSequence",false },
  {0x0029,0x007d,"MMCPrivate",VR::CS,VM::VM1,"EchoPlanarPulseSequence",false },
  {0x0029,0x007e,"MMCPrivate",VR::CS,VM::VM1,"SpectrallySelectedSuppression",false },
  {0x0029,0x007f,"MMCPrivate",VR::CS,VM::VM1,"OversamplingPhase",false },
  {0x0029,0x0080,"MMCPrivate",VR::CS,VM::VM1,"SegmentedKSpaceTraversal",false },
  {0x0029,0x0081,"MMCPrivate",VR::CS,VM::VM1,"CoverageOfKSpace",false },
  {0x0029,0x0082,"MMCPrivate",VR::SQ,VM::VM1,"MRTimingAndRelatedParametersSequence",false },
  {0x0029,0x0083,"MMCPrivate",VR::US,VM::VM1,"RFEchoTrainLength",false },
  {0x0029,0x0084,"MMCPrivate",VR::US,VM::VM1,"GradientEchoTrainLength",false },
  {0x0029,0x0085,"MMCPrivate",VR::CS,VM::VM1,"GradientOutputType",false },
  {0x0029,0x0086,"MMCPrivate",VR::FD,VM::VM1,"GradientOutput",false },
  {0x0029,0x0087,"MMCPrivate",VR::SQ,VM::VM1,"MRFOVGeometrySequence",false },
  {0x0029,0x0088,"MMCPrivate",VR::US,VM::VM1,"MRAcquisitionFrequencyEncodingSteps",false },
  {0x0029,0x0089,"MMCPrivate",VR::US,VM::VM1,"MRAcquisitionPhaseEncodingStepsInPlane",false },
  {0x0029,0x008a,"MMCPrivate",VR::US,VM::VM1,"MRAcquisitionPhaseEncodingStepsOutOfPlane",false },
  {0x0029,0x008b,"MMCPrivate",VR::SQ,VM::VM1,"MRTransmitCoilSequence",false },
  {0x0029,0x008c,"MMCPrivate",VR::SH,VM::VM1,"TransmitCoilName",false },
  {0x0029,0x008d,"MMCPrivate",VR::LO,VM::VM1,"TransmitCoilManufacturerName",false },
  {0x0029,0x008e,"MMCPrivate",VR::CS,VM::VM1,"TransmitCoilType",false },
  {0x0029,0x008f,"MMCPrivate",VR::SQ,VM::VM1,"MREchoSequence",false },
  {0x0029,0x0090,"MMCPrivate",VR::FD,VM::VM1,"EffectiveEchoTime",false },
  {0x0029,0x0091,"MMCPrivate",VR::SQ,VM::VM1,"MRMetaboliteMapSequence",false },
  {0x0029,0x0092,"MMCPrivate",VR::ST,VM::VM1,"MetaboliteMapDescription",false },
  {0x0029,0x0093,"MMCPrivate",VR::SQ,VM::VM1,"MetaboliteMapCodeSequence",false },
  {0x0029,0x0094,"MMCPrivate",VR::SH,VM::VM1,"MetaboliteMapCodeValue",false },
  {0x0029,0x0095,"MMCPrivate",VR::SH,VM::VM1,"MetaboliteMapCodingSchemeDesignator",false },
  {0x0029,0x0096,"MMCPrivate",VR::SH,VM::VM1,"MetaboliteMapCodingSchemeVersion",false },
  {0x0029,0x0097,"MMCPrivate",VR::LO,VM::VM1,"MetaboliteMapCodeMeaning",false },
  {0x0029,0x0098,"MMCPrivate",VR::SQ,VM::VM1,"MRImagingModifierSequence",false },
  {0x0029,0x0099,"MMCPrivate",VR::CS,VM::VM1,"MagnetizationTransfer",false },
  {0x0029,0x009a,"MMCPrivate",VR::CS,VM::VM1,"BloodSignalNulling",false },
  {0x0029,0x009b,"MMCPrivate",VR::CS,VM::VM1,"Tagging",false },
  {0x0029,0x009c,"MMCPrivate",VR::FD,VM::VM1,"TagSpacingFirstDimension",false },
  {0x0029,0x009d,"MMCPrivate",VR::FD,VM::VM1,"TagSpacingSecondDimension",false },
  {0x0029,0x009e,"MMCPrivate",VR::FD,VM::VM1,"TagAngleFirstAxis",false },
  {0x0029,0x009f,"MMCPrivate",VR::SS,VM::VM1,"TagAngleSecondAxis",false },
  {0x0029,0x00a0,"MMCPrivate",VR::FD,VM::VM1,"TagThickness",false },
  {0x0029,0x00a1,"MMCPrivate",VR::FD,VM::VM1,"TaggingDelay",false },
  {0x0029,0x00a2,"MMCPrivate",VR::FD,VM::VM1,"TransmitterFrequency",false },
  {0x0029,0x00a3,"MMCPrivate",VR::DS,VM::VM1,"PixelBandwidth",false },
  {0x0029,0x00a4,"MMCPrivate",VR::SQ,VM::VM1,"MRVelocityEncodingSequence",false },
  {0x0029,0x00a5,"MMCPrivate",VR::FD,VM::VM1,"VelocityEncodingDirection",false },
  {0x0029,0x00a6,"MMCPrivate",VR::FD,VM::VM1,"VelocityEncodingMinimumValue",false },
  {0x0029,0x00a7,"MMCPrivate",VR::FD,VM::VM1,"VelocityEncodingMaximumValue",false },
  {0x0029,0x00a8,"MMCPrivate",VR::SQ,VM::VM1,"MRImageFrameTypeSequence",false },
  {0x0029,0x00a9,"MMCPrivate",VR::CS,VM::VM1,"FrameType",false },
  {0x0029,0x00aa,"MMCPrivate",VR::CS,VM::VM1,"PixelPresentation",false },
  {0x0029,0x00ab,"MMCPrivate",VR::CS,VM::VM1,"VolumetricProperties",false },
  {0x0029,0x00ac,"MMCPrivate",VR::CS,VM::VM1,"VolumeBasedCalculationTechnique",false },
  {0x0029,0x00ad,"MMCPrivate",VR::IS,VM::VM1,"FilmedCount",false },
  {0x0029,0x00ae,"MMCPrivate",VR::LO,VM::VM1,"IsTransferred",false },
  {0x0029,0x00af,"MMCPrivate",VR::LO,VM::VM1,"IsArchived",false },
  {0x0029,0x00b0,"MMCPrivate",VR::LO,VM::VM1,"MppsStepStatus",false },
  {0x0029,0x00b1,"MMCPrivate",VR::LO,VM::VM1,"CommitmentStatus",false },
  {0x0029,0x00b2,"MMCPrivate",VR::LO,VM::VM1,"IsStorageCommitted",false },
  {0x0029,0x00b3,"MMCPrivate",VR::LO,VM::VM1,"IsDicom",false },
  {0x0029,0x00b4,"MMCPrivate",VR::LO,VM::VM1,"IsAllowCascadeSave",false },
  {0x0029,0x00b5,"MMCPrivate",VR::LO,VM::VM1,"IsAllowCascadeProtect",false },
  {0x0029,0x00b6,"MMCPrivate",VR::LO,VM::VM1,"IsDeleted",false },
  {0x0029,0x00b7,"MMCPrivate",VR::OB,VM::VM1,"ApplicationData",false },
  {0x0029,0x00b8,"MMCPrivate",VR::LO,VM::VM1,"IsAllowCascadeSave",false },
  {0x0029,0x00b9,"MMCPrivate",VR::LO,VM::VM1,"IsAllowCascadeProtect",false },
  {0x0029,0x00ba,"MMCPrivate",VR::LO,VM::VM1,"IsDeleted",false },
  {0x0029,0x00bb,"MMCPrivate",VR::IS,VM::VM1,"VOI1",false },
  {0x0029,0x00bc,"MMCPrivate",VR::IS,VM::VM1,"VOI2",false },
  {0x0029,0x00bd,"MMCPrivate",VR::UI,VM::VM1,"BackgroundImageInstanceUID",false },
  {0x0029,0x00be,"MMCPrivate",VR::LO,VM::VM1,"IsStoredToPortableMedia",false },
  {0x0029,0x00bf,"MMCPrivate",VR::DS,VM::VM1,"Voi1",false },
  {0x0029,0x00c0,"MMCPrivate",VR::FD,VM::VM1,"SelectiveIRColumn",false },
  {0x0029,0x00c1,"MMCPrivate",VR::DS,VM::VM1,"Voi2",false },
  {0x0029,0x00c2,"MMCPrivate",VR::DS,VM::VM1,"MixingTime",false },
  {0x0029,0x00c3,"MMCPrivate",VR::DS,VM::VM1,"SelectiveIRPosition",false },
  {0x0029,0x00c4,"MMCPrivate",VR::DS,VM::VM1,"SelectiveIRRow",false },
  {0x0029,0x00c5,"MMCPrivate",VR::DS,VM::VM1,"SelectiveIRColumn",false },
  {0x0029,0x00c6,"MMCPrivate",VR::DS,VM::VM1,"SelectiveIROrientation",false },
  {0x0029,0x00c7,"MMCPrivate",VR::LO,VM::VM1,"SelectiveIRThickness",false },
  {0x0029,0x00c8,"MMCPrivate",VR::SH,VM::VM1,"RephaseOrderSlice",false },
  {0x0029,0x00c9,"MMCPrivate",VR::SH,VM::VM1,"RephaseOrderPhase",false },
  {0x0029,0x00ca,"MMCPrivate",VR::SH,VM::VM1,"RephaseOrderFreq",false },
  {0x0029,0x00cb,"MMCPrivate",VR::ST,VM::VM1,"MetaboliteMapDescription",false },
  {0x0029,0x00cc,"MMCPrivate",VR::SQ,VM::VM1,"volumeLocalizationSeq",false },
  {0x0029,0x00cd,"MMCPrivate",VR::FD,VM::VM1,"SlabThickness",false },
  {0x0029,0x00ce,"MMCPrivate",VR::FD,VM::VM1,"SlabOrientation",false },
  {0x0029,0x00cf,"MMCPrivate",VR::FD,VM::VM1,"MidSlabPosition",false },
  {0x0029,0x00d0,"MMCPrivate",VR::LO,VM::VM1,"AcqModeSliceDir",false },
  {0x0029,0x00d1,"MMCPrivate",VR::LO,VM::VM1,"IRThicknessRatio",false },
  {0x0029,0x00d2,"MMCPrivate",VR::LO,VM::VM1,"BBIRThicknessRatio",false },
  {0x0029,0x00d3,"MMCPrivate",VR::LO,VM::VM1,"DeltaAngle",false },
  {0x0029,0x00d4,"MMCPrivate",VR::IS,VM::VM1,"MultiFrameFrameNumber",false },
  {0x0029,0x00d5,"MMCPrivate",VR::UI,VM::VM1,"EnhancedSopInstanceUid",false },
  {0x0029,0x00d6,"MMCPrivate",VR::LO,VM::VM1,"PolarityOfPhaseEncoding",false },
  {0x0029,0x00d7,"MMCPrivate",VR::OB,VM::VM1,"PresentationStates",false },
  {0x0029,0x00d9,"MMCPrivate",VR::DS,VM::VM3,"Magnetic Field Direction",false },
  {0x0041,0x0001,"MMCPrivate",VR::OB,VM::VM1,"RawDataAppData",false },
  {0x0041,0x0002,"MMCPrivate",VR::SQ,VM::VM1,"RawDataIndex",false },
  {0x0041,0x0003,"MMCPrivate",VR::LO,VM::VM1,"ChannelNumber",false },
  {0x0041,0x0004,"MMCPrivate",VR::LO,VM::VM1,"AxisDirection",false },
  {0x0041,0x0005,"MMCPrivate",VR::LO,VM::VM1,"SlabNumbe",false },
  {0x0041,0x0006,"MMCPrivate",VR::LO,VM::VM1,"CardiacPhaseNumbe",false },
  {0x0041,0x0007,"MMCPrivate",VR::LO,VM::VM1,"EchoNumber",false },
  {0x0041,0x0008,"MMCPrivate",VR::LO,VM::VM1,"SliceEncodeNumber",false },
  {0x0041,0x0009,"MMCPrivate",VR::LO,VM::VM1,"NsaNumber",false },
  {0x0041,0x000a,"MMCPrivate",VR::OB,VM::VM1,"RawData",false },
  {0x0041,0x000b,"MMCPrivate",VR::SS,VM::VM1,"RawDataMRInfo",false },
  {0x0041,0x000c,"MMCPrivate",VR::IS,VM::VM1,"NumberOfVoxels",false },
  {0x0041,0x000d,"MMCPrivate",VR::DS,VM::VM1,"MixingTime",false },
  {0x0041,0x000e,"MMCPrivate",VR::DS,VM::VM1,"ADDiff",false },
  {0x0041,0x000f,"MMCPrivate",VR::LO,VM::VM1,"ScanTime",false },
  {0x0041,0x0010,"MMCPrivate",VR::LO,VM::VM1,"NumPreSat",false },
  {0x0041,0x0011,"MMCPrivate",VR::LO,VM::VM1,"IsStoredToPortableMedia",false },
  {0x0041,0x0012,"MMCPrivate",VR::DS,VM::VM1,"Voi1",false },
  {0x0041,0x0013,"MMCPrivate",VR::DS,VM::VM1,"Voi2",false },
  {0x0041,0x0014,"MMCPrivate",VR::DS,VM::VM1,"VoxelSize",false },
  {0x0041,0x0015,"MMCPrivate",VR::IS,VM::VM1,"FreqPoint",false },
  {0x0041,0x0016,"MMCPrivate",VR::SH,VM::VM1,"LowOrderShim",false },
  {0x0041,0x0017,"MMCPrivate",VR::SH,VM::VM1,"EccLevel",false },
  {0x0041,0x0018,"MMCPrivate",VR::FL,VM::VM1,"FwhmHz",false },
  {0x0041,0x0019,"MMCPrivate",VR::FL,VM::VM1,"FwhmPpm",false },
  {0x0041,0x001a,"MMCPrivate",VR::FL,VM::VM1,"WaterSupRate",false },
  {0x0071,0x0001,"MMCPrivate",VR::FL,VM::VM1,"ForegroundTransparency",false },
  {0x0071,0x0002,"MMCPrivate",VR::LO,VM::VM1,"IsDisplayBackgroundImage",false },
  {0x0071,0x0003,"MMCPrivate",VR::FL,VM::VM1,"ForegroundHorizontalShift",false },
  {0x0071,0x0004,"MMCPrivate",VR::FL,VM::VM1,"ForegroundVerticalShift",false },
  {0x0071,0x0005,"MMCPrivate",VR::FL,VM::VM1,"ForegroundRotationAngle",false },
  {0x0071,0x0006,"MMCPrivate",VR::FL,VM::VM1,"ForegroundMagnification",false },
  {0x0071,0x0007,"MMCPrivate",VR::OB,VM::VM1,"ApplicationData",false },
  {0x0021,0x0001,"Mayo/IBM Archive Project",VR::UN,VM::VM1,"?",false },
  {0x0021,0x0010,"Mayo/IBM Archive Project",VR::UN,VM::VM1,"?",false },
  {0x0021,0x0011,"Mayo/IBM Archive Project",VR::UN,VM::VM1,"?",false },
  {0x0021,0x0012,"Mayo/IBM Archive Project",VR::UN,VM::VM1,"?",false },
  {0x0021,0x0013,"Mayo/IBM Archive Project",VR::UN,VM::VM1,"?",false },
  {0x0021,0x0014,"Mayo/IBM Archive Project",VR::UN,VM::VM1,"?",false },
  {0x0021,0x0015,"Mayo/IBM Archive Project",VR::UN,VM::VM1,"?",false },
  {0x0021,0x0016,"Mayo/IBM Archive Project",VR::UN,VM::VM1,"?",false },
  {0x0021,0x0017,"Mayo/IBM Archive Project",VR::UN,VM::VM1,"?",false },
  {0x0021,0x0018,"Mayo/IBM Archive Project",VR::UN,VM::VM1,"?",false },
  {0x0021,0x0019,"Mayo/IBM Archive Project",VR::UN,VM::VM1,"?",false },
  {0x0021,0x001a,"Mayo/IBM Archive Project",VR::UN,VM::VM1,"?",false },
  {0x0021,0x001b,"Mayo/IBM Archive Project",VR::UN,VM::VM1,"?",false },
  {0x0021,0x001c,"Mayo/IBM Archive Project",VR::UN,VM::VM1,"?",false },
  {0x0021,0x001d,"Mayo/IBM Archive Project",VR::UN,VM::VM1,"?",false },
  {0x0021,0x001e,"Mayo/IBM Archive Project",VR::UN,VM::VM1,"?",false },
  {0x0021,0x001f,"Mayo/IBM Archive Project",VR::UN,VM::VM1,"?",false },
  {0x0021,0x0020,"Mayo/IBM Archive Project",VR::UN,VM::VM1,"?",false },
  {0x0021,0x0040,"Mayo/IBM Archive Project",VR::UN,VM::VM1,"?",false },
  {0x0021,0x0041,"Mayo/IBM Archive Project",VR::UN,VM::VM1,"?",false },
  {0x0021,0x0050,"Mayo/IBM Archive Project",VR::UN,VM::VM1,"?",false },
  {0x0021,0x0060,"Mayo/IBM Archive Project",VR::UN,VM::VM1,"?",false },
  {0x0021,0x0065,"Mayo/IBM Archive Project",VR::UN,VM::VM1,"?",false },
  {0x0019,0x0001,"MeVis BreastCare",VR::LO,VM::VM1,"?",false },
  {0x0071,0x0001,"MeVis BreastCare",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0011,"MeVis eD: Geometry Information",VR::UN,VM::VM1,"GeoScannerOrigin",false },
  {0x0009,0x0021,"MeVis eD: Geometry Information",VR::UN,VM::VM1,"SlicesID",false },
  {0x0009,0x0010,"MeVis eD: Slice Information",VR::UN,VM::VM1,"SlicesSOPInstanceUID",false },
  {0x0021,0x0010,"MeVis eD: Timepoint Information",VR::LT,VM::VM1,"TPDateTime",false },
  {0x0021,0x0011,"MeVis eD: Timepoint Information",VR::CS,VM::VM1,"TPDateTimeType",false },
  {0x0021,0x0012,"MeVis eD: Timepoint Information",VR::UN,VM::VM1,"TPSeriesDescription",false },
  {0x0021,0x0013,"MeVis eD: Timepoint Information",VR::UN,VM::VM1,"TPGradientDirections",false },
  {0x0021,0x0021,"MeVis eD: Timepoint Information",VR::UN,VM::VM1,"GeometryID",false },
  {0x0009,0x0010,"MeVis eatDicom",VR::UN,VM::VM1,"EatDicomVersion",false },
  {0x0009,0x0011,"MeVis eatDicom",VR::UN,VM::VM1,"EatDicomOptions",false },
  {0x0009,0x0021,"MeVis eatDicom",VR::UN,VM::VM1,"TimepointsID",false },
  {0x5473,0x0003,"MedIns HP Extensions",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0000,"Mitra Markup 1.0",VR::OB,VM::VM1_n,"Markup1",false },
  {0x0029,0x0001,"Mitra Markup 1.0",VR::OB,VM::VM1_n,"Markup2",false },
  {0x0029,0x0002,"Mitra Markup 1.0",VR::OB,VM::VM1_n,"Markup3",false },
  {0x0029,0x0003,"Mitra Markup 1.0",VR::OB,VM::VM1_n,"Markup4",false },
  {0x0029,0x0004,"Mitra Markup 1.0",VR::OB,VM::VM1_n,"Markup5",false },
  {0x0029,0x0005,"Mitra Markup 1.0",VR::OB,VM::VM1_n,"Markup6",false },
  {0x0029,0x0006,"Mitra Markup 1.0",VR::OB,VM::VM1_n,"Markup7",false },
  {0x0029,0x0007,"Mitra Markup 1.0",VR::OB,VM::VM1_n,"Markup8",false },
  {0x0029,0x0008,"Mitra Markup 1.0",VR::OB,VM::VM1_n,"Markup9",false },
  {0x0029,0x0009,"Mitra Markup 1.0",VR::OB,VM::VM1_n,"Markup10",false },
  {0x0029,0x0010,"Mitra Markup 1.0",VR::OB,VM::VM1_n,"Markup11",false },
  {0x0029,0x0011,"Mitra Markup 1.0",VR::OB,VM::VM1_n,"Markup12",false },
  {0x0029,0x0012,"Mitra Markup 1.0",VR::OB,VM::VM1,"Markup13",false },
  {0x0029,0x0013,"Mitra Markup 1.0",VR::OB,VM::VM1,"Markup14",false },
  {0x0029,0x0014,"Mitra Markup 1.0",VR::OB,VM::VM1,"Markup15",false },
  {0x0029,0x0000,"Mitra Object Document 1.0",VR::OB,VM::VM1,"IMPAX object document",false },
  {0x0029,0x0010,"Mitra Object Document 1.0",VR::OB,VM::VM1,"IMPAX markup XML stored",false },
  {0x1455,0x0000,"Mortara_Inc",VR::OW,VM::VM1,"ELI Interpretation Vector",false },
  {0x1455,0x0001,"Mortara_Inc",VR::UN,VM::VM1,"Custom ID",false },
  {0x1455,0x0002,"Mortara_Inc",VR::UT,VM::VM1,"Race",false },
  {0x1455,0x0003,"Mortara_Inc",VR::UT,VM::VM1,"Social Security Number",false },
  {0x1455,0x0004,"Mortara_Inc",VR::UT,VM::VM1,"Attending Physician",false },
  {0x1455,0x0005,"Mortara_Inc",VR::UT,VM::VM1,"Procedural Diagnosis",false },
  {0x1455,0x0006,"Mortara_Inc",VR::UT,VM::VM1,"Note1",false },
  {0x1455,0x0007,"Mortara_Inc",VR::UT,VM::VM1,"Note2",false },
  {0x1455,0x0008,"Mortara_Inc",VR::LO,VM::VM1,"Order Request Number",false },
  {0x1455,0x0010,"Mortara_Inc",VR::LO,VM::VM1,"Manufacturer Name",false },
  {0x7777,0x0002,"NUD_PRIVATE",VR::UT,VM::VM1,"Interfile",false },
  {0x7777,0x0005,"NUD_PRIVATE",VR::IS,VM::VM1,"?",false },
  {0x8101,0x0000,"ObjectModel (ID, Version, Place, PlaceDescription)",VR::LO,VM::VM1,"ID",false },
  {0x8101,0x0010,"ObjectModel (ID, Version, Place, PlaceDescription)",VR::LO,VM::VM1,"Version",false },
  {0x0009,0x0000,"PAPYRUS",VR::LT,VM::VM1,"Original File Name",false },
  {0x0009,0x0010,"PAPYRUS",VR::LT,VM::VM1,"Original File Location",false },
  {0x0009,0x0018,"PAPYRUS",VR::LT,VM::VM1,"Data Set Identifier",false },
  {0x0041,0x0000,"PAPYRUS",VR::LT,VM::VM1_n,"Papyrus Comments",false },
  {0x0041,0x0010,"PAPYRUS",VR::US,VM::VM1,"Folder Type",false },
  {0x0041,0x0011,"PAPYRUS",VR::LT,VM::VM1,"Patient Folder Data Set ID",false },
  {0x0041,0x0020,"PAPYRUS",VR::LT,VM::VM1,"Folder Name",false },
  {0x0041,0x0030,"PAPYRUS",VR::DA,VM::VM1,"Creation Date",false },
  {0x0041,0x0032,"PAPYRUS",VR::TM,VM::VM1,"Creation Time",false },
  {0x0041,0x0034,"PAPYRUS",VR::DA,VM::VM1,"Modified Date",false },
  {0x0041,0x0036,"PAPYRUS",VR::TM,VM::VM1,"Modified Time",false },
  {0x0041,0x0040,"PAPYRUS",VR::LT,VM::VM1_n,"Owner Name",false },
  {0x0041,0x0050,"PAPYRUS",VR::LT,VM::VM1,"Folder Status",false },
  {0x0041,0x0060,"PAPYRUS",VR::UL,VM::VM1,"Number of Images",false },
  {0x0041,0x0062,"PAPYRUS",VR::UL,VM::VM1,"Number of Other",false },
  {0x0041,0x00a0,"PAPYRUS",VR::LT,VM::VM1_n,"External Folder Element DSID",false },
  {0x0041,0x00a1,"PAPYRUS",VR::US,VM::VM1_n,"External Folder Element Data Set Type",false },
  {0x0041,0x00a2,"PAPYRUS",VR::LT,VM::VM1_n,"External Folder Element File Location",false },
  {0x0041,0x00a3,"PAPYRUS",VR::UL,VM::VM1_n,"External Folder Element Length",false },
  {0x0041,0x00b0,"PAPYRUS",VR::LT,VM::VM1_n,"Internal Folder Element DSID",false },
  {0x0041,0x00b1,"PAPYRUS",VR::US,VM::VM1_n,"Internal Folder Element Data Set Type",false },
  {0x0041,0x00b2,"PAPYRUS",VR::UL,VM::VM1_n,"Internal Offset To Data Set",false },
  {0x0041,0x00b3,"PAPYRUS",VR::UL,VM::VM1_n,"Internal Offset To Image",false },
  {0x0041,0x0000,"PAPYRUS 3.0",VR::LT,VM::VM1,"Comments",false },
  {0x0041,0x0010,"PAPYRUS 3.0",VR::SQ,VM::VM1,"Pointer Sequence",false },
  {0x0041,0x0011,"PAPYRUS 3.0",VR::UL,VM::VM1,"Image Pointer",false },
  {0x0041,0x0012,"PAPYRUS 3.0",VR::UL,VM::VM1,"Pixel Offset",false },
  {0x0041,0x0013,"PAPYRUS 3.0",VR::SQ,VM::VM1,"Image Identifier Sequence",false },
  {0x0041,0x0014,"PAPYRUS 3.0",VR::SQ,VM::VM1,"External PAPYRUS-File Reference Sequence",false },
  {0x0041,0x0015,"PAPYRUS 3.0",VR::US,VM::VM1,"Number of images",false },
  {0x0041,0x0021,"PAPYRUS 3.0",VR::UI,VM::VM1,"Referenced SOP Class UID",false },
  {0x0041,0x0022,"PAPYRUS 3.0",VR::UI,VM::VM1,"Referenced SOP Instance UID",false },
  {0x0041,0x0031,"PAPYRUS 3.0",VR::LO,VM::VM1,"Referenced File Name",false },
  {0x0041,0x0032,"PAPYRUS 3.0",VR::LO,VM::VM1_n,"Referenced File Path",false },
  {0x0041,0x0041,"PAPYRUS 3.0",VR::UI,VM::VM1,"Referenced Image SOP Class UID",false },
  {0x0041,0x0042,"PAPYRUS 3.0",VR::UI,VM::VM1,"Referenced Image SOP Instance UID",false },
  {0x0041,0x0050,"PAPYRUS 3.0",VR::SQ,VM::VM1,"Image Sequence",false },
  // UIN:
  {0x6001,0x0000,"PAPYRUS 3.0",VR::IS,VM::VM1,"Overlay ID",false },
  {0x6001,0x0001,"PAPYRUS 3.0",VR::LT,VM::VM1_n,"Linked Overlays",false },
  {0x6001,0x0010,"PAPYRUS 3.0",VR::US,VM::VM1,"Overlays Rows",false },
  {0x6001,0x0011,"PAPYRUS 3.0",VR::US,VM::VM1,"Overlays Cols",false },
  {0x6001,0x0040,"PAPYRUS 3.0",VR::LO,VM::VM1,"Overlays Type",false },
  {0x6001,0x0050,"PAPYRUS 3.0",VR::US,VM::VM1,"Overlays Origin",false },
  {0x6001,0x0060,"PAPYRUS 3.0",VR::LO,VM::VM1,"Editable",false },
  {0x6001,0x0070,"PAPYRUS 3.0",VR::LO,VM::VM1,"Overlay Font",false },
  {0x6001,0x0072,"PAPYRUS 3.0",VR::LO,VM::VM1,"Overlay Style",false },
  {0x6001,0x0074,"PAPYRUS 3.0",VR::US,VM::VM1,"Overlay Font Size",false },
  {0x6001,0x0076,"PAPYRUS 3.0",VR::LO,VM::VM1,"Overlay Color",false },
  {0x6001,0x0078,"PAPYRUS 3.0",VR::US,VM::VM1,"Shadow Size",false },
  {0x6001,0x0080,"PAPYRUS 3.0",VR::LO,VM::VM1,"Fill Pattern",false },
  {0x6001,0x0082,"PAPYRUS 3.0",VR::US,VM::VM1,"Overlay Pen Size",false },
  {0x6001,0x00A0,"PAPYRUS 3.0",VR::LO,VM::VM1,"Label",false },
  {0x6001,0x00A2,"PAPYRUS 3.0",VR::LT,VM::VM1,"Post It Text",false },
  {0x6001,0x00A4,"PAPYRUS 3.0",VR::US,VM::VM1,"Anchor Point",false },
  {0x6001,0x00b0,"PAPYRUS 3.0",VR::LO,VM::VM1,"ROI Type",false },
  {0x6001,0x00b2,"PAPYRUS 3.0",VR::LT,VM::VM1,"Attached Annotation",false },
  {0x6001,0x00ba,"PAPYRUS 3.0",VR::US,VM::VM1_n,"Contour Points",false },
  {0x6001,0x00bc,"PAPYRUS 3.0",VR::US,VM::VM1_n,"Mask Data",false },
  {0x6001,0x00c0,"PAPYRUS 3.0",VR::SQ,VM::VM1,"UIN Overlay Sequence",false },
  {0x0009,0x0010,"PHILIPS MR",VR::LO,VM::VM1,"SPI Release",false },
  {0x0009,0x0012,"PHILIPS MR",VR::DA,VM::VM1,"date",false },
  {0x0019,0x0000,"PHILIPS MR R5.5/PART",VR::DS,VM::VM1,"Field of View",false },
  {0x0019,0x0000,"PHILIPS MR R5.6/PART",VR::DS,VM::VM1,"Field of View",false },
  {0x0019,0x0001,"PHILIPS MR SPECTRO;1",VR::US,VM::VM1,"?",false },
  {0x0019,0x0002,"PHILIPS MR SPECTRO;1",VR::US,VM::VM1,"?",false },
  {0x0019,0x0003,"PHILIPS MR SPECTRO;1",VR::US,VM::VM1,"?",false },
  {0x0019,0x0004,"PHILIPS MR SPECTRO;1",VR::US,VM::VM1,"?",false },
  {0x0019,0x0005,"PHILIPS MR SPECTRO;1",VR::US,VM::VM1,"?",false },
  {0x0019,0x0006,"PHILIPS MR SPECTRO;1",VR::US,VM::VM1,"?",false },
  {0x0019,0x0007,"PHILIPS MR SPECTRO;1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0008,"PHILIPS MR SPECTRO;1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0009,"PHILIPS MR SPECTRO;1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0010,"PHILIPS MR SPECTRO;1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0012,"PHILIPS MR SPECTRO;1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0013,"PHILIPS MR SPECTRO;1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0014,"PHILIPS MR SPECTRO;1",VR::US,VM::VM1,"?",false },
  {0x0019,0x0015,"PHILIPS MR SPECTRO;1",VR::US,VM::VM1_n,"?",false },
  {0x0019,0x0016,"PHILIPS MR SPECTRO;1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0017,"PHILIPS MR SPECTRO;1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0018,"PHILIPS MR SPECTRO;1",VR::UN,VM::VM1,"?",false },
  {0x0019,0x0020,"PHILIPS MR SPECTRO;1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0021,"PHILIPS MR SPECTRO;1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0022,"PHILIPS MR SPECTRO;1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0023,"PHILIPS MR SPECTRO;1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0024,"PHILIPS MR SPECTRO;1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0025,"PHILIPS MR SPECTRO;1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0026,"PHILIPS MR SPECTRO;1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0027,"PHILIPS MR SPECTRO;1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0028,"PHILIPS MR SPECTRO;1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0029,"PHILIPS MR SPECTRO;1",VR::IS,VM::VM1_n,"?",false },
  {0x0019,0x0031,"PHILIPS MR SPECTRO;1",VR::US,VM::VM1,"?",false },
  {0x0019,0x0032,"PHILIPS MR SPECTRO;1",VR::US,VM::VM1,"?",false },
  {0x0019,0x0041,"PHILIPS MR SPECTRO;1",VR::LT,VM::VM1,"?",false },
  {0x0019,0x0042,"PHILIPS MR SPECTRO;1",VR::IS,VM::VM2,"?",false },
  {0x0019,0x0043,"PHILIPS MR SPECTRO;1",VR::IS,VM::VM2,"?",false },
  {0x0019,0x0045,"PHILIPS MR SPECTRO;1",VR::US,VM::VM1,"?",false },
  {0x0019,0x0046,"PHILIPS MR SPECTRO;1",VR::US,VM::VM1,"?",false },
  {0x0019,0x0047,"PHILIPS MR SPECTRO;1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0048,"PHILIPS MR SPECTRO;1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0049,"PHILIPS MR SPECTRO;1",VR::US,VM::VM1,"?",false },
  {0x0019,0x0050,"PHILIPS MR SPECTRO;1",VR::UN,VM::VM1,"?",false },
  {0x0019,0x0060,"PHILIPS MR SPECTRO;1",VR::US,VM::VM1,"?",false },
  {0x0019,0x0061,"PHILIPS MR SPECTRO;1",VR::US,VM::VM1,"?",false },
  {0x0019,0x0070,"PHILIPS MR SPECTRO;1",VR::UN,VM::VM1,"?",false },
  {0x0019,0x0071,"PHILIPS MR SPECTRO;1",VR::IS,VM::VM1_n,"?",false },
  {0x0019,0x0072,"PHILIPS MR SPECTRO;1",VR::US,VM::VM1,"?",false },
  {0x0019,0x0073,"PHILIPS MR SPECTRO;1",VR::US,VM::VM1,"?",false },
  {0x0019,0x0074,"PHILIPS MR SPECTRO;1",VR::US,VM::VM1,"?",false },
  {0x0019,0x0076,"PHILIPS MR SPECTRO;1",VR::US,VM::VM1,"?",false },
  {0x0019,0x0077,"PHILIPS MR SPECTRO;1",VR::US,VM::VM1,"?",false },
  {0x0019,0x0078,"PHILIPS MR SPECTRO;1",VR::US,VM::VM1,"?",false },
  {0x0019,0x0079,"PHILIPS MR SPECTRO;1",VR::US,VM::VM1,"?",false },
  {0x0019,0x0080,"PHILIPS MR SPECTRO;1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0009,"PHILIPS MR/LAST",VR::DS,VM::VM1,"Main Magnetic Field",false },
  {0x0019,0x000e,"PHILIPS MR/LAST",VR::IS,VM::VM1,"Flow Compensation",false },
  {0x0019,0x00b1,"PHILIPS MR/LAST",VR::IS,VM::VM1,"Minimum RR Interval",false },
  {0x0019,0x00b2,"PHILIPS MR/LAST",VR::IS,VM::VM1,"Maximum RR Interval",false },
  {0x0019,0x00b3,"PHILIPS MR/LAST",VR::IS,VM::VM1,"Number of Rejections",false },
  {0x0019,0x00b4,"PHILIPS MR/LAST",VR::IS,VM::VM1_n,"Number of RR Intervals",false },
  {0x0019,0x00b5,"PHILIPS MR/LAST",VR::IS,VM::VM1,"Arrhythmia Rejection",false },
  {0x0019,0x00b7,"PHILIPS MR/LAST",VR::IS,VM::VM1,"?",false },
  {0x0019,0x00c0,"PHILIPS MR/LAST",VR::DS,VM::VM1_n,"?",false },
  {0x0019,0x00c6,"PHILIPS MR/LAST",VR::IS,VM::VM1,"Cycled Multiple Slice",false },
  {0x0019,0x00ce,"PHILIPS MR/LAST",VR::IS,VM::VM1,"REST",false },
  {0x0019,0x00d5,"PHILIPS MR/LAST",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00d6,"PHILIPS MR/LAST",VR::IS,VM::VM1,"Fourier Interpolation",false },
  {0x0019,0x00d9,"PHILIPS MR/LAST",VR::IS,VM::VM1_n,"?",false },
  {0x0019,0x00e0,"PHILIPS MR/LAST",VR::IS,VM::VM1,"Prepulse",false },
  {0x0019,0x00e1,"PHILIPS MR/LAST",VR::DS,VM::VM1,"Prepulse Delay",false },
  {0x0019,0x00e2,"PHILIPS MR/LAST",VR::IS,VM::VM1,"?",false },
  {0x0019,0x00e3,"PHILIPS MR/LAST",VR::DS,VM::VM3,"?",false },
  {0x0019,0x00e4,"PHILIPS MR/LAST",VR::IS,VM::VM1,"?",false },
  {0x0019,0x00e5,"PHILIPS MR/LAST",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00f0,"PHILIPS MR/LAST",VR::LT,VM::VM1,"WS Protocol String 1",false },
  {0x0019,0x00f1,"PHILIPS MR/LAST",VR::LT,VM::VM1,"WS Protocol String 2",false },
  {0x0019,0x00f2,"PHILIPS MR/LAST",VR::LT,VM::VM1,"WS Protocol String 3",false },
  {0x0019,0x00f3,"PHILIPS MR/LAST",VR::LT,VM::VM1,"WS Protocol String 4",false },
  {0x0021,0x0000,"PHILIPS MR/LAST",VR::IS,VM::VM1,"?",false },
  {0x0021,0x0010,"PHILIPS MR/LAST",VR::IS,VM::VM1,"?",false },
  {0x0021,0x0020,"PHILIPS MR/LAST",VR::IS,VM::VM1,"?",false },
  {0x0021,0x0021,"PHILIPS MR/LAST",VR::DS,VM::VM1,"Slice Gap",false },
  {0x0021,0x0022,"PHILIPS MR/LAST",VR::DS,VM::VM1,"Stack Radial Angle",false },
  {0x0027,0x0000,"PHILIPS MR/LAST",VR::US,VM::VM1,"?",false },
  {0x0027,0x0011,"PHILIPS MR/LAST",VR::US,VM::VM1_n,"?",false },
  {0x0027,0x0012,"PHILIPS MR/LAST",VR::DS,VM::VM1_n,"?",false },
  {0x0027,0x0013,"PHILIPS MR/LAST",VR::DS,VM::VM1_n,"?",false },
  {0x0027,0x0014,"PHILIPS MR/LAST",VR::DS,VM::VM1_n,"?",false },
  {0x0027,0x0015,"PHILIPS MR/LAST",VR::DS,VM::VM1_n,"?",false },
  {0x0027,0x0016,"PHILIPS MR/LAST",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0010,"PHILIPS MR/LAST",VR::DS,VM::VM1,"FP Min",false },
  {0x0029,0x0020,"PHILIPS MR/LAST",VR::DS,VM::VM1,"FP Max",false },
  {0x0029,0x0030,"PHILIPS MR/LAST",VR::DS,VM::VM1,"Scaled Minimum",false },
  {0x0029,0x0040,"PHILIPS MR/LAST",VR::DS,VM::VM1,"Scaled Maximum",false },
  {0x0029,0x0050,"PHILIPS MR/LAST",VR::DS,VM::VM1,"Window Minimum",false },
  {0x0029,0x0060,"PHILIPS MR/LAST",VR::DS,VM::VM1,"Window Maximum",false },
  {0x0029,0x0061,"PHILIPS MR/LAST",VR::IS,VM::VM1,"?",false },
  {0x0029,0x0062,"PHILIPS MR/LAST",VR::IS,VM::VM1,"?",false },
  {0x0029,0x0070,"PHILIPS MR/LAST",VR::DS,VM::VM1,"?",false },
  {0x0029,0x0071,"PHILIPS MR/LAST",VR::DS,VM::VM1,"?",false },
  {0x0029,0x0072,"PHILIPS MR/LAST",VR::IS,VM::VM1,"?",false },
  {0x0029,0x0080,"PHILIPS MR/LAST",VR::IS,VM::VM2,"View Center",false },
  {0x0029,0x0081,"PHILIPS MR/LAST",VR::IS,VM::VM2,"View Size",false },
  {0x0029,0x0082,"PHILIPS MR/LAST",VR::IS,VM::VM1,"View Zoom",false },
  {0x0029,0x0083,"PHILIPS MR/LAST",VR::IS,VM::VM1,"View Transform",false },
  {0x0041,0x0007,"PHILIPS MR/LAST",VR::LO,VM::VM1,"?",false },
  {0x0041,0x0009,"PHILIPS MR/LAST",VR::DS,VM::VM1,"?",false },
  {0x6001,0x0000,"PHILIPS MR/LAST",VR::LT,VM::VM1,"?",false },
  {0x0019,0x0000,"PHILIPS MR/PART",VR::DS,VM::VM1_2,"Field of View",false },
  {0x0019,0x0001,"PHILIPS MR/PART",VR::DS,VM::VM1_n,"Stack Type",false },
  {0x0019,0x0002,"PHILIPS MR/PART",VR::DS,VM::VM1_n,"?",false },
  {0x0019,0x0003,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0005,"PHILIPS MR/PART",VR::DS,VM::VM1,"CC Angulation",false },
  {0x0019,0x0006,"PHILIPS MR/PART",VR::DS,VM::VM1,"AP Angulation",false },
  {0x0019,0x0007,"PHILIPS MR/PART",VR::DS,VM::VM1,"LR Angulation",false },
  {0x0019,0x0008,"PHILIPS MR/PART",VR::IS,VM::VM1,"Patient Orientation 1",false },
  {0x0019,0x0009,"PHILIPS MR/PART",VR::IS,VM::VM1,"Patient Orientation",false },
  {0x0019,0x000a,"PHILIPS MR/PART",VR::IS,VM::VM1,"Slice Orientation",false },
  {0x0019,0x000b,"PHILIPS MR/PART",VR::DS,VM::VM1,"LR Offcenter",false },
  {0x0019,0x000c,"PHILIPS MR/PART",VR::DS,VM::VM1,"CC Offcenter",false },
  {0x0019,0x000d,"PHILIPS MR/PART",VR::DS,VM::VM1,"AP Offcenter",false },
  {0x0019,0x000e,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0019,0x000f,"PHILIPS MR/PART",VR::IS,VM::VM1,"Number of Slices",false },
  {0x0019,0x0010,"PHILIPS MR/PART",VR::DS,VM::VM1,"Slice Factor",false },
  {0x0019,0x0011,"PHILIPS MR/PART",VR::DS,VM::VM1_n,"Echo Times",false },
  {0x0019,0x0014,"PHILIPS MR/PART",VR::CS,VM::VM1,"?",false },
  {0x0019,0x0015,"PHILIPS MR/PART",VR::IS,VM::VM1,"Dynamic Study",false },
  {0x0019,0x0018,"PHILIPS MR/PART",VR::DS,VM::VM1,"Heartbeat Interval",false },
  {0x0019,0x0019,"PHILIPS MR/PART",VR::DS,VM::VM1,"Repetition Time FFE",false },
  {0x0019,0x001a,"PHILIPS MR/PART",VR::DS,VM::VM1,"FFE Flip Angle",false },
  {0x0019,0x001b,"PHILIPS MR/PART",VR::IS,VM::VM1,"Number of Scans",false },
  {0x0019,0x001c,"PHILIPS MR/PART",VR::CS,VM::VM1,"?",false },
  {0x0019,0x001d,"PHILIPS MR/PART",VR::CS,VM::VM1,"?",false },
  {0x0019,0x001e,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0021,"PHILIPS MR/PART",VR::DS,VM::VM1_n,"?",false },
  {0x0019,0x0022,"PHILIPS MR/PART",VR::DS,VM::VM1,"Dynamic Scan Time Begin",false },
  {0x0019,0x0023,"PHILIPS MR/PART",VR::DS,VM::VM1_n,"?",false },
  {0x0019,0x0024,"PHILIPS MR/PART",VR::IS,VM::VM1_n,"?",false },
  {0x0019,0x0025,"PHILIPS MR/PART",VR::DS,VM::VM1_n,"?",false },
  {0x0019,0x0026,"PHILIPS MR/PART",VR::DS,VM::VM1_n,"?",false },
  {0x0019,0x0027,"PHILIPS MR/PART",VR::DS,VM::VM1_n,"?",false },
  {0x0019,0x0028,"PHILIPS MR/PART",VR::LO,VM::VM1_n,"?",false },
  {0x0019,0x0029,"PHILIPS MR/PART",VR::DS,VM::VM1_n,"?",false },
  {0x0019,0x0030,"PHILIPS MR/PART",VR::LO,VM::VM1_n,"?",false },
  {0x0019,0x0031,"PHILIPS MR/PART",VR::DS,VM::VM1_n,"?",false },
  {0x0019,0x0040,"PHILIPS MR/PART",VR::US,VM::VM1,"?",false },
  {0x0019,0x0045,"PHILIPS MR/PART",VR::IS,VM::VM1,"Reconstruction Resolution",false },
  {0x0019,0x0050,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0051,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0052,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0053,"PHILIPS MR/PART",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0054,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0055,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0056,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0057,"PHILIPS MR/PART",VR::US,VM::VM1,"?",false },
  {0x0019,0x0058,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0059,"PHILIPS MR/PART",VR::US,VM::VM1,"?",false },
  {0x0019,0x0060,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0061,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0062,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0063,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0064,"PHILIPS MR/PART",VR::DS,VM::VM1,"Repetition Time SE",false },
  {0x0019,0x0065,"PHILIPS MR/PART",VR::DS,VM::VM1,"Repetition Time IR",false },
  {0x0019,0x0066,"PHILIPS MR/PART",VR::US,VM::VM1,"?",false },
  {0x0019,0x0067,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0069,"PHILIPS MR/PART",VR::IS,VM::VM1,"Number of Phases",false },
  {0x0019,0x006a,"PHILIPS MR/PART",VR::IS,VM::VM1,"Cardiac Frequency",false },
  {0x0019,0x006b,"PHILIPS MR/PART",VR::DS,VM::VM1,"Inversion Delay",false },
  {0x0019,0x006c,"PHILIPS MR/PART",VR::DS,VM::VM1,"Gate Delay",false },
  {0x0019,0x006d,"PHILIPS MR/PART",VR::DS,VM::VM1,"Gate Width",false },
  {0x0019,0x006e,"PHILIPS MR/PART",VR::DS,VM::VM1,"Trigger Delay Time",false },
  {0x0019,0x0070,"PHILIPS MR/PART",VR::DS,VM::VM1_n,"?",false },
  {0x0019,0x0080,"PHILIPS MR/PART",VR::IS,VM::VM1,"Number of Chemical Shifts",false },
  {0x0019,0x0081,"PHILIPS MR/PART",VR::DS,VM::VM1,"Chemical Shift",false },
  {0x0019,0x0084,"PHILIPS MR/PART",VR::IS,VM::VM1,"Number of Rows",false },
  {0x0019,0x0085,"PHILIPS MR/PART",VR::IS,VM::VM1,"Number of Samples",false },
  {0x0019,0x008a,"PHILIPS MR/PART",VR::CS,VM::VM1,"?",false },
  {0x0019,0x008b,"PHILIPS MR/PART",VR::CS,VM::VM1,"?",false },
  {0x0019,0x008c,"PHILIPS MR/PART",VR::CS,VM::VM1,"?",false },
  {0x0019,0x008d,"PHILIPS MR/PART",VR::CS,VM::VM1,"?",false },
  {0x0019,0x008e,"PHILIPS MR/PART",VR::CS,VM::VM1,"?",false },
  {0x0019,0x008f,"PHILIPS MR/PART",VR::CS,VM::VM1,"?",false },
  {0x0019,0x0090,"PHILIPS MR/PART",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0091,"PHILIPS MR/PART",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0092,"PHILIPS MR/PART",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0093,"PHILIPS MR/PART",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0094,"PHILIPS MR/PART",VR::LO,VM::VM1,"Magnetization Transfer Contrast",false },
  {0x0019,0x0095,"PHILIPS MR/PART",VR::LO,VM::VM1,"Spectral Presaturation With Inversion Recovery",false },
  {0x0019,0x0096,"PHILIPS MR/PART",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0097,"PHILIPS MR/PART",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0099,"PHILIPS MR/PART",VR::LO,VM::VM1,"?",false },
  {0x0019,0x00a0,"PHILIPS MR/PART",VR::IS,VM::VM1,"?",false },
  {0x0019,0x00a1,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00a3,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00a4,"PHILIPS MR/PART",VR::CS,VM::VM1,"?",false },
  {0x0019,0x00b4,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00b5,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00b6,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00c0,"PHILIPS MR/PART",VR::DS,VM::VM1,"Trigger Delay Times",false },
  {0x0019,0x00c8,"PHILIPS MR/PART",VR::IS,VM::VM1,"?",false },
  {0x0019,0x00c9,"PHILIPS MR/PART",VR::IS,VM::VM1,"Foldover Direction Transverse",false },
  {0x0019,0x00ca,"PHILIPS MR/PART",VR::IS,VM::VM1,"Foldover Direction Sagittal",false },
  {0x0019,0x00cb,"PHILIPS MR/PART",VR::IS,VM::VM1,"Foldover Direction Coronal",false },
  {0x0019,0x00cc,"PHILIPS MR/PART",VR::IS,VM::VM1,"?",false },
  {0x0019,0x00cd,"PHILIPS MR/PART",VR::IS,VM::VM1,"?",false },
  {0x0019,0x00ce,"PHILIPS MR/PART",VR::IS,VM::VM1,"?",false },
  {0x0019,0x00cf,"PHILIPS MR/PART",VR::IS,VM::VM1,"Number of Echoes",false },
  {0x0019,0x00d0,"PHILIPS MR/PART",VR::IS,VM::VM1,"Scan Resolution",false },
  {0x0019,0x00d1,"PHILIPS MR/PART",VR::US,VM::VM1,"?",false },
  {0x0019,0x00d2,"PHILIPS MR/PART",VR::LO,VM::VM2,"Water Fat Shift",false },
  {0x0019,0x00d3,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00d4,"PHILIPS MR/PART",VR::IS,VM::VM1,"Artifact Reduction",false },
  {0x0019,0x00d5,"PHILIPS MR/PART",VR::IS,VM::VM1,"?",false },
  {0x0019,0x00d6,"PHILIPS MR/PART",VR::IS,VM::VM1,"?",false },
  {0x0019,0x00d7,"PHILIPS MR/PART",VR::DS,VM::VM1,"Scan Percentage",false },
  {0x0019,0x00d8,"PHILIPS MR/PART",VR::IS,VM::VM1,"Halfscan",false },
  {0x0019,0x00d9,"PHILIPS MR/PART",VR::IS,VM::VM1,"EPI Factor",false },
  {0x0019,0x00da,"PHILIPS MR/PART",VR::IS,VM::VM1,"Turbo Factor",false },
  {0x0019,0x00db,"PHILIPS MR/PART",VR::IS,VM::VM1,"?",false },
  {0x0019,0x00e0,"PHILIPS MR/PART",VR::IS,VM::VM1,"Percentage of Scan Completed",false },
  {0x0019,0x00e1,"PHILIPS MR/PART",VR::IS,VM::VM1,"Prepulse Delay",false },
  {0x0019,0x00e3,"PHILIPS MR/PART",VR::DS,VM::VM1,"Phase Contrast Velocity",false },
  {0x0019,0x00f0,"PHILIPS MR/PART",VR::IS,VM::VM1,"?",false },
  {0x0019,0x00f6,"PHILIPS MR/PART",VR::CS,VM::VM1,"?",false },
  {0x0019,0x00f7,"PHILIPS MR/PART",VR::CS,VM::VM1,"?",false },
  {0x0019,0x00f8,"PHILIPS MR/PART",VR::CS,VM::VM1,"?",false },
  {0x0019,0x00f9,"PHILIPS MR/PART",VR::CS,VM::VM1,"?",false },
  {0x0019,0x00fa,"PHILIPS MR/PART",VR::CS,VM::VM1,"?",false },
  {0x0019,0x00fb,"PHILIPS MR/PART",VR::CS,VM::VM1,"?",false },
  {0x0019,0x00fc,"PHILIPS MR/PART",VR::CS,VM::VM1,"Resonance Frequency",false },
  {0x0019,0x00fe,"PHILIPS MR/PART",VR::CS,VM::VM1,"Resonance Frequency",false },
  {0x0021,0x0000,"PHILIPS MR/PART",VR::DA,VM::VM1,"Scan Date",false },
  {0x0021,0x0006,"PHILIPS MR/PART",VR::LO,VM::VM1,"?",false },
  {0x0021,0x0007,"PHILIPS MR/PART",VR::DT,VM::VM1,"?",false },
  {0x0021,0x0008,"PHILIPS MR/PART",VR::LO,VM::VM1,"?",false },
  {0x0021,0x0009,"PHILIPS MR/PART",VR::CS,VM::VM1,"?",false },
  {0x0021,0x000a,"PHILIPS MR/PART",VR::CS,VM::VM1,"?",false },
  {0x0021,0x000f,"PHILIPS MR/PART",VR::LO,VM::VM1,"?",false },
  {0x0021,0x0010,"PHILIPS MR/PART",VR::IS,VM::VM1,"Image Type",false },
  {0x0021,0x0013,"PHILIPS MR/PART",VR::CS,VM::VM1,"?",false },
  {0x0021,0x0014,"PHILIPS MR/PART",VR::LO,VM::VM1,"?",false },
  {0x0021,0x0015,"PHILIPS MR/PART",VR::US,VM::VM1,"?",false },
  {0x0021,0x0020,"PHILIPS MR/PART",VR::IS,VM::VM1,"Slice Number",false },
  {0x0021,0x0021,"PHILIPS MR/PART",VR::IS,VM::VM1,"Slice Gap",false },
  {0x0021,0x0030,"PHILIPS MR/PART",VR::IS,VM::VM1,"Echo Number",false },
  {0x0021,0x0031,"PHILIPS MR/PART",VR::DS,VM::VM1,"Patient Reference ID",false },
  {0x0021,0x0035,"PHILIPS MR/PART",VR::IS,VM::VM1,"Chemical Shift Number",false },
  {0x0021,0x0040,"PHILIPS MR/PART",VR::IS,VM::VM1,"Phase Number",false },
  {0x0021,0x0050,"PHILIPS MR/PART",VR::IS,VM::VM1,"Dynamic Scan Number",false },
  {0x0021,0x0060,"PHILIPS MR/PART",VR::IS,VM::VM1,"Number of Rows In Object",false },
  {0x0021,0x0061,"PHILIPS MR/PART",VR::IS,VM::VM1_n,"Row Number",false },
  {0x0021,0x0062,"PHILIPS MR/PART",VR::IS,VM::VM1_n,"?",false },
  {0x0029,0x0000,"PHILIPS MR/PART",VR::DS,VM::VM2,"?",false },
  {0x0029,0x0004,"PHILIPS MR/PART",VR::US,VM::VM1,"?",false },
  {0x0029,0x0010,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0029,0x0011,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0029,0x0020,"PHILIPS MR/PART",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0031,"PHILIPS MR/PART",VR::DS,VM::VM2,"?",false },
  {0x0029,0x0032,"PHILIPS MR/PART",VR::DS,VM::VM2,"?",false },
  {0x0029,0x0050,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0029,0x0051,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0029,0x0052,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0029,0x0053,"PHILIPS MR/PART",VR::DS,VM::VM1,"?",false },
  {0x0029,0x00c3,"PHILIPS MR/PART",VR::IS,VM::VM1,"Scan Resolution",false },
  {0x0029,0x00c4,"PHILIPS MR/PART",VR::IS,VM::VM1,"Field of View",false },
  {0x0029,0x00d5,"PHILIPS MR/PART",VR::LT,VM::VM1,"Slice Thickness",false },
  {0x0009,0x0010,"PHILIPS MR/PART 12",VR::US,VM::VM1,"?",false },
  {0x0019,0x0010,"PHILIPS MR/PART 6",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0000,"PHILIPS MR/PART 7",VR::IS,VM::VM1,"?",false },
  {0x7051,0x0000,"PHILIPS NM -Private",VR::US,VM::VM1,"Current Segment",false },
  {0x7051,0x0001,"PHILIPS NM -Private",VR::US,VM::VM1,"Number of Segments",false },
  {0x7051,0x0002,"PHILIPS NM -Private",VR::FL,VM::VM1,"Segment Start Position",false },
  {0x7051,0x0003,"PHILIPS NM -Private",VR::FL,VM::VM1,"Segment Stop Position",false },
  {0x7051,0x0004,"PHILIPS NM -Private",VR::FL,VM::VM1,"Rel. COR offset - X dir.",false },
  {0x7051,0x0005,"PHILIPS NM -Private",VR::FL,VM::VM1,"Rel. COR offset - Z dir.",false },
  {0x7051,0x0006,"PHILIPS NM -Private",VR::US,VM::VM1,"Current Rotation Number",false },
  {0x7051,0x0007,"PHILIPS NM -Private",VR::US,VM::VM1,"Number of Rotations",false },
  {0x7051,0x0010,"PHILIPS NM -Private",VR::DS,VM::VM3,"Alignment Translations",false },
  {0x7051,0x0011,"PHILIPS NM -Private",VR::DS,VM::VM3,"Alignment Rotations",false },
  {0x7051,0x0012,"PHILIPS NM -Private",VR::DS,VM::VM1,"Alignment Timestamp",false },
  {0x7051,0x0015,"PHILIPS NM -Private",VR::UI,VM::VM1,"Related Xray Series Instance UID",false },
  {0x7051,0x0016,"PHILIPS NM -Private",VR::UI,VM::VM1,"?UID?",false },
  {0x7051,0x0025,"PHILIPS NM -Private",VR::LO,VM::VM1,"??",false },
  {0x7051,0x0026,"PHILIPS NM -Private",VR::DS,VM::VM1,"??",false },
  {0x7051,0x0027,"PHILIPS NM -Private",VR::DS,VM::VM1,"??",false },
  {0x7051,0x0028,"PHILIPS NM -Private",VR::IS,VM::VM1,"??",false },
  {0x7051,0x0029,"PHILIPS NM -Private",VR::IS,VM::VM1,"??",false },
  {0x7051,0x0001,"PHILIPS XCT -Private",VR::DS,VM::VM1,"Attenuation threshold",false },
  {0x7051,0x0002,"PHILIPS XCT -Private",VR::DS,VM::VM1,"DLP Estimate",false },
  {0x7051,0x0003,"PHILIPS XCT -Private",VR::LO,VM::VM1,"??",false },
  {0x5503,0x0012,"PHILIPS UNDOCUMENTED",VR::SL,VM::VM1,"Slice Thickness",false },
  {0x5503,0x0024,"PHILIPS UNDOCUMENTED",VR::UT,VM::VM1,"Plane",false },
  {0x5503,0x0030,"PHILIPS UNDOCUMENTED",VR::SS,VM::VM1,"Number of Slices",false },
  {0x5503,0x00a5,"PHILIPS UNDOCUMENTED",VR::UT,VM::VM1,"Prepulse",false },
  {0x5503,0x00ae,"PHILIPS UNDOCUMENTED",VR::SS,VM::VM1,"Contrast Agent Flag",false },
  {0x5503,0x00bb,"PHILIPS UNDOCUMENTED",VR::SL,VM::VM1,"Coil ID",false },
  {0x5503,0x00c2,"PHILIPS UNDOCUMENTED",VR::UT,VM::VM1,"Anatomy",false },
  {0x5503,0x00c3,"PHILIPS UNDOCUMENTED",VR::UT,VM::VM1,"Indication",false },
  {0x2001,0x0001,"PHILIPS UNKNOWN",VR::UN,VM::VM1,"MRSeriesNrOfEchoes",false },
  {0x2001,0x0002,"PHILIPS UNKNOWN",VR::UN,VM::VM1,"MRSeriesScanningTechniqueDescription",false },
  {0x2001,0x0005,"PHILIPS UNKNOWN",VR::UN,VM::VM1,"StackSequence",false },
  {0x2001,0x0006,"PHILIPS UNKNOWN",VR::UN,VM::VM1,"Examination Source",false },
  {0x2001,0x0007,"PHILIPS UNKNOWN",VR::UN,VM::VM1,"MRSeriesAcquisitionNumber",false },
  {0x2001,0x0008,"PHILIPS UNKNOWN",VR::UN,VM::VM1,"MRSeriesNrOfDynamicScans",false },
  {0x2005,0x0002,"PHILIPS UNKNOWN",VR::UN,VM::VM1,"MRSeriesNrOfChemicalShifts",false },
  {0x2005,0x0003,"PHILIPS UNKNOWN",VR::UN,VM::VM1,"MRSeriesRepetitionTime",false },
  {0x2005,0x0036,"PHILIPS UNKNOWN",VR::UN,VM::VM1,"MRVolumeSelection",false },
  {0x0019,0x0011,"PHILIPS-MR-1",VR::IS,VM::VM1,"Chemical Shift Number",false },
  {0x0019,0x0012,"PHILIPS-MR-1",VR::IS,VM::VM1,"Phase Number (Philips)",false },
  {0x0021,0x0001,"PHILIPS-MR-1",VR::IS,VM::VM1,"Reconstruction Number",false },
  {0x0021,0x0002,"PHILIPS-MR-1",VR::IS,VM::VM1,"Slice Number",false },
  {0x0055,0x0001,"PMOD_1",VR::FD,VM::VM1_n,"Frame Start Times Vector",false },
  {0x0055,0x0002,"PMOD_1",VR::FD,VM::VM3_3n,"Frame Positions Vector",false },
  {0x0055,0x0003,"PMOD_1",VR::FD,VM::VM6_6n,"Frame Orientations Vector",false },
  {0x0055,0x0004,"PMOD_1",VR::FD,VM::VM1_n,"Frame Durations (ms) Vector",false },
  {0x0055,0x0005,"PMOD_1",VR::FD,VM::VM1_n,"Frame Rescale Slope Vector",false },
  {0x7fe1,0x0001,"PMOD_GENPET",VR::UT,VM::VM1,"Slices Names",false },
  {0x7fe1,0x0002,"PMOD_GENPET",VR::UT,VM::VM1,"Gene Codes",false },
  {0x7fe1,0x0003,"PMOD_GENPET",VR::UT,VM::VM1,"Gene Labels",false },
  {0x0089,0x0020,"PMS-THORA-5.1",VR::SQ,VM::VM1,"Stamp Image Sequence",false },
  {0x0029,0x0001,"PMTF INFORMATION DATA",VR::SQ,VM::VM1,"?DTI Seq?",false },
  {0x0029,0x0031,"PMTF INFORMATION DATA",VR::LO,VM::VM1,"PMTF Information 1",false },
  {0x0029,0x0032,"PMTF INFORMATION DATA",VR::UL,VM::VM1,"PMTF Information 2",false },
  {0x0029,0x0033,"PMTF INFORMATION DATA",VR::UL,VM::VM1,"PMTF Information 3",false },
  {0x0029,0x0034,"PMTF INFORMATION DATA",VR::CS,VM::VM1,"PMTF Information 4",false },
  {0x0029,0x0089,"PMTF INFORMATION DATA",VR::LO,VM::VM1,"?MRImage?",false },
  {0x0029,0x0090,"PMTF INFORMATION DATA",VR::OB,VM::VM1,"Reversed EVRLE DataSet",false },
  {0x7015,0x0073,"PMTF INFORMATION DATA",VR::SQ,VM::VM1,"?",false },
  {0x0009,0x0002,"POLYTRON-SMS 2.5",VR::UN,VM::VM1,"Private Data 1",false },
  {0x0009,0x0003,"POLYTRON-SMS 2.5",VR::UN,VM::VM1,"Private Data 2",false },
  {0x0009,0x0004,"POLYTRON-SMS 2.5",VR::UN,VM::VM1,"Private Data 3",false },
  {0x0009,0x0006,"POLYTRON-SMS 2.5",VR::UN,VM::VM1,"Private Data 3",false },
  {0x0009,0x0010,"POLYTRON-SMS 2.5",VR::UN,VM::VM1,"Private Data 4",false },
  {0x0009,0x0011,"POLYTRON-SMS 2.5",VR::UN,VM::VM1,"Private Data 5",false },
  {0x0009,0x0012,"POLYTRON-SMS 2.5",VR::UN,VM::VM1,"Private Data 6",false },
  {0x0089,0x0010,"POLYTRON-SMS 2.5",VR::UN,VM::VM1,"Private Data",false },
  {0x0021,0x0001,"PRIVATE ORIGINAL ATTRIBUTES",VR::SQ,VM::VM1,"Private Original Attribute Sequence",false },
  {0x0021,0x0005,"PRIVATE ORIGINAL ATTRIBUTES",VR::DT,VM::VM1,"Replacement Date and Time",false },
  {0x0021,0x0010,"PRIVATE ORIGINAL ATTRIBUTES",VR::LO,VM::VM1,"Replacement Originator Name",false },
  {0x0021,0x0012,"PRIVATE ORIGINAL ATTRIBUTES",VR::LO,VM::VM1,"Replacement Reason",false },
  {0x0021,0x0070,"PRIVATE ORIGINAL ATTRIBUTES",VR::PN,VM::VM1_n,"Original Operators'Name",false },
  {0x2007,0x00b4,"Philips EV Imaging DD 017",VR::FL,VM::VM2,"?",false },
  {0x2007,0x00ba,"Philips EV Imaging DD 017",VR::FD,VM::VM1,"?",false },
  {0x2007,0x00bb,"Philips EV Imaging DD 017",VR::FD,VM::VM1,"?",false },
  {0x2007,0x00bc,"Philips EV Imaging DD 017",VR::SL,VM::VM1,"?",false },
  {0x2007,0x00d0,"Philips EV Imaging DD 017",VR::FD,VM::VM1,"?",false },
  {0x2007,0x00d1,"Philips EV Imaging DD 017",VR::FD,VM::VM1,"?",false },
  {0x2007,0x00d2,"Philips EV Imaging DD 017",VR::SL,VM::VM1,"?",false },
  {0x2007,0x00c1,"Philips EV Imaging DD 017",VR::UI,VM::VM1,"? SOP Instance UID ?",false },
  {0x2007,0x00c2,"Philips EV Imaging DD 017",VR::DA,VM::VM1,"?",false },
  {0x2007,0x00c3,"Philips EV Imaging DD 017",VR::TM,VM::VM1,"?",false },
  {0x2007,0x00c4,"Philips EV Imaging DD 017",VR::UL,VM::VM1,"?",false },
  {0x2007,0x0000,"Philips EV Imaging DD 019",VR::LO,VM::VM1,"?",false },
  {0x2007,0x0001,"Philips EV Imaging DD 019",VR::UT,VM::VM1,"?",false },
  {0x2007,0x0004,"Philips EV Imaging DD 021",VR::SL,VM::VM1,"Vol Seg Edit Connectivity",false },
  {0x2007,0x0051,"Philips EV Imaging DD 021",VR::US_SS,VM::VM1,"?US or SS?",false },
  {0x2007,0x0000,"Philips EV Imaging DD 022",VR::ST,VM::VM1,"?",false },
  {0x2007,0x0001,"Philips EV Imaging DD 022",VR::LO,VM::VM2,"?",false },
  {0x2007,0x0012,"Philips EV Imaging DD 022",VR::LO,VM::VM1,"Volume Sequence Capture",false },
  {0x2007,0x0017,"Philips EV Imaging DD 022",VR::UI,VM::VM1,"Original SOP Instance",false },
  {0x2007,0x0033,"Philips EV Imaging DD 022",VR::LO,VM::VM1,"?",false },
  {0x2007,0x0034,"Philips EV Imaging DD 022",VR::CS,VM::VM1,"?",false },
  {0x7777,0x0001,"Philips EnVisor",VR::DA,VM::VM1,"Date 3D/PanView internal 'dataset' file was acquired",false },
  {0x7777,0x0002,"Philips EnVisor",VR::CS,VM::VM1,"Type of 3D or PanView",false },
  {0x7777,0x0003,"Philips EnVisor",VR::LO,VM::VM1,"Internal 3D/PanView software version number",false },
  {0x7777,0x0014,"Philips EnVisor",VR::LT,VM::VM1,"Private string",false },
  {0x2001,0x0001,"PHILIPS IMAGING DD 001",VR::FL,VM::VM1,"Chemical Shift",false },
  {0x2001,0x0002,"PHILIPS IMAGING DD 001",VR::IS,VM::VM1,"Chemical Shift Number MR",false },
  {0x2001,0x0003,"PHILIPS IMAGING DD 001",VR::FL,VM::VM1,"Diffusion B-Factor",false },
  {0x2001,0x0004,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Diffusion Direction",false },
  {0x2001,0x0005,"PHILIPS IMAGING DD 001",VR::SS,VM::VM1,"Graphic Annotation Parent ID",false },
  {0x2001,0x0006,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Image Enhanced",false },
  {0x2001,0x0007,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Image Type ED ES",false },
  {0x2001,0x0008,"PHILIPS IMAGING DD 001",VR::IS,VM::VM1,"Phase Number",false },
  {0x2001,0x0009,"PHILIPS IMAGING DD 001",VR::FL,VM::VM1,"MRImagePrepulseDelay",false },
  {0x2001,0x000a,"PHILIPS IMAGING DD 001",VR::IS,VM::VM1,"Slice Number MR (aka ImagePlaneNumber)",false },
  {0x2001,0x000b,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Slice Orientation (aka ImagePlaneOrientation)",false },
  {0x2001,0x000c,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"?MRSeriesArrhythmiaRejection",false },
  {0x2001,0x000e,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"?MRSeriesCardiacCycled",false },
  {0x2001,0x000f,"PHILIPS IMAGING DD 001",VR::SS,VM::VM1,"?MRSeriesCardiacGateWidth",false },
  {0x2001,0x0010,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Cardiac Sync",false },
  {0x2001,0x0011,"PHILIPS IMAGING DD 001",VR::FL,VM::VM1,"Diffusion Echo Time",false },
  {0x2001,0x0012,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Dynamic Series",false },
  {0x2001,0x0013,"PHILIPS IMAGING DD 001",VR::SL,VM::VM1,"EPI Factor",false },
  {0x2001,0x0014,"PHILIPS IMAGING DD 001",VR::SL,VM::VM1,"Number of Echoes",false },
  {0x2001,0x0015,"PHILIPS IMAGING DD 001",VR::SS,VM::VM1,"Number of Locations",false },
  {0x2001,0x0016,"PHILIPS IMAGING DD 001",VR::SS,VM::VM1,"Number of PC Directions",false },
  {0x2001,0x0017,"PHILIPS IMAGING DD 001",VR::SL,VM::VM1,"Number of Phases MR",false },
  {0x2001,0x0018,"PHILIPS IMAGING DD 001",VR::SL,VM::VM1,"Number of Slices MR",false },
  {0x2001,0x0019,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Partial Matrix Scanned",false },
  {0x2001,0x001a,"PHILIPS IMAGING DD 001",VR::FL,VM::VM1_n,"PC Velocity",false },
  {0x2001,0x001b,"PHILIPS IMAGING DD 001",VR::FL,VM::VM1,"Prepulse Delay",false },
  {0x2001,0x001c,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Prepulse Type",false },
  {0x2001,0x001d,"PHILIPS IMAGING DD 001",VR::IS,VM::VM1,"Reconstruction Number MR",false },
  {0x2001,0x001e,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"?MRSeriesReformatAccuracy",false },
  {0x2001,0x001f,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Respiration Sync",false },
  {0x2001,0x0020,"PHILIPS IMAGING DD 001",VR::LO,VM::VM1,"Scanning Technique Description MR",false },
  {0x2001,0x0021,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"SPIR",false },
  {0x2001,0x0022,"PHILIPS IMAGING DD 001",VR::FL,VM::VM1,"Water Fat Shift",false },
  {0x2001,0x0023,"PHILIPS IMAGING DD 001",VR::DS,VM::VM1,"Flip Angle Philips",false },
  {0x2001,0x0024,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Interactive",false },
  {0x2001,0x0025,"PHILIPS IMAGING DD 001",VR::SH,VM::VM1,"Echo Time Display MR",false },
  {0x2001,0x0026,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Presentation State Subtraction Active",false },
  {0x2001,0x0028,"PHILIPS IMAGING DD 001",VR::FL,VM::VM1,"Edge Enhancement Gain Factor Sub",false },
  {0x2001,0x0029,"PHILIPS IMAGING DD 001",VR::FL,VM::VM1,"Edge Enhancement Gain Factor Non Sub",false },
  {0x2001,0x002a,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Edge Enhancement Taste Adapt Sub",false },
  {0x2001,0x002b,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Edge Enhancement Taste Non Sub",false },
  {0x2001,0x002c,"PHILIPS IMAGING DD 001",VR::FL,VM::VM1,"Harmonization Factor",false },
  {0x2001,0x002d,"PHILIPS IMAGING DD 001",VR::SS,VM::VM1,"Number of Stack Slices",false },
  {0x2001,0x002f,"PHILIPS IMAGING DD 001",VR::FL,VM::VM1,"Harmonization Gain",false },
  {0x2001,0x0030,"PHILIPS IMAGING DD 001",VR::UL,VM::VM1,"Log Subtraction Gain Step",false },
  {0x2001,0x0031,"PHILIPS IMAGING DD 001",VR::US,VM::VM1,"Mixing NR of Mask Image Numbers",false },
  {0x2001,0x0032,"PHILIPS IMAGING DD 001",VR::FL,VM::VM1_n,"Stack Radial Angle",false },
  {0x2001,0x0033,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1_n,"Stack Radial Axis",false },
  {0x2001,0x0034,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Mixing Mask Operation",false },
  {0x2001,0x0035,"PHILIPS IMAGING DD 001",VR::SS,VM::VM1_n,"Stack Slice Number",false },
  {0x2001,0x0036,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1_n,"Stack Type",false },
  {0x2001,0x0037,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Mixing Operation Type",false },
  {0x2001,0x0039,"PHILIPS IMAGING DD 001",VR::FL,VM::VM1,"?",false },
  {0x2001,0x003a,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Pixel Shift",false },
  {0x2001,0x003d,"PHILIPS IMAGING DD 001",VR::UL,VM::VM1,"?ContourFillColor",false },
  {0x2001,0x003f,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Interpolation Method",false },
  {0x2001,0x0042,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Subtraction Land Marking Active",false },
  {0x2001,0x0046,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Graphic Line Style",false },
  {0x2001,0x0047,"PHILIPS IMAGING DD 001",VR::FL,VM::VM1,"Graphic Line Width",false },
  {0x2001,0x0048,"PHILIPS IMAGING DD 001",VR::SS,VM::VM1,"Graphic Annotation ID",false },
  {0x2001,0x004b,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Poly Line Interpolation Method",false },
  {0x2001,0x004c,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Poly Line Begin Point Style",false },
  {0x2001,0x004d,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Poly Line End Point Style",false },
  {0x2001,0x004e,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Window Smoothing Taste",false },
  {0x2001,0x004f,"PHILIPS IMAGING DD 001",VR::FD,VM::VM1,"Harmonization Offset",false },
  {0x2001,0x0050,"PHILIPS IMAGING DD 001",VR::LO,VM::VM1,"?GraphicMarkerType",false },
  {0x2001,0x0051,"PHILIPS IMAGING DD 001",VR::IS,VM::VM1,"?OverlayPlaneID",false },
  {0x2001,0x0052,"PHILIPS IMAGING DD 001",VR::UI,VM::VM1,"?ImagePresentationStateUID",false },
  {0x2001,0x0053,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Window Invert",false },
  {0x2001,0x0055,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Graphic Line Color",false },
  {0x2001,0x0056,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Graphic Type (Private)",false },
  {0x2001,0x005a,"PHILIPS IMAGING DD 001",VR::ST,VM::VM1,"Graphic Annotation Model",false },
  {0x2001,0x005d,"PHILIPS IMAGING DD 001",VR::ST,VM::VM1,"Measurement Text Units",false },
  {0x2001,0x005e,"PHILIPS IMAGING DD 001",VR::ST,VM::VM1,"Measurement Text Type",false },
  {0x2001,0x005f,"PHILIPS IMAGING DD 001",VR::SQ,VM::VM1,"Stack Sequence",false },
  {0x2001,0x0060,"PHILIPS IMAGING DD 001",VR::SL,VM::VM1,"Number of Stacks",false },
  {0x2001,0x0061,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"?SeriesTransmitted",false },
  {0x2001,0x0062,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"?SeriesCommitted",false },
  {0x2001,0x0063,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Examination Source",false },
  {0x2001,0x0064,"PHILIPS IMAGING DD 001",VR::SH,VM::VM1,"Text Type",false },
  {0x2001,0x0065,"PHILIPS IMAGING DD 001",VR::SQ,VM::VM1,"?GraphicOverlayPlane",false },
  {0x2001,0x0067,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"?",false },
  {0x2001,0x0068,"PHILIPS IMAGING DD 001",VR::SQ,VM::VM1,"?LinearModalityGLTrafo",false },
  {0x2001,0x0069,"PHILIPS IMAGING DD 001",VR::SQ,VM::VM1,"Display Shutter Sequence",false },
  {0x2001,0x006a,"PHILIPS IMAGING DD 001",VR::SQ,VM::VM1,"?SpatialTransformation",false },
  {0x2001,0x006b,"PHILIPS IMAGING DD 001",VR::SQ,VM::VM1,"Has Edge Enhancement Sequence",false },
  {0x2001,0x006d,"PHILIPS IMAGING DD 001",VR::LO,VM::VM1,"Text Font",false },
  {0x2001,0x006e,"PHILIPS IMAGING DD 001",VR::SH,VM::VM1,"Series Type",false },
  {0x2001,0x006f,"PHILIPS IMAGING DD 001",VR::SQ,VM::VM1,"Mixing Sequence",false },
  {0x2001,0x0071,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"?GraphicConstraint",false },
  {0x2001,0x0073,"PHILIPS IMAGING DD 001",VR::SQ,VM::VM1,"Referenced Mask Image Sequence",false },
  {0x2001,0x0074,"PHILIPS IMAGING DD 001",VR::DS,VM::VM1_n,"Window Center Sub",false },
  {0x2001,0x0075,"PHILIPS IMAGING DD 001",VR::DS,VM::VM1_n,"Window Width Sub",false },
  {0x2001,0x0076,"PHILIPS IMAGING DD 001",VR::UL,VM::VM1,"?MRNumberOfFrames",false },
  {0x2001,0x0077,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"GL TrafoType",false },
  {0x2001,0x0079,"PHILIPS IMAGING DD 001",VR::SQ,VM::VM1,"Harmonisation Sequence",false },
  {0x2001,0x007a,"PHILIPS IMAGING DD 001",VR::FL,VM::VM1,"?WindowRoundingFactor",false },
  {0x2001,0x007b,"PHILIPS IMAGING DD 001",VR::IS,VM::VM1,"Acquisition Number",false },
  {0x2001,0x007e,"PHILIPS IMAGING DD 001",VR::US,VM::VM1,"Edge Enhancement Gain Taste",false },
  {0x2001,0x007f,"PHILIPS IMAGING DD 001",VR::US,VM::VM1,"Edge Enhancement Gain Taste Sub",false },
  {0x2001,0x0080,"PHILIPS IMAGING DD 001",VR::LO,VM::VM1,"Text Anchor Point Alignment",false },
  {0x2001,0x0081,"PHILIPS IMAGING DD 001",VR::IS,VM::VM1,"Number of Dynamic Scans",false },
  {0x2001,0x0082,"PHILIPS IMAGING DD 001",VR::IS,VM::VM1,"Echo Train Length",false },
  {0x2001,0x0083,"PHILIPS IMAGING DD 001",VR::DS,VM::VM1,"Imaging Frequency",false },
  {0x2001,0x0084,"PHILIPS IMAGING DD 001",VR::DS,VM::VM1,"Inversion Time",false },
  {0x2001,0x0085,"PHILIPS IMAGING DD 001",VR::DS,VM::VM1,"Magnetic Field Strength",false },
  {0x2001,0x0086,"PHILIPS IMAGING DD 001",VR::IS,VM::VM1,"?NrOfPhaseEncodingSteps?",false },
  {0x2001,0x0087,"PHILIPS IMAGING DD 001",VR::SH,VM::VM1,"Imaged Nucleus",false },
  {0x2001,0x0088,"PHILIPS IMAGING DD 001",VR::DS,VM::VM1,"Number of Averages",false },
  {0x2001,0x0089,"PHILIPS IMAGING DD 001",VR::DS,VM::VM1,"Phase FOV Percent",false },
  {0x2001,0x008a,"PHILIPS IMAGING DD 001",VR::DS,VM::VM1,"Sampling Percent",false },
  {0x2001,0x008b,"PHILIPS IMAGING DD 001",VR::SH,VM::VM1,"?Transmitting Coil?",false },
  {0x2001,0x0093,"PHILIPS IMAGING DD 001",VR::LO,VM::VM1,"Text Style",false },
  {0x2001,0x0094,"PHILIPS IMAGING DD 001",VR::LO,VM::VM1,"Processing Order Specialization",false },
  {0x2001,0x0097,"PHILIPS IMAGING DD 001",VR::SH,VM::VM1,"Nucleus",false },
  {0x2001,0x009a,"PHILIPS IMAGING DD 001",VR::SQ,VM::VM1,"Graphic Number Sequence",false },
  {0x2001,0x009b,"PHILIPS IMAGING DD 001",VR::UL,VM::VM1,"Graphic Number",false },
  {0x2001,0x009c,"PHILIPS IMAGING DD 001",VR::LO,VM::VM1,"?GraphicAnnotationLabel",false },
  {0x2001,0x009d,"PHILIPS IMAGING DD 001",VR::LO,VM::VM1,"PIIM Subtraction Type",false },
  {0x2001,0x009f,"PHILIPS IMAGING DD 001",VR::US,VM::VM2,"Pixel Processing Kernel Size",false },
  {0x2001,0x00a1,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Is Raw Image",false },
  {0x2001,0x00a2,"PHILIPS IMAGING DD 001",VR::US,VM::VM1,"LOG Subtraction Curve Taste",false },
  {0x2001,0x00a3,"PHILIPS IMAGING DD 001",VR::UL,VM::VM1,"Text Color Foreground",false },
  {0x2001,0x00a4,"PHILIPS IMAGING DD 001",VR::UL,VM::VM1,"Text Color Background",false },
  {0x2001,0x00a5,"PHILIPS IMAGING DD 001",VR::UL,VM::VM1,"Text Color Shadow",false },
  {0x2001,0x00c1,"PHILIPS IMAGING DD 001",VR::LO,VM::VM1,"Nested Object Type Name",false },
  {0x2001,0x00c8,"PHILIPS IMAGING DD 001",VR::LO,VM::VM1,"?ExamCardName?",false },
  {0x2001,0x00cc,"PHILIPS IMAGING DD 001",VR::ST,VM::VM1,"?SeriesDerivationDescription",false },
  {0x2001,0x00da,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"?",false },
  {0x2001,0x00f1,"PHILIPS IMAGING DD 001",VR::FL,VM::VM1_n,"Prospective Motion Correction",false },
  {0x2001,0x00f2,"PHILIPS IMAGING DD 001",VR::FL,VM::VM1,"Retrospective Motion Correction",false },
//  {0x2005,0x0014,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Diffusion",false },
//  {0x2005,0x0020,"PHILIPS IMAGING DD 001",VR::SL,VM::VM1,"Number of Chemical Shifts",false },
//  {0x2005,0x0030,"PHILIPS IMAGING DD 001",VR::FL,VM::VM1,"Repetition Time",false },
//  {0x2005,0x0033,"PHILIPS IMAGING DD 001",VR::FL,VM::VM1,"Scan Duration",false },
//  {0x2005,0x0035,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Data Type",false },
//  {0x2005,0x008b,"PHILIPS IMAGING DD 001",VR::SH,VM::VM1,"Transmitting Coil",false },
//  {0x2005,0x009f,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Spectral Selective Excitation Pulse",false },
//  {0x2005,0x00a1,"PHILIPS IMAGING DD 001",VR::CS,VM::VM1,"Syncra Scan Type",false },

  {0x2001,0x0001,"Philips Imaging DD 001",VR::FL,VM::VM1,"Chemical Shift",false },
  {0x2001,0x0002,"Philips Imaging DD 001",VR::IS,VM::VM1,"Chemical Shift Number MR",false },
  {0x2001,0x0003,"Philips Imaging DD 001",VR::FL,VM::VM1,"Diffusion B-Factor",false },
  {0x2001,0x0004,"Philips Imaging DD 001",VR::CS,VM::VM1,"Diffusion Direction",false },
  {0x2001,0x0005,"Philips Imaging DD 001",VR::SS,VM::VM1,"Graphic Annotation Parent ID",false },
  {0x2001,0x0006,"Philips Imaging DD 001",VR::CS,VM::VM1,"Image Enhanced",false },
  {0x2001,0x0007,"Philips Imaging DD 001",VR::CS,VM::VM1,"Image Type ED ES",false },
  {0x2001,0x0008,"Philips Imaging DD 001",VR::IS,VM::VM1,"Phase Number",false },
  {0x2001,0x0009,"Philips Imaging DD 001",VR::FL,VM::VM1,"?ImagePrepulseDelay?",false },
  {0x2001,0x000a,"Philips Imaging DD 001",VR::IS,VM::VM1,"Slice Number MR",false },
  {0x2001,0x000b,"Philips Imaging DD 001",VR::CS,VM::VM1,"Slice Orientation",false },
  {0x2001,0x000c,"Philips Imaging DD 001",VR::CS,VM::VM1,"?ArrhythmiaRejection",false },
  {0x2001,0x000e,"Philips Imaging DD 001",VR::CS,VM::VM1,"?CardiacCycled",false },
  {0x2001,0x000f,"Philips Imaging DD 001",VR::SS,VM::VM1,"?CardiacGateWidth",false },
  {0x2001,0x0010,"Philips Imaging DD 001",VR::CS,VM::VM1,"Cardiac Sync",false },
  {0x2001,0x0011,"Philips Imaging DD 001",VR::FL,VM::VM1,"Diffusion Echo Time",false },
  {0x2001,0x0012,"Philips Imaging DD 001",VR::CS,VM::VM1,"Dynamic Series",false },
  {0x2001,0x0013,"Philips Imaging DD 001",VR::SL,VM::VM1,"EPI Factor",false },
  {0x2001,0x0014,"Philips Imaging DD 001",VR::SL,VM::VM1,"Number of Echoes",false },
  {0x2001,0x0015,"Philips Imaging DD 001",VR::SS,VM::VM1,"Number of Locations",false },
  {0x2001,0x0016,"Philips Imaging DD 001",VR::SS,VM::VM1,"Number of PC Directions",false },
  {0x2001,0x0017,"Philips Imaging DD 001",VR::SL,VM::VM1,"Number of Phases MR",false },
  {0x2001,0x0018,"Philips Imaging DD 001",VR::SL,VM::VM1,"Number of Slices MR",false },
  {0x2001,0x0019,"Philips Imaging DD 001",VR::CS,VM::VM1,"Partial Matrix Scanned",false },
  {0x2001,0x001a,"Philips Imaging DD 001",VR::FL,VM::VM1_n,"PC Velocity",false },
  {0x2001,0x001b,"Philips Imaging DD 001",VR::FL,VM::VM1,"Prepulse Delay",false },
  {0x2001,0x001c,"Philips Imaging DD 001",VR::CS,VM::VM1,"Prepulse Type",false },
  {0x2001,0x001d,"Philips Imaging DD 001",VR::IS,VM::VM1,"Reconstruction Number MR",false },
  {0x2001,0x001e,"Philips Imaging DD 001",VR::CS,VM::VM1,"?MRSeriesReformatAccuracy",false },
  {0x2001,0x001f,"Philips Imaging DD 001",VR::CS,VM::VM1,"Respiration Sync",false },
  {0x2001,0x0020,"Philips Imaging DD 001",VR::LO,VM::VM1,"Scanning Technique Description MR",false },
  {0x2001,0x0021,"Philips Imaging DD 001",VR::CS,VM::VM1,"SPIR",false },
  {0x2001,0x0022,"Philips Imaging DD 001",VR::FL,VM::VM1,"Water Fat Shift",false },
  {0x2001,0x0023,"Philips Imaging DD 001",VR::DS,VM::VM1,"Flip Angle Philips",false },
  {0x2001,0x0024,"Philips Imaging DD 001",VR::CS,VM::VM1,"Interactive",false },
  {0x2001,0x0025,"Philips Imaging DD 001",VR::SH,VM::VM1,"Echo Time Display MR",false },
  {0x2001,0x0026,"Philips Imaging DD 001",VR::CS,VM::VM1,"Presentation State Subtraction Active",false },
  {0x2001,0x0028,"Philips Imaging DD 001",VR::FL,VM::VM1,"Edge Enhancement Gain Factor Sub",false },
  {0x2001,0x0029,"Philips Imaging DD 001",VR::FL,VM::VM1,"Edge Enhancement Gain Factor Non Sub",false },
  {0x2001,0x002a,"Philips Imaging DD 001",VR::CS,VM::VM1,"Edge Enhancement Taste Adapt Sub",false },
  {0x2001,0x002b,"Philips Imaging DD 001",VR::CS,VM::VM1,"Edge Enhancement Taste Non Sub",false },
  {0x2001,0x002c,"Philips Imaging DD 001",VR::FL,VM::VM1,"Harmonization Factor",false },
  {0x2001,0x002d,"Philips Imaging DD 001",VR::SS,VM::VM1,"Number of Stack Slices",false },
  {0x2001,0x002f,"Philips Imaging DD 001",VR::FL,VM::VM1,"Harmonization Gain",false },
  {0x2001,0x0030,"Philips Imaging DD 001",VR::UL,VM::VM1,"Log Subtraction Gain Step",false },
  {0x2001,0x0031,"Philips Imaging DD 001",VR::US,VM::VM1,"Mixing NR of Mask Image Numbers",false },
  {0x2001,0x0032,"Philips Imaging DD 001",VR::FL,VM::VM1_n,"Stack Radial Angle",false },
  {0x2001,0x0033,"Philips Imaging DD 001",VR::CS,VM::VM1_n,"Stack Radial Axis",false },
  {0x2001,0x0034,"Philips Imaging DD 001",VR::CS,VM::VM1,"Mixing Mask Operation",false },
  {0x2001,0x0035,"Philips Imaging DD 001",VR::SS,VM::VM1_n,"Stack Slice Number",false },
  {0x2001,0x0036,"Philips Imaging DD 001",VR::CS,VM::VM1_n,"Stack Type",false },
  {0x2001,0x0037,"Philips Imaging DD 001",VR::CS,VM::VM1,"Mixing Operation Type",false },
  {0x2001,0x0039,"Philips Imaging DD 001",VR::FL,VM::VM1,"?",false },
  {0x2001,0x003a,"Philips Imaging DD 001",VR::CS,VM::VM1,"Pixel Shift",false },
  {0x2001,0x003d,"Philips Imaging DD 001",VR::UL,VM::VM1,"?ContourFillColor",false },
  {0x2001,0x003f,"Philips Imaging DD 001",VR::CS,VM::VM1,"Interpolation Method",false },
  {0x2001,0x0042,"Philips Imaging DD 001",VR::CS,VM::VM1,"Subtraction Land Marking Active",false },
  {0x2001,0x0046,"Philips Imaging DD 001",VR::CS,VM::VM1,"Graphic Line Style",false },
  {0x2001,0x0047,"Philips Imaging DD 001",VR::FL,VM::VM1,"Graphic Line Width",false },
  {0x2001,0x0048,"Philips Imaging DD 001",VR::SS,VM::VM1,"Graphic Annotation ID",false },
  {0x2001,0x004b,"Philips Imaging DD 001",VR::CS,VM::VM1,"Poly Line Interpolation Method",false },
  {0x2001,0x004c,"Philips Imaging DD 001",VR::CS,VM::VM1,"Poly Line Begin Point Style",false },
  {0x2001,0x004d,"Philips Imaging DD 001",VR::CS,VM::VM1,"Poly Line End Point Style",false },
  {0x2001,0x004e,"Philips Imaging DD 001",VR::CS,VM::VM1,"Window Smoothing Taste",false },
  {0x2001,0x004f,"Philips Imaging DD 001",VR::FD,VM::VM1,"Harmonization Offset",false },
  {0x2001,0x0050,"Philips Imaging DD 001",VR::LO,VM::VM1,"?GraphicMarkerType",false },
  {0x2001,0x0051,"Philips Imaging DD 001",VR::IS,VM::VM1,"?OverlayPlaneID",false },
  {0x2001,0x0052,"Philips Imaging DD 001",VR::UI,VM::VM1,"?ImagePresentationStateUID",false },
  {0x2001,0x0053,"Philips Imaging DD 001",VR::CS,VM::VM1,"Window Invert",false },
  {0x2001,0x0055,"Philips Imaging DD 001",VR::CS,VM::VM1,"Graphic Line Color",false },
  {0x2001,0x0056,"Philips Imaging DD 001",VR::CS,VM::VM1,"Graphic Type (Private)",false },
  {0x2001,0x005a,"Philips Imaging DD 001",VR::ST,VM::VM1,"Graphic Annotation Model",false },
  {0x2001,0x005d,"Philips Imaging DD 001",VR::ST,VM::VM1,"Measurement Text Units",false },
  {0x2001,0x005e,"Philips Imaging DD 001",VR::ST,VM::VM1,"Measurement Text Type",false },
  {0x2001,0x005f,"Philips Imaging DD 001",VR::SQ,VM::VM1,"Stack Sequence",false },
  {0x2001,0x0060,"Philips Imaging DD 001",VR::SL,VM::VM1,"Number of Stacks",false },
  {0x2001,0x0061,"Philips Imaging DD 001",VR::CS,VM::VM1,"?SeriesTransmitted",false },
  {0x2001,0x0062,"Philips Imaging DD 001",VR::CS,VM::VM1,"?SeriesCommitted",false },
  {0x2001,0x0063,"Philips Imaging DD 001",VR::CS,VM::VM1,"Examination Source",false },
  {0x2001,0x0064,"Philips Imaging DD 001",VR::SH,VM::VM1,"Text Type",false },
  {0x2001,0x0065,"Philips Imaging DD 001",VR::SQ,VM::VM1,"?GraphicOverlayPlane",false },
  {0x2001,0x0067,"Philips Imaging DD 001",VR::CS,VM::VM1,"?",false },
  {0x2001,0x0068,"Philips Imaging DD 001",VR::SQ,VM::VM1,"?LinearModalityGLTrafo",false },
  {0x2001,0x0069,"Philips Imaging DD 001",VR::SQ,VM::VM1,"Display Shutter Sequence",false },
  {0x2001,0x006a,"Philips Imaging DD 001",VR::SQ,VM::VM1,"?SpatialTransformation",false },
  {0x2001,0x006b,"Philips Imaging DD 001",VR::SQ,VM::VM1,"Has Edge Enhancement Sequence",false },
  {0x2001,0x006d,"Philips Imaging DD 001",VR::LO,VM::VM1,"Text Font",false },
  {0x2001,0x006e,"Philips Imaging DD 001",VR::SH,VM::VM1,"Series Type",false },
  {0x2001,0x006f,"Philips Imaging DD 001",VR::SQ,VM::VM1,"Mixing Sequence",false },
  {0x2001,0x0071,"Philips Imaging DD 001",VR::CS,VM::VM1,"?GraphicConstraint",false },
  {0x2001,0x0073,"Philips Imaging DD 001",VR::SQ,VM::VM1,"Referenced Mask Image Sequence",false },
  {0x2001,0x0074,"Philips Imaging DD 001",VR::DS,VM::VM1_n,"Window Center Sub",false },
  {0x2001,0x0075,"Philips Imaging DD 001",VR::DS,VM::VM1_n,"Window Width Sub",false },
  {0x2001,0x0076,"Philips Imaging DD 001",VR::UL,VM::VM1,"?MRNumberOfFrames",false },
  {0x2001,0x0077,"Philips Imaging DD 001",VR::CS,VM::VM1,"GL TrafoType",false },
  {0x2001,0x0079,"Philips Imaging DD 001",VR::SQ,VM::VM1,"Harmonisation Sequence",false },
  {0x2001,0x007a,"Philips Imaging DD 001",VR::FL,VM::VM1,"?WindowRoundingFactor",false },
  {0x2001,0x007b,"Philips Imaging DD 001",VR::IS,VM::VM1,"Acquisition Number",false },
  {0x2001,0x007e,"Philips Imaging DD 001",VR::US,VM::VM1,"Edge Enhancement Gain Taste",false },
  {0x2001,0x007f,"Philips Imaging DD 001",VR::US,VM::VM1,"Edge Enhancement Gain Taste Sub",false },
  {0x2001,0x0080,"Philips Imaging DD 001",VR::LO,VM::VM1,"Text Anchor Point Alignment",false },
  {0x2001,0x0081,"Philips Imaging DD 001",VR::IS,VM::VM1,"Number of Dynamic Scans",false },
  {0x2001,0x0082,"Philips Imaging DD 001",VR::IS,VM::VM1,"Echo Train Length",false },
  {0x2001,0x0083,"Philips Imaging DD 001",VR::DS,VM::VM1,"Imaging Frequency",false },
  {0x2001,0x0084,"Philips Imaging DD 001",VR::DS,VM::VM1,"Inversion Time",false },
  {0x2001,0x0085,"Philips Imaging DD 001",VR::DS,VM::VM1,"Magnetic Field Strength",false },
  {0x2001,0x0086,"Philips Imaging DD 001",VR::IS,VM::VM1,"?NrOfPhaseEncodingSteps?",false },
  {0x2001,0x0087,"Philips Imaging DD 001",VR::SH,VM::VM1,"Imaged Nucleus",false },
  {0x2001,0x0088,"Philips Imaging DD 001",VR::DS,VM::VM1,"Number of Averages",false },
  {0x2001,0x0089,"Philips Imaging DD 001",VR::DS,VM::VM1,"Phase FOV Percent",false },
  {0x2001,0x008a,"Philips Imaging DD 001",VR::DS,VM::VM1,"Sampling Percent",false },
  {0x2001,0x008b,"Philips Imaging DD 001",VR::SH,VM::VM1,"?Transmitting Coil?",false },
  {0x2001,0x0093,"Philips Imaging DD 001",VR::LO,VM::VM1,"Text Style",false },
  {0x2001,0x0094,"Philips Imaging DD 001",VR::LO,VM::VM1,"Processing Order Specialization",false },
  {0x2001,0x0097,"Philips Imaging DD 001",VR::SH,VM::VM1,"Nucleus",false },
  {0x2001,0x009a,"Philips Imaging DD 001",VR::SQ,VM::VM1,"Graphic Number Sequence",false },
  {0x2001,0x009b,"Philips Imaging DD 001",VR::UL,VM::VM1,"Graphic Number",false },
  {0x2001,0x009c,"Philips Imaging DD 001",VR::LO,VM::VM1,"?GraphicAnnotationLabel",false },
  {0x2001,0x009d,"Philips Imaging DD 001",VR::LO,VM::VM1,"PIIM Subtraction Type",false },
  {0x2001,0x009f,"Philips Imaging DD 001",VR::US,VM::VM2,"Pixel Processing Kernel Size",false },
  {0x2001,0x00a1,"Philips Imaging DD 001",VR::CS,VM::VM1,"Is Raw Image",false },
  {0x2001,0x00a2,"Philips Imaging DD 001",VR::US,VM::VM1,"LOG Subtraction Curve Taste",false },
  {0x2001,0x00a3,"Philips Imaging DD 001",VR::UL,VM::VM1,"Text Color Foreground",false },
  {0x2001,0x00a4,"Philips Imaging DD 001",VR::UL,VM::VM1,"Text Color Background",false },
  {0x2001,0x00a5,"Philips Imaging DD 001",VR::UL,VM::VM1,"Text Color Shadow",false },
  {0x2001,0x00c1,"Philips Imaging DD 001",VR::LO,VM::VM1,"Nested Object Type Name",false },
  {0x2001,0x00c8,"Philips Imaging DD 001",VR::LO,VM::VM1,"?ExamCardName?",false },
  {0x2001,0x00cc,"Philips Imaging DD 001",VR::ST,VM::VM1,"?SeriesDerivationDescription",false },
  {0x2001,0x00da,"Philips Imaging DD 001",VR::CS,VM::VM1,"?",false },
  {0x2001,0x00de,"Philips Imaging DD 001",VR::UN,VM::VM1,"?",false },
  {0x2001,0x00f1,"Philips Imaging DD 001",VR::FL,VM::VM1_n,"Prospective Motion Correction",false },
  {0x2001,0x00f2,"Philips Imaging DD 001",VR::FL,VM::VM1_n,"Retrospective Motion Correction",false },
  {0x2001,0x00f3,"Philips Imaging DD 001",VR::CS,VM::VM1,"Unknown Tag & Data",false },
  {0x2005,0x0014,"Philips Imaging DD 001",VR::CS,VM::VM1,"Diffusion",false },
  {0x2005,0x0020,"Philips Imaging DD 001",VR::SL,VM::VM1,"Number of Chemical Shifts",false },
  {0x2005,0x0030,"Philips Imaging DD 001",VR::FL,VM::VM1_n,"Repetition Time",false },
  {0x2005,0x0033,"Philips Imaging DD 001",VR::FL,VM::VM1,"Scan Duration",false },
  {0x2005,0x0035,"Philips Imaging DD 001",VR::CS,VM::VM1,"Data Type",false },
  {0x2005,0x008b,"Philips Imaging DD 001",VR::SH,VM::VM1,"Transmitting Coil",false },
  {0x2005,0x009f,"Philips Imaging DD 001",VR::CS,VM::VM1,"Spectral Selective Excitation Pulse",false },
  {0x2005,0x00a1,"Philips Imaging DD 001",VR::CS,VM::VM1,"Syncra Scan Type",false },
  {0x2001,0x0001,"Philips Imaging DD 002",VR::US,VM::VM1,"?",false },
  {0x2001,0x0002,"Philips Imaging DD 002",VR::FD,VM::VM1,"?",false },
  {0x2001,0x0013,"Philips Imaging DD 002",VR::SS,VM::VM1,"?",false },
  {0x2001,0x0014,"Philips Imaging DD 002",VR::FD,VM::VM1,"?",false },
  {0x2001,0x0015,"Philips Imaging DD 002",VR::FD,VM::VM1,"?",false },
  {0x2001,0x0016,"Philips Imaging DD 002",VR::FD,VM::VM1,"?",false },
  {0x2001,0x0017,"Philips Imaging DD 002",VR::FD,VM::VM1,"?",false },
  {0x2001,0x0018,"Philips Imaging DD 002",VR::CS,VM::VM1,"?",false },
  {0x2001,0x0019,"Philips Imaging DD 002",VR::FD,VM::VM1,"?",false },
  {0x2001,0x001a,"Philips Imaging DD 002",VR::FD,VM::VM1,"?",false },
  {0x2001,0x001b,"Philips Imaging DD 002",VR::FD,VM::VM1,"?",false },
  {0x2001,0x001c,"Philips Imaging DD 002",VR::FD,VM::VM1,"?",false },
  {0x2001,0x001d,"Philips Imaging DD 002",VR::FD,VM::VM1,"?",false },
  {0x2001,0x001e,"Philips Imaging DD 002",VR::FD,VM::VM1,"?",false },
  {0x2001,0x001f,"Philips Imaging DD 002",VR::FD,VM::VM1,"?",false },
  {0x2001,0x0020,"Philips Imaging DD 002",VR::FD,VM::VM1,"?",false },
  {0x2001,0x0021,"Philips Imaging DD 002",VR::FD,VM::VM1,"?",false },
  {0x2001,0x0022,"Philips Imaging DD 002",VR::FD,VM::VM1,"?",false },
  {0x2001,0x0023,"Philips Imaging DD 002",VR::FD,VM::VM1,"?",false },
  {0x2001,0x0024,"Philips Imaging DD 002",VR::FD,VM::VM1,"?",false },
  {0x2001,0x0025,"Philips Imaging DD 002",VR::FD,VM::VM1,"?",false },
  {0x2001,0x0026,"Philips Imaging DD 002",VR::FD,VM::VM1,"?",false },
  {0x2001,0x0027,"Philips Imaging DD 002",VR::FD,VM::VM1,"?",false },
  {0x2001,0x0028,"Philips Imaging DD 002",VR::US,VM::VM1,"?",false },
  {0x2001,0x0029,"Philips Imaging DD 002",VR::US,VM::VM1,"?",false },
  {0x2001,0x002a,"Philips Imaging DD 002",VR::US,VM::VM1,"?",false },
  {0x2001,0x002b,"Philips Imaging DD 002",VR::SS,VM::VM1,"?",false },
  {0x2001,0x002c,"Philips Imaging DD 002",VR::FD,VM::VM1,"?",false },
  {0x2001,0x002d,"Philips Imaging DD 002",VR::FD,VM::VM1,"?",false },
  {0x2001,0x002e,"Philips Imaging DD 002",VR::SS,VM::VM1,"?",false },
  {0x2001,0x002f,"Philips Imaging DD 002",VR::SS,VM::VM1,"?",false },
  {0x2001,0x0030,"Philips Imaging DD 002",VR::SS,VM::VM1,"?",false },
  {0x2001,0x0031,"Philips Imaging DD 002",VR::SS,VM::VM1,"?",false },
  {0x2001,0x0032,"Philips Imaging DD 002",VR::SS,VM::VM1,"?",false },
  {0x2001,0x0033,"Philips Imaging DD 002",VR::SS,VM::VM1,"?",false },
  {0x2001,0x0034,"Philips Imaging DD 002",VR::SS,VM::VM1,"?",false },
  {0x2001,0x0035,"Philips Imaging DD 002",VR::FD,VM::VM1,"?",false },
  {0x2001,0x0036,"Philips Imaging DD 002",VR::FD,VM::VM1,"?",false },
  {0x2001,0x0037,"Philips Imaging DD 002",VR::FD,VM::VM1,"?",false },
  {0x2001,0x0039,"Philips Imaging DD 002",VR::CS,VM::VM1,"?",false },
  {0x2001,0x003a,"Philips Imaging DD 002",VR::SQ,VM::VM1,"?",false },
  {0x2001,0x003b,"Philips Imaging DD 002",VR::SQ,VM::VM1,"?",false },
  {0x2001,0x003c,"Philips Imaging DD 002",VR::SQ,VM::VM1,"?",false },
  {0x2001,0x003d,"Philips Imaging DD 002",VR::SQ,VM::VM1,"?",false },
  {0x2001,0x003e,"Philips Imaging DD 002",VR::SS,VM::VM1,"?",false },
  {0x2001,0x003f,"Philips Imaging DD 002",VR::SS,VM::VM1,"?",false },
  {0x2001,0x0040,"Philips Imaging DD 002",VR::SS,VM::VM1,"?",false },
  {0x2001,0x006b,"Philips Imaging DD 002",VR::LO,VM::VM1,"?",false },
  {0x4007,0x0000,"Philips Imaging DD 065",VR::CS,VM::VM1,"?",false },
  {0x4001,0x0000,"Philips Imaging DD 067",VR::SQ,VM::VM1,"?",false },
  {0x4001,0x0001,"Philips Imaging DD 067",VR::CS,VM::VM1,"?",false },
  {0x4001,0x0008,"Philips Imaging DD 067",VR::CS,VM::VM1,"?",false },
  {0x4001,0x0010,"Philips Imaging DD 070",VR::SQ,VM::VM1,"?",false },
  {0x4001,0x0011,"Philips Imaging DD 070",VR::SQ,VM::VM1,"?",false },
  {0x4001,0x0012,"Philips Imaging DD 070",VR::SQ,VM::VM1,"?",false },
  {0x4001,0x0016,"Philips Imaging DD 070",VR::ST,VM::VM1,"?",false },
  {0x4001,0x0017,"Philips Imaging DD 070",VR::ST,VM::VM1,"?",false },
  {0x4001,0x0018,"Philips Imaging DD 070",VR::ST,VM::VM1,"?",false },
  {0x4001,0x001c,"Philips Imaging DD 070",VR::SQ,VM::VM1,"?",false },
  {0x4001,0x001d,"Philips Imaging DD 070",VR::LT,VM::VM1,"?",false },
  {0x4007,0x0048,"Philips Imaging DD 073",VR::FL,VM::VM1,"?",false },
  {0x1001,0x0003,"Philips Imaging DD 124",VR::LO,VM::VM1,"?",false }, /* MR Import Filter 3/21/2014 1:57:35 PM */
  {0x2001,0x0000,"Philips Imaging DD 129",VR::SQ,VM::VM1,"PresentationStateSequence",false },
  {0x2001,0x0001,"Philips Imaging DD 129",VR::SQ,VM::VM1,"?",false },
  {0x2005,0x0000,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"?AngulationAP?",false },
  {0x2005,0x0001,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"?AngulationFH?",false },
  {0x2005,0x0002,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"?AngulationRL?",false },
  {0x2005,0x0003,"PHILIPS MR IMAGING DD 001",VR::IS,VM::VM1,"?ImageAnnotationCount",false },
  {0x2005,0x0004,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"?DisplayOrientation?",false },
  {0x2005,0x0005,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"Synergy Reconstruction Type",false },
  {0x2005,0x0006,"PHILIPS MR IMAGING DD 001",VR::UI,VM::VM1,"?",false },
  {0x2005,0x0007,"PHILIPS MR IMAGING DD 001",VR::IS,VM::VM1,"?ImageLineCount",false },
  {0x2005,0x0008,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"?OffCentreAP?",false },
  {0x2005,0x0009,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"?OffCentreFH?",false },
  {0x2005,0x000a,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"?OffCentreRL?",false },
  {0x2005,0x000b,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"?MaxFP?",false },
  {0x2005,0x000c,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"?MinFP?",false },
  {0x2005,0x000d,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"?ScaleIntercept?",false },
  {0x2005,0x000e,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"?ScaleSlope?",false },
  {0x2005,0x000f,"PHILIPS MR IMAGING DD 001",VR::DS,VM::VM1,"Window Center Philips",false },
  {0x2005,0x0010,"PHILIPS MR IMAGING DD 001",VR::DS,VM::VM1,"Window Width Philips",false },
  {0x2005,0x0011,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"Image Type Philips",false },
  {0x2005,0x0012,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"?CardiacGating?",false },
  {0x2005,0x0013,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"?DevelopmentMode?",false },
  {0x2005,0x0014,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"Diffusion",false },
  {0x2005,0x0015,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"?FatSaturationTechnique?",false },
  {0x2005,0x0016,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"?FlowCompensation",false },
  {0x2005,0x0017,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"?FourierInterpolation",false },
  {0x2005,0x0018,"PHILIPS MR IMAGING DD 001",VR::LO,VM::VM1,"?HardcopyProtocol",false },
  {0x2005,0x0019,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"?InverseReconstructed",false },
  {0x2005,0x001a,"PHILIPS MR IMAGING DD 001",VR::SS,VM::VM1,"?LabelSyntax",false },
  {0x2005,0x001b,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"?MagnetiPrepared",false },
  {0x2005,0x001c,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"?MagnetTransferConst",false },
  {0x2005,0x001d,"PHILIPS MR IMAGING DD 001",VR::SS,VM::VM1,"?MeasurementScanResolution",false },
  {0x2005,0x001e,"PHILIPS MR IMAGING DD 001",VR::SH,VM::VM1,"?MIPProtocol",false },
  {0x2005,0x001f,"PHILIPS MR IMAGING DD 001",VR::SH,VM::VM1,"?MPRProtocol",false },
  {0x2005,0x0020,"PHILIPS MR IMAGING DD 001",VR::SL,VM::VM1,"Number of Chemical Shift",false },
  {0x2005,0x0021,"PHILIPS MR IMAGING DD 001",VR::SS,VM::VM1,"NumberOfMixes", false },
  {0x2005,0x0022,"PHILIPS MR IMAGING DD 001",VR::IS,VM::VM1,"NumberOfReferences", false },
  {0x2005,0x0023,"PHILIPS MR IMAGING DD 001",VR::SS,VM::VM1,"NumberOfSlabs", false },
  {0x2005,0x0025,"PHILIPS MR IMAGING DD 001",VR::SS,VM::VM1,"NumberOfVolumes", false },
  {0x2005,0x0026,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"OverSampleingPhase", false },
  {0x2005,0x0027,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"PackageMode", false },
  {0x2005,0x0028,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"PartialFourierFrequency", false },
  {0x2005,0x0029,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"PartialFourierPhase", false },
  {0x2005,0x002a,"PHILIPS MR IMAGING DD 001",VR::IS,VM::VM1,"PatientReferenceID", false },
  {0x2005,0x002b,"PHILIPS MR IMAGING DD 001",VR::SS,VM::VM1,"PercentScanComplete", false },
  {0x2005,0x002c,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"PhaseEncodedRecording", false },
  {0x2005,0x002d,"PHILIPS MR IMAGING DD 001",VR::IS,VM::VM1,"PlanScanSurveyNumberOfImages", false },
  {0x2005,0x002e,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"PPGPPUGating", false },
  {0x2005,0x002f,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"SpatialPresaturation", false },
  {0x2005,0x0030,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1_n,"Repetition Time",false },
  {0x2005,0x0031,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"RespiratoryGating", false },
  {0x2005,0x0032,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"?SampleRepresentation",false },
  {0x2005,0x0033,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"Scan Duration",false },
  {0x2005,0x0034,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"MRSegmentedKSpace", false },
  {0x2005,0x0035,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"Data Type",false },
  {0x2005,0x0036,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"SeriesIsCardiac", false },
  {0x2005,0x0037,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"SeriesIsSpectro", false },
  {0x2005,0x0038,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"Spoiled", false },
  {0x2005,0x0039,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"SteadyState", false },
  {0x2005,0x003a,"PHILIPS MR IMAGING DD 001",VR::SH,VM::VM1,"SubAnatomy", false },
  {0x2005,0x003b,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"TimeReversedSteadyState", false },
  {0x2005,0x003c,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"SeriesTone", false },
  {0x2005,0x003d,"PHILIPS MR IMAGING DD 001",VR::SS,VM::VM1,"NumberOfRRIntervalRanges", false },
  {0x2005,0x003e,"PHILIPS MR IMAGING DD 001",VR::SL,VM::VM1_n,"RRIntervalsDistribution", false },
  {0x2005,0x003f,"PHILIPS MR IMAGING DD 001",VR::SL,VM::VM1,"PlanScanAcquisitionNo", false },
  {0x2005,0x0040,"PHILIPS MR IMAGING DD 001",VR::SL,VM::VM1,"ChemicalShiftNo", false },
  {0x2005,0x0041,"PHILIPS MR IMAGING DD 001",VR::SL,VM::VM1,"PlanScanDynamicScanNo", false },
  {0x2005,0x0042,"PHILIPS MR IMAGING DD 001",VR::SL,VM::VM1,"PlanScanSurveyEchoNo", false },
  {0x2005,0x0043,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"PlanScanImageType", false },
  {0x2005,0x0044,"PHILIPS MR IMAGING DD 001",VR::SL,VM::VM1,"PlanScanPhaseNo", false },
  {0x2005,0x0045,"PHILIPS MR IMAGING DD 001",VR::SL,VM::VM1,"PlanScanReconstructionNo", false },
  {0x2005,0x0046,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"PlanScanScanningSequence", false },
  {0x2005,0x0047,"PHILIPS MR IMAGING DD 001",VR::SL,VM::VM1,"PlanScanSliceNo", false },
  {0x2005,0x0048,"PHILIPS MR IMAGING DD 001",VR::IS,VM::VM1,"?MRReferenceAcquisitionNo",false },
  {0x2005,0x0049,"PHILIPS MR IMAGING DD 001",VR::IS,VM::VM1,"?MRReferenceChemicalShiftNo",false },
  {0x2005,0x004a,"PHILIPS MR IMAGING DD 001",VR::IS,VM::VM1,"?MRReferenceDynamicScanNo",false },
  {0x2005,0x004b,"PHILIPS MR IMAGING DD 001",VR::IS,VM::VM1,"?MRReferenceEchoNo",false },
  {0x2005,0x004c,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"?MRReferenceEntity",false },
  {0x2005,0x004d,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"?MRReferenceImageType",false },
  {0x2005,0x004e,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"?MRSlabFovRL",false },
  {0x2005,0x004f,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"?MRSlabOffcentreAP",false },
  {0x2005,0x0050,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"?MRSlabOffcentreFH",false },
  {0x2005,0x0051,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"?MRSlabOffcentreRL",false },
  {0x2005,0x0052,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"?MRSlabType",false },
  {0x2005,0x0053,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"?MRSlabViewAxis",false },
  {0x2005,0x0054,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"SVS Voxel Angulation AP (deg)",false },
  {0x2005,0x0055,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"SVS Voxel Angulation FH (deg)",false },
  {0x2005,0x0056,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"SVS Voxel Angulation LR (deg)",false },
  {0x2005,0x0057,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"SVS Voxel Size AP (mm)",false },
  {0x2005,0x0058,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"SVS Voxel Size FH (mm)",false },
  {0x2005,0x0059,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"SVS Voxel Size LR (mm)",false },
  {0x2005,0x005a,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"SVS Voxel Offcenter AP (mm)",false },
  {0x2005,0x005b,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"SVS Voxel Offcenter FH (mm)",false },
  {0x2005,0x005c,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"SVS Voxel Offcenter LR (mm)",false },
  {0x2005,0x005d,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"VolumeType", false },
  {0x2005,0x005e,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"VolumeViewAxis", false },
  {0x2005,0x005f,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"StudyOrigin", false },
  {0x2005,0x0060,"PHILIPS MR IMAGING DD 001",VR::IS,VM::VM1,"StudySequenceNumber", false },
  {0x2005,0x0061,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"ImagePrepulseType", false },
  {0x2005,0x0063,"PHILIPS MR IMAGING DD 001",VR::SS,VM::VM1,"fMRIStatusIndication", false },
  {0x2005,0x0064,"PHILIPS MR IMAGING DD 001",VR::IS,VM::VM1,"?MRReferencePhaseNo",false },
  {0x2005,0x0065,"PHILIPS MR IMAGING DD 001",VR::IS,VM::VM1,"?MRReferenceReconstructionNo",false },
  {0x2005,0x0066,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"?MRReferenceScanningSequence",false },
  {0x2005,0x0067,"PHILIPS MR IMAGING DD 001",VR::IS,VM::VM1,"?MRReferenceSliceNo",false },
  {0x2005,0x0068,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"?MRReferenceType",false },
  {0x2005,0x0069,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"?MRSlabAngulationAP",false },
  {0x2005,0x006a,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"?MRSlabAngulationFH",false },
  {0x2005,0x006b,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"?MRSlabAngulationRL",false },
  {0x2005,0x006c,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"?MRSlabFovAP",false },
  {0x2005,0x006d,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"?MRSlabFovFH",false },
  {0x2005,0x006e,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"ImageScanningSequencePrivate", false },
  {0x2005,0x006f,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"?AcquisitionType?", false },
  {0x2005,0x0070,"PHILIPS MR IMAGING DD 001",VR::LO,VM::VM1,"?MRSeriesHardcopyProtocolEV",false },
  {0x2005,0x0071,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"StackAngulationAP", false },
  {0x2005,0x0072,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"StackAngulationFH", false },
  {0x2005,0x0073,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"StackAngulationRL", false },
  {0x2005,0x0074,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"StackFovAP", false },
  {0x2005,0x0075,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"StackFovFH", false },
  {0x2005,0x0076,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"StackFovRL", false },
  {0x2005,0x0078,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"StackOffcentreAP", false },
  {0x2005,0x0079,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"StackOffcentreFH", false },
  {0x2005,0x007a,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"StackOffcentreRL", false },
  {0x2005,0x007b,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"StackPreparationDirection", false },
  {0x2005,0x007e,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"Z-Spacing (aka Stack Slice Distance)",false },
  {0x2005,0x0080,"PHILIPS MR IMAGING DD 001",VR::SQ,VM::VM1,"SeriesPlanScan", false },
  {0x2005,0x0081,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"StackViewAxis", false },
  {0x2005,0x0083,"PHILIPS MR IMAGING DD 001",VR::SQ,VM::VM1,"SeriesSlab", false },
  {0x2005,0x0084,"PHILIPS MR IMAGING DD 001",VR::SQ,VM::VM1,"SeriesReference", false },
  {0x2005,0x0085,"PHILIPS MR IMAGING DD 001",VR::SQ,VM::VM1,"SeriesVolume", false },
  {0x2005,0x0086,"PHILIPS MR IMAGING DD 001",VR::SS,VM::VM1,"NumberOfGeometry", false },
  {0x2005,0x009e,"PHILIPS MR IMAGING DD 001",VR::SQ,VM::VM1,"SeriesGeom", false },
  {0x2005,0x009f,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"MRSeriesSpectralSelectiveExcitationPulse", false },
  {0x2005,0x00a0,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"Philips Temporal Position Identifier (aka ImageDynamicScanBeginTime)",false },
  {0x2005,0x00a1,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"Syncra Scan Type",false },
  {0x2005,0x00a2,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"IsCOCA", false },
  {0x2005,0x00a3,"PHILIPS MR IMAGING DD 001",VR::IS,VM::VM1,"StackCoilID", false },
  {0x2005,0x00a4,"PHILIPS MR IMAGING DD 001",VR::IS,VM::VM1,"StackCBBCoil1", false },
  {0x2005,0x00a5,"PHILIPS MR IMAGING DD 001",VR::IS,VM::VM1,"StackCBBCoil2", false },
  {0x2005,0x00a6,"PHILIPS MR IMAGING DD 001",VR::IS,VM::VM1,"StackChannelCombi", false },
  {0x2005,0x00a7,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"StackCoilConn", false },
  {0x2005,0x00a8,"PHILIPS MR IMAGING DD 001",VR::DS,VM::VM1,"PrivateInversionTime", false },
  {0x2005,0x00a9,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"SeriesGeometryCorrection", false },
  {0x2005,0x00b0,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"Diffusion Direction RL (degree)",false },
  {0x2005,0x00b1,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"Diffusion Direction AP (degree)",false },
  {0x2005,0x00b2,"PHILIPS MR IMAGING DD 001",VR::FL,VM::VM1,"Diffusion Direction FH (degree)",false },
  {0x2005,0x00c0,"PHILIPS MR IMAGING DD 001",VR::CS,VM::VM1,"?SequenceVariant?", false },
  /* pay attention to the case */
  {0x2005,0x0000,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"?AngulationAP?",false },
  {0x2005,0x0001,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"?AngulationFH?",false },
  {0x2005,0x0002,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"?AngulationRL?",false },
  {0x2005,0x0003,"Philips MR Imaging DD 001",VR::IS,VM::VM1,"?ImageAnnotationCount",false },
  {0x2005,0x0004,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"?DisplayOrientation?",false },
  {0x2005,0x0005,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"Synergy Reconstruction Type",false },
  {0x2005,0x0006,"Philips MR Imaging DD 001",VR::UI,VM::VM1,"?",false },
  {0x2005,0x0007,"Philips MR Imaging DD 001",VR::IS,VM::VM1,"ImageLineCount",false },
  {0x2005,0x0008,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"?OffCentreAP?",false },
  {0x2005,0x0009,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"?OffCentreFH?",false },
  {0x2005,0x000a,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"?OffCentreRL?",false },
  {0x2005,0x000b,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"?MaxFP?",false },
  {0x2005,0x000c,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"?MinFP?",false },
  {0x2005,0x000d,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"?ScaleIntercept?",false },
  {0x2005,0x000e,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"?ScaleSlope?",false },
  {0x2005,0x000f,"Philips MR Imaging DD 001",VR::DS,VM::VM1,"Window Center Philips",false },
  {0x2005,0x0010,"Philips MR Imaging DD 001",VR::DS,VM::VM1,"Window Width Philips",false },
  {0x2005,0x0011,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"Image Type Philips",false },
  {0x2005,0x0012,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"?CardiacGating?",false },
  {0x2005,0x0013,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"?DevelopmentMode?",false },
  {0x2005,0x0014,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"Diffusion",false },
  {0x2005,0x0015,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"?FatSaturationTechnique?",false },
  {0x2005,0x0016,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"?FlowCompensation",false },
  {0x2005,0x0017,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"?FourierInterpolation",false },
  {0x2005,0x0018,"Philips MR Imaging DD 001",VR::LO,VM::VM1,"?MRHardcopyProtocol",false },
  {0x2005,0x0019,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"?InverseReconstructed",false },
  {0x2005,0x001a,"Philips MR Imaging DD 001",VR::SS,VM::VM1,"?LabelSyntax",false },
  {0x2005,0x001b,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"?MagnetiPrepared",false },
  {0x2005,0x001c,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"?MagnetTransferConst",false },
  {0x2005,0x001d,"Philips MR Imaging DD 001",VR::SS,VM::VM1,"?MeasurementScanResolution",false },
  {0x2005,0x001e,"Philips MR Imaging DD 001",VR::SH,VM::VM1,"?MIPProtocol",false },
  {0x2005,0x001f,"Philips MR Imaging DD 001",VR::SH,VM::VM1,"?MPRProtocol",false },
  {0x2005,0x0020,"Philips MR Imaging DD 001",VR::SL,VM::VM1,"Number of Chemical Shift",false },
  {0x2005,0x0021,"Philips MR Imaging DD 001",VR::SS,VM::VM1,"NumberOfMixes", false },
  {0x2005,0x0022,"Philips MR Imaging DD 001",VR::IS,VM::VM1,"NumberOfReferences", false },
  {0x2005,0x0023,"Philips MR Imaging DD 001",VR::SS,VM::VM1,"NumberOfSlabs", false },
  {0x2005,0x0025,"Philips MR Imaging DD 001",VR::SS,VM::VM1,"NumberOfVolumes", false },
  {0x2005,0x0026,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"OverSampleingPhase", false },
  {0x2005,0x0027,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"PackageMode", false },
  {0x2005,0x0028,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"PartialFourierFrequency", false },
  {0x2005,0x0029,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"PartialFourierPhase", false },
  {0x2005,0x002a,"Philips MR Imaging DD 001",VR::IS,VM::VM1,"PatientReferenceID", false },
  {0x2005,0x002b,"Philips MR Imaging DD 001",VR::SS,VM::VM1,"PercentScanComplete", false },
  {0x2005,0x002c,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"PhaseEncodedRecording", false },
  {0x2005,0x002d,"Philips MR Imaging DD 001",VR::IS,VM::VM1,"PlanScanSurveyNumberOfImages", false },
  {0x2005,0x002e,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"PPGPPUGating", false },
  {0x2005,0x002f,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"SpatialPresaturation", false },
  {0x2005,0x0030,"Philips MR Imaging DD 001",VR::FL,VM::VM1_n,"Repetition Time",false },
  {0x2005,0x0031,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"RespiratoryGating", false },
  {0x2005,0x0032,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"?SampleRepresentation",false },
  {0x2005,0x0033,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"Scan Duration",false },
  {0x2005,0x0034,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"MRSegmentedKSpace", false },
  {0x2005,0x0035,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"Data Type",false },
  {0x2005,0x0036,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"SeriesIsCardiac", false },
  {0x2005,0x0037,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"SeriesIsSpectro", false },
  {0x2005,0x0038,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"Spoiled", false },
  {0x2005,0x0039,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"SteadyState", false },
  {0x2005,0x003a,"Philips MR Imaging DD 001",VR::SH,VM::VM1,"SubAnatomy", false },
  {0x2005,0x003b,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"TimeReversedSteadyState", false },
  {0x2005,0x003c,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"SeriesTone", false },
  {0x2005,0x003d,"Philips MR Imaging DD 001",VR::SS,VM::VM1,"NumberOfRRIntervalRanges", false },
  {0x2005,0x003e,"Philips MR Imaging DD 001",VR::SL,VM::VM1_n,"RRIntervalsDistribution", false },
  {0x2005,0x003f,"Philips MR Imaging DD 001",VR::SL,VM::VM1,"PlanScanAcquisitionNo", false },
  {0x2005,0x0040,"Philips MR Imaging DD 001",VR::SL,VM::VM1,"ChemicalShiftNo", false },
  {0x2005,0x0041,"Philips MR Imaging DD 001",VR::SL,VM::VM1,"PlanScanDynamicScanNo", false },
  {0x2005,0x0042,"Philips MR Imaging DD 001",VR::SL,VM::VM1,"PlanScanSurveyEchoNo", false },
  {0x2005,0x0043,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"PlanScanImageType", false },
  {0x2005,0x0044,"Philips MR Imaging DD 001",VR::SL,VM::VM1,"PlanScanPhaseNo", false },
  {0x2005,0x0045,"Philips MR Imaging DD 001",VR::SL,VM::VM1,"PlanScanReconstructionNo", false },
  {0x2005,0x0046,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"PlanScanScanningSequence", false },
  {0x2005,0x0047,"Philips MR Imaging DD 001",VR::SL,VM::VM1,"PlanScanSliceNo", false },
  {0x2005,0x0048,"Philips MR Imaging DD 001",VR::IS,VM::VM1,"?MRReferenceAcquisitionNo",false },
  {0x2005,0x0049,"Philips MR Imaging DD 001",VR::IS,VM::VM1,"?MRReferenceChemicalShiftNo",false },
  {0x2005,0x004a,"Philips MR Imaging DD 001",VR::IS,VM::VM1,"?MRReferenceDynamicScanNo",false },
  {0x2005,0x004b,"Philips MR Imaging DD 001",VR::IS,VM::VM1,"?MRReferenceEchoNo",false },
  {0x2005,0x004c,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"?MRReferenceEntity",false },
  {0x2005,0x004d,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"?MRReferenceImageType",false },
  {0x2005,0x004e,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"?MRSlabFovRL",false },
  {0x2005,0x004f,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"?MRSlabOffcentreAP",false },
  {0x2005,0x0050,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"?MRSlabOffcentreFH",false },
  {0x2005,0x0051,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"?MRSlabOffcentreRL",false },
  {0x2005,0x0052,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"?MRSlabType",false },
  {0x2005,0x0053,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"?MRSlabViewAxis",false },
  {0x2005,0x0054,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"SVS Voxel Angulation AP (deg)",false },
  {0x2005,0x0055,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"SVS Voxel Angulation FH (deg)",false },
  {0x2005,0x0056,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"SVS Voxel Angulation LR (deg)",false },
  {0x2005,0x0057,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"SVS Voxel Size AP (mm)",false },
  {0x2005,0x0058,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"SVS Voxel Size FH (mm)",false },
  {0x2005,0x0059,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"SVS Voxel Size LR (mm)",false },
  {0x2005,0x005a,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"SVS Voxel Offcenter AP (mm)",false },
  {0x2005,0x005b,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"SVS Voxel Offcenter FH (mm)",false },
  {0x2005,0x005c,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"SVS Voxel Offcenter LR (mm)",false },
  {0x2005,0x005d,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"VolumeType", false },
  {0x2005,0x005e,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"VolumeViewAxis", false },
  {0x2005,0x005f,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"StudyOrigin", false },
  {0x2005,0x0060,"Philips MR Imaging DD 001",VR::IS,VM::VM1,"StudySequenceNumber", false },
  {0x2005,0x0061,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"ImagePrepulseType", false },
  {0x2005,0x0063,"Philips MR Imaging DD 001",VR::SS,VM::VM1,"fMRIStatusIndication", false },
  {0x2005,0x0064,"Philips MR Imaging DD 001",VR::IS,VM::VM1,"?MRReferencePhaseNo",false },
  {0x2005,0x0065,"Philips MR Imaging DD 001",VR::IS,VM::VM1,"?MRReferenceReconstructionNo",false },
  {0x2005,0x0066,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"?MRReferenceScanningSequence",false },
  {0x2005,0x0067,"Philips MR Imaging DD 001",VR::IS,VM::VM1,"?MRReferenceSliceNo",false },
  {0x2005,0x0068,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"?MRReferenceType",false },
  {0x2005,0x0069,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"?MRSlabAngulationAP",false },
  {0x2005,0x006a,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"?MRSlabAngulationFH",false },
  {0x2005,0x006b,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"?MRSlabAngulationRL",false },
  {0x2005,0x006c,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"?MRSlabFovAP",false },
  {0x2005,0x006d,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"?MRSlabFovFH",false },
  {0x2005,0x006e,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"?Scanning Sequence / Sequence Type?",false },
  {0x2005,0x006f,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"?MR Acquisition Type?",false },
  {0x2005,0x0070,"Philips MR Imaging DD 001",VR::LO,VM::VM1,"?MRSeriesHardcopyProtocolEV",false },
  {0x2005,0x0071,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"StackAngulationAP", false },
  {0x2005,0x0072,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"StackAngulationFH", false },
  {0x2005,0x0073,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"StackAngulationRL", false },
  {0x2005,0x0074,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"StackFovAP", false },
  {0x2005,0x0075,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"StackFovFH", false },
  {0x2005,0x0076,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"StackFovRL", false },
  {0x2005,0x0078,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"StackOffcentreAP", false },
  {0x2005,0x0079,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"StackOffcentreFH", false },
  {0x2005,0x007a,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"StackOffcentreRL", false },
  {0x2005,0x007b,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"StackPreparationDirection", false },
  {0x2005,0x007e,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"Z-Spacing (aka Stack Slice Distance)",false },
  {0x2005,0x0080,"Philips MR Imaging DD 001",VR::SQ,VM::VM1,"SeriesPlanScan", false },
  {0x2005,0x0081,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"StackViewAxis", false },
  {0x2005,0x0083,"Philips MR Imaging DD 001",VR::SQ,VM::VM1,"SeriesSlab", false },
  {0x2005,0x0084,"Philips MR Imaging DD 001",VR::SQ,VM::VM1,"SeriesReference", false },
  {0x2005,0x0085,"Philips MR Imaging DD 001",VR::SQ,VM::VM1,"SeriesVolume", false },
  {0x2005,0x0086,"Philips MR Imaging DD 001",VR::SS,VM::VM1,"NumberOfGeometry", false },
  {0x2005,0x009e,"Philips MR Imaging DD 001",VR::SQ,VM::VM1,"SeriesGeom", false },
  {0x2005,0x009f,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"MRSeriesSpectralSelectiveExcitationPulse", false },
  {0x2005,0x00a0,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"Philips Temporal Position Identifier (aka ImageDynamicScanBeginTime)",false },
  {0x2005,0x00a1,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"Syncra Scan Type",false },
  {0x2005,0x00a2,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"IsCOCA", false },
  {0x2005,0x00a3,"Philips MR Imaging DD 001",VR::IS,VM::VM1,"StackCoilID", false },
  {0x2005,0x00a4,"Philips MR Imaging DD 001",VR::IS,VM::VM1,"StackCBBCoil1", false },
  {0x2005,0x00a5,"Philips MR Imaging DD 001",VR::IS,VM::VM1,"StackCBBCoil2", false },
  {0x2005,0x00a6,"Philips MR Imaging DD 001",VR::IS,VM::VM1,"StackChannelCombi", false },
  {0x2005,0x00a7,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"StackCoilConn", false },
  {0x2005,0x00a8,"Philips MR Imaging DD 001",VR::DS,VM::VM1,"PrivateInversionTime", false },
  {0x2005,0x00a9,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"?Oversampling Phase (aka SeriesGeometryCorrection)?",false },
  {0x2005,0x00b0,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"Diffusion Direction RL (degree)",false },
  {0x2005,0x00b1,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"Diffusion Direction AP (degree)",false },
  {0x2005,0x00b2,"Philips MR Imaging DD 001",VR::FL,VM::VM1,"Diffusion Direction FH (degree)",false },
  {0x2005,0x00c0,"Philips MR Imaging DD 001",VR::CS,VM::VM1,"?SequenceVariant?", false },
  /* end of dups */
  {0x2005,0x0032,"Philips MR Imaging DD 002",VR::SQ,VM::VM1,"Protocol Data Sequence",false },
  {0x2005,0x0034,"Philips MR Imaging DD 002",VR::LT,VM::VM1,"?SeriesTransactionUID?",false },
  {0x2005,0x0037,"Philips MR Imaging DD 002",VR::LO,VM::VM1,"Protocol Data Name",false },
  {0x2005,0x0038,"Philips MR Imaging DD 002",VR::LO,VM::VM1,"?MRApplicationName",false },
  {0x2005,0x0039,"Philips MR Imaging DD 002",VR::LO,VM::VM1,"Protocol Data Type",false },
  {0x2005,0x0040,"Philips MR Imaging DD 002",VR::LO,VM::VM1,"?MRVersionStr",false },
  {0x2005,0x0041,"Philips MR Imaging DD 002",VR::LO,VM::VM1,"?MRCommentStr",false },
  {0x2005,0x0043,"Philips MR Imaging DD 002",VR::SL,VM::VM1,"Protocol Data Block Length (non-padded)",false },
  {0x2005,0x0044,"Philips MR Imaging DD 002",VR::OW,VM::VM1,"Protocol Data Block",false },
  {0x2005,0x0047,"Philips MR Imaging DD 002",VR::CS,VM::VM1,"Protocol Data Boolean",false },
  {0x2005,0x0099,"Philips MR Imaging DD 002",VR::UL,VM::VM1,"?NumberOfRequestExcerpts?",false },
  {0x2005,0x0000,"Philips MR Imaging DD 003",VR::UL,VM::VM1,"?NumberOfSOPCommon?",false },
  {0x2005,0x0001,"Philips MR Imaging DD 003",VR::UL,VM::VM1,"?NoOfFilmConsumption?",false },
  {0x2005,0x0013,"Philips MR Imaging DD 003",VR::UL,VM::VM1,"?NumberOfCodes?",false },
  {0x2005,0x0034,"Philips MR Imaging DD 003",VR::SL,VM::VM1,"?MRNumberOfImagePerSeriesRef",false },
  {0x2005,0x0043,"Philips MR Imaging DD 003",VR::SS,VM::VM1,"?NoDateOfLastCalibration?",false },
  {0x2005,0x0044,"Philips MR Imaging DD 003",VR::SS,VM::VM1,"?NoTimeOfLastCalibration?",false },
  {0x2005,0x0045,"Philips MR Imaging DD 003",VR::SS,VM::VM1,"?NrOfSoftwareVersion?",false },
  {0x2005,0x0046,"Philips MR Imaging DD 003",VR::SS,VM::VM1,"?",false },
  {0x2005,0x0047,"Philips MR Imaging DD 003",VR::SS,VM::VM1,"?NrOfPatientOtherNames?",false },
  {0x2005,0x0048,"Philips MR Imaging DD 003",VR::SS,VM::VM1,"?NrOfReqRecipeOfResults?",false },
  {0x2005,0x0049,"Philips MR Imaging DD 003",VR::SS,VM::VM1,"?NrOfSeriesOperatorsName?",false },
  {0x2005,0x0050,"Philips MR Imaging DD 003",VR::SS,VM::VM1,"?NrOfSeriesPerfPhysiName?",false },
  {0x2005,0x0051,"Philips MR Imaging DD 003",VR::SS,VM::VM1,"?NrOfStudyAdmittingDiagnosticDescr?",false },
  {0x2005,0x0052,"Philips MR Imaging DD 003",VR::SS,VM::VM1,"?NrOfStudyPatientContrastAllergies?",false },
  {0x2005,0x0053,"Philips MR Imaging DD 003",VR::SS,VM::VM1,"?NrOfStudyPatientMedicalAlerts?",false },
  {0x2005,0x0054,"Philips MR Imaging DD 003",VR::SS,VM::VM1,"?NrOfStudyPhysiciansOfRecord?",false },
  {0x2005,0x0055,"Philips MR Imaging DD 003",VR::SS,VM::VM1,"?NrOfStudyPhysiReadingStudy?",false },
  {0x2005,0x0056,"Philips MR Imaging DD 003",VR::SS,VM::VM1,"?MRNrSCSoftwareVersions",false },
  {0x2005,0x0057,"Philips MR Imaging DD 003",VR::SS,VM::VM1,"?MRNrRunningAttributes",false },
  {0x2005,0x0070,"Philips MR Imaging DD 003",VR::OW,VM::VM1,"FID storage (array of ieee floats)",false },
  {0x2005,0x0081,"Philips MR Imaging DD 003",VR::UI,VM::VM1,"?DefaultImageUID",false },
  {0x2005,0x0082,"Philips MR Imaging DD 003",VR::CS,VM::VM1,"?RunningAttributes",false },
  {0x2005,0x0000,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"?MRSpectrumExtraNumber",false },
  {0x2005,0x0001,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"?MRSpectrumKxCoordinate",false },
  {0x2005,0x0002,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"?MRSpectrumKyCoordinate",false },
  {0x2005,0x0003,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"?MRSpectrumLocationNumber",false },
  {0x2005,0x0004,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"?MRSpectrumMixNumber",false },
  {0x2005,0x0005,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"?MRSpectrumXCoordinate",false },
  {0x2005,0x0006,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"?MRSpectrumYCoordinate",false },
  {0x2005,0x0007,"Philips MR Imaging DD 004",VR::FL,VM::VM1,"?MRSpectrumDCLevel",false },
  {0x2005,0x0008,"Philips MR Imaging DD 004",VR::FL,VM::VM1,"?MRSpectrumNoiseLevel",false },
  {0x2005,0x0009,"Philips MR Imaging DD 004",VR::FL,VM::VM1,"?MRSpectrumBeginTime",false },
  {0x2005,0x0010,"Philips MR Imaging DD 004",VR::FL,VM::VM1,"Echo Time (ms) (Spectroscopy only)",false },
  {0x2005,0x0012,"Philips MR Imaging DD 004",VR::FL,VM::VM1,"Inversion Time (ms) (Spectroscopy only)",false },
  {0x2005,0x0013,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"?MRSpectrumNumber",false },
  {0x2005,0x0014,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"?MRSpectrumNumberOfAverages",false },
  {0x2005,0x0015,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"?MRSpectrumNumberOfSamples",false },
  {0x2005,0x0016,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"?MRSpectrumScanSequenceNumber",false },
  {0x2005,0x0017,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"?MRSpectrumNumberOfPeaks",false },
  {0x2005,0x0018,"Philips MR Imaging DD 004",VR::SQ,VM::VM1,"?MRSpectrumPeak",false },
  {0x2005,0x0025,"Philips MR Imaging DD 004",VR::CS,VM::VM1,"MRSpectroSIB0Correction",false },
  {0x2005,0x0026,"Philips MR Imaging DD 004",VR::FL,VM::VM1,"MRSpectroB0EchoTopPosition", false },
  {0x2005,0x0027,"Philips MR Imaging DD 004",VR::CS,VM::VM1,"MRSpectroComplexComponent", false },
  {0x2005,0x0028,"Philips MR Imaging DD 004",VR::CS,VM::VM1,"MRSpectroDataOrigin", false },
  {0x2005,0x0029,"Philips MR Imaging DD 004",VR::FL,VM::VM1,"MRSpectroEchoTopPosition", false },
  {0x2005,0x0030,"Philips MR Imaging DD 004",VR::CS,VM::VM1_n,"MRInPlaneTransforms", false },
  {0x2005,0x0031,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"MRNumberOfSpectraAcquired", false },
  {0x2005,0x0033,"Philips MR Imaging DD 004",VR::FL,VM::VM3,"MRPhaseEncodingEchoTopPositions", false },
  {0x2005,0x0034,"Philips MR Imaging DD 004",VR::CS,VM::VM1,"MRPhysicalQuantityForChemicalShift", false },
  {0x2005,0x0035,"Philips MR Imaging DD 004",VR::CS,VM::VM1,"MRPhysicalQuantitySpatial", false },
  {0x2005,0x0036,"Philips MR Imaging DD 004",VR::FL,VM::VM1,"MRReferenceFrequency", false },
  {0x2005,0x0037,"Philips MR Imaging DD 004",VR::FL,VM::VM1,"MRSampleOffset", false },
  {0x2005,0x0038,"Philips MR Imaging DD 004",VR::FL,VM::VM1,"MRSamplePitch", false },
  {0x2005,0x0039,"Philips MR Imaging DD 004",VR::SS,VM::VM2,"MRSearchIntervalForPeaks", false },
  {0x2005,0x0040,"Philips MR Imaging DD 004",VR::CS,VM::VM1,"MRSignalDomainForChemicalShift", false },
  {0x2005,0x0041,"Philips MR Imaging DD 004",VR::CS,VM::VM1,"MRSignalDomainSpatial", false },
  {0x2005,0x0042,"Philips MR Imaging DD 004",VR::CS,VM::VM1,"MRSignalType", false },
  {0x2005,0x0043,"Philips MR Imaging DD 004",VR::CS,VM::VM1,"MRSpectroAdditionalRotations", false },
  {0x2005,0x0044,"Philips MR Imaging DD 004",VR::SS,VM::VM1_n,"MRSpectroDisplayRanges", false },
  {0x2005,0x0045,"Philips MR Imaging DD 004",VR::CS,VM::VM1,"MRSpectroEchoAcquisition", false },
  {0x2005,0x0046,"Philips MR Imaging DD 004",VR::CS,VM::VM1,"MRSpectroFrequencyUnit", false },
  {0x2005,0x0047,"Philips MR Imaging DD 004",VR::FL,VM::VM1,"MRSpectroGamma", false },
  {0x2005,0x0048,"Philips MR Imaging DD 004",VR::CS,VM::VM1,"MRSpectroHiddenLineRemoval", false },
  {0x2005,0x0049,"Philips MR Imaging DD 004",VR::FL,VM::VM1,"MRSpectroHorizontalShift", false },
  {0x2005,0x0050,"Philips MR Imaging DD 004",VR::FL,VM::VM2,"MRSpectroHorizontalWindow", false },
  {0x2005,0x0051,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"MRSpectroNumberOfDisplayRanges", false },
  {0x2005,0x0052,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"MRSpectroNumberOfEchoPulses", false },
  {0x2005,0x0053,"Philips MR Imaging DD 004",VR::LO,VM::VM1_n,"MRSpectroProcessingHistory", false },
  {0x2005,0x0054,"Philips MR Imaging DD 004",VR::CS,VM::VM1,"MRSpectroScanType", false },
  {0x2005,0x0055,"Philips MR Imaging DD 004",VR::FL,VM::VM1_n,"MRSpectroSICSIntervals", false },
  {0x2005,0x0056,"Philips MR Imaging DD 004",VR::CS,VM::VM1,"MRSpectroSIMode", false },
  {0x2005,0x0057,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"Sampling Frequency (aka MRSpectroSpectralBW)",false },
  {0x2005,0x0058,"Philips MR Imaging DD 004",VR::LO,VM::VM1,"MRSpectroTitleLine", false },
  {0x2005,0x0059,"Philips MR Imaging DD 004",VR::FL,VM::VM1,"MRSpectroTurboEchoSpacing", false },
  {0x2005,0x0060,"Philips MR Imaging DD 004",VR::FL,VM::VM1,"MRSpectroVerticalShift", false },
  {0x2005,0x0061,"Philips MR Imaging DD 004",VR::FL,VM::VM2,"MRSpectroVerticalWindow", false },
  {0x2005,0x0062,"Philips MR Imaging DD 004",VR::FL,VM::VM1,"MRSpectroOffset", false },
  {0x2005,0x0063,"Philips MR Imaging DD 004",VR::FL,VM::VM1,"MRSpectrumPitch", false },
  {0x2005,0x0064,"Philips MR Imaging DD 004",VR::CS,VM::VM1,"MRVolumeSelection", false },
  {0x2005,0x0070,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"MRNoMixesSpectro", false },
  {0x2005,0x0071,"Philips MR Imaging DD 004",VR::SQ,VM::VM1,"MRSeriesSPMix", false },
  {0x2005,0x0072,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"?SPMixTResolution",false },
  {0x2005,0x0073,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"?SPMixKXResolution",false },
  {0x2005,0x0074,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"?SPMixKYResolution",false },
  {0x2005,0x0075,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"?SPMixFResolution",false },
  {0x2005,0x0076,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"?SPMixXResolution",false },
  {0x2005,0x0077,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"?SPMixYResolution",false },
  {0x2005,0x0078,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"?SPMixNoSpectraIntended",false },
  {0x2005,0x0079,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"?SPMixNoAverages",false },
  {0x2005,0x0081,"Philips MR Imaging DD 004",VR::IS,VM::VM1,"MRScanoGramSurveyNumberOfImages", false },
  {0x2005,0x0082,"Philips MR Imaging DD 004",VR::UL,VM::VM1,"MRNumberOfProcedureCodes", false },
  {0x2005,0x0083,"Philips MR Imaging DD 004",VR::CS,VM::VM1_n,"?SortAttributes",false },
  {0x2005,0x0084,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"?MRNrSortAttributes",false },
  {0x2005,0x0085,"Philips MR Imaging DD 004",VR::CS,VM::VM1,"?ImageDisplayDirection",false },
  {0x2005,0x0086,"Philips MR Imaging DD 004",VR::CS,VM::VM1,"?InsetScanogram",false },
  {0x2005,0x0087,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"?MRDisplayLayoutNrColumns",false },
  {0x2005,0x0088,"Philips MR Imaging DD 004",VR::SS,VM::VM1,"?MRDisplayLayoutNrRows",false },
  {0x2005,0x0089,"Philips MR Imaging DD 004",VR::SQ,VM::VM1,"?ViewingProtocol",false },
  {0x2005,0x0090,"Philips MR Imaging DD 004",VR::CS,VM::VM1,"MRStackCoilFunction", false },
  {0x2005,0x0091,"Philips MR Imaging DD 004",VR::PN,VM::VM1,"?PatientNameJobInParams",false },
  {0x2005,0x0092,"Philips MR Imaging DD 004",VR::IS,VM::VM1,"MRGeolinkID", false },
  {0x2005,0x0093,"Philips MR Imaging DD 004",VR::IS,VM::VM1,"MRStationNo", false },
  {0x2005,0x0095,"Philips MR Imaging DD 004",VR::ST,VM::VM1,"?ViewProcedureString",false },
  {0x2005,0x0096,"Philips MR Imaging DD 004",VR::CS,VM::VM1,"MRFlowImagesPresent", false },
  {0x2005,0x0097,"Philips MR Imaging DD 004",VR::LO,VM::VM1,"Anatomic Region Code Value",false },
  {0x2005,0x0098,"Philips MR Imaging DD 004",VR::CS,VM::VM1,"MRMobiviewEnabled", false },
  {0x2005,0x0099,"Philips MR Imaging DD 004",VR::CS,VM::VM1,"MRIViewBoldEnabled", false },
  /* new group */
  {0x2005,0x0000,"Philips MR Imaging DD 005",VR::CS,VM::VM1,"MRVolumeViewEnabled", false },
  {0x2005,0x0001,"Philips MR Imaging DD 005",VR::UL,VM::VM1,"MRNumberOfStudyReference", false },
  {0x2005,0x0002,"Philips MR Imaging DD 005",VR::SQ,VM::VM1,"SPSCode", false },
  {0x2005,0x0003,"Philips MR Imaging DD 005",VR::UL,VM::VM1,"MRNumberOfSPSCodes", false },
  {0x2005,0x0004,"Philips MR Imaging DD 005",VR::SS,VM::VM1,"?",false },
  {0x2005,0x0006,"Philips MR Imaging DD 005",VR::SS,VM::VM1,"?",false },
  {0x2005,0x0007,"Philips MR Imaging DD 005",VR::SS,VM::VM1,"MRNrOfSpecificCharacterSet", false },
  {0x2005,0x0009,"Philips MR Imaging DD 005",VR::DS,VM::VM1,"Rescale Intercept Philips",false },
  {0x2005,0x000a,"Philips MR Imaging DD 005",VR::DS,VM::VM1,"Rescale Slope Philips",false },
  {0x2005,0x000b,"Philips MR Imaging DD 005",VR::LO,VM::VM1,"Rescale Type Philips",false },
  {0x2005,0x000e,"Philips MR Imaging DD 005",VR::SQ,VM::VM1,"PrivateSharedSq", false },
  {0x2005,0x000f,"Philips MR Imaging DD 005",VR::SQ,VM::VM1,"Acquisition Parameters (Philips) Sequence",false },
  {0x2005,0x0010,"Philips MR Imaging DD 005",VR::IS,VM::VM1,"?MFConvTreatSpectorMixNo",false },
  {0x2005,0x0011,"Philips MR Imaging DD 005",VR::UI,VM::VM1,"MFPrivateReferencedSOPInstanceUID", false },
  {0x2005,0x0012,"Philips MR Imaging DD 005",VR::IS,VM::VM1,"MRImageDiffBValueNumber", false },
  {0x2005,0x0013,"Philips MR Imaging DD 005",VR::IS,VM::VM1,"MRImageGradientOrientationNumber", false },
  {0x2005,0x0014,"Philips MR Imaging DD 005",VR::SL,VM::VM1,"MRSeriesNrOfDiffBValues", false },
  {0x2005,0x0015,"Philips MR Imaging DD 005",VR::SL,VM::VM1,"MRSeriesNrOfDiffGradOrients", false },
  {0x2005,0x0016,"Philips MR Imaging DD 005",VR::CS,VM::VM1,"MRSeriesPlanMode", false },
  {0x2005,0x0017,"Philips MR Imaging DD 005",VR::FD,VM::VM1,"B_matrix / DiffusionBMatrix", false },
  {0x2005,0x0018,"Philips MR Imaging DD 005",VR::CS,VM::VM3,"Operating Mode Type",false },
  {0x2005,0x0019,"Philips MR Imaging DD 005",VR::CS,VM::VM3,"Operating Mode",false },
  {0x2005,0x001a,"Philips MR Imaging DD 005",VR::CS,VM::VM1,"MRFatSaturationTechnique", false },
  {0x2005,0x001b,"Philips MR Imaging DD 005",VR::IS,VM::VM1,"MRVersionNumberDeletedImages", false },
  {0x2005,0x001c,"Philips MR Imaging DD 005",VR::IS,VM::VM1,"MRVersionNumberDeletedSpectra", false },
  {0x2005,0x001d,"Philips MR Imaging DD 005",VR::IS,VM::VM1,"MRVersionNumberDeletedBlobsets", false },
  {0x2005,0x001e,"Philips MR Imaging DD 005",VR::UL,VM::VM1,"LUT1Offset?",false },
  {0x2005,0x001f,"Philips MR Imaging DD 005",VR::UL,VM::VM1,"LUT1Range?",false },
  {0x2005,0x0020,"Philips MR Imaging DD 005",VR::UL,VM::VM1,"?LUT1BeginColor",false },
  {0x2005,0x0021,"Philips MR Imaging DD 005",VR::UL,VM::VM1,"?LUT1EndColor",false },
  {0x2005,0x0022,"Philips MR Imaging DD 005",VR::UL,VM::VM1,"?LUT2Offset",false },
  {0x2005,0x0023,"Philips MR Imaging DD 005",VR::UL,VM::VM1,"?LUT2Range",false },
  {0x2005,0x0024,"Philips MR Imaging DD 005",VR::UL,VM::VM1,"?LUT2BeginColor",false },
  {0x2005,0x0025,"Philips MR Imaging DD 005",VR::UL,VM::VM1,"?LUT2EndColor",false },
  {0x2005,0x0026,"Philips MR Imaging DD 005",VR::CS,VM::VM1,"ViewingHardcopyOnly", false },
  {0x2005,0x0027,"Philips MR Imaging DD 005",VR::CS,VM::VM1,"?PrivateEMR",false },
  {0x2005,0x0028,"Philips MR Imaging DD 005",VR::SL,VM::VM1,"MRSeriesNrOfLabelTypes", false },
  {0x2005,0x0029,"Philips MR Imaging DD 005",VR::CS,VM::VM1,"?MRImageLabelType",false },
  {0x2005,0x002a,"Philips MR Imaging DD 005",VR::CS,VM::VM1,"ExamPrintStatus", false },
  {0x2005,0x002b,"Philips MR Imaging DD 005",VR::CS,VM::VM1,"ExamExportStatus", false },
  {0x2005,0x002c,"Philips MR Imaging DD 005",VR::CS,VM::VM1,"ExamStorageCommitStatus", false },
  {0x2005,0x002d,"Philips MR Imaging DD 005",VR::CS,VM::VM1,"ExamMediaWriteStatus", false },
  {0x2005,0x002e,"Philips MR Imaging DD 005",VR::FL,VM::VM1,"?MRSeriesDBdt",false },
  {0x2005,0x002f,"Philips MR Imaging DD 005",VR::FL,VM::VM1,"?MRSeriesProtonSAR",false },
  {0x2005,0x0030,"Philips MR Imaging DD 005",VR::FL,VM::VM1,"?MRSeriesNonProtonSAR",false },
  {0x2005,0x0031,"Philips MR Imaging DD 005",VR::FL,VM::VM1,"?MRSeriesLocalSAR",false },
  {0x2005,0x0032,"Philips MR Imaging DD 005",VR::CS,VM::VM1,"MRSeriesSafetyOverrideMode", false },
  {0x2005,0x0035,"Philips MR Imaging DD 005",VR::CS,VM::VM1,"SpectroExamcard", false },
  {0x2005,0x0036,"Philips MR Imaging DD 005",VR::UI,VM::VM1,"?MRRefSeriesInstanceUID",false },
  {0x2005,0x0037,"Philips MR Imaging DD 005",VR::CS,VM::VM1,"?ColorLUTType",false },
  {0x2005,0x0038,"Philips MR Imaging DD 005",VR::LT,VM::VM1,"?",false },
  {0x2005,0x0039,"Philips MR Imaging DD 005",VR::LT,VM::VM1,"?",false },
  {0x2005,0x003a,"Philips MR Imaging DD 005",VR::LT,VM::VM1,"DataDictionaryContentsVersion", false },
  {0x2005,0x003b,"Philips MR Imaging DD 005",VR::CS,VM::VM1,"MRIsCoilSurvey", false },
  {0x2005,0x003c,"Philips MR Imaging DD 005",VR::FL,VM::VM1,"MRStackTablePosLong", false },
  {0x2005,0x003d,"Philips MR Imaging DD 005",VR::FL,VM::VM1,"MRStackTablePosLat", false },
  {0x2005,0x003e,"Philips MR Imaging DD 005",VR::FL,VM::VM1,"MRStackPosteriorCoilPos", false },
  {0x2005,0x003f,"Philips MR Imaging DD 005",VR::CS,VM::VM1,"AIMDLimitsApplied", false },
  {0x2005,0x0040,"Philips MR Imaging DD 005",VR::FL,VM::VM1,"AIMDHeadSARLimit", false },
  {0x2005,0x0041,"Philips MR Imaging DD 005",VR::FL,VM::VM1,"AIMDWholeBodySARLimit", false },
  {0x2005,0x0042,"Philips MR Imaging DD 005",VR::FL,VM::VM1,"AIMDB1RMSLimit", false },
  {0x2005,0x0043,"Philips MR Imaging DD 005",VR::FL,VM::VM1,"AIMDdbDtLimit", false },
  {0x2005,0x0044,"Philips MR Imaging DD 005",VR::IS,VM::VM1,"TFEFactor", false },
  {0x2005,0x0045,"Philips MR Imaging DD 005",VR::CS,VM::VM1,"AttenuationCorrection", false },
  {0x2005,0x0046,"Philips MR Imaging DD 005",VR::FL,VM::VM1,"?FWHMShim",false },
  {0x2005,0x0047,"Philips MR Imaging DD 005",VR::FL,VM::VM1,"PowerOptimization", false },
  {0x2005,0x0048,"Philips MR Imaging DD 005",VR::FL,VM::VM1,"?CoilQ",false },
  {0x2005,0x0049,"Philips MR Imaging DD 005",VR::FL,VM::VM1,"?ReceiverGain",false },
  {0x2005,0x004a,"Philips MR Imaging DD 005",VR::FL,VM::VM1,"DataWindowDuration", false },
  {0x2005,0x004b,"Philips MR Imaging DD 005",VR::FL,VM::VM1,"MixingTime", false },
  {0x2005,0x004c,"Philips MR Imaging DD 005",VR::FL,VM::VM1,"FirstEchoTime", false },
  {0x2005,0x004d,"Philips MR Imaging DD 005",VR::CS,VM::VM1,"IsB0Series", false },
  {0x2005,0x004e,"Philips MR Imaging DD 005",VR::CS,VM::VM1,"IsB1Series", false },
  {0x2005,0x004f,"Philips MR Imaging DD 005",VR::CS,VM::VM1,"VolumeSelect", false },
  {0x2005,0x0050,"Philips MR Imaging DD 005",VR::SS,VM::VM1,"MRNrOfPatientOtherIDs", false },
  {0x2005,0x0051,"Philips MR Imaging DD 005",VR::IS,VM::VM1,"?", false },
  {0x2005,0x0052,"Philips MR Imaging DD 005",VR::UI,VM::VM1,"?", false },
  {0x2005,0x0055,"Philips MR Imaging DD 005",VR::FD,VM::VM3,"ImageVelocityEncodingDirection", false },
  {0x2005,0x0092,"Philips MR Imaging DD 005",VR::FL,VM::VM1,"Specific Energy Dose",false },
  {0x2005,0x0053,"Philips MR Imaging DD 006",VR::FL,VM::VM1,"MRE Frequency",false },
  {0x2005,0x0054,"Philips MR Imaging DD 006",VR::FL,VM::VM1,"MRE Amplitude",false },
  {0x2005,0x0055,"Philips MR Imaging DD 006",VR::FL,VM::VM1,"MREMEG Frequency",false },
  {0x2005,0x0056,"Philips MR Imaging DD 006",VR::FL,VM::VM1,"MREMEG Pairs",false },
  {0x2005,0x0057,"Philips MR Imaging DD 006",VR::CS,VM::VM1,"MREMEG Direction",false },
  {0x2005,0x0058,"Philips MR Imaging DD 006",VR::FL,VM::VM1,"MREMEG Amplitude",false },
  {0x2005,0x0059,"Philips MR Imaging DD 006",VR::FL,VM::VM1,"MRE Number of Phase Delays",false },
  {0x2005,0x0060,"Philips MR Imaging DD 006",VR::IS,VM::VM1,"MRE Number of Motion Cycles",false },
  {0x2005,0x0061,"Philips MR Imaging DD 006",VR::FL,VM::VM1,"MRE Motion Meg Phase Delay",false },
  {0x2005,0x0062,"Philips MR Imaging DD 006",VR::LT,VM::VM1,"MRE Inversion Algorithm Version",false },
  {0x2005,0x0063,"Philips MR Imaging DD 006",VR::CS,VM::VM1,"Sagittal Slice Order",false },
  {0x2005,0x0064,"Philips MR Imaging DD 006",VR::CS,VM::VM1,"Coronal Slice Order",false },
  {0x2005,0x0065,"Philips MR Imaging DD 006",VR::CS,VM::VM1,"Transversal Slice Order",false },
  {0x2005,0x0066,"Philips MR Imaging DD 006",VR::CS,VM::VM1,"Series Orientation ",false },
  {0x2005,0x0067,"Philips MR Imaging DD 006",VR::IS,VM::VM1,"MR Stack Reverse ",false },
  {0x2005,0x0068,"Philips MR Imaging DD 006",VR::IS,VM::VM1,"MRE Phase Delay Number",false },
  {0x2005,0x0071,"Philips MR Imaging DD 006",VR::IS,VM::VM1,"Number Of Inversion Delays",false },
  {0x2005,0x0072,"Philips MR Imaging DD 006",VR::FL,VM::VM1,"Inversion Delay Time",false },
  {0x2005,0x0073,"Philips MR Imaging DD 006",VR::IS,VM::VM1,"Inversion Delay Number",false },
  {0x2005,0x0074,"Philips MR Imaging DD 006",VR::DS,VM::VM1,"Max DB DT",false },
  {0x2005,0x0075,"Philips MR Imaging DD 006",VR::DS,VM::VM1,"Max SAR",false },
  {0x2005,0x0076,"Philips MR Imaging DD 006",VR::LT,VM::VM1,"SAR Type",false },
  {0x2005,0x0078,"Philips MR Imaging DD 006",VR::CS,VM::VM1,"Metal Implant Status",false },
  {0x2005,0x0079,"Philips MR Imaging DD 006",VR::CS,VM::VM1_n,"Orientation Mirror Flip",false },
  {0x2005,0x0081,"Philips MR Imaging DD 006",VR::CS,VM::VM1,"SAR Operation Mode",false },
  {0x2005,0x0082,"Philips MR Imaging DD 006",VR::IS,VM::VM1,"Spatial Gradient",false },
  {0x2005,0x0083,"Philips MR Imaging DD 006",VR::LT,VM::VM1,"Additional Constraints",false },
  {0x2005,0x0085,"Philips MR Imaging DD 006",VR::DS,VM::VM1,"PIIM_GRADIENT_SLEW_RATE",false },
  {0x2005,0x0086,"Philips MR Imaging DD 006",VR::LT,VM::VM1,"?",false },
  {0x2005,0x0087,"Philips MR Imaging DD 006",VR::DS,VM::VM1,"PIIM_MR_STUDY_B1RMS",false },
  {0x2005,0x0095,"Philips MR Imaging DD 006",VR::CS,VM::VM1,"?",false },
  {0x2005,0x0096,"Philips MR Imaging DD 006",VR::IS,VM::VM1,"?",false },
  {0x2005,0x0097,"Philips MR Imaging DD 006",VR::CS,VM::VM1,"?",false },
  {0x2005,0x0099,"Philips MR Imaging DD 006",VR::SL,VM::VM1,"?",false },
  {0x2005,0x0000,"Philips MR Imaging DD 007",VR::CS,VM::VM1,"?",false },
  {0x2005,0x0001,"Philips MR Imaging DD 007",VR::FL,VM::VM1,"?",false },
  {0x2005,0x0002,"Philips MR Imaging DD 007",VR::FL,VM::VM1,"?",false },
  {0x2005,0x0003,"Philips MR Imaging DD 007",VR::FL,VM::VM1,"?",false },
  /* new group */
  {0x7043,0x0000,"Philips NM Private Group",VR::SH,VM::VM1,"?",false },
  {0x0511,0x0000,"Philips PET Private Group",VR::US,VM::VM1,"?",false },
  {0x0511,0x0001,"Philips PET Private Group",VR::US,VM::VM1,"?",false },
  {0x0511,0x0002,"Philips PET Private Group",VR::OB,VM::VM1,"?",false },
  {0x0511,0x0003,"Philips PET Private Group",VR::OB,VM::VM1,"?",false },
  {0x0511,0x0032,"Philips PET Private Group",VR::DS,VM::VM1,"?",false },
  {0x0511,0x0050,"Philips PET Private Group",VR::DS,VM::VM1,"?",false },
  {0x7053,0x0000,"Philips PET Private Group",VR::DS,VM::VM1,"SUV Scale Factor",false },
  {0x7053,0x0001,"Philips PET Private Group",VR::OB,VM::VM1,"Private",false },
  {0x7053,0x0002,"Philips PET Private Group",VR::OB,VM::VM1,"Private",false },
  {0x7053,0x0003,"Philips PET Private Group",VR::ST,VM::VM1,"Original image file name",false },
  {0x7053,0x0004,"Philips PET Private Group",VR::OB,VM::VM1,"File Data",false },
  {0x7053,0x0005,"Philips PET Private Group",VR::LO,VM::VM1,"Worklist Info File Name",false },
  {0x7053,0x0006,"Philips PET Private Group",VR::OB,VM::VM1,"Private",false },
  {0x7053,0x0007,"Philips PET Private Group",VR::SQ,VM::VM1,"Acquisition File Sequence",false },
  {0x7053,0x0008,"Philips PET Private Group",VR::SQ,VM::VM1,"?",false },
  {0x7053,0x0009,"Philips PET Private Group",VR::DS,VM::VM1,"Activity Concentration Scale Factor",false },
  {0x7053,0x000b,"Philips PET Private Group",VR::DS,VM::VM1,"?",false },
  {0x7053,0x000c,"Philips PET Private Group",VR::DS,VM::VM1,"?",false },
  {0x7053,0x000d,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x000e,"Philips PET Private Group",VR::DS,VM::VM1,"?",false },
  {0x7053,0x000f,"Philips PET Private Group",VR::UL,VM::VM1,"Segment Size",false },
  {0x7053,0x0010,"Philips PET Private Group",VR::US,VM::VM1,"Segment Number",false },
  {0x7053,0x0011,"Philips PET Private Group",VR::US,VM::VM1,"Number of Segments",false },
  {0x7053,0x0012,"Philips PET Private Group",VR::SQ,VM::VM1,"File Data Sequence",false },
  {0x7053,0x0013,"Philips PET Private Group",VR::SS,VM::VM1,"Private",false },
  {0x7053,0x0014,"Philips PET Private Group",VR::SS,VM::VM1,"Private",false },
  {0x7053,0x0015,"Philips PET Private Group",VR::SS,VM::VM1,"Private",false },
  {0x7053,0x0016,"Philips PET Private Group",VR::SS,VM::VM1,"Private",false },
  {0x7053,0x0017,"Philips PET Private Group",VR::SS,VM::VM1,"Private",false },
  {0x7053,0x0018,"Philips PET Private Group",VR::SS,VM::VM1,"Private",false },
  {0x7053,0x0050,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x0051,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x0052,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x0053,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x0054,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x0055,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x0056,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x0057,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x0058,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x0059,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x005a,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x005b,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x005c,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x005d,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x005e,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x005f,"Philips PET Private Group",VR::LO,VM::VM1,"?",false },
  {0x7053,0x0060,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x0061,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x0062,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x0063,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x0064,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x0065,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x0066,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x0067,"Philips PET Private Group",VR::LO,VM::VM1,"?",false },
  {0x7053,0x0068,"Philips PET Private Group",VR::LO,VM::VM1,"?",false },
  {0x7053,0x0069,"Philips PET Private Group",VR::LO,VM::VM1,"?",false },
  {0x7053,0x006a,"Philips PET Private Group",VR::LT,VM::VM1,"?",false },
  {0x7053,0x006b,"Philips PET Private Group",VR::SH,VM::VM1,"?",false },
  {0x7053,0x006c,"Philips PET Private Group",VR::LO,VM::VM1,"?",false },
  {0x7053,0x006d,"Philips PET Private Group",VR::FL,VM::VM1,"?",false },
  {0x7053,0x006e,"Philips PET Private Group",VR::FL,VM::VM1,"?",false },
  {0x7053,0x006f,"Philips PET Private Group",VR::FL,VM::VM1,"?",false },
  {0x7053,0x0070,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x0071,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x0072,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x0073,"Philips PET Private Group",VR::SH,VM::VM1,"?",false },
  {0x7053,0x0074,"Philips PET Private Group",VR::SH,VM::VM1,"?",false },
  {0x7053,0x0075,"Philips PET Private Group",VR::SH,VM::VM1,"?",false },
  {0x7053,0x0076,"Philips PET Private Group",VR::SH,VM::VM1,"?",false },
  {0x7053,0x0077,"Philips PET Private Group",VR::FL,VM::VM1,"?",false },
  {0x7053,0x0078,"Philips PET Private Group",VR::FL,VM::VM1,"?",false },
  {0x7053,0x00c2,"Philips PET Private Group",VR::UI,VM::VM1,"PET-CT Multi Modality Name",false },
  {0x7053,0x00d0,"Philips PET Private Group",VR::SS,VM::VM1,"?",false },
  {0x7053,0x00d1,"Philips PET Private Group",VR::FL,VM::VM1,"?",false },
  {0x7053,0x00d2,"Philips PET Private Group",VR::DT,VM::VM1,"?",false },
  {0x7053,0x00d3,"Philips PET Private Group",VR::DT,VM::VM1,"?",false },
  {0x7053,0x00d5,"Philips PET Private Group",VR::FL,VM::VM1,"?",false },
  {0x7053,0x00d6,"Philips PET Private Group",VR::FL,VM::VM1,"?",false },
  {0x7053,0x00d7,"Philips PET Private Group",VR::FL,VM::VM1,"?",false },
  {0x7053,0x00d8,"Philips PET Private Group",VR::FL,VM::VM1,"?",false },
  {0x7053,0x00d9,"Philips PET Private Group",VR::FL,VM::VM1,"?",false },
  {0x200b,0x0000,"Philips RAD Imaging DD 001",VR::PN,VM::VM1,"?",false },
  {0x200b,0x0001,"Philips RAD Imaging DD 001",VR::US,VM::VM1,"?",false },
  {0x200b,0x0002,"Philips RAD Imaging DD 001",VR::US,VM::VM1,"?",false },
  {0x200b,0x0005,"Philips RAD Imaging DD 001",VR::IS,VM::VM1,"?",false },
  {0x200b,0x0011,"Philips RAD Imaging DD 001",VR::LO,VM::VM1,"?",false },
  {0x200b,0x0027,"Philips RAD Imaging DD 001",VR::DT,VM::VM1,"?",false },
  {0x200b,0x0028,"Philips RAD Imaging DD 001",VR::DS,VM::VM1,"?",false },
  {0x200b,0x0029,"Philips RAD Imaging DD 001",VR::DS,VM::VM1,"?",false },
  {0x200b,0x002a,"Philips RAD Imaging DD 001",VR::UL,VM::VM1,"?",false },
  {0x200b,0x002b,"Philips RAD Imaging DD 001",VR::DA,VM::VM1,"?",false },
  {0x200b,0x002c,"Philips RAD Imaging DD 001",VR::TM,VM::VM1,"?",false },
  {0x200b,0x002d,"Philips RAD Imaging DD 001",VR::LO,VM::VM1,"?",false },
  {0x200b,0x003b,"Philips RAD Imaging DD 001",VR::LO,VM::VM1,"?",false },
  {0x200b,0x0040,"Philips RAD Imaging DD 001",VR::SH,VM::VM1,"?",false },
  {0x200b,0x0041,"Philips RAD Imaging DD 001",VR::SH,VM::VM1,"?",false },
  {0x200b,0x0042,"Philips RAD Imaging DD 001",VR::UI,VM::VM1,"?",false },
  {0x200b,0x0043,"Philips RAD Imaging DD 001",VR::UI,VM::VM1,"?",false },
  {0x200b,0x0047,"Philips RAD Imaging DD 001",VR::DA,VM::VM1,"?",false },
  {0x200b,0x0048,"Philips RAD Imaging DD 001",VR::SH,VM::VM1,"?",false },
  {0x200b,0x004c,"Philips RAD Imaging DD 001",VR::SH,VM::VM1,"?",false },
  {0x200b,0x004d,"Philips RAD Imaging DD 001",VR::SH,VM::VM1,"?",false },
  {0x200b,0x004f,"Philips RAD Imaging DD 001",VR::DT,VM::VM1,"?",false },
  {0x200b,0x0052,"Philips RAD Imaging DD 001",VR::SH,VM::VM1,"?",false },
  {0x200b,0x0000,"Philips RAD Imaging DD 097",VR::ST,VM::VM1,"?",false },
  {0x200b,0x0001,"Philips RAD Imaging DD 097",VR::CS,VM::VM1,"?",false },
  {0x200b,0x0002,"Philips RAD Imaging DD 097",VR::SS,VM::VM1,"?",false },
  {0x200b,0x0050,"Philips RAD Imaging DD 097",VR::SS,VM::VM1,"?",false },
  {0x200b,0x0051,"Philips RAD Imaging DD 097",VR::SS,VM::VM1,"?",false },
  {0x200b,0x0052,"Philips RAD Imaging DD 097",VR::SS,VM::VM1,"?",false },
  {0x200b,0x0053,"Philips RAD Imaging DD 097",VR::SS,VM::VM1,"?",false },
  {0x200b,0x0054,"Philips RAD Imaging DD 097",VR::ST,VM::VM1,"?",false },
  {0x200b,0x0060,"Philips RAD Imaging DD 097",VR::LO,VM::VM1,"?",false },
  {0x200b,0x0063,"Philips RAD Imaging DD 097",VR::LT,VM::VM1,"?",false },
  {0x200b,0x0065,"Philips RAD Imaging DD 097",VR::SS,VM::VM1,"?",false },
  {0x200b,0x006e,"Philips RAD Imaging DD 097",VR::US,VM::VM1,"?",false },
  {0x200b,0x006f,"Philips RAD Imaging DD 097",VR::UI,VM::VM1,"?",false },
  {0x200b,0x0072,"Philips RAD Imaging DD 097",VR::FD,VM::VM1,"?",false },
  {0x200b,0x0073,"Philips RAD Imaging DD 097",VR::SS,VM::VM1,"?",false },
  {0x200b,0x0074,"Philips RAD Imaging DD 097",VR::IS,VM::VM1,"?",false },
  {0x200b,0x0075,"Philips RAD Imaging DD 097",VR::CS,VM::VM1,"?",false },
  {0x200b,0x0076,"Philips RAD Imaging DD 097",VR::SH,VM::VM1,"?",false },
  {0x200b,0x0078,"Philips RAD Imaging DD 097",VR::IS,VM::VM1_n,"?",false },
  {0x200b,0x0079,"Philips RAD Imaging DD 097",VR::IS,VM::VM1,"?",false },
  {0x200b,0x007a,"Philips RAD Imaging DD 097",VR::IS,VM::VM1,"?",false },
  {0x200b,0x007b,"Philips RAD Imaging DD 097",VR::US,VM::VM1,"?",false },
  {0x200b,0x007c,"Philips RAD Imaging DD 097",VR::US,VM::VM1,"?",false },
  {0x200b,0x007d,"Philips RAD Imaging DD 097",VR::IS,VM::VM1_n,"?",false },
  {0x200b,0x007e,"Philips RAD Imaging DD 097",VR::UI,VM::VM1,"?",false },
  {0x200b,0x0081,"Philips RAD Imaging DD 097",VR::SH,VM::VM1,"?",false },
  {0x200b,0x0082,"Philips RAD Imaging DD 097",VR::SH,VM::VM1,"?",false },
  {0x200b,0x0085,"Philips RAD Imaging DD 097",VR::IS,VM::VM1,"?",false },
  {0x200b,0x0086,"Philips RAD Imaging DD 097",VR::CS,VM::VM1,"?",false },
  {0x200b,0x0088,"Philips RAD Imaging DD 097",VR::CS,VM::VM1,"?",false },
  {0x200b,0x0089,"Philips RAD Imaging DD 097",VR::LO,VM::VM1,"?",false },
  {0x200b,0x0090,"Philips RAD Imaging DD 097",VR::DS,VM::VM1,"?",false },
  {0x200b,0x0096,"Philips RAD Imaging DD 097",VR::SH,VM::VM1,"?",false },
  {0x200b,0x0099,"Philips RAD Imaging DD 097",VR::SH,VM::VM1,"?",false },
  {0x200b,0x009a,"Philips RAD Imaging DD 097",VR::FD,VM::VM1,"?",false },
  {0x200b,0x009b,"Philips RAD Imaging DD 097",VR::FD,VM::VM1,"?",false },
  {0x200b,0x009f,"Philips RAD Imaging DD 097",VR::SQ,VM::VM1,"?",false },
  {0x200b,0x00a0,"Philips RAD Imaging DD 097",VR::LT,VM::VM1,"?",false },
  {0x200b,0x00a1,"Philips RAD Imaging DD 097",VR::CS,VM::VM1,"?bool?",false },
  {0x0029,0x0050,"Philips US Imaging 60",VR::CS,VM::VM1,"?COLOR/CPA/NONE?",false },
  {0x0031,0x0030,"Philips US Imaging 60",VR::UL,VM::VM1,"Private data",false },
  {0x0031,0x0031,"Philips US Imaging 60",VR::UL,VM::VM1,"Private data",false },
  {0x0031,0x0032,"Philips US Imaging 60",VR::UL,VM::VM1,"Private data",false },
  {0x200d,0x0005,"Philips US Imaging DD 017",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0007,"Philips US Imaging DD 021",VR::LO,VM::VM1,"?CUSTOM_SELECTION_13?",false },
  {0x200d,0x0037,"Philips US Imaging DD 023",VR::DA,VM::VM1,"?",false },
  {0x200d,0x0038,"Philips US Imaging DD 023",VR::TM,VM::VM1,"?",false },
  {0x200d,0x0045,"Philips US Imaging DD 023",VR::IS,VM::VM1,"?",false },
  {0x200d,0x0059,"Philips US Imaging DD 023",VR::IS,VM::VM1,"?",false },
  {0x200d,0x0000,"Philips US Imaging DD 033",VR::OB,VM::VM1,"Philips Confidential v.1",false },
  {0x200d,0x0001,"Philips US Imaging DD 033",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0002,"Philips US Imaging DD 033",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0003,"Philips US Imaging DD 033",VR::LO,VM::VM6,"?",false },
  {0x200d,0x0004,"Philips US Imaging DD 033",VR::LO,VM::VM6,"?",false },
  {0x200d,0x0005,"Philips US Imaging DD 033",VR::LO,VM::VM1_n,"?",false },
  {0x200d,0x0006,"Philips US Imaging DD 033",VR::LO,VM::VM6,"?",false },
  {0x200d,0x0007,"Philips US Imaging DD 033",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0008,"Philips US Imaging DD 033",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0009,"Philips US Imaging DD 033",VR::LO,VM::VM1,"?",false },
  {0x200d,0x000a,"Philips US Imaging DD 033",VR::LO,VM::VM1,"?",false },
  {0x200d,0x000b,"Philips US Imaging DD 033",VR::OB,VM::VM1,"Raw Data US",false },
  {0x200d,0x000d,"Philips US Imaging DD 033",VR::LO,VM::VM1,"UDM USD DATATYPE DIN",false },
  {0x200d,0x000f,"Philips US Imaging DD 033",VR::OB,VM::VM1,"XML: UAA / CaptureContext",false },
  {0x200d,0x0010,"Philips US Imaging DD 033",VR::IS,VM::VM1,"Number of Frames",false },
  {0x200d,0x0011,"Philips US Imaging DD 033",VR::IS,VM::VM1,"Max Alloc Buffer (Zlib)",false },
  {0x200d,0x0014,"Philips US Imaging DD 033",VR::IS,VM::VM1,"?",false },
  {0x200d,0x0021,"Philips US Imaging DD 033",VR::IS,VM::VM1,"?Zero?",false },
  {0x200d,0x0022,"Philips US Imaging DD 033",VR::IS,VM::VM1,"?Zero?",false },
  {0x200d,0x0001,"Philips US Imaging DD 034",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0002,"Philips US Imaging DD 034",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0003,"Philips US Imaging DD 034",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0004,"Philips US Imaging DD 034",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0005,"Philips US Imaging DD 034",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0006,"Philips US Imaging DD 034",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0007,"Philips US Imaging DD 034",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0008,"Philips US Imaging DD 034",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0009,"Philips US Imaging DD 034",VR::LO,VM::VM1,"?",false },
  {0x200d,0x000a,"Philips US Imaging DD 034",VR::LO,VM::VM1,"?",false },
  {0x200d,0x000b,"Philips US Imaging DD 034",VR::LO,VM::VM1,"?",false },
  {0x200d,0x000c,"Philips US Imaging DD 034",VR::LO,VM::VM1,"?",false },
  {0x200d,0x000d,"Philips US Imaging DD 034",VR::LO,VM::VM1,"?",false },
  {0x200d,0x000e,"Philips US Imaging DD 034",VR::LO,VM::VM1,"?",false },
  {0x200d,0x000f,"Philips US Imaging DD 034",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0010,"Philips US Imaging DD 034",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0021,"Philips US Imaging DD 034",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0022,"Philips US Imaging DD 034",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0023,"Philips US Imaging DD 034",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0024,"Philips US Imaging DD 034",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0001,"Philips US Imaging DD 035",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0002,"Philips US Imaging DD 035",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0003,"Philips US Imaging DD 035",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0004,"Philips US Imaging DD 035",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0005,"Philips US Imaging DD 035",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0007,"Philips US Imaging DD 035",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0008,"Philips US Imaging DD 035",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0009,"Philips US Imaging DD 035",VR::LO,VM::VM1,"?",false },
  {0x200d,0x000a,"Philips US Imaging DD 035",VR::LO,VM::VM1,"?",false },
  {0x200d,0x000b,"Philips US Imaging DD 035",VR::LO,VM::VM1,"?",false },
  {0x200d,0x000c,"Philips US Imaging DD 035",VR::LO,VM::VM1,"?",false },
  {0x200d,0x000d,"Philips US Imaging DD 035",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0001,"Philips US Imaging DD 036",VR::LO,VM::VM3,"Cols/Rows/Frames",false },
  {0x200d,0x0002,"Philips US Imaging DD 036",VR::LO,VM::VM3,"Cols/Rows/Frames",false },
  {0x200d,0x0003,"Philips US Imaging DD 036",VR::LO,VM::VM3,"Voxel Spacing",false },
  {0x200d,0x0004,"Philips US Imaging DD 036",VR::LO,VM::VM3,"?",false },
  {0x200d,0x0001,"Philips US Imaging DD 038",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0001,"Philips US Imaging DD 039",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0004,"Philips US Imaging DD 039",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0005,"Philips US Imaging DD 039",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0006,"Philips US Imaging DD 039",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0007,"Philips US Imaging DD 039",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0008,"Philips US Imaging DD 039",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0009,"Philips US Imaging DD 039",VR::LO,VM::VM1,"?",false },
  {0x200d,0x000a,"Philips US Imaging DD 039",VR::LO,VM::VM1,"?",false },
  {0x200d,0x000b,"Philips US Imaging DD 039",VR::LO,VM::VM1,"?",false },
  {0x200d,0x000c,"Philips US Imaging DD 039",VR::LO,VM::VM1,"?",false },
  {0x200d,0x000d,"Philips US Imaging DD 039",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0001,"Philips US Imaging DD 040",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0002,"Philips US Imaging DD 040",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0003,"Philips US Imaging DD 040",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0004,"Philips US Imaging DD 040",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0005,"Philips US Imaging DD 040",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0006,"Philips US Imaging DD 040",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0007,"Philips US Imaging DD 040",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0020,"Philips US Imaging DD 040",VR::LO,VM::VM1,"?",false },
  {0x200d,0x001e,"Philips US Imaging DD 041",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0015,"Philips US Imaging DD 042",VR::IS,VM::VM1,"?",false },
  {0x200d,0x0016,"Philips US Imaging DD 042",VR::FD,VM::VM1,"?",false },
  {0x200d,0x0050,"Philips US Imaging DD 042",VR::LO,VM::VM3,"?",false },
  {0x200d,0x0051,"Philips US Imaging DD 042",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0052,"Philips US Imaging DD 042",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0053,"Philips US Imaging DD 042",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0054,"Philips US Imaging DD 042",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0055,"Philips US Imaging DD 042",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0056,"Philips US Imaging DD 042",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0057,"Philips US Imaging DD 042",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0058,"Philips US Imaging DD 042",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0059,"Philips US Imaging DD 042",VR::LO,VM::VM1,"?",false },
  {0x200d,0x005a,"Philips US Imaging DD 042",VR::LO,VM::VM1,"?",false },
  {0x200d,0x005b,"Philips US Imaging DD 042",VR::LO,VM::VM1,"?",false },
  {0x200d,0x005c,"Philips US Imaging DD 042",VR::LO,VM::VM1,"?",false },
  {0x200d,0x005d,"Philips US Imaging DD 042",VR::LO,VM::VM1,"?",false },
  {0x200d,0x005e,"Philips US Imaging DD 042",VR::LO,VM::VM1,"?",false },
  {0x200d,0x005f,"Philips US Imaging DD 042",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0060,"Philips US Imaging DD 042",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0070,"Philips US Imaging DD 042",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0071,"Philips US Imaging DD 042",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0072,"Philips US Imaging DD 042",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0073,"Philips US Imaging DD 042",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0074,"Philips US Imaging DD 042",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0075,"Philips US Imaging DD 042",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0076,"Philips US Imaging DD 042",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0077,"Philips US Imaging DD 042",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0005,"Philips US Imaging DD 043",VR::SH,VM::VM1,"?",false },
  {0x200d,0x0006,"Philips US Imaging DD 043",VR::IS,VM::VM1,"?",false },
  {0x200d,0x00f1,"Philips US Imaging DD 045",VR::SQ,VM::VM1,"Private DataType Image Sequence",false },
  {0x200d,0x00f3,"Philips US Imaging DD 045",VR::OB,VM::VM1,"image buffer (Zlib/none compressed)",false },
  {0x200d,0x00f8,"Philips US Imaging DD 045",VR::SQ,VM::VM1,"Private DataType Image Groups Sequence",false },
  {0x200d,0x00fa,"Philips US Imaging DD 045",VR::CS,VM::VM1,"ZLib/None",false },
  {0x200d,0x00fb,"Philips US Imaging DD 045",VR::OB,VM::VM1,"Frames header array",false },
  {0x200d,0x0000,"Philips US Imaging DD 066",VR::OB,VM::VM1,"Philips Confidential v.1",false },
  {0x200d,0x0000,"Philips US Imaging DD 109",VR::US,VM::VM1,"?",false },
  {0x200d,0x0001,"Philips US Imaging DD 109",VR::SQ,VM::VM1,"?",false },
  {0x200d,0x0002,"Philips US Imaging DD 109",VR::ST,VM::VM1,"?",false },
  {0x200d,0x0003,"Philips US Imaging DD 109",VR::CS,VM::VM1,"?",false },
  {0x200d,0x0004,"Philips US Imaging DD 109",VR::SL,VM::VM4,"?",false },
  {0x200d,0x0005,"Philips US Imaging DD 109",VR::UL,VM::VM3,"?",false },
  {0x200d,0x0006,"Philips US Imaging DD 109",VR::UL,VM::VM3,"?",false },
  {0x200d,0x0007,"Philips US Imaging DD 109",VR::CS,VM::VM1,"?",false },
  {0x200d,0x0008,"Philips US Imaging DD 109",VR::CS,VM::VM1,"?",false },
  {0x200d,0x0009,"Philips US Imaging DD 109",VR::OB,VM::VM1,"?",false },
  {0x200d,0x000a,"Philips US Imaging DD 109",VR::UL,VM::VM1,"?",false },
  {0x200d,0x000b,"Philips US Imaging DD 109",VR::OB,VM::VM1,"?",false },
  {0x200d,0x000c,"Philips US Imaging DD 109",VR::UL,VM::VM1,"?",false },
  {0x200d,0x000d,"Philips US Imaging DD 109",VR::SQ,VM::VM1,"?",false },
  {0x200d,0x000e,"Philips US Imaging DD 109",VR::CS,VM::VM1,"?",false },
  {0x200d,0x000f,"Philips US Imaging DD 109",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0010,"Philips US Imaging DD 109",VR::SL,VM::VM1,"?",false },
  {0x200d,0x0011,"Philips US Imaging DD 109",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0012,"Philips US Imaging DD 109",VR::SL,VM::VM1,"?",false },
  {0x200d,0x0013,"Philips US Imaging DD 109",VR::US,VM::VM1_n,"?",false },
  {0x200d,0x0014,"Philips US Imaging DD 109",VR::CS,VM::VM1,"?",false },
  {0x200d,0x0001,"Philips US Imaging DD 113",VR::LO,VM::VM1,"?",false },
  {0x200d,0x0002,"Philips US Imaging DD 113",VR::UL,VM::VM1,"?",false },
  {0x200d,0x0003,"Philips US Imaging DD 113",VR::UL,VM::VM1,"?",false },
  {0x200d,0x0004,"Philips US Imaging DD 113",VR::UL,VM::VM1,"?",false },
  {0x200d,0x0005,"Philips US Imaging DD 113",VR::UL,VM::VM1,"?",false },
  {0x200d,0x0006,"Philips US Imaging DD 113",VR::UL,VM::VM1,"?",false },
  {0x200d,0x0007,"Philips US Imaging DD 113",VR::CS,VM::VM1,"?",false },
  {0x200d,0x0008,"Philips US Imaging DD 113",VR::UL,VM::VM1,"?",false }, // 16bits RLE
  {0x200d,0x0009,"Philips US Imaging DD 113",VR::UL,VM::VM1,"?",false },
  {0x200d,0x000a,"Philips US Imaging DD 113",VR::FD,VM::VM1_n,"?",false },
  {0x200d,0x000b,"Philips US Imaging DD 113",VR::CS,VM::VM1,"?",false },
  {0x200d,0x000c,"Philips US Imaging DD 113",VR::UL,VM::VM1,"?",false },
  {0x200d,0x000d,"Philips US Imaging DD 113",VR::UL,VM::VM1,"?",false },
  {0x200d,0x000e,"Philips US Imaging DD 113",VR::UL,VM::VM1,"?",false },
  {0x200d,0x000f,"Philips US Imaging DD 113",VR::DS,VM::VM1_n,"?",false },
  {0x200d,0x0010,"Philips US Imaging DD 113",VR::DS,VM::VM1_n,"?",false },
  {0x200d,0x0011,"Philips US Imaging DD 113",VR::SL,VM::VM1,"?",false },
  {0x200d,0x0012,"Philips US Imaging DD 113",VR::UL,VM::VM1,"?",false },
  {0x200d,0x0013,"Philips US Imaging DD 113",VR::FL,VM::VM1_n,"?",false },
  {0x200d,0x0014,"Philips US Imaging DD 113",VR::US_SS,VM::VM1_n,"?",false },
  {0x200d,0x0015,"Philips US Imaging DD 113",VR::UL,VM::VM1,"?",false },
  {0x200d,0x0017,"Philips US Imaging DD 113",VR::FD,VM::VM1,"?",false },
  {0x200d,0x0018,"Philips US Imaging DD 113",VR::FD,VM::VM1,"?",false },
  {0x200d,0x0019,"Philips US Imaging DD 113",VR::FD,VM::VM1,"?",false },
  {0x200d,0x001a,"Philips US Imaging DD 113",VR::FD,VM::VM1,"?",false },
  {0x200d,0x001b,"Philips US Imaging DD 113",VR::FD,VM::VM1,"?",false },
  {0x200d,0x001c,"Philips US Imaging DD 113",VR::FD,VM::VM1,"?",false },
  {0x200d,0x001d,"Philips US Imaging DD 113",VR::UL,VM::VM1,"?",false },
  {0x200d,0x001e,"Philips US Imaging DD 113",VR::UL,VM::VM1,"?",false },
  {0x200d,0x001f,"Philips US Imaging DD 113",VR::FD,VM::VM1,"?",false },
  {0x200d,0x0020,"Philips US Imaging DD 113",VR::FD,VM::VM1,"?",false },
  {0x200d,0x0021,"Philips US Imaging DD 113",VR::FD,VM::VM1,"?",false },
  {0x200d,0x0022,"Philips US Imaging DD 113",VR::UL,VM::VM1,"?",false },
  {0x200d,0x0024,"Philips US Imaging DD 113",VR::UL,VM::VM1,"?",false },
  {0x200d,0x0025,"Philips US Imaging DD 113",VR::UL,VM::VM1,"?",false },
  {0x200d,0x0026,"Philips US Imaging DD 113",VR::UL,VM::VM1,"?",false },
  {0x200d,0x0027,"Philips US Imaging DD 113",VR::UL,VM::VM1,"?",false },
  {0x200d,0x0028,"Philips US Imaging DD 113",VR::UL,VM::VM1,"?",false },
  {0x200d,0x0031,"Philips US Imaging DD 113",VR::UL,VM::VM1,"?",false },
  {0x200d,0x0002,"Philips US Private 3D",VR::IS,VM::VM1,"Threedpr Echo Vision Setting",false },
  {0x200d,0x0003,"Philips US Private 3D",VR::IS,VM::VM1,"Threedpr Color Vision Setting",false },
  {0x200d,0x0004,"Philips US Private 3D",VR::IS,VM::VM1,"Threedpr Transparency",false },
  {0x200d,0x0006,"Philips US Private 3D",VR::FD,VM::VM1,"Threedpr Vision Vol Brightness",false },
  {0x200d,0x0008,"Philips US Private 3D",VR::IS,VM::VM1,"Threedpr Low Threshold",false },
  {0x200d,0x0009,"Philips US Private 3D",VR::IS,VM::VM1,"Threedpr Image Layout",false },
  {0x200d,0x000a,"Philips US Private 3D",VR::IS,VM::VM1,"Threedpr Xres Filtering Enabled",false },
  {0x200d,0x000b,"Philips US Private 3D",VR::IS,VM::VM1,"Threedpr Echo Twod Chroma Map Enabled",false },
  {0x200d,0x000c,"Philips US Private 3D",VR::IS,VM::VM1,"Threedpr Box Outlined Enabled",false },
  {0x200d,0x000d,"Philips US Private 3D",VR::IS,VM::VM1,"Threedpr Smoothing Enabled",false },
  {0x200d,0x000e,"Philips US Private 3D",VR::IS,VM::VM1,"Threedpr Reference Graphic Enabled",false },
  {0x200d,0x000f,"Philips US Private 3D",VR::IS,VM::VM1,"Threedpr Map Inverted",false },
  {0x200d,0x0010,"Philips US Private 3D",VR::IS,VM::VM1,"Threedpr Color Stream Display Enabled",false },
  {0x200d,0x0011,"Philips US Private 3D",VR::IS,VM::VM1,"Threedpr Echo Stream Display Enabled",false },
  {0x200d,0x0012,"Philips US Private 3D",VR::LO,VM::VM1,"Vdb Param Color 3d Estimate Scale(2)",false },
  {0x200d,0x0013,"Philips US Private 3D",VR::LO,VM::VM1,"Vdb Param Color 3d Estimate Apex Position (1)",false },
  {0x200d,0x0014,"Philips US Private 3D",VR::LO,VM::VM1,"Vdb Param Color 3d Estimate Color Offset",false },
  {0x200d,0x0015,"Philips US Private 3D",VR::LO,VM::VM1,"Vdb Param Echo 3d Estimate Dimension(2)",false },
  {0x200d,0x0016,"Philips US Private 3D",VR::SQ,VM::VM1,"Private Native Threed Data Sequence(1)",false },
  {0x200d,0x0017,"Philips US Private 3D",VR::IS,VM::VM1,"Threedpr Basic Vol Brightness",false },
  {0x200d,0x0018,"Philips US Private 3D",VR::LO,VM::VM1,"Vdb Param Echo 3d Estimate Apex Position (2)",false },
  {0x200d,0x001a,"Philips US Private 3D",VR::IS,VM::VM1,"Threedpr Echo Smoothing",false },
  {0x200d,0x001b,"Philips US Private 3D",VR::IS,VM::VM1,"Threedpr Color Smoothing",false },
  {0x200d,0x001e,"Philips US Private 3D",VR::IS,VM::VM1,"Threedpr Num Completed Trim Planes",false },
  {0x200d,0x001f,"Philips US Private 3D",VR::IS,VM::VM1,"Threedpr Depth",false },
  {0x200d,0x0020,"Philips US Private 3D",VR::SQ,VM::VM1,"Private Native Data Stream Array",false },
  {0x200d,0x0021,"Philips US Private 3D",VR::IS,VM::VM1,"Threedpr Wall Filter",false },
  {0x200d,0x0023,"Philips US Private 3D",VR::FD,VM::VM1,"Threedpr Baseline",false },
  {0x200d,0x0024,"Philips US Private 3D",VR::FD,VM::VM1,"Threedpr Gain",false },
  {0x200d,0x0025,"Philips US Private 3D",VR::FD,VM::VM1,"Threedpr Compress",false },
  {0x200d,0x0026,"Philips US Private 3D",VR::FD,VM::VM1,"Threedpr Light Brightness",false },
  {0x200d,0x0027,"Philips US Private 3D",VR::FD,VM::VM1,"Threedpr Elevation Scale Factor",false },
  {0x200d,0x0028,"Philips US Private 3D",VR::FD,VM::VM1,"Threedpr Zoom Factor",false },
  {0x200d,0x0029,"Philips US Private 3D",VR::FD,VM::VM1,"Threedpr Color Write Priority",false },
  {0x200d,0x002a,"Philips US Private 3D",VR::FD,VM::VM1,"Threedpr Rotation Angle X",false },
  {0x200d,0x002b,"Philips US Private 3D",VR::FD,VM::VM1,"Threedpr Rotation Angle Y",false },
  {0x200d,0x002c,"Philips US Private 3D",VR::FD,VM::VM1,"Threedpr Data Voi Center",false },
  {0x200d,0x002d,"Philips US Private 3D",VR::FD,VM::VM1,"Threedpr View Translation",false },
  {0x200d,0x002e,"Philips US Private 3D",VR::FD,VM::VM1,"Threedpr Data Voi Min Point",false },
  {0x200d,0x002f,"Philips US Private 3D",VR::FD,VM::VM1,"Threedpr Data Voi Max Point",false },
  {0x200d,0x0030,"Philips US Private 3D",VR::FD,VM::VM1,"Threedpr Box Crop Min Point",false },
  {0x200d,0x0031,"Philips US Private 3D",VR::FD,VM::VM1,"Threedpr Box Crop Max Point",false },
  {0x200d,0x0032,"Philips US Private 3D",VR::FD,VM::VM1,"Threedpr Mpr Rotation Matrix",false },
  {0x200d,0x0033,"Philips US Private 3D",VR::FD,VM::VM1,"Threedpr Vol Rotation Matrix",false },
  {0x200d,0x0034,"Philips US Private 3D",VR::FD,VM::VM1,"Threedpr Trim Plane Equation",false },
  {0x200d,0x0035,"Philips US Private 3D",VR::IS,VM::VM1,"Threedpr Subpage Data Version",false },
  {0x200d,0x0036,"Philips US Private 3D",VR::IS,VM::VM1,"Threedpr Is Arbitrary Crop",false },
  {0x200d,0x0037,"Philips US Private 3D",VR::IS,VM::VM1,"Threedpr Arbitrary Crop Dist From Center",false },
  {0x200d,0x0038,"Philips US Private 3D",VR::FD,VM::VM1,"Threedpr Arbitrary Crop Rotation Matrix",false },
  {0x200d,0x0039,"Philips US Private 3D",VR::FD,VM::VM1,"Threedpr Color Gain",false },
  {0x200d,0x0040,"Philips US Private 3D",VR::IS,VM::VM1,"Threedpr Color Wall Filter Index",false },
  {0x200d,0x0041,"Philips US Private 3D",VR::IS,VM::VM1,"Threedpr Cursor Enabled",false },
  {0x200d,0x0042,"Philips US Private 3D",VR::IS,VM::VM1,"Threedpr Bondbox Graphic Enabled",false },
  {0x2003,0x0000,"Philips X-ray Imaging DD 001",VR::CS,VM::VM1,"?",false },
  {0x2003,0x0001,"Philips X-ray Imaging DD 001",VR::LO,VM::VM1,"?",false },
  {0x2003,0x0002,"Philips X-ray Imaging DD 001",VR::FD,VM::VM3,"?",false },
  {0x2003,0x0003,"Philips X-ray Imaging DD 001",VR::LO,VM::VM1,"?",false },
  {0x2003,0x0004,"Philips X-ray Imaging DD 001",VR::SQ,VM::VM1,"Private enhanced Table Sequence",false },
  {0x2003,0x0006,"Philips X-ray Imaging DD 001",VR::SL,VM::VM1,"?",false },
  {0x2003,0x0009,"Philips X-ray Imaging DD 001",VR::SL,VM::VM1,"?",false },
  {0x2003,0x0010,"Philips X-ray Imaging DD 001",VR::LO,VM::VM1,"?",false },
  {0x2003,0x0011,"Philips X-ray Imaging DD 001",VR::SL,VM::VM1,"?",false },
  {0x2003,0x0012,"Philips X-ray Imaging DD 001",VR::SL,VM::VM1,"?",false },
  {0x2003,0x0013,"Philips X-ray Imaging DD 001",VR::SL,VM::VM1,"?",false },
  {0x2003,0x0014,"Philips X-ray Imaging DD 001",VR::FD,VM::VM1,"?",false },
  {0x2003,0x0015,"Philips X-ray Imaging DD 001",VR::FD,VM::VM1,"?",false },
  {0x2003,0x0016,"Philips X-ray Imaging DD 001",VR::SL,VM::VM1,"?",false },
  {0x2003,0x0017,"Philips X-ray Imaging DD 001",VR::SL,VM::VM1,"?",false },
  {0x2003,0x0018,"Philips X-ray Imaging DD 001",VR::SL,VM::VM1,"?",false },
  {0x2003,0x0019,"Philips X-ray Imaging DD 001",VR::SL,VM::VM1,"?",false },
  {0x2003,0x0022,"Philips X-ray Imaging DD 001",VR::SL,VM::VM1,"?",false },
  {0x2003,0x0024,"Philips X-ray Imaging DD 001",VR::FD,VM::VM4,"?",false },
  {0x2003,0x0025,"Philips X-ray Imaging DD 001",VR::SL,VM::VM1,"?",false },
  {0x2003,0x0026,"Philips X-ray Imaging DD 001",VR::SL,VM::VM1,"?",false },
  {0x2003,0x0027,"Philips X-ray Imaging DD 001",VR::SH,VM::VM1,"?",false },
  {0x2003,0x0028,"Philips X-ray Imaging DD 001",VR::SH,VM::VM1,"?",false },
  {0x2003,0x0029,"Philips X-ray Imaging DD 001",VR::FD,VM::VM1,"?",false },
  {0x2003,0x002a,"Philips X-ray Imaging DD 001",VR::LO,VM::VM1,"?",false },
  {0x2003,0x002b,"Philips X-ray Imaging DD 001",VR::FD,VM::VM1,"?",false },
  {0x2003,0x002c,"Philips X-ray Imaging DD 001",VR::SH,VM::VM1,"?",false },
  {0x2003,0x002d,"Philips X-ray Imaging DD 001",VR::SL,VM::VM1_n,"?",false },
  {0x2003,0x002e,"Philips X-ray Imaging DD 001",VR::SQ,VM::VM1,"?",false },
  {0x2003,0x002f,"Philips X-ray Imaging DD 001",VR::CS,VM::VM1,"IsResolvePatientMixApplied",false },
  {0x2003,0x0030,"Philips X-ray Imaging DD 001",VR::CS,VM::VM1,"?",false },
  {0x2003,0x0031,"Philips X-ray Imaging DD 001",VR::CS,VM::VM1,"?",false },
  {0x2003,0x0032,"Philips X-ray Imaging DD 001",VR::UI,VM::VM1,"?",false },
  {0x7101,0x0000,"Picker MR Private Group",VR::OB,VM::VM1,"?Protocol Data 0 (weird gzip)?",false },
  {0x7101,0x0001,"Picker MR Private Group",VR::SL,VM::VM1,"?Protocol Data 0 Length?",false },
  {0x7101,0x0002,"Picker MR Private Group",VR::OB,VM::VM1,"?Protocol Data 2 (weird gzip)?",false },
  {0x7101,0x0003,"Picker MR Private Group",VR::SL,VM::VM1,"?Protocol Data 2 Length?",false },
  {0x7101,0x0004,"Picker MR Private Group",VR::SH,VM::VM1,"?Time?",false },
  {0x7101,0x0005,"Picker MR Private Group",VR::SH,VM::VM2,"?ImageType?",false },
  {0x7101,0x0006,"Picker MR Private Group",VR::SH,VM::VM4,"?MR Diffusion Info?",false },
  {0x7101,0x0010,"Picker MR Private Group",VR::DS,VM::VM1,"?",false },
  {0x7001,0x0001,"Picker NM Private Group",VR::UI,VM::VM1,"Private",false },
  {0x7001,0x0002,"Picker NM Private Group",VR::OB,VM::VM1,"Private",false },
  {0x7001,0x0003,"Picker NM Private Group",VR::OB,VM::VM1,"Private",false },
  {0x7001,0x0004,"Picker NM Private Group",VR::OB,VM::VM1,"Private",false },
  {0x7001,0x0005,"Picker NM Private Group",VR::OB,VM::VM1,"Private",false },
  {0x7001,0x0006,"Picker NM Private Group",VR::OB,VM::VM1,"Private",false },
  {0x7001,0x0007,"Picker NM Private Group",VR::OB,VM::VM1,"Private",false },
  {0x7001,0x0008,"Picker NM Private Group",VR::OB,VM::VM1,"Private",false },
  {0x7001,0x0009,"Picker NM Private Group",VR::OB,VM::VM1,"Private",false },
  {0x7001,0x0010,"Picker NM Private Group",VR::SQ,VM::VM1,"Private",false },
  {0x7001,0x0011,"Picker NM Private Group",VR::LO,VM::VM1,"Private",false },
  {0x7001,0x0012,"Picker NM Private Group",VR::OB,VM::VM1,"Private",false },
  {0x7001,0x0013,"Picker NM Private Group",VR::US,VM::VM1,"Private",false },
  {0x7001,0x0014,"Picker NM Private Group",VR::OB,VM::VM1,"Private",false },
  {0x7001,0x0015,"Picker NM Private Group",VR::OB,VM::VM1,"Private",false },
  {0x7001,0x0016,"Picker NM Private Group",VR::OB,VM::VM1,"Private",false },
  {0x7001,0x0017,"Picker NM Private Group",VR::LO,VM::VM1,"Contains a text string that includes some of the same information that appears on the Odyssey Clinical Index (image directory) for this Image.",false },
  {0x7043,0x0000,"Picker NM Private Group",VR::SH,VM::VM1,"Private.",false },
  {0x0009,0x0000,"QCA",VR::CS,VM::VM1,"Analysis Type",false },
  {0x0009,0x0004,"QCA",VR::LO,VM::VM1,"Segment Name",false },
  {0x0009,0x0012,"QCA",VR::DS,VM::VM1,"Pre Catheter size",false },
  {0x0009,0x0013,"QCA",VR::DS,VM::VM1,"Pre Reference Diameter",false },
  {0x0009,0x0014,"QCA",VR::DS,VM::VM1,"Pre Minimum Lumen Diameter",false },
  {0x0009,0x0015,"QCA",VR::DS,VM::VM1,"Pre Average Diameter",false },
  {0x0009,0x0016,"QCA",VR::DS,VM::VM1,"Pre Stenosis Length",false },
  {0x0009,0x0017,"QCA",VR::DS,VM::VM1,"Pre Stenosis %",false },
  {0x0009,0x0018,"QCA",VR::DS,VM::VM1,"Pre Geometric Area Reduction %",false },
  {0x0009,0x0022,"QCA",VR::DS,VM::VM1,"Post Catheter Size",false },
  {0x0009,0x0023,"QCA",VR::DS,VM::VM1,"Post Reference Diameter",false },
  {0x0009,0x0024,"QCA",VR::DS,VM::VM1,"Post Minimum Lumen Diameter",false },
  {0x0009,0x0025,"QCA",VR::DS,VM::VM1,"Post Average Diameter",false },
  {0x0009,0x0026,"QCA",VR::DS,VM::VM1,"Post Stenosis Length",false },
  {0x0009,0x0027,"QCA",VR::DS,VM::VM1,"Post Stenosis %",false },
  {0x0009,0x0028,"QCA",VR::DS,VM::VM1,"Post Geometric Area Reduction %",false },
  {0x0009,0x0000,"QCA_RESULTS",VR::CS,VM::VM1,"Analysis Views",false },
  {0x0009,0x0010,"QCA_RESULTS",VR::LO,VM::VM1,"Segment",false },
  {0x0009,0x0011,"QCA_RESULTS",VR::LO,VM::VM1,"Pre Catheter Name",false },
  {0x0009,0x0012,"QCA_RESULTS",VR::DS,VM::VM1,"Pre Catheter Size",false },
  {0x0009,0x0013,"QCA_RESULTS",VR::DS,VM::VM1,"Pre Reference Diameter",false },
  {0x0009,0x0014,"QCA_RESULTS",VR::DS,VM::VM1,"Pre Minimum Lumen Diameter",false },
  {0x0009,0x0015,"QCA_RESULTS",VR::DS,VM::VM1,"Pre Average Diameter",false },
  {0x0009,0x0016,"QCA_RESULTS",VR::DS,VM::VM1,"Pre Stenosis Length",false },
  {0x0009,0x0017,"QCA_RESULTS",VR::IS,VM::VM1,"Pre Stenosis %",false },
  {0x0009,0x0018,"QCA_RESULTS",VR::IS,VM::VM1,"Pre Geometric Area Reduction %",false },
  {0x0009,0x0021,"QCA_RESULTS",VR::LO,VM::VM1,"Post Catheter Name",false },
  {0x0009,0x0022,"QCA_RESULTS",VR::DS,VM::VM1,"Post Catheter Size",false },
  {0x0009,0x0023,"QCA_RESULTS",VR::DS,VM::VM1,"Post Reference Diameter",false },
  {0x0009,0x0024,"QCA_RESULTS",VR::DS,VM::VM1,"Post Minimum Lumen Diameter",false },
  {0x0009,0x0025,"QCA_RESULTS",VR::DS,VM::VM1,"Post Average Diameter",false },
  {0x0009,0x0026,"QCA_RESULTS",VR::DS,VM::VM1,"Post Stenosis Length",false },
  {0x0009,0x0027,"QCA_RESULTS",VR::IS,VM::VM1,"Post Stenosis %",false },
  {0x0009,0x0028,"QCA_RESULTS",VR::IS,VM::VM1,"Post Geometric Area Reduction %",false },
  {0x0009,0x0040,"QCA_RESULTS",VR::IS,VM::VM1,"Calibration Frame",false },
  {0x0009,0x0041,"QCA_RESULTS",VR::IS,VM::VM1,"End Diastolic Frame",false },
  {0x0009,0x0042,"QCA_RESULTS",VR::IS,VM::VM1,"End Systolic Frame",false },
  {0x0009,0x0043,"QCA_RESULTS",VR::DS,VM::VM1,"End Diastolic Volume",false },
  {0x0009,0x0044,"QCA_RESULTS",VR::DS,VM::VM1,"End Systolic Volume",false },
  {0x0009,0x0045,"QCA_RESULTS",VR::DS,VM::VM1,"Stroke Volume",false },
  {0x0009,0x0047,"QCA_RESULTS",VR::DS,VM::VM1,"Ejection Fraction",false },
  {0x0009,0x0048,"QCA_RESULTS",VR::DS,VM::VM1,"Body Surface Area",false },
  {0x0009,0x0049,"QCA_RESULTS",VR::SH,VM::VM1,"Artery Territory Region",false },
  {0x0009,0x0050,"QCA_RESULTS",VR::IS,VM::VM1,"Number of Diseased Vessels",false },
  {0x0009,0x0051,"QCA_RESULTS",VR::DS,VM::VM1,"Hypokinesis in Region",false },
  {0x0009,0x0052,"QCA_RESULTS",VR::DS,VM::VM1,"Hyperkinesis in Opposite Region",false },
  {0x0009,0x0053,"QCA_RESULTS",VR::IS,VM::VM1,"Percent Total LV Hypokinesis",false },
  {0x0009,0x0055,"QCA_RESULTS",VR::DS,VM::VM1,"Calibration Factor",false },
  {0x0009,0x0001,"QUASAR_INTERNAL_USE",VR::UL,VM::VM1_n,"Rate Vector",false },
  {0x0009,0x0002,"QUASAR_INTERNAL_USE",VR::UL,VM::VM1_n,"Count Vector",false },
  {0x0009,0x0003,"QUASAR_INTERNAL_USE",VR::UL,VM::VM1_n,"Time Vector",false },
  {0x0009,0x0007,"QUASAR_INTERNAL_USE",VR::US,VM::VM1_n,"Angle Vector",false },
  {0x0009,0x0008,"QUASAR_INTERNAL_USE",VR::US,VM::VM1,"Camera Shape",false },
  {0x0009,0x0010,"QUASAR_INTERNAL_USE",VR::US,VM::VM1,"WholeBody Spots",false },
  {0x0009,0x0011,"QUASAR_INTERNAL_USE",VR::US,VM::VM1,"Worklist Flag",false },
  {0x0009,0x0012,"QUASAR_INTERNAL_USE",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0013,"QUASAR_INTERNAL_USE",VR::ST,VM::VM1,"Sequence Type",false },
  {0x0009,0x0014,"QUASAR_INTERNAL_USE",VR::ST,VM::VM1,"Sequence Name",false },
  {0x0009,0x0015,"QUASAR_INTERNAL_USE",VR::UL,VM::VM1,"Avr RR Time Vector",false },
  {0x0009,0x0016,"QUASAR_INTERNAL_USE",VR::UL,VM::VM1,"Low Limit Vector",false },
  {0x0009,0x0017,"QUASAR_INTERNAL_USE",VR::UL,VM::VM1,"High Limit Vector",false },
  {0x0009,0x0018,"QUASAR_INTERNAL_USE",VR::UL,VM::VM1,"Begin Index Vector",false },
  {0x0009,0x0019,"QUASAR_INTERNAL_USE",VR::UL,VM::VM1,"End Index Vector",false },
  {0x0009,0x001a,"QUASAR_INTERNAL_USE",VR::UL,VM::VM1,"Raw Time Vector",false },
  {0x0009,0x001b,"QUASAR_INTERNAL_USE",VR::LO,VM::VM1,"Image Type String",false },
  {0x0009,0x001d,"QUASAR_INTERNAL_USE",VR::US,VM::VM1,"?",false },
  {0x0009,0x001e,"QUASAR_INTERNAL_USE",VR::ST,VM::VM1,"?",false },
  {0x0009,0x0022,"QUASAR_INTERNAL_USE",VR::FL,VM::VM1,"?",false },
  {0x0009,0x0023,"QUASAR_INTERNAL_USE",VR::US,VM::VM1,"?",false },
  {0x0009,0x0039,"QUASAR_INTERNAL_USE",VR::UI,VM::VM1,"?",false },
  {0x0009,0x0040,"QUASAR_INTERNAL_USE",VR::DA,VM::VM1,"?",false },
  {0x0009,0x0041,"QUASAR_INTERNAL_USE",VR::TM,VM::VM1,"?",false },
  {0x0009,0x0042,"QUASAR_INTERNAL_USE",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0044,"QUASAR_INTERNAL_USE",VR::SH,VM::VM1,"?",false },
  {0x0037,0x0010,"QUASAR_INTERNAL_USE",VR::SQ,VM::VM1,"?",false },
  {0x0037,0x001b,"QUASAR_INTERNAL_USE",VR::LO,VM::VM1,"?",false },
  {0x0037,0x0030,"QUASAR_INTERNAL_USE",VR::LO,VM::VM1,"?",false },
  {0x0037,0x0040,"QUASAR_INTERNAL_USE",VR::LO,VM::VM1,"?",false },
  {0x0037,0x0050,"QUASAR_INTERNAL_USE",VR::LO,VM::VM1,"?",false },
  {0x0037,0x0060,"QUASAR_INTERNAL_USE",VR::LO,VM::VM1,"?",false },
  {0x0037,0x0070,"QUASAR_INTERNAL_USE",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0040,"QVA",VR::IS,VM::VM1,"Calibration Frame",false },
  {0x0009,0x0041,"QVA",VR::IS,VM::VM1,"End Diastolic Frame",false },
  {0x0009,0x0042,"QVA",VR::IS,VM::VM1,"End Systolic Frame",false },
  {0x0009,0x0043,"QVA",VR::DS,VM::VM1,"End Diastolic Volume",false },
  {0x0009,0x0044,"QVA",VR::DS,VM::VM1,"End Systolic Volume",false },
  {0x0009,0x0045,"QVA",VR::DS,VM::VM1,"Stroke Volume",false },
  {0x0009,0x0046,"QVA",VR::DS,VM::VM1,"Cardiac Output",false },
  {0x0009,0x0047,"QVA",VR::DS,VM::VM1,"Ejection Fraction",false },
  {0x0009,0x0048,"QVA",VR::DS,VM::VM1,"Body Surface Area",false },
  {0x0009,0x0049,"QVA",VR::SH,VM::VM1,"Artery Territory Region",false },
  {0x0009,0x0050,"QVA",VR::IS,VM::VM1,"Number of Diseased Vessels",false },
  {0x0009,0x0051,"QVA",VR::DS,VM::VM1,"Hypokinesis in Region",false },
  {0x0009,0x0052,"QVA",VR::DS,VM::VM1,"Hyperkinesis in Region",false },
  {0x0009,0x0053,"QVA",VR::IS,VM::VM1,"percent of chords with hyperkinesis <- 2 SD",false },
  {0x0009,0x0054,"QVA",VR::IS,VM::VM1,"percent of chords with alkinesis/dyskinesis",false },
  {0x0009,0x0055,"QVA",VR::DS,VM::VM1,"Calibration Factor",false },
  {0x0039,0x0095,"REPORT_FROM_APP",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0024,"RadWorksMarconi",VR::US,VM::VM1,"<internal",false },
  {0x3111,0x0002,"RadWorksTBR",VR::CS,VM::VM1,"AMI Compression Type (RET)",false },
  {0x3111,0x00ff,"RadWorksTBR",VR::SQ,VM::VM1,"<internal",false },
  {0x0021,0x0001,"SCHICK TECHNOLOGIES - Change Item Creator ID",VR::UI,VM::VM1,"Change List Item UID",false },
  {0x0021,0x0002,"SCHICK TECHNOLOGIES - Change Item Creator ID",VR::US,VM::VM1,"Operation",false },
  {0x0021,0x0003,"SCHICK TECHNOLOGIES - Change Item Creator ID",VR::DT,VM::VM1,"Date",false },
  {0x0021,0x0004,"SCHICK TECHNOLOGIES - Change Item Creator ID",VR::PN,VM::VM1,"User Name",false },
  {0x0021,0x0005,"SCHICK TECHNOLOGIES - Change Item Creator ID",VR::OB,VM::VM1,"Additional Change Data",false },
  {0x0021,0x0001,"SCHICK TECHNOLOGIES - Change List Creator ID",VR::UI,VM::VM1,"Reference Image Instance UID",false },
  {0x0021,0x0002,"SCHICK TECHNOLOGIES - Change List Creator ID",VR::SQ,VM::VM1,"Sequence",false },
  {0x0029,0x0001,"SCHICK TECHNOLOGIES - Image Security Creator ID",VR::UL,VM::VM1,"Image Checksum",false },
  {0x0021,0x0001,"SCHICK TECHNOLOGIES - Note Item Creator ID",VR::ST,VM::VM1,"Text",false },
  {0x0021,0x0002,"SCHICK TECHNOLOGIES - Note Item Creator ID",VR::DT,VM::VM1,"Date",false },
  {0x0021,0x0003,"SCHICK TECHNOLOGIES - Note Item Creator ID",VR::PN,VM::VM1,"User Name",false },
  {0x0021,0x0004,"SCHICK TECHNOLOGIES - Note Item Creator ID",VR::US,VM::VM2,"Coordinates (X, Y)",false },
  {0x0021,0x0001,"SCHICK TECHNOLOGIES - Note List Creator ID",VR::UI,VM::VM1,"Reference Image Instance UID",false },
  {0x0021,0x0002,"SCHICK TECHNOLOGIES - Note List Creator ID",VR::SQ,VM::VM1,"Item",false },
  {0x0009,0x0001,"SCHICK TECHNOLOGIES - Viewset Creator ID",VR::UI,VM::VM1,"Instance UID",false },
  {0x0009,0x0002,"SCHICK TECHNOLOGIES - Viewset Creator ID",VR::LO,VM::VM1,"Name",false },
  {0x0009,0x0003,"SCHICK TECHNOLOGIES - Viewset Creator ID",VR::SQ,VM::VM1,"Sequence",false },
  {0x0009,0x0004,"SCHICK TECHNOLOGIES - Viewset Creator ID",VR::UL,VM::VM1,"Mode",false },
  {0x0009,0x0005,"SCHICK TECHNOLOGIES - Viewset Creator ID",VR::UL,VM::VM1,"Type",false },
  {0x0009,0x0001,"SCHICK TECHNOLOGIES - Viewset Item Creator ID",VR::UI,VM::VM1,"Image Instance UID",false },
  {0x0009,0x0002,"SCHICK TECHNOLOGIES - Viewset Item Creator ID",VR::US,VM::VM1,"Number",false },
  {0x0009,0x0003,"SCHICK TECHNOLOGIES - Viewset Item Creator ID",VR::US,VM::VM1,"Order",false },
  {0x0009,0x0004,"SCHICK TECHNOLOGIES - Viewset Item Creator ID",VR::US,VM::VM1,"Type",false },
  {0x0009,0x0005,"SCHICK TECHNOLOGIES - Viewset Item Creator ID",VR::US,VM::VM1,"Orientation",false },
  {0x0009,0x0006,"SCHICK TECHNOLOGIES - Viewset Item Creator ID",VR::US,VM::VM4,"Rect (Left, Top, Width, Height)",false },
  {0x0009,0x0007,"SCHICK TECHNOLOGIES - Viewset Item Creator ID",VR::US,VM::VM1,"Caption Height",false },
  {0x0009,0x0008,"SCHICK TECHNOLOGIES - Viewset Item Creator ID",VR::US,VM::VM1,"Source",false },
  {0x0009,0x0009,"SCHICK TECHNOLOGIES - Viewset Item Creator ID",VR::US,VM::VM1,"Default Width",false },
  {0x0009,0x000a,"SCHICK TECHNOLOGIES - Viewset Item Creator ID",VR::US,VM::VM1,"Default Height",false },
  {0x0009,0x0001,"SECTRA_Ident_01",VR::LO,VM::VM1,"Request number",false },
  {0x0009,0x0002,"SECTRA_Ident_01",VR::LO,VM::VM1,"Examination number",false },
  {0x0009,0x0004,"SECTRA_Ident_01",VR::LO,VM::VM1,"Series ID",false },
  {0x0009,0x0005,"SECTRA_Ident_01",VR::LO,VM::VM1,"Series Order",false },
  {0x0009,0x0006,"SECTRA_Ident_01",VR::LO,VM::VM1,"File Name",false },
  {0x0009,0x0007,"SECTRA_Ident_01",VR::LO,VM::VM1,"Image Data ID",false },
  {0x0029,0x0001,"SECTRA_ImageInfo_01",VR::OB,VM::VM1,"Image info",false },
  {0x0029,0x0002,"SECTRA_ImageInfo_01",VR::CS,VM::VM1,"Marking",false },
  {0x0029,0x0003,"SECTRA_ImageInfo_01",VR::LO,VM::VM1,"No Decompression",false },
  {0x0029,0x0004,"SECTRA_ImageInfo_01",VR::OB,VM::VM1,"Image Info new",false },
  {0x0029,0x0005,"SECTRA_ImageInfo_01",VR::SS,VM::VM1,"?",false },
  {0x6001,0x0001,"SECTRA_OverlayInfo_01",VR::LO,VM::VM1,"Sectra Overlay",false },
  {0x0031,0x0098,"SEGAMI MIML",VR::OW,VM::VM1,"?Binary XML with PHI?",false },
  {0x0035,0x0097,"SEGAMI__MEMO",VR::SH,VM::VM1,"?",false },
  {0x0035,0x0098,"SEGAMI__MEMO",VR::LT,VM::VM1,"?",false },
  {0x0019,0x0000,"SET WINDOW",VR::UN,VM::VM1,"Set Window Image Filter",false },
  {0x0019,0x0001,"SET WINDOW",VR::UN,VM::VM1,"Set Window Magnification Power",false },
  {0x0029,0x0001,"SHS MagicView 300",VR::FD,VM::VM1_n,"Zoom Factor",false },
  {0x0029,0x0002,"SHS MagicView 300",VR::FD,VM::VM1,"Pan Factor X",false },
  {0x0029,0x0003,"SHS MagicView 300",VR::FD,VM::VM1,"Pan Factor Y",false },
  {0x0029,0x0004,"SHS MagicView 300",VR::LO,VM::VM1,"NmPalette",false },
  {0x0019,0x0010,"SIEMENS CM VA0  ACQU",VR::LO,VM::VM1,"Parameter File Name",false },
  {0x0019,0x0011,"SIEMENS CM VA0  ACQU",VR::LO,VM::VM1,"Sequence File Name",false },
  {0x0019,0x0012,"SIEMENS CM VA0  ACQU",VR::LO,VM::VM1,"Sequence File Owner",false },
  {0x0019,0x0013,"SIEMENS CM VA0  ACQU",VR::LO,VM::VM1,"Sequence Description",false },
  {0x0019,0x0014,"SIEMENS CM VA0  ACQU",VR::LO,VM::VM1,"EPI File Name",false },
  {0x0019,0x0020,"SIEMENS CM VA0  ACQU",VR::LO,VM::VM1,"Mode File Name",false },
  {0x0019,0x0030,"SIEMENS CM VA0  ACQU",VR::LO,VM::VM1,"Convolution Kernel Comment 1",false },
  {0x0019,0x0031,"SIEMENS CM VA0  ACQU",VR::LO,VM::VM1,"Convolution Kernel Comment 2",false },
  {0x0019,0x0032,"SIEMENS CM VA0  ACQU",VR::LO,VM::VM1,"Convolution Kernel Comment 3",false },
  {0x0019,0x0033,"SIEMENS CM VA0  ACQU",VR::LO,VM::VM1,"Convolution Kernel Comment 4",false },
  {0x0019,0x0034,"SIEMENS CM VA0  ACQU",VR::LO,VM::VM1,"Convolution Kernel Comment 5",false },
  {0x0019,0x0035,"SIEMENS CM VA0  ACQU",VR::LO,VM::VM1,"Convolution Kernel Comment 6",false },
  {0x0019,0x0036,"SIEMENS CM VA0  ACQU",VR::LO,VM::VM1,"Convolution Kernel Comment 7",false },
  {0x0019,0x0037,"SIEMENS CM VA0  ACQU",VR::LO,VM::VM1,"Convolution Kernel Comment 8",false },
  {0x0019,0x0038,"SIEMENS CM VA0  ACQU",VR::LO,VM::VM1,"Convolution Kernel Comment 9",false },
  {0x0019,0x0039,"SIEMENS CM VA0  ACQU",VR::LO,VM::VM1,"Convolution Kernel Comment 10",false },
  {0x0019,0x003a,"SIEMENS CM VA0  ACQU",VR::LO,VM::VM1,"Reading Code RZ1 Comment 1",false },
  {0x0019,0x003b,"SIEMENS CM VA0  ACQU",VR::LO,VM::VM1,"Reading Code RZ1 Comment 2",false },
  {0x0019,0x003c,"SIEMENS CM VA0  ACQU",VR::LO,VM::VM1,"Reading Code RZ2 Comment 1",false },
  {0x0009,0x0000,"SIEMENS CM VA0  CMS",VR::IS,VM::VM1,"Number of Measurements",false },
  {0x0009,0x0010,"SIEMENS CM VA0  CMS",VR::CS,VM::VM1,"Storage Mode",false },
  {0x0009,0x0012,"SIEMENS CM VA0  CMS",VR::UL,VM::VM1,"Evaluation Mask Image",false },
  // SIEMENS_MAGNETOM-12-MONO2-Uncompressed.dcm seems to think the following 2 are UL and not IS...
  // but clearly SIEMENS_ImageLocationUN.dcm indicates those two are VR:IS...
  {0x0009,0x0013,"SIEMENS CM VA0  CMS",VR::IS,VM::VM1,"Int ??",false },
  {0x0009,0x0014,"SIEMENS CM VA0  CMS",VR::IS,VM::VM1,"Int ??",false },
  {0x0009,0x0026,"SIEMENS CM VA0  CMS",VR::DA,VM::VM1,"Table Zero Date",false },
  {0x0009,0x0027,"SIEMENS CM VA0  CMS",VR::TM,VM::VM1,"Table Zero Time",false },
  {0x0011,0x000a,"SIEMENS CM VA0  CMS",VR::LO,VM::VM1,"?",false },
  {0x0011,0x0010,"SIEMENS CM VA0  CMS",VR::DA,VM::VM1,"Registration Date",false },
  {0x0011,0x0011,"SIEMENS CM VA0  CMS",VR::TM,VM::VM1,"Registration Time",false },
  {0x0011,0x0022,"SIEMENS CM VA0  CMS",VR::LO,VM::VM1,"?",false },
  {0x0011,0x0023,"SIEMENS CM VA0  CMS",VR::DS,VM::VM1,"Used Patient Weight",false },
  {0x0011,0x0040,"SIEMENS CM VA0  CMS",VR::IS,VM::VM1,"Organ Code",false },
  {0x0013,0x0000,"SIEMENS CM VA0  CMS",VR::PN,VM::VM1,"Modifying Physician",false },
  {0x0013,0x0010,"SIEMENS CM VA0  CMS",VR::DA,VM::VM1,"Modification Date",false },
  {0x0013,0x0012,"SIEMENS CM VA0  CMS",VR::TM,VM::VM1,"Modification Time",false },
  {0x0013,0x0020,"SIEMENS CM VA0  CMS",VR::PN,VM::VM1,"Patient Name",false },
  {0x0013,0x0022,"SIEMENS CM VA0  CMS",VR::LO,VM::VM1,"Patient Id",false },
  {0x0013,0x0030,"SIEMENS CM VA0  CMS",VR::DA,VM::VM1,"Patient Birthdate",false },
  {0x0013,0x0031,"SIEMENS CM VA0  CMS",VR::DS,VM::VM1,"Patient Weight",false },
  {0x0013,0x0032,"SIEMENS CM VA0  CMS",VR::PN,VM::VM1,"Patient Maiden Name",false },
  {0x0013,0x0033,"SIEMENS CM VA0  CMS",VR::PN,VM::VM1,"Referring Physician",false },
  {0x0013,0x0034,"SIEMENS CM VA0  CMS",VR::LO,VM::VM1,"Admitting Diagnosis",false },
  {0x0013,0x0035,"SIEMENS CM VA0  CMS",VR::CS,VM::VM1,"Patient Sex",false },
  {0x0013,0x0040,"SIEMENS CM VA0  CMS",VR::LO,VM::VM1,"Procedure Description",false },
  {0x0013,0x0042,"SIEMENS CM VA0  CMS",VR::CS,VM::VM1,"Rest Direction",false },
  {0x0013,0x0044,"SIEMENS CM VA0  CMS",VR::CS,VM::VM1,"Patient Position",false },
  {0x0013,0x0046,"SIEMENS CM VA0  CMS",VR::SH,VM::VM1,"View Direction",false },
  {0x0013,0x0050,"SIEMENS CM VA0  CMS",VR::LO,VM::VM1,"?",false },
  {0x0013,0x0051,"SIEMENS CM VA0  CMS",VR::LO,VM::VM1,"?",false },
  {0x0013,0x0052,"SIEMENS CM VA0  CMS",VR::LO,VM::VM1,"?",false },
  {0x0013,0x0053,"SIEMENS CM VA0  CMS",VR::LO,VM::VM1,"?",false },
  {0x0013,0x0054,"SIEMENS CM VA0  CMS",VR::LO,VM::VM1,"?",false },
  {0x0013,0x0055,"SIEMENS CM VA0  CMS",VR::LO,VM::VM1,"?",false },
  {0x0013,0x0056,"SIEMENS CM VA0  CMS",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0010,"SIEMENS CM VA0  CMS",VR::IS,VM::VM1,"Net Frequency",false },
  {0x0019,0x0020,"SIEMENS CM VA0  CMS",VR::CS,VM::VM1,"Measurement Mode",false },
  {0x0019,0x0030,"SIEMENS CM VA0  CMS",VR::CS,VM::VM1,"Calculation Mode",false },
  {0x0019,0x0050,"SIEMENS CM VA0  CMS",VR::IS,VM::VM1,"Noise Level",false },
  {0x0019,0x0060,"SIEMENS CM VA0  CMS",VR::IS,VM::VM1,"Number of Data Bytes",false },
  {0x0019,0x0070,"SIEMENS CM VA0  CMS",VR::DS,VM::VM1_n,"?",false },
  {0x0019,0x0080,"SIEMENS CM VA0  CMS",VR::LO,VM::VM1,"?",false },
  {0x0021,0x0020,"SIEMENS CM VA0  CMS",VR::DS,VM::VM1_n,"Field of View",false },
  {0x0021,0x0022,"SIEMENS CM VA0  CMS",VR::DS,VM::VM1,"Image Magnification Factor",false },
  {0x0021,0x0024,"SIEMENS CM VA0  CMS",VR::DS,VM::VM1_n,"Image Scroll Offset",false },
  {0x0021,0x0026,"SIEMENS CM VA0  CMS",VR::IS,VM::VM1,"Image Pixel Offset",false },
  {0x0021,0x0030,"SIEMENS CM VA0  CMS",VR::CS,VM::VM1,"View Direction",false },
  {0x0021,0x0032,"SIEMENS CM VA0  CMS",VR::CS,VM::VM1,"Rest Direction",false },
  {0x0021,0x0060,"SIEMENS CM VA0  CMS",VR::DS,VM::VM1_n,"Image Position",false },
  {0x0021,0x0061,"SIEMENS CM VA0  CMS",VR::DS,VM::VM1_n,"Image Normal",false },
  {0x0021,0x0063,"SIEMENS CM VA0  CMS",VR::DS,VM::VM1,"Image Distance",false },
  {0x0021,0x0065,"SIEMENS CM VA0  CMS",VR::US,VM::VM1,"Image Positioning History Mask",false },
  {0x0021,0x006a,"SIEMENS CM VA0  CMS",VR::DS,VM::VM1_n,"Image Row",false },
  {0x0021,0x006b,"SIEMENS CM VA0  CMS",VR::DS,VM::VM1_n,"Image Column",false },
  {0x0021,0x0070,"SIEMENS CM VA0  CMS",VR::CS,VM::VM1_n,"Patient Orientation Set 1",false },
  {0x0021,0x0071,"SIEMENS CM VA0  CMS",VR::CS,VM::VM1_n,"Patient Orientation Set 2",false },
  {0x0021,0x0080,"SIEMENS CM VA0  CMS",VR::LO,VM::VM1,"Study Name",false },
  {0x0021,0x0082,"SIEMENS CM VA0  CMS",VR::SH,VM::VM1,"Study Type",false },
  {0x0029,0x0010,"SIEMENS CM VA0  CMS",VR::CS,VM::VM1,"Window Style",false },
  {0x0029,0x0011,"SIEMENS CM VA0  CMS",VR::CS,VM::VM1,"Window Look",false },
  {0x0029,0x0012,"SIEMENS CM VA0  CMS",VR::UL,VM::VM1,"Window Table Index",false },
  {0x0029,0x0013,"SIEMENS CM VA0  CMS",VR::CS,VM::VM1,"Window Scale",false },
  {0x0029,0x0020,"SIEMENS CM VA0  CMS",VR::CS,VM::VM1_n,"Pixel Quality Code",false },
  {0x0029,0x0022,"SIEMENS CM VA0  CMS",VR::IS,VM::VM1_n,"Pixel Quality Value",false },
  {0x0029,0x0050,"SIEMENS CM VA0  CMS",VR::CS,VM::VM1,"Archive Code",false },
  {0x0029,0x0051,"SIEMENS CM VA0  CMS",VR::CS,VM::VM1,"Exposure Code",false },
  {0x0029,0x0052,"SIEMENS CM VA0  CMS",VR::IS,VM::VM1,"Sort Code",false },
  {0x0029,0x0053,"SIEMENS CM VA0  CMS",VR::CS,VM::VM1,"Remote Copy Code",false },
  {0x0029,0x0060,"SIEMENS CM VA0  CMS",VR::IS,VM::VM1,"Splash",false },
  {0x0051,0x0010,"SIEMENS CM VA0  CMS",VR::LO,VM::VM1_n,"Image Text",false },
  {0x6021,0x0000,"SIEMENS CM VA0  CMS",VR::CS,VM::VM1,"Image Graphics Format Code",false },
  {0x6021,0x0010,"SIEMENS CM VA0  CMS",VR::OB,VM::VM1,"Image Graphics",false },
  {0x7fe1,0x0000,"SIEMENS CM VA0  CMS",VR::OB,VM::VM1_n,"Binary Data",false },
  {0x0019,0x003d,"SIEMENS CM VA0  COAD",VR::LO,VM::VM1,"Reading Code RZ2 Comment 2",false },
  {0x0009,0x0010,"SIEMENS CM VA0  LAB",VR::LO,VM::VM1,"Generator Identification Label",false },
  {0x0009,0x0011,"SIEMENS CM VA0  LAB",VR::LO,VM::VM1,"Gantry Identification Label",false },
  {0x0009,0x0012,"SIEMENS CM VA0  LAB",VR::LO,VM::VM1,"XRay Tube Identification Label",false },
  {0x0009,0x0013,"SIEMENS CM VA0  LAB",VR::LO,VM::VM1,"Detector Identification Label",false },
  {0x0009,0x0014,"SIEMENS CM VA0  LAB",VR::LO,VM::VM1,"DAS Identification Label",false },
  {0x0009,0x0015,"SIEMENS CM VA0  LAB",VR::LO,VM::VM1,"SMI Identification Label",false },
  {0x0009,0x0016,"SIEMENS CM VA0  LAB",VR::LO,VM::VM1,"CPU Identification Label",false },
  {0x0009,0x0020,"SIEMENS CM VA0  LAB",VR::SH,VM::VM1,"Header Version",false },
  {0x6023,0x0000,"SIEMENS CM VA0  OST",VR::LO,VM::VM1,"Osteo Contour Comment",false },
  {0x6023,0x0010,"SIEMENS CM VA0  OST",VR::SS,VM::VM1_n,"Osteo Contour Buffer",false },
  {0x0029,0x0010,"SIEMENS CSA ENVELOPE",VR::OB,VM::VM1,"syngo Report Data",false },
  {0x0029,0x0011,"SIEMENS CSA ENVELOPE",VR::OB,VM::VM1,"syngo Report Presentation",false },
  {0x0029,0x0008,"SIEMENS CSA HEADER",VR::CS,VM::VM1,"CSA Image Header Type",false },
  {0x0029,0x0009,"SIEMENS CSA HEADER",VR::LO,VM::VM1,"CSA Image Header Version",false },
  {0x0029,0x0010,"SIEMENS CSA HEADER",VR::OB,VM::VM1,"CSA Image Header Info",false },
  {0x0029,0x0018,"SIEMENS CSA HEADER",VR::CS,VM::VM1,"CSA Series Header Type",false },
  {0x0029,0x0019,"SIEMENS CSA HEADER",VR::LO,VM::VM1,"CSA Series Header Version",false },
  {0x0029,0x0020,"SIEMENS CSA HEADER",VR::OB,VM::VM1,"CSA Series Header Info",false },
  {0x0029,0x0008,"SIEMENS CSA NON-IMAGE",VR::CS,VM::VM1,"CSA Data Type",false },
  {0x0029,0x0009,"SIEMENS CSA NON-IMAGE",VR::LO,VM::VM1,"CSA Data Version",false },
  {0x0029,0x0010,"SIEMENS CSA NON-IMAGE",VR::OB,VM::VM1,"CSA Data Info",false },
  {0x7fe1,0x0010,"SIEMENS CSA NON-IMAGE",VR::OB,VM::VM1,"CSA Data",false },
  {0x0029,0x0008,"SIEMENS CSA REPORT",VR::CS,VM::VM1,"syngo Report Type",false },
  {0x0029,0x0009,"SIEMENS CSA REPORT",VR::LO,VM::VM1,"syngo Report",false },
  {0x0029,0x0015,"SIEMENS CSA REPORT",VR::US,VM::VM1,"SR Variant",false },
  {0x0029,0x0017,"SIEMENS CSA REPORT",VR::UI,VM::VM1,"SC SOP Instance UID",false },
  {0x0019,0x0010,"SIEMENS CT VA0  COAD",VR::IS,VM::VM1,"Distance Source to SourceSide Collimator",false },
  {0x0019,0x0011,"SIEMENS CT VA0  COAD",VR::IS,VM::VM1,"Distance Source to DetectorSide Collimator",false },
  {0x0019,0x0020,"SIEMENS CT VA0  COAD",VR::IS,VM::VM1,"Number of Possible Channels",false },
  {0x0019,0x0021,"SIEMENS CT VA0  COAD",VR::IS,VM::VM1,"Mean Channel Number",false },
  {0x0019,0x0022,"SIEMENS CT VA0  COAD",VR::DS,VM::VM1,"Detector Spacing",false },
  {0x0019,0x0023,"SIEMENS CT VA0  COAD",VR::DS,VM::VM1,"Detector Center",false },
  {0x0019,0x0024,"SIEMENS CT VA0  COAD",VR::DS,VM::VM1,"Reading Integration Time",false },
  {0x0019,0x0050,"SIEMENS CT VA0  COAD",VR::DS,VM::VM1,"Detector Alignment",false },
  {0x0019,0x0052,"SIEMENS CT VA0  COAD",VR::DS,VM::VM1,"Table Feed",false },
  {0x0019,0x0054,"SIEMENS CT VA0  COAD",VR::DS,VM::VM1,"Rotation Time",false },
  {0x0019,0x0060,"SIEMENS CT VA0  COAD",VR::DS,VM::VM1,"Focus Alignment",false },
  {0x0019,0x0065,"SIEMENS CT VA0  COAD",VR::UL,VM::VM1,"Focal Spot Deflection Amplitude",false },
  {0x0019,0x0066,"SIEMENS CT VA0  COAD",VR::UL,VM::VM1,"Focal Spot Deflection Phase",false },
  {0x0019,0x0067,"SIEMENS CT VA0  COAD",VR::UL,VM::VM1,"Focal Spot Deflection Offset",false },
  {0x0019,0x0070,"SIEMENS CT VA0  COAD",VR::DS,VM::VM1,"Water Scaling Factor",false },
  {0x0019,0x0071,"SIEMENS CT VA0  COAD",VR::DS,VM::VM1,"Interpolation Factor",false },
  {0x0019,0x0080,"SIEMENS CT VA0  COAD",VR::CS,VM::VM1,"Patient Region",false },
  {0x0019,0x0082,"SIEMENS CT VA0  COAD",VR::CS,VM::VM1,"Patient Phase of Life",false },
  {0x0019,0x0090,"SIEMENS CT VA0  COAD",VR::DS,VM::VM1,"Osteo Offset",false },
  {0x0019,0x0092,"SIEMENS CT VA0  COAD",VR::DS,VM::VM1,"Osteo Regression Line Slope",false },
  {0x0019,0x0093,"SIEMENS CT VA0  COAD",VR::DS,VM::VM1,"Osteo Regression Line Intercept",false },
  {0x0019,0x0094,"SIEMENS CT VA0  COAD",VR::IS,VM::VM1,"Osteo Standardization Code",false },
  {0x0019,0x0096,"SIEMENS CT VA0  COAD",VR::IS,VM::VM1,"Osteo Phantom Number",false },
  {0x0019,0x00a0,"SIEMENS CT VA0  COAD",VR::DS,VM::VM1,"Spiral Location Begin",false },
  {0x0019,0x00a1,"SIEMENS CT VA0  COAD",VR::DS,VM::VM1,"Spiral Location End",false },
  {0x0019,0x00a2,"SIEMENS CT VA0  COAD",VR::CS,VM::VM1,"Spiral z Profile",false },
  {0x0019,0x00a3,"SIEMENS CT VA0  COAD",VR::CS,VM::VM1,"Spiral Reconstruction Direction",false },
  {0x0019,0x00a4,"SIEMENS CT VA0  COAD",VR::DS,VM::VM1,"Spiral Reconstruction Increment",false },
  {0x0019,0x00a5,"SIEMENS CT VA0  COAD",VR::DS,VM::VM1,"Scan Tilt Plane Distance",false },
  {0x0019,0x00a6,"SIEMENS CT VA0  COAD",VR::UL,VM::VM1_n,"Mode Processing Step Mask",false },
  {0x0019,0x00a7,"SIEMENS CT VA0  COAD",VR::UL,VM::VM1,"Mode Processing Control Mask",false },
  {0x0019,0x00a8,"SIEMENS CT VA0  COAD",VR::UL,VM::VM1_n,"Mode Master Mask",false },
  {0x0019,0x00a9,"SIEMENS CT VA0  COAD",VR::IS,VM::VM1,"Mode Store Step",false },
  {0x0019,0x00aa,"SIEMENS CT VA0  COAD",VR::CS,VM::VM1,"Focus Size Type",false },
  {0x0019,0x00ab,"SIEMENS CT VA0  COAD",VR::IS,VM::VM1,"Mode Adaptive Filter Kind",false },
  {0x0019,0x00ac,"SIEMENS CT VA0  COAD",VR::IS,VM::VM1,"Mode MBH Kind",false },
  {0x0019,0x00ad,"SIEMENS CT VA0  COAD",VR::IS,VM::VM1,"Mode Balancing Kind",false },
  {0x0019,0x00ae,"SIEMENS CT VA0  COAD",VR::IS,VM::VM1,"Number of Readings per Rotation",false },
  {0x0019,0x00af,"SIEMENS CT VA0  COAD",VR::DS,VM::VM1,"Correction Table Reference Time",false },
  {0x0019,0x00b0,"SIEMENS CT VA0  COAD",VR::DS,VM::VM1,"Feed per Rotation",false },
  {0x0019,0x00b1,"SIEMENS CT VA0  COAD",VR::SH,VM::VM1,"Admin Data Version",true},
  {0x0019,0x00bd,"SIEMENS CT VA0  COAD",VR::IS,VM::VM1,"Pulmo Trigger Level",false },
  {0x0019,0x00be,"SIEMENS CT VA0  COAD",VR::DS,VM::VM1,"Expiratoric Reserve",false },
  {0x0019,0x00bf,"SIEMENS CT VA0  COAD",VR::DS,VM::VM1,"Vital Capacity",false },
  {0x0019,0x00c0,"SIEMENS CT VA0  COAD",VR::DS,VM::VM1,"Pulmo Water",false },
  {0x0019,0x00c1,"SIEMENS CT VA0  COAD",VR::DS,VM::VM1,"Pulmo Air",false },
  {0x0019,0x00c2,"SIEMENS CT VA0  COAD",VR::DA,VM::VM1,"Pulmo Date",false },
  {0x0019,0x00c3,"SIEMENS CT VA0  COAD",VR::TM,VM::VM1,"Pulmo Time",false },
  {0x0019,0x00c4,"SIEMENS CT VA0  COAD",VR::UL,VM::VM1,"?",false },
  {0x0019,0x00c5,"SIEMENS CT VA0  COAD",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0010,"SIEMENS CT VA0  GEN",VR::DS,VM::VM1,"Source Side Collimator Aperture",false },
  {0x0019,0x0011,"SIEMENS CT VA0  GEN",VR::DS,VM::VM1,"Detector Side Collimator Aperture",false },
  {0x0019,0x0020,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"Exposure Time current",false },
  {0x0019,0x0021,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"Exposure current",false },
  {0x0019,0x0025,"SIEMENS CT VA0  GEN",VR::DS,VM::VM1,"Generator Power current",false },
  {0x0019,0x0026,"SIEMENS CT VA0  GEN",VR::DS,VM::VM1,"Generator Voltage current",false },
  {0x0019,0x0040,"SIEMENS CT VA0  GEN",VR::UL,VM::VM1,"Master Control Mask",false },
  {0x0019,0x0042,"SIEMENS CT VA0  GEN",VR::US,VM::VM1_n,"Processing Mask",false },
  {0x0019,0x0044,"SIEMENS CT VA0  GEN",VR::UL,VM::VM1_n,"IRS Data Number",false },
  {0x0019,0x0045,"SIEMENS CT VA0  GEN",VR::UL,VM::VM1_n,"IRS Data Length",false },
  {0x0019,0x0062,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"Number of Virtual Channels",false },
  {0x0019,0x0070,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"Number of Reading",false },
  {0x0019,0x0071,"SIEMENS CT VA0  GEN",VR::CS,VM::VM1,"Reading Code",false },
  {0x0019,0x0074,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"Number of Projections",false },
  {0x0019,0x0075,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"Number of Bytes",false },
  {0x0019,0x0080,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1_n,"Reconstruction Algorithm Set",false },
  {0x0019,0x0081,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"Reconstruction Algorithm Index",false },
  {0x0019,0x0082,"SIEMENS CT VA0  GEN",VR::LO,VM::VM1,"Regeneration Software Version",false },
  {0x0019,0x0084,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"Multiscan Reconstruction Increment",false },
  {0x0019,0x0085,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"Multiscan Reconstruction Width",false },
  {0x0019,0x0088,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"Range Number",false },
  {0x0021,0x0010,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"Rotation Angle",false },
  {0x0021,0x0011,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"Start Angle",false },
  {0x0021,0x0020,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"Scan",false },
  {0x0021,0x0030,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"Tube Position",false },
  {0x0021,0x0032,"SIEMENS CT VA0  GEN",VR::DS,VM::VM1,"Length of Topogram",false },
  {0x0021,0x0034,"SIEMENS CT VA0  GEN",VR::DS,VM::VM1,"Correction Factor",false },
  {0x0021,0x0036,"SIEMENS CT VA0  GEN",VR::DS,VM::VM1,"Maximum Table Position",false },
  {0x0021,0x0040,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"Table Move Direction Code",false },
  {0x0021,0x0045,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"VOI Start Row",false },
  {0x0021,0x0046,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"VOI Stop Row",false },
  {0x0021,0x0047,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"VOI Start Column",false },
  {0x0021,0x0048,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"VOI Stop Column",false },
  {0x0021,0x0049,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"VOI Start Slice",false },
  {0x0021,0x004a,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"VOI Stop Slice",false },
  {0x0021,0x0050,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"Vector Start Row",false },
  {0x0021,0x0051,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"Vector Row Step",false },
  {0x0021,0x0052,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"Vector Start Column",false },
  {0x0021,0x0053,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"Vector Column Step",false },
  {0x0021,0x0060,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"Range Type Code",false },
  {0x0021,0x0062,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"Reference Type Code",false },
  {0x0021,0x0070,"SIEMENS CT VA0  GEN",VR::DS,VM::VM1_n,"Object Orientation",false },
  {0x0021,0x0072,"SIEMENS CT VA0  GEN",VR::DS,VM::VM1_n,"Light Orientation",false },
  {0x0021,0x0075,"SIEMENS CT VA0  GEN",VR::DS,VM::VM1,"Light Brightness",false },
  {0x0021,0x0076,"SIEMENS CT VA0  GEN",VR::DS,VM::VM1,"Light Contrast",false },
  {0x0021,0x007a,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1_n,"Overlay Threshold",false },
  {0x0021,0x007b,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1_n,"Surface Threshold",false },
  {0x0021,0x007c,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1_n,"Grey Scale Threshold",false },
  {0x0021,0x00a0,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1_n,"Field of Interest",false },
  {0x0021,0x00a1,"SIEMENS CT VA0  GEN",VR::CS,VM::VM1,"Image Filter Label",false },
  {0x0021,0x00a2,"SIEMENS CT VA0  GEN",VR::CS,VM::VM1,"Exemplary Element 1",false },
  {0x0021,0x00a3,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1_n,"Exemplary Element 2",false },
  {0x0021,0x00a4,"SIEMENS CT VA0  GEN",VR::LO,VM::VM1,"Exemplary Element 3",false },
  {0x0021,0x00a5,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"Reference Data Set 1",false },
  {0x0021,0x00a6,"SIEMENS CT VA0  GEN",VR::IS,VM::VM1,"Reference Data Set 2",false },
  {0x0021,0x00a7,"SIEMENS CT VA0  GEN",VR::LO,VM::VM1,"Window Program Name",false },
  {0x0009,0x0010,"SIEMENS CT VA0  IDE",VR::LO,VM::VM1,"SOMATOM Software Name",false },
  {0x0009,0x0030,"SIEMENS CT VA0  IDE",VR::CS,VM::VM1,"Data Set Subtype 1",false },
  {0x0009,0x0031,"SIEMENS CT VA0  IDE",VR::SH,VM::VM1,"Data Set Subtype 2",false },
  {0x0009,0x0032,"SIEMENS CT VA0  IDE",VR::SH,VM::VM1,"Data Set Subtype 3",false },
  {0x0009,0x0033,"SIEMENS CT VA0  IDE",VR::SH,VM::VM1,"Data Set Subtype 4",false },
  {0x0009,0x0034,"SIEMENS CT VA0  IDE",VR::SH,VM::VM1,"Data Set Subtype 5",false },
  {0x0009,0x0040,"SIEMENS CT VA0  IDE",VR::LO,VM::VM1,"Directory Text",false },
  {0x0009,0x0042,"SIEMENS CT VA0  IDE",VR::LO,VM::VM1,"Exposure Time Text",false },
  {0x0009,0x0050,"SIEMENS CT VA0  IDE",VR::DA,VM::VM1,"Scan Date",false },
  {0x0009,0x0051,"SIEMENS CT VA0  IDE",VR::TM,VM::VM1,"Scan Time",false },
  {0x0009,0x0020,"SIEMENS CT VA0  ORI",VR::LO,VM::VM1,"Original Data Set Owner",false },
  {0x0009,0x0030,"SIEMENS CT VA0  ORI",VR::SH,VM::VM1,"Original Modality",false },
  {0x0021,0x0080,"SIEMENS CT VA0  ORMR",VR::DS,VM::VM1,"Repetition Time",false },
  {0x0021,0x0081,"SIEMENS CT VA0  ORMR",VR::DS,VM::VM1,"Echo Time",false },
  {0x0021,0x0082,"SIEMENS CT VA0  ORMR",VR::DS,VM::VM1,"Inversion Time",false },
  {0x0021,0x0083,"SIEMENS CT VA0  ORMR",VR::IS,VM::VM1,"Number of Averages",false },
  {0x0021,0x0084,"SIEMENS CT VA0  ORMR",VR::DS,VM::VM1,"Imaging Frequency",false },
  {0x0021,0x0085,"SIEMENS CT VA0  ORMR",VR::LO,VM::VM1,"Imaged Nucleus",false },
  {0x0021,0x0086,"SIEMENS CT VA0  ORMR",VR::IS,VM::VM1,"Echo Number",false },
  {0x0021,0x0090,"SIEMENS CT VA0  ORMR",VR::DS,VM::VM1,"Flip Angle",false },
  {0x6021,0x0000,"SIEMENS CT VA0  OST",VR::LO,VM::VM1,"Osteo Contour Comment",false },
  {0x6021,0x0010,"SIEMENS CT VA0  OST",VR::US,VM::VM256,"Osteo Contour Buffer",false },
  {0x0021,0x0010,"SIEMENS CT VA0  RAW",VR::UL,VM::VM1_n,"Creation Mask service",false },
  {0x0021,0x0020,"SIEMENS CT VA0  RAW",VR::UL,VM::VM1_n,"Evaluation Mask service",false },
  {0x0021,0x0030,"SIEMENS CT VA0  RAW",VR::US,VM::VM1_n,"Extended Processing Mask",false },
  {0x0021,0x0040,"SIEMENS CT VA0  RAW",VR::UL,VM::VM1_n,"Manipulation Mask",false },
  {0x0021,0x0041,"SIEMENS CT VA0  RAW",VR::UL,VM::VM1_n,"Composition Mask",false },
  {0x0021,0x0042,"SIEMENS CT VA0  RAW",VR::UL,VM::VM1_n,"Alteration Mask",false },
  {0x0021,0x0043,"SIEMENS CT VA0  RAW",VR::UL,VM::VM1_n,"Generation Mask",false },
  {0x0021,0x0044,"SIEMENS CT VA0  RAW",VR::UL,VM::VM1_n,"Function Mask",false },
  {0x0021,0x0050,"SIEMENS CT VA0  RAW",VR::CS,VM::VM1,"Mode Mirroring Code",false },
  {0x0021,0x0010,"SIEMENS DFR.01",VR::US,VM::VM1,"Graphic Evaluation Mask",false },
  {0x0021,0x0011,"SIEMENS DFR.01",VR::US,VM::VM1,"Type 1st Graphic Element",false },
  {0x0021,0x0012,"SIEMENS DFR.01",VR::LO,VM::VM1,"Description 1st Graphic Element",false },
  {0x0021,0x0013,"SIEMENS DFR.01",VR::US,VM::VM1,"Type 2nd Graphic Element",false },
  {0x0021,0x0014,"SIEMENS DFR.01",VR::LO,VM::VM1,"Description 2nd Graphic Element",false },
  {0x0021,0x0015,"SIEMENS DFR.01",VR::US,VM::VM1,"Type 3rd Graphic Element",false },
  {0x0021,0x0016,"SIEMENS DFR.01",VR::LO,VM::VM1,"Description 3rd Graphic Element",false },
  {0x0021,0x0017,"SIEMENS DFR.01",VR::US,VM::VM1,"Type 4th Graphic Element",false },
  {0x0021,0x0018,"SIEMENS DFR.01",VR::LO,VM::VM1,"Description 4th Graphic Element",false },
  {0x0021,0x0019,"SIEMENS DFR.01",VR::US,VM::VM1,"Type 5th Graphic Element",false },
  {0x0021,0x0020,"SIEMENS DFR.01",VR::LO,VM::VM1,"Description 5th Graphic Element",false },
  {0x0021,0x0021,"SIEMENS DFR.01",VR::US,VM::VM1,"Type 6th Graphic Element",false },
  {0x0021,0x0022,"SIEMENS DFR.01",VR::LO,VM::VM1,"Description 6th Graphic Element",false },
  {0x0021,0x0023,"SIEMENS DFR.01",VR::US,VM::VM1,"Type 7th Graphic Element",false },
  {0x0021,0x0024,"SIEMENS DFR.01",VR::LO,VM::VM1,"Description 7th Graphic Element",false },
  {0x0021,0x0025,"SIEMENS DFR.01",VR::US,VM::VM1,"Type 8th Graphic Element",false },
  {0x0021,0x0026,"SIEMENS DFR.01",VR::LO,VM::VM1,"Description 8th Graphic Element",false },
  {0x0021,0x0027,"SIEMENS DFR.01",VR::US,VM::VM1,"Type 9th Graphic Element",false },
  {0x0021,0x0028,"SIEMENS DFR.01",VR::LO,VM::VM1,"Description 9th Graphic Element",false },
  {0x0021,0x0029,"SIEMENS DFR.01",VR::US,VM::VM1,"Type 10th Graphic Element",false },
  {0x0021,0x0030,"SIEMENS DFR.01",VR::LO,VM::VM1,"Description 10th Graphic Element",false },
  {0x0017,0x0011,"SIEMENS DFR.01 MANIPULATED",VR::US,VM::VM1,"Window Width Native Current",false },
  {0x0017,0x0012,"SIEMENS DFR.01 MANIPULATED",VR::US,VM::VM1,"Window Center Native",false },
  {0x0017,0x0014,"SIEMENS DFR.01 MANIPULATED",VR::US,VM::VM1,"Window Width Subtraction Current",false },
  {0x0017,0x0015,"SIEMENS DFR.01 MANIPULATED",VR::US,VM::VM1,"Window Center Subtraction Current",false },
  {0x0017,0x0025,"SIEMENS DFR.01 MANIPULATED",VR::DS,VM::VM1,"Edge Enhancement Native Current",false },
  {0x0017,0x0027,"SIEMENS DFR.01 MANIPULATED",VR::DS,VM::VM1,"Edge Enhancement Subtraction Current",false },
  {0x0017,0x0029,"SIEMENS DFR.01 MANIPULATED",VR::IS,VM::VM1,"Edge Enhancement %",false },
  {0x0017,0x0030,"SIEMENS DFR.01 MANIPULATED",VR::IS,VM::VM1,"Harmonization %",false },
  {0x0017,0x0031,"SIEMENS DFR.01 MANIPULATED",VR::US,VM::VM1,"DSA Representation Current",false },
  {0x0017,0x0032,"SIEMENS DFR.01 MANIPULATED",VR::US,VM::VM1,"Number of Mask Current",false },
  {0x0017,0x0033,"SIEMENS DFR.01 MANIPULATED",VR::US,VM::VM1,"RightLeft Mark Current",false },
  {0x0017,0x0035,"SIEMENS DFR.01 MANIPULATED",VR::US,VM::VM1,"Number of Series Representative",false },
  {0x0017,0x0037,"SIEMENS DFR.01 MANIPULATED",VR::US,VM::VM1,"Rectangle Shutter Horizontal Current",false },
  {0x0017,0x0038,"SIEMENS DFR.01 MANIPULATED",VR::US,VM::VM1,"Rectangle Shutter Vertical Aktuell",false },
  {0x0017,0x0071,"SIEMENS DFR.01 MANIPULATED",VR::LO,VM::VM1,"Landmark Current",false },
  {0x0017,0x0072,"SIEMENS DFR.01 MANIPULATED",VR::US,VM::VM1,"Zoom Factor",false },
  {0x0017,0x0073,"SIEMENS DFR.01 MANIPULATED",VR::US,VM::VM1,"Roam Horizontal",false },
  {0x0017,0x0074,"SIEMENS DFR.01 MANIPULATED",VR::US,VM::VM1,"Roam Vertical",false },
  {0x0017,0x0077,"SIEMENS DFR.01 MANIPULATED",VR::IS,VM::VM1,"Pixel Shift Horizontal",false },
  {0x0017,0x0078,"SIEMENS DFR.01 MANIPULATED",VR::IS,VM::VM1,"Pixel Shift Vertical",false },
  {0x0017,0x0079,"SIEMENS DFR.01 MANIPULATED",VR::US,VM::VM1,"Image Flip",false },
  {0x0017,0x007a,"SIEMENS DFR.01 MANIPULATED",VR::US,VM::VM1,"Invers Current",false },
  {0x0017,0x0080,"SIEMENS DFR.01 MANIPULATED",VR::US,VM::VM1,"?",false },
  {0x0017,0x0083,"SIEMENS DFR.01 MANIPULATED",VR::SH,VM::VM1,"Documentation Flag",false },
  {0x0017,0x0084,"SIEMENS DFR.01 MANIPULATED",VR::LO,VM::VM1,"Right Marker",false },
  {0x0017,0x0091,"SIEMENS DFR.01 MANIPULATED",VR::US,VM::VM1,"Summation Count Filling",false },
  {0x0017,0x0092,"SIEMENS DFR.01 MANIPULATED",VR::US,VM::VM1,"Summation Count Mask",false },
  {0x0017,0x0093,"SIEMENS DFR.01 MANIPULATED",VR::US,VM::VM1,"PeakOp Start",false },
  {0x0017,0x0094,"SIEMENS DFR.01 MANIPULATED",VR::US,VM::VM1,"PeakOp End",false },
  {0x0017,0x00a0,"SIEMENS DFR.01 MANIPULATED",VR::IS,VM::VM1,"Cine Rate",false },
  {0x0017,0x00a1,"SIEMENS DFR.01 MANIPULATED",VR::LO,VM::VM1,"Image Name",false },
  {0x0017,0x00a2,"SIEMENS DFR.01 MANIPULATED",VR::LO,VM::VM1,"Image Name Extension 1",false },
  {0x0017,0x00a3,"SIEMENS DFR.01 MANIPULATED",VR::LO,VM::VM1,"Image Name Extension 2",false },
  {0x0017,0x0010,"SIEMENS DFR.01 ORIGINAL",VR::CS,VM::VM1,"Image Type",false },
  {0x0017,0x0011,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1,"Window Width Native Original",false },
  {0x0017,0x0012,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1,"Window Center Native",false },
  {0x0017,0x0014,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1,"Window Width Subtraction Original",false },
  {0x0017,0x0015,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1,"Window Center Subtraction Original",false },
  {0x0017,0x0016,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1,"LUT Number for Autowindow",false },
  {0x0017,0x0018,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1,"?",false },
  {0x0017,0x0021,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1,"Edge Enhancement Kernelsize Original",false },
  {0x0017,0x0022,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1,"Edge Enhancement Mode Original",false },
  {0x0017,0x0023,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1_n,"Edge Enhancement LUTs Original",false },
  {0x0017,0x0024,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1,"Filter Type Native Original",false },
  {0x0017,0x0025,"SIEMENS DFR.01 ORIGINAL",VR::DS,VM::VM1,"Edge Enhancement Native Original",false },
  {0x0017,0x0026,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1,"Filter Type Subtraction Original",false },
  {0x0017,0x0027,"SIEMENS DFR.01 ORIGINAL",VR::DS,VM::VM1,"Edge Enhancement Subtraction Original",false },
  {0x0017,0x002a,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1,"?",false },
  {0x0017,0x0030,"SIEMENS DFR.01 ORIGINAL",VR::IS,VM::VM1,"?",false },
  {0x0017,0x0031,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1,"DSA Representation Original",false },
  {0x0017,0x0032,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1,"Number of Maske Original",false },
  {0x0017,0x0033,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1,"RightLeft Mark Original",false },
  {0x0017,0x0034,"SIEMENS DFR.01 ORIGINAL",VR::DS,VM::VM1,"Slice Position",false },
  {0x0017,0x0037,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1,"Rectangle Shutter Horizontal Original",false },
  {0x0017,0x0038,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1,"Rectangle Vertical Original",false },
  {0x0017,0x0041,"SIEMENS DFR.01 ORIGINAL",VR::DS,VM::VM1,"Dose",false },
  {0x0017,0x0042,"SIEMENS DFR.01 ORIGINAL",VR::DS,VM::VM1,"Dose Area Product",false },
  {0x0017,0x0043,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1,"Focus",false },
  {0x0017,0x0044,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1,"SDM",false },
  {0x0017,0x0045,"SIEMENS DFR.01 ORIGINAL",VR::DS,VM::VM1,"Image Intensifer Format",false },
  {0x0017,0x0046,"SIEMENS DFR.01 ORIGINAL",VR::DS,VM::VM1,"Density Correction",false },
  {0x0017,0x0047,"SIEMENS DFR.01 ORIGINAL",VR::LO,VM::VM1,"Exposure Curves",false },
  {0x0017,0x0048,"SIEMENS DFR.01 ORIGINAL",VR::IS,VM::VM1,"?",false },
  {0x0017,0x0049,"SIEMENS DFR.01 ORIGINAL",VR::LO,VM::VM1,"Organ Program Name",false },
  {0x0017,0x004a,"SIEMENS DFR.01 ORIGINAL",VR::IS,VM::VM1,"Count of Images in Series",false },
  {0x0017,0x0051,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1,"Image Intensifier Zoom",false },
  {0x0017,0x0052,"SIEMENS DFR.01 ORIGINAL",VR::UL,VM::VM1,"ROSA Receiver ErrCount",false },
  {0x0017,0x0061,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1,"Table Position Peri",false },
  {0x0017,0x0062,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1_n,"Series Numbers of Petri",false },
  {0x0017,0x0071,"SIEMENS DFR.01 ORIGINAL",VR::LO,VM::VM1,"Landmark Original",false },
  {0x0017,0x0072,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1,"?",false },
  {0x0017,0x0073,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1,"?",false },
  {0x0017,0x0074,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1,"?",false },
  {0x0017,0x0079,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1,"?",false },
  {0x0017,0x007a,"SIEMENS DFR.01 ORIGINAL",VR::US,VM::VM1,"Invers Original",false },
  {0x0017,0x007b,"SIEMENS DFR.01 ORIGINAL",VR::IS,VM::VM1,"?",false },
  {0x0017,0x00a0,"SIEMENS DFR.01 ORIGINAL",VR::IS,VM::VM1,"?",false },
  {0x0017,0x00b1,"SIEMENS DFR.01 ORIGINAL",VR::CS,VM::VM1,"L2 Image",false },
  {0x0017,0x00b2,"SIEMENS DFR.01 ORIGINAL",VR::IS,VM::VM1,"L2 Acquisition Monitor",false },
  {0x0017,0x00c1,"SIEMENS DFR.01 ORIGINAL",VR::LO,VM::VM1,"Folder UID",false },
  {0x0017,0x00c2,"SIEMENS DFR.01 ORIGINAL",VR::LO,VM::VM1,"Series UID",false },
  {0x0017,0x00c3,"SIEMENS DFR.01 ORIGINAL",VR::LO,VM::VM1,"Subseries UID",false },
  {0x0021,0x0010,"SIEMENS DFR.01 SERVICE",VR::US,VM::VM1,"1 Plot",false },
  {0x0021,0x0011,"SIEMENS DFR.01 SERVICE",VR::US,VM::VM1,"Width 1 Plot",false },
  {0x0021,0x0012,"SIEMENS DFR.01 SERVICE",VR::US,VM::VM1,"2 Plot",false },
  {0x0021,0x0013,"SIEMENS DFR.01 SERVICE",VR::US,VM::VM1,"Width 2 Plot",false },
  {0x0021,0x0020,"SIEMENS DFR.01 SERVICE",VR::US,VM::VM1_5,"xCoordinate Region of Interest",false },
  {0x0021,0x0021,"SIEMENS DFR.01 SERVICE",VR::US,VM::VM1_5,"yCoordinate Region of Interest",false },
  {0x0021,0x0022,"SIEMENS DFR.01 SERVICE",VR::US,VM::VM1_5,"Width Region of Interest",false },
  {0x0021,0x0023,"SIEMENS DFR.01 SERVICE",VR::US,VM::VM1_5,"Height Region of Interest",false },
  {0x0009,0x0010,"SIEMENS DICOM",VR::UN,VM::VM1,"?",false },
  {0x0009,0x0012,"SIEMENS DICOM",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0010,"SIEMENS DLR.01",VR::LO,VM::VM1,"Measurement Mode",false },
  {0x0019,0x0011,"SIEMENS DLR.01",VR::LO,VM::VM1,"Image Type",false },
  {0x0019,0x0015,"SIEMENS DLR.01",VR::LO,VM::VM1,"Software Version",false },
  {0x0019,0x0020,"SIEMENS DLR.01",VR::LO,VM::VM1,"MPM Code",false },
  {0x0019,0x0021,"SIEMENS DLR.01",VR::LO,VM::VM1,"Latitude",false },
  {0x0019,0x0022,"SIEMENS DLR.01",VR::LO,VM::VM1,"Sensitivity",false },
  {0x0019,0x0023,"SIEMENS DLR.01",VR::LO,VM::VM1,"EDR",false },
  {0x0019,0x0024,"SIEMENS DLR.01",VR::LO,VM::VM1,"l fix",false },
  {0x0019,0x0025,"SIEMENS DLR.01",VR::LO,VM::VM1,"s fix",false },
  {0x0019,0x0026,"SIEMENS DLR.01",VR::LO,VM::VM1,"Preset Mode",false },
  {0x0019,0x0027,"SIEMENS DLR.01",VR::LO,VM::VM1,"Region",false },
  {0x0019,0x0028,"SIEMENS DLR.01",VR::LO,VM::VM1,"Subregion",false },
  {0x0019,0x0030,"SIEMENS DLR.01",VR::LO,VM::VM1,"Orientation",false },
  {0x0019,0x0031,"SIEMENS DLR.01",VR::LO,VM::VM1,"Mark on Film",false },
  {0x0019,0x0032,"SIEMENS DLR.01",VR::LO,VM::VM1,"Rotation on DRC",false },
  {0x0019,0x0040,"SIEMENS DLR.01",VR::LO,VM::VM1,"Reader Type",false },
  {0x0019,0x0041,"SIEMENS DLR.01",VR::LO,VM::VM1,"Sub Modality",false },
  {0x0019,0x0042,"SIEMENS DLR.01",VR::LO,VM::VM1,"Reader Serial Number",false },
  {0x0019,0x0050,"SIEMENS DLR.01",VR::LO,VM::VM1,"Cassette Scale",false },
  {0x0019,0x0051,"SIEMENS DLR.01",VR::LO,VM::VM1,"Cassette Matrix",false },
  {0x0019,0x0052,"SIEMENS DLR.01",VR::LO,VM::VM1,"Cassette Submatrix",false },
  {0x0019,0x0053,"SIEMENS DLR.01",VR::LO,VM::VM1,"Barcode",false },
  {0x0019,0x0060,"SIEMENS DLR.01",VR::LO,VM::VM1,"GT Contrast Type",false },
  {0x0019,0x0061,"SIEMENS DLR.01",VR::LO,VM::VM1,"GA Rotation Amount",false },
  {0x0019,0x0062,"SIEMENS DLR.01",VR::LO,VM::VM1,"GC Rotation Center",false },
  {0x0019,0x0063,"SIEMENS DLR.01",VR::LO,VM::VM1,"GS Density Shift",false },
  {0x0019,0x0064,"SIEMENS DLR.01",VR::US,VM::VM1,"RN Frequency Rank",false },
  {0x0019,0x0065,"SIEMENS DLR.01",VR::LO,VM::VM1,"RE Frequency Enhancement",false },
  {0x0019,0x0066,"SIEMENS DLR.01",VR::LO,VM::VM1,"RT Frequency Type",false },
  {0x0019,0x0067,"SIEMENS DLR.01",VR::LO,VM::VM1,"Kernel Length",false },
  {0x0019,0x0068,"SIEMENS DLR.01",VR::UL,VM::VM1,"Kernel Mode",false },
  {0x0019,0x0069,"SIEMENS DLR.01",VR::UL,VM::VM1,"Convulsion Mode",false },
  {0x0019,0x0070,"SIEMENS DLR.01",VR::LO,VM::VM1,"PLA Source",false },
  {0x0019,0x0071,"SIEMENS DLR.01",VR::LO,VM::VM1,"PLA Destination",false },
  {0x0019,0x0075,"SIEMENS DLR.01",VR::LO,VM::VM1,"UID Original Image",false },
  {0x0019,0x0076,"SIEMENS DLR.01",VR::LO,VM::VM1,"Filterset",false },
  {0x0019,0x0080,"SIEMENS DLR.01",VR::LO,VM::VM1,"Reader Header",false },
  {0x0019,0x0090,"SIEMENS DLR.01",VR::LO,VM::VM1,"PLA of Secondary Destination",false },
  {0x0019,0x00a0,"SIEMENS DLR.01",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00a1,"SIEMENS DLR.01",VR::DS,VM::VM1,"?",false },
  {0x0041,0x0010,"SIEMENS DLR.01",VR::US,VM::VM1,"Number of Hardcopies",false },
  {0x0041,0x0020,"SIEMENS DLR.01",VR::CS,VM::VM1,"Film Format",false },
  {0x0041,0x0030,"SIEMENS DLR.01",VR::LO,VM::VM1,"Film Size",false },
  {0x0041,0x0031,"SIEMENS DLR.01",VR::LO,VM::VM1,"Full Film Format",false },
  {0x0003,0x0008,"SIEMENS ISI",VR::US,VM::VM1,"ISI Command Field",false },
  {0x0003,0x0011,"SIEMENS ISI",VR::US,VM::VM1,"Attach ID Application Code",false },
  {0x0003,0x0012,"SIEMENS ISI",VR::UL,VM::VM1,"Attach ID Message Count",false },
  {0x0003,0x0013,"SIEMENS ISI",VR::DA,VM::VM1,"Attach ID Date",false },
  {0x0003,0x0014,"SIEMENS ISI",VR::TM,VM::VM1,"Attach ID Time",false },
  {0x0003,0x0020,"SIEMENS ISI",VR::US,VM::VM1,"Message Type",false },
  {0x0003,0x0030,"SIEMENS ISI",VR::DA,VM::VM1,"Max Waiting Date",false },
  {0x0003,0x0031,"SIEMENS ISI",VR::TM,VM::VM1,"Max Waiting Time",false },
  {0x0009,0x0001,"SIEMENS ISI",VR::UN,VM::VM1,"RIS Patient Info IMGEF",false },
  {0x0009,0x0014,"SIEMENS ISI",VR::SQ,VM::VM1,"RIS Patient Info IMGEF",false },
  {0x0011,0x0003,"SIEMENS ISI",VR::LO,VM::VM1,"Patient UID",false },
  {0x0011,0x0004,"SIEMENS ISI",VR::LO,VM::VM1,"Patient ID",false },
  {0x0011,0x000a,"SIEMENS ISI",VR::LO,VM::VM1,"Case ID",false },
  {0x0011,0x0022,"SIEMENS ISI",VR::LO,VM::VM1,"Request ID",false },
  {0x0011,0x0023,"SIEMENS ISI",VR::LO,VM::VM1,"Examination UID",false },
  {0x0011,0x0041,"SIEMENS ISI",VR::LO,VM::VM1,"Medical Alerts",false },
  {0x0011,0x0042,"SIEMENS ISI",VR::LO,VM::VM1,"Contrast Allergy",false },
  {0x0011,0x00a1,"SIEMENS ISI",VR::DA,VM::VM1,"Patient Registration Date",false },
  {0x0011,0x00a2,"SIEMENS ISI",VR::TM,VM::VM1,"Patient Registration Time",false },
  {0x0011,0x00b0,"SIEMENS ISI",VR::LO,VM::VM1,"Patient Last Name",false },
  {0x0011,0x00b2,"SIEMENS ISI",VR::LO,VM::VM1,"Patient First Name",false },
  {0x0011,0x00b4,"SIEMENS ISI",VR::CS,VM::VM1,"Patient Hospital Status",false },
  {0x0011,0x00bc,"SIEMENS ISI",VR::LO,VM::VM1,"Current Location Time",false },
  {0x0011,0x00c0,"SIEMENS ISI",VR::CS,VM::VM1,"Patient Insurance Status",false },
  {0x0011,0x00d0,"SIEMENS ISI",VR::LO,VM::VM1,"Patient Billing Type",false },
  {0x0011,0x00d2,"SIEMENS ISI",VR::LO,VM::VM1,"Patient Billing Address",false },
  {0x0031,0x0012,"SIEMENS ISI",VR::LO,VM::VM1,"Examination Reason",false },
  {0x0031,0x0030,"SIEMENS ISI",VR::DA,VM::VM1,"Requested Date",false },
  {0x0031,0x0032,"SIEMENS ISI",VR::TM,VM::VM1,"Worklist Request Start Time",false },
  {0x0031,0x0033,"SIEMENS ISI",VR::TM,VM::VM1,"Worklist Request End Time",false },
  {0x0031,0x004a,"SIEMENS ISI",VR::TM,VM::VM1,"Requested Time",false },
  {0x0031,0x0080,"SIEMENS ISI",VR::LO,VM::VM1,"Requested Location",false },
  {0x0055,0x0046,"SIEMENS ISI",VR::PN,VM::VM1,"Current Ward",false },
  {0x0193,0x0002,"SIEMENS ISI",VR::LO,VM::VM1,"RIS Key",false },
  {0x0307,0x0001,"SIEMENS ISI",VR::SQ,VM::VM1,"RIS Worklist IMGEF",false },
  {0x0309,0x0001,"SIEMENS ISI",VR::SQ,VM::VM1,"RIS Report IMGEF",false },
  {0x4009,0x0001,"SIEMENS ISI",VR::LO,VM::VM1,"Report ID",false },
  {0x4009,0x0020,"SIEMENS ISI",VR::CS,VM::VM1,"Report Status",false },
  {0x4009,0x0030,"SIEMENS ISI",VR::DA,VM::VM1,"Report Creation Date",false },
  {0x4009,0x0070,"SIEMENS ISI",VR::PN,VM::VM1,"Report Approving Physician",false },
  {0x4009,0x00e0,"SIEMENS ISI",VR::LO,VM::VM1,"Report Text",false },
  {0x4009,0x00e1,"SIEMENS ISI",VR::LO,VM::VM1,"Report Author",false },
  {0x4009,0x00e2,"SIEMENS ISI",VR::PN,VM::VM1,"Report Transcriber",false },
  {0x4009,0x00e3,"SIEMENS ISI",VR::PN,VM::VM1,"Reporting Radiologist",false },
  {0x0009,0x0010,"SIEMENS MED",VR::LO,VM::VM1,"Recognition Code",false },
  {0x0009,0x0030,"SIEMENS MED",VR::US,VM::VM1,"Byte Offset of Original Header",false },
  {0x0009,0x0031,"SIEMENS MED",VR::US,VM::VM1,"Length of Original Header",false },
  {0x0009,0x0040,"SIEMENS MED",VR::US,VM::VM1,"Byte Offset of Pixel Matrix",false },
  {0x0009,0x0041,"SIEMENS MED",VR::UL,VM::VM1,"Length of Pixel Matrix",false },
  {0x0009,0x0050,"SIEMENS MED",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0051,"SIEMENS MED",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0080,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0009,0x00f5,"SIEMENS MED",VR::LO,VM::VM1,"PDM EFID Placeholder",false },
  {0x0009,0x00f6,"SIEMENS MED",VR::LO,VM::VM1,"PDM Data Object Type Extension",false },
  {0x0011,0x0010,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0017,0x0000,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0017,0x0020,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0017,0x0060,"SIEMENS MED",VR::LO,VM::VM1,"",false },
  {0x0017,0x0080,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0019,0x0001,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0002,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0003,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0004,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0019,0x0005,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0019,0x0006,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x0007,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x0008,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0009,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x000a,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0019,0x000b,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0019,0x000c,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x000d,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0019,0x000e,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0019,0x000f,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x0011,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0019,0x0012,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0019,0x0013,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0014,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0015,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0016,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0017,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0018,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0019,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x001a,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x001b,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x001c,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x001d,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0020,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x0021,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x0022,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x0023,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0024,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x0025,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x0026,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x0027,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x0030,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x0031,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x0032,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x0041,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0042,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0043,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0044,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0045,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x0047,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x0048,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0019,0x0049,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x004a,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x004b,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x004c,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x004e,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x004f,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0050,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0051,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0052,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0053,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0054,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0019,0x0055,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0056,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0057,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0058,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0059,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x005a,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0019,0x005b,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x005c,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x005d,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x005e,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x005f,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0060,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0061,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0062,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0063,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0064,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0065,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0066,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0067,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0068,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x0069,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x006a,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x006b,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x006c,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x006d,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x006e,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x006f,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0070,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0019,0x0071,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0019,0x0072,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x0073,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x0074,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0075,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0076,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0077,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0078,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x007f,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0080,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0081,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0082,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x0083,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0084,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0085,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0086,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0087,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0088,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x0089,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x008a,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x008d,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x008e,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x0091,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x0092,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x0093,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x0094,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0095,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x0096,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x0097,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x00a0,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x00a1,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x00a2,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x00a3,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x00a4,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x00c0,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0019,0x00c1,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x00c2,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0019,0x00f0,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x001f,0x0000,"SIEMENS MED",VR::LO,VM::VM1,"",false },
  {0x001f,0x0002,"SIEMENS MED",VR::LO,VM::VM1,"",false },
  {0x001f,0x0050,"SIEMENS MED",VR::LO,VM::VM1,"",false },
  {0x001f,0x0070,"SIEMENS MED",VR::LO,VM::VM1,"",false },
  {0x0021,0x0010,"SIEMENS MED",VR::DS,VM::VM1,"Zoom",false },
  {0x0021,0x0011,"SIEMENS MED",VR::DS,VM::VM1_n,"Target",false },
  {0x0021,0x0012,"SIEMENS MED",VR::IS,VM::VM1,"Tube Angle During Topogram",false },
  {0x0021,0x0020,"SIEMENS MED",VR::US,VM::VM1,"ROI Mask",false },
  {0x0031,0x0001,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0031,0x0003,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0031,0x0004,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0031,0x0005,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0031,0x0006,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0031,0x0007,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0031,0x0008,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0031,0x0009,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0031,0x000a,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0031,0x000b,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0031,0x000c,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0031,0x000d,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0031,0x000e,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0031,0x000f,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0031,0x0010,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0031,0x0011,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0041,0x0012,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0041,0x0013,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0041,0x0021,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0041,0x0030,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0041,0x0032,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0041,0x0034,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0041,0x0036,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0041,0x0040,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0041,0x0050,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0041,0x0052,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0041,0x0060,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0041,0x0070,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0041,0x0071,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0051,0x0010,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0051,0x0011,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0051,0x0012,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0051,0x0013,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0051,0x0014,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0051,0x0015,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0051,0x0018,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0053,0x0001,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0053,0x0002,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0053,0x0003,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0053,0x0010,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0053,0x0011,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0053,0x0012,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0053,0x0013,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0053,0x0014,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0053,0x0015,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0053,0x0016,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0053,0x0017,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0053,0x0018,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0053,0x0019,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0053,0x001a,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0053,0x001b,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0053,0x001c,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0053,0x001d,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0053,0x0030,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0053,0x0031,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0053,0x0032,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0053,0x0033,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0053,0x0040,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0053,0x0041,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0053,0x0042,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0053,0x0043,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0053,0x0044,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0001,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0002,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0003,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0020,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0021,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0055,0x0022,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0023,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0055,0x0024,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0025,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0055,0x0029,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x002a,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0030,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0031,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0055,0x0032,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0033,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0055,0x0034,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0035,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0055,0x0038,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0039,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x003a,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0040,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0048,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0050,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0051,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0052,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0053,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0055,0x0054,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0055,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0056,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0057,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0058,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0059,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0055,0x005a,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x005b,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x005c,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0055,0x005d,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x005e,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x005f,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0060,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0061,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0064,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0065,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0066,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0067,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0068,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0069,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x006a,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x006b,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x006c,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x006d,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0070,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0071,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0072,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0073,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0074,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0075,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0076,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0077,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0078,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0080,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0081,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0082,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0083,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0084,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0085,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0086,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0087,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0088,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0089,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0055,0x008a,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x008b,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x008c,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x008d,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x008e,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x008f,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0090,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0091,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0092,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x0055,0x0093,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0094,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0095,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0096,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0097,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x0098,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x0099,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x009a,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x009b,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x009c,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x009d,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x009e,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x009f,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x00a0,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x00a1,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x00a2,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x00a3,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x00a4,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x00a5,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x00a6,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x0055,0x00a7,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x00a8,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x00a9,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x00b0,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0055,0x00b1,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0057,0x0010,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0057,0x0020,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x5001,0x0001,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x5001,0x0002,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x5001,0x0003,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x5001,0x0004,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x5001,0x0005,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x5001,0x0006,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x5001,0x0007,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x5001,0x0008,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x5001,0x0009,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x5001,0x000a,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x5001,0x000b,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x7001,0x0010,"SIEMENS MED",VR::SS,VM::VM1,"?",false },
  {0x7003,0x0010,"SIEMENS MED",VR::SL,VM::VM1,"?",false },
  {0x7005,0x0010,"SIEMENS MED",VR::SS,VM::VM1,"?",false },
  {0x7fe1,0x0010,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x7fe3,0x0014,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x7fe3,0x0015,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x7fe3,0x0016,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x7fe3,0x0017,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x7fe3,0x0018,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x7fe3,0x0019,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x7fe3,0x001a,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x7fe3,0x001b,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x7fe3,0x001c,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x7fe3,0x001e,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x7fe3,0x001f,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x7fe3,0x0020,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x7fe3,0x0021,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x7fe3,0x0022,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x7fe3,0x0023,"SIEMENS MED",VR::SL,VM::VM1,"",false },
  {0x7fe3,0x0024,"SIEMENS MED",VR::ST,VM::VM1,"",false },
  {0x7fe3,0x0025,"SIEMENS MED",VR::SS,VM::VM1,"",false },
  {0x0029,0x0004,"SIEMENS MED DISPLAY",VR::CS,VM::VM1,"Photometric Interpretation",false },
  {0x0029,0x0010,"SIEMENS MED DISPLAY",VR::US,VM::VM1,"Rows",false },
  {0x0029,0x0011,"SIEMENS MED DISPLAY",VR::US,VM::VM1,"Columns",false },
  {0x0029,0x0020,"SIEMENS MED DISPLAY",VR::US,VM::VM1,"?",false },
  {0x0029,0x0021,"SIEMENS MED DISPLAY",VR::US,VM::VM1,"?",false },
  {0x0029,0x0050,"SIEMENS MED DISPLAY",VR::US,VM::VM2,"Origin",false },
  {0x0029,0x0080,"SIEMENS MED DISPLAY",VR::US,VM::VM1,"?",false },
  {0x0029,0x0099,"SIEMENS MED DISPLAY",VR::LO,VM::VM1,"Shutter Type",false },
  {0x0029,0x00a0,"SIEMENS MED DISPLAY",VR::US,VM::VM1,"Rows of Rectangular Shutter",false },
  {0x0029,0x00a1,"SIEMENS MED DISPLAY",VR::US,VM::VM1,"Columns of Rectangular Shutter",false },
  {0x0029,0x00a2,"SIEMENS MED DISPLAY",VR::US,VM::VM1,"Origin of Rectangular Shutter",false },
  {0x0029,0x00b0,"SIEMENS MED DISPLAY",VR::US,VM::VM1,"Radius of Circular Shutter",false },
  {0x0029,0x00b2,"SIEMENS MED DISPLAY",VR::US,VM::VM1,"Origin of Circular Shutter",false },
  {0x0029,0x00c1,"SIEMENS MED DISPLAY",VR::US,VM::VM1,"Contour of Irregular Shutter",false },
  {0x0029,0x0099,"SIEMENS MED DISPLAY 0000",VR::LO,VM::VM1,"?",false },
  {0x0029,0x00c1,"SIEMENS MED DISPLAY 0000",VR::US,VM::VM1_n,"?",false },
  {0x0021,0x0000,"SIEMENS MED ECAT FILE INFO",VR::OB,VM::VM1,"ECAT_Main_Header",false },
  {0x0021,0x0001,"SIEMENS MED ECAT FILE INFO",VR::OB,VM::VM1,"ECAT_Image_Subheader",false },
  {0x0029,0x0010,"SIEMENS MED HG",VR::US,VM::VM1,"List of Group Numbers",false },
  {0x0029,0x0015,"SIEMENS MED HG",VR::LO,VM::VM1,"List of Shadow Owner Codes",false },
  {0x0029,0x0020,"SIEMENS MED HG",VR::US,VM::VM1,"List of Element Numbers",false },
  {0x0029,0x0030,"SIEMENS MED HG",VR::US,VM::VM1,"List of Total Display Length",false },
  {0x0029,0x0040,"SIEMENS MED HG",VR::LO,VM::VM1_n,"List of Display Prefix",false },
  {0x0029,0x0050,"SIEMENS MED HG",VR::LO,VM::VM1_n,"List of Display Postfix",false },
  {0x0029,0x0060,"SIEMENS MED HG",VR::US,VM::VM1,"List of Text Position",false },
  {0x0029,0x0070,"SIEMENS MED HG",VR::LO,VM::VM1,"List of Text Concatenation",false },
  {0x0029,0x005a,"SIEMENS MED MAMMO",VR::CS,VM::VM1,"?",false },
  {0x0029,0x0010,"SIEMENS MED MG",VR::US,VM::VM1,"List of Group Numbers",false },
  {0x0029,0x0015,"SIEMENS MED MG",VR::LO,VM::VM1,"List of Shadow Owner Codes",false },
  {0x0029,0x0020,"SIEMENS MED MG",VR::US,VM::VM1,"List of Element Numbers",false },
  {0x0029,0x0030,"SIEMENS MED MG",VR::US,VM::VM1,"List of Total Display Length",false },
  {0x0029,0x0040,"SIEMENS MED MG",VR::LO,VM::VM1_n,"List of Display Prefix",false },
  {0x0029,0x0050,"SIEMENS MED MG",VR::LO,VM::VM1_n,"List of Display Postfix",false },
  {0x0029,0x0060,"SIEMENS MED MG",VR::US,VM::VM1,"List of Text Position",false },
  {0x0029,0x0070,"SIEMENS MED MG",VR::LO,VM::VM1,"List of Text Concatenation",false },
  {0x0009,0x0080,"SIEMENS MED NM",VR::LO,VM::VM1,"Main Acquisition View",false },
  {0x0011,0x0010,"SIEMENS MED NM",VR::CS,VM::VM1,"Organ System",false },
  {0x0017,0x0000,"SIEMENS MED NM",VR::ST,VM::VM1,"Unknown 0017 0000",false },
  {0x0017,0x0020,"SIEMENS MED NM",VR::CS,VM::VM1,"Scanning Sequence",false },
  {0x0017,0x0060,"SIEMENS MED NM",VR::DS,VM::VM1,"KVP",false },
  {0x0017,0x0070,"SIEMENS MED NM",VR::ST,VM::VM1,"?",false },
  {0x0017,0x0080,"SIEMENS MED NM",VR::ST,VM::VM1_n,"Unknown 0017 0080",false },
  {0x0019,0x0001,"SIEMENS MED NM",VR::SS,VM::VM1,"Number of detectors",false },
  {0x0019,0x0002,"SIEMENS MED NM",VR::SS,VM::VM1,"Number of scalers",false },
  {0x0019,0x0003,"SIEMENS MED NM",VR::SS,VM::VM1,"Scaler Index",false },
  {0x0019,0x0004,"SIEMENS MED NM",VR::DS,VM::VM1,"Radionuclide Units",false },
  {0x0019,0x0005,"SIEMENS MED NM",VR::DS,VM::VM1,"Radiopharmaceutical",false },
  {0x0019,0x0006,"SIEMENS MED NM",VR::FL,VM::VM1,"Display Magnification",false },
  {0x0019,0x0007,"SIEMENS MED NM",VR::SL,VM::VM1,"Scaler Counts",false },
  {0x0019,0x0008,"SIEMENS MED NM",VR::SS,VM::VM1,"Processed Flag 0 Raw 1 Process",false },
  {0x0019,0x0009,"SIEMENS MED NM",VR::SS,VM::VM1,"MicroDELTA Parameter Blocks",false },
  {0x0019,0x000a,"SIEMENS MED NM",VR::LO,VM::VM1,"MicroDELTA File Extension",false },
  {0x0019,0x000b,"SIEMENS MED NM",VR::LO,VM::VM1,"MicroDELTA Write ID",false },
  {0x0019,0x000c,"SIEMENS MED NM",VR::SS,VM::VM1,"MicroDELTA Write Version",false },
  {0x0019,0x000d,"SIEMENS MED NM",VR::CS,VM::VM1,"MicroDELTA Mode",false },
  {0x0019,0x000e,"SIEMENS MED NM",VR::LO,VM::VM1,"MicroDELTA ESFR",false },
  {0x0019,0x000f,"SIEMENS MED NM",VR::SL,VM::VM1,"Siemens ICON Data Type",false },
  {0x0019,0x0010,"SIEMENS MED NM",VR::OW,VM::VM1,"MicroDELTA Expansion Blocks",false },
  {0x0019,0x0011,"SIEMENS MED NM",VR::CS,VM::VM1,"MicroDELTA Version",false },
  {0x0019,0x0012,"SIEMENS MED NM",VR::LO,VM::VM1,"MicroDELTA Trend",false },
  {0x0019,0x0013,"SIEMENS MED NM",VR::SS,VM::VM1,"MicroDELTA Number of Overlays",false },
  {0x0019,0x0014,"SIEMENS MED NM",VR::SS,VM::VM1,"MicroDELTA Overlay Block",false },
  {0x0019,0x0015,"SIEMENS MED NM",VR::SS,VM::VM1,"MicroDELTA Color Table Extra",false },
  {0x0019,0x0016,"SIEMENS MED NM",VR::SS,VM::VM1,"Number of Views",false },
  {0x0019,0x0017,"SIEMENS MED NM",VR::SS,VM::VM1,"MicroDELTA Thyroid Markers",false },
  {0x0019,0x0018,"SIEMENS MED NM",VR::SS,VM::VM1,"MicroDELTA Color Table Used",false },
  {0x0019,0x0019,"SIEMENS MED NM",VR::SS,VM::VM1,"Save Screen Flag 1 SaveScreen",false },
  {0x0019,0x001a,"SIEMENS MED NM",VR::SS,VM::VM1,"MicroDELTA Overlay Map",false },
  {0x0019,0x001b,"SIEMENS MED NM",VR::SS,VM::VM1,"SGI Converted LookUp Table",false },
  {0x0019,0x001c,"SIEMENS MED NM",VR::SS,VM::VM1,"MicroDELTA Mystery1",false },
  {0x0019,0x001d,"SIEMENS MED NM",VR::SS,VM::VM1,"MicroDELTA Mystery2",false },
  {0x0019,0x0020,"SIEMENS MED NM",VR::FL,VM::VM1,"Time of Lasix Administration",false },
  {0x0019,0x0021,"SIEMENS MED NM",VR::FL,VM::VM1,"Lasix Dose mg",false },
  {0x0019,0x0022,"SIEMENS MED NM",VR::FL,VM::VM1,"Bladder Esidual Volume cc",false },
  {0x0019,0x0023,"SIEMENS MED NM",VR::SS,VM::VM1,"Bladder Image Extension",false },
  {0x0019,0x0024,"SIEMENS MED NM",VR::FL,VM::VM1,"Serum Creatinine mg/dl",false },
  {0x0019,0x0025,"SIEMENS MED NM",VR::FL,VM::VM1,"TER",false },
  {0x0019,0x0026,"SIEMENS MED NM",VR::FL,VM::VM1,"TER Lower Limit",false },
  {0x0019,0x0027,"SIEMENS MED NM",VR::FL,VM::VM1,"Body Surface Area m2",false },
  {0x0019,0x0030,"SIEMENS MED NM",VR::FL,VM::VM1,"CTQ Normal Margin",false },
  {0x0019,0x0031,"SIEMENS MED NM",VR::FL,VM::VM1,"CTQ Abnormal Sagittal Margin",false },
  {0x0019,0x0032,"SIEMENS MED NM",VR::FL,VM::VM1,"CTQ Abnormal Oblique Margin",false },
  {0x0019,0x0041,"SIEMENS MED NM",VR::SS,VM::VM1,"Accepted Beats",false },
  {0x0019,0x0042,"SIEMENS MED NM",VR::SS,VM::VM1,"Rejected Beats",false },
  {0x0019,0x0043,"SIEMENS MED NM",VR::SS,VM::VM1,"Percent of R-R Interval Framed",false },
  {0x0019,0x0044,"SIEMENS MED NM",VR::SS,VM::VM1,"Beats Rejected Post PVC",false },
  {0x0019,0x0045,"SIEMENS MED NM",VR::FL,VM::VM1,"Trigger Source",false },
  {0x0019,0x0047,"SIEMENS MED NM",VR::FL,VM::VM1,"Beat Rejection Parameter",false },
  {0x0019,0x0048,"SIEMENS MED NM",VR::SH,VM::VM1,"Rejection Scheme",false },
  {0x0019,0x0049,"SIEMENS MED NM",VR::FL,VM::VM1,"PVC Threshold msec",false },
  {0x0019,0x004a,"SIEMENS MED NM",VR::SS,VM::VM1,"Framing Mode 1 Forward 2 F/B",false },
  {0x0019,0x004b,"SIEMENS MED NM",VR::SS,VM::VM1,"Autobeat Window 0 No 1 Yes",false },
  {0x0019,0x004c,"SIEMENS MED NM",VR::FL,VM::VM1,"MicroDELTA Acquisition Time per Window msec",false },
  {0x0019,0x004e,"SIEMENS MED NM",VR::SS,VM::VM1,"CTQ Unknown Entry Count",false },
  {0x0019,0x004f,"SIEMENS MED NM",VR::SS,VM::VM1,"CTQ Matrix Size",false },
  {0x0019,0x0050,"SIEMENS MED NM",VR::SS,VM::VM1,"CTQ has been used",false },
  {0x0019,0x0051,"SIEMENS MED NM",VR::SS,VM::VM1,"CTQ Male Entry Count",false },
  {0x0019,0x0052,"SIEMENS MED NM",VR::SS,VM::VM1,"CTQ Female Entry Count",false },
  {0x0019,0x0053,"SIEMENS MED NM",VR::SS,VM::VM1,"CTQ Stress/Rest Time Delay",false },
  {0x0019,0x0054,"SIEMENS MED NM",VR::CS,VM::VM1,"CTQ Time Delay Units",false },
  {0x0019,0x0055,"SIEMENS MED NM",VR::SS,VM::VM1,"CTQ Min Oblique Stack Count",false },
  {0x0019,0x0056,"SIEMENS MED NM",VR::SS,VM::VM1,"CTQ Max Oblique Stack Count",false },
  {0x0019,0x0057,"SIEMENS MED NM",VR::SS,VM::VM1,"CTQ Min Sagittal Stack Count",false },
  {0x0019,0x0058,"SIEMENS MED NM",VR::SS,VM::VM1,"CTQ Max Sagittal Stack Count",false },
  {0x0019,0x0059,"SIEMENS MED NM",VR::SS,VM::VM1,"CTQ Min Radius Percent",false },
  {0x0019,0x005a,"SIEMENS MED NM",VR::UL,VM::VM1,"CTQ Find Max or Average Flag",false },
  {0x0019,0x005b,"SIEMENS MED NM",VR::SS,VM::VM1,"CTQ Basal Slice Index Offset",false },
  {0x0019,0x005c,"SIEMENS MED NM",VR::SS,VM::VM1,"CTQ Starting Line Index",false },
  {0x0019,0x005d,"SIEMENS MED NM",VR::SS,VM::VM1,"CTQ Line Counts",false },
  {0x0019,0x005e,"SIEMENS MED NM",VR::SS,VM::VM1,"CTQ Radii Counts",false },
  {0x0019,0x005f,"SIEMENS MED NM",VR::SS,VM::VM1,"CTQ Radial Angle Increment",false },
  {0x0019,0x0060,"SIEMENS MED NM",VR::SS,VM::VM1,"Whole Body Processed Flag 0:Raw 1:Img1-Img2 2:Im",false },
  {0x0019,0x0061,"SIEMENS MED NM",VR::SS,VM::VM1,"Whole Body View",false },
  {0x0019,0x0062,"SIEMENS MED NM",VR::SS,VM::VM1,"WB Display Field Length in cm",false },
  {0x0019,0x0063,"SIEMENS MED NM",VR::SS,VM::VM1,"WB Scan Length in cm",false },
  {0x0019,0x0064,"SIEMENS MED NM",VR::SS,VM::VM1,"WB Scan Speed cm/min",false },
  {0x0019,0x0065,"SIEMENS MED NM",VR::SS,VM::VM1,"WB Scan Speed Mode 0 Manual 1 Auto",false },
  {0x0019,0x0066,"SIEMENS MED NM",VR::SS,VM::VM1,"WB Scan Start 0 Head 1 Feet",false },
  {0x0019,0x0067,"SIEMENS MED NM",VR::SS,VM::VM1,"WB Scan Direction 0 Lt to Rt 1 Rt to Lt",false },
  {0x0019,0x0068,"SIEMENS MED NM",VR::FL,VM::VM1,"WB Aperature Width in cm",false },
  {0x0019,0x0069,"SIEMENS MED NM",VR::SS,VM::VM1,"WB Number of Passes",false },
  {0x0019,0x006a,"SIEMENS MED NM",VR::SS,VM::VM1,"WB Scan Time msec",false },
  {0x0019,0x006b,"SIEMENS MED NM",VR::FL,VM::VM1,"WB Anterior kCounts View 1",false },
  {0x0019,0x006c,"SIEMENS MED NM",VR::FL,VM::VM1,"WB Anterior kCounts View 2",false },
  {0x0019,0x006d,"SIEMENS MED NM",VR::FL,VM::VM1,"WB Count Density",false },
  {0x0019,0x006e,"SIEMENS MED NM",VR::SS,VM::VM1,"WB Reference View Flag 1 Ref",false },
  {0x0019,0x006f,"SIEMENS MED NM",VR::FL,VM::VM1,"WB Coun Density Speed cm/min",false },
  {0x0019,0x0070,"SIEMENS MED NM",VR::CS,VM::VM1,"WB Scan Type",false },
  {0x0019,0x0071,"SIEMENS MED NM",VR::LO,VM::VM1,"WB IRIS Corr. File Name",false },
  {0x0019,0x0072,"SIEMENS MED NM",VR::FL,VM::VM1,"WB Posterior kCounts View 1",false },
  {0x0019,0x0073,"SIEMENS MED NM",VR::FL,VM::VM1,"WB Posterior kCounts View 2",false },
  {0x0019,0x0074,"SIEMENS MED NM",VR::SS,VM::VM1,"WB Ant View Status No Img 1 Complete 2 Aborted",false },
  {0x0019,0x0075,"SIEMENS MED NM",VR::SS,VM::VM1,"WB Post View Status No Img 1 Complete 2 Aborted",false },
  {0x0019,0x0076,"SIEMENS MED NM",VR::SS,VM::VM1,"WB Post Scan Time msec",false },
  {0x0019,0x0077,"SIEMENS MED NM",VR::SS,VM::VM1,"WB Ant Pixel Overflow 1 Yes",false },
  {0x0019,0x0078,"SIEMENS MED NM",VR::SS,VM::VM1,"WB Post Pixel Overflow 1 Yes",false },
  {0x0019,0x007f,"SIEMENS MED NM",VR::SS,VM::VM1,"Segment Type 0 Segments 2 Time/Segment",false },
  {0x0019,0x0080,"SIEMENS MED NM",VR::SS,VM::VM1,"Number of Segments",false },
  {0x0019,0x0081,"SIEMENS MED NM",VR::SS,VM::VM1,"PVC Rejection Flag 1 Reject",false },
  {0x0019,0x0082,"SIEMENS MED NM",VR::SL,VM::VM1,"Total Counts Framed",false },
  {0x0019,0x0083,"SIEMENS MED NM",VR::SS,VM::VM1,"Total Beats Framed",false },
  {0x0019,0x0084,"SIEMENS MED NM",VR::SS,VM::VM1,"Present Time of Day",false },
  {0x0019,0x0085,"SIEMENS MED NM",VR::SS,VM::VM1,"Framed Counts",false },
  {0x0019,0x0086,"SIEMENS MED NM",VR::SS,VM::VM1,"Time per Segment msec",false },
  {0x0019,0x0087,"SIEMENS MED NM",VR::LT,VM::VM1,"BBA Miscellaneous Info",false },
  {0x0019,0x0088,"SIEMENS MED NM",VR::SL,VM::VM1,"Tolerance %",false },
  {0x0019,0x0089,"SIEMENS MED NM",VR::SS,VM::VM1,"SWG",false },
  {0x0019,0x008a,"SIEMENS MED NM",VR::SL,VM::VM1,"R-R Interval msec",false },
  {0x0019,0x008d,"SIEMENS MED NM",VR::SL,VM::VM1,"Total Counts in Acquisition",false },
  {0x0019,0x008e,"SIEMENS MED NM",VR::SL,VM::VM1,"Rejection Type",false },
  {0x0019,0x0091,"SIEMENS MED NM",VR::FL,VM::VM1,"Phase Total Counts",false },
  {0x0019,0x0092,"SIEMENS MED NM",VR::SL,VM::VM1,"Phase Pause msec",false },
  {0x0019,0x0093,"SIEMENS MED NM",VR::SL,VM::VM1,"Phase Start msec",false },
  {0x0019,0x0094,"SIEMENS MED NM",VR::SS,VM::VM1,"Window Center Method 0 Avg 1 Peak",false },
  {0x0019,0x0095,"SIEMENS MED NM",VR::SS,VM::VM1,"Number of Beats to Average",false },
  {0x0019,0x0096,"SIEMENS MED NM",VR::FL,VM::VM1,"Histogram Bin Interval msec",false },
  {0x0019,0x0097,"SIEMENS MED NM",VR::SS,VM::VM1,"Counts in Peak Bin",false },
  {0x0019,0x00a0,"SIEMENS MED NM",VR::SS,VM::VM1,"VPDs Detected",false },
  {0x0019,0x00a1,"SIEMENS MED NM",VR::SS,VM::VM1,"Number of Phases",false },
  {0x0019,0x00a2,"SIEMENS MED NM",VR::SL,VM::VM1,"Requested Frame Duration msec",false },
  {0x0019,0x00a3,"SIEMENS MED NM",VR::SL,VM::VM1,"Frames per Phase",false },
  {0x0019,0x00a4,"SIEMENS MED NM",VR::SL,VM::VM1,"Frame Pause msec",false },
  {0x0019,0x00a5,"SIEMENS MED NM",VR::SS,VM::VM1_n,"Number of repeats per phase",false },
  {0x0019,0x00a6,"SIEMENS MED NM",VR::SS,VM::VM1_n,"Cycles per repeat",false },
  {0x0019,0x00a7,"SIEMENS MED NM",VR::SL,VM::VM1_n,"Repeat start time",false },
  {0x0019,0x00a8,"SIEMENS MED NM",VR::SL,VM::VM1_n,"Repeat stop time",false },
  {0x0019,0x00a9,"SIEMENS MED NM",VR::SL,VM::VM1_n,"Effective repeat time",false },
  {0x0019,0x00aa,"SIEMENS MED NM",VR::SS,VM::VM1_n,"Acquired cycles per repeat",false },
  {0x0019,0x00c0,"SIEMENS MED NM",VR::FL,VM::VM1,"Head Tilt degrees",false },
  {0x0019,0x00c1,"SIEMENS MED NM",VR::SS,VM::VM1,"Uniformity Correction Flag",false },
  {0x0019,0x00c2,"SIEMENS MED NM",VR::SS,VM::VM1,"Overflow Flag",false },
  {0x0019,0x00c3,"SIEMENS MED NM",VR::ST,VM::VM1,"?",false },
  {0x0019,0x00f0,"SIEMENS MED NM",VR::FL,VM::VM1,"Pixel Dimensions",false },
  {0x001f,0x0050,"SIEMENS MED NM",VR::DS,VM::VM1,"Location RET",false },
  {0x0021,0x0000,"SIEMENS MED NM",VR::OB,VM::VM1,"ECAT File Menu Header",false },
  {0x0021,0x0001,"SIEMENS MED NM",VR::OB,VM::VM1,"ECAT File Subheader",false },
  {0x0021,0x0010,"SIEMENS MED NM",VR::ST,VM::VM1,"?",false },
  {0x0031,0x0001,"SIEMENS MED NM",VR::LO,VM::VM1,"Camera Name",false },
  {0x0031,0x0002,"SIEMENS MED NM",VR::SS,VM::VM1,"Unknown Element 0031 02",false },
  {0x0031,0x0003,"SIEMENS MED NM",VR::FL,VM::VM1,"External Zoom",false },
  {0x0031,0x0004,"SIEMENS MED NM",VR::FL,VM::VM1,"CCD Energy Level",false },
  {0x0031,0x0005,"SIEMENS MED NM",VR::SS,VM::VM1,"CCD Window Width",false },
  {0x0031,0x0006,"SIEMENS MED NM",VR::FL,VM::VM1,"CCD Window Offset",false },
  {0x0031,0x0007,"SIEMENS MED NM",VR::SS,VM::VM1,"CCD Peaking Mode",false },
  {0x0031,0x0008,"SIEMENS MED NM",VR::LO,VM::VM1,"CCD Autopeak Function",false },
  {0x0031,0x0009,"SIEMENS MED NM",VR::SS,VM::VM1,"CCD Count Rate Mode",false },
  {0x0031,0x000a,"SIEMENS MED NM",VR::SS,VM::VM1,"Analog Output Device",false },
  {0x0031,0x000b,"SIEMENS MED NM",VR::SS,VM::VM1,"Communication Type",false },
  {0x0031,0x000c,"SIEMENS MED NM",VR::SS,VM::VM1,"Control Unit 1 ICON 2 DOT 3 ScintiView",false },
  {0x0031,0x000d,"SIEMENS MED NM",VR::SS,VM::VM1,"WAM Mode",false },
  {0x0031,0x000e,"SIEMENS MED NM",VR::SS,VM::VM1,"Event Control Mode",false },
  {0x0031,0x000f,"SIEMENS MED NM",VR::FL,VM::VM1,"Zoom X Offset 0.1 mm",false },
  {0x0031,0x0010,"SIEMENS MED NM",VR::FL,VM::VM1,"Zoom Y Offset 0.1 mm",false },
  {0x0031,0x0011,"SIEMENS MED NM",VR::SS,VM::VM1,"Gantry Zero Position",false },
  {0x0031,0x0012,"SIEMENS MED NM",VR::SS,VM::VM1,"?",false },
  {0x0031,0x0013,"SIEMENS MED NM",VR::ST,VM::VM1,"?",false },
  {0x0031,0x0014,"SIEMENS MED NM",VR::ST,VM::VM1,"?",false },
  {0x0031,0x0015,"SIEMENS MED NM",VR::SL,VM::VM1_n,"?",false },
  {0x0031,0x0016,"SIEMENS MED NM",VR::SL,VM::VM1_n,"?",false },
  {0x0031,0x0017,"SIEMENS MED NM",VR::SS,VM::VM1_n,"?",false },
  {0x0031,0x0020,"SIEMENS MED NM",VR::ST,VM::VM1,"?",false },
  {0x0031,0x0021,"SIEMENS MED NM",VR::SS,VM::VM1,"?",false },
  {0x0033,0x0000,"SIEMENS MED NM",VR::FL,VM::VM1_n,"Flood correction Matrix Det 1 up to SR 2.0",false },
  {0x0033,0x0001,"SIEMENS MED NM",VR::FL,VM::VM1_n,"Flood correction Matrix Det 2 up to SR 2.0",false },
  {0x0033,0x0010,"SIEMENS MED NM",VR::FL,VM::VM1_n,"COR Data for Detector 1",false },
  {0x0033,0x0011,"SIEMENS MED NM",VR::FL,VM::VM1_n,"COR Data for Detector 2",false },
  {0x0033,0x0014,"SIEMENS MED NM",VR::FL,VM::VM1,"MHR ( Y-Shift) data for detector 1",false },
  {0x0033,0x0015,"SIEMENS MED NM",VR::FL,VM::VM1,"MHR ( Y-Shift) data for detector 2",false },
  {0x0033,0x0018,"SIEMENS MED NM",VR::FL,VM::VM1_n,"NCO Data for detector 1",false },
  {0x0033,0x0019,"SIEMENS MED NM",VR::FL,VM::VM1_n,"NCO Data for detector 2",false },
  {0x0033,0x0020,"SIEMENS MED NM",VR::FL,VM::VM1,"Bed correction angle",false },
  {0x0033,0x0021,"SIEMENS MED NM",VR::FL,VM::VM1,"Gantry correction angle",false },
  {0x0033,0x0022,"SIEMENS MED NM",VR::SS,VM::VM1_n,"Bed U/D correction data",false },
  {0x0033,0x0023,"SIEMENS MED NM",VR::SS,VM::VM1_n,"?",false },
  {0x0033,0x0024,"SIEMENS MED NM",VR::FL,VM::VM1,"Backprojection angle head 1",false },
  {0x0033,0x0025,"SIEMENS MED NM",VR::FL,VM::VM1,"Backprojection angle head 2",false },
  {0x0033,0x0028,"SIEMENS MED NM",VR::SL,VM::VM1,"Number of point sources used for NCO and MHR",false },
  {0x0033,0x0029,"SIEMENS MED NM",VR::FL,VM::VM2,"Crystal thickness",false },
  {0x0033,0x0030,"SIEMENS MED NM",VR::LO,VM::VM1,"Preset name used for acquisition",false },
  {0x0033,0x0031,"SIEMENS MED NM",VR::FL,VM::VM1,"Camera config angle",false },
  {0x0033,0x0032,"SIEMENS MED NM",VR::LO,VM::VM1,"Crystal type Startburst or not",false },
  {0x0033,0x0033,"SIEMENS MED NM",VR::SL,VM::VM1,"Gantry step for COIN acquisitions",false },
  {0x0033,0x0034,"SIEMENS MED NM",VR::FL,VM::VM1,"Bed step for wholebody or Coin acquisitions",false },
  {0x0033,0x0035,"SIEMENS MED NM",VR::FL,VM::VM1,"Weight factor table for coincidence acquisitions",false },
  {0x0033,0x0036,"SIEMENS MED NM",VR::FL,VM::VM1,"Transaxial acceptance width for coincidence",false },
  {0x0033,0x0037,"SIEMENS MED NM",VR::SL,VM::VM1,"Starburst flags",false },
  {0x0033,0x0038,"SIEMENS MED NM",VR::FL,VM::VM1,"Pixel scale factor",false },
  {0x0035,0x0000,"SIEMENS MED NM",VR::LO,VM::VM1,"Specialized tomo type",false },
  {0x0035,0x0001,"SIEMENS MED NM",VR::LO,VM::VM1,"Energy window type",false },
  {0x0035,0x0002,"SIEMENS MED NM",VR::SS,VM::VM1,"Start and end row illuminated by wind position",false },
  {0x0035,0x0003,"SIEMENS MED NM",VR::LO,VM::VM1,"Blank scan image for profile",false },
  {0x0035,0x0004,"SIEMENS MED NM",VR::SS,VM::VM1,"Repeat number of the original dynamic SPECT",false },
  {0x0035,0x0005,"SIEMENS MED NM",VR::SS,VM::VM1,"Phase number of the original dynamic SPECT",false },
  {0x0037,0x0000,"SIEMENS MED NM",VR::OW,VM::VM1,"Flood correction matrix Detector 1",false },
  {0x0037,0x0080,"SIEMENS MED NM",VR::OW,VM::VM1,"Flood correction matrix Detector 2",false },
  {0x0039,0x0000,"SIEMENS MED NM",VR::LT,VM::VM1,"Toshiba CBF activity results",false },
  {0x0041,0x0001,"SIEMENS MED NM",VR::SL,VM::VM1,"?",false },
  {0x0041,0x0002,"SIEMENS MED NM",VR::SL,VM::VM1,"?",false },
  {0x0041,0x0003,"SIEMENS MED NM",VR::FL,VM::VM1,"?",false },
  {0x0041,0x0005,"SIEMENS MED NM",VR::FD,VM::VM1_n,"?",false },
  {0x0041,0x0006,"SIEMENS MED NM",VR::FD,VM::VM1_n,"?",false },
  {0x0041,0x0007,"SIEMENS MED NM",VR::FD,VM::VM1_n,"?",false },
  {0x0041,0x0008,"SIEMENS MED NM",VR::FD,VM::VM1_n,"?",false },
  {0x0041,0x0011,"SIEMENS MED NM",VR::FL,VM::VM1_n,"?",false },
  {0x0041,0x0021,"SIEMENS MED NM",VR::LO,VM::VM1,"Text Relating to Data Object",false },
  {0x0041,0x0030,"SIEMENS MED NM",VR::DA,VM::VM1,"Folder Creation Date",false },
  {0x0041,0x0032,"SIEMENS MED NM",VR::TM,VM::VM1,"Folder Creation Time",false },
  {0x0041,0x0034,"SIEMENS MED NM",VR::DA,VM::VM1,"Folder Modification Date",false },
  {0x0041,0x0036,"SIEMENS MED NM",VR::TM,VM::VM1,"Folder Modification Time",false },
  {0x0041,0x0040,"SIEMENS MED NM",VR::LO,VM::VM1,"Folder Owner Name",false },
  {0x0041,0x0050,"SIEMENS MED NM",VR::SS,VM::VM1,"Folder Status",false },
  {0x0041,0x0070,"SIEMENS MED NM",VR::SL,VM::VM1,"Number of Images in Folder",false },
  {0x0041,0x0071,"SIEMENS MED NM",VR::SL,VM::VM1,"Number of Non-Image Objects",false },
  {0x0043,0x0001,"SIEMENS MED NM",VR::FL,VM::VM1_n,"?",false },
  {0x0043,0x0003,"SIEMENS MED NM",VR::FL,VM::VM1_n,"?",false },
  {0x0043,0x0004,"SIEMENS MED NM",VR::FL,VM::VM1_n,"?",false },
  {0x0051,0x0010,"SIEMENS MED NM",VR::SS,VM::VM1,"Number of Windows",false },
  {0x0051,0x0011,"SIEMENS MED NM",VR::SS,VM::VM1,"Average Heart Rate BPM",false },
  {0x0051,0x0012,"SIEMENS MED NM",VR::SS,VM::VM1,"Accepted Beats in Window",false },
  {0x0051,0x0013,"SIEMENS MED NM",VR::SS,VM::VM1,"Average R-R Interval msec",false },
  {0x0051,0x0014,"SIEMENS MED NM",VR::SS,VM::VM1,"Window Width %",false },
  {0x0051,0x0015,"SIEMENS MED NM",VR::SL,VM::VM1,"Hottest Pixel in Window",false },
  {0x0051,0x0018,"SIEMENS MED NM",VR::SL,VM::VM1,"Total Counts in Window",false },
  {0x0053,0x0001,"SIEMENS MED NM",VR::SS,VM::VM1,"Terminate on Pixel Overflow Flag 1 Terminate",false },
  {0x0053,0x0002,"SIEMENS MED NM",VR::SS,VM::VM1,"Terminate on List Buffer Overflow Flag",false },
  {0x0053,0x0003,"SIEMENS MED NM",VR::SS,VM::VM1,"First Frame by Counts",false },
  {0x0053,0x0010,"SIEMENS MED NM",VR::SS,VM::VM1,"Elapsed Time Start Condition 0 None 1 Any 2 All",false },
  {0x0053,0x0011,"SIEMENS MED NM",VR::SL,VM::VM1,"Elapsed Time Start msec",false },
  {0x0053,0x0012,"SIEMENS MED NM",VR::SS,VM::VM1,"Count Rate Start Condition 0 None 1 Any 2 All",false },
  {0x0053,0x0013,"SIEMENS MED NM",VR::SL,VM::VM1,"Count Rate Start",false },
  {0x0053,0x0014,"SIEMENS MED NM",VR::SS,VM::VM1,"Start Heart Beat Count Condition 0 None 1 Any 2",false },
  {0x0053,0x0015,"SIEMENS MED NM",VR::SL,VM::VM1,"Start Heart Beat Count",false },
  {0x0053,0x0016,"SIEMENS MED NM",VR::SS,VM::VM1,"Start Heart Rate Threshold Condition",false },
  {0x0053,0x0017,"SIEMENS MED NM",VR::SS,VM::VM1,"Start Heart Rate Increasing Flag",false },
  {0x0053,0x0018,"SIEMENS MED NM",VR::SL,VM::VM1,"Start Heart Rate",false },
  {0x0053,0x0019,"SIEMENS MED NM",VR::SL,VM::VM1,"Manual Stop Flag 0 No 1 Yes",false },
  {0x0053,0x001a,"SIEMENS MED NM",VR::SS,VM::VM1,"Elapsed Time Stop Condition 0 None 1 Any 2 All",false },
  {0x0053,0x001b,"SIEMENS MED NM",VR::SL,VM::VM1,"Elapsed Time Stop",false },
  {0x0053,0x001c,"SIEMENS MED NM",VR::SS,VM::VM1,"Stop Heart Beat Count Condition 0 None 1 Any 2 A",false },
  {0x0053,0x001d,"SIEMENS MED NM",VR::SL,VM::VM1,"Stop Heart Beat Count",false },
  {0x0053,0x0030,"SIEMENS MED NM",VR::SS,VM::VM1,"Info Density Stop Condition",false },
  {0x0053,0x0031,"SIEMENS MED NM",VR::SS,VM::VM1,"Info Density Stop Scaler",false },
  {0x0053,0x0032,"SIEMENS MED NM",VR::SS,VM::VM1,"Info Density Stop Region",false },
  {0x0053,0x0033,"SIEMENS MED NM",VR::SL,VM::VM1,"Info Density Stop Count",false },
  {0x0053,0x0040,"SIEMENS MED NM",VR::SS,VM::VM1,"Stop Condition Status 0 None 1 Any 2 All",false },
  {0x0053,0x0041,"SIEMENS MED NM",VR::SL,VM::VM1,"Stop Count",false },
  {0x0053,0x0042,"SIEMENS MED NM",VR::SS,VM::VM1,"Stop Count Condition Status 0 None 1 Any 2 All",false },
  {0x0053,0x0043,"SIEMENS MED NM",VR::SL,VM::VM1,"Stop Time",false },
  {0x0053,0x0044,"SIEMENS MED NM",VR::SS,VM::VM1,"Stop Time Condition Status 0 None 1 Any 2 All",false },
  {0x0055,0x0002,"SIEMENS MED NM",VR::SS,VM::VM1,"Unknown 55-02",false },
  {0x0055,0x0004,"SIEMENS MED NM",VR::SS,VM::VM1,"Prompt window width",false },
  {0x0055,0x0005,"SIEMENS MED NM",VR::SS,VM::VM1,"Random window width",false },
  {0x0055,0x0020,"SIEMENS MED NM",VR::SS,VM::VM1,"Flood Corr. Flag 1 Corr",false },
  {0x0055,0x0021,"SIEMENS MED NM",VR::DS,VM::VM1,"Flood Acquisition Date",false },
  {0x0055,0x0030,"SIEMENS MED NM",VR::SS,VM::VM1,"COR Corr. Flag 1 Corr",false },
  {0x0055,0x0031,"SIEMENS MED NM",VR::DA,VM::VM1,"COR Acquisition Date",false },
  {0x0055,0x0038,"SIEMENS MED NM",VR::OW,VM::VM1,"COR Data",false },
  {0x0055,0x0040,"SIEMENS MED NM",VR::SS,VM::VM1,"Noncircular Orbit Flag 1 NCO",false },
  {0x0055,0x0048,"SIEMENS MED NM",VR::FL,VM::VM1,"NCO Values",false },
  {0x0055,0x004c,"SIEMENS MED NM",VR::SL,VM::VM1,"?",false },
  {0x0055,0x004d,"SIEMENS MED NM",VR::SL,VM::VM1,"?",false },
  {0x0055,0x0050,"SIEMENS MED NM",VR::SS,VM::VM1,"Prone Flag 0 Supine 1 Prone",false },
  {0x0055,0x0051,"SIEMENS MED NM",VR::SS,VM::VM1,"Orbit 0 Circular 1 Elliptical",false },
  {0x0055,0x0052,"SIEMENS MED NM",VR::SS,VM::VM1,"Total Rotation in Degrees",false },
  {0x0055,0x0053,"SIEMENS MED NM",VR::CS,VM::VM1,"Zero Location (T=Top B=Bottom)",false },
  {0x0055,0x0054,"SIEMENS MED NM",VR::FL,VM::VM1,"Angular Step",false },
  {0x0055,0x0055,"SIEMENS MED NM",VR::FL,VM::VM1,"Start Angle",false },
  {0x0055,0x0056,"SIEMENS MED NM",VR::SS,VM::VM1,"Decay Corr. Flag 1 Corr",false },
  {0x0055,0x0057,"SIEMENS MED NM",VR::SS,VM::VM1,"Scatter Corr. Flag 1 Corr",false },
  {0x0055,0x0058,"SIEMENS MED NM",VR::FL,VM::VM1,"Radial Map Info in 0.1 mm Inc.",false },
  {0x0055,0x0059,"SIEMENS MED NM",VR::CS,VM::VM1,"Direction H HeadIn F FeetIn",false },
  {0x0055,0x005a,"SIEMENS MED NM",VR::FL,VM::VM1,"Table Traverse cm",false },
  {0x0055,0x005b,"SIEMENS MED NM",VR::SS,VM::VM1,"Multiplexor Flag 1 Multiplexed",false },
  {0x0055,0x005c,"SIEMENS MED NM",VR::DS,VM::VM1,"Pixel Acquisition Date",false },
  {0x0055,0x005d,"SIEMENS MED NM",VR::SS,VM::VM1,"Angle Type 1 360o 2 128 Steps",false },
  {0x0055,0x005e,"SIEMENS MED NM",VR::SL,VM::VM1,"Orbit Definition",false },
  {0x0055,0x005f,"SIEMENS MED NM",VR::FL,VM::VM1,"Arm Length in cm",false },
  {0x0055,0x0060,"SIEMENS MED NM",VR::SS,VM::VM1,"Auto Window Beats",false },
  {0x0055,0x0061,"SIEMENS MED NM",VR::SS,VM::VM1,"Non-gated File Extension",false },
  {0x0055,0x0064,"SIEMENS MED NM",VR::SS,VM::VM1,"TIC Error Values",false },
  {0x0055,0x0065,"SIEMENS MED NM",VR::SS,VM::VM1,"X1",false },
  {0x0055,0x0066,"SIEMENS MED NM",VR::SS,VM::VM1,"Y1",false },
  {0x0055,0x0067,"SIEMENS MED NM",VR::SS,VM::VM1,"X2",false },
  {0x0055,0x0068,"SIEMENS MED NM",VR::SS,VM::VM1,"Y2",false },
  {0x0055,0x0069,"SIEMENS MED NM",VR::SS,VM::VM1,"X3",false },
  {0x0055,0x006a,"SIEMENS MED NM",VR::SS,VM::VM1,"Y3",false },
  {0x0055,0x006b,"SIEMENS MED NM",VR::SS,VM::VM1,"X4",false },
  {0x0055,0x006c,"SIEMENS MED NM",VR::SS,VM::VM1,"Y4",false },
  {0x0055,0x006d,"SIEMENS MED NM",VR::SS,VM::VM1,"Pixel Size Flag 1 Valid Size",false },
  {0x0055,0x0070,"SIEMENS MED NM",VR::FL,VM::VM1,"X Focal Length",false },
  {0x0055,0x0071,"SIEMENS MED NM",VR::FL,VM::VM1,"Y Focal Length",false },
  {0x0055,0x0072,"SIEMENS MED NM",VR::FL,VM::VM1,"X Gains",false },
  {0x0055,0x0073,"SIEMENS MED NM",VR::FL,VM::VM1,"Y Gains",false },
  {0x0055,0x0074,"SIEMENS MED NM",VR::FL,VM::VM1,"X Offset",false },
  {0x0055,0x0075,"SIEMENS MED NM",VR::FL,VM::VM1,"Y Offset",false },
  {0x0055,0x0076,"SIEMENS MED NM",VR::FL,VM::VM1,"X Principal Ray",false },
  {0x0055,0x0077,"SIEMENS MED NM",VR::FL,VM::VM1,"Y Principal Ray",false },
  {0x0055,0x0078,"SIEMENS MED NM",VR::FL,VM::VM1,"Total Flood Counts",false },
  {0x0055,0x007e,"SIEMENS MED NM",VR::FL,VM::VM1_n,"Collimator thickness",false },
  {0x0055,0x007f,"SIEMENS MED NM",VR::FL,VM::VM1_n,"Collimator angular resolution",false },
  {0x0055,0x0080,"SIEMENS MED NM",VR::SS,VM::VM1,"File Altered Flag 1 Altered",false },
  {0x0055,0x0081,"SIEMENS MED NM",VR::SS,VM::VM1,"Raw Data Extension",false },
  {0x0055,0x0082,"SIEMENS MED NM",VR::SS,VM::VM1,"Sagittal Extension",false },
  {0x0055,0x0083,"SIEMENS MED NM",VR::SS,VM::VM1,"Coronal Extension",false },
  {0x0055,0x0084,"SIEMENS MED NM",VR::SS,VM::VM1,"Transverse Extension",false },
  {0x0055,0x0085,"SIEMENS MED NM",VR::SS,VM::VM1,"Volume Rendered Extension",false },
  {0x0055,0x0086,"SIEMENS MED NM",VR::SL,VM::VM1,"Plane of Transverse Bottom Slice",false },
  {0x0055,0x0087,"SIEMENS MED NM",VR::SL,VM::VM1,"Plane of Transverse Top Slice",false },
  {0x0055,0x0088,"SIEMENS MED NM",VR::SL,VM::VM1,"Slice Step in Pixels",false },
  {0x0055,0x008a,"SIEMENS MED NM",VR::SL,VM::VM1,"Y-Axis Smooth Slice Weights",false },
  {0x0055,0x008b,"SIEMENS MED NM",VR::SL,VM::VM1,"Number of Transverse Slices",false },
  {0x0055,0x008c,"SIEMENS MED NM",VR::SS,VM::VM1,"Attenuation Flag 1 Chang 2 Sorenson",false },
  {0x0055,0x008d,"SIEMENS MED NM",VR::SS,VM::VM1,"Sagittal Right Limit",false },
  {0x0055,0x008e,"SIEMENS MED NM",VR::SS,VM::VM1,"Sagittal Left Limit",false },
  {0x0055,0x008f,"SIEMENS MED NM",VR::SS,VM::VM1,"Coronal Posterior Limit",false },
  {0x0055,0x0090,"SIEMENS MED NM",VR::SS,VM::VM1,"Coronal Anterior Limit",false },
  {0x0055,0x0091,"SIEMENS MED NM",VR::SS,VM::VM1,"Isotope Used in Reconstruction",false },
  {0x0055,0x0093,"SIEMENS MED NM",VR::FL,VM::VM1,"Filter Cutoff Frequency",false },
  {0x0055,0x0094,"SIEMENS MED NM",VR::FL,VM::VM1,"Filter Alpha Coefficient",false },
  {0x0055,0x0095,"SIEMENS MED NM",VR::FL,VM::VM1,"Filter Order",false },
  {0x0055,0x0096,"SIEMENS MED NM",VR::FL,VM::VM1,"Attenuation Coefficient",false },
  {0x0055,0x0097,"SIEMENS MED NM",VR::SS,VM::VM1,"Percent Ray",false },
  {0x0055,0x0098,"SIEMENS MED NM",VR::SS,VM::VM1,"Astigmatic Collimator 1 Neuro 2 Cardiac",false },
  {0x0055,0x0099,"SIEMENS MED NM",VR::SS,VM::VM1,"XP0",false },
  {0x0055,0x009a,"SIEMENS MED NM",VR::SS,VM::VM1,"ZP0",false },
  {0x0055,0x009b,"SIEMENS MED NM",VR::SS,VM::VM1,"XP1",false },
  {0x0055,0x009c,"SIEMENS MED NM",VR::SS,VM::VM1,"ZP1",false },
  {0x0055,0x009d,"SIEMENS MED NM",VR::FL,VM::VM1,"Rotation Matrix",false },
  {0x0055,0x009e,"SIEMENS MED NM",VR::FL,VM::VM1,"X0",false },
  {0x0055,0x009f,"SIEMENS MED NM",VR::FL,VM::VM1,"X1",false },
  {0x0055,0x00a0,"SIEMENS MED NM",VR::FL,VM::VM1,"Y0",false },
  {0x0055,0x00a1,"SIEMENS MED NM",VR::FL,VM::VM1,"Y1",false },
  {0x0055,0x00a2,"SIEMENS MED NM",VR::FL,VM::VM1,"Z0",false },
  {0x0055,0x00a3,"SIEMENS MED NM",VR::FL,VM::VM1,"Z1",false },
  {0x0055,0x00a4,"SIEMENS MED NM",VR::SS,VM::VM1,"Transverse Scale",false },
  {0x0055,0x00a5,"SIEMENS MED NM",VR::FL,VM::VM1,"Pixel Size X Direction mm",false },
  {0x0055,0x00a6,"SIEMENS MED NM",VR::FL,VM::VM1,"Pixel Size Y Direction mm",false },
  {0x0055,0x00a7,"SIEMENS MED NM",VR::SS,VM::VM1,"Regions File Extension",false },
  {0x0055,0x00a8,"SIEMENS MED NM",VR::SS,VM::VM1,"Gated SPECT Bins Reconstructed",false },
  {0x0055,0x00a9,"SIEMENS MED NM",VR::SS,VM::VM1,"Vertical Oblique Image",false },
  {0x0055,0x00b0,"SIEMENS MED NM",VR::SS,VM::VM1,"Oblique Region First Slice",false },
  {0x0055,0x00b1,"SIEMENS MED NM",VR::SS,VM::VM1,"Oblique Region Last Slice",false },
  {0x0055,0x00c0,"SIEMENS MED NM",VR::SS,VM::VM1_n,"?",false },
  {0x0055,0x00c2,"SIEMENS MED NM",VR::SS,VM::VM1,"?",false },
  {0x0055,0x00c3,"SIEMENS MED NM",VR::SS,VM::VM1,"?",false },
  {0x0055,0x00c4,"SIEMENS MED NM",VR::SS,VM::VM1,"?",false },
  {0x0055,0x00d0,"SIEMENS MED NM",VR::SS,VM::VM1,"?",false },
  {0x0057,0x0001,"SIEMENS MED NM",VR::LO,VM::VM1,"e.soft DICOM original image type",false },
  {0x0057,0x0002,"SIEMENS MED NM",VR::FL,VM::VM1,"Dose calibration factor",false },
  {0x0057,0x0003,"SIEMENS MED NM",VR::LO,VM::VM1,"Units",false },
  {0x0057,0x0004,"SIEMENS MED NM",VR::LO,VM::VM1,"Decay correction",false },
  {0x0057,0x0005,"SIEMENS MED NM",VR::SL,VM::VM1_n,"Radio nuclide half life",false },
  {0x0057,0x0006,"SIEMENS MED NM",VR::FL,VM::VM1,"Rescale intercept",false },
  {0x0057,0x0007,"SIEMENS MED NM",VR::FL,VM::VM1,"Rescale Slope",false },
  {0x0057,0x0008,"SIEMENS MED NM",VR::FL,VM::VM1,"Frame reference time",false },
  {0x0057,0x0009,"SIEMENS MED NM",VR::SL,VM::VM1,"Number of Radiopharmaceutical information seq",false },
  {0x0057,0x000a,"SIEMENS MED NM",VR::FL,VM::VM1,"Decay factor",false },
  {0x0057,0x000b,"SIEMENS MED NM",VR::LO,VM::VM1,"Counts source",false },
  {0x0057,0x000c,"SIEMENS MED NM",VR::SL,VM::VM1_n,"Radio Nuclide positron fraction",false },
  {0x0057,0x0010,"SIEMENS MED NM",VR::SS,VM::VM1,"Low Count",false },
  {0x0057,0x0020,"SIEMENS MED NM",VR::SS,VM::VM1,"Vertical Oblique Image",false },
  {0x5001,0x0002,"SIEMENS MED NM",VR::SL,VM::VM1,"Number of Points",false },
  {0x5001,0x0003,"SIEMENS MED NM",VR::FL,VM::VM1,"Minimal Value on X-Axis",false },
  {0x5001,0x0004,"SIEMENS MED NM",VR::FL,VM::VM1,"Maximal Value on X-Axis",false },
  {0x5001,0x0005,"SIEMENS MED NM",VR::FL,VM::VM1,"Minimal Value on Y-Axis",false },
  {0x5001,0x0006,"SIEMENS MED NM",VR::FL,VM::VM1,"Maximal Value on Y-Axis",false },
  {0x5001,0x0007,"SIEMENS MED NM",VR::LO,VM::VM1,"Curve Name",false },
  {0x5001,0x0008,"SIEMENS MED NM",VR::LO,VM::VM1,"Curve Function Label",false },
  {0x5001,0x0009,"SIEMENS MED NM",VR::SH,VM::VM1,"X Axis Label",false },
  {0x5001,0x000a,"SIEMENS MED NM",VR::SH,VM::VM1,"Y Axis Label",false },
  {0x5001,0x000b,"SIEMENS MED NM",VR::OW,VM::VM1,"Curve Data",false },
  {0x7fe1,0x0010,"SIEMENS MED NM",VR::OW,VM::VM1,"Compressed Pixel Data",false },
  {0x7fe3,0x0000,"SIEMENS MED NM",VR::UL,VM::VM1,"Length of Group 0x7FE3",false },
  {0x7fe3,0x0014,"SIEMENS MED NM",VR::SL,VM::VM1,"Lowest Pixel Value in THIS Frame",false },
  {0x7fe3,0x0015,"SIEMENS MED NM",VR::SL,VM::VM1,"Highest Pixel Value in THIS Frame",false },
  {0x7fe3,0x0016,"SIEMENS MED NM",VR::SL,VM::VM1,"Total Counts in THIS Frame",false },
  {0x7fe3,0x0017,"SIEMENS MED NM",VR::SL,VM::VM1,"Duration of THIS Frame msec",false },
  {0x7fe3,0x0018,"SIEMENS MED NM",VR::FL,VM::VM1,"Frame Angle",false },
  {0x7fe3,0x0019,"SIEMENS MED NM",VR::LO,VM::VM1,"Frame Name",false },
  {0x7fe3,0x001a,"SIEMENS MED NM",VR::FL,VM::VM1,"View Number",false },
  {0x7fe3,0x001b,"SIEMENS MED NM",VR::SS,VM::VM1,"Pixel Overflow Flag 1 Pixel Overflow",false },
  {0x7fe3,0x001c,"SIEMENS MED NM",VR::SS,VM::VM1,"Buffer Overflow Flag 1 Buffer Overflow",false },
  {0x7fe3,0x001e,"SIEMENS MED NM",VR::SL,VM::VM1,"Start Time of THIS Frame From Start of Study",false },
  {0x7fe3,0x001f,"SIEMENS MED NM",VR::SL,VM::VM1,"ROI Maximum Pixel Value",false },
  {0x7fe3,0x0020,"SIEMENS MED NM",VR::SL,VM::VM1,"ROI Average Pixel Value",false },
  {0x7fe3,0x0021,"SIEMENS MED NM",VR::SL,VM::VM1,"ROI Total Counts",false },
  {0x7fe3,0x0022,"SIEMENS MED NM",VR::SL,VM::VM1,"ROI Total Pixels",false },
  {0x7fe3,0x0023,"SIEMENS MED NM",VR::SL,VM::VM1,"ROI Center Coordinates",false },
  {0x7fe3,0x0024,"SIEMENS MED NM",VR::LO,VM::VM1,"ROI Name",false },
  {0x7fe3,0x0025,"SIEMENS MED NM",VR::SS,VM::VM1,"ROI Data",false },
  {0x7fe3,0x0026,"SIEMENS MED NM",VR::OW,VM::VM1,"?",false },
  {0x7fe3,0x0027,"SIEMENS MED NM",VR::OW,VM::VM1,"?",false },
  {0x7fe3,0x0028,"SIEMENS MED NM",VR::OW,VM::VM1,"?",false },
  {0x7fe3,0x0029,"SIEMENS MED NM",VR::OW,VM::VM1,"?",false },
  {0x0039,0x0081,"SIEMENS MED OCS ACKNOWLEDGE FLAG",VR::IS,VM::VM1,"Acknowledge Flag",false },
  {0x0039,0x0079,"SIEMENS MED OCS AE TITLE",VR::CS,VM::VM1,"AE Title",false },
  {0x0039,0x0081,"SIEMENS MED OCS CALIBRATION DATE",VR::CS,VM::VM1,"Calibration Date",false },
  {0x0039,0x0084,"SIEMENS MED OCS CBT ARC INCREMENT",VR::IS,VM::VM1_n,"Arc Increment",false },
  {0x0039,0x0083,"SIEMENS MED OCS CBT ENDING GANTRY ANGLE",VR::DS,VM::VM1_n,"Ending Projection Angle",false },
  {0x0039,0x0091,"SIEMENS MED OCS CBT MISSING PROJECTION FRAMES",VR::IS,VM::VM1_n,"Missing Projection Frames",false },
  {0x0039,0x0093,"SIEMENS MED OCS CBT PATIENT_ORIENTATION",VR::CS,VM::VM1,"Patient Orientation",false },
  {0x0039,0x0086,"SIEMENS MED OCS CBT PROJECTION GANTRY POSITIONS",VR::DS,VM::VM1_n,"Gantry Positions",false },
  {0x0039,0x0090,"SIEMENS MED OCS CBT RECON FILTER TYPE",VR::IS,VM::VM1,"Reconstruction Filter Type",false },
  {0x0039,0x0089,"SIEMENS MED OCS CBT RECON NUM OF SLICES",VR::IS,VM::VM1,"Number of Reconstruction Slices",false },
  {0x0039,0x0088,"SIEMENS MED OCS CBT RECON SLICE HEIGHT",VR::IS,VM::VM1,"Reconstruction Slice Height",false },
  {0x0039,0x0087,"SIEMENS MED OCS CBT RECON SLICE WIDTH",VR::IS,VM::VM1,"Reconstruction Slice Width",false },
  {0x0039,0x0085,"SIEMENS MED OCS CBT ROTATION DIRECTION",VR::CS,VM::VM1_n,"Gantry Rotation Direction",false },
  {0x0039,0x0082,"SIEMENS MED OCS CBT STARTING GANTRY ANGLE",VR::DS,VM::VM1_n,"Starting Projection Angle",false },
  {0x0039,0x0096,"SIEMENS MED OCS CBT TOTAL_MU",VR::DS,VM::VM1,"Total MU delivered during Cone Beam",false },
  {0x0039,0x0083,"SIEMENS MED OCS COMMON EDGE",VR::IS,VM::VM1,"Common Edge",false },
  {0x0039,0x0080,"SIEMENS MED OCS FIELD ID",VR::CS,VM::VM1,"Field ID",false },
  {0x0039,0x0079,"SIEMENS MED OCS FIELD NAME",VR::CS,VM::VM1,"Field Name",false },
  {0x0039,0x0095,"SIEMENS MED OCS FP Alignment Calibration Date / Time",VR::DT,VM::VM1,"Flat Panel Alignment last calibration Date and Time",false },
  {0x0039,0x0097,"SIEMENS MED OCS I0 MU SCALE FACTOR",VR::IS,VM::VM1,"I0-MU Scale Factor",false },
  {0x0039,0x0082,"SIEMENS MED OCS ICC MATRIX",VR::DS,VM::VM2,"Image Center in Image Coordinates",false },
  {0x0039,0x0076,"SIEMENS MED OCS NUMBER OF FRAMES",VR::IS,VM::VM1,"Number of Frames",false },
  {0x0039,0x0077,"SIEMENS MED OCS NUMBER OF SUB FRAMES",VR::IS,VM::VM1,"Number of Sub-Frames",false },
  {0x0039,0x0094,"SIEMENS MED OCS PI ALIGNMENT IMAGE",VR::IS,VM::VM1,"Alignment Offset Valid Flag",false },
  {0x0039,0x0092,"SIEMENS MED OCS PI VALID PIXEL MAP",VR::IS,VM::VM1,"Valid Pixel Map Flag",false },
  {0x0039,0x0076,"SIEMENS MED OCS REFERENCE IMAGE",VR::UI,VM::VM1,"Reference Image",false },
  {0x0039,0x0088,"SIEMENS MED OCS REFERENCE LOCKED",VR::IS,VM::VM1,"Locked Flag",false },
  {0x0039,0x0087,"SIEMENS MED OCS SC IMG PIX SPC",VR::DS,VM::VM2,"Non-RT Image Pixel Spacing",false },
  {0x0039,0x0086,"SIEMENS MED OCS SC IMG SAD",VR::DS,VM::VM1,"Non-RT Image SAD",false },
  {0x0039,0x0085,"SIEMENS MED OCS SC IMG SID",VR::DS,VM::VM1,"Non-RT Image SID",false },
  {0x0039,0x0084,"SIEMENS MED OCS SCALED IMG PIX SPC",VR::DS,VM::VM2,"Scaled Image Pixel Spacing",false },
  {0x0039,0x0078,"SIEMENS MED OCS SITE NAME",VR::CS,VM::VM1,"Site Name",false },
  {0x0039,0x0098,"SIEMENS MED OCS TABLE LATERAL POSITION",VR::DS,VM::VM1,"Lateral Table Position",false },
  {0x0039,0x0099,"SIEMENS MED OCS TABLE LONGITUDINAL POSITION",VR::DS,VM::VM1,"Longitudinal Table Position",false },
  {0x0039,0x0078,"SIEMENS MED OCS TABLE OFFSET",VR::DS,VM::VM3,"Offset",false },
  {0x0039,0x0075,"SIEMENS MED OCS TABLE VERTICAL POSITION",VR::DS,VM::VM1,"vertical Table Position",false },
  {0x0039,0x0080,"SIEMENS MED OCS TRANS MATRIX",VR::DS,VM::VM5,"Transformation Matrix",false },
  {0x0019,0x0000,"SIEMENS MED SMS USG ANTARES",VR::SH,VM::VM1,"Private Creator Version",false },
  {0x0019,0x0003,"SIEMENS MED SMS USG ANTARES",VR::FD,VM::VM1,"Frame Rate",false },
  {0x0019,0x000c,"SIEMENS MED SMS USG ANTARES",VR::US,VM::VM1,"Burned in Graphics",false },
  {0x0019,0x000d,"SIEMENS MED SMS USG ANTARES",VR::SH,VM::VM1,"SieClear Index",false },
  {0x0019,0x0020,"SIEMENS MED SMS USG ANTARES",VR::SH,VM::VM1,"B-Mode Submode",false },
  {0x0019,0x0021,"SIEMENS MED SMS USG ANTARES",VR::FD,VM::VM1,"B-Mode Dynamic Range",false },
  {0x0019,0x0022,"SIEMENS MED SMS USG ANTARES",VR::FD,VM::VM1,"B-Mode Overall Gain",false },
  {0x0019,0x0023,"SIEMENS MED SMS USG ANTARES",VR::US,VM::VM1,"B-Mode Res/Speed Index",false },
  {0x0019,0x0024,"SIEMENS MED SMS USG ANTARES",VR::US,VM::VM1,"B-Mode Edge Enhance Index",false },
  {0x0019,0x0025,"SIEMENS MED SMS USG ANTARES",VR::US,VM::VM1,"B-Mode Persistence Index",false },
  {0x0019,0x0026,"SIEMENS MED SMS USG ANTARES",VR::US,VM::VM1,"B-Mode Map Index",false },
  {0x0019,0x0027,"SIEMENS MED SMS USG ANTARES",VR::US,VM::VM1,"???",false },
  {0x0019,0x0028,"SIEMENS MED SMS USG ANTARES",VR::US,VM::VM1,"???",false },
  {0x0019,0x0029,"SIEMENS MED SMS USG ANTARES",VR::US,VM::VM1,"???",false },
  {0x0019,0x002a,"SIEMENS MED SMS USG ANTARES",VR::US,VM::VM1,"B-Mode Tint Type",false },
  {0x0019,0x002d,"SIEMENS MED SMS USG ANTARES",VR::US,VM::VM1,"B-Mode Tint Index",false },
  {0x0019,0x002e,"SIEMENS MED SMS USG ANTARES",VR::SH,VM::VM1,"ClarifyVE Index",false },
  {0x0019,0x0031,"SIEMENS MED SMS USG ANTARES",VR::US,VM::VM1,"?",false },
  {0x0019,0x003a,"SIEMENS MED SMS USG ANTARES",VR::UN,VM::VM1,"Image Flag",false },
  {0x0019,0x003b,"SIEMENS MED SMS USG ANTARES",VR::LT,VM::VM1,"Protocol Info CalcGroups",false },
  {0x0019,0x0040,"SIEMENS MED SMS USG ANTARES",VR::UN,VM::VM1,"Color Flow state",false },
  {0x0019,0x0041,"SIEMENS MED SMS USG ANTARES",VR::UN,VM::VM1,"Color Flow Wall Filter Index",false },
  {0x0019,0x0042,"SIEMENS MED SMS USG ANTARES",VR::UN,VM::VM1,"Color Flow Submode",false },
  {0x0019,0x0043,"SIEMENS MED SMS USG ANTARES",VR::UN,VM::VM1,"Color Flow Overall Gain",false },
  {0x0019,0x0044,"SIEMENS MED SMS USG ANTARES",VR::UN,VM::VM1,"Color Flow Res/Speed Index",false },
  {0x0019,0x0046,"SIEMENS MED SMS USG ANTARES",VR::UN,VM::VM1,"Color Flow Smooth Index",false },
  {0x0019,0x0047,"SIEMENS MED SMS USG ANTARES",VR::UN,VM::VM1,"Color Flow Persistence Index",false },
  {0x0019,0x0048,"SIEMENS MED SMS USG ANTARES",VR::UN,VM::VM1,"Color Flow Map Index",false },
  {0x0019,0x0049,"SIEMENS MED SMS USG ANTARES",VR::UN,VM::VM1,"Color Flow Priority Index",false },
  {0x0019,0x0054,"SIEMENS MED SMS USG ANTARES",VR::UN,VM::VM1,"Color Flow Maximum Velocity",false },
  {0x0019,0x0060,"SIEMENS MED SMS USG ANTARES",VR::UN,VM::VM1,"Doppler Dynamic Range",false },
  {0x0019,0x0061,"SIEMENS MED SMS USG ANTARES",VR::UN,VM::VM1,"Doppler Overall Gain",false },
  {0x0019,0x0062,"SIEMENS MED SMS USG ANTARES",VR::UN,VM::VM1,"Doppler Wall Filter",false },
  {0x0019,0x0063,"SIEMENS MED SMS USG ANTARES",VR::UN,VM::VM1,"Doppler Gate Size",false },
  {0x0019,0x0065,"SIEMENS MED SMS USG ANTARES",VR::UN,VM::VM1,"Doppler Map Index",false },
  {0x0019,0x0066,"SIEMENS MED SMS USG ANTARES",VR::UN,VM::VM1,"Doppler Submode",false },
  {0x0019,0x0067,"SIEMENS MED SMS USG ANTARES",VR::US,VM::VM1,"?",false },
  {0x0019,0x0069,"SIEMENS MED SMS USG ANTARES",VR::UN,VM::VM1,"Doppler Time/Freq Res Index",false },
  {0x0019,0x006a,"SIEMENS MED SMS USG ANTARES",VR::US,VM::VM1,"Doppler Trace Inverted",false },
  {0x0019,0x006c,"SIEMENS MED SMS USG ANTARES",VR::US,VM::VM1,"Doppler Tint Type",false },
  {0x0019,0x0072,"SIEMENS MED SMS USG ANTARES",VR::UN,VM::VM1,"Doppler Tint Index",false },
  {0x0019,0x0080,"SIEMENS MED SMS USG ANTARES",VR::UN,VM::VM1,"M-Mode Dynamic Range",false },
  {0x0019,0x0081,"SIEMENS MED SMS USG ANTARES",VR::UN,VM::VM1,"M-Mode Overall Gain",false },
  {0x0019,0x0082,"SIEMENS MED SMS USG ANTARES",VR::UN,VM::VM1,"M-Mode Edge Enhance Index",false },
  {0x0019,0x0083,"SIEMENS MED SMS USG ANTARES",VR::UN,VM::VM1,"M-Mode Map Index",false },
  {0x0019,0x0086,"SIEMENS MED SMS USG ANTARES",VR::UN,VM::VM1,"M-Mode Tint Type",false },
  {0x0019,0x0087,"SIEMENS MED SMS USG ANTARES",VR::UN,VM::VM1,"M-Mode Submode",false },
  {0x0019,0x0088,"SIEMENS MED SMS USG ANTARES",VR::UN,VM::VM1,"M-Mode Tint Index",false },
  {0x0019,0x00a0,"SIEMENS MED SMS USG ANTARES",VR::LT,VM::VM1,"?",false },
  {0x0039,0x0000,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"Release Version",false },
  {0x0039,0x0003,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"VolumeAcquisitionDuration",false },
  {0x0039,0x0004,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"VolumeRawDataType",false },
  {0x0039,0x0005,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"ScanType",false },
  {0x0039,0x0006,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"ZlateralMin",false },
  {0x0039,0x0007,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"ZlateralSpan",false },
  {0x0039,0x0008,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"ZRadiusOfCurvature",false },
  {0x0039,0x0009,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"WobbleCorrection",false },
  {0x0039,0x0010,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"ScaleAlongWidth",false },
  {0x0039,0x0011,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"ScaleAlongHeight",false },
  {0x0039,0x0012,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"ScaleAlongDepth",false },
  {0x0039,0x0013,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"BufferSize",false },
  {0x0039,0x0014,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"AcquisitionRate",false },
  {0x0039,0x0015,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DepthMinCm",false },
  {0x0039,0x0016,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"IsLeftRightFlippedEn",false },
  {0x0039,0x0017,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"IsUpDownFlippedEn",false },
  {0x0039,0x0018,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"IsVolumeGeomAccurate",false },
  {0x0039,0x0019,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"BByteMaskOffset",false },
  {0x0039,0x0020,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"BByteMaskSize",false },
  {0x0039,0x0021,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DepthMaxCm",false },
  {0x0039,0x0022,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"AcqPlaneRotationDeg",false },
  {0x0039,0x0023,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"BeamAxialSpan",false },
  {0x0039,0x0024,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"BeamLateralMin",false },
  {0x0039,0x0025,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"BeamLateralSpan",false },
  {0x0039,0x0026,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"BeamAxialMin",false },
  {0x0039,0x0027,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"NumDisplaySamples",false },
  {0x0039,0x0028,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DVolumeWidth",false },
  {0x0039,0x0029,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DVolumeDepth",false },
  {0x0039,0x0030,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DVolumeHeight",false },
  {0x0039,0x0031,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DVolumePosX",false },
  {0x0039,0x0032,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DVolumePosY",false },
  {0x0039,0x0033,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DVolumePosZ",false },
  {0x0039,0x0034,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DBeamAxialMin",false },
  {0x0039,0x0035,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DBeamAxialSpan",false },
  {0x0039,0x0036,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DBeamLateralMin",false },
  {0x0039,0x0037,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DBeamLateralSpan",false },
  {0x0039,0x0038,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"NumOfVolumesInSequence",false },
  {0x0039,0x0039,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DByteMaskOffset",false },
  {0x0039,0x0040,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DByteMaskSize",false },
  {0x0039,0x0050,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"PrivateCreatorVersionOfBookmark",false },
  {0x0039,0x0051,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"BCutPlaneEnable",false },
  {0x0039,0x0052,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"BMprColorMapIndex",false },
  {0x0039,0x0053,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"BMprDynamicRangeDb",false },
  {0x0039,0x0054,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"BMprGrayMapIndex",false },
  {0x0039,0x0055,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"BVolumeRenderMode",false },
  {0x0039,0x0056,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"BVrBrightness",false },
  {0x0039,0x0057,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"BVrContrast",false },
  {0x0039,0x0058,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"BVrColorMapIndex",false },
  {0x0039,0x0059,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"BVrDynamicRangeDb",false },
  {0x0039,0x005a,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"BVrGrayMapIndex",false },
  {0x0039,0x005b,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"BVrOpacity",false },
  {0x0039,0x005c,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"BVrThresholdHigh",false },
  {0x0039,0x005d,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"BVrThresholdLow",false },
  {0x0039,0x005e,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"BPreProcessFilterMix",false },
  {0x0039,0x005f,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"CCutPlaneEnable",false },
  {0x0039,0x0060,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"CFrontClipMode",false },
  {0x0039,0x0061,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"CMprColorMapIndex",false },
  {0x0039,0x0062,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"CMprColorFlowPriorityIndex",false },
  {0x0039,0x0063,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"CVolumeRenderMode",false },
  {0x0039,0x0064,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"CVrColorMapIndex",false },
  {0x0039,0x0065,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"CVrColorFlowPriorityIndex",false },
  {0x0039,0x0066,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"CVrOpacity",false },
  {0x0039,0x0067,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"CVrThresholdHigh",false },
  {0x0039,0x0068,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"CVrThresholdLow",false },
  {0x0039,0x0069,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"VoiMode",false },
  {0x0039,0x006a,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"VoiRotationOffsetDeg",false },
  {0x0039,0x006b,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"VoiSizeRatioX",false },
  {0x0039,0x006c,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"VoiSizeRatioY",false },
  {0x0039,0x006d,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"VoiSizeRatioZ",false },
  {0x0039,0x006e,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"VoiSyncPlane",false },
  {0x0039,0x006f,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"VoiViewMode",false },
  {0x0039,0x0070,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"VrOrientationA",false },
  {0x0039,0x0071,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"MprOrientationA",false },
  {0x0039,0x0072,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"VrOffsetVector",false },
  {0x0039,0x0073,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"BlendingRatio",false },
  {0x0039,0x0074,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"FusionBlendMode",false },
  {0x0039,0x0075,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"QualityFactor",false },
  {0x0039,0x0076,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"RendererType",false },
  {0x0039,0x0077,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"SliceMode",false },
  {0x0039,0x0078,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"ActiveQuad",false },
  {0x0039,0x0079,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"ScreenMode",false },
  {0x0039,0x007a,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"CutPlaneSide",false },
  {0x0039,0x007b,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"WireframeMode",false },
  {0x0039,0x007c,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"CrossmarkMode",false },
  {0x0039,0x007d,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"MprDisplayType",false },
  {0x0039,0x007e,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"VolumeDisplayType",false },
  {0x0039,0x007f,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"LastReset",false },
  {0x0039,0x0080,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"LastNonFullScreenMode",false },
  {0x0039,0x0081,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"MprToolIndex",false },
  {0x0039,0x0082,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"VoiToolIndex",false },
  {0x0039,0x0083,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"ToolLoopMode",false },
  {0x0039,0x0084,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"VolumeArbMode",false },
  {0x0039,0x0085,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"MprZoomEn",false },
  {0x0039,0x0086,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"IsVolumeZoomEn",false },
  {0x0039,0x0087,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"ZoomLevelMpr",false },
  {0x0039,0x0088,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"ZoomLevelVolume",false },
  {0x0039,0x0089,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"IsAutoRotateEn",false },
  {0x0039,0x008a,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"AutoRotateAxis",false },
  {0x0039,0x008b,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"AutoRotateRangeIndex",false },
  {0x0039,0x008c,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"AutoRotateSpeedIndex",false },
  {0x0039,0x008d,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"CVrBrightness",false },
  {0x0039,0x008e,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"CFlowStateIndex",false },
  {0x0039,0x008f,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"BSubmodeIndex",false },
  {0x0039,0x0090,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"CSubmodeIndex",false },
  {0x0039,0x0091,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DICOMAttrNameCutPlane",false },
  {0x0039,0x0092,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"BookmarkChunkId",false },
  {0x0039,0x0093,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"SequenceMinChunkId",false },
  {0x0039,0x0094,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"SequenceMaxChunkId",false },
  {0x0039,0x0095,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"VolumeRateHz",false },
  {0x0039,0x009a,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"VoiPositionOffsetX",false },
  {0x0039,0x009b,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"VoiPositionOffsetY",false },
  {0x0039,0x009c,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"VoiPositionOffsetZ",false },
  {0x0039,0x009d,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"VrToolIndex",false },
  {0x0039,0x009e,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"ShadingPercent",false },
  {0x0039,0x009f,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"VolumeType",false },
  {0x0039,0x00a0,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DICOMAttrNameVrQuadDisplayType",false },
  {0x0039,0x00a1,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DICOMAttrNameMprCenterLocation",false },
  {0x0039,0x00e0,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DICOMAttrNameSliceRangeType",false },
  {0x0039,0x00e1,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DICOMAttrNameSliceMPRPlane",false },
  {0x0039,0x00e2,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DICOMAttrNameSliceLayout",false },
  {0x0039,0x00e3,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DICOMAttrNameSliceSpacing",false },
  {0x0039,0x00e4,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DICOMAttrNameThinVrMode",false },
  {0x0039,0x00e5,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DICOMAttrNameThinVrThickness",false },
  {0x0039,0x00e6,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DICOMAttrNameVoiPivotX",false },
  {0x0039,0x00e7,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DICOMAttrNameVoiPivotY",false },
  {0x0039,0x00e8,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DICOMAttrNameVoiPivotZ",false },
  {0x0039,0x00e9,"SIEMENS MED SMS USG ANTARES 3D VOLUME",VR::UN,VM::VM1,"DICOMAttrNameCTopVoiQuad",false },
  {0x0019,0x0000,"SIEMENS MED SMS USG S2000",VR::SH,VM::VM1,"Private Creator Version",false },
  {0x0019,0x0003,"SIEMENS MED SMS USG S2000",VR::FD,VM::VM1,"Frame Rate",false },
  {0x0019,0x000c,"SIEMENS MED SMS USG S2000",VR::US,VM::VM1,"Burned in Graphics",false },
  {0x0019,0x000d,"SIEMENS MED SMS USG S2000",VR::SH,VM::VM1,"SieClear Index",false },
  {0x0019,0x0020,"SIEMENS MED SMS USG S2000",VR::SH,VM::VM1,"B-Mode Submode",false },
  {0x0019,0x0021,"SIEMENS MED SMS USG S2000",VR::FD,VM::VM1,"B-Mode Dynamic Range",false },
  {0x0019,0x0022,"SIEMENS MED SMS USG S2000",VR::FD,VM::VM1,"B-Mode Overall Gain",false },
  {0x0019,0x0023,"SIEMENS MED SMS USG S2000",VR::US,VM::VM1,"B-Mode Res/Speed Index",false },
  {0x0019,0x0024,"SIEMENS MED SMS USG S2000",VR::US,VM::VM1,"B-Mode Edge Enhance Index",false },
  {0x0019,0x0025,"SIEMENS MED SMS USG S2000",VR::US,VM::VM1,"B-Mode Persistence Index",false },
  {0x0019,0x0026,"SIEMENS MED SMS USG S2000",VR::US,VM::VM1,"B-Mode Map Index",false },
  {0x0019,0x0027,"SIEMENS MED SMS USG S2000",VR::US,VM::VM1,"??",false },
  {0x0019,0x0028,"SIEMENS MED SMS USG S2000",VR::US,VM::VM1,"??",false },
  {0x0019,0x0029,"SIEMENS MED SMS USG S2000",VR::US,VM::VM1,"??",false },
  {0x0019,0x002a,"SIEMENS MED SMS USG S2000",VR::US,VM::VM1,"B-Mode Tint Type",false },
  {0x0019,0x002d,"SIEMENS MED SMS USG S2000",VR::US,VM::VM1,"B-Mode Tint Index",false },
  {0x0019,0x002e,"SIEMENS MED SMS USG S2000",VR::SH,VM::VM1,"ClarifyVE Index",false },
  {0x0019,0x003a,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"Image Flag",false },
  {0x0019,0x003b,"SIEMENS MED SMS USG S2000",VR::LT,VM::VM1,"IPDA/MDA Data",false },
  {0x0019,0x003c,"SIEMENS MED SMS USG S2000",VR::UT,VM::VM1,"Thumbnail Data",false },
  {0x0019,0x0040,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"Color Flow state",false },
  {0x0019,0x0041,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"Color Flow Wall Filter Index",false },
  {0x0019,0x0042,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"Color Flow Submode",false },
  {0x0019,0x0043,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"Color Flow Overall Gain",false },
  {0x0019,0x0044,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"Color Flow Res/Speed Index",false },
  {0x0019,0x0046,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"Color Flow Smooth Index",false },
  {0x0019,0x0047,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"Color Flow Persistence Index",false },
  {0x0019,0x0048,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"Color Flow Map Index",false },
  {0x0019,0x0049,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"Color Flow Priority Index",false },
  {0x0019,0x0054,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"Color Flow Maximum Velocity",false },
  {0x0019,0x0060,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"Doppler Dynamic Range",false },
  {0x0019,0x0061,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"Doppler Overall Gain",false },
  {0x0019,0x0062,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"S2000 Private Doppler Wall Filter",false },
  {0x0019,0x0063,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"Attributes Doppler Gate Size",false },
  {0x0019,0x0065,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"Doppler Map Index",false },
  {0x0019,0x0066,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"Doppler Submode",false },
  {0x0019,0x0069,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"Doppler Time/Freq Res Index",false },
  {0x0019,0x006a,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"Doppler Trace Inverted",false },
  {0x0019,0x006c,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"Doppler Tint Type",false },
  {0x0019,0x0072,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"Doppler Tint Index",false },
  {0x0019,0x0080,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"M-Mode Dynamic Range",false },
  {0x0019,0x0081,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"M-Mode Overall Gain",false },
  {0x0019,0x0082,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"M-Mode Edge Enhance Index",false },
  {0x0019,0x0083,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"M-Mode Map Index",false },
  {0x0019,0x0086,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"M-Mode Tint Type",false },
  {0x0019,0x0087,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"M-Mode Submode",false },
  {0x0019,0x0088,"SIEMENS MED SMS USG S2000",VR::UN,VM::VM1,"M-Mode Tint Index",false },
  {0x0019,0x0001,"SIEMENS MED SP DXMG WH AWS 1",VR::UL,VM::VM1_n,"AEC Coordinates",false },
  {0x0019,0x0002,"SIEMENS MED SP DXMG WH AWS 1",VR::US,VM::VM2,"AEC Coordinates Size",false },
  {0x0019,0x0010,"SIEMENS MED SP DXMG WH AWS 1",VR::ST,VM::VM1,"Derivation Description",false },
  {0x0041,0x0002,"SIEMENS MED SP DXMG WH AWS 1",VR::LO,VM::VM1,"Reason for the Requested Procedure",false },
  {0x0055,0x0001,"SIEMENS MED SP DXMG WH AWS 1",VR::LO,VM::VM1,"Projection View Display String",false },
  {0x0029,0x0008,"SIEMENS MEDCOM HEADER",VR::CS,VM::VM1,"MedCom Header Type",false },
  {0x0029,0x0009,"SIEMENS MEDCOM HEADER",VR::LO,VM::VM1,"MedCom Header Version",false },
  {0x0029,0x0010,"SIEMENS MEDCOM HEADER",VR::OB,VM::VM1,"MedCom Header Info",false },
  {0x0029,0x0020,"SIEMENS MEDCOM HEADER",VR::OB,VM::VM1,"MedCom History Information",false },
  {0x0029,0x0031,"SIEMENS MEDCOM HEADER",VR::LO,VM::VM1,"PMTF Information 1",false },
  {0x0029,0x0032,"SIEMENS MEDCOM HEADER",VR::UL,VM::VM1,"PMTF Information 2",false },
  {0x0029,0x0033,"SIEMENS MEDCOM HEADER",VR::UL,VM::VM1,"PMTF Information 3",false },
  {0x0029,0x0034,"SIEMENS MEDCOM HEADER",VR::CS,VM::VM1,"PMTF Information 4",false },
  {0x0029,0x0035,"SIEMENS MEDCOM HEADER",VR::UL,VM::VM1,"PMTF Information 5",false },
  {0x0029,0x0040,"SIEMENS MEDCOM HEADER",VR::SQ,VM::VM1,"Application Header Sequence",false },
  {0x0029,0x0041,"SIEMENS MEDCOM HEADER",VR::CS,VM::VM1,"Application Header Type",false },
  {0x0029,0x0042,"SIEMENS MEDCOM HEADER",VR::LO,VM::VM1,"Application Header ID",false },
  {0x0029,0x0043,"SIEMENS MEDCOM HEADER",VR::LO,VM::VM1,"Application Header Version",false },
  {0x0029,0x0044,"SIEMENS MEDCOM HEADER",VR::OB,VM::VM1,"Application Header Info",false },
  {0x0029,0x0050,"SIEMENS MEDCOM HEADER",VR::LO,VM::VM8,"Workflow Control Flags",false },
  {0x0029,0x0051,"SIEMENS MEDCOM HEADER",VR::CS,VM::VM1,"Arch. Management Flag Keep Online",false },
  {0x0029,0x0052,"SIEMENS MEDCOM HEADER",VR::CS,VM::VM1,"Arch. Mgmnt Flag Do Not Archive",false },
  {0x0029,0x0053,"SIEMENS MEDCOM HEADER",VR::CS,VM::VM1,"Image Location Status",false },
  {0x0029,0x0054,"SIEMENS MEDCOM HEADER",VR::DS,VM::VM1,"Estimated Retrieve Time",false },
  {0x0029,0x0055,"SIEMENS MEDCOM HEADER",VR::DS,VM::VM1,"Data Size of Retrieved Images",false },
  {0x0029,0x0070,"SIEMENS MEDCOM HEADER",VR::SQ,VM::VM1,"Siemens Link Sequence",false },
  {0x0029,0x0071,"SIEMENS MEDCOM HEADER",VR::AT,VM::VM1,"Referenced Tag",false },
  {0x0029,0x0072,"SIEMENS MEDCOM HEADER",VR::CS,VM::VM1,"Referenced Tag Type",false },
  {0x0029,0x0073,"SIEMENS MEDCOM HEADER",VR::UL,VM::VM1,"Referenced Value Length",false },
  {0x0029,0x0074,"SIEMENS MEDCOM HEADER",VR::CS,VM::VM1,"Referenced Object Device Type",false },
  {0x0029,0x0075,"SIEMENS MEDCOM HEADER",VR::OB,VM::VM1,"Referenced Object Device Location",false },
  {0x0029,0x0076,"SIEMENS MEDCOM HEADER",VR::OB,VM::VM1,"Referenced Object ID",false },
  {0x0029,0x0008,"SIEMENS MEDCOM HEADER2",VR::CS,VM::VM1,"CSA Image Header Type",false },
  {0x0029,0x0009,"SIEMENS MEDCOM HEADER2",VR::LO,VM::VM1,"CSA Image Header Version",false },
  {0x0029,0x0010,"SIEMENS MEDCOM HEADER2",VR::OB,VM::VM1,"CSA Image Header Info",false },
  {0x0029,0x0018,"SIEMENS MEDCOM HEADER2",VR::CS,VM::VM1,"CSA Series Header Type",false },
  {0x0029,0x0019,"SIEMENS MEDCOM HEADER2",VR::LO,VM::VM1,"CSA Series Header Version",false },
  {0x0029,0x0020,"SIEMENS MEDCOM HEADER2",VR::OB,VM::VM1,"CSA Series Header Info",false },
  {0x0029,0x0060,"SIEMENS MEDCOM HEADER2",VR::LO,VM::VM1,"Series Workflow Status",false },
  {0x0029,0x0008,"SIEMENS MEDCOM OOG",VR::CS,VM::VM1,"MedCom OOG Type",false },
  {0x0029,0x0009,"SIEMENS MEDCOM OOG",VR::LO,VM::VM1,"MedCom OOG Version",false },
  {0x0029,0x0010,"SIEMENS MEDCOM OOG",VR::OB,VM::VM1,"MedCom OOG Info",false },
  {0x0019,0x0008,"SIEMENS MR HEADER",VR::CS,VM::VM1,"CSA Image Header Type",false },
  {0x0019,0x0009,"SIEMENS MR HEADER",VR::LO,VM::VM1,"CSA Image Header Version ??",false },
  {0x0019,0x000a,"SIEMENS MR HEADER",VR::US,VM::VM1,"NumberOfImagesInMosaic",false },
  {0x0019,0x000b,"SIEMENS MR HEADER",VR::DS,VM::VM1,"SliceMeasurementDuration",false },
  {0x0019,0x000c,"SIEMENS MR HEADER",VR::IS,VM::VM1,"B_value",false },
  {0x0019,0x000d,"SIEMENS MR HEADER",VR::CS,VM::VM1,"DiffusionDirectionality",false },
  {0x0019,0x000e,"SIEMENS MR HEADER",VR::FD,VM::VM3,"DiffusionGradientDirection",false },
  {0x0019,0x000f,"SIEMENS MR HEADER",VR::SH,VM::VM1,"GradientMode",false },
  {0x0019,0x0011,"SIEMENS MR HEADER",VR::SH,VM::VM1,"FlowCompensation",false },
  {0x0019,0x0012,"SIEMENS MR HEADER",VR::SL,VM::VM3,"TablePositionOrigin",false },
  {0x0019,0x0013,"SIEMENS MR HEADER",VR::SL,VM::VM3,"ImaAbsTablePosition",false },
  {0x0019,0x0014,"SIEMENS MR HEADER",VR::IS,VM::VM3,"ImaRelTablePosition",false },
  {0x0019,0x0015,"SIEMENS MR HEADER",VR::FD,VM::VM3,"SlicePosition_PCS",false },
  {0x0019,0x0016,"SIEMENS MR HEADER",VR::DS,VM::VM1,"TimeAfterStart",false },
  {0x0019,0x0017,"SIEMENS MR HEADER",VR::DS,VM::VM1,"SliceResolution",false },
  {0x0019,0x0018,"SIEMENS MR HEADER",VR::IS,VM::VM1,"RealDwellTime",false },
  {0x0019,0x0023,"SIEMENS MR HEADER",VR::IS,VM::VM1,"FMRIStimulInfo",false },
  {0x0019,0x0024,"SIEMENS MR HEADER",VR::FD,VM::VM2,"?",false },
  {0x0019,0x0025,"SIEMENS MR HEADER",VR::FD,VM::VM3,"RBMoCoTrans",false },
  {0x0019,0x0026,"SIEMENS MR HEADER",VR::FD,VM::VM3,"RBMoCoRot",false },
  {0x0019,0x0027,"SIEMENS MR HEADER",VR::FD,VM::VM6,"B_matrix",false },
  {0x0019,0x0028,"SIEMENS MR HEADER",VR::FD,VM::VM1,"BandwidthPerPixelPhaseEncode",false },
  {0x0019,0x0029,"SIEMENS MR HEADER",VR::FD,VM::VM1_n,"MosaicRefAcqTimes",false },
  {0x0051,0x0008,"SIEMENS MR HEADER",VR::CS,VM::VM1,"CSA Image Header Type",false },
  {0x0051,0x0009,"SIEMENS MR HEADER",VR::LO,VM::VM1,"CSA Image Header Version ??",false },
  {0x0051,0x000a,"SIEMENS MR HEADER",VR::LO,VM::VM1,"Meas Duration",false },
  {0x0051,0x000b,"SIEMENS MR HEADER",VR::LO,VM::VM1,"Acquisition Matrix (Reconstruction Matrix if Interpolated)",false },
  {0x0051,0x000c,"SIEMENS MR HEADER",VR::LO,VM::VM1,"Field of View",false },
  {0x0051,0x000d,"SIEMENS MR HEADER",VR::SH,VM::VM1,"Slice Position",false },
  {0x0051,0x000e,"SIEMENS MR HEADER",VR::LO,VM::VM1,"Slice Orientation",false },
  {0x0051,0x000f,"SIEMENS MR HEADER",VR::LO,VM::VM1,"Coil String",false },
  {0x0051,0x0011,"SIEMENS MR HEADER",VR::LO,VM::VM1,"PAT Mode (Ima PAT Mode)",false },
  {0x0051,0x0012,"SIEMENS MR HEADER",VR::SH,VM::VM1,"Rel Table Position",false },
  {0x0051,0x0013,"SIEMENS MR HEADER",VR::SH,VM::VM1,"Positive PCS Directions",false },
  {0x0051,0x0014,"SIEMENS MR HEADER",VR::SH,VM::VM1,"?v150_through?",false },
  {0x0051,0x0015,"SIEMENS MR HEADER",VR::SH,VM::VM1,"?Data Filter?",false },
  {0x0051,0x0016,"SIEMENS MR HEADER",VR::LO,VM::VM1,"?Image Type?",false },
  {0x0051,0x0017,"SIEMENS MR HEADER",VR::SH,VM::VM1,"?Slice Thickness?",false },
  {0x0051,0x0018,"SIEMENS MR HEADER",VR::SH,VM::VM1,"?",false },
  {0x0051,0x0019,"SIEMENS MR HEADER",VR::LO,VM::VM1,"?Scan options?",false },
  {0x0019,0x0012,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1,"Magnetic Field Strength",false },
  {0x0019,0x0014,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1,"ADC Voltage",false },
  {0x0019,0x0016,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1_n,"ADC Offset",false },
  {0x0019,0x0020,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1,"Transmitter Amplitude",false },
  {0x0019,0x0021,"SIEMENS MR VA0  COAD",VR::IS,VM::VM1,"Number of Transmitter Amplitudes",false },
  {0x0019,0x0022,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1,"Transmitter Attenuator",false },
  {0x0019,0x0024,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1,"Transmitter Calibration",false },
  {0x0019,0x0026,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1,"Transmitter Reference",false },
  {0x0019,0x0050,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1,"Receiver Total Gain",false },
  {0x0019,0x0051,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1,"Receiver Amplifier Gain",false },
  {0x0019,0x0052,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1,"Receiver Preamplifier Gain",false },
  {0x0019,0x0054,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1,"Receiver Cable Attenuation",false },
  {0x0019,0x0055,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1,"Receiver Reference Gain",false },
  {0x0019,0x0056,"SIEMENS MR VA0  COAD",VR::IS,VM::VM1,"Receiver Filter Frequency",false },
  {0x0019,0x0060,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1,"Reconstruction Scale Factor",false },
  {0x0019,0x0062,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1,"Reference Scale Factor",false },
  {0x0019,0x0070,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1,"Phase Gradient Amplitude",false },
  {0x0019,0x0071,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1,"Readout Gradient Amplitude",false },
  {0x0019,0x0072,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1,"Selection Gradient Amplitude",false },
  {0x0019,0x0080,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1_n,"Gradient Delay Time",false },
  {0x0019,0x0082,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1,"Total Gradient Delay Time",false },
  {0x0019,0x0090,"SIEMENS MR VA0  COAD",VR::LO,VM::VM1,"Sensitivity Correction Label",false },
  {0x0019,0x0091,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1_5,"Saturation Phase Encoding Vector Coronal Component",false },
  {0x0019,0x0092,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1_5,"Saturation Readout Vector Coronal Component",false },
  {0x0019,0x00a0,"SIEMENS MR VA0  COAD",VR::IS,VM::VM1,"RF Watchdog Mask",false },
  {0x0019,0x00a1,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1,"EPI Reconstruction Slope",false },
  {0x0019,0x00a2,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1,"RF Power Error Indicator",false },
  {0x0019,0x00a5,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1_n,"Specific Absorption Rate SAR",false },
  {0x0019,0x00a6,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1_n,"Specific Energy Dose SED",false },
  {0x0019,0x00b0,"SIEMENS MR VA0  COAD",VR::UL,VM::VM1,"Adjustment Status Mask",false },
  {0x0019,0x00c1,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1_n,"EPI Capacity",false },
  {0x0019,0x00c2,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1_n,"EPI Inductance",false },
  {0x0019,0x00c3,"SIEMENS MR VA0  COAD",VR::IS,VM::VM1_n,"EPI Switch Configuration Code",false },
  {0x0019,0x00c4,"SIEMENS MR VA0  COAD",VR::IS,VM::VM1_n,"EPI Switch Hardware Code",false },
  {0x0019,0x00c5,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1_n,"EPI Switch Delay Time",false },
  {0x0019,0x00d1,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1,"Flow Sensitivity",false },
  {0x0019,0x00d2,"SIEMENS MR VA0  COAD",VR::CS,VM::VM1,"Calculation Submode",false },
  {0x0019,0x00d3,"SIEMENS MR VA0  COAD",VR::DS,VM::VM1,"Field of View Ratio",false },
  {0x0019,0x00d4,"SIEMENS MR VA0  COAD",VR::IS,VM::VM1,"Base Raw Matrix Size",false },
  {0x0019,0x00d5,"SIEMENS MR VA0  COAD",VR::IS,VM::VM1,"2D Phase Oversampling Lines",false },
  {0x0019,0x00d6,"SIEMENS MR VA0  COAD",VR::IS,VM::VM1,"3D Phase Oversampling Part",false },
  {0x0019,0x00d7,"SIEMENS MR VA0  COAD",VR::IS,VM::VM1,"Echo Line Position",false },
  {0x0019,0x00d8,"SIEMENS MR VA0  COAD",VR::IS,VM::VM1,"Echo Column Position",false },
  {0x0019,0x00d9,"SIEMENS MR VA0  COAD",VR::IS,VM::VM1,"Lines per Segment",false },
  {0x0019,0x00da,"SIEMENS MR VA0  COAD",VR::CS,VM::VM1,"Phase Coding Direction",false },
  {0x0019,0x0010,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1,"Total Measurement Time nominal",false },
  {0x0019,0x0011,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1,"Total Measurement Time current",false },
  {0x0019,0x0012,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1,"Start Delay Time",false },
  {0x0019,0x0013,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1,"Dwell Time",false },
  {0x0019,0x0014,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"Number of Phases",false },
  {0x0019,0x0016,"SIEMENS MR VA0  GEN",VR::UL,VM::VM1_n,"Sequence Control Mask",false },
  {0x0019,0x0018,"SIEMENS MR VA0  GEN",VR::UL,VM::VM1,"Measurement Status Mask",false },
  {0x0019,0x0020,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"Number of Fourier Lines nominal",false },
  {0x0019,0x0021,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"Number of Fourier Lines current",false },
  {0x0019,0x0026,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"Number of Fourier Lines after Zero",false },
  {0x0019,0x0028,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"First Measured Fourier Line",false },
  {0x0019,0x0030,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"Acquisition Columns",false },
  {0x0019,0x0031,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"Reconstruction Columns",false },
  {0x0019,0x0040,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"AC Element Number",false },
  {0x0019,0x0041,"SIEMENS MR VA0  GEN",VR::UL,VM::VM1,"AC Element Select Mask",false },
  {0x0019,0x0042,"SIEMENS MR VA0  GEN",VR::UL,VM::VM1,"AC Element Data Mask",false },
  {0x0019,0x0043,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1_n,"AC Element To ADC Connect",false },
  {0x0019,0x0044,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1_n,"AC Element Noise Level",false },
  {0x0019,0x0045,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"AC ADC Pair Number",false },
  {0x0019,0x0046,"SIEMENS MR VA0  GEN",VR::UL,VM::VM1,"AC Combination Mask",false },
  {0x0019,0x0050,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"Number of Averages current",false },
  {0x0019,0x0060,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1,"Flip Angle",false },
  {0x0019,0x0070,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"Number of Prescans",false },
  {0x0019,0x0081,"SIEMENS MR VA0  GEN",VR::CS,VM::VM1,"Filter Type for Raw Data",false },
  {0x0019,0x0082,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1_4,"Filter Parameter for Raw Data",false },
  {0x0019,0x0083,"SIEMENS MR VA0  GEN",VR::CS,VM::VM1,"Filter Type for Image Data",false },
  {0x0019,0x0084,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1_4,"Filter Parameter for Image Data",false },
  {0x0019,0x0085,"SIEMENS MR VA0  GEN",VR::CS,VM::VM1,"Filter Type for Phase Correction",false },
  {0x0019,0x0086,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1_4,"Filter Parameter for Phase Correction",false },
  {0x0019,0x0087,"SIEMENS MR VA0  GEN",VR::CS,VM::VM1,"Normalization Filter Type for Image Data",false },
  {0x0019,0x0088,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1_4,"Normalization Filter Parameter for Image Data",false },
  {0x0019,0x0090,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"Number of Saturation Regions",false },
  {0x0019,0x0091,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1_5,"Saturation Phase Encoding Vector Sagittal Compo...",false },
  {0x0019,0x0092,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1_5,"Saturation Readout Vector Sagittal Component",false },
  {0x0019,0x0093,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"EPI Stimulation Monitor Mode",false },
  {0x0019,0x0094,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1,"Image Rotation Angle",false },
  {0x0019,0x0096,"SIEMENS MR VA0  GEN",VR::UL,VM::VM1_n,"Coil ID Mask",false },
  {0x0019,0x0097,"SIEMENS MR VA0  GEN",VR::UL,VM::VM1_n,"Coil Class Mask",false },
  {0x0019,0x0098,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1_n,"Coil Position",false },
  {0x0019,0x00a0,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1,"EPI Reconstruction Phase",false },
  {0x0019,0x00a1,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1,"EPI Reconstruction Slope",false },
  {0x0021,0x0020,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"Phase Corrections Rows sequence",false },
  {0x0021,0x0021,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"Phase Corrections Columns sequence",false },
  {0x0021,0x0022,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"Phase Corrections Rows reconstruction",false },
  {0x0021,0x0024,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"Phase Corrections Columns reconstruction",false },
  {0x0021,0x0030,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"Number of 3D Raw Partitions nominal",false },
  {0x0021,0x0031,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"Number of 3D Raw Partitions current",false },
  {0x0021,0x0034,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"Number of 3D Image Partitions",false },
  {0x0021,0x0036,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"Actual 3D Image Partition Number",false },
  {0x0021,0x0039,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1,"Slab Thickness",false },
  {0x0021,0x0040,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"Number of Slices nominal",false },
  {0x0021,0x0041,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"Number of Slices current",false },
  {0x0021,0x0042,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"Current Slice Number",false },
  {0x0021,0x0043,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"Current Group Number",false },
  {0x0021,0x0044,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1,"Current Slice Distance Factor",false },
  {0x0021,0x0045,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"MIP Start Row",false },
  {0x0021,0x0046,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"MIP Stop Row",false },
  {0x0021,0x0047,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"MIP Start Column",false },
  {0x0021,0x0048,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"MIP Stop Column",false },
  {0x0021,0x0049,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"MIP Start Slice",false },
  {0x0021,0x004a,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"MIP Stop Slice",false },
  {0x0021,0x004f,"SIEMENS MR VA0  GEN",VR::CS,VM::VM1,"Order of Slices",false },
  {0x0021,0x0050,"SIEMENS MR VA0  GEN",VR::UL,VM::VM1,"Signal Mask",false },
  {0x0021,0x0052,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"Delay after Trigger",false },
  {0x0021,0x0053,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"RRInterval",false },
  {0x0021,0x0054,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1,"Number of Trigger Pulses",false },
  {0x0021,0x0056,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1,"Repetition Time effective",false },
  {0x0021,0x0057,"SIEMENS MR VA0  GEN",VR::CS,VM::VM1,"Gate Phase",false },
  {0x0021,0x0058,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1,"Gate Threshold",false },
  {0x0021,0x0059,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1,"Gated Ratio",false },
  {0x0021,0x0060,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"Number of Interpolated Images",false },
  {0x0021,0x0070,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"Number of Echoes",false },
  {0x0021,0x0072,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1,"Second Echo Time",false },
  {0x0021,0x0073,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1,"Second Repetition Time",false },
  {0x0021,0x0080,"SIEMENS MR VA0  GEN",VR::IS,VM::VM1,"Cardiac Code",false },
  {0x0021,0x0091,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1_n,"Saturation Phase Encoding Vector Transversal Co...",false },
  {0x0021,0x0092,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1_n,"Saturation Readout Vector Transversal Component",false },
  {0x0021,0x0093,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1,"EPI Change Value of Magnitude",false },
  {0x0021,0x0094,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1,"EPI Change Value of X Component",false },
  {0x0021,0x0095,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1,"EPI Change Value of Y Component",false },
  {0x0021,0x0096,"SIEMENS MR VA0  GEN",VR::DS,VM::VM1,"EPI Change Value of Z Component",false },
  {0x0021,0x0000,"SIEMENS MR VA0  RAW",VR::CS,VM::VM1,"Sequence Type",false },
  {0x0021,0x0001,"SIEMENS MR VA0  RAW",VR::IS,VM::VM1,"Vector Size original",false },
  {0x0021,0x0002,"SIEMENS MR VA0  RAW",VR::IS,VM::VM1,"Vector Size extended",false },
  {0x0021,0x0003,"SIEMENS MR VA0  RAW",VR::DS,VM::VM1,"Acquired Spectral Range",false },
  {0x0021,0x0004,"SIEMENS MR VA0  RAW",VR::DS,VM::VM1_n,"VOI Position",false },
  {0x0021,0x0005,"SIEMENS MR VA0  RAW",VR::DS,VM::VM1_n,"VOI Size",false },
  {0x0021,0x0006,"SIEMENS MR VA0  RAW",VR::IS,VM::VM1_n,"CSI Matrix Size original",false },
  {0x0021,0x0007,"SIEMENS MR VA0  RAW",VR::IS,VM::VM1_n,"CSI Matrix Size extended",false },
  {0x0021,0x0008,"SIEMENS MR VA0  RAW",VR::DS,VM::VM1_n,"Spatial Grid Shift",false },
  {0x0021,0x0009,"SIEMENS MR VA0  RAW",VR::DS,VM::VM1,"Signal Limits Minimum",false },
  {0x0021,0x0010,"SIEMENS MR VA0  RAW",VR::DS,VM::VM1,"Signal Limits Maximum",false },
  {0x0021,0x0011,"SIEMENS MR VA0  RAW",VR::DS,VM::VM1,"Spec Info Mask",false },
  {0x0021,0x0012,"SIEMENS MR VA0  RAW",VR::DS,VM::VM1,"EPI Time Rate of Change of Magnitude",false },
  {0x0021,0x0013,"SIEMENS MR VA0  RAW",VR::DS,VM::VM1,"EPI Time Rate of Change of X Component",false },
  {0x0021,0x0014,"SIEMENS MR VA0  RAW",VR::DS,VM::VM1,"EPI Time Rate of Change of Y Component",false },
  {0x0021,0x0015,"SIEMENS MR VA0  RAW",VR::DS,VM::VM1,"EPI Time Rate of Change of Z Component",false },
  {0x0021,0x0016,"SIEMENS MR VA0  RAW",VR::DS,VM::VM1,"EPI Time Rate of Change Legal Limit 1",false },
  {0x0021,0x0017,"SIEMENS MR VA0  RAW",VR::IS,VM::VM1,"EPI Operation Mode Flag",false },
  {0x0021,0x0018,"SIEMENS MR VA0  RAW",VR::DS,VM::VM1,"EPI Field Calculation Safety Factor",false },
  {0x0021,0x0019,"SIEMENS MR VA0  RAW",VR::DS,VM::VM1,"EPI Legal Limit 1 of Change Value",false },
  {0x0021,0x0020,"SIEMENS MR VA0  RAW",VR::DS,VM::VM1,"EPI Legal Limit 2 of Change Value",false },
  {0x0021,0x0021,"SIEMENS MR VA0  RAW",VR::DS,VM::VM1,"EPI Rise Time",false },
  {0x0021,0x0030,"SIEMENS MR VA0  RAW",VR::DS,VM::VM1_n,"AC ADC Offset",false },
  {0x0021,0x0031,"SIEMENS MR VA0  RAW",VR::DS,VM::VM1_n,"AC Preamplifier Gain",false },
  {0x0021,0x0050,"SIEMENS MR VA0  RAW",VR::CS,VM::VM1,"Saturation Type",false },
  {0x0021,0x0051,"SIEMENS MR VA0  RAW",VR::DS,VM::VM1_n,"Saturation Normal Vector",false },
  {0x0021,0x0052,"SIEMENS MR VA0  RAW",VR::DS,VM::VM1_n,"Saturation Position Vector",false },
  {0x0021,0x0053,"SIEMENS MR VA0  RAW",VR::DS,VM::VM1_n,"Saturation Thickness",false },
  {0x0021,0x0054,"SIEMENS MR VA0  RAW",VR::DS,VM::VM1_n,"Saturation Width",false },
  {0x0021,0x0055,"SIEMENS MR VA0  RAW",VR::DS,VM::VM1_n,"Saturation Distance",false },
  {0x7fe3,0x0000,"SIEMENS NUMARIS II",VR::LO,VM::VM1,"Image Graphics Format Code",false },
  {0x7fe3,0x0010,"SIEMENS NUMARIS II",VR::OB,VM::VM1,"Image Graphics",false },
  {0x7fe3,0x0020,"SIEMENS NUMARIS II",VR::OB,VM::VM1,"Image Graphics Dummy",false },
  {0x0011,0x0020,"SIEMENS RA GEN",VR::UL,VM::VM1,"Total Fluoro TimePatient",false },
  {0x0011,0x0025,"SIEMENS RA GEN",VR::UL,VM::VM1,"Dose Area ProductPatient",false },
  {0x0011,0x0030,"SIEMENS RA GEN",VR::LO,VM::VM1,"Patient DOB",false },
  {0x0011,0x0035,"SIEMENS RA GEN",VR::LO,VM::VM1,"Patient Initial Puck Counter",false },
  {0x0011,0x0040,"SIEMENS RA GEN",VR::CS,VM::VM1,"Patient Initial Position",false },
  {0x0019,0x0015,"SIEMENS RA GEN",VR::CS,VM::VM1,"Acquired Plane",false },
  {0x0019,0x0020,"SIEMENS RA GEN",VR::UL,VM::VM1,"Scene Flag",false },
  {0x0019,0x0022,"SIEMENS RA GEN",VR::UL,VM::VM1,"ReferencePhotofile Flag",false },
  {0x0019,0x0024,"SIEMENS RA GEN",VR::LO,VM::VM1,"Scene Name",false },
  {0x0019,0x0026,"SIEMENS RA GEN",VR::US,VM::VM1,"Acquisition Index",false },
  {0x0019,0x0028,"SIEMENS RA GEN",VR::US,VM::VM1,"Mixed Pulse Mode",false },
  {0x0019,0x002a,"SIEMENS RA GEN",VR::US,VM::VM1,"Number of Positions",false },
  {0x0019,0x002c,"SIEMENS RA GEN",VR::US,VM::VM1,"Number of Phases",false },
  {0x0019,0x002e,"SIEMENS RA GEN",VR::US,VM::VM1_n,"Framerate for 8 Positions",false },
  {0x0019,0x0030,"SIEMENS RA GEN",VR::US,VM::VM1_n,"FramesPhase",false },
  {0x0019,0x0032,"SIEMENS RA GEN",VR::US,VM::VM1,"Stepping Direction",false },
  {0x0019,0x0034,"SIEMENS RA GEN",VR::US,VM::VM1,"?",false },
  {0x0019,0x0036,"SIEMENS RA GEN",VR::US,VM::VM1,"?",false },
  {0x0019,0x0038,"SIEMENS RA GEN",VR::US,VM::VM1,"?",false },
  {0x0019,0x003a,"SIEMENS RA GEN",VR::US,VM::VM1,"?",false },
  {0x0019,0x003c,"SIEMENS RA GEN",VR::US,VM::VM1,"?",false },
  {0x0019,0x003e,"SIEMENS RA GEN",VR::US,VM::VM1,"?",false },
  {0x0019,0x0040,"SIEMENS RA GEN",VR::US,VM::VM1,"?",false },
  {0x0019,0x0042,"SIEMENS RA GEN",VR::US,VM::VM1,"?",false },
  {0x0019,0x0044,"SIEMENS RA GEN",VR::US,VM::VM1,"Image Transfer Delay",false },
  {0x0019,0x0046,"SIEMENS RA GEN",VR::UL,VM::VM1,"InversNot Invers",false },
  {0x0019,0x0048,"SIEMENS RA GEN",VR::US,VM::VM1,"?",false },
  {0x0019,0x004a,"SIEMENS RA GEN",VR::US,VM::VM1,"?",false },
  {0x0019,0x004c,"SIEMENS RA GEN",VR::US,VM::VM1,"Blanking Circle Diameter",false },
  {0x0019,0x0050,"SIEMENS RA GEN",VR::US,VM::VM1,"Stand Data Valid",false },
  {0x0019,0x0052,"SIEMENS RA GEN",VR::US,VM::VM1,"Table Tilt",false },
  {0x0019,0x0054,"SIEMENS RA GEN",VR::US,VM::VM1,"Table Axis Rotation",false },
  {0x0019,0x0056,"SIEMENS RA GEN",VR::US,VM::VM1,"Table Longitudinal Position",false },
  {0x0019,0x0058,"SIEMENS RA GEN",VR::US,VM::VM1,"Table Side Offset",false },
  {0x0019,0x005a,"SIEMENS RA GEN",VR::US,VM::VM1,"Table Isocenter Height",false },
  {0x0019,0x005c,"SIEMENS RA GEN",VR::UN,VM::VM1,"?",false },
  {0x0019,0x005e,"SIEMENS RA GEN",VR::UL,VM::VM1,"Collimation Data Valid",false },
  {0x0019,0x0060,"SIEMENS RA GEN",VR::UL,VM::VM1,"PERI Sequence Number",false },
  {0x0019,0x0062,"SIEMENS RA GEN",VR::UL,VM::VM1,"PERI Total Scenes",false },
  {0x0019,0x0064,"SIEMENS RA GEN",VR::US,VM::VM1,"PERI Overlap Top",false },
  {0x0019,0x0066,"SIEMENS RA GEN",VR::UL,VM::VM1,"PERI Overlap Bottom",false },
  {0x0019,0x0068,"SIEMENS RA GEN",VR::UL,VM::VM1,"Raw Image Number",false },
  {0x0019,0x006a,"SIEMENS RA GEN",VR::UL,VM::VM1,"XRay Data Valid",false },
  {0x0019,0x0070,"SIEMENS RA GEN",VR::US,VM::VM1_n,"?",false },
  {0x0019,0x0072,"SIEMENS RA GEN",VR::US,VM::VM1_n,"?",false },
  {0x0019,0x0074,"SIEMENS RA GEN",VR::US,VM::VM1_n,"?",false },
  {0x0019,0x0076,"SIEMENS RA GEN",VR::UL,VM::VM1,"Filling Average Factor",false },
  {0x0019,0x0078,"SIEMENS RA GEN",VR::US,VM::VM1_n,"?",false },
  {0x0019,0x007a,"SIEMENS RA GEN",VR::US,VM::VM1_n,"?",false },
  {0x0019,0x007c,"SIEMENS RA GEN",VR::US,VM::VM1_n,"?",false },
  {0x0019,0x007e,"SIEMENS RA GEN",VR::US,VM::VM1_n,"?",false },
  {0x0019,0x0080,"SIEMENS RA GEN",VR::US,VM::VM1_n,"?",false },
  {0x0019,0x0082,"SIEMENS RA GEN",VR::US,VM::VM1_n,"?",false },
  {0x0019,0x0084,"SIEMENS RA GEN",VR::US,VM::VM1_n,"?",false },
  {0x0019,0x0086,"SIEMENS RA GEN",VR::US,VM::VM1_n,"?",false },
  {0x0019,0x0088,"SIEMENS RA GEN",VR::US,VM::VM1_n,"?",false },
  {0x0019,0x008a,"SIEMENS RA GEN",VR::US,VM::VM1_n,"?",false },
  {0x0019,0x008c,"SIEMENS RA GEN",VR::US,VM::VM1_n,"?",false },
  {0x0019,0x008e,"SIEMENS RA GEN",VR::US,VM::VM1_n,"?",false },
  {0x0019,0x0092,"SIEMENS RA GEN",VR::US,VM::VM1_n,"?",false },
  {0x0019,0x0094,"SIEMENS RA GEN",VR::US,VM::VM1_n,"?",false },
  {0x0019,0x0096,"SIEMENS RA GEN",VR::US,VM::VM1_n,"?",false },
  {0x0019,0x0098,"SIEMENS RA GEN",VR::US,VM::VM1_n,"?",false },
  {0x0019,0x009a,"SIEMENS RA GEN",VR::US,VM::VM1_n,"?",false },
  {0x0019,0x009c,"SIEMENS RA GEN",VR::UL,VM::VM1,"II Level CalibrationFactor",false },
  {0x0019,0x009e,"SIEMENS RA GEN",VR::UL,VM::VM1,"Nativ Review Flag",false },
  {0x0021,0x0015,"SIEMENS RA GEN",VR::US,VM::VM1,"Images in Study",false },
  {0x0021,0x0020,"SIEMENS RA GEN",VR::US,VM::VM1,"Scenes in Study",false },
  {0x0021,0x0025,"SIEMENS RA GEN",VR::US,VM::VM1,"Images in Photofile",false },
  {0x0021,0x0030,"SIEMENS RA GEN",VR::US,VM::VM1,"Images in All Scenes",false },
  {0x0021,0x0040,"SIEMENS RA GEN",VR::US,VM::VM1,"?",false },
  {0x0019,0x0020,"SIEMENS RA PLANE A",VR::UL,VM::VM1,"System Calibration Factor Plane A",false },
  {0x0019,0x0022,"SIEMENS RA PLANE A",VR::US,VM::VM1,"XRay Parameter Set Number",false },
  {0x0019,0x0024,"SIEMENS RA PLANE A",VR::US,VM::VM1,"XRay System",false },
  {0x0019,0x0026,"SIEMENS RA PLANE A",VR::US,VM::VM1,"?",false },
  {0x0019,0x0028,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Acquired Display Mode",false },
  {0x0019,0x002a,"SIEMENS RA PLANE A",VR::US,VM::VM1,"?",false },
  {0x0019,0x002c,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Acquisition Framerate",false },
  {0x0019,0x002e,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Maximum Frame Limit",false },
  {0x0019,0x0030,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Native Filter Type",false },
  {0x0019,0x0032,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Subtracted Filter Type",false },
  {0x0019,0x0034,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Filter Factor Native",false },
  {0x0019,0x0036,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Anatomic Blackground Factor",false },
  {0x0019,0x0038,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Window Upper Limit Native",false },
  {0x0019,0x003a,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Window Lower Limit Native",false },
  {0x0019,0x003c,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Window Brightness Phase 1 Sub",false },
  {0x0019,0x003e,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Window Brightness Phase 2 Sub",false },
  {0x0019,0x0040,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Window Contrast Phase 1 Sub",false },
  {0x0019,0x0042,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Window Contrast Phase 2 Sub",false },
  {0x0019,0x0044,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Filter Factor Subtraction",false },
  {0x0019,0x0046,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Peak Opacified",false },
  {0x0019,0x0048,"SIEMENS RA PLANE A",VR::UL,VM::VM1,"Mask Frame",false },
  {0x0019,0x004a,"SIEMENS RA PLANE A",VR::UL,VM::VM1,"BIH Frame",false },
  {0x0019,0x004c,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Center Beam Angulation CaudCran",false },
  {0x0019,0x004e,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Center Beam Angulation LR Anterior",false },
  {0x0019,0x0050,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Longitudinal Position",false },
  {0x0019,0x0052,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Side Offset",false },
  {0x0019,0x0054,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Isocenter Height",false },
  {0x0019,0x0056,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Image Twist",false },
  {0x0019,0x0058,"SIEMENS RA PLANE A",VR::US,VM::VM1,"SourceImage Distance",false },
  {0x0019,0x005a,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Mechanical Magnet Factor",false },
  {0x0019,0x005c,"SIEMENS RA PLANE A",VR::UL,VM::VM1,"Calibration Flag",false },
  {0x0019,0x005e,"SIEMENS RA PLANE A",VR::UL,VM::VM1,"Calibration Angle CranCaud",false },
  {0x0019,0x0060,"SIEMENS RA PLANE A",VR::UL,VM::VM1,"Calibration Angle RAOLAO",false },
  {0x0019,0x0062,"SIEMENS RA PLANE A",VR::UL,VM::VM1,"Calibration Table to Floor Distance",false },
  {0x0019,0x0064,"SIEMENS RA PLANE A",VR::UL,VM::VM1,"Calibration Isocenter to Floor Distance",false },
  {0x0019,0x0066,"SIEMENS RA PLANE A",VR::UL,VM::VM1,"Calibration Isocenter to Source Distance",false },
  {0x0019,0x0068,"SIEMENS RA PLANE A",VR::UL,VM::VM1,"Calibration Source to II Distance",false },
  {0x0019,0x006a,"SIEMENS RA PLANE A",VR::UL,VM::VM1,"Calibration II Zoom",false },
  {0x0019,0x006c,"SIEMENS RA PLANE A",VR::UL,VM::VM1,"Calibration II Field",false },
  {0x0019,0x006e,"SIEMENS RA PLANE A",VR::UL,VM::VM1,"Calibration Factor",false },
  {0x0019,0x0070,"SIEMENS RA PLANE A",VR::UL,VM::VM1,"Calibration Object to Image Distance",false },
  {0x0019,0x0072,"SIEMENS RA PLANE A",VR::UL,VM::VM1_n,"Calibration System Factor",false },
  {0x0019,0x0074,"SIEMENS RA PLANE A",VR::UL,VM::VM1_n,"Calibration System Correction",false },
  {0x0019,0x0076,"SIEMENS RA PLANE A",VR::UL,VM::VM1_n,"Calibration System II Formats",false },
  {0x0019,0x0078,"SIEMENS RA PLANE A",VR::UL,VM::VM1,"Calibration Gantry Data Valid",false },
  {0x0019,0x007a,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Collimator Square Breadth",false },
  {0x0019,0x007c,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Collimator Square Height",false },
  {0x0019,0x007e,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Collimator Square Diameter",false },
  {0x0019,0x0080,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Collimator Finger Turn Angle",false },
  {0x0019,0x0082,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Collimator Finger Position",false },
  {0x0019,0x0084,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Collimator Diaphragm Turn Angle",false },
  {0x0019,0x0086,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Collimator Diagphragm Position 1",false },
  {0x0019,0x0088,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Collimator Diaphragm Position 2",false },
  {0x0019,0x008a,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Collimator Diaphragm Mode",false },
  {0x0019,0x008c,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Collimator Beam Limit Breadth",false },
  {0x0019,0x008e,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Collimator Beam Limit Height",false },
  {0x0019,0x0090,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Collimator Beam Limit Diameter",false },
  {0x0019,0x0092,"SIEMENS RA PLANE A",VR::US,VM::VM1,"XRay Control Mode",false },
  {0x0019,0x0094,"SIEMENS RA PLANE A",VR::US,VM::VM1,"XRay System",false },
  {0x0019,0x0096,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Focal Spot",false },
  {0x0019,0x0098,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Exposure Control",false },
  {0x0019,0x009a,"SIEMENS RA PLANE A",VR::US,VM::VM1,"XRay Voltage",false },
  {0x0019,0x009c,"SIEMENS RA PLANE A",VR::US,VM::VM1,"XRay Current",false },
  {0x0019,0x009e,"SIEMENS RA PLANE A",VR::UL,VM::VM1,"XRay Current Time Product",false },
  {0x0019,0x00a0,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Pulse Time",false },
  {0x0019,0x00a2,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Scene Time Fluoro Clock",false },
  {0x0019,0x00a4,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Maximum PulseRate",false },
  {0x0019,0x00a6,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Pulses per Scene",false },
  {0x0019,0x00a8,"SIEMENS RA PLANE A",VR::UL,VM::VM1,"Dose Area Product of Scene",false },
  {0x0019,0x00aa,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Dose",false },
  {0x0019,0x00ac,"SIEMENS RA PLANE A",VR::US,VM::VM1,"Dose Rate",false },
  {0x0019,0x0020,"SIEMENS RA PLANE B",VR::UL,VM::VM1,"System Calibration Factor Plane B",false },
  {0x0019,0x0022,"SIEMENS RA PLANE B",VR::US,VM::VM1,"XRay Parameter Set Number",false },
  {0x0019,0x0024,"SIEMENS RA PLANE B",VR::US,VM::VM1,"XRay System",false },
  {0x0019,0x0026,"SIEMENS RA PLANE B",VR::US,VM::VM1,"?",false },
  {0x0019,0x0028,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Acquired Display Mode",false },
  {0x0019,0x002a,"SIEMENS RA PLANE B",VR::US,VM::VM1,"?",false },
  {0x0019,0x002c,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Acquisition Framerate",false },
  {0x0019,0x002e,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Maximum Frame Limit",false },
  {0x0019,0x0030,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Native Filter Type",false },
  {0x0019,0x0032,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Subtracted Filter Type",false },
  {0x0019,0x0034,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Filter Factor Native",false },
  {0x0019,0x0036,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Anatomic Blackground Factor",false },
  {0x0019,0x0038,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Window Upper Limit Native",false },
  {0x0019,0x003a,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Window Lower Limit Native",false },
  {0x0019,0x003c,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Window Brightness Phase 1 Sub",false },
  {0x0019,0x003e,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Window Brightness Phase 2 Sub",false },
  {0x0019,0x0040,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Window Contrast Phase 1 Sub",false },
  {0x0019,0x0042,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Window Contrast Phase 2 Sub",false },
  {0x0019,0x0044,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Filter Factor Subtraction",false },
  {0x0019,0x0046,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Peak Opacified",false },
  {0x0019,0x0048,"SIEMENS RA PLANE B",VR::UL,VM::VM1,"Mask Frame",false },
  {0x0019,0x004a,"SIEMENS RA PLANE B",VR::UL,VM::VM1,"BIH Frame",false },
  {0x0019,0x004c,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Center Beam Angulation CaudCran",false },
  {0x0019,0x004e,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Center Beam Angulation LR Anterior",false },
  {0x0019,0x0050,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Longitudinal Position",false },
  {0x0019,0x0052,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Side Offset",false },
  {0x0019,0x0054,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Isocenter Height",false },
  {0x0019,0x0056,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Image Twist",false },
  {0x0019,0x0058,"SIEMENS RA PLANE B",VR::US,VM::VM1,"SourceImage Distance",false },
  {0x0019,0x005a,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Mechanical Magnet Factor",false },
  {0x0019,0x005c,"SIEMENS RA PLANE B",VR::UL,VM::VM1,"Calibration Flag",false },
  {0x0019,0x005e,"SIEMENS RA PLANE B",VR::UL,VM::VM1,"Calibration Angle CranCaud",false },
  {0x0019,0x0060,"SIEMENS RA PLANE B",VR::UL,VM::VM1,"Calibration Angle RAOLAO",false },
  {0x0019,0x0062,"SIEMENS RA PLANE B",VR::UL,VM::VM1,"Calibration Table to Floor Distance",false },
  {0x0019,0x0064,"SIEMENS RA PLANE B",VR::UL,VM::VM1,"Calibration Isocenter to Floor Distance",false },
  {0x0019,0x0066,"SIEMENS RA PLANE B",VR::UL,VM::VM1,"Calibration Isocenter to Source Distance",false },
  {0x0019,0x0068,"SIEMENS RA PLANE B",VR::UL,VM::VM1,"Calibration Source to II Distance",false },
  {0x0019,0x006a,"SIEMENS RA PLANE B",VR::UL,VM::VM1,"Calibration II Zoom",false },
  {0x0019,0x006c,"SIEMENS RA PLANE B",VR::UL,VM::VM1,"Calibration II Field",false },
  {0x0019,0x006e,"SIEMENS RA PLANE B",VR::UL,VM::VM1,"Calibration Factor",false },
  {0x0019,0x0070,"SIEMENS RA PLANE B",VR::UL,VM::VM1,"Calibration Object to Image Distance",false },
  {0x0019,0x0072,"SIEMENS RA PLANE B",VR::UL,VM::VM1_n,"Calibration System Factor",false },
  {0x0019,0x0074,"SIEMENS RA PLANE B",VR::UL,VM::VM1_n,"Calibration System Correction",false },
  {0x0019,0x0076,"SIEMENS RA PLANE B",VR::UL,VM::VM1_n,"Calibration System II Formats",false },
  {0x0019,0x0078,"SIEMENS RA PLANE B",VR::UL,VM::VM1,"Calibration Gantry Data Valid",false },
  {0x0019,0x007a,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Collimator Square Breadth",false },
  {0x0019,0x007c,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Collimator Square Height",false },
  {0x0019,0x007e,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Collimator Square Diameter",false },
  {0x0019,0x0080,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Collimator Finger Turn Angle",false },
  {0x0019,0x0082,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Collimator Finger Position",false },
  {0x0019,0x0084,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Collimator Diaphragm Turn Angle",false },
  {0x0019,0x0086,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Collimator Diagphragm Position 1",false },
  {0x0019,0x0088,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Collimator Diaphragm Position 2",false },
  {0x0019,0x008a,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Collimator Diaphragm Mode",false },
  {0x0019,0x008c,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Collimator Beam Limit Breadth",false },
  {0x0019,0x008e,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Collimator Beam Limit Height",false },
  {0x0019,0x0090,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Collimator Beam Limit Diameter",false },
  {0x0019,0x0092,"SIEMENS RA PLANE B",VR::US,VM::VM1,"XRay Control Mode",false },
  {0x0019,0x0094,"SIEMENS RA PLANE B",VR::US,VM::VM1,"XRay System",false },
  {0x0019,0x0096,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Focal Spot",false },
  {0x0019,0x0098,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Exposure Control",false },
  {0x0019,0x009a,"SIEMENS RA PLANE B",VR::US,VM::VM1,"XRay Voltage",false },
  {0x0019,0x009c,"SIEMENS RA PLANE B",VR::US,VM::VM1,"XRay Current",false },
  {0x0019,0x009e,"SIEMENS RA PLANE B",VR::UL,VM::VM1,"XRay Current Time Product",false },
  {0x0019,0x00a0,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Pulse Time",false },
  {0x0019,0x00a2,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Scene Time Fluoro Clock",false },
  {0x0019,0x00a4,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Maximum PulseRate",false },
  {0x0019,0x00a6,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Pulses per Scene",false },
  {0x0019,0x00a8,"SIEMENS RA PLANE B",VR::UL,VM::VM1,"Dose Area Product of Scene",false },
  {0x0019,0x00aa,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Dose",false },
  {0x0019,0x00ac,"SIEMENS RA PLANE B",VR::US,VM::VM1,"Dose Rate",false },
  {0x0011,0x0010,"SIEMENS RIS",VR::LO,VM::VM1,"Patient UID",false },
  {0x0011,0x0011,"SIEMENS RIS",VR::LO,VM::VM1,"Patient ID",false },
  {0x0011,0x0020,"SIEMENS RIS",VR::DA,VM::VM1,"Patient Registration Date",false },
  {0x0011,0x0021,"SIEMENS RIS",VR::TM,VM::VM1,"Patient Registration Time",false },
  {0x0011,0x0030,"SIEMENS RIS",VR::PN,VM::VM1,"Patient NameRIS",false },
  {0x0011,0x0031,"SIEMENS RIS",VR::PN,VM::VM1,"Patient Prename",false },
  {0x0011,0x0040,"SIEMENS RIS",VR::LO,VM::VM1,"Patient Hospital Status",false },
  {0x0011,0x0041,"SIEMENS RIS",VR::LO,VM::VM1,"Medical Alerts",false },
  {0x0011,0x0042,"SIEMENS RIS",VR::LO,VM::VM1,"Contrast Allergy",false },
  {0x0031,0x0010,"SIEMENS RIS",VR::LO,VM::VM1,"Request UID",false },
  {0x0031,0x0045,"SIEMENS RIS",VR::LO,VM::VM1,"Requesting Physician",false },
  {0x0031,0x0050,"SIEMENS RIS",VR::LO,VM::VM1,"Requested Physician",false },
  {0x0033,0x0010,"SIEMENS RIS",VR::LO,VM::VM1,"Patient Study UID",false },
  {0x0019,0x0001,"SIEMENS SIENET",VR::DS,VM::VM1,"?",false },
  {0x0021,0x0000,"SIEMENS SMS-AX  ACQ 1.0",VR::US,VM::VM1,"Acquisition Type",false },
  {0x0021,0x0001,"SIEMENS SMS-AX  ACQ 1.0",VR::US,VM::VM1,"Acquisition Mode",false },
  {0x0021,0x0002,"SIEMENS SMS-AX  ACQ 1.0",VR::US,VM::VM1,"Footswitch Index",false },
  {0x0021,0x0003,"SIEMENS SMS-AX  ACQ 1.0",VR::US,VM::VM1,"Acquisition Room",false },
  {0x0021,0x0004,"SIEMENS SMS-AX  ACQ 1.0",VR::SL,VM::VM1,"Current Time Product",false },
  {0x0021,0x0005,"SIEMENS SMS-AX  ACQ 1.0",VR::SL,VM::VM1,"Dose",false },
  {0x0021,0x0006,"SIEMENS SMS-AX  ACQ 1.0",VR::SL,VM::VM1,"Skin Dose Percent",false },
  {0x0021,0x0007,"SIEMENS SMS-AX  ACQ 1.0",VR::SL,VM::VM1,"Skin Dose Accumulation",false },
  {0x0021,0x0008,"SIEMENS SMS-AX  ACQ 1.0",VR::SL,VM::VM1,"Skin Dose Rate",false },
  {0x0021,0x0009,"SIEMENS SMS-AX  ACQ 1.0",VR::UL,VM::VM1,"Impac Filename",false },
  {0x0021,0x000a,"SIEMENS SMS-AX  ACQ 1.0",VR::UL,VM::VM1,"Copper Filter",false },
  {0x0021,0x000b,"SIEMENS SMS-AX  ACQ 1.0",VR::US,VM::VM1,"Measuring Field",false },
  {0x0021,0x000c,"SIEMENS SMS-AX  ACQ 1.0",VR::SS,VM::VM3,"Post Blanking Circle",false },
  {0x0021,0x000d,"SIEMENS SMS-AX  ACQ 1.0",VR::SS,VM::VM2_2n,"Dyna Angles",false },
  {0x0021,0x000e,"SIEMENS SMS-AX  ACQ 1.0",VR::SS,VM::VM1,"Total Steps",false },
  {0x0021,0x000f,"SIEMENS SMS-AX  ACQ 1.0",VR::SL,VM::VM3_3n,"Dyna X-Ray Info",false },
  {0x0021,0x0010,"SIEMENS SMS-AX  ACQ 1.0",VR::US,VM::VM1,"Modality LUT Input Gamma",false },
  {0x0021,0x0011,"SIEMENS SMS-AX  ACQ 1.0",VR::US,VM::VM1,"Modality LUT Output Gamma",false },
  {0x0021,0x0012,"SIEMENS SMS-AX  ACQ 1.0",VR::OB,VM::VM1_n,"SH_STPAR",false },
  {0x0021,0x0013,"SIEMENS SMS-AX  ACQ 1.0",VR::US,VM::VM1,"Acquisition Zoom",false },
  {0x0021,0x0014,"SIEMENS SMS-AX  ACQ 1.0",VR::SS,VM::VM1,"Dyna Angulation Step Width",false },
  {0x0021,0x0015,"SIEMENS SMS-AX  ACQ 1.0",VR::US,VM::VM1,"Harmonization",false },
  {0x0021,0x0016,"SIEMENS SMS-AX  ACQ 1.0",VR::US,VM::VM1,"DR Single Flag",false },
  {0x0021,0x0017,"SIEMENS SMS-AX  ACQ 1.0",VR::SL,VM::VM1,"Source to Isocenter",false },
  {0x0021,0x0018,"SIEMENS SMS-AX  ACQ 1.0",VR::US,VM::VM1,"Pressure Data",false },
  {0x0021,0x0019,"SIEMENS SMS-AX  ACQ 1.0",VR::SL,VM::VM1,"ECG Index Array",false },
  {0x0021,0x001a,"SIEMENS SMS-AX  ACQ 1.0",VR::US,VM::VM1,"FD Flag",false },
  {0x0021,0x001b,"SIEMENS SMS-AX  ACQ 1.0",VR::OB,VM::VM1,"SH_ZOOM",false },
  {0x0021,0x001c,"SIEMENS SMS-AX  ACQ 1.0",VR::OB,VM::VM1,"SH_COLPAR",false },
  {0x0021,0x001d,"SIEMENS SMS-AX  ACQ 1.0",VR::US,VM::VM1,"K-Factor",false },
  {0x0021,0x001e,"SIEMENS SMS-AX  ACQ 1.0",VR::US,VM::VM8,"EVE",false },
  {0x0021,0x001f,"SIEMENS SMS-AX  ACQ 1.0",VR::SL,VM::VM1,"Total Scene Time",false },
  {0x0021,0x0020,"SIEMENS SMS-AX  ACQ 1.0",VR::US,VM::VM1,"Restore Flag",false },
  {0x0021,0x0021,"SIEMENS SMS-AX  ACQ 1.0",VR::US,VM::VM1,"Stand Movement Flag",false },
  {0x0021,0x0022,"SIEMENS SMS-AX  ACQ 1.0",VR::US,VM::VM1,"FD Rows",false },
  {0x0021,0x0023,"SIEMENS SMS-AX  ACQ 1.0",VR::US,VM::VM1,"FD Columns",false },
  {0x0021,0x0024,"SIEMENS SMS-AX  ACQ 1.0",VR::US,VM::VM1,"Table Movement Flag",false },
  {0x0021,0x0025,"SIEMENS SMS-AX  ACQ 1.0",VR::LO,VM::VM1,"Original Organ Program Name",false },
  {0x0021,0x0026,"SIEMENS SMS-AX  ACQ 1.0",VR::DS,VM::VM1,"Crispy XPI Filter",false },
  {0x0021,0x0028,"SIEMENS SMS-AX  ACQ 1.0",VR::SQ,VM::VM1,"?",false },
  {0x0021,0x0030,"SIEMENS SMS-AX  ACQ 1.0",VR::OB,VM::VM1,"?",false },
  {0x0021,0x0040,"SIEMENS SMS-AX  ACQ 1.0",VR::US,VM::VM3,"?",false },
  {0x0021,0x0041,"SIEMENS SMS-AX  ACQ 1.0",VR::LO,VM::VM1,"?",false },
  {0x0021,0x0042,"SIEMENS SMS-AX  ACQ 1.0",VR::US,VM::VM1_n,"?",false },
  {0x0021,0x0057,"SIEMENS SMS-AX  ACQ 1.0",VR::SS,VM::VM1_3,"?",false },
  {0x0021,0x0058,"SIEMENS SMS-AX  ACQ 1.0",VR::DS,VM::VM1,"?",false },
  {0x0021,0x0071,"SIEMENS SMS-AX  ACQ 1.0",VR::DS,VM::VM1,"?",false },
  {0x0025,0x0000,"SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0",VR::US,VM::VM1,"View Native",false },
  {0x0025,0x0001,"SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0",VR::US,VM::VM1,"Original Series Number",false },
  {0x0025,0x0002,"SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0",VR::US,VM::VM1,"Original Image Number",false },
  {0x0025,0x0003,"SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0",VR::US,VM::VM1,"Win Center",false },
  {0x0025,0x0004,"SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0",VR::US,VM::VM1,"Win Width",false },
  {0x0025,0x0005,"SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0",VR::US,VM::VM1,"Win Brightness",false },
  {0x0025,0x0006,"SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0",VR::US,VM::VM1,"Win Contrast",false },
  {0x0025,0x0007,"SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0",VR::US,VM::VM1,"Original Frame Number",false },
  {0x0025,0x0008,"SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0",VR::US,VM::VM1,"Original Mask Frame Number",false },
  {0x0025,0x0009,"SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0",VR::US,VM::VM1,"Opac",false },
  {0x0025,0x000a,"SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0",VR::US,VM::VM1,"Original Number of Frames",false },
  {0x0025,0x000b,"SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0",VR::DS,VM::VM1,"Original Scene Duration",false },
  {0x0025,0x000c,"SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0",VR::LO,VM::VM1,"Identifier LOID",false },
  {0x0025,0x000d,"SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0",VR::SS,VM::VM1_n,"Original Scene VFR Info",false },
  {0x0025,0x000e,"SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0",VR::SS,VM::VM1,"Original Frame ECG Position",false },
  {0x0025,0x000f,"SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0",VR::SS,VM::VM1,"Original ECG 1st Frame Offset (retired)",false },
  {0x0025,0x0010,"SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0",VR::SS,VM::VM1,"Zoom Flag",false },
  {0x0025,0x0011,"SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0",VR::US,VM::VM1,"Flex",false },
  {0x0025,0x0012,"SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0",VR::US,VM::VM1,"Number of Mask Frames",false },
  {0x0025,0x0013,"SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0",VR::US,VM::VM1,"Number of Fill Frames",false },
  {0x0025,0x0014,"SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0",VR::IS,VM::VM1,"Series Number",false },
  {0x0025,0x0015,"SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0",VR::IS,VM::VM1,"Image Number",false },
  {0x0025,0x0016,"SIEMENS SMS-AX  ORIGINAL IMAGE INFO 1.0",VR::IS,VM::VM1,"?",false },
  {0x0023,0x0000,"SIEMENS SMS-AX  QUANT 1.0",VR::DS,VM::VM2,"Horizontal Calibration Pixel Size",false },
  {0x0023,0x0001,"SIEMENS SMS-AX  QUANT 1.0",VR::DS,VM::VM2,"Vertical Calibration Pixel Size",false },
  {0x0023,0x0002,"SIEMENS SMS-AX  QUANT 1.0",VR::LO,VM::VM1,"Calibration Object",false },
  {0x0023,0x0003,"SIEMENS SMS-AX  QUANT 1.0",VR::DS,VM::VM1,"Calibration Object Size",false },
  {0x0023,0x0004,"SIEMENS SMS-AX  QUANT 1.0",VR::LO,VM::VM1,"Calibration Method",false },
  {0x0023,0x0005,"SIEMENS SMS-AX  QUANT 1.0",VR::ST,VM::VM1,"Filename",false },
  {0x0023,0x0006,"SIEMENS SMS-AX  QUANT 1.0",VR::IS,VM::VM1,"Frame Number",false },
  {0x0023,0x0007,"SIEMENS SMS-AX  QUANT 1.0",VR::IS,VM::VM2,"Calibration Factor Multiplicity",false },
  {0x0023,0x0008,"SIEMENS SMS-AX  QUANT 1.0",VR::IS,VM::VM1,"Calibration TOD Value",false },
  {0x0019,0x0000,"SIEMENS SMS-AX  VIEW 1.0",VR::US,VM::VM1,"Review Mode",false },
  {0x0019,0x0001,"SIEMENS SMS-AX  VIEW 1.0",VR::US,VM::VM1,"Anatomical Background Percent",false },
  {0x0019,0x0002,"SIEMENS SMS-AX  VIEW 1.0",VR::US,VM::VM1,"Number of Phases",false },
  {0x0019,0x0003,"SIEMENS SMS-AX  VIEW 1.0",VR::US,VM::VM1,"Apply Anatomical Background",false },
  {0x0019,0x0004,"SIEMENS SMS-AX  VIEW 1.0",VR::SS,VM::VM4_4n,"Pixel Shift Array",false },
  {0x0019,0x0005,"SIEMENS SMS-AX  VIEW 1.0",VR::US,VM::VM1,"Brightness",false },
  {0x0019,0x0006,"SIEMENS SMS-AX  VIEW 1.0",VR::US,VM::VM1,"Contrast",false },
  {0x0019,0x0007,"SIEMENS SMS-AX  VIEW 1.0",VR::US,VM::VM1,"Enabled",false },
  {0x0019,0x0008,"SIEMENS SMS-AX  VIEW 1.0",VR::US,VM::VM1,"Native Edge Enh. Percent Gain",false },
  {0x0019,0x0009,"SIEMENS SMS-AX  VIEW 1.0",VR::SS,VM::VM1,"Native Edge Enh. LUT Index",false },
  {0x0019,0x000a,"SIEMENS SMS-AX  VIEW 1.0",VR::SS,VM::VM1,"Native Edge Enh. Kernel Size",false },
  {0x0019,0x000b,"SIEMENS SMS-AX  VIEW 1.0",VR::US,VM::VM1,"Subtr. Edge Enh. Percent Gain",false },
  {0x0019,0x000c,"SIEMENS SMS-AX  VIEW 1.0",VR::SS,VM::VM1,"Subtr. Edge Enh. LUT Index",false },
  {0x0019,0x000d,"SIEMENS SMS-AX  VIEW 1.0",VR::SS,VM::VM1,"Subtr. Edge Enh. Kernel Size",false },
  {0x0019,0x000e,"SIEMENS SMS-AX  VIEW 1.0",VR::US,VM::VM1,"Fade Percent",false },
  {0x0019,0x000f,"SIEMENS SMS-AX  VIEW 1.0",VR::US,VM::VM1,"Flipped Before Laterality Applied",false },
  {0x0019,0x0010,"SIEMENS SMS-AX  VIEW 1.0",VR::US,VM::VM1,"Apply Fade",false },
  {0x0019,0x0011,"SIEMENS SMS-AX  VIEW 1.0",VR::US,VM::VM1,"RefImages Taken",false },
  {0x0019,0x0012,"SIEMENS SMS-AX  VIEW 1.0",VR::US,VM::VM1,"Zoom",false },
  {0x0019,0x0013,"SIEMENS SMS-AX  VIEW 1.0",VR::SS,VM::VM1,"Pan X",false },
  {0x0019,0x0014,"SIEMENS SMS-AX  VIEW 1.0",VR::SS,VM::VM1,"Pan Y",false },
  {0x0019,0x0015,"SIEMENS SMS-AX  VIEW 1.0",VR::SS,VM::VM1,"Native Edge Enh. Adv Perc. Gain",false },
  {0x0019,0x0016,"SIEMENS SMS-AX  VIEW 1.0",VR::SS,VM::VM1,"Subtr. Edge Enh. Adv Perc. Gain",false },
  {0x0019,0x0017,"SIEMENS SMS-AX  VIEW 1.0",VR::US,VM::VM1,"Invert Flag",false },
  {0x0019,0x001a,"SIEMENS SMS-AX  VIEW 1.0",VR::OB,VM::VM1,"Quant 1K Overlay",false },
  {0x0019,0x001b,"SIEMENS SMS-AX  VIEW 1.0",VR::US,VM::VM1,"Original Resolution",false },
  {0x0019,0x001c,"SIEMENS SMS-AX  VIEW 1.0",VR::DS,VM::VM1,"Auto Window Center",false },
  {0x0019,0x001d,"SIEMENS SMS-AX  VIEW 1.0",VR::DS,VM::VM1,"Auto Window Width",false },
  {0x0019,0x0006,"SIEMENS Selma",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0007,"SIEMENS Selma",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0008,"SIEMENS Selma",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0026,"SIEMENS Selma",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0029,"SIEMENS Selma",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0030,"SIEMENS Selma",VR::US,VM::VM1,"?",false },
  {0x0019,0x0031,"SIEMENS Selma",VR::US,VM::VM1,"?",false },
  {0x0019,0x0032,"SIEMENS Selma",VR::US,VM::VM1,"?",false },
  {0x0019,0x0033,"SIEMENS Selma",VR::US,VM::VM1,"?",false },
  {0x0019,0x0034,"SIEMENS Selma",VR::US,VM::VM1,"?",false },
  {0x0019,0x0035,"SIEMENS Selma",VR::US,VM::VM1,"?",false },
  {0x0029,0x0062,"SIEMENS SYNGO ADVANCED PRESENTATION",VR::ST,VM::VM1,"Segmentation Display Data UID",false },
  {0x0029,0x0069,"SIEMENS SYNGO ADVANCED PRESENTATION",VR::ST,VM::VM1,"??",false },
  {0x0029,0x006a,"SIEMENS SYNGO ADVANCED PRESENTATION",VR::LO,VM::VM1,"??",false },
  {0x0029,0x0070,"SIEMENS SYNGO ADVANCED PRESENTATION",VR::DS,VM::VM3,"??",false },
  {0x0029,0x007a,"SIEMENS SYNGO ADVANCED PRESENTATION",VR::DS,VM::VM1,"??",false },
  {0x0029,0x007b,"SIEMENS SYNGO ADVANCED PRESENTATION",VR::FL,VM::VM1_n,"??",false },
  {0x0077,0x0010,"SIEMENS SYNGO EVIDENCE DOCUMENT DATA",VR::LO,VM::VM1,"Evidence Document Template Name",false },
  {0x0077,0x0011,"SIEMENS SYNGO EVIDENCE DOCUMENT DATA",VR::DS,VM::VM1,"Evidence Document Template Version",false },
  {0x0077,0x0020,"SIEMENS SYNGO EVIDENCE DOCUMENT DATA",VR::OB,VM::VM1,"Dictionary.ValueCollectionOfstringClinicalFindingEntry (UTF-16 XML)",false },
  {0x0077,0x0021,"SIEMENS SYNGO EVIDENCE DOCUMENT DATA",VR::OB,VM::VM1,"ArrayOfmetaDataEntry (syngo.Services.Reporting.EvidenceDocument.Core)",false },
  {0x0077,0x0030,"SIEMENS SYNGO EVIDENCE DOCUMENT DATA",VR::DS,VM::VM1,"Framework Version",false },
  {0x0077,0x0040,"SIEMENS SYNGO EVIDENCE DOCUMENT DATA",VR::OB,VM::VM1,"SerUid/SopCl/SopUid/StuUid (UTF-16 XML)",false },
  {0x0077,0x0050,"SIEMENS SYNGO EVIDENCE DOCUMENT DATA",VR::LO,VM::VM1,"?? UID ??",false },
  {0x0077,0x0060,"SIEMENS SYNGO EVIDENCE DOCUMENT DATA",VR::OB,VM::VM1,"ApplicationData (syngo.Services.Reporting.EvidenceDocument)",false },
  {0x0077,0x0070,"SIEMENS SYNGO EVIDENCE DOCUMENT DATA",VR::LO,VM::VM1,"??",false },
  {0x0077,0x0071,"SIEMENS SYNGO EVIDENCE DOCUMENT DATA",VR::LO,VM::VM1,"??",false },
  {0x0077,0x0072,"SIEMENS SYNGO EVIDENCE DOCUMENT DATA",VR::OB,VM::VM1,"supportedTemplates (XML)",false },
  {0x0077,0x0080,"SIEMENS SYNGO EVIDENCE DOCUMENT DATA",VR::OB,VM::VM1,"VolumeCatalog (syngo.Services.Reporting.EvidenceDocument)",false },
  {0x0071,0x0001,"SIEMENS WH SR 1.0",VR::LO,VM::VM1,"?",false },
  {0x0071,0x0002,"SIEMENS WH SR 1.0",VR::LO,VM::VM1,"?",false },
  {0x0017,0x000a,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::SS,VM::VM1,"Internal Data",false },
  {0x0017,0x000b,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::SS,VM::VM1,"Internal Data",false },
  {0x0017,0x000c,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::SS,VM::VM1,"Internal Data",false },
  {0x0017,0x000d,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::FL,VM::VM1,"Internal Data",false },
  {0x0017,0x000e,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::LO,VM::VM1,"Internal Data",false },
  {0x0017,0x000f,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::LO,VM::VM1,"Internal Data",false },
  {0x0017,0x0014,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x0016,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x0017,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x0018,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x0019,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x001a,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x001b,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x001c,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x001e,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x001f,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::SS,VM::VM1,"Internal Data",false },
  {0x0017,0x0020,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::SS,VM::VM1,"Internal Data",false },
  {0x0017,0x0021,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x0022,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x0023,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x0024,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x0025,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x0026,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x0027,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x0028,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::FL,VM::VM1,"Internal Data",false },
  {0x0017,0x0029,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::FL,VM::VM1,"Internal Data",false },
  {0x0017,0x0048,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::SS,VM::VM1,"Internal Data",false },
  {0x0017,0x0049,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::SS,VM::VM1,"Internal Data",false },
  {0x0017,0x004a,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::SS,VM::VM1,"Internal Data",false },
  {0x0017,0x004b,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::SS,VM::VM1,"Internal Data",false },
  {0x0017,0x004c,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::LO,VM::VM1,"Internal Data",false },
  {0x0017,0x004d,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::SS,VM::VM1,"Internal Data",false },
  {0x0017,0x004e,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::LO,VM::VM1,"Internal Data",false },
  {0x0017,0x004f,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::LO,VM::VM1,"Internal Data",false },
  {0x0017,0x0050,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::SS,VM::VM1,"Internal Data",false },
  {0x0017,0x0051,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::FL,VM::VM1,"Internal Data",false },
  {0x0017,0x0052,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::SS,VM::VM1,"Internal Data",false },
  {0x0017,0x0053,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::SS,VM::VM1,"Internal Data",false },
  {0x0017,0x0054,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::SS,VM::VM1,"Internal Data",false },
  {0x0017,0x0055,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::SS,VM::VM1,"Internal Data",false },
  {0x0017,0x005a,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::OW,VM::VM1,"Internal Data",false },
  {0x0017,0x005b,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::OW,VM::VM1,"Internal Data",false },
  {0x0017,0x005c,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::OW,VM::VM1,"Internal Data",false },
  {0x0017,0x0064,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x0066,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x0067,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x0068,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x0085,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x0086,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x0087,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x0088,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x0089,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x008a,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x008b,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x008c,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::FL,VM::VM1,"Internal Data",false },
  {0x0017,0x008d,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::FL,VM::VM1,"Internal Data",false },
  {0x0017,0x008e,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x008f,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x00a0,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x00a1,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x00a2,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x00a3,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x00a4,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x00a5,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x00a6,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::US,VM::VM1,"Internal Data",false },
  {0x0017,0x00b0,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::DS,VM::VM1,"Internal Data",false },
  {0x0017,0x00c0,"SIEMENS_FLCOMPACT_VA01A_PROC",VR::LO,VM::VM1,"Internal Data",false },
  {0x0009,0x0001,"SIENET",VR::US,VM::VM1,"SIENET Command Field",false },
  {0x0009,0x0014,"SIENET",VR::LO,VM::VM1,"Receiver PLA",false },
  {0x0009,0x0016,"SIENET",VR::US,VM::VM1,"Transfer Priority",false },
  {0x0009,0x0029,"SIENET",VR::LO,VM::VM1,"Actual User",false },
  {0x0009,0x0070,"SIENET",VR::DS,VM::VM1,"?",false },
  {0x0009,0x0071,"SIENET",VR::DS,VM::VM1,"?",false },
  {0x0009,0x0072,"SIENET",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0073,"SIENET",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0074,"SIENET",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0075,"SIENET",VR::LO,VM::VM1,"?",false },
  {0x0091,0x0020,"SIENET",VR::PN,VM::VM1_n,"Patient Name",false },
  {0x0095,0x0001,"SIENET",VR::LO,VM::VM1,"Examination Folder ID",false },
  {0x0095,0x0004,"SIENET",VR::SL,VM::VM1,"Folder Reported Status",false },
  {0x0095,0x0005,"SIENET",VR::LO,VM::VM1,"Folder Reporting Radiologist",false },
  {0x0095,0x0007,"SIENET",VR::LO,VM::VM1,"SIENET ISA PLA",false },
  {0x0095,0x000c,"SIENET",VR::SL,VM::VM1,"Folder Rebuild Status",false },
  {0x0095,0x0020,"SIENET",VR::UN,VM::VM1,"?",false },
  {0x0095,0x00fb,"SIENET",VR::ST,VM::VM1,"?",false },
  {0x0099,0x0002,"SIENET",VR::UL,VM::VM1,"Key Images",false },
  {0x0099,0x0005,"SIENET",VR::SL,VM::VM1,"Image Number",false },
  {0x0009,0x0010,"SPI",VR::LO,VM::VM1,"Comments",false },
  {0x0009,0x0015,"SPI",VR::LO,VM::VM1,"Unique Identifier",false },
  {0x0009,0x0040,"SPI",VR::US,VM::VM1,"Data Object Type",false },
  {0x0009,0x0041,"SPI",VR::SH,VM::VM1,"Data Object Subtype",false },
  {0x0011,0x0010,"SPI",VR::LO,VM::VM1,"Organ",false },
  {0x0011,0x0015,"SPI",VR::LO,VM::VM1,"Allergy Indication",false },
  {0x0011,0x0020,"SPI",VR::CS,VM::VM1,"Pregnancy",false },
  {0x0029,0x0060,"SPI",VR::LO,VM::VM1,"Compression Algorithm",false },
  {0x0009,0x0010,"SPI RELEASE 1",VR::LO,VM::VM1,"Comments",false },
  {0x0009,0x0015,"SPI RELEASE 1",VR::LO,VM::VM1,"Unique Identifier",false },
  {0x0009,0x0040,"SPI RELEASE 1",VR::US,VM::VM1,"Data Object Type",false },
  {0x0009,0x0041,"SPI RELEASE 1",VR::SH,VM::VM1,"Data Object Subtype",false },
  {0x0011,0x0010,"SPI RELEASE 1",VR::LO,VM::VM1,"Organ",false },
  {0x0011,0x0015,"SPI RELEASE 1",VR::LO,VM::VM1,"Allergy Indication",false },
//  {0x0011,0x0020,"SPI RELEASE 1",VR::CS,VM::VM1,"Pregnancy",false },
//  {0x0029,0x0060,"SPI RELEASE 1",VR::LO,VM::VM1,"Compression Algorithm",false },
  {0x0009,0x0008,"SPI Release 1",VR::CS,VM::VM1,"?",false },
  {0x0009,0x0010,"SPI Release 1",VR::LO,VM::VM1,"Comments",false },
  {0x0009,0x0015,"SPI Release 1",VR::LO,VM::VM1,"UID",false },
  {0x0009,0x0040,"SPI Release 1",VR::US,VM::VM1,"Data Object Type",false },
  {0x0009,0x0041,"SPI Release 1",VR::SH,VM::VM1,"Data Object Subtype",false },
  {0x0011,0x0010,"SPI Release 1",VR::LO,VM::VM1,"Organ",false },
  {0x0011,0x0015,"SPI Release 1",VR::LO,VM::VM1,"Allergy Indication",false },
  {0x0011,0x0020,"SPI Release 1",VR::LO,VM::VM1,"Pregnancy",false },
  {0x0029,0x0060,"SPI Release 1",VR::LO,VM::VM1,"Compression Algorithm",false },
  {0x0009,0x0000,"SPI-P Release 1",VR::LT,VM::VM1,"Data Object Recognition Code",false },
  {0x0009,0x0004,"SPI-P Release 1",VR::LO,VM::VM1_n,"Image Data Consistency",false },
  {0x0009,0x0008,"SPI-P Release 1",VR::US,VM::VM1,"?",false },
  {0x0009,0x0012,"SPI-P Release 1",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0015,"SPI-P Release 1",VR::LO,VM::VM1,"Unique Identifier",false },
  {0x0009,0x0016,"SPI-P Release 1",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0018,"SPI-P Release 1",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0021,"SPI-P Release 1",VR::LT,VM::VM1,"?",false },
  {0x0009,0x0022,"SPI-P Release 1",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0031,"SPI-P Release 1",VR::LO,VM::VM1,"PACS Unique Identifier",false },
  {0x0009,0x0034,"SPI-P Release 1",VR::LO,VM::VM1,"Cluster Unique Identifier",false },
  {0x0009,0x0038,"SPI-P Release 1",VR::LO,VM::VM1,"System Unique Identifier",false },
  {0x0009,0x0039,"SPI-P Release 1",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0040,"SPI-P Release 1",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0041,"SPI-P Release 1",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0051,"SPI-P Release 1",VR::LO,VM::VM1,"Study Unique Identifier",false },
  {0x0009,0x0061,"SPI-P Release 1",VR::LO,VM::VM1,"Series Unique Identifier",false },
  {0x0009,0x0091,"SPI-P Release 1",VR::LT,VM::VM1,"?",false },
  {0x0009,0x00a0,"SPI-P Release 1",VR::LT,VM::VM1,"?",false },
  {0x0009,0x00f2,"SPI-P Release 1",VR::LT,VM::VM1,"?",false },
  {0x0009,0x00f3,"SPI-P Release 1",VR::UN,VM::VM1,"?",false },
  {0x0009,0x00f4,"SPI-P Release 1",VR::LT,VM::VM1,"?",false },
  {0x0009,0x00f5,"SPI-P Release 1",VR::UN,VM::VM1,"?",false },
  {0x0009,0x00f7,"SPI-P Release 1",VR::LT,VM::VM1,"?",false },
  {0x0011,0x0010,"SPI-P Release 1",VR::LT,VM::VM1,"Patient Entry ID",false },
  {0x0011,0x0020,"SPI-P Release 1",VR::UN,VM::VM1,"?",false },
  {0x0011,0x0021,"SPI-P Release 1",VR::UN,VM::VM1,"?",false },
  {0x0011,0x0022,"SPI-P Release 1",VR::UN,VM::VM1,"?",false },
  {0x0011,0x0030,"SPI-P Release 1",VR::UN,VM::VM1,"?",false },
  {0x0011,0x0031,"SPI-P Release 1",VR::UN,VM::VM1,"?",false },
  {0x0011,0x0032,"SPI-P Release 1",VR::UN,VM::VM1,"?",false },
  {0x0019,0x0000,"SPI-P Release 1",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0001,"SPI-P Release 1",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0002,"SPI-P Release 1",VR::UN,VM::VM1,"?",false },
  {0x0019,0x0010,"SPI-P Release 1",VR::US,VM::VM1,"Mains Frequency",false },
  {0x0019,0x0025,"SPI-P Release 1",VR::LO,VM::VM1_n,"Original Pixel Data Quality",false },
  {0x0019,0x0030,"SPI-P Release 1",VR::US,VM::VM1,"ECG Triggering",false },
  {0x0019,0x0031,"SPI-P Release 1",VR::UN,VM::VM1,"ECG 1 Offset",false },
  {0x0019,0x0032,"SPI-P Release 1",VR::UN,VM::VM1,"ECG 2 Offset 1",false },
  {0x0019,0x0033,"SPI-P Release 1",VR::UN,VM::VM1,"ECG 2 Offset 2",false },
  {0x0019,0x0050,"SPI-P Release 1",VR::US,VM::VM1,"Video Scan Mode",false },
  {0x0019,0x0051,"SPI-P Release 1",VR::US,VM::VM1,"Video LineRate",false },
  {0x0019,0x0060,"SPI-P Release 1",VR::US,VM::VM1,"Xray Technique",false },
  {0x0019,0x0061,"SPI-P Release 1",VR::DS,VM::VM1,"Image Identifier Format",false },
  {0x0019,0x0062,"SPI-P Release 1",VR::US,VM::VM1,"Iris Diaphragm",false },
  {0x0019,0x0063,"SPI-P Release 1",VR::CS,VM::VM1,"Filter",false },
  {0x0019,0x0064,"SPI-P Release 1",VR::CS,VM::VM1,"Cine Parallel",false },
  {0x0019,0x0065,"SPI-P Release 1",VR::CS,VM::VM1,"Cine Master",false },
  {0x0019,0x0070,"SPI-P Release 1",VR::US,VM::VM1,"Exposure Channel",false },
  {0x0019,0x0071,"SPI-P Release 1",VR::UN,VM::VM1,"Exposure Channel First Image",false },
  {0x0019,0x0072,"SPI-P Release 1",VR::US,VM::VM1,"Processing Channel",false },
  {0x0019,0x0080,"SPI-P Release 1",VR::DS,VM::VM1,"Acquisition Delay",false },
  {0x0019,0x0081,"SPI-P Release 1",VR::UN,VM::VM1,"Relative Image Time",false },
  {0x0019,0x0090,"SPI-P Release 1",VR::CS,VM::VM1,"Video White Compression",false },
  {0x0019,0x00a0,"SPI-P Release 1",VR::US,VM::VM1,"Angulation",false },
  {0x0019,0x00a1,"SPI-P Release 1",VR::US,VM::VM1,"Rotation",false },
  {0x0021,0x0012,"SPI-P Release 1",VR::LT,VM::VM1,"Series Unique Identifier",false },
  {0x0021,0x0014,"SPI-P Release 1",VR::LT,VM::VM1,"?",false },
  {0x0029,0x0000,"SPI-P Release 1",VR::DS,VM::VM4,"?",false },
  {0x0029,0x0020,"SPI-P Release 1",VR::DS,VM::VM1,"Pixel Aspect Ratio",false },
  {0x0029,0x0025,"SPI-P Release 1",VR::LO,VM::VM1_n,"Processed Pixel Data Quality",false },
  {0x0029,0x0030,"SPI-P Release 1",VR::LT,VM::VM1,"?",false },
  {0x0029,0x0038,"SPI-P Release 1",VR::US,VM::VM1,"?",false },
  {0x0029,0x0060,"SPI-P Release 1",VR::LT,VM::VM1,"?",false },
  {0x0029,0x0061,"SPI-P Release 1",VR::LT,VM::VM1,"?",false },
  {0x0029,0x0067,"SPI-P Release 1",VR::LT,VM::VM1,"?",false },
  {0x0029,0x0070,"SPI-P Release 1",VR::LT,VM::VM1,"Window ID",false },
  {0x0029,0x0071,"SPI-P Release 1",VR::CS,VM::VM1,"Video Invert Subtracted",false },
  {0x0029,0x0072,"SPI-P Release 1",VR::CS,VM::VM1,"Video Invert Nonsubtracted",false },
  {0x0029,0x0077,"SPI-P Release 1",VR::CS,VM::VM1,"Window Select Status",false },
  {0x0029,0x0078,"SPI-P Release 1",VR::LT,VM::VM1,"ECG Display Printing ID",false },
  {0x0029,0x0079,"SPI-P Release 1",VR::CS,VM::VM1,"ECG Display Printing",false },
  {0x0029,0x007e,"SPI-P Release 1",VR::CS,VM::VM1,"ECG Display Printing Enable Status",false },
  {0x0029,0x007f,"SPI-P Release 1",VR::CS,VM::VM1,"ECG Display Printing Select Status",false },
  {0x0029,0x0080,"SPI-P Release 1",VR::LT,VM::VM1,"Physiological Display ID",false },
  {0x0029,0x0081,"SPI-P Release 1",VR::US,VM::VM1,"Preferred Physiological Channel Display",false },
  {0x0029,0x008e,"SPI-P Release 1",VR::CS,VM::VM1,"Physiological Display Enable Status",false },
  {0x0029,0x008f,"SPI-P Release 1",VR::CS,VM::VM1,"Physiological Display Select Status",false },
  {0x0029,0x0090,"SPI-P Release 1",VR::DS,VM::VM1_n,"?",false },
  {0x0029,0x0091,"SPI-P Release 1",VR::US,VM::VM1,"?",false },
  {0x0029,0x009f,"SPI-P Release 1",VR::CS,VM::VM1,"?",false },
  {0x0029,0x00a0,"SPI-P Release 1",VR::DS,VM::VM1_n,"?",false },
  {0x0029,0x00a1,"SPI-P Release 1",VR::US,VM::VM1,"?",false },
  {0x0029,0x00af,"SPI-P Release 1",VR::CS,VM::VM1,"?",false },
  {0x0029,0x00b0,"SPI-P Release 1",VR::DS,VM::VM1_n,"?",false },
  {0x0029,0x00b1,"SPI-P Release 1",VR::US,VM::VM1,"?",false },
  {0x0029,0x00bf,"SPI-P Release 1",VR::CS,VM::VM1,"?",false },
  {0x0029,0x00c0,"SPI-P Release 1",VR::LT,VM::VM1,"Functional Shutter ID",false },
  {0x0029,0x00c1,"SPI-P Release 1",VR::US,VM::VM1,"Field Of Shutter",false },
  {0x0029,0x00c5,"SPI-P Release 1",VR::LT,VM::VM1,"Field Of Shutter Rectangle",false },
  {0x0029,0x00ce,"SPI-P Release 1",VR::CS,VM::VM1,"Shutter Enable Status",false },
  {0x0029,0x00cf,"SPI-P Release 1",VR::CS,VM::VM1,"Shutter Select Status",false },
  {0x7fe1,0x0010,"SPI-P Release 1",VR::UN,VM::VM1,"Pixel Data",false },
  {0x0009,0x00c0,"SPI-P Release 1;1",VR::LT,VM::VM1,"?",false },
  {0x0009,0x00c1,"SPI-P Release 1;1",VR::LT,VM::VM1,"?",false },
  {0x0019,0x0000,"SPI-P Release 1;1",VR::UN,VM::VM1,"Physiological Data Type",false },
  {0x0019,0x0001,"SPI-P Release 1;1",VR::UN,VM::VM1,"Physiological Data Channel And Kind",false },
  {0x0019,0x0002,"SPI-P Release 1;1",VR::US,VM::VM1,"Sample Bits Allocated",false },
  {0x0019,0x0003,"SPI-P Release 1;1",VR::US,VM::VM1,"Sample Bits Stored",false },
  {0x0019,0x0004,"SPI-P Release 1;1",VR::US,VM::VM1,"Sample High Bit",false },
  {0x0019,0x0005,"SPI-P Release 1;1",VR::US,VM::VM1,"Sample Representation",false },
  {0x0019,0x0006,"SPI-P Release 1;1",VR::UN,VM::VM1,"Smallest Sample Value",false },
  {0x0019,0x0007,"SPI-P Release 1;1",VR::UN,VM::VM1,"Largest Sample Value",false },
  {0x0019,0x0008,"SPI-P Release 1;1",VR::UN,VM::VM1,"Number Of Samples",false },
  {0x0019,0x0009,"SPI-P Release 1;1",VR::UN,VM::VM1,"Sample Data",false },
  {0x0019,0x000a,"SPI-P Release 1;1",VR::UN,VM::VM1,"Sample Rate",false },
  {0x0019,0x0010,"SPI-P Release 1;1",VR::UN,VM::VM1,"Physiological Data Type 2",false },
  {0x0019,0x0011,"SPI-P Release 1;1",VR::UN,VM::VM1,"Physiological Data Channel And Kind 2",false },
  {0x0019,0x0012,"SPI-P Release 1;1",VR::US,VM::VM1,"Sample Bits Allocated 2",false },
  {0x0019,0x0013,"SPI-P Release 1;1",VR::US,VM::VM1,"Sample Bits Stored 2",false },
  {0x0019,0x0014,"SPI-P Release 1;1",VR::US,VM::VM1,"Sample High Bit 2",false },
  {0x0019,0x0015,"SPI-P Release 1;1",VR::US,VM::VM1,"Sample Representation 2",false },
  {0x0019,0x0016,"SPI-P Release 1;1",VR::UN,VM::VM1,"Smallest Sample Value 2",false },
  {0x0019,0x0017,"SPI-P Release 1;1",VR::UN,VM::VM1,"Largest Sample Value 2",false },
  {0x0019,0x0018,"SPI-P Release 1;1",VR::UN,VM::VM1,"Number Of Samples 2",false },
  {0x0019,0x0019,"SPI-P Release 1;1",VR::UN,VM::VM1,"Sample Data 2",false },
  {0x0019,0x001a,"SPI-P Release 1;1",VR::UN,VM::VM1,"Sample Rate 2",false },
  {0x0021,0x0001,"SPI-P Release 1;1",VR::LO,VM::VM1,"?",false },
  {0x0021,0x0041,"SPI-P Release 1;1",VR::LO,VM::VM1,"?",false },
  {0x0021,0x0042,"SPI-P Release 1;1",VR::US,VM::VM1,"?",false },
  {0x0021,0x0043,"SPI-P Release 1;1",VR::US,VM::VM1,"?",false },
  {0x0021,0x0044,"SPI-P Release 1;1",VR::DS,VM::VM1,"?",false },
  {0x0021,0x0045,"SPI-P Release 1;1",VR::LO,VM::VM1,"?",false },
  {0x0021,0x0050,"SPI-P Release 1;1",VR::US,VM::VM1,"?",false },
  {0x0021,0x0051,"SPI-P Release 1;1",VR::US,VM::VM1,"?",false },
  {0x0021,0x0052,"SPI-P Release 1;1",VR::US,VM::VM1,"?",false },
  {0x0021,0x0060,"SPI-P Release 1;1",VR::IS,VM::VM1_n,"?",false },
  {0x0021,0x0061,"SPI-P Release 1;1",VR::IS,VM::VM1_n,"?",false },
  {0x0021,0x0067,"SPI-P Release 1;1",VR::LO,VM::VM1_n,"?",false },
  {0x0021,0x006a,"SPI-P Release 1;1",VR::DS,VM::VM1_n,"?",false },
  {0x0021,0x006b,"SPI-P Release 1;1",VR::DS,VM::VM1_n,"?",false },
  {0x0021,0x006c,"SPI-P Release 1;1",VR::US,VM::VM1_n,"?",false },
  {0x0021,0x006d,"SPI-P Release 1;1",VR::DS,VM::VM1_n,"?",false },
  {0x0021,0x006e,"SPI-P Release 1;1",VR::US,VM::VM1_n,"?",false },
  {0x0021,0x006f,"SPI-P Release 1;1",VR::DS,VM::VM1_n,"?",false },
  {0x0029,0x0000,"SPI-P Release 1;1",VR::LT,VM::VM1,"Zoom ID",false },
  {0x0029,0x0001,"SPI-P Release 1;1",VR::DS,VM::VM1_n,"Zoom Rectangle",false },
  {0x0029,0x0003,"SPI-P Release 1;1",VR::DS,VM::VM1,"Zoom Factor",false },
  {0x0029,0x0004,"SPI-P Release 1;1",VR::US,VM::VM1,"Zoom Function",false },
  {0x0029,0x000e,"SPI-P Release 1;1",VR::CS,VM::VM1,"Zoom Enable Status",false },
  {0x0029,0x000f,"SPI-P Release 1;1",VR::CS,VM::VM1,"Zoom Select Status",false },
  {0x0029,0x0040,"SPI-P Release 1;1",VR::LT,VM::VM1,"Magnifying Glass ID",false },
  {0x0029,0x0041,"SPI-P Release 1;1",VR::DS,VM::VM1_n,"Magnifying Glass Rectangle",false },
  {0x0029,0x0043,"SPI-P Release 1;1",VR::DS,VM::VM1,"Magnifying Glass Factor",false },
  {0x0029,0x0044,"SPI-P Release 1;1",VR::US,VM::VM1,"Magnifying Glass Function",false },
  {0x0029,0x004e,"SPI-P Release 1;1",VR::CS,VM::VM1,"Magnifying Glass Enable Status",false },
  {0x0029,0x004f,"SPI-P Release 1;1",VR::CS,VM::VM1,"Magnifying Glass Select Status",false },
  {0x0029,0x0000,"SPI-P Release 1;2",VR::LT,VM::VM1,"Subtraction Mask ID",false },
  {0x0029,0x0004,"SPI-P Release 1;2",VR::UN,VM::VM1,"Masking Function",false },
  {0x0029,0x000c,"SPI-P Release 1;2",VR::UN,VM::VM1,"Proprietary Masking Parameters",false },
  {0x0029,0x001e,"SPI-P Release 1;2",VR::CS,VM::VM1,"Subtraction Mask Enable Status",false },
  {0x0029,0x001f,"SPI-P Release 1;2",VR::CS,VM::VM1,"Subtraction Mask Select Status",false },
  {0x0029,0x0000,"SPI-P Release 1;3",VR::LT,VM::VM1,"Image Enhancement ID",false },
  {0x0029,0x0001,"SPI-P Release 1;3",VR::LT,VM::VM1,"Image Enhancement",false },
  {0x0029,0x0002,"SPI-P Release 1;3",VR::LT,VM::VM1,"Convolution ID",false },
  {0x0029,0x0003,"SPI-P Release 1;3",VR::LT,VM::VM1,"Convolution Type",false },
  {0x0029,0x0004,"SPI-P Release 1;3",VR::LT,VM::VM1,"Convolution Kernel Size ID",false },
  {0x0029,0x0005,"SPI-P Release 1;3",VR::US,VM::VM2,"Convolution Kernel Size",false },
  {0x0029,0x0006,"SPI-P Release 1;3",VR::US,VM::VM1_n,"Convolution Kernel",false },
  {0x0029,0x000c,"SPI-P Release 1;3",VR::DS,VM::VM1,"Enhancement Gain",false },
  {0x0029,0x001e,"SPI-P Release 1;3",VR::CS,VM::VM1,"Image Enhancement Enable Status",false },
  {0x0029,0x001f,"SPI-P Release 1;3",VR::CS,VM::VM1,"Image Enhancement Select Status",false },
  {0x0011,0x0018,"SPI-P Release 2;1",VR::LT,VM::VM1,"?",false },
  {0x0023,0x000d,"SPI-P Release 2;1",VR::UI,VM::VM1,"?",false },
  {0x0023,0x000e,"SPI-P Release 2;1",VR::UI,VM::VM1,"?",false },
  {0x0019,0x0000,"SPI-P-CTBE Release 1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0002,"SPI-P-CTBE Release 1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0003,"SPI-P-CTBE Release 1",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0004,"SPI-P-CTBE Release 1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0005,"SPI-P-CTBE Release 1",VR::DS,VM::VM1,"?",false },
  {0x0019,0x000b,"SPI-P-CTBE Release 1",VR::DS,VM::VM1,"?",false },
  {0x0019,0x000c,"SPI-P-CTBE Release 1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0014,"SPI-P-CTBE Release 1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0018,"SPI-P-CTBE Release 1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0019,"SPI-P-CTBE Release 1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x001a,"SPI-P-CTBE Release 1",VR::DS,VM::VM1,"?",false },
  {0x0019,0x001b,"SPI-P-CTBE Release 1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x001c,"SPI-P-CTBE Release 1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x001d,"SPI-P-CTBE Release 1",VR::DS,VM::VM1,"?",false },
  {0x0021,0x0000,"SPI-P-CTBE-Private Release 1",VR::DS,VM::VM1,"?",false },
  {0x0009,0x0000,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0010,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0020,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0030,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0040,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0050,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0060,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0070,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0075,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0080,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0009,0x0090,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0008,"SPI-P-GV-CT Release 1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0009,"SPI-P-GV-CT Release 1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x000a,"SPI-P-GV-CT Release 1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0010,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0020,"SPI-P-GV-CT Release 1",VR::TM,VM::VM1,"?",false },
  {0x0019,0x0050,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0060,"SPI-P-GV-CT Release 1",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0061,"SPI-P-GV-CT Release 1",VR::US,VM::VM1,"?",false },
  {0x0019,0x0063,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0064,"SPI-P-GV-CT Release 1",VR::US,VM::VM1,"?",false },
  {0x0019,0x0065,"SPI-P-GV-CT Release 1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0070,"SPI-P-GV-CT Release 1",VR::LT,VM::VM1,"?",false },
  {0x0019,0x0080,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0081,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0090,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0019,0x00a0,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0019,0x00a1,"SPI-P-GV-CT Release 1",VR::US,VM::VM1,"?",false },
  {0x0019,0x00a2,"SPI-P-GV-CT Release 1",VR::US,VM::VM1,"?",false },
  {0x0019,0x00a3,"SPI-P-GV-CT Release 1",VR::US,VM::VM1,"?",false },
  {0x0019,0x00b0,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0019,0x00b1,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0021,0x0020,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0021,0x0030,"SPI-P-GV-CT Release 1",VR::DS,VM::VM1,"?",false },
  {0x0021,0x0040,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0021,0x0050,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0021,0x0060,"SPI-P-GV-CT Release 1",VR::DS,VM::VM1,"?",false },
  {0x0021,0x0070,"SPI-P-GV-CT Release 1",VR::DS,VM::VM1,"?",false },
  {0x0021,0x0080,"SPI-P-GV-CT Release 1",VR::DS,VM::VM1,"?",false },
  {0x0021,0x0090,"SPI-P-GV-CT Release 1",VR::DS,VM::VM1,"?",false },
  {0x0021,0x00a0,"SPI-P-GV-CT Release 1",VR::US,VM::VM1,"?",false },
  {0x0021,0x00a1,"SPI-P-GV-CT Release 1",VR::DS,VM::VM1,"?",false },
  {0x0021,0x00a2,"SPI-P-GV-CT Release 1",VR::DS,VM::VM1,"?",false },
  {0x0021,0x00a3,"SPI-P-GV-CT Release 1",VR::LT,VM::VM1,"?",false },
  {0x0021,0x00a4,"SPI-P-GV-CT Release 1",VR::LT,VM::VM1,"?",false },
  {0x0021,0x00b0,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0021,0x00c0,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0010,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0030,"SPI-P-GV-CT Release 1",VR::UL,VM::VM1,"?",false },
  {0x0029,0x0031,"SPI-P-GV-CT Release 1",VR::UL,VM::VM1,"?",false },
  {0x0029,0x0032,"SPI-P-GV-CT Release 1",VR::UL,VM::VM1,"?",false },
  {0x0029,0x0033,"SPI-P-GV-CT Release 1",VR::UL,VM::VM1,"?",false },
  {0x0029,0x0080,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0090,"SPI-P-GV-CT Release 1",VR::LO,VM::VM1,"?",false },
  {0x0029,0x00d0,"SPI-P-GV-CT Release 1",VR::IS,VM::VM1,"?",false },
  {0x0029,0x00d1,"SPI-P-GV-CT Release 1",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0010,"SPI-P-PCR Release 2",VR::US,VM::VM1,"Reader Mode",false },
  {0x0019,0x0020,"SPI-P-PCR Release 2",VR::ST,VM::VM1,"MRM Code",false },
  {0x0019,0x0030,"SPI-P-PCR Release 2",VR::US,VM::VM1,"?",false },
  {0x0019,0x0040,"SPI-P-PCR Release 2",VR::DS,VM::VM1,"Latitude",false },
  {0x0019,0x0060,"SPI-P-PCR Release 2",VR::ST,VM::VM1,"View Name",false },
  {0x0019,0x0080,"SPI-P-PCR Release 2",VR::US,VM::VM1,"?",false },
  {0x0019,0x00a1,"SPI-P-PCR Release 2",VR::ST,VM::VM1,"?",false },
  {0x0019,0x00a3,"SPI-P-PCR Release 2",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00a4,"SPI-P-PCR Release 2",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00a5,"SPI-P-PCR Release 2",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00a6,"SPI-P-PCR Release 2",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00a7,"SPI-P-PCR Release 2",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00a8,"SPI-P-PCR Release 2",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00a9,"SPI-P-PCR Release 2",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00aa,"SPI-P-PCR Release 2",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00ab,"SPI-P-PCR Release 2",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00ac,"SPI-P-PCR Release 2",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00ad,"SPI-P-PCR Release 2",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00ae,"SPI-P-PCR Release 2",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00af,"SPI-P-PCR Release 2",VR::ST,VM::VM1,"?",false },
  {0x0019,0x00b0,"SPI-P-PCR Release 2",VR::ST,VM::VM1,"?",false },
  {0x0019,0x00b1,"SPI-P-PCR Release 2",VR::ST,VM::VM1,"?",false },
  {0x0019,0x00b2,"SPI-P-PCR Release 2",VR::IS,VM::VM1,"?",false },
  {0x0019,0x00b3,"SPI-P-PCR Release 2",VR::IS,VM::VM1,"?",false },
  {0x0019,0x00b4,"SPI-P-PCR Release 2",VR::IS,VM::VM1,"?",false },
  {0x0019,0x00b5,"SPI-P-PCR Release 2",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00b6,"SPI-P-PCR Release 2",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00b7,"SPI-P-PCR Release 2",VR::ST,VM::VM1,"?",false },
  {0x0019,0x00b8,"SPI-P-PCR Release 2",VR::DS,VM::VM1,"?",false },
  {0x0019,0x00b9,"SPI-P-PCR Release 2",VR::ST,VM::VM1,"?",false },
  {0x0019,0x00ba,"SPI-P-PCR Release 2",VR::ST,VM::VM1,"?",false },
  {0x0021,0x0000,"SPI-P-Private-CWS Release 1",VR::LT,VM::VM1,"Window Of Images ID",false },
  {0x0021,0x0001,"SPI-P-Private-CWS Release 1",VR::CS,VM::VM1,"Window Of Images Type",false },
  {0x0021,0x0002,"SPI-P-Private-CWS Release 1",VR::IS,VM::VM1_n,"WindowOfImagesScope",false },
  {0x0019,0x0010,"SPI-P-Private-DCI Release 1",VR::UN,VM::VM1,"?",false },
  {0x0019,0x0011,"SPI-P-Private-DCI Release 1",VR::UN,VM::VM1,"?",false },
  {0x0019,0x0012,"SPI-P-Private-DCI Release 1",VR::UN,VM::VM1,"?",false },
  {0x0019,0x0013,"SPI-P-Private-DCI Release 1",VR::UN,VM::VM1,"?",false },
  {0x0019,0x0014,"SPI-P-Private-DCI Release 1",VR::UN,VM::VM1,"?",false },
  {0x0019,0x0015,"SPI-P-Private-DCI Release 1",VR::UN,VM::VM1,"?",false },
  {0x0019,0x0016,"SPI-P-Private-DCI Release 1",VR::UN,VM::VM1,"?",false },
  {0x0019,0x0017,"SPI-P-Private-DCI Release 1",VR::UN,VM::VM1,"?",false },
  {0x0019,0x0000,"SPI-P-Private-DiDi Release 1",VR::LT,VM::VM1,"Post Mode String",false },
  {0x0019,0x0001,"SPI-P-Private-DiDi Release 1",VR::LT,VM::VM1,"Post Data",false },
  {0x0019,0x0010,"SPI-P-Private-DiDi Release 1",VR::LT,VM::VM1,"Image Header",false },
  {0x0019,0x0020,"SPI-P-Private-DiDi Release 1",VR::LT,VM::VM1,"SCPreName",false },
  {0x0019,0x0021,"SPI-P-Private-DiDi Release 1",VR::IS,VM::VM1,"SCPostNo",false },
  {0x0019,0x0022,"SPI-P-Private-DiDi Release 1",VR::LT,VM::VM1,"SCSOPInstanceUID",false },
  {0x0019,0x0023,"SPI-P-Private-DiDi Release 1",VR::LT,VM::VM1,"SCCreationDate",false },
  {0x0019,0x0024,"SPI-P-Private-DiDi Release 1",VR::LT,VM::VM1,"SCProcessingCategory",false },
  {0x0019,0x0025,"SPI-P-Private-DiDi Release 1",VR::LT,VM::VM1,"SCPostMethod",false },
  {0x0019,0x0026,"SPI-P-Private-DiDi Release 1",VR::LT,VM::VM1,"SCPostString",false },
  {0x0019,0x0027,"SPI-P-Private-DiDi Release 1",VR::LT,VM::VM1,"SCPostString2",false },
  {0x0019,0x0028,"SPI-P-Private-DiDi Release 1",VR::LT,VM::VM1,"SCMode",false },
  {0x0019,0x0029,"SPI-P-Private-DiDi Release 1",VR::IS,VM::VM1,"SCStCommit",false },
  {0x0019,0x002a,"SPI-P-Private-DiDi Release 1",VR::IS,VM::VM1,"SCExported",false },
  {0x0019,0x002b,"SPI-P-Private-DiDi Release 1",VR::IS,VM::VM1,"SCPrinted",false },
  {0x0019,0x002c,"SPI-P-Private-DiDi Release 1",VR::LT,VM::VM1,"SCOperator",false },
  {0x0019,0x0040,"SPI-P-Private-DiDi Release 1",VR::LT,VM::VM1,"SCPreName",false },
  {0x0019,0x0041,"SPI-P-Private-DiDi Release 1",VR::IS,VM::VM1,"SCPreType",false },
  {0x0021,0x0040,"SPI-P-Private_CDS Release 1",VR::IS,VM::VM1,"?",false },
  {0x0029,0x0000,"SPI-P-Private_CDS Release 1",VR::UN,VM::VM1,"?",false },
  {0x0029,0x0010,"SPI-P-Private_CDS Release 1",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0030,"SPI-P-Private_ICS Release 1",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0031,"SPI-P-Private_ICS Release 1",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0007,"SPI-P-Private_ICS Release 1",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0008,"SPI-P-Private_ICS Release 1",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x000c,"SPI-P-Private_ICS Release 1",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x000d,"SPI-P-Private_ICS Release 1",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x000e,"SPI-P-Private_ICS Release 1",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x000f,"SPI-P-Private_ICS Release 1",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0010,"SPI-P-Private_ICS Release 1",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0012,"SPI-P-Private_ICS Release 1",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0018,"SPI-P-Private_ICS Release 1",VR::FD,VM::VM1,"?",false },
  {0x0029,0x001b,"SPI-P-Private_ICS Release 1",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x001c,"SPI-P-Private_ICS Release 1",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x001d,"SPI-P-Private_ICS Release 1",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x001e,"SPI-P-Private_ICS Release 1",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0020,"SPI-P-Private_ICS Release 1",VR::UN,VM::VM1,"?",false },
  {0x0029,0x0021,"SPI-P-Private_ICS Release 1",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x003d,"SPI-P-Private_ICS Release 1",VR::LO,VM::VM1,"?",false },
  {0x0029,0x003e,"SPI-P-Private_ICS Release 1",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0048,"SPI-P-Private_ICS Release 1",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0049,"SPI-P-Private_ICS Release 1",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x004c,"SPI-P-Private_ICS Release 1",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x004d,"SPI-P-Private_ICS Release 1",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x004e,"SPI-P-Private_ICS Release 1",VR::UN,VM::VM1,"?",false },
  {0x0029,0x004f,"SPI-P-Private_ICS Release 1",VR::UN,VM::VM1,"?",false },
  {0x0029,0x0050,"SPI-P-Private_ICS Release 1",VR::UN,VM::VM1,"?",false },
  {0x0029,0x0051,"SPI-P-Private_ICS Release 1",VR::UN,VM::VM1,"?",false },
  {0x0029,0x0066,"SPI-P-Private_ICS Release 1",VR::DS,VM::VM4,"?",false },
  {0x0029,0x0067,"SPI-P-Private_ICS Release 1",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0068,"SPI-P-Private_ICS Release 1",VR::US,VM::VM1,"?",false },
  {0x0029,0x006a,"SPI-P-Private_ICS Release 1",VR::LO,VM::VM1,"?",false },
  {0x0029,0x006b,"SPI-P-Private_ICS Release 1",VR::US,VM::VM1,"?",false },
  {0x0029,0x0072,"SPI-P-Private_ICS Release 1",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0091,"SPI-P-Private_ICS Release 1",VR::IS,VM::VM1,"?",false },
  {0x0029,0x0093,"SPI-P-Private_ICS Release 1",VR::LO,VM::VM1_n,"?",false },
  {0x0029,0x0094,"SPI-P-Private_ICS Release 1",VR::LO,VM::VM1_n,"?",false },
  {0x0029,0x0000,"SPI-P-Private_ICS Release 1;1",VR::SL,VM::VM1,"?",false },
  {0x0029,0x0005,"SPI-P-Private_ICS Release 1;1",VR::FL,VM::VM1,"?",false },
  {0x0029,0x0006,"SPI-P-Private_ICS Release 1;1",VR::FL,VM::VM1,"?",false },
  {0x0029,0x0020,"SPI-P-Private_ICS Release 1;1",VR::FL,VM::VM1,"?",false },
  {0x0029,0x0021,"SPI-P-Private_ICS Release 1;1",VR::FL,VM::VM1,"?",false },
  {0x0029,0x0022,"SPI-P-Private_ICS Release 1;1",VR::US,VM::VM1,"?",false },
  {0x0029,0x0023,"SPI-P-Private_ICS Release 1;1",VR::FD,VM::VM1,"?",false },
  {0x0029,0x0024,"SPI-P-Private_ICS Release 1;1",VR::FD,VM::VM1,"?",false },
  {0x0029,0x0025,"SPI-P-Private_ICS Release 1;1",VR::FD,VM::VM1,"?",false },
  {0x0029,0x0026,"SPI-P-Private_ICS Release 1;1",VR::FD,VM::VM1,"?",false },
  {0x0029,0x0030,"SPI-P-Private_ICS Release 1;1",VR::ST,VM::VM1,"?",false },
  {0x0029,0x0033,"SPI-P-Private_ICS Release 1;1",VR::SL,VM::VM1,"?",false },
  {0x0029,0x006b,"SPI-P-Private_ICS Release 1;1",VR::FD,VM::VM1,"?",false },
  {0x0029,0x008b,"SPI-P-Private_ICS Release 1;1",VR::FD,VM::VM1,"?",false },
  {0x0029,0x008c,"SPI-P-Private_ICS Release 1;1",VR::SL,VM::VM1,"?",false },
  {0x0029,0x008d,"SPI-P-Private_ICS Release 1;1",VR::FD,VM::VM1,"?",false },
  {0x0029,0x00c1,"SPI-P-Private_ICS Release 1;1",VR::US,VM::VM1,"?",false },
  {0x0029,0x00cb,"SPI-P-Private_ICS Release 1;1",VR::FD,VM::VM1,"?",false },
  {0x0029,0x00cc,"SPI-P-Private_ICS Release 1;1",VR::FD,VM::VM1,"?",false },
  {0x0029,0x00cd,"SPI-P-Private_ICS Release 1;1",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x00d4,"SPI-P-Private_ICS Release 1;1",VR::FL,VM::VM1,"?",false },
  {0x0029,0x00d5,"SPI-P-Private_ICS Release 1;1",VR::FL,VM::VM1,"?",false },
  {0x0029,0x00d6,"SPI-P-Private_ICS Release 1;1",VR::ST,VM::VM1,"?",false },
  {0x0029,0x0000,"SPI-P-Private_ICS Release 1;2",VR::FD,VM::VM1,"?",false },
  {0x0029,0x0001,"SPI-P-Private_ICS Release 1;2",VR::FD,VM::VM1,"?",false },
  {0x0029,0x0002,"SPI-P-Private_ICS Release 1;2",VR::FD,VM::VM1,"?",false },
  {0x0029,0x0003,"SPI-P-Private_ICS Release 1;2",VR::SL,VM::VM1,"?",false },
  {0x0029,0x0004,"SPI-P-Private_ICS Release 1;2",VR::SL,VM::VM1,"?",false },
  {0x0029,0x0005,"SPI-P-Private_ICS Release 1;2",VR::SL,VM::VM1,"?",false },
  {0x0029,0x0030,"SPI-P-Private_ICS Release 1;2",VR::LT,VM::VM1,"?",false },
  {0x0029,0x00a0,"SPI-P-Private_ICS Release 1;2",VR::SL,VM::VM1,"?",false },
  {0x0029,0x00a1,"SPI-P-Private_ICS Release 1;2",VR::FD,VM::VM1,"?",false },
  {0x0029,0x00a2,"SPI-P-Private_ICS Release 1;2",VR::SL,VM::VM1,"?",false },
  {0x0029,0x00a3,"SPI-P-Private_ICS Release 1;2",VR::SL,VM::VM1,"?",false },
  {0x0029,0x00a5,"SPI-P-Private_ICS Release 1;2",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x00a6,"SPI-P-Private_ICS Release 1;2",VR::SL,VM::VM1,"?",false },
  {0x0029,0x00d9,"SPI-P-Private_ICS Release 1;2",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0000,"SPI-P-Private_ICS Release 1;3",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0001,"SPI-P-Private_ICS Release 1;3",VR::ST,VM::VM1,"?",false },
  {0x0029,0x0060,"SPI-P-Private_ICS Release 1;3",VR::US,VM::VM1,"?",false },
  {0x0029,0x006a,"SPI-P-Private_ICS Release 1;3",VR::FL,VM::VM1,"?",false },
  {0x0029,0x00b0,"SPI-P-Private_ICS Release 1;3",VR::SL,VM::VM1,"?",false },
  {0x0029,0x00b1,"SPI-P-Private_ICS Release 1;3",VR::FL,VM::VM1,"?",false },
  {0x0029,0x00b2,"SPI-P-Private_ICS Release 1;3",VR::FL,VM::VM1,"?",false },
  {0x0029,0x00b4,"SPI-P-Private_ICS Release 1;3",VR::FL,VM::VM1,"?",false },
  {0x0029,0x00b9,"SPI-P-Private_ICS Release 1;3",VR::FL,VM::VM1,"?",false },
  {0x0029,0x00c0,"SPI-P-Private_ICS Release 1;3",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x00c1,"SPI-P-Private_ICS Release 1;3",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x00c2,"SPI-P-Private_ICS Release 1;3",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x00c3,"SPI-P-Private_ICS Release 1;3",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x00c4,"SPI-P-Private_ICS Release 1;3",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x00c5,"SPI-P-Private_ICS Release 1;3",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0000,"SPI-P-Private_ICS Release 1;4",VR::UL,VM::VM1,"?",false },
  {0x0029,0x0001,"SPI-P-Private_ICS Release 1;4",VR::FL,VM::VM1_n,"?",false },
  {0x0029,0x0002,"SPI-P-Private_ICS Release 1;4",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0006,"SPI-P-Private_ICS Release 1;4",VR::UL,VM::VM1,"?",false },
  {0x0029,0x0007,"SPI-P-Private_ICS Release 1;4",VR::SL,VM::VM1,"?",false },
  {0x0029,0x0008,"SPI-P-Private_ICS Release 1;4",VR::US,VM::VM1,"?",false },
  {0x0029,0x0009,"SPI-P-Private_ICS Release 1;4",VR::FL,VM::VM1,"?",false },
  {0x0029,0x000a,"SPI-P-Private_ICS Release 1;4",VR::FL,VM::VM1,"?",false },
  {0x0029,0x000b,"SPI-P-Private_ICS Release 1;4",VR::FL,VM::VM1,"?",false },
  {0x0029,0x000c,"SPI-P-Private_ICS Release 1;4",VR::FL,VM::VM1,"?",false },
  {0x0029,0x000e,"SPI-P-Private_ICS Release 1;4",VR::US,VM::VM1,"?",false },
  {0x0029,0x000f,"SPI-P-Private_ICS Release 1;4",VR::US,VM::VM1,"?",false },
  {0x0029,0x0010,"SPI-P-Private_ICS Release 1;4",VR::SL,VM::VM1,"?",false },
  {0x0029,0x0090,"SPI-P-Private_ICS Release 1;4",VR::SL,VM::VM1,"?",false },
  {0x0029,0x009a,"SPI-P-Private_ICS Release 1;4",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x009c,"SPI-P-Private_ICS Release 1;4",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x00a0,"SPI-P-Private_ICS Release 1;4",VR::US,VM::VM1,"?",false },
  {0x0029,0x00a1,"SPI-P-Private_ICS Release 1;4",VR::US,VM::VM1,"?",false },
  {0x0029,0x00b1,"SPI-P-Private_ICS Release 1;4",VR::US,VM::VM1,"?",false },
  {0x0029,0x00b2,"SPI-P-Private_ICS Release 1;4",VR::US,VM::VM1,"?",false },
  {0x0029,0x00b3,"SPI-P-Private_ICS Release 1;4",VR::SL,VM::VM1,"?",false },
  {0x0029,0x00b5,"SPI-P-Private_ICS Release 1;4",VR::FL,VM::VM1,"?",false },
  {0x0029,0x00b6,"SPI-P-Private_ICS Release 1;4",VR::SL,VM::VM1,"?",false },
  {0x0029,0x00bc,"SPI-P-Private_ICS Release 1;4",VR::SL,VM::VM1,"?",false },
  {0x0029,0x00bd,"SPI-P-Private_ICS Release 1;4",VR::SL,VM::VM1,"?",false },
  {0x0029,0x00be,"SPI-P-Private_ICS Release 1;4",VR::US,VM::VM1,"?",false },
  {0x0029,0x00d6,"SPI-P-Private_ICS Release 1;4",VR::FD,VM::VM1,"?",false },
  {0x0029,0x00d7,"SPI-P-Private_ICS Release 1;4",VR::FD,VM::VM1,"?",false },
  {0x0029,0x00d8,"SPI-P-Private_ICS Release 1;4",VR::UN,VM::VM1,"?",false },
  {0x0029,0x00d9,"SPI-P-Private_ICS Release 1;4",VR::UN,VM::VM1,"?",false },
  {0x0029,0x00da,"SPI-P-Private_ICS Release 1;4",VR::UN,VM::VM1,"?",false },
  {0x0029,0x00dc,"SPI-P-Private_ICS Release 1;4",VR::FD,VM::VM1,"?",false },
  {0x0029,0x00dd,"SPI-P-Private_ICS Release 1;4",VR::FD,VM::VM1,"?",false },
  {0x0029,0x00e0,"SPI-P-Private_ICS Release 1;4",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0050,"SPI-P-Private_ICS Release 1;5",VR::CS,VM::VM1,"?",false },
  {0x0029,0x0055,"SPI-P-Private_ICS Release 1;5",VR::CS,VM::VM1,"?",false },
  {0x0029,0x0001,"SPI-P-Private_ICS Release 1;6",VR::FL,VM::VM1,"?",false },
  {0x0029,0x0002,"SPI-P-Private_ICS Release 1;6",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0003,"SPI-P-Private_ICS Release 1;6",VR::UL,VM::VM1,"?",false },
  {0x0029,0x0004,"SPI-P-Private_ICS Release 1;6",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0006,"SPI-P-Private_ICS Release 1;6",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0007,"SPI-P-Private_ICS Release 1;6",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0008,"SPI-P-Private_ICS Release 1;6",VR::US,VM::VM1,"?",false },
  {0x0029,0x000a,"SPI-P-Private_ICS Release 1;6",VR::US,VM::VM1,"?",false },
  {0x0029,0x000b,"SPI-P-Private_ICS Release 1;6",VR::US,VM::VM1,"?",false },
  {0x0029,0x000c,"SPI-P-Private_ICS Release 1;6",VR::US,VM::VM1,"?",false },
  {0x0029,0x000d,"SPI-P-Private_ICS Release 1;6",VR::US,VM::VM1,"?",false },
  {0x0029,0x000f,"SPI-P-Private_ICS Release 1;6",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0010,"SPI-P-Private_ICS Release 1;6",VR::US,VM::VM1,"?",false },
  {0x0029,0x0014,"SPI-P-Private_ICS Release 1;6",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0015,"SPI-P-Private_ICS Release 1;6",VR::UL,VM::VM2,"?",false },
  {0x0029,0x0016,"SPI-P-Private_ICS Release 1;6",VR::FL,VM::VM1,"?",false },
  {0x0029,0x0017,"SPI-P-Private_ICS Release 1;6",VR::FL,VM::VM1,"?",false },
  {0x0029,0x001a,"SPI-P-Private_ICS Release 1;6",VR::UL,VM::VM1,"?",false },
  {0x0029,0x001d,"SPI-P-Private_ICS Release 1;6",VR::US,VM::VM1,"?",false },
  {0x0029,0x001f,"SPI-P-Private_ICS Release 1;6",VR::FL,VM::VM1,"?",false },
  {0x0029,0x0020,"SPI-P-Private_ICS Release 1;6",VR::FL,VM::VM1,"?",false },
  {0x0029,0x0021,"SPI-P-Private_ICS Release 1;6",VR::UL,VM::VM1,"?",false },
  {0x0029,0x0023,"SPI-P-Private_ICS Release 1;6",VR::US,VM::VM1,"?",false },
  {0x0029,0x0024,"SPI-P-Private_ICS Release 1;6",VR::UL,VM::VM1,"?",false },
  {0x0029,0x002a,"SPI-P-Private_ICS Release 1;6",VR::UL,VM::VM1,"?",false },
  {0x0029,0x002b,"SPI-P-Private_ICS Release 1;6",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x002c,"SPI-P-Private_ICS Release 1;6",VR::UL,VM::VM1,"?",false },
  {0x0029,0x002d,"SPI-P-Private_ICS Release 1;6",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x002e,"SPI-P-Private_ICS Release 1;6",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x002f,"SPI-P-Private_ICS Release 1;6",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0030,"SPI-P-Private_ICS Release 1;6",VR::US,VM::VM1,"?",false },
  {0x0029,0x0032,"SPI-P-Private_ICS Release 1;6",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0033,"SPI-P-Private_ICS Release 1;6",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0034,"SPI-P-Private_ICS Release 1;6",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0035,"SPI-P-Private_ICS Release 1;6",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0036,"SPI-P-Private_ICS Release 1;6",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0037,"SPI-P-Private_ICS Release 1;6",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0038,"SPI-P-Private_ICS Release 1;6",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0039,"SPI-P-Private_ICS Release 1;6",VR::FL,VM::VM1,"?",false },
  {0x0029,0x003d,"SPI-P-Private_ICS Release 1;6",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0043,"SPI-P-Private_ICS Release 1;6",VR::US,VM::VM1,"?",false },
  {0x0029,0x0044,"SPI-P-Private_ICS Release 1;6",VR::US,VM::VM1,"?",false },
  {0x0029,0x0048,"SPI-P-Private_ICS Release 1;6",VR::UL,VM::VM1,"?",false },
  {0x0029,0x004f,"SPI-P-Private_ICS Release 1;6",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x005c,"SPI-P-Private_ICS Release 1;6",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x005d,"SPI-P-Private_ICS Release 1;6",VR::UL,VM::VM1_n,"?",false },
  {0x0029,0x005e,"SPI-P-Private_ICS Release 1;6",VR::UL,VM::VM1,"?",false },
  {0x0029,0x005f,"SPI-P-Private_ICS Release 1;6",VR::US,VM::VM1,"?",false },
  {0x0029,0x0062,"SPI-P-Private_ICS Release 1;6",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0063,"SPI-P-Private_ICS Release 1;6",VR::US,VM::VM1,"?",false },
  {0x0029,0x0065,"SPI-P-Private_ICS Release 1;6",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0066,"SPI-P-Private_ICS Release 1;6",VR::UL,VM::VM1,"?",false },
  {0x0029,0x0067,"SPI-P-Private_ICS Release 1;6",VR::UL,VM::VM1,"?",false },
  {0x0029,0x0068,"SPI-P-Private_ICS Release 1;6",VR::UL,VM::VM1,"?",false },
  {0x0029,0x0069,"SPI-P-Private_ICS Release 1;6",VR::UL,VM::VM1,"?",false },
  {0x0029,0x006a,"SPI-P-Private_ICS Release 1;6",VR::US,VM::VM1,"?",false },
  {0x0029,0x006b,"SPI-P-Private_ICS Release 1;6",VR::US,VM::VM1,"?",false },
  {0x0029,0x006c,"SPI-P-Private_ICS Release 1;6",VR::FL,VM::VM1,"?",false },
  {0x0029,0x006d,"SPI-P-Private_ICS Release 1;6",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0011,"SPI-P-Private_ICS Release 1;8",VR::SQ,VM::VM1,"?",false },
  {0x0019,0x0010,"SPI-P-XSB-DCI Release 1",VR::LT,VM::VM1,"Video Beam Boost",false },
  {0x0019,0x0011,"SPI-P-XSB-DCI Release 1",VR::US,VM::VM1,"Channel Generating Video Sync",false },
  {0x0019,0x0012,"SPI-P-XSB-DCI Release 1",VR::US,VM::VM1,"Video Gain",false },
  {0x0019,0x0013,"SPI-P-XSB-DCI Release 1",VR::US,VM::VM1,"Video Offset",false },
  {0x0019,0x0020,"SPI-P-XSB-DCI Release 1",VR::DS,VM::VM1,"RTD Data Compression Factor",false },
  {0x0019,0x0000,"SPI-P-XSB-VISUB Release 1",VR::UN,VM::VM1,"?",false },
  {0x0019,0x0010,"SPI-P-XSB-VISUB Release 1",VR::UN,VM::VM1,"?",false },
  {0x0019,0x0011,"SPI-P-XSB-VISUB Release 1",VR::UN,VM::VM1,"?",false },
  {0x0019,0x0012,"SPI-P-XSB-VISUB Release 1",VR::UN,VM::VM1,"?",false },
  {0x0019,0x0020,"SPI-P-XSB-VISUB Release 1",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0040,"SPI-P-XSB-VISUB Release 1",VR::UN,VM::VM1,"?",false },
  {0x0019,0x0050,"SPI-P-XSB-VISUB Release 1",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0000,"SPI-P-XSB-VISUB Release 1",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0001,"SPI-P-XSB-VISUB Release 1",VR::DS,VM::VM1,"?",false },
  {0x0029,0x0002,"SPI-P-XSB-VISUB Release 1",VR::CS,VM::VM1,"?",false },
  {0x0029,0x000f,"SPI-P-XSB-VISUB Release 1",VR::CS,VM::VM1,"?",false },
  {0x0029,0x0010,"SPI-P-XSB-VISUB Release 1",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0011,"SPI-P-XSB-VISUB Release 1",VR::DS,VM::VM1,"?",false },
  {0x0029,0x0012,"SPI-P-XSB-VISUB Release 1",VR::CS,VM::VM1,"?",false },
  {0x0029,0x001f,"SPI-P-XSB-VISUB Release 1",VR::CS,VM::VM1,"?",false },
  {0x0029,0x0020,"SPI-P-XSB-VISUB Release 1",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0021,"SPI-P-XSB-VISUB Release 1",VR::DS,VM::VM1,"?",false },
  {0x0029,0x0022,"SPI-P-XSB-VISUB Release 1",VR::CS,VM::VM1,"?",false },
  {0x0029,0x002f,"SPI-P-XSB-VISUB Release 1",VR::CS,VM::VM1,"?",false },
  {0x0029,0x0030,"SPI-P-XSB-VISUB Release 1",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0031,"SPI-P-XSB-VISUB Release 1",VR::DS,VM::VM1,"?",false },
  {0x0029,0x0032,"SPI-P-XSB-VISUB Release 1",VR::CS,VM::VM1,"?",false },
  {0x0029,0x003f,"SPI-P-XSB-VISUB Release 1",VR::CS,VM::VM1,"?",false },
  {0x0073,0x0001,"STENTOR",VR::ST,VM::VM1,"?",false },
  {0x0073,0x0002,"STENTOR",VR::ST,VM::VM1,"Stentor Remote IP-Address Element",false },
  {0x0073,0x0003,"STENTOR",VR::ST,VM::VM1,"Stentor Remote AETitle Element",false },
  {0x0073,0x0004,"STENTOR",VR::ST,VM::VM1,"Stentor Local AETitle Element",false },
  {0x0073,0x0006,"STENTOR",VR::LO,VM::VM1,"Stentor Transfer Syntax Value",false },
  {0x0017,0x0000,"SVISION",VR::LO,VM::VM1,"Extended Body Part",false },
  {0x0017,0x0010,"SVISION",VR::LO,VM::VM1,"Extended View Position",false },
  {0x0017,0x0020,"SVISION",VR::SQ,VM::VM1,"Scheduled Procedure Step",false },
  {0x0017,0x00a0,"SVISION",VR::IS,VM::VM1,"Fixed Grid System",false },
  {0x0017,0x00f0,"SVISION",VR::IS,VM::VM1,"Images SOP Class",false },
  {0x0019,0x0000,"SVISION",VR::IS,VM::VM1,"AEC Field",false },
  {0x0019,0x0001,"SVISION",VR::IS,VM::VM1,"AEC Film Screen",false },
  {0x0019,0x0002,"SVISION",VR::IS,VM::VM1,"AEC Density",false },
  {0x0019,0x0010,"SVISION",VR::IS,VM::VM1,"Patient Thickness",false },
  {0x0019,0x0016,"SVISION",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0018,"SVISION",VR::IS,VM::VM1,"Beam Distance",false },
  {0x0019,0x0020,"SVISION",VR::IS,VM::VM1,"Workstation Number",false },
  {0x0019,0x0028,"SVISION",VR::IS,VM::VM1,"Tube Number",false },
  {0x0019,0x0030,"SVISION",VR::IS,VM::VM1,"Bucky Grid",false },
  {0x0019,0x0034,"SVISION",VR::IS,VM::VM1,"Focus",false },
  {0x0019,0x0038,"SVISION",VR::IS,VM::VM1,"Age Group",false },
  {0x0019,0x0040,"SVISION",VR::IS,VM::VM1,"Collimator Distance X",false },
  {0x0019,0x0041,"SVISION",VR::IS,VM::VM1,"Collimator Distance Y",false },
  {0x0019,0x0050,"SVISION",VR::IS,VM::VM1,"Bucky Height",false },
  {0x0019,0x0060,"SVISION",VR::IS,VM::VM1,"Bucky Angle",false },
  {0x0019,0x0068,"SVISION",VR::IS,VM::VM1,"C-Arm Angle",false },
  {0x0019,0x0069,"SVISION",VR::IS,VM::VM1,"Collimator Angle",false },
  {0x0019,0x0070,"SVISION",VR::IS,VM::VM1,"Filter Number",false },
  {0x0019,0x0074,"SVISION",VR::LO,VM::VM1,"Filter Material 1",false },
  {0x0019,0x0075,"SVISION",VR::LO,VM::VM1,"Filter Material 2",false },
  {0x0019,0x0078,"SVISION",VR::DS,VM::VM1,"Filter Thickness 1",false },
  {0x0019,0x0079,"SVISION",VR::DS,VM::VM1,"Filter Thickness 2",false },
  {0x0019,0x0080,"SVISION",VR::IS,VM::VM1,"Bucky Format",false },
  {0x0019,0x0081,"SVISION",VR::LO,VM::VM1,"Object Position",false },
  {0x0019,0x0090,"SVISION",VR::LO,VM::VM1,"Desk Command",false },
  {0x0019,0x0091,"SVISION",VR::IS,VM::VM1,"Central Beam X",false },
  {0x0019,0x0092,"SVISION",VR::IS,VM::VM1,"Central Beam Y",false },
  {0x0019,0x0093,"SVISION",VR::IS,VM::VM1,"Tube Turn Angle",false },
  {0x0019,0x0094,"SVISION",VR::IS,VM::VM1,"Stand drive level",false },
  {0x0019,0x00a0,"SVISION",VR::DS,VM::VM1,"Extended Exposure Time",false },
  {0x0019,0x00a1,"SVISION",VR::DS,VM::VM1,"Actual Exposure Time",false },
  {0x0019,0x00a8,"SVISION",VR::DS,VM::VM1,"Extended X-ray Tube Current",false },
  {0x0019,0x00b0,"SVISION",VR::IS,VM::VM1,"Dose Indicator",false },
  {0x0019,0x00b1,"SVISION",VR::IS,VM::VM1,"Shift Reference Value",false },
  {0x0019,0x00f0,"SVISION",VR::IS,VM::VM1,"?",false },
  {0x0021,0x0000,"SVISION",VR::DS,VM::VM1,"Noise Reduction",false },
  {0x0021,0x0001,"SVISION",VR::DS,VM::VM1,"Contrast Amplification",false },
  {0x0021,0x0002,"SVISION",VR::DS,VM::VM1,"Edge Contrast Boosting",false },
  {0x0021,0x0003,"SVISION",VR::DS,VM::VM1,"Latitude Reduction",false },
  {0x0021,0x0010,"SVISION",VR::LO,VM::VM1,"Find Range Algorithm",false },
  {0x0021,0x0011,"SVISION",VR::DS,VM::VM1,"Threshold C-Algorithm",false },
  {0x0021,0x0020,"SVISION",VR::LO,VM::VM1,"Sensometric Curve",false },
  {0x0021,0x0030,"SVISION",VR::DS,VM::VM1,"Lower Window Offset",false },
  {0x0021,0x0031,"SVISION",VR::DS,VM::VM1,"Upper Window Offset",false },
  {0x0021,0x0040,"SVISION",VR::DS,VM::VM1,"Min. Printable Density",false },
  {0x0021,0x0041,"SVISION",VR::DS,VM::VM1,"Max. Printable Density",false },
  {0x0021,0x0050,"SVISION",VR::DS,VM::VM1,"Min. Window Latitude",false },
  {0x0021,0x0051,"SVISION",VR::DS,VM::VM1,"Max. Window Latitude",false },
  {0x0021,0x0052,"SVISION",VR::DS,VM::VM1,"Rel. Window Alignment",false },
  {0x0021,0x0060,"SVISION",VR::DS,VM::VM1,"Decomposition Layer",false },
  {0x0021,0x0090,"SVISION",VR::DS,VM::VM1,"Brightness",false },
  {0x0021,0x0091,"SVISION",VR::DS,VM::VM1,"Contrast",false },
  {0x0021,0x0092,"SVISION",VR::DS,VM::VM1,"Shape Factor",false },
  {0x0023,0x0000,"SVISION",VR::LO,VM::VM1,"Image Laterality",false },
  {0x0023,0x0001,"SVISION",VR::IS,VM::VM1,"Letter Position",false },
  {0x0023,0x0002,"SVISION",VR::IS,VM::VM1,"Burned In Annotation",false },
  {0x0023,0x0003,"SVISION",VR::LO,VM::VM1,"?",false },
  {0x0023,0x00f0,"SVISION",VR::IS,VM::VM1,"Image SOP Class",false },
  {0x0025,0x0000,"SVISION",VR::IS,VM::VM1,"Original Image",false },
  {0x0025,0x0001,"SVISION",VR::IS,VM::VM1,"Not Processed Image",false },
  {0x0025,0x0002,"SVISION",VR::IS,VM::VM1,"Cut Out Image",false },
  {0x0025,0x0003,"SVISION",VR::IS,VM::VM1,"Duplicated Image",false },
  {0x0025,0x0004,"SVISION",VR::IS,VM::VM1,"Stored Image",false },
  {0x0025,0x0005,"SVISION",VR::IS,VM::VM1,"Retrieved Image",false },
  {0x0025,0x0006,"SVISION",VR::IS,VM::VM1,"New Image",false },
  {0x0025,0x0007,"SVISION",VR::IS,VM::VM1,"Media Stored Image",false },
  {0x0025,0x0008,"SVISION",VR::IS,VM::VM1,"Image State",false },
  {0x0025,0x0009,"SVISION",VR::IS,VM::VM1,"Image Stitched Manually",false },
  {0x0025,0x000a,"SVISION",VR::IS,VM::VM1,"Image Stitched Automatically",false },
  {0x0025,0x0020,"SVISION",VR::LO,VM::VM1,"Source Image File",false },
  {0x0025,0x0021,"SVISION",VR::LO,VM::VM1,"Source UID",false },
  {0x0027,0x0000,"SVISION",VR::IS,VM::VM1,"Number of Series",false },
  {0x0027,0x0001,"SVISION",VR::IS,VM::VM1,"Number of Studies",false },
  {0x0027,0x0010,"SVISION",VR::DT,VM::VM1,"Oldest Series",false },
  {0x0027,0x0011,"SVISION",VR::DT,VM::VM1,"Newest Series",false },
  {0x0027,0x0012,"SVISION",VR::DT,VM::VM1,"Oldest Study",false },
  {0x0027,0x0013,"SVISION",VR::DT,VM::VM1,"Newest Study",false },
  {0x0029,0x0000,"SVISION",VR::IS,VM::VM1,"Key Note Instance UID",false },
  {0x0029,0x0001,"SVISION",VR::IS,VM::VM1,"Storage State",false },
  {0x0029,0x0002,"SVISION",VR::IS,VM::VM1,"Referenced Image SOP Class",false },
  {0x0029,0x0003,"SVISION",VR::LO,VM::VM1,"Ref. Image Instance UID",false },
  {0x0029,0x0004,"SVISION",VR::IS,VM::VM1,"Rel. Presentation State Number",false },
  {0x0029,0x0005,"SVISION",VR::LO,VM::VM1,"Rel. Presentation State UID",false },
  {0x0019,0x0020,"Siemens Ultrasound Miscellaneous",VR::SH,VM::VM1,"?",false },
  {0x0023,0x0000,"Siemens: Thorax/Multix FD Image Stamp",VR::US,VM::VM1,"?",false },
  {0x0023,0x0001,"Siemens: Thorax/Multix FD Image Stamp",VR::US,VM::VM1,"?",false },
  {0x0023,0x0002,"Siemens: Thorax/Multix FD Image Stamp",VR::US,VM::VM1,"?",false },
  {0x0023,0x0003,"Siemens: Thorax/Multix FD Image Stamp",VR::US,VM::VM1,"?",false },
  {0x0023,0x0004,"Siemens: Thorax/Multix FD Image Stamp",VR::US,VM::VM1,"?",false },
  {0x0019,0x0000,"Siemens: Thorax/Multix FD Lab Settings",VR::LT,VM::VM1,"Internal Value",false },
  {0x0019,0x0001,"Siemens: Thorax/Multix FD Lab Settings",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0002,"Siemens: Thorax/Multix FD Lab Settings",VR::LO,VM::VM1,"Total Area Dose Product",false },
  {0x0019,0x0003,"Siemens: Thorax/Multix FD Lab Settings",VR::US,VM::VM1,"?",false },
  {0x0019,0x0004,"Siemens: Thorax/Multix FD Lab Settings",VR::LO,VM::VM1,"?",false },
  {0x0019,0x0005,"Siemens: Thorax/Multix FD Lab Settings",VR::US,VM::VM1,"Cassette Image",false },
  {0x0019,0x0006,"Siemens: Thorax/Multix FD Lab Settings",VR::FD,VM::VM1,"Table Object Distance",false },
  {0x0019,0x0007,"Siemens: Thorax/Multix FD Lab Settings",VR::FD,VM::VM1,"Table Detector Distance",false },
  {0x0019,0x0008,"Siemens: Thorax/Multix FD Lab Settings",VR::US,VM::VM1,"Ortho Step Distance",false },
  {0x0021,0x0001,"Siemens: Thorax/Multix FD Lab Settings",VR::SS,VM::VM1,"Label Left Orientation",false },
  {0x0021,0x0002,"Siemens: Thorax/Multix FD Lab Settings",VR::FL,VM::VM1,"Label Left Percentage X",false },
  {0x0021,0x0003,"Siemens: Thorax/Multix FD Lab Settings",VR::FL,VM::VM1,"Label Left Percentage X",false },
  {0x0021,0x0004,"Siemens: Thorax/Multix FD Lab Settings",VR::US,VM::VM1,"Label Right",false },
  {0x0021,0x0006,"Siemens: Thorax/Multix FD Lab Settings",VR::FL,VM::VM1,"Label Right Percentage X",false },
  {0x0021,0x0007,"Siemens: Thorax/Multix FD Lab Settings",VR::FL,VM::VM1,"Label Right Percentage X",false },
  {0x0021,0x0008,"Siemens: Thorax/Multix FD Lab Settings",VR::US,VM::VM1,"Auto Window Flag",false },
  {0x0021,0x0009,"Siemens: Thorax/Multix FD Lab Settings",VR::SL,VM::VM1,"Auto Window Center",false },
  {0x0021,0x000a,"Siemens: Thorax/Multix FD Lab Settings",VR::SL,VM::VM1,"Auto Window Width",false },
  {0x0021,0x000b,"Siemens: Thorax/Multix FD Lab Settings",VR::SS,VM::VM1,"Filter ID",false },
  {0x0021,0x000c,"Siemens: Thorax/Multix FD Lab Settings",VR::FL,VM::VM1,"Degree of Edge Enhancement",false },
  {0x0021,0x000d,"Siemens: Thorax/Multix FD Lab Settings",VR::SS,VM::VM1,"Half Kernel Size",false },
  {0x0021,0x000e,"Siemens: Thorax/Multix FD Lab Settings",VR::US,VM::VM1,"Dose Control Value",false },
  {0x0021,0x000f,"Siemens: Thorax/Multix FD Lab Settings",VR::US,VM::VM1,"Deleted Image Flag",false },
  {0x0021,0x0010,"Siemens: Thorax/Multix FD Lab Settings",VR::US,VM::VM1,"Same Size Support",false },
  {0x0021,0x0011,"Siemens: Thorax/Multix FD Lab Settings",VR::SS,VM::VM1,"Label Left Init Position",false },
  {0x0021,0x0012,"Siemens: Thorax/Multix FD Lab Settings",VR::SS,VM::VM1,"Label Right Init Position",false },
  {0x0021,0x0013,"Siemens: Thorax/Multix FD Lab Settings",VR::SS,VM::VM1,"Patient Rotation",false },
  {0x0021,0x0014,"Siemens: Thorax/Multix FD Lab Settings",VR::US,VM::VM1,"Anatomic Correct View Flag",false },
  {0x0021,0x0015,"Siemens: Thorax/Multix FD Lab Settings",VR::SS,VM::VM1,"Auto Window Shift",false },
  {0x0021,0x0016,"Siemens: Thorax/Multix FD Lab Settings",VR::DS,VM::VM1,"Auto Window Expansion",false },
  {0x0021,0x0017,"Siemens: Thorax/Multix FD Lab Settings",VR::LO,VM::VM1,"System Type",false },
  {0x0021,0x0030,"Siemens: Thorax/Multix FD Lab Settings",VR::SH,VM::VM1,"Anatomic Sort Number",false },
  {0x0021,0x0031,"Siemens: Thorax/Multix FD Lab Settings",VR::SH,VM::VM1,"Acquisition Sort Number",false },
  {0x0021,0x0000,"Siemens: Thorax/Multix FD Post Processing",VR::US,VM::VM1,"Label Left",false },
  {0x0021,0x0001,"Siemens: Thorax/Multix FD Post Processing",VR::SS,VM::VM1,"Label Left Orientation",false },
  {0x0021,0x0002,"Siemens: Thorax/Multix FD Post Processing",VR::FL,VM::VM1,"Label Left Percentage X",false },
  {0x0021,0x0003,"Siemens: Thorax/Multix FD Post Processing",VR::FL,VM::VM1,"Label Left Percentage X",false },
  {0x0021,0x0004,"Siemens: Thorax/Multix FD Post Processing",VR::US,VM::VM1,"Label Right",false },
  {0x0021,0x0005,"Siemens: Thorax/Multix FD Post Processing",VR::SS,VM::VM1,"Label Right Orientation",false },
  {0x0021,0x0006,"Siemens: Thorax/Multix FD Post Processing",VR::FL,VM::VM1,"Label Right Percentage X",false },
  {0x0021,0x0007,"Siemens: Thorax/Multix FD Post Processing",VR::FL,VM::VM1,"Label Right Percentage X",false },
  {0x0021,0x0008,"Siemens: Thorax/Multix FD Post Processing",VR::US,VM::VM1,"Auto Window Flag",false },
  {0x0021,0x0009,"Siemens: Thorax/Multix FD Post Processing",VR::SL,VM::VM1,"Auto Window Center",false },
  {0x0021,0x000a,"Siemens: Thorax/Multix FD Post Processing",VR::SL,VM::VM1,"Auto Window Width",false },
  {0x0021,0x000b,"Siemens: Thorax/Multix FD Post Processing",VR::SS,VM::VM1,"Filter ID",false },
  {0x0021,0x000c,"Siemens: Thorax/Multix FD Post Processing",VR::FL,VM::VM1,"Degree of Edge Enhancement",false },
  {0x0021,0x000d,"Siemens: Thorax/Multix FD Post Processing",VR::SS,VM::VM1,"Half Kernel Size",false },
  {0x0021,0x000e,"Siemens: Thorax/Multix FD Post Processing",VR::US,VM::VM1,"Dose Control Value",false },
  {0x0021,0x000f,"Siemens: Thorax/Multix FD Post Processing",VR::US,VM::VM1,"Deleted Image Flag",false },
  {0x0021,0x0010,"Siemens: Thorax/Multix FD Post Processing",VR::US,VM::VM1,"Same Size Support",false },
  {0x0021,0x0011,"Siemens: Thorax/Multix FD Post Processing",VR::SS,VM::VM1,"Label Left Init Position",false },
  {0x0021,0x0012,"Siemens: Thorax/Multix FD Post Processing",VR::SS,VM::VM1,"Label Right Init Position",false },
  {0x0021,0x0013,"Siemens: Thorax/Multix FD Post Processing",VR::SS,VM::VM1,"Patient Rotation",false },
  {0x0021,0x0014,"Siemens: Thorax/Multix FD Post Processing",VR::US,VM::VM1,"Anatomic Correct View Flag",false },
  {0x0021,0x0015,"Siemens: Thorax/Multix FD Post Processing",VR::SS,VM::VM1,"Auto Window Shift",false },
  {0x0021,0x0016,"Siemens: Thorax/Multix FD Post Processing",VR::DS,VM::VM1,"Auto Window Expansion",false },
  {0x0025,0x0000,"Siemens: Thorax/Multix FD Raw Image Settings",VR::SS,VM::VM1,"Raw Image Amplification",false },
  {0x0025,0x0001,"Siemens: Thorax/Multix FD Raw Image Settings",VR::SS,VM::VM1,"Gamma LUT",false },
  {0x0025,0x0002,"Siemens: Thorax/Multix FD Raw Image Settings",VR::US,VM::VM1,"Collimator Manual",false },
  {0x0025,0x0003,"Siemens: Thorax/Multix FD Raw Image Settings",VR::FL,VM::VM1,"Collimator Up Left X Percentage",false },
  {0x0025,0x0004,"Siemens: Thorax/Multix FD Raw Image Settings",VR::FL,VM::VM1,"Collimator Up Right X Percentage",false },
  {0x0025,0x0005,"Siemens: Thorax/Multix FD Raw Image Settings",VR::FL,VM::VM1,"Collimator Low Left X Percentage",false },
  {0x0025,0x0006,"Siemens: Thorax/Multix FD Raw Image Settings",VR::FL,VM::VM1,"Collimator Low Right X Percentage",false },
  {0x0025,0x0007,"Siemens: Thorax/Multix FD Raw Image Settings",VR::FL,VM::VM1,"Collimator Up Left Y Percentage",false },
  {0x0025,0x0008,"Siemens: Thorax/Multix FD Raw Image Settings",VR::FL,VM::VM1,"Collimator Up Right Y Percentage",false },
  {0x0025,0x0009,"Siemens: Thorax/Multix FD Raw Image Settings",VR::FL,VM::VM1,"Collimator Low Left Y Percentage",false },
  {0x0025,0x000a,"Siemens: Thorax/Multix FD Raw Image Settings",VR::FL,VM::VM1,"Collimator Low Right Y Percentage",false },
  {0x0025,0x000b,"Siemens: Thorax/Multix FD Raw Image Settings",VR::US,VM::VM1,"Set Gray",false },
  {0x0025,0x000c,"Siemens: Thorax/Multix FD Raw Image Settings",VR::SS,VM::VM1,"Harmonization Kernel",false },
  {0x0025,0x000d,"Siemens: Thorax/Multix FD Raw Image Settings",VR::FL,VM::VM1,"Harmonization Gain",false },
  {0x0025,0x000e,"Siemens: Thorax/Multix FD Raw Image Settings",VR::SS,VM::VM1,"Edge Enhancement Kernel",false },
  {0x0025,0x000f,"Siemens: Thorax/Multix FD Raw Image Settings",VR::FL,VM::VM1,"Edge Enhancement Gain",false },
  {0x0025,0x0010,"Siemens: Thorax/Multix FD Raw Image Settings",VR::LT,VM::VM1,"Raw Image File Name",false },
  {0x0025,0x0011,"Siemens: Thorax/Multix FD Raw Image Settings",VR::SS,VM::VM1,"Global Offset",false },
  {0x0025,0x0012,"Siemens: Thorax/Multix FD Raw Image Settings",VR::SS,VM::VM1,"Image Columns",false },
  {0x0025,0x0013,"Siemens: Thorax/Multix FD Raw Image Settings",VR::SS,VM::VM1,"Image Rows",false },
  {0x0025,0x0014,"Siemens: Thorax/Multix FD Raw Image Settings",VR::SS,VM::VM1,"Image Alignment",false },
  {0x0025,0x0015,"Siemens: Thorax/Multix FD Raw Image Settings",VR::SS,VM::VM1,"Active Image Columns",false },
  {0x0025,0x0016,"Siemens: Thorax/Multix FD Raw Image Settings",VR::SS,VM::VM1,"Active Image Rows",false },
  {0x0025,0x0017,"Siemens: Thorax/Multix FD Raw Image Settings",VR::LO,VM::VM1,"System Type",false },
  {0x0025,0x0018,"Siemens: Thorax/Multix FD Raw Image Settings",VR::US,VM::VM1,"Auto Gain",false },
  {0x0025,0x0019,"Siemens: Thorax/Multix FD Raw Image Settings",VR::US,VM::VM1,"Ortho Sub Sampling",false },
  {0x0025,0x001a,"Siemens: Thorax/Multix FD Raw Image Settings",VR::US,VM::VM2,"Image Crop Upper Left",false },
  {0x0025,0x001b,"Siemens: Thorax/Multix FD Raw Image Settings",VR::US,VM::VM2,"Image Crop Upper Right",false },
  {0x0025,0x001c,"Siemens: Thorax/Multix FD Raw Image Settings",VR::US,VM::VM2,"Image Crop Lower Left",false },
  {0x0025,0x001d,"Siemens: Thorax/Multix FD Raw Image Settings",VR::US,VM::VM2,"Image Crop Lower Right",false },
  {0x0025,0x0030,"Siemens: Thorax/Multix FD Raw Image Settings",VR::US,VM::VM1,"Manual Cropping",false },
  {0x0025,0x0031,"Siemens: Thorax/Multix FD Raw Image Settings",VR::SS,VM::VM1,"Gamma LUT Parameter 1",false },
  {0x0025,0x0032,"Siemens: Thorax/Multix FD Raw Image Settings",VR::DS,VM::VM1,"Gamma LUT Parameter 2",false },
  {0x0025,0x0033,"Siemens: Thorax/Multix FD Raw Image Settings",VR::SS,VM::VM1,"Gamma LUT Parameter 3",false },
  {0x0025,0x0034,"Siemens: Thorax/Multix FD Raw Image Settings",VR::SS,VM::VM1,"Gamma LUT Parameter 4",false },
  {0x0017,0x0000,"Siemens: Thorax/Multix FD Version",VR::LO,VM::VM1,"Build Date",false },
  {0x0017,0x0001,"Siemens: Thorax/Multix FD Version",VR::LO,VM::VM1,"IOD Version",false },
  {0x0029,0x0011,"Silhouette Annot V1.0",VR::IS,VM::VM1,"Annotation Name",false },
  {0x0029,0x0012,"Silhouette Annot V1.0",VR::LT,VM::VM1,"Annotation Font",false },
  {0x0029,0x0013,"Silhouette Annot V1.0",VR::LT,VM::VM1,"Annotation Text Foreground Color",false },
  {0x0029,0x0014,"Silhouette Annot V1.0",VR::LT,VM::VM1,"Annotation Text Background Color",false },
  {0x0029,0x0015,"Silhouette Annot V1.0",VR::UL,VM::VM1,"Annotation Text Backing Mode",false },
  {0x0029,0x0016,"Silhouette Annot V1.0",VR::UL,VM::VM1,"Annotation Text Justification",false },
  {0x0029,0x0017,"Silhouette Annot V1.0",VR::UL,VM::VM1,"Annotation Text Location",false },
  {0x0029,0x0018,"Silhouette Annot V1.0",VR::LT,VM::VM1,"Annotation Text String",false },
  {0x0029,0x0019,"Silhouette Annot V1.0",VR::UL,VM::VM1,"Annotation Text Attach Mode",false },
  {0x0029,0x0020,"Silhouette Annot V1.0",VR::UL,VM::VM1,"Annotation Text Cursor Mode",false },
  {0x0029,0x0021,"Silhouette Annot V1.0",VR::UL,VM::VM1,"Annotation Text Shadow Offset X",false },
  {0x0029,0x0022,"Silhouette Annot V1.0",VR::UL,VM::VM1,"Annotation Text Shadow Offset Y",false },
  {0x0029,0x0023,"Silhouette Annot V1.0",VR::LT,VM::VM1,"Annotation Line Color",false },
  {0x0029,0x0024,"Silhouette Annot V1.0",VR::UL,VM::VM1,"Annotation Line Thickness",false },
  {0x0029,0x0025,"Silhouette Annot V1.0",VR::UL,VM::VM1,"Annotation Line Type",false },
  {0x0029,0x0026,"Silhouette Annot V1.0",VR::UL,VM::VM1,"Annotation Line Style",false },
  {0x0029,0x0027,"Silhouette Annot V1.0",VR::UL,VM::VM1,"Annotation Line Dash Length",false },
  {0x0029,0x0028,"Silhouette Annot V1.0",VR::UL,VM::VM1,"Annotation Line Attach Mode",false },
  {0x0029,0x0029,"Silhouette Annot V1.0",VR::UL,VM::VM1,"Annotation Line Point Count",false },
  {0x0029,0x0030,"Silhouette Annot V1.0",VR::FD,VM::VM1,"Annotation Line Points",false },
  {0x0029,0x0031,"Silhouette Annot V1.0",VR::UL,VM::VM1,"Annotation Line Control Size",false },
  {0x0029,0x0032,"Silhouette Annot V1.0",VR::LT,VM::VM1,"Annotation Marker Color",false },
  {0x0029,0x0033,"Silhouette Annot V1.0",VR::UL,VM::VM1,"Annotation Marker Type",false },
  {0x0029,0x0034,"Silhouette Annot V1.0",VR::UL,VM::VM1,"Annotation Marker Size",false },
  {0x0029,0x0035,"Silhouette Annot V1.0",VR::FD,VM::VM1,"Annotation Marker Location",false },
  {0x0029,0x0036,"Silhouette Annot V1.0",VR::UL,VM::VM1,"Annotation Marker Attach Mode",false },
  {0x0029,0x0037,"Silhouette Annot V1.0",VR::LT,VM::VM1,"Annotation Geom Color",false },
  {0x0029,0x0038,"Silhouette Annot V1.0",VR::UL,VM::VM1,"Annotation Geom Thickness",false },
  {0x0029,0x0039,"Silhouette Annot V1.0",VR::UL,VM::VM1,"Annotation Geom Line Style",false },
  {0x0029,0x0040,"Silhouette Annot V1.0",VR::UL,VM::VM1,"Annotation Geom Dash Length",false },
  {0x0029,0x0041,"Silhouette Annot V1.0",VR::UL,VM::VM1,"Annotation Geom Fill Pattern",false },
  {0x0029,0x0042,"Silhouette Annot V1.0",VR::UL,VM::VM1,"Annotation Interactivity",false },
  {0x0029,0x0043,"Silhouette Annot V1.0",VR::FD,VM::VM1,"Annotation Arrow Length",false },
  {0x0029,0x0044,"Silhouette Annot V1.0",VR::FD,VM::VM1,"Annotation Arrow Angle",false },
  {0x0029,0x0045,"Silhouette Annot V1.0",VR::UL,VM::VM1,"Annotation Don't Save",false },
  {0x0029,0x0000,"Silhouette Graphics Export V1.0",VR::UI,VM::VM1,"?",false },
  {0x0029,0x0011,"Silhouette Line V1.0",VR::IS,VM::VM1,"Line Name",false },
  {0x0029,0x0012,"Silhouette Line V1.0",VR::LT,VM::VM1,"Line Name Font",false },
  {0x0029,0x0013,"Silhouette Line V1.0",VR::UL,VM::VM1,"Line Name Display",false },
  {0x0029,0x0014,"Silhouette Line V1.0",VR::LT,VM::VM1,"Line Normal Color",false },
  {0x0029,0x0015,"Silhouette Line V1.0",VR::UL,VM::VM1,"Line Type",false },
  {0x0029,0x0016,"Silhouette Line V1.0",VR::UL,VM::VM1,"Line Thickness",false },
  {0x0029,0x0017,"Silhouette Line V1.0",VR::UL,VM::VM1,"Line Style",false },
  {0x0029,0x0018,"Silhouette Line V1.0",VR::UL,VM::VM1,"Line Dash Length",false },
  {0x0029,0x0019,"Silhouette Line V1.0",VR::UL,VM::VM1,"Line Interactivity",false },
  {0x0029,0x0020,"Silhouette Line V1.0",VR::LT,VM::VM1,"Line Measurement Color",false },
  {0x0029,0x0021,"Silhouette Line V1.0",VR::LT,VM::VM1,"Line Measurement Font",false },
  {0x0029,0x0022,"Silhouette Line V1.0",VR::UL,VM::VM1,"Line Measurement Dash Length",false },
  {0x0029,0x0023,"Silhouette Line V1.0",VR::UL,VM::VM1,"Line Point Space",false },
  {0x0029,0x0024,"Silhouette Line V1.0",VR::FD,VM::VM1,"Line Points",false },
  {0x0029,0x0025,"Silhouette Line V1.0",VR::UL,VM::VM1,"Line Control Point Size",false },
  {0x0029,0x0026,"Silhouette Line V1.0",VR::UL,VM::VM1,"Line Control Point Space",false },
  {0x0029,0x0027,"Silhouette Line V1.0",VR::FD,VM::VM1,"Line Control Points",false },
  {0x0029,0x0028,"Silhouette Line V1.0",VR::LT,VM::VM1,"Line Label",false },
  {0x0029,0x0029,"Silhouette Line V1.0",VR::UL,VM::VM1,"Line Don't Save",false },
  {0x0029,0x0011,"Silhouette ROI V1.0",VR::IS,VM::VM1,"ROI Name",false },
  {0x0029,0x0012,"Silhouette ROI V1.0",VR::LT,VM::VM1,"ROI Name Font",false },
  {0x0029,0x0013,"Silhouette ROI V1.0",VR::LT,VM::VM1,"ROI Normal Color",false },
  {0x0029,0x0014,"Silhouette ROI V1.0",VR::UL,VM::VM1,"ROI Fill Pattern",false },
  {0x0029,0x0015,"Silhouette ROI V1.0",VR::UL,VM::VM1,"ROI Bp Seg",false },
  {0x0029,0x0016,"Silhouette ROI V1.0",VR::UN,VM::VM1,"ROI Bp Seg Pairs",false },
  {0x0029,0x0017,"Silhouette ROI V1.0",VR::UL,VM::VM1,"ROI Seed Space",false },
  {0x0029,0x0018,"Silhouette ROI V1.0",VR::UN,VM::VM1,"ROI Seeds",false },
  {0x0029,0x0019,"Silhouette ROI V1.0",VR::UL,VM::VM1,"ROI Line Thickness",false },
  {0x0029,0x0020,"Silhouette ROI V1.0",VR::UL,VM::VM1,"ROI Line Style",false },
  {0x0029,0x0021,"Silhouette ROI V1.0",VR::UL,VM::VM1,"ROI Line Dash Length",false },
  {0x0029,0x0022,"Silhouette ROI V1.0",VR::UL,VM::VM1,"ROI Interactivity",false },
  {0x0029,0x0023,"Silhouette ROI V1.0",VR::UL,VM::VM1,"ROI Name Position",false },
  {0x0029,0x0024,"Silhouette ROI V1.0",VR::UL,VM::VM1,"ROI Name Display",false },
  {0x0029,0x0025,"Silhouette ROI V1.0",VR::LT,VM::VM1,"ROI Label",false },
  {0x0029,0x0026,"Silhouette ROI V1.0",VR::UL,VM::VM1,"ROI Shape",false },
  {0x0029,0x0027,"Silhouette ROI V1.0",VR::FD,VM::VM1,"ROI Shape Tilt",false },
  {0x0029,0x0028,"Silhouette ROI V1.0",VR::UL,VM::VM1,"ROI Shape Points Count",false },
  {0x0029,0x0029,"Silhouette ROI V1.0",VR::UL,VM::VM1,"ROI Shape Points Space",false },
  {0x0029,0x0030,"Silhouette ROI V1.0",VR::FD,VM::VM1,"ROI Shape Points",false },
  {0x0029,0x0031,"Silhouette ROI V1.0",VR::UL,VM::VM1,"ROI Shape Control Points Count",false },
  {0x0029,0x0032,"Silhouette ROI V1.0",VR::UL,VM::VM1,"ROI Shape Control Points Space",false },
  {0x0029,0x0033,"Silhouette ROI V1.0",VR::FD,VM::VM1,"ROI Shape Control Points",false },
  {0x0029,0x0034,"Silhouette ROI V1.0",VR::UL,VM::VM1,"ROI Don't Save",false },
  {0x0029,0x0041,"Silhouette Sequence Ids V1.0",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0042,"Silhouette Sequence Ids V1.0",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0043,"Silhouette Sequence Ids V1.0",VR::SQ,VM::VM1,"?",false },
  {0x0029,0x0013,"Silhouette V1.0",VR::UL,VM::VM1,"?",false },
  {0x0029,0x0014,"Silhouette V1.0",VR::UL,VM::VM1,"?",false },
  {0x0029,0x0017,"Silhouette V1.0",VR::UN,VM::VM1,"?",false },
  {0x0029,0x0018,"Silhouette V1.0",VR::UN,VM::VM1,"?",false },
  {0x0029,0x0019,"Silhouette V1.0",VR::UL,VM::VM1,"?",false },
  {0x0029,0x001a,"Silhouette V1.0",VR::UN,VM::VM1,"?",false },
  {0x0029,0x001b,"Silhouette V1.0",VR::UL,VM::VM1,"?",false },
  {0x0029,0x001c,"Silhouette V1.0",VR::UL,VM::VM1,"?",false },
  {0x0029,0x001d,"Silhouette V1.0",VR::UN,VM::VM1,"?",false },
  {0x0029,0x001e,"Silhouette V1.0",VR::UN,VM::VM1,"?",false },
  {0x0029,0x0021,"Silhouette V1.0",VR::US,VM::VM1,"?",false },
  {0x0029,0x0022,"Silhouette V1.0",VR::US,VM::VM1,"?",false },
  {0x0029,0x0023,"Silhouette V1.0",VR::US,VM::VM1,"?",false },
  {0x0029,0x0024,"Silhouette V1.0",VR::US,VM::VM1,"?",false },
  {0x0029,0x0025,"Silhouette V1.0",VR::US,VM::VM1,"?",false },
  {0x0029,0x0027,"Silhouette V1.0",VR::UN,VM::VM1,"?",false },
  {0x0029,0x0028,"Silhouette V1.0",VR::UN,VM::VM1,"?",false },
  {0x0029,0x0029,"Silhouette V1.0",VR::UN,VM::VM1,"?",false },
  {0x0029,0x0030,"Silhouette V1.0",VR::UN,VM::VM1,"?",false },
  {0x0029,0x0052,"Silhouette V1.0",VR::US,VM::VM1,"?",false },
  {0x0029,0x0053,"Silhouette V1.0",VR::LT,VM::VM1,"?",false },
  {0x0029,0x0054,"Silhouette V1.0",VR::UN,VM::VM1,"?",false },
  {0x0029,0x0055,"Silhouette V1.0",VR::LT,VM::VM1,"?",false },
  {0x0029,0x0056,"Silhouette V1.0",VR::LT,VM::VM1,"?",false },
  {0x0029,0x0057,"Silhouette V1.0",VR::UN,VM::VM1,"?",false },
  {0x0029,0x0013,"Silhouette VRS 3.0",VR::UL,VM::VM1,"?",false },
  {0x0029,0x0014,"Silhouette VRS 3.0",VR::UL,VM::VM1,"Length in Byte of Image",false },
  {0x0029,0x0017,"Silhouette VRS 3.0",VR::FD,VM::VM2,"?",false },
  {0x0029,0x0018,"Silhouette VRS 3.0",VR::FD,VM::VM2,"?",false },
  {0x0029,0x0019,"Silhouette VRS 3.0",VR::UL,VM::VM1,"?",false },
  {0x0029,0x001a,"Silhouette VRS 3.0",VR::FD,VM::VM1,"?",false },
  {0x0029,0x001b,"Silhouette VRS 3.0",VR::UL,VM::VM1,"?",false },
  {0x0029,0x001c,"Silhouette VRS 3.0",VR::UL,VM::VM1,"?",false },
  {0x0029,0x001d,"Silhouette VRS 3.0",VR::UL,VM::VM3,"?",false },
  {0x0029,0x001e,"Silhouette VRS 3.0",VR::UL,VM::VM3,"?",false },
  {0x0029,0x0027,"Silhouette VRS 3.0",VR::US,VM::VM1,"?",false },
  {0x0029,0x0028,"Silhouette VRS 3.0",VR::OB,VM::VM1,"?",false },
  {0x0029,0x0030,"Silhouette VRS 3.0",VR::US,VM::VM1,"?",false },
  {0x0029,0x0032,"Silhouette VRS 3.0",VR::US,VM::VM1,"?",false },
  {0x0029,0x0034,"Silhouette VRS 3.0",VR::LO,VM::VM1,"?",false },
  {0x0029,0x0035,"Silhouette VRS 3.0",VR::CS,VM::VM1,"?",false },
  {0x0029,0x0036,"Silhouette VRS 3.0",VR::US,VM::VM1,"?",false },
  {0xf001,0x0000,"Sound Technologies",VR::CS,VM::VM1,"Patient Species",false },
  {0xf001,0x0001,"Sound Technologies",VR::CS,VM::VM1,"Patient Breed",false },
  {0xf001,0x0002,"Sound Technologies",VR::CS,VM::VM1,"Patient Category Size",false },
  {0xf001,0x0003,"Sound Technologies",VR::CS,VM::VM1,"Patient Sex Extended (including spay/neuter)",false },
  {0xf001,0x0004,"Sound Technologies",VR::CS,VM::VM1,"Image View",false },
  {0xf001,0x0005,"Sound Technologies",VR::CS,VM::VM1,"Anatomy Imaged",false },
  {0xf001,0x0006,"Sound Technologies",VR::CS,VM::VM1,"Image enhancements",false },
  {0xf001,0x0007,"Sound Technologies",VR::CS,VM::VM1,"Detector Settings",false },
  {0xf001,0x0008,"Sound Technologies",VR::CS,VM::VM1,"Application Version",false },
  {0xf001,0x0009,"Sound Technologies",VR::LO,VM::VM1,"Image Laterality Extended",false },
  {0xf001,0x000a,"Sound Technologies",VR::PN,VM::VM1,"Client Name",false },
  {0xf001,0x000b,"Sound Technologies",VR::UI,VM::VM1,"Reference Study Instance Uid",false },
  {0xf001,0x000c,"Sound Technologies",VR::UI,VM::VM1,"Reference Series Instance Uid",false },
  {0xf001,0x000d,"Sound Technologies",VR::LO,VM::VM1,"Exam Ref Id",false },
  {0xf001,0x000e,"Sound Technologies",VR::ST,VM::VM1,"Physician Of Record Address",false },
  {0xf001,0x000f,"Sound Technologies",VR::SH,VM::VM1,"Physician Of Record Phone Numbers",false },
  {0xf001,0x0010,"Sound Technologies",VR::LT,VM::VM1,"Reason For Study",false },
  {0xf001,0x0011,"Sound Technologies",VR::LO,VM::VM1,"Protocol",false },
  {0xf001,0x0012,"Sound Technologies",VR::LO,VM::VM1,"CaptureInputType",false },
  {0xf001,0x0013,"Sound Technologies",VR::LT,VM::VM1,"Exam Complaint",false },
  {0xf001,0x0014,"Sound Technologies",VR::LO,VM::VM1,"Exam Web Code",false },
  {0xf001,0x0015,"Sound Technologies",VR::LO,VM::VM1,"Exam Category",false },
  {0xf001,0x0016,"Sound Technologies",VR::SH,VM::VM1,"Exam Diagnosis",false },
  {0xf001,0x0017,"Sound Technologies",VR::PN,VM::VM1,"Exam Created By",false },
  {0xf001,0x0018,"Sound Technologies",VR::LO,VM::VM1,"Exam Created By Group",false },
  {0xf001,0x0019,"Sound Technologies",VR::DT,VM::VM1,"Exam Required By DateTime",false },
  {0xf001,0x001a,"Sound Technologies",VR::LO,VM::VM1,"Capture Type",false },
  {0xf001,0x001b,"Sound Technologies",VR::IS,VM::VM1,"Telemed Exam ID",false },
  {0xf001,0x001c,"Sound Technologies",VR::LO,VM::VM1,"Exam Created By Guid",false },
  {0xf001,0x001d,"Sound Technologies",VR::LO,VM::VM1,"Client Name Guid",false },
  {0x7eff,0x0021,"TMDTK",VR::SQ,VM::VM1,"?",false },
  {0x7eff,0x002e,"TMDTK",VR::LT,VM::VM1,"?",false },
  {0x0029,0x0008,"TOSHIBA COMAPL HEADER",VR::CS,VM::VM1,"COMAPL Header Type",false },
  {0x0029,0x0009,"TOSHIBA COMAPL HEADER",VR::LO,VM::VM1,"COMAPL Header Version",false },
  {0x0029,0x0010,"TOSHIBA COMAPL HEADER",VR::SQ,VM::VM1,"COMAPL Header Info",false },
  {0x0029,0x0020,"TOSHIBA COMAPL HEADER",VR::OB,VM::VM1,"COMAPL History Information",false },
  {0x0029,0x0031,"TOSHIBA COMAPL HEADER",VR::LO,VM::VM1_n,"",false },
  {0x0029,0x0032,"TOSHIBA COMAPL HEADER",VR::UL,VM::VM1_n,"",false },
  {0x0029,0x0033,"TOSHIBA COMAPL HEADER",VR::UL,VM::VM1_n,"",false },
  {0x0029,0x0034,"TOSHIBA COMAPL HEADER",VR::CS,VM::VM1_n,"",false },
  {0x0029,0x0008,"TOSHIBA COMAPL OOG",VR::CS,VM::VM1,"COMAPL OOG Type",false },
  {0x0029,0x0009,"TOSHIBA COMAPL OOG",VR::LO,VM::VM1,"COMAPL OOG Version",false },
  {0x0029,0x0010,"TOSHIBA COMAPL OOG",VR::LO,VM::VM1,"",false },
  {0x0029,0x0031,"TOSHIBA COMAPL OOG",VR::LO,VM::VM1_n,"",false },
  {0x0029,0x0032,"TOSHIBA COMAPL OOG",VR::UL,VM::VM1_n,"",false },
  {0x0029,0x0033,"TOSHIBA COMAPL OOG",VR::UL,VM::VM1_n,"",false },
  {0x0029,0x0034,"TOSHIBA COMAPL OOG",VR::CS,VM::VM1_n,"",false },
  {0x7015,0x0000,"TOSHIBA ENCRYPTED SR DATA",VR::OB,VM::VM1,"Toshiba US Private Data",false },
  {0x0029,0x0008,"TOSHIBA MDW HEADER",VR::CS,VM::VM1,"Image Header Type",false },
  {0x0029,0x0009,"TOSHIBA MDW HEADER",VR::LO,VM::VM1,"Image Header Version",false },
  {0x0029,0x0010,"TOSHIBA MDW HEADER",VR::OB,VM::VM1,"Image Header Info",false },
  {0x0029,0x0011,"TOSHIBA MDW HEADER",VR::LO,VM::VM1,"Image Header Version",false },
  {0x0029,0x0012,"TOSHIBA MDW HEADER",VR::OB,VM::VM1,"Image Header Info",false },
  {0x0029,0x0018,"TOSHIBA MDW HEADER",VR::CS,VM::VM1,"Series Header Type",false },
  {0x0029,0x0019,"TOSHIBA MDW HEADER",VR::LO,VM::VM1,"Series Header Version",false },
  {0x0029,0x0020,"TOSHIBA MDW HEADER",VR::OB,VM::VM1,"Series Header Info",false },
  {0x0029,0x0031,"TOSHIBA MDW HEADER",VR::LO,VM::VM1,"PMTF Information 1",false },
  {0x0029,0x0032,"TOSHIBA MDW HEADER",VR::UL,VM::VM1,"PMTF Information 2",false },
  {0x0029,0x0033,"TOSHIBA MDW HEADER",VR::UL,VM::VM1,"PMTF Information 3",false },
  {0x0029,0x0034,"TOSHIBA MDW HEADER",VR::CS,VM::VM1,"PMTF Information 4",false },
  {0x0029,0x0008,"TOSHIBA MDW NON-IMAGE",VR::CS,VM::VM1,"Application Header Type",false },
  {0x0029,0x0009,"TOSHIBA MDW NON-IMAGE",VR::LO,VM::VM1,"Application Header Version",false },
  {0x0029,0x0010,"TOSHIBA MDW NON-IMAGE",VR::OB,VM::VM1,"Application Header Data",false },
  {0x0029,0x0020,"TOSHIBA MDW NON-IMAGE",VR::OB,VM::VM1,"Application Header Data",false },
  {0x7fe1,0x0010,"TOSHIBA MDW NON-IMAGE",VR::OB,VM::VM1,"Toshiba US Private Data",false },
  {0x0009,0x0001,"TOSHIBA_MEC_1.0",VR::LT,VM::VM1,"?",false },
  {0x0009,0x0002,"TOSHIBA_MEC_1.0",VR::US,VM::VM1_n,"?",false },
  {0x0009,0x0003,"TOSHIBA_MEC_1.0",VR::US,VM::VM1_n,"?",false },
  {0x0009,0x0004,"TOSHIBA_MEC_1.0",VR::US,VM::VM1_n,"?",false },
  {0x0009,0x0006,"TOSHIBA_MEC_1.0",VR::US,VM::VM6,"?",false },
  {0x0009,0x0007,"TOSHIBA_MEC_1.0",VR::UN,VM::VM1,"?",false },
  {0x0009,0x0008,"TOSHIBA_MEC_1.0",VR::CS,VM::VM1,"?",false },
  {0x0009,0x0009,"TOSHIBA_MEC_1.0",VR::CS,VM::VM1,"?",false },
  {0x0009,0x0012,"TOSHIBA_MEC_1.0",VR::IS,VM::VM1,"?",false },
  {0x0009,0x0013,"TOSHIBA_MEC_1.0",VR::IS,VM::VM1,"?",false },
  {0x0009,0x0018,"TOSHIBA_MEC_1.0",VR::CS,VM::VM1,"?",false },
  {0x0009,0x0019,"TOSHIBA_MEC_1.0",VR::CS,VM::VM1,"?",false },
  {0x0009,0x0082,"TOSHIBA_MEC_1.0",VR::LO,VM::VM1_n,"?",false },
  {0x0009,0x0085,"TOSHIBA_MEC_1.0",VR::US,VM::VM6,"?",false },
  {0x0011,0x0001,"TOSHIBA_MEC_1.0",VR::LT,VM::VM1,"?Patient's Name?",false },
  {0x0011,0x0002,"TOSHIBA_MEC_1.0",VR::US,VM::VM1_n,"?",false },
  {0x0011,0x0080,"TOSHIBA_MEC_1.0",VR::LT,VM::VM1,"?Patient's Name?",false },
  {0x0019,0x0001,"TOSHIBA_MEC_1.0",VR::US,VM::VM1_n,"?",false },
  {0x0019,0x0002,"TOSHIBA_MEC_1.0",VR::US,VM::VM1_n,"?",false },
  {0x0021,0x0001,"TOSHIBA_MEC_1.0",VR::US,VM::VM1_n,"?",false },
  {0x0021,0x0002,"TOSHIBA_MEC_1.0",VR::US,VM::VM1_n,"?",false },
  {0x0021,0x0003,"TOSHIBA_MEC_1.0",VR::US,VM::VM1_n,"?",false },
  {0x0021,0x0016,"TOSHIBA_MEC_1.0",VR::LO,VM::VM1,"?",false },
  {0x0021,0x0017,"TOSHIBA_MEC_1.0",VR::CS,VM::VM1_n,"?",false },
  {0x0021,0x0020,"TOSHIBA_MEC_1.0",VR::US,VM::VM4,"?",false },
  {0x7ff1,0x0001,"TOSHIBA_MEC_1.0",VR::US,VM::VM1_n,"?",false },
  {0x7ff1,0x0002,"TOSHIBA_MEC_1.0",VR::US,VM::VM1_n,"?",false },
  {0x7ff1,0x0003,"TOSHIBA_MEC_1.0",VR::US,VM::VM1_n,"?",false },
  {0x7ff1,0x0004,"TOSHIBA_MEC_1.0",VR::SH,VM::VM1_n,"?File Type Remarks?",false },
  {0x7ff1,0x0010,"TOSHIBA_MEC_1.0",VR::US,VM::VM1_n,"?",false },
  {0x7005,0x0000,"TOSHIBA_MEC_CT3",VR::OB,VM::VM1,"CT Private Data 1",false },
  {0x7005,0x0003,"TOSHIBA_MEC_CT3",VR::SH,VM::VM1,"Cardiac R-R Mean Time",false },
  {0x7005,0x0004,"TOSHIBA_MEC_CT3",VR::SH,VM::VM1,"Cardiac Reconstruction Gating Phase in Percent",false },
  {0x7005,0x0005,"TOSHIBA_MEC_CT3",VR::SH,VM::VM1,"Cardiac Reconstruction Gating Phase in ms",false },
  {0x7005,0x0006,"TOSHIBA_MEC_CT3",VR::SH,VM::VM1,"Cardiac Reconstruction Mode",false },
  {0x7005,0x0007,"TOSHIBA_MEC_CT3",VR::DS,VM::VM2,"Reconstruction Center",false },
  {0x7005,0x0008,"TOSHIBA_MEC_CT3",VR::DS,VM::VM1,"Detector Slice Thickness in mm",false },
  {0x7005,0x0009,"TOSHIBA_MEC_CT3",VR::LO,VM::VM1,"Number of Detector rows to Reconstruct",false },
  {0x7005,0x000a,"TOSHIBA_MEC_CT3",VR::DS,VM::VM1,"Table Speed in mm/rot",false },
  {0x7005,0x000b,"TOSHIBA_MEC_CT3",VR::SH,VM::VM1,"Filter",false },
  {0x7005,0x000c,"TOSHIBA_MEC_CT3",VR::SH,VM::VM1,"Reconstruction Correction Type",false },
  {0x7005,0x000d,"TOSHIBA_MEC_CT3",VR::CS,VM::VM1,"Organ",false },
  {0x7005,0x000e,"TOSHIBA_MEC_CT3",VR::SH,VM::VM1,"File Type Remarks",false },
  {0x7005,0x000f,"TOSHIBA_MEC_CT3",VR::SH,VM::VM1,"Direction",false },
  {0x7005,0x0010,"TOSHIBA_MEC_CT3",VR::OB,VM::VM1_n,"CT Private Data 2",false },
  {0x7005,0x0011,"TOSHIBA_MEC_CT3",VR::LT,VM::VM1,"Series Comment",false },
  {0x7005,0x0012,"TOSHIBA_MEC_CT3",VR::SH,VM::VM1,"Position",false },
  {0x7005,0x0013,"TOSHIBA_MEC_CT3",VR::US,VM::VM1,"Expert Plan No.",false },
  {0x7005,0x0014,"TOSHIBA_MEC_CT3",VR::US,VM::VM1,"Reconstruction ROI No.",false },
  {0x7005,0x0015,"TOSHIBA_MEC_CT3",VR::US,VM::VM1,"Special Helical ACQ No.",false },
  {0x7005,0x0016,"TOSHIBA_MEC_CT3",VR::UI,VM::VM1,"Volume UID",false },
  {0x7005,0x0017,"TOSHIBA_MEC_CT3",VR::US,VM::VM1,"Total Frame Count in the Volume",false },
  {0x7005,0x0018,"TOSHIBA_MEC_CT3",VR::US,VM::VM1,"Frame No.",false },
  {0x7005,0x0019,"TOSHIBA_MEC_CT3",VR::UL,VM::VM1,"Frame Sort Key",false },
  {0x7005,0x001a,"TOSHIBA_MEC_CT3",VR::US,VM::VM1,"Frame Sort Order",false },
  {0x7005,0x001b,"TOSHIBA_MEC_CT3",VR::SH,VM::VM1,"Convolution Kernel",false },
  {0x7005,0x001c,"TOSHIBA_MEC_CT3",VR::LO,VM::VM1,"Contrast/Bolus Agent",false },
  {0x7005,0x001d,"TOSHIBA_MEC_CT3",VR::UL,VM::VM1,"Reconstruction Number",false },
  {0x7005,0x001e,"TOSHIBA_MEC_CT3",VR::UL,VM::VM1,"Raw Data Number",false },
  {0x7005,0x001f,"TOSHIBA_MEC_CT3",VR::LO,VM::VM1,"Volume Number",false },
  {0x7005,0x0020,"TOSHIBA_MEC_CT3",VR::UL,VM::VM1,"Local Series Number",false },
  {0x7005,0x0021,"TOSHIBA_MEC_CT3",VR::LO,VM::VM1,"Decrease in Artifact Filter",false },
  {0x7005,0x0022,"TOSHIBA_MEC_CT3",VR::DS,VM::VM1,"Reconstruction Interval",false },
  {0x7005,0x0023,"TOSHIBA_MEC_CT3",VR::DS,VM::VM1,"Pitch Factor",false },
  {0x7005,0x0024,"TOSHIBA_MEC_CT3",VR::DA,VM::VM1,"The Acquisition Date of NRA",false },
  {0x7005,0x0025,"TOSHIBA_MEC_CT3",VR::UL,VM::VM1,"Large Data File Attribute",false },
  {0x7005,0x0026,"TOSHIBA_MEC_CT3",VR::CS,VM::VM1_8,"Large Data File Name",false },
  {0x7005,0x0028,"TOSHIBA_MEC_CT3",VR::SQ,VM::VM1,"Enhanced CT Private Sequence",false },
  {0x7005,0x0029,"TOSHIBA_MEC_CT3",VR::UI,VM::VM1,"Frame UID",false },
  {0x7005,0x0030,"TOSHIBA_MEC_CT3",VR::CS,VM::VM1,"Main Modality in Study",false },
  {0x7005,0x0035,"TOSHIBA_MEC_CT3",VR::DS,VM::VM2,"Scan Range",false },
  {0x7005,0x0036,"TOSHIBA_MEC_CT3",VR::OB,VM::VM1_n,"CT Private Data 3",false },
  {0x7005,0x0037,"TOSHIBA_MEC_CT3",VR::IS,VM::VM1,"Total Frames",false },
  {0x7005,0x0038,"TOSHIBA_MEC_CT3",VR::IS,VM::VM1,"Start Frame",false },
  {0x7005,0x0039,"TOSHIBA_MEC_CT3",VR::IS,VM::VM1,"End Frame",false },
  {0x7005,0x0040,"TOSHIBA_MEC_CT3",VR::FD,VM::VM1,"DLP Dose Length Product",false },
  {0x7005,0x0041,"TOSHIBA_MEC_CT3",VR::SH,VM::VM1,"Raw Slice Information",false },
  {0x7005,0x0042,"TOSHIBA_MEC_CT3",VR::US,VM::VM1,"Local Frame No",false },
  {0x7005,0x0043,"TOSHIBA_MEC_CT3",VR::DS,VM::VM3,"Volume Vector",false },
  {0x7005,0x0044,"TOSHIBA_MEC_CT3",VR::US,VM::VM1,"Volume Type",false },
  {0x7005,0x0045,"TOSHIBA_MEC_CT3",VR::DS,VM::VM1,"Relative Table Position of 4D Volume",false },
  {0x7005,0x0046,"TOSHIBA_MEC_CT3",VR::DS,VM::VM1,"Absolute Table Position of 4D Volume",false },
  {0x7005,0x0047,"TOSHIBA_MEC_CT3",VR::DS,VM::VM1,"Slice Pitch of 4D Volume",false },
  {0x7005,0x0048,"TOSHIBA_MEC_CT3",VR::LO,VM::VM1,"Respiratory Gating Inf.",false },
  {0x7005,0x0049,"TOSHIBA_MEC_CT3",VR::SH,VM::VM1,"Respiration Phase",false },
  {0x7005,0x0063,"TOSHIBA_MEC_CT3",VR::FD,VM::VM1,"CTDIw",false },
  {0x7005,0x00f1,"TOSHIBA_MEC_CT3",VR::CS,VM::VM1,"Protect Mark for Image, Curve or Private Record",false },
  {0x7005,0x00f2,"TOSHIBA_MEC_CT3",VR::CS,VM::VM1,"Protect Mark for Series Record",false },
  {0x7005,0x00f3,"TOSHIBA_MEC_CT3",VR::CS,VM::VM1,"Protect Mark for Study Record",false },
  {0x0019,0x0001,"TOSHIBA_MEC_CT_1.0",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0002,"TOSHIBA_MEC_CT_1.0",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0003,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM1_n,"?",false },
  {0x0019,0x0004,"TOSHIBA_MEC_CT_1.0",VR::LT,VM::VM1,"?",false },
  {0x0019,0x0005,"TOSHIBA_MEC_CT_1.0",VR::LT,VM::VM1,"?",false },
  {0x0019,0x0006,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM1_n,"?",false },
  {0x0019,0x0007,"TOSHIBA_MEC_CT_1.0",VR::OB,VM::VM1,"?",false },
  {0x0019,0x0008,"TOSHIBA_MEC_CT_1.0",VR::LT,VM::VM1,"Orientation Head Feet",false },
  {0x0019,0x0009,"TOSHIBA_MEC_CT_1.0",VR::LT,VM::VM1,"View Direction",false },
  {0x0019,0x000a,"TOSHIBA_MEC_CT_1.0",VR::LT,VM::VM1,"Orientation Supine Prone",false },
  {0x0019,0x000b,"TOSHIBA_MEC_CT_1.0",VR::DS,VM::VM1,"Location",false },
  {0x0019,0x000c,"TOSHIBA_MEC_CT_1.0",VR::CS,VM::VM1,"Scan View Direction",false },
  {0x0019,0x000d,"TOSHIBA_MEC_CT_1.0",VR::TM,VM::VM1,"Time",false },
  {0x0019,0x000e,"TOSHIBA_MEC_CT_1.0",VR::DS,VM::VM1,"?",false },
  {0x0019,0x0011,"TOSHIBA_MEC_CT_1.0",VR::IS,VM::VM1,"?",false },
  {0x0019,0x0022,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"?",false },
  {0x0019,0x0023,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM6,"?",false },
  {0x0019,0x0030,"TOSHIBA_MEC_CT_1.0",VR::LO,VM::VM1,"?Protocol Name?",false },
  {0x0019,0x0041,"TOSHIBA_MEC_CT_1.0",VR::LO,VM::VM1,"?Date?",false },
  {0x0019,0x0086,"TOSHIBA_MEC_CT_1.0",VR::DS,VM::VM1,"?Nominal Cardiac Trigger Delay Time?",false },
  {0x0019,0x0087,"TOSHIBA_MEC_CT_1.0",VR::DS,VM::VM1,"?Cardiac R-R Interval Specified?",false },
  {0x7ff1,0x0001,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM1_n,"?",false },
  {0x7ff1,0x0002,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM1_n,"?",false },
  {0x7ff1,0x0003,"TOSHIBA_MEC_CT_1.0",VR::IS,VM::VM1,"?",false },
  {0x7ff1,0x0004,"TOSHIBA_MEC_CT_1.0",VR::IS,VM::VM1,"?",false },
  {0x7ff1,0x0005,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM1_n,"?",false },
  {0x7ff1,0x0007,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM1_n,"?",false },
  {0x7ff1,0x0008,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM1_n,"?",false },
  {0x7ff1,0x0009,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM1_n,"?",false },
  {0x7ff1,0x000a,"TOSHIBA_MEC_CT_1.0",VR::LT,VM::VM1,"?",false },
  {0x7ff1,0x000b,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM1_n,"?",false },
  {0x7ff1,0x000c,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM1_n,"?",false },
  {0x7ff1,0x000d,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM1_n,"?",false },
  {0x7ff1,0x0011,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM1,"?",false },
  {0x7ff1,0x0012,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"?",false },
  {0x7ff1,0x0013,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"?",false },
  {0x7ff1,0x0014,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM1,"?",false },
  {0x7ff1,0x0015,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM1,"?",false },
  {0x7ff1,0x0016,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"?",false },
  {0x7ff1,0x0017,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM1,"?",false },
  {0x7ff1,0x0018,"TOSHIBA_MEC_CT_1.0",VR::LO,VM::VM1,"?Decrease in Artifact Filter?",false },
  {0x7ff1,0x0019,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM4,"?",false },
  {0x7ff1,0x001a,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM1,"?",false },
  {0x7ff1,0x0021,"TOSHIBA_MEC_CT_1.0",VR::LO,VM::VM1,"?Date?",false },
  {0x7ff1,0x0022,"TOSHIBA_MEC_CT_1.0",VR::LO,VM::VM1,"?Time?",false },
  {0x7ff1,0x0023,"TOSHIBA_MEC_CT_1.0",VR::CS,VM::VM1,"?",false },
  {0x7ff1,0x0032,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM1,"?",false },
  {0x7ff1,0x0033,"TOSHIBA_MEC_CT_1.0",VR::LO,VM::VM1,"?",false },
  {0x7ff1,0x0034,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"?",false },
  {0x7ff1,0x0035,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM1,"?",false },
  {0x7ff1,0x0038,"TOSHIBA_MEC_CT_1.0",VR::IS,VM::VM2,"?",false },
  {0x7ff1,0x003c,"TOSHIBA_MEC_CT_1.0",VR::IS,VM::VM1,"?",false },
  {0x7ff1,0x003d,"TOSHIBA_MEC_CT_1.0",VR::DS,VM::VM1,"?",false },
  {0x7ff1,0x003e,"TOSHIBA_MEC_CT_1.0",VR::CS,VM::VM1,"?",false },
  {0x7ff1,0x003f,"TOSHIBA_MEC_CT_1.0",VR::DS,VM::VM1,"?",false },
  {0x7ff1,0x0042,"TOSHIBA_MEC_CT_1.0",VR::IS,VM::VM1,"?",false },
  {0x7ff1,0x0045,"TOSHIBA_MEC_CT_1.0",VR::CS,VM::VM1,"?Some kind of Date?",false },
  {0x7ff1,0x0048,"TOSHIBA_MEC_CT_1.0",VR::LO,VM::VM1,"?",false },
  {0x7ff1,0x004b,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"?Reconstruction Number?",false },
  {0x7ff1,0x004d,"TOSHIBA_MEC_CT_1.0",VR::CS,VM::VM1,"?EP?",false },
  {0x7ff1,0x0050,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"??",false },
  {0x7ff1,0x0052,"TOSHIBA_MEC_CT_1.0",VR::DS,VM::VM1,"?",false },
  {0x7ff1,0x0053,"TOSHIBA_MEC_CT_1.0",VR::DS,VM::VM1,"?",false },
  {0x7ff1,0x0054,"TOSHIBA_MEC_CT_1.0",VR::DS,VM::VM1,"?",false },
  {0x7ff1,0x0056,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"?",false },
  {0x7ff1,0x0058,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"?",false },
  {0x7ff1,0x005b,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"?",false },
  {0x7ff1,0x005c,"TOSHIBA_MEC_CT_1.0",VR::IS,VM::VM1,"?",false },
  {0x7ff1,0x005d,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"?",false },
  {0x7ff1,0x005f,"TOSHIBA_MEC_CT_1.0",VR::DA,VM::VM1,"?ACR-NEMA DA encoded?",false },
  {0x7ff1,0x0060,"TOSHIBA_MEC_CT_1.0",VR::DS,VM::VM1,"?",false },
  {0x7ff1,0x0061,"TOSHIBA_MEC_CT_1.0",VR::UI,VM::VM1,"?",false },
  {0x7ff1,0x0063,"TOSHIBA_MEC_CT_1.0",VR::LO,VM::VM1,"?",false },
  {0x7ff1,0x0065,"TOSHIBA_MEC_CT_1.0",VR::LO,VM::VM1,"?",false },
  {0x7ff1,0x0066,"TOSHIBA_MEC_CT_1.0",VR::DS,VM::VM3,"?Volume Vector?",false },
  {0x7ff1,0x0075,"TOSHIBA_MEC_CT_1.0",VR::UI,VM::VM1,"?UID?",false },
  {0x7ff1,0x0078,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"?",false },
  {0x7ff1,0x0079,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"?",false },
  {0x7ff1,0x007a,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"?",false },
  {0x7ff1,0x007b,"TOSHIBA_MEC_CT_1.0",VR::UI,VM::VM1,"?Volume UID?",false },
  {0x7ff1,0x007c,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM1,"?",false },
  {0x7ff1,0x007d,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM1,"?",false },
  {0x7ff1,0x007e,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"?Frame Sort Key?",false },
  {0x7ff1,0x007f,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM1,"?",false },
  {0x7ff1,0x0080,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM1,"?",false },
  {0x7ff1,0x0081,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"?",false },
  {0x7ff1,0x0082,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"?",false },
  {0x7ff1,0x0083,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM1_n,"?",false },
  {0x7ff1,0x0084,"TOSHIBA_MEC_CT_1.0",VR::IS,VM::VM2,"?",false },
  {0x7ff1,0x0086,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"?",false },
  {0x7ff1,0x0087,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"?",false },
  {0x7ff1,0x0088,"TOSHIBA_MEC_CT_1.0",VR::LO,VM::VM1,"?",false },
  {0x7ff1,0x0089,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"?Local Series Number?",false },
  {0x7ff1,0x008f,"TOSHIBA_MEC_CT_1.0",VR::LO,VM::VM1,"?",false },
  {0x7ff1,0x0090,"TOSHIBA_MEC_CT_1.0",VR::IS,VM::VM1,"?",false },
  {0x7ff1,0x0091,"TOSHIBA_MEC_CT_1.0",VR::IS,VM::VM1,"?",false },
  {0x7ff1,0x0092,"TOSHIBA_MEC_CT_1.0",VR::IS,VM::VM1,"?",false },
  {0x7ff1,0x0093,"TOSHIBA_MEC_CT_1.0",VR::IS,VM::VM4,"?",false },
  {0x7ff1,0x0094,"TOSHIBA_MEC_CT_1.0",VR::LO,VM::VM1,"?",false },
  {0x7ff1,0x0095,"TOSHIBA_MEC_CT_1.0",VR::LO,VM::VM1,"?",false },
  {0x7ff1,0x0096,"TOSHIBA_MEC_CT_1.0",VR::LO,VM::VM1,"?",false },
  {0x7ff1,0x009a,"TOSHIBA_MEC_CT_1.0",VR::LO,VM::VM1,"?",false },
  {0x7ff1,0x009c,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM1,"?",false },
  {0x7ff1,0x009d,"TOSHIBA_MEC_CT_1.0",VR::DS,VM::VM1,"?",false },
  {0x7ff1,0x009e,"TOSHIBA_MEC_CT_1.0",VR::DS,VM::VM1,"?",false },
  {0x7ff1,0x00a0,"TOSHIBA_MEC_CT_1.0",VR::DS,VM::VM1,"?",false },
  {0x7ff1,0x00a1,"TOSHIBA_MEC_CT_1.0",VR::IS,VM::VM1_n,"?",false },
  {0x7ff1,0x00a2,"TOSHIBA_MEC_CT_1.0",VR::IS,VM::VM1,"?",false },
  {0x7ff1,0x00a3,"TOSHIBA_MEC_CT_1.0",VR::CS,VM::VM1,"?",false },
  {0x7ff1,0x00a7,"TOSHIBA_MEC_CT_1.0",VR::IS,VM::VM1,"?",false },
  {0x7ff1,0x00a8,"TOSHIBA_MEC_CT_1.0",VR::IS,VM::VM1,"?",false },
  {0x7ff1,0x00a9,"TOSHIBA_MEC_CT_1.0",VR::IS,VM::VM1,"?",false },
  {0x7ff1,0x00ae,"TOSHIBA_MEC_CT_1.0",VR::SQ,VM::VM1,"?",false },
  {0x7ff1,0x00b1,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"??",false },
  {0x7ff1,0x00b2,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"??",false },
  {0x7ff1,0x00b3,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"??",false },
  {0x7ff1,0x00b4,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"??",false },
  {0x7ff1,0x00b5,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"??",false },
  {0x7ff1,0x00bc,"TOSHIBA_MEC_CT_1.0",VR::IS,VM::VM1,"??",false },
  {0x7ff1,0x00cd,"TOSHIBA_MEC_CT_1.0",VR::IS,VM::VM1,"??",false },
  {0x7ff1,0x00ce,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM1_n,"??",false },
  {0x7ff1,0x00cf,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM1,"??",false },
  {0x7ff1,0x00d2,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM4,"??",false },
  {0x7ff1,0x00d3,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM4,"??",false },
  {0x7ff1,0x00d4,"TOSHIBA_MEC_CT_1.0",VR::US,VM::VM4,"??",false },
  {0x7ff1,0x00d5,"TOSHIBA_MEC_CT_1.0",VR::LO,VM::VM1,"??",false },
  {0x7ff1,0x00d6,"TOSHIBA_MEC_CT_1.0",VR::DS,VM::VM1,"?Reconstruction Angle?",false },
  {0x7ff1,0x00d8,"TOSHIBA_MEC_CT_1.0",VR::IS,VM::VM1,"?Number of Frames?",false },
  {0x7ff1,0x00d9,"TOSHIBA_MEC_CT_1.0",VR::DS,VM::VM1,"?Slice Pitch of 4D Volume?",false },
  {0x7ff1,0x00e6,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"?",false },
  {0x7ff1,0x00e7,"TOSHIBA_MEC_CT_1.0",VR::UL,VM::VM1,"?",false },
  {0x7ff1,0x00e8,"TOSHIBA_MEC_CT_1.0",VR::LO,VM::VM1,"?",false },
  {0x7ff1,0x00e9,"TOSHIBA_MEC_CT_1.0",VR::LO,VM::VM1,"?",false },
  {0x7ff1,0x00eb,"TOSHIBA_MEC_CT_1.0",VR::DS,VM::VM1,"?",false },
  {0x7ff1,0x00ed,"TOSHIBA_MEC_CT_1.0",VR::LO,VM::VM1,"?",false },
  {0x7ff1,0x00f0,"TOSHIBA_MEC_CT_1.0",VR::SQ,VM::VM1,"?",false },
  {0x7ff1,0x0007,"TOSHIBA_MEC_CT_2.0",VR::UL,VM::VM1,"??",false },
  {0x7ff1,0x0008,"TOSHIBA_MEC_CT_2.0",VR::UL,VM::VM1,"??",false },
  {0x7ff1,0x0009,"TOSHIBA_MEC_CT_2.0",VR::UL,VM::VM1,"??",false },
  {0x7ff1,0x000a,"TOSHIBA_MEC_CT_2.0",VR::UL,VM::VM1,"??",false },
  {0x7ff1,0x000b,"TOSHIBA_MEC_CT_2.0",VR::LO,VM::VM1,"??",false },
  {0x7ff1,0x000f,"TOSHIBA_MEC_CT_2.0",VR::CS,VM::VM1,"??",false },
  {0x7ff1,0x0010,"TOSHIBA_MEC_CT_2.0",VR::CS,VM::VM1,"?Filter Type?",false },
  {0x7ff1,0x0014,"TOSHIBA_MEC_CT_2.0",VR::LO,VM::VM1,"?Convolution Kernel Group?",false },
  {0x7ff1,0x0016,"TOSHIBA_MEC_CT_2.0",VR::LO,VM::VM1,"??",false },
  {0x7ff1,0x0019,"TOSHIBA_MEC_CT_2.0",VR::DS,VM::VM1,"?Acquisition Duration?",false },
  {0x7ff1,0x001a,"TOSHIBA_MEC_CT_2.0",VR::UL,VM::VM1,"??",false },
  {0x7ff1,0x001b,"TOSHIBA_MEC_CT_2.0",VR::UL,VM::VM1,"??",false },
  {0x7ff1,0x001e,"TOSHIBA_MEC_CT_2.0",VR::DS,VM::VM1,"?X-Ray Tube Current in mA?",false },
  {0x700d,0x0000,"TOSHIBA_MEC_MR3",VR::DS,VM::VM1,"Scale Factor",false },
  {0x700d,0x0001,"TOSHIBA_MEC_MR3",VR::OB,VM::VM1,"Acquisition Order",false },
  {0x700d,0x0002,"TOSHIBA_MEC_MR3",VR::DS,VM::VM9,"Orientation Vector",false },
  {0x700d,0x0003,"TOSHIBA_MEC_MR3",VR::SS,VM::VM1,"Flip Flag",false },
  {0x700d,0x0004,"TOSHIBA_MEC_MR3",VR::OB,VM::VM1,"Rotate Information",false },
  {0x700d,0x0005,"TOSHIBA_MEC_MR3",VR::DS,VM::VM4,"FOV",false },
  {0x700d,0x0006,"TOSHIBA_MEC_MR3",VR::US,VM::VM4,"Image Matrix",false },
  {0x700d,0x0007,"TOSHIBA_MEC_MR3",VR::OB,VM::VM1,"Image Information",false },
  {0x700d,0x0008,"TOSHIBA_MEC_MR3",VR::OB,VM::VM1,"Original Data",false },
  {0x700d,0x0009,"TOSHIBA_MEC_MR3",VR::SS,VM::VM1,"Original Data Flag",false },
  {0x700d,0x000c,"TOSHIBA_MEC_MR3",VR::CS,VM::VM1,"Receiver Gain Correction Check Flag",false },
  {0x700d,0x0020,"TOSHIBA_MEC_MR3",VR::SH,VM::VM1,"Identification Flag of 3D GDC",false },
  {0x700d,0x0080,"TOSHIBA_MEC_MR3",VR::US,VM::VM1,"Number of PAC channel",false },
  {0x700d,0x0081,"TOSHIBA_MEC_MR3",VR::US,VM::VM1,"Reference mode",false },
  {0x700d,0x0082,"TOSHIBA_MEC_MR3",VR::SQ,VM::VM1,"Gain value group for MRS",false },
  {0x700d,0x0083,"TOSHIBA_MEC_MR3",VR::FL,VM::VM1,"Gain value of each channel for MRS",false },
  {0x700d,0x0084,"TOSHIBA_MEC_MR3",VR::FL,VM::VM1,"Phase value of each channel for MRS",false },
  {0x700d,0x0085,"TOSHIBA_MEC_MR3",VR::FL,VM::VM1,"Reference gain value of each channel for MRS",false },
  {0x700d,0x0086,"TOSHIBA_MEC_MR3",VR::FL,VM::VM1,"Reference gain value of each channel for MRS",false },
  {0x700d,0x0087,"TOSHIBA_MEC_MR3",VR::FL,VM::VM1,"Receiver gain of reference scan",false },
  {0x700d,0x0088,"TOSHIBA_MEC_MR3",VR::UL,VM::VM1,"Flag of water Sat pulse",false },
  {0x700d,0x0089,"TOSHIBA_MEC_MR3",VR::FL,VM::VM1,"Selected contrast TE",false },
  {0x700d,0x008a,"TOSHIBA_MEC_MR3",VR::SQ,VM::VM1,"Raw Data Set Sequence",false },
  {0x700d,0x008b,"TOSHIBA_MEC_MR3",VR::DS,VM::VM1,"Multi Slice number",false },
  {0x700d,0x008c,"TOSHIBA_MEC_MR3",VR::DS,VM::VM1,"Multi Coverage number",false },
  {0x700d,0x008d,"TOSHIBA_MEC_MR3",VR::SQ,VM::VM1,"Raw Data Sequence",false },
  {0x700d,0x008e,"TOSHIBA_MEC_MR3",VR::DS,VM::VM1,"Coil Channel No.",false },
  {0x700d,0x008f,"TOSHIBA_MEC_MR3",VR::OB,VM::VM1,"MRS Raw Data",false },
  {0x700d,0x0090,"TOSHIBA_MEC_MR3",VR::OF,VM::VM1,"MRS Ref Raw Data",false },
  {0x700d,0x0091,"TOSHIBA_MEC_MR3",VR::FL,VM::VM1,"Receiver gain of prescan",false },
  {0x0009,0x0000,"TOSHIBA_MEC_OT3",VR::LO,VM::VM1,"HIS/RIS Study ID",false },
  {0x7019,0x0045,"TOSHIBA_MEC_OT3",VR::SQ,VM::VM1,"?",false },
  {0x7019,0x0046,"TOSHIBA_MEC_OT3",VR::DA,VM::VM1,"?",false },
  {0x7019,0x0047,"TOSHIBA_MEC_OT3",VR::TM,VM::VM1,"?",false },
  {0x7019,0x0071,"TOSHIBA_MEC_OT3",VR::CS,VM::VM1,"Indicator of creator",false },
  {0x7019,0x0072,"TOSHIBA_MEC_OT3",VR::LO,VM::VM1,"ID of quality assurance reviewer",false },
  {0x7019,0x0073,"TOSHIBA_MEC_OT3",VR::PN,VM::VM1,"Name of quality assurance reviewer",false },
  {0x7019,0x0074,"TOSHIBA_MEC_OT3",VR::DA,VM::VM1,"Date of completion of quality assurance",false },
  {0x7019,0x0075,"TOSHIBA_MEC_OT3",VR::TM,VM::VM1,"Time of completion of quality assurance",false },
  {0x7019,0x0076,"TOSHIBA_MEC_OT3",VR::LO,VM::VM1,"Host name of quality assurance station",false },
  {0x7019,0x0077,"TOSHIBA_MEC_OT3",VR::CS,VM::VM1,"Station name and version of quality assurance station",false },
  {0x7019,0x0078,"TOSHIBA_MEC_OT3",VR::CS,VM::VM1,"Flag of completion of quality assurance",false },
  {0x7019,0x0079,"TOSHIBA_MEC_OT3",VR::CS,VM::VM1,"Status of quality assurance review",false },
  {0x7019,0x007a,"TOSHIBA_MEC_OT3",VR::UI,VM::VM1,"Original SOP Instance UID",false },
  {0x7019,0x007b,"TOSHIBA_MEC_OT3",VR::UI,VM::VM1,"Original Series Instance UID",false },
  {0x7019,0x007c,"TOSHIBA_MEC_OT3",VR::UI,VM::VM1,"Original Study Instance UID",false },
  {0x7019,0x007d,"TOSHIBA_MEC_OT3",VR::US,VM::VM1,"Rotation information for image display from original image",false },
  {0x7019,0x007e,"TOSHIBA_MEC_OT3",VR::SQ,VM::VM1,"Sequence of LR mark",false },
  {0x7019,0x007f,"TOSHIBA_MEC_OT3",VR::US,VM::VM1,"Rotation status at creation of LR mark",false },
  {0x7019,0x0080,"TOSHIBA_MEC_OT3",VR::OB_OW,VM::VM1,"Original image data on area of original LR mark",false },
  {0x7019,0x0081,"TOSHIBA_MEC_OT3",VR::SS,VM::VM1,"X start position of LR mark",false },
  {0x7019,0x0082,"TOSHIBA_MEC_OT3",VR::SS,VM::VM1,"Y start position of LR mark",false },
  {0x7019,0x0083,"TOSHIBA_MEC_OT3",VR::US,VM::VM1,"Size of LR mark in X direction",false },
  {0x7019,0x0084,"TOSHIBA_MEC_OT3",VR::US,VM::VM1,"Size of LR mark in Y direction",false },
  {0x7019,0x0085,"TOSHIBA_MEC_OT3",VR::LO,VM::VM1,"Previous Patient ID",false },
  {0x7019,0x0086,"TOSHIBA_MEC_OT3",VR::CS,VM::VM1,"Previous modality",false },
  {0x7079,0x0010,"TOSHIBA_MEC_XA3",VR::LO,VM::VM1,"",false },
  {0x7079,0x0021,"TOSHIBA_MEC_XA3",VR::SH,VM::VM5,"",false },
  {0x7079,0x0022,"TOSHIBA_MEC_XA3",VR::IS,VM::VM2,"",false },
  {0x7079,0x0023,"TOSHIBA_MEC_XA3",VR::IS,VM::VM2,"",false },
  {0x7079,0x0024,"TOSHIBA_MEC_XA3",VR::DS,VM::VM1,"",false },
  {0x7079,0x0025,"TOSHIBA_MEC_XA3",VR::SH,VM::VM1,"",false },
  {0x7079,0x0026,"TOSHIBA_MEC_XA3",VR::DS,VM::VM2,"",false },
  {0x7079,0x0027,"TOSHIBA_MEC_XA3",VR::US,VM::VM2,"",false },
  {0x7079,0x0028,"TOSHIBA_MEC_XA3",VR::US,VM::VM1,"",false },
  {0x7079,0x0029,"TOSHIBA_MEC_XA3",VR::US,VM::VM10,"",false },
  {0x7079,0x002a,"TOSHIBA_MEC_XA3",VR::US,VM::VM5,"",false },
  {0x7079,0x002b,"TOSHIBA_MEC_XA3",VR::UN,VM::VM1,"",false },
  {0x7079,0x002c,"TOSHIBA_MEC_XA3",VR::SH,VM::VM3,"",false },
  {0x7079,0x002d,"TOSHIBA_MEC_XA3",VR::SS,VM::VM2,"",false },
  {0x7079,0x002e,"TOSHIBA_MEC_XA3",VR::SH,VM::VM1,"",false },
  {0x7079,0x002f,"TOSHIBA_MEC_XA3",VR::SH,VM::VM1,"",false },
  {0x7079,0x0030,"TOSHIBA_MEC_XA3",VR::US,VM::VM1,"",false },
  {0x7079,0x0031,"TOSHIBA_MEC_XA3",VR::US,VM::VM1_n,"",false },
  {0x7079,0x0032,"TOSHIBA_MEC_XA3",VR::SH,VM::VM2,"",false },
  {0x7079,0x0033,"TOSHIBA_MEC_XA3",VR::US,VM::VM1_n,"",false },
  {0x7079,0x0034,"TOSHIBA_MEC_XA3",VR::SH,VM::VM1,"",false },
  {0x7079,0x0035,"TOSHIBA_MEC_XA3",VR::SH,VM::VM1,"",false },
  {0x7079,0x0036,"TOSHIBA_MEC_XA3",VR::SH,VM::VM1,"",false },
  {0x7079,0x0037,"TOSHIBA_MEC_XA3",VR::US,VM::VM4,"",false },
  {0x7079,0x0038,"TOSHIBA_MEC_XA3",VR::SS,VM::VM1_n,"",false },
  {0x7079,0x0039,"TOSHIBA_MEC_XA3",VR::SS,VM::VM1_n,"",false },
  {0x7079,0x003a,"TOSHIBA_MEC_XA3",VR::US,VM::VM2,"",false },
  {0x7079,0x003b,"TOSHIBA_MEC_XA3",VR::SH,VM::VM1,"",false },
  {0x7079,0x003c,"TOSHIBA_MEC_XA3",VR::DS,VM::VM2,"",false },
  {0x7079,0x003d,"TOSHIBA_MEC_XA3",VR::SS,VM::VM2,"",false },
  {0x7079,0x003e,"TOSHIBA_MEC_XA3",VR::SH,VM::VM1,"",false },
  {0x7079,0x003f,"TOSHIBA_MEC_XA3",VR::US,VM::VM2,"",false },
  {0x7079,0x0040,"TOSHIBA_MEC_XA3",VR::SH,VM::VM2,"",false },
  {0x7079,0x0041,"TOSHIBA_MEC_XA3",VR::SS,VM::VM2,"",false },
  {0x7079,0x0042,"TOSHIBA_MEC_XA3",VR::SH,VM::VM1,"",false },
  {0x7079,0x0043,"TOSHIBA_MEC_XA3",VR::SH,VM::VM1,"",false },
  {0x7079,0x0044,"TOSHIBA_MEC_XA3",VR::US,VM::VM1,"",false },
  {0x7079,0x0045,"TOSHIBA_MEC_XA3",VR::SH,VM::VM1,"",false },
  {0x7079,0x0046,"TOSHIBA_MEC_XA3",VR::SS,VM::VM1,"",false },
  {0x7079,0x0047,"TOSHIBA_MEC_XA3",VR::SS,VM::VM18,"",false },
  {0x7079,0x0048,"TOSHIBA_MEC_XA3",VR::US,VM::VM12,"",false },
  {0x7079,0x0049,"TOSHIBA_MEC_XA3",VR::US,VM::VM3,"",false },
  {0x7079,0x004a,"TOSHIBA_MEC_XA3",VR::US,VM::VM1,"",false },
  {0x7079,0x004b,"TOSHIBA_MEC_XA3",VR::LO,VM::VM3,"",false },
  {0x7079,0x004c,"TOSHIBA_MEC_XA3",VR::OB,VM::VM1,"",false },
  {0x7079,0x004d,"TOSHIBA_MEC_XA3",VR::LO,VM::VM3,"",false },
  {0x7079,0x004e,"TOSHIBA_MEC_XA3",VR::SL,VM::VM35,"",false },
  {0x7079,0x004f,"TOSHIBA_MEC_XA3",VR::SH,VM::VM3,"",false },
  {0x7079,0x0050,"TOSHIBA_MEC_XA3",VR::US,VM::VM1_n,"",false },
  {0x7079,0x0051,"TOSHIBA_MEC_XA3",VR::US,VM::VM1_n,"",false },
  {0x7079,0x0052,"TOSHIBA_MEC_XA3",VR::US,VM::VM1,"",false },
  {0x7079,0x0053,"TOSHIBA_MEC_XA3",VR::SH,VM::VM1,"",false },
  {0x7079,0x0054,"TOSHIBA_MEC_XA3",VR::US,VM::VM5,"",false },
  {0x7079,0x0055,"TOSHIBA_MEC_XA3",VR::UL,VM::VM47_47n,"",false },
  {0x7079,0x0056,"TOSHIBA_MEC_XA3",VR::US,VM::VM1,"",false },
  {0x7079,0x0057,"TOSHIBA_MEC_XA3",VR::US,VM::VM1_n,"",false },
  {0x7079,0x0058,"TOSHIBA_MEC_XA3",VR::US,VM::VM1,"",false },
  {0x7079,0x0059,"TOSHIBA_MEC_XA3",VR::US,VM::VM2_2n,"",false },
  {0x7079,0x005a,"TOSHIBA_MEC_XA3",VR::US,VM::VM1,"",false },
  {0x7079,0x005b,"TOSHIBA_MEC_XA3",VR::US,VM::VM2,"",false },
  {0x7079,0x005c,"TOSHIBA_MEC_XA3",VR::US,VM::VM2_2n,"",false },
  {0x7079,0x005d,"TOSHIBA_MEC_XA3",VR::SH,VM::VM1,"",false },
  {0x7079,0x005e,"TOSHIBA_MEC_XA3",VR::US,VM::VM2,"",false },
  {0x7079,0x005f,"TOSHIBA_MEC_XA3",VR::SS,VM::VM4_4n,"",false },
  {0x7079,0x0060,"TOSHIBA_MEC_XA3",VR::US,VM::VM1_n,"",false },
  {0x7079,0x0061,"TOSHIBA_MEC_XA3",VR::SS,VM::VM30_30n,"",false },
  {0x7079,0x0062,"TOSHIBA_MEC_XA3",VR::US,VM::VM4,"",false },
  {0x7079,0x0063,"TOSHIBA_MEC_XA3",VR::US,VM::VM1_n,"",false },
  {0x7079,0x0064,"TOSHIBA_MEC_XA3",VR::US,VM::VM1_n,"",false },
  {0x7079,0x0065,"TOSHIBA_MEC_XA3",VR::US,VM::VM1_n,"",false },
  {0x7079,0x0066,"TOSHIBA_MEC_XA3",VR::US,VM::VM1_n,"",false },
  {0x7079,0x0067,"TOSHIBA_MEC_XA3",VR::US,VM::VM1_n,"",false },
  {0x7079,0x0068,"TOSHIBA_MEC_XA3",VR::US,VM::VM1_n,"",false },
  {0x7079,0x0069,"TOSHIBA_MEC_XA3",VR::US,VM::VM1_n,"",false },
  {0x7079,0x006a,"TOSHIBA_MEC_XA3",VR::US,VM::VM1_n,"",false },
  {0x7079,0x006b,"TOSHIBA_MEC_XA3",VR::US,VM::VM1_n,"",false },
  {0x7079,0x006c,"TOSHIBA_MEC_XA3",VR::US,VM::VM1,"",false },
  {0x7079,0x006d,"TOSHIBA_MEC_XA3",VR::DS,VM::VM1,"",false },
  {0x7079,0x006e,"TOSHIBA_MEC_XA3",VR::LO,VM::VM1,"",false },
  {0x7079,0x006f,"TOSHIBA_MEC_XA3",VR::DS,VM::VM1,"",false },
  {0x7079,0x0070,"TOSHIBA_MEC_XA3",VR::DS,VM::VM1,"",false },
  {0x7079,0x0071,"TOSHIBA_MEC_XA3",VR::DS,VM::VM1_n,"",false },
  {0x7079,0x0072,"TOSHIBA_MEC_XA3",VR::DS,VM::VM1_n,"",false },
  {0x7079,0x0073,"TOSHIBA_MEC_XA3",VR::SL,VM::VM4,"",false },
  {0x7079,0x0074,"TOSHIBA_MEC_XA3",VR::SL,VM::VM28,"",false },
  {0x7079,0x0075,"TOSHIBA_MEC_XA3",VR::SH,VM::VM1,"",false },
  {0x7079,0x0076,"TOSHIBA_MEC_XA3",VR::US,VM::VM1,"",false },
  {0x7079,0x0077,"TOSHIBA_MEC_XA3",VR::SL,VM::VM1,"",false },
  {0x7079,0x0078,"TOSHIBA_MEC_XA3",VR::US,VM::VM1,"",false },
  {0x7079,0x0079,"TOSHIBA_MEC_XA3",VR::US,VM::VM1,"",false },
  {0x7079,0x007b,"TOSHIBA_MEC_XA3",VR::SH,VM::VM1,"?",false },
  {0x7079,0x0080,"TOSHIBA_MEC_XA3",VR::LO,VM::VM1,"",false },
  {0x7079,0x0081,"TOSHIBA_MEC_XA3",VR::SS,VM::VM1_n,"",false },
  {0x7079,0x0082,"TOSHIBA_MEC_XA3",VR::SS,VM::VM1_n,"",false },
  {0x7015,0x0010,"TOSHIBA_SR",VR::OB,VM::VM1,"?",false },
  {0x7015,0x0060,"TOSHIBA_SR",VR::OB,VM::VM1,"?",false },
  {0x0011,0x0001,"V1",VR::OB,VM::VM1,"User Data",false },
  {0x0011,0x0002,"V1",VR::DS,VM::VM1,"Normalization Coefficient",false },
  {0x0011,0x0003,"V1",VR::DS,VM::VM1_n,"Receiving Gain",false },
  {0x0011,0x0004,"V1",VR::DS,VM::VM1,"Mean Image Noise",false },
  {0x0055,0x0010,"VEPRO VIM 5.0 DATA",VR::OB,VM::VM1,"VIFHEADER2",false },
  {0x0055,0x0020,"VEPRO VIM 5.0 DATA",VR::OB,VM::VM1,"VIMDATA2",false },
  {0x0055,0x0030,"VEPRO VIM 5.0 DATA",VR::OB,VM::VM1,"ICONDATA2",false },
  {0x0055,0x0051,"VEPRO VIM 5.0 DATA",VR::UI,VM::VM1,"VEPRO UID ??",false },
  {0x0055,0x0065,"VEPRO VIM 5.0 DATA",VR::OB,VM::VM1,"VEPRO HASHNUMBER ??",false },
  {0x0057,0x0010,"VEPRO BROKER 1.0",VR::SQ,VM::VM1,"VEPRO Data replace sequence",false },
  {0x0057,0x0020,"VEPRO BROKER 1.0 DATA REPLACE",VR::SQ,VM::VM1,"Original data sequence",false },
  {0x0057,0x0030,"VEPRO BROKER 1.0 DATA REPLACE",VR::SQ,VM::VM1,"Replaced data sequence",false },
  {0x0057,0x0040,"VEPRO BROKER 1.0 DATA REPLACE",VR::DA,VM::VM1,"Date of data replacement",false },
  {0x0057,0x0041,"VEPRO BROKER 1.0 DATA REPLACE",VR::TM,VM::VM1,"Time of data replacement",false },
  {0x0057,0x0042,"VEPRO BROKER 1.0 DATA REPLACE",VR::LO,VM::VM1,"Dicom receive node",false },
  {0x0057,0x0043,"VEPRO BROKER 1.0 DATA REPLACE",VR::LO,VM::VM1,"Application name",false },
  {0x0057,0x0044,"VEPRO BROKER 1.0 DATA REPLACE",VR::LO,VM::VM1,"Computer name",false },
  {0x0059,0x0040,"VEPRO DICOM RECEIVE DATA 1.0",VR::DA,VM::VM1,"Receive date",false },
  {0x0059,0x0041,"VEPRO DICOM RECEIVE DATA 1.0",VR::TM,VM::VM1,"Receive time",false },
  {0x0059,0x0042,"VEPRO DICOM RECEIVE DATA 1.0",VR::ST,VM::VM1,"Receive node",false },
  {0x0059,0x0043,"VEPRO DICOM RECEIVE DATA 1.0",VR::ST,VM::VM1,"Receive application",false },
  {0x0059,0x0050,"VEPRO DICOM RECEIVE DATA 1.0",VR::ST,VM::VM1,"Receive local computer",false },
  {0x0059,0x0051,"VEPRO DICOM RECEIVE DATA 1.0",VR::ST,VM::VM1,"Receive local ae title",false },
  {0x0059,0x0060,"VEPRO DICOM RECEIVE DATA 1.0",VR::ST,VM::VM1,"Receive remote computer",false },
  {0x0059,0x0061,"VEPRO DICOM RECEIVE DATA 1.0",VR::ST,VM::VM1,"Receive remote ae title",false },
  {0x0059,0x0070,"VEPRO DICOM RECEIVE DATA 1.0",VR::UI,VM::VM1,"Receive original transfer syntax",false },
  {0x0059,0x0010,"VEPRO DICOM TRANSFER 1.0",VR::SQ,VM::VM1,"Dicom Transfer Info",false },
  {0x0055,0x0020,"VEPRO VIF 3.0 DATA",VR::OB,VM::VM1,"VIF DATA",false },
  {0x0055,0x0030,"VEPRO VIF 3.0 DATA",VR::OB,VM::VM1,"Icon data",false },
  {0x0055,0x0065,"VEPRO VIF 3.0 DATA",VR::OB,VM::VM1,"Image hash value",false },
  {0x3253,0x0000,"Varian Medical Systems VISION 3253",VR::OB,VM::VM1,"Extended Interface Data",false },
  {0x3253,0x0001,"Varian Medical Systems VISION 3253",VR::IS,VM::VM1,"Extended Interface Length",false },
  {0x3253,0x0002,"Varian Medical Systems VISION 3253",VR::LO,VM::VM1,"Extended Interface Format",false },
  {0x3273,0x0000,"Varian Medical Systems VISION 3273",VR::DS,VM::VM3,"RT Image Isocenter Position",false },
  {0x3273,0x0001,"Varian Medical Systems VISION 3273",VR::CS,VM::VM1,"RT Image Patient Position",false },
  {0x0065,0x0093,"Viewing Protocol",VR::CS,VM::VM1,"?",false },
  {0x5653,0x0010,"Vital Images SW 3.4",VR::OB,VM::VM1,"Saved Workflow",false },
  {0x5653,0x0011,"Vital Images SW 3.4",VR::LO,VM::VM1,"Saved Workflow File Name",false },
  {0x5653,0x0012,"Vital Images SW 3.4",VR::OB,VM::VM1,"Saved Workflow File Data",false },
  {0x5653,0x0013,"Vital Images SW 3.4",VR::SL,VM::VM1,"Saved Workflow File Length",false },
  {0x5653,0x0014,"Vital Images SW 3.4",VR::SQ,VM::VM1,"Saved Workflow File Sequence",false },
  {0x5653,0x0015,"Vital Images SW 3.4",VR::SQ,VM::VM1,"Image Sequence",false },
  {0x5653,0x0016,"Vital Images SW 3.4",VR::SL,VM::VM1,"Volume Interpolated Slices",false },
  {0x5653,0x0017,"Vital Images SW 3.4",VR::UI,VM::VM1,"Volume SOP Instance UID",false },
  {0x5653,0x0018,"Vital Images SW 3.4",VR::SH,VM::VM1,"Saved Workflow Mark",false },
  {0x5653,0x0019,"Vital Images SW 3.4",VR::UI,VM::VM1,"?",false },
  {0x5653,0x0022,"Vital Images SW 3.4",VR::UI,VM::VM1,"?",false },
  {0x5653,0x0023,"Vital Images SW 3.4",VR::LO,VM::VM1,"?",false },
  {0x1135,0x0000,"Voxar 2.16.124.113543.6003.1999.12.20.12.5.0",VR::UI,VM::VM1,"Private Voxar Capture UID",false },
  {0x1135,0x0001,"Voxar 2.16.124.113543.6003.1999.12.20.12.5.0",VR::OB,VM::VM1,"Private Voxar Saved State",false },
  {0x1135,0x0002,"Voxar 2.16.124.113543.6003.1999.12.20.12.5.0",VR::SQ,VM::VM1,"Private Voxar Icon Image Sequence",false },
  {0x1135,0x0003,"Voxar 2.16.124.113543.6003.1999.12.20.12.5.0",VR::ST,VM::VM1,"Private Voxar Top Right Display Demographics",false },
  {0x1135,0x0004,"Voxar 2.16.124.113543.6003.1999.12.20.12.5.0",VR::ST,VM::VM1,"Private Voxar Bottom Left Display Demographics",false },
  {0x1135,0x0005,"Voxar 2.16.124.113543.6003.1999.12.20.12.5.0",VR::IS,VM::VM1,"Private Voxar Span State Schema",false },
  {0x1135,0x0006,"Voxar 2.16.124.113543.6003.1999.12.20.12.5.0",VR::UL,VM::VM1,"Private Voxar Live Image Restored Size",false },
  {0x1135,0x0007,"Voxar 2.16.124.113543.6003.1999.12.20.12.5.0",VR::UI,VM::VM1,"Private Voxar CaptureType",false },
  {0x1135,0x0008,"Voxar 2.16.124.113543.6003.1999.12.20.12.5.0",VR::SQ,VM::VM1,"Private Voxar Capture Source Series UID Sequence",false },
  {0x1135,0x0009,"Voxar 2.16.124.113543.6003.1999.12.20.12.5.0",VR::UI,VM::VM1,"Private Voxar Capture Source Series UID",false },
  {0x1135,0x0010,"Voxar 2.16.124.113543.6003.1999.12.20.12.5.0",VR::IS,VM::VM1,"Private Voxar Background Gray For Capture",false },
  {0x1135,0x0011,"Voxar 2.16.124.113543.6003.1999.12.20.12.5.0",VR::IS,VM::VM1,"Private Voxar Calcium Scoring Report Schema",false },
  {0x1135,0x0012,"Voxar 2.16.124.113543.6003.1999.12.20.12.5.0",VR::IS,VM::VM1,"Private Voxar Source Series Number",false },
  {0x1135,0x0013,"Voxar 2.16.124.113543.6003.1999.12.20.12.5.0",VR::LO,VM::VM1,"Private Voxar Bottom Left Display Demographics",false },
  {0x0009,0x0001,"WG12 Supplement 43",VR::SQ,VM::VM1,"Event Timer Sequence",false },
  {0x0009,0x0002,"WG12 Supplement 43",VR::FD,VM::VM1,"Event Time Interval",false },
  {0x0009,0x0003,"WG12 Supplement 43",VR::SQ,VM::VM1,"Event Code Sequence",false },
  {0x0019,0x0001,"WG12 Supplement 43",VR::FD,VM::VM1,"Focus Depth(s)",false },
  {0x0019,0x0003,"WG12 Supplement 43",VR::SQ,VM::VM1,"Excluded Intervals Sequence",false },
  {0x0019,0x0004,"WG12 Supplement 43",VR::DT,VM::VM1,"Exclusion Start Datetime",false },
  {0x0019,0x0005,"WG12 Supplement 43",VR::FD,VM::VM1,"Exclusion Duration",false },
  {0x0019,0x0006,"WG12 Supplement 43",VR::SQ,VM::VM1,"US Image Description Sequence",false },
  {0x0019,0x0007,"WG12 Supplement 43",VR::SQ,VM::VM1,"Image Data Type Sequence",false },
  {0x0019,0x0008,"WG12 Supplement 43",VR::CS,VM::VM1,"Data Type",false },
  {0x0019,0x0009,"WG12 Supplement 43",VR::SQ,VM::VM1,"Transducer Scan Geometry Code Sequence",false },
  {0x0019,0x000b,"WG12 Supplement 43",VR::CS,VM::VM1,"Aliased Data Type",false },
  {0x0019,0x000c,"WG12 Supplement 43",VR::CS,VM::VM1,"Position Measuring Device Used",false },
  {0x0019,0x000d,"WG12 Supplement 43",VR::SQ,VM::VM1,"Transducer Scanning Configuration Code Sequence",false },
  {0x0019,0x000e,"WG12 Supplement 43",VR::SQ,VM::VM1,"Transducer Beam Steering Code Sequence",false },
  {0x0019,0x000f,"WG12 Supplement 43",VR::SQ,VM::VM1,"Transducer Access Code Sequence",false },
  {0x0021,0x0001,"WG12 Supplement 43",VR::FD,VM::VM1,"Image Position (Volume)",false },
  {0x0021,0x0002,"WG12 Supplement 43",VR::FD,VM::VM1,"Image Orientation (Volume)",false },
  {0x0021,0x0007,"WG12 Supplement 43",VR::CS,VM::VM1,"Ultrasound Acquisition Geometry",false },
  {0x0021,0x0008,"WG12 Supplement 43",VR::FD,VM::VM1,"Apex Position",false },
  {0x0021,0x0009,"WG12 Supplement 43",VR::FD,VM::VM1,"Volume to Transducer Mapping Matrix",false },
  {0x0021,0x000a,"WG12 Supplement 43",VR::FD,VM::VM1,"Volume to Table Mapping Matrix",false },
  {0x0021,0x000c,"WG12 Supplement 43",VR::CS,VM::VM1,"Patient Frame of Reference Source",false },
  {0x0021,0x000d,"WG12 Supplement 43",VR::FD,VM::VM1,"Temporal Position Time Offset",false },
  {0x0021,0x000e,"WG12 Supplement 43",VR::SQ,VM::VM1,"Plane Position (Volume) Sequence",false },
  {0x0021,0x000f,"WG12 Supplement 43",VR::SQ,VM::VM1,"Plane Orientation (Volume) Sequence",false },
  {0x0021,0x0010,"WG12 Supplement 43",VR::SQ,VM::VM1,"Temporal Position Sequence",false },
  {0x0021,0x0011,"WG12 Supplement 43",VR::CS,VM::VM1,"Dimension Organization Type",false },
  {0x0029,0x0001,"WG12 Supplement 43",VR::SQ,VM::VM1,"Data Frame Assignment Sequence",false },
  {0x0029,0x0002,"WG12 Supplement 43",VR::CS,VM::VM1,"Data Path Assignment",false },
  {0x0029,0x0003,"WG12 Supplement 43",VR::US,VM::VM1,"Bits Mapped to Color Lookup Table",false },
  {0x0029,0x0004,"WG12 Supplement 43",VR::SQ,VM::VM1,"Opacity 1 LUT Sequence",false },
  {0x0029,0x0005,"WG12 Supplement 43",VR::CS,VM::VM1,"Opacity 1 LUT Transfer Function",false },
  {0x0029,0x0006,"WG12 Supplement 43",VR::FD,VM::VM1,"Opacity Constant",false },
  {0x0029,0x0007,"WG12 Supplement 43",VR::US,VM::VM1,"Opacity Lookup Table Descriptor",false },
  {0x0029,0x0008,"WG12 Supplement 43",VR::OW,VM::VM1,"Opacity Lookup Table Data",false },
  {0x0029,0x000b,"WG12 Supplement 43",VR::SQ,VM::VM1,"Enhanced Palette Color Lookup Table Sequence",false },
  {0x0029,0x000c,"WG12 Supplement 43",VR::SQ,VM::VM1,"Opacity 2 LUT Sequence",false },
  {0x0029,0x000d,"WG12 Supplement 43",VR::CS,VM::VM1,"Opacity 2 LUT Transfer Function",false },
  {0x0029,0x000e,"WG12 Supplement 43",VR::CS,VM::VM1,"Data Path ID",false },
  {0x0029,0x000f,"WG12 Supplement 43",VR::CS,VM::VM1,"RGB LUT Transfer Function",false },
  {0x0029,0x0010,"WG12 Supplement 43",VR::CS,VM::VM1,"Alpha LUT Transfer Function",false },
  {0x0041,0x0001,"WG12 Supplement 43",VR::CS,VM::VM1,"Performed Protocol Type",false },
  {0x3119,0x0010,"http://www.gemedicalsystems.com/it_solutions/bamwallthickness/1.0",VR::DT,VM::VM1,"BAM WallThickness Session Date/Time",false },
  {0x3119,0x0020,"http://www.gemedicalsystems.com/it_solutions/bamwallthickness/1.0",VR::PN,VM::VM1,"BAM WallThickness Session Creator",false },
  {0x3119,0x0030,"http://www.gemedicalsystems.com/it_solutions/bamwallthickness/1.0",VR::CS,VM::VM1,"BAM WallThickness Session Completion Flag",false },
  {0x3119,0x0040,"http://www.gemedicalsystems.com/it_solutions/bamwallthickness/1.0",VR::SQ,VM::VM1,"BAM WallThickness File Sequence",false },
  {0x3119,0x0050,"http://www.gemedicalsystems.com/it_solutions/bamwallthickness/1.0",VR::ST,VM::VM1,"BAM WallThickness File Name",false },
  {0x3119,0x0060,"http://www.gemedicalsystems.com/it_solutions/bamwallthickness/1.0",VR::OB,VM::VM1,"BAM WallThickness File Content",false },
  {0x3117,0x0010,"http://www.gemedicalsystems.com/it_solutions/orthoview/2.1",VR::DT,VM::VM1,"OrthoView Session Date/Time",false },
  {0x3117,0x0020,"http://www.gemedicalsystems.com/it_solutions/orthoview/2.1",VR::PN,VM::VM1,"OrthoView Session Creator",false },
  {0x3117,0x0030,"http://www.gemedicalsystems.com/it_solutions/orthoview/2.1",VR::CS,VM::VM1,"OrthoView Session Completion Flag",false },
  {0x3117,0x0040,"http://www.gemedicalsystems.com/it_solutions/orthoview/2.1",VR::SQ,VM::VM1,"OrthoView File Sequence",false },
  {0x3117,0x0050,"http://www.gemedicalsystems.com/it_solutions/orthoview/2.1",VR::ST,VM::VM1,"OrthoView File Name",false },
  {0x3117,0x0060,"http://www.gemedicalsystems.com/it_solutions/orthoview/2.1",VR::OB,VM::VM1,"OrthoView File Content",false },
  {0x3115,0x0001,"http://www.gemedicalsystems.com/it_solutions/rad_pacs/",VR::UT,VM::VM1,"Reference to pacs study",false },
  {0x3115,0x0002,"http://www.gemedicalsystems.com/it_solutions/rad_pacs/",VR::UT,VM::VM1,"Reference to pacs image",false },
  {0x3115,0x0003,"http://www.gemedicalsystems.com/it_solutions/rad_pacs/",VR::CS,VM::VM1,"Pacs examnotes flag",false },
  {0x0021,0x00ae,"syngoDynamics",VR::OB,VM::VM1,"?",false },
  //{0x0029,0x0000,"1.2.840.113663.1",VR::UL,VM::VM1,"Group Length 0029",true },
  //{0x0019,0x0000,"1.2.840.113681",VR::UL,VM::VM1,"Group Length 0019",true },
  //{0x0087,0x0000,"1.2.840.113708.794.1.1.2.0",VR::UL,VM::VM1,"Group Length 0087",true },
  //{0x3711,0x0000,"A.L.I. Technologies, Inc.",VR::UL,VM::VM1,"Group Length 3711",true },
  //{0x0009,0x0000,"ACUSON",VR::UL,VM::VM1,"Group Length 0009",true },
  //{0x7fdf,0x0000,"ACUSON: 1.2.840.11386.1.0",VR::UL,VM::VM1,"Group Length 7fdf",true },
  //{0x7043,0x0000,"ADAC_IMG",VR::UL,VM::VM1,"Group Length 7043",true },
  //{0x0003,0x0000,"AEGIS_DICOM_2.00",VR::UL,VM::VM1,"Group Length 0003",true },
  //{0x0005,0x0000,"AEGIS_DICOM_2.00",VR::UL,VM::VM1,"Group Length 0005",true },
  //{0x1369,0x0000,"AEGIS_DICOM_2.00",VR::UL,VM::VM1,"Group Length 1369",true },
  //{0x0031,0x0000,"AGFA PACS Archive Mirroring 1.0",VR::UL,VM::VM1,"Group Length 0031",true },
  //{0x0071,0x0000,"AGFA-AG_HPState",VR::UL,VM::VM1,"Group Length 0071",true },
  //{0x3101,0x0000,"AMI Annotations_01",VR::UL,VM::VM1,"Group Length 3101",true },
  //{0x3107,0x0000,"AMI ImageTransform_01",VR::UL,VM::VM1,"Group Length 3107",true },
  //{0x3105,0x0000,"AMI Sequence AnnotElements_01",VR::UL,VM::VM1,"Group Length 3105",true },
  //{0x3103,0x0000,"AMI Sequence Annotations_01",VR::UL,VM::VM1,"Group Length 3103",true },
  //{0x0023,0x0000,"AMICAS0",VR::UL,VM::VM1,"Group Length 0023",true },
  //{0x0027,0x0000,"APEX_PRIVATE",VR::UL,VM::VM1,"Group Length 0027",true },
  //{0x0011,0x0000,"Acuson X500",VR::UL,VM::VM1,"Group Length 0011",true },
  //{0x0013,0x0000,"Acuson X500",VR::UL,VM::VM1,"Group Length 0013",true },
  //{0x0015,0x0000,"Acuson X500",VR::UL,VM::VM1,"Group Length 0015",true },
  //{0x0017,0x0000,"Acuson X500",VR::UL,VM::VM1,"Group Length 0017",true },
  //{0x3113,0x0000,"Applicare/RadStore/Version 1.0",VR::UL,VM::VM1,"Group Length 3113",true },
  //{0x3109,0x0000,"Applicare/RadWorks/Version 5.0",VR::UL,VM::VM1,"Group Length 3109",true },
  //{0x4103,0x0000,"Applicare/RadWorks/Version 6.0",VR::UL,VM::VM1,"Group Length 4103",true },
  //{0x4105,0x0000,"Applicare/RadWorks/Version 6.0",VR::UL,VM::VM1,"Group Length 4105",true },
  //{0x4107,0x0000,"Applicare/RadWorks/Version 6.0",VR::UL,VM::VM1,"Group Length 4107",true },
  //{0x0021,0x0000,"BRIT Systems, Inc.",VR::UL,VM::VM1,"Group Length 0021",true },
  //{0x0063,0x0000,"BioPri3D",VR::UL,VM::VM1,"Group Length 0063",true },
  //{0x3411,0x0000,"BrainLAB_BeamProfile",VR::UL,VM::VM1,"Group Length 3411",true },
  //{0x3273,0x0000,"BrainLAB_PatientSetup",VR::UL,VM::VM1,"Group Length 3273",true },
  //{0x3033,0x0000,"CEMAX-ICON",VR::UL,VM::VM1,"Group Length 3033",true },
  //{0x0089,0x0000,"DIDI TO PCR 1.1",VR::UL,VM::VM1,"Group Length 0089",true },
  //{0x4453,0x0000,"DR Systems, Inc.",VR::UL,VM::VM1,"Group Length 4453",true },
  //{0x7fa1,0x0000,"ELGEMS_XPERT",VR::UL,VM::VM1,"Group Length 7fa1",true },
  //{0x00e1,0x0000,"ELSCINT1",VR::UL,VM::VM1,"Group Length 00e1",true },
  //{0x01e1,0x0000,"ELSCINT1",VR::UL,VM::VM1,"Group Length 01e1",true },
  //{0x01f1,0x0000,"ELSCINT1",VR::UL,VM::VM1,"Group Length 01f1",true },
  //{0x01f3,0x0000,"ELSCINT1",VR::UL,VM::VM1,"Group Length 01f3",true },
  //{0x01f7,0x0000,"ELSCINT1",VR::UL,VM::VM1,"Group Length 01f7",true },
  //{0x0601,0x0000,"ELSCINT1",VR::UL,VM::VM1,"Group Length 0601",true },
  //{0x07a1,0x0000,"ELSCINT1",VR::UL,VM::VM1,"Group Length 07a1",true },
  //{0x07a3,0x0000,"ELSCINT1",VR::UL,VM::VM1,"Group Length 07a3",true },
  //{0x5001,0x0000,"ELSCINT1",VR::UL,VM::VM1,"Group Length 5001",true },
  //{0x6001,0x0000,"ELSCINT1",VR::UL,VM::VM1,"Group Length 6001",true },
  //{0x200d,0x0000,"Eclipse 60",VR::UL,VM::VM1,"Group Length 200d",true },
  //{0x7777,0x0000,"Eclipse 60",VR::UL,VM::VM1,"Group Length 7777",true },
  //{0x0025,0x0000,"FDMS 1.0",VR::UL,VM::VM1,"Group Length 0025",true },
  //{0x0032,0x0000,"FDMS 1.0",VR::UL,VM::VM1,"Group Length 0032",true },
  //{0x2011,0x0000,"FDMS 1.0",VR::UL,VM::VM1,"Group Length 2011",true },
  //{0x50f1,0x0000,"FDMS 1.0",VR::UL,VM::VM1,"Group Length 50f1",true },
  //{0x0045,0x0000,"GE LUT Asymmetry Parameter",VR::UL,VM::VM1,"Group Length 0045",true },
  //{0x004b,0x0000,"GEIIS",VR::UL,VM::VM1,"Group Length 004b",true },
  //{0x0905,0x0000,"GEIIS",VR::UL,VM::VM1,"Group Length 0905",true },
  //{0x0907,0x0000,"GEIIS",VR::UL,VM::VM1,"Group Length 0907",true },
  //{0x7fd1,0x0000,"GEIIS",VR::UL,VM::VM1,"Group Length 7fd1",true },
  //{0x0903,0x0000,"GEIIS PACS",VR::UL,VM::VM1,"Group Length 0903",true },
  //{0x0047,0x0000,"GEMS_3DSTATE_001",VR::UL,VM::VM1,"Group Length 0047",true },
  //{0x0057,0x0000,"GEMS_ADWSoft_3D2",VR::UL,VM::VM1,"Group Length 0057",true },
  //{0x0039,0x0000,"GEMS_ADWSoft_DPO",VR::UL,VM::VM1,"Group Length 0039",true },
  //{0x0033,0x0000,"GEMS_CTHD_01",VR::UL,VM::VM1,"Group Length 0033",true },
  //{0x0049,0x0000,"GEMS_CT_CARDIAC_001",VR::UL,VM::VM1,"Group Length 0049",true },
  //{0x0051,0x0000,"GEMS_CT_VES_01",VR::UL,VM::VM1,"Group Length 0051",true },
  //{0x0037,0x0000,"GEMS_DRS_1",VR::UL,VM::VM1,"Group Length 0037",true },
  //{0x0035,0x0000,"GEMS_GENIE_1",VR::UL,VM::VM1,"Group Length 0035",true },
  //{0x0055,0x0000,"GEMS_GENIE_1",VR::UL,VM::VM1,"Group Length 0055",true },
  //{0x0043,0x0000,"GEMS_IMPS_01",VR::UL,VM::VM1,"Group Length 0043",true },
  //{0x7001,0x0000,"GEMS_MR_RAW_01",VR::UL,VM::VM1,"Group Length 7001",true },
  //{0x2001,0x0000,"GEMS_PARM_01",VR::UL,VM::VM1,"Group Length 2001",true },
  //{0x5003,0x0000,"GEMS_PETD_01",VR::UL,VM::VM1,"Group Length 5003",true },
  //{0x5005,0x0000,"GEMS_PETD_01",VR::UL,VM::VM1,"Group Length 5005",true },
  //{0x0249,0x0000,"GEMS_RTEN_01",VR::UL,VM::VM1,"Group Length 0249",true },
  //{0x6003,0x0000,"GEMS_Ultrasound_ImageGroup_001",VR::UL,VM::VM1,"Group Length 6003",true },
  //{0x7f01,0x0000,"HOLOGIC, Inc.",VR::UL,VM::VM1,"Group Length 7f01",true },
  //{0x1269,0x0000,"IMS s.r.l. Biopsy Private Code",VR::UL,VM::VM1,"Group Length 1269",true },
  //{0x1271,0x0000,"IMS s.r.l. Mammography Private Code",VR::UL,VM::VM1,"Group Length 1271",true },
  //{0x0041,0x0000,"INTEGRIS 1.0",VR::UL,VM::VM1,"Group Length 0041",true },
  //{0x3f01,0x0000,"INTELERAD MEDICAL SYSTEMS",VR::UL,VM::VM1,"Group Length 3f01",true },
  //{0x8003,0x0000,"Image (ID, Version, Size, Dump, GUID)",VR::UL,VM::VM1,"Group Length 8003",true },
  //{0x7fe1,0x0000,"KRETZ_US",VR::UL,VM::VM1,"Group Length 7fe1",true },
  //{0x3005,0x0000,"MDS NORDION OTP ANATOMY MODELLING",VR::UL,VM::VM1,"Group Length 3005",true },
  //{0x5473,0x0000,"MedIns HP Extensions",VR::UL,VM::VM1,"Group Length 5473",true },
  //{0x1455,0x0000,"Mortara_Inc",VR::UL,VM::VM1,"Group Length 1455",true },
  //{0x8101,0x0000,"ObjectModel (ID, Version, Place, PlaceDescription)",VR::UL,VM::VM1,"Group Length 8101",true },
  //{0x7051,0x0000,"PHILIPS NM -Private",VR::UL,VM::VM1,"Group Length 7051",true },
  //{0x5503,0x0000,"PHILIPS UNDOCUMENTED",VR::UL,VM::VM1,"Group Length 5503",true },
  //{0x2005,0x0000,"PHILIPS UNKNOWN",VR::UL,VM::VM1,"Group Length 2005",true },
  //{0x7015,0x0000,"PMTF INFORMATION DATA",VR::UL,VM::VM1,"Group Length 7015",true },
  //{0x2007,0x0000,"Philips EV Imaging DD 021",VR::UL,VM::VM1,"Group Length 2007",true },
  //{0x4007,0x0000,"Philips Imaging DD 065",VR::UL,VM::VM1,"Group Length 4007",true },
  //{0x4001,0x0000,"Philips Imaging DD 067",VR::UL,VM::VM1,"Group Length 4001",true },
  //{0x0511,0x0000,"Philips PET Private Group",VR::UL,VM::VM1,"Group Length 0511",true },
  //{0x7053,0x0000,"Philips PET Private Group",VR::UL,VM::VM1,"Group Length 7053",true },
  //{0x200b,0x0000,"Philips RAD Imaging DD 001",VR::UL,VM::VM1,"Group Length 200b",true },
  //{0x2003,0x0000,"Philips X-ray Imaging DD 001",VR::UL,VM::VM1,"Group Length 2003",true },
  //{0x7101,0x0000,"Picker MR Private Group",VR::UL,VM::VM1,"Group Length 7101",true },
  //{0x6021,0x0000,"SIEMENS CM VA0  CMS",VR::UL,VM::VM1,"Group Length 6021",true },
  //{0x6023,0x0000,"SIEMENS CM VA0  OST",VR::UL,VM::VM1,"Group Length 6023",true },
  //{0x0193,0x0000,"SIEMENS ISI",VR::UL,VM::VM1,"Group Length 0193",true },
  //{0x0307,0x0000,"SIEMENS ISI",VR::UL,VM::VM1,"Group Length 0307",true },
  //{0x0309,0x0000,"SIEMENS ISI",VR::UL,VM::VM1,"Group Length 0309",true },
  //{0x4009,0x0000,"SIEMENS ISI",VR::UL,VM::VM1,"Group Length 4009",true },
  //{0x001f,0x0000,"SIEMENS MED",VR::UL,VM::VM1,"Group Length 001f",true },
  //{0x0053,0x0000,"SIEMENS MED",VR::UL,VM::VM1,"Group Length 0053",true },
  //{0x7003,0x0000,"SIEMENS MED",VR::UL,VM::VM1,"Group Length 7003",true },
  //{0x7005,0x0000,"SIEMENS MED",VR::UL,VM::VM1,"Group Length 7005",true },
  //{0x7fe3,0x0000,"SIEMENS MED",VR::UL,VM::VM1,"Group Length 7fe3",true },
  //{0x0091,0x0000,"SIENET",VR::UL,VM::VM1,"Group Length 0091",true },
  //{0x0095,0x0000,"SIENET",VR::UL,VM::VM1,"Group Length 0095",true },
  //{0x0099,0x0000,"SIENET",VR::UL,VM::VM1,"Group Length 0099",true },
  //{0x0073,0x0000,"STENTOR",VR::UL,VM::VM1,"Group Length 0073",true },
  //{0xf001,0x0000,"Sound Technologies",VR::UL,VM::VM1,"Group Length f001",true },
  //{0x7ff1,0x0000,"TOSHIBA_MEC_1.0",VR::UL,VM::VM1,"Group Length 7ff1",true },
  //{0x700d,0x0000,"TOSHIBA_MEC_MR3",VR::UL,VM::VM1,"Group Length 700d",true },
  //{0x7019,0x0000,"TOSHIBA_MEC_OT3",VR::UL,VM::VM1,"Group Length 7019",true },
  //{0x7079,0x0000,"TOSHIBA_MEC_XA3",VR::UL,VM::VM1,"Group Length 7079",true },
  //{0x3253,0x0000,"Varian Medical Systems VISION 3253",VR::UL,VM::VM1,"Group Length 3253",true },
  //{0x0065,0x0000,"Viewing Protocol",VR::UL,VM::VM1,"Group Length 0065",true },
  //{0x1135,0x0000,"Voxar 2.16.124.113543.6003.1999.12.20.12.5.0",VR::UL,VM::VM1,"Group Length 1135",true },
  //{0x3115,0x0000,"http://www.gemedicalsystems.com/it_solutions/rad_pacs/",VR::UL,VM::VM1,"Group Length 3115",true },

  {0xffff,0xffff,"GDCM Private Sentinel",VR::INVALID,VM::VM0,"",true }, // Dummy invalid element
  {0xffff,0xffff,"",VR::INVALID,VM::VM0,nullptr,true } // Guard (will NOT be added)
};
} //end anonymous namespace
namespace gdcm {

/*
void Dict::LoadDefault()
{
   unsigned int i = 0;
   DICT_ENTRY n = DICOMV3DataDict[i];
   while( n.name != 0 )
   {
   if( n.group % 2 == 0 )
     {
     gdcm_assert( n.owner == 0 );
     Tag t(n.group, n.element);
     DictEntry e( n.name, n.vr, n.vm, n.ret );
     AddDictEntry( t, e );
     n = DICOMV3DataDict[++i];
     }
   }
}
*/

void PrivateDict::LoadDefault()
{
   unsigned int i = 0;
   DICT_ENTRY n = DICOMV3DataDict[i];
   while( n.name != nullptr )
   {
//   if( n.group % 2 != 0 )
     {
     gdcm_assert( n.owner != nullptr );
     gdcm_assert( n.name );
     gdcm_assert( n.group % 2 != 0 || n.group == 0xffff );
     gdcm_assert( n.element <= 0xff || n.element == 0xffff );
     PrivateTag t(n.group, n.element,n.owner);
     DictEntry e( n.name, "", n.vr, n.vm, n.ret );
     AddDictEntry( t, e );
     }
   n = DICOMV3DataDict[++i];
   }
}

} // end namespace gdcm
#endif // GDCMPRIVATEDEFAULTDICTS_CXX
